#ifndef RESOLUTIONPERFORMANCEPROCESSOR_H
#define RESOLUTIONPERFORMANCEPROCESSOR_H

#include <marlin/Processor.h>
#include "marlin/Global.h"

#include <lcio.h>
#include <string>
#include "TrackFitterBase.h"
#include <EVENT/TrackerHit.h>

//MarlinTPC
#include "ModuleDistortionCalculator.h"

//GEAR
#include "gear/TPCParameters.h"

//ROOT
#include "TH1.h"
#include "TH2.h"
#include "TNtuple.h"


// In case you want AIDA histograms

namespace marlintpc
{

/**  Resolution calculator
 *
 *   Determines the resolution via the geometric mean method.
 *
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A collection of tracks (lcio::Track). <br>
 *  Remember that LCIO::TRBIT_HITS has to be set, so the hits are stored.
 *
 *  <h4>Output</h4>
 *  \li Histograms with residual distribution including all hits (n)
 *  and leaving out the test hit (n-1)
 *
 *
 * @param InputCollectionName: Name of input collection, Track
 * @param xyResidualRange: Histograms will have range (-_histoRange, _histoRange), default _histoRange = 3mm
 * @param xyResidualBins: Histograms will be binned with _histoBin, default _histoBin = 254
 * @param zResidualRange: Histograms will have range (-_histoRange, _histoRange), default _histoRange = 5mm
 * @param zResidualBins: Histograms will be binned with _histoBin, default _histoBin = 254
 * @param bfieldScaleFactor: scales the input magnetic field (can be used to switch it off), default _bfieldScaleFactor = 1
 * @param meanCalculationMethod: Defines the method of calculating the mean of the histograms. (default _meanMethod = 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails
 * @param ModuleWithBestRow: Module that has the best row, default _bestModule = 0
 * @param RowWithBestPerformance: Row with the best resolution performance, default _bestRow = 0
 * 
 * @author Felix Mueller, DESY
 *
 */

class ResolutionPerformanceProcessor : public marlin::Processor
{
public:

	virtual Processor*  newProcessor() { return new ResolutionPerformanceProcessor; }

	ResolutionPerformanceProcessor();

	virtual void init();

	virtual void processRunHeader(lcio::LCRunHeader* run);

	virtual void processEvent(lcio::LCEvent* evt);

	virtual void check(lcio::LCEvent* evt);

	virtual void end();



protected:
	/* the place for protected and private member data and functions */
	std::string _inputColName; ///< Name of the input collection

private:
	double calcHitPhi(TrackerHit * firstHit, TrackerHit * secondHit);

	double _bfieldScaleFactor; ///< scale factor for magnetic field (default: 1.0)
	int _xyResolutionBins; ///< Number of bins in the xy-residual histogram
	float _xyResolutionRange; ///< Range of the xy-residual histogram
	int _zResolutionBins; ///< Number of bins in the z-residual histogram
	float _zResolutionRange; ///< Range of the z-residual histogram
	int _chargeBins; ///< Number of bins on the charge axis in the residuals vs charge histograms
	int _chargeUpperBound; ///< Upper bound of the charge axis in the residuals vs charge histograms 
	bool _useHitsWithOverflow; ///< switch for skipping of Hits wit overflow pulses

	/**
	 * Histograms containing distances and residuals for each module and row
	 */
	std::map< std::pair<int,int>, TH1*> _zDistanceRow;
	std::map< std::pair<int,int>, TH1*> _xyDistanceRow;
	std::map< std::pair<int,int>, TH1*> _zResidualRow;
	std::map< std::pair<int,int>, TH1*> _xyResidualRow;
	std::map< std::pair<int,int>, TH1*> _phiRow;

	std::map< std::pair<int,int>, TH2*> _zResidualVSChargeRow;

	/**
	 * Histograms containing distances and residuals for all modules and rows
	 */
	TH1 * _zDistanceAll;
	TH1 * _xyDistanceAll;
	TH1 * _zResidualAll;
	TH1 * _xyResidualAll;

	TH2 * _zResidualVSChargeAllRows;
	TH2 * _xyResidualVSChargeAllRows;

	TNtuple * _allRowsTuple;
	TNtuple * _singleRowsTuple;
	TNtuple * _bestRowTuple;

	TNtuple * _zResidualVSChargeAllRowsTuple;
	TNtuple * _zResidualVSChargeSingleRowsTuple;

	TrackFitterBase * _trackFitter;
	int _meanMethodZ; ///< Defines the method of calculating the mean of the z-histograms. (default 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails, 4:half gaussian fit in 3 sigma on rising slope)
	int _meanMethodXY; ///< Defines the method of calculating the mean of the xy-histograms. (default 2:Gaussian fit within range of 3 sigma, 1:RMS90, 3:rms after 5 sigma iterative cut of tails, 4:half gaussian fit in 3 sigma on rising slope)
	int _bestModule; ///< Module with the best row
	int _bestRow; ///< Row with the best resolution

	ModuleDistortionCalculator * _myModuleDistortionCalculator;
	const gear::TPCParameters* _gearParameters;

	float _maxChargeCut;///<optional cut on charge of maximum pulse to determine if hit is evaluated for resolution
	float _minChargeCut;///<optional cut on charge of maximum pulse to determine if hit is evaluated for resolution
	int _minBinsRiseTimeCut;///<optional cut on the number of bins on the rising slope of the maximum pulse to determine if hit is evaluated for resolution
};
} // namespace marlintpc
#endif // RESOLUTIONPERFORMANCEPROCESSOR_H
