#ifndef TIMEPIXTRACKEVENTDISPLAYPROCESSOR_H
#define TIMEPIXTRACKEVENTDISPLAYPROCESSOR_H

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <Exceptions.h>
#include <gear/BField.h>
#include "SimpleHelixTrackModel.h"

// ROOT
#include "TList.h"  // TH2D->GetListOfFunctions()
#include "TCanvas.h"
#include "TF1.h"
#include "TMarker.h"
#include "TGraph.h"
#include "TLegend.h"
#include "TMath.h"
#include "TH2D.h"
#include "TFile.h"

#include <marlin/Processor.h>
#include <lcio.h>
#include <sstream>
#include <string>
#include <vector>
#include <iomanip> // setw()


namespace AIDA{
  class IHistogram1D;
  class IHistogram2D;
}

namespace marlintpc
{
/** TimePixTrackEventDisplayProcessor draws all TrackerHits in an event as black dots (kFullDotMedium).
 *  Also Tracks are draw on top of their corresponding Hits where the tracks and hits have the same color.
 *  Tracks are only drawn in the xy plane.
 *\n
 *\n
 * @author M. Rogowski
 *
 */

class TimePixTrackEventDisplayProcessor : public marlin::Processor
{
 public:

  virtual Processor*  newProcessor() { return new TimePixTrackEventDisplayProcessor; }

  TimePixTrackEventDisplayProcessor();
  ~TimePixTrackEventDisplayProcessor()
  {
  }

  virtual void init();

  virtual void processRunHeader(lcio::LCRunHeader* run);

  virtual void processEvent(lcio::LCEvent* evt);

  virtual void check(lcio::LCEvent* evt);

  virtual void end();

 public:
  /**   Return TGraph with points on the helix projection in xy-plane\n
   *    \param myTrack: the track that provides the parameters for evaluations of the points\n
   *    \n
   *    Return value:\n
   *    \li an instance of TGraph filled with the evaluated points\n
   */
  static TGraph * DrawHelixXY( const Track * const myTrack );


  /** Returns TGraph with two points in the xy-plane.\n
   *  \param myTrack: Pointer to Track
   *  \n
   *  Return value:\n
   *  \li An instance of TGraph containing the evaluated points
   */
  static TGraph * DrawStraightLineXY( const Track * const myTrack );

  /** Returns TGraph with two points in the zx-plane.\n
   *  \param myTrack: Pointer to Track
   *  \n
   *  Return value:\n
   *  \li An instance of TGraph containing the evaluated points
   */
  static TGraph * DrawStraightLineZX( const Track * const myTrack );

  /** Returns TGraph with points on the helix projection in zyx-plane.\n
   *  \param myTrack: Pointer to Track
   *  \n
   *  Return value:\n
   *  \li An instance of TGraph containing the evaluated points
   */
  static TGraph * DrawHelixZX( const Track * const myTrack );

 protected:
  std::string _inputColNameTracks ;   ///<- name of the input collection for tracks
  std::string _TrackerHitColName; ///<- the name of an collection that contains all TrackerHits; used to draw full event
  int _outputIsTransient;       ///<- if not 0 the output collection is set transient (default 0)
  /* the place for protected and private member data and functions */
  TH2D * _FullXYHist;     ///<- vector of 2D histograms, displays all hits in event, XY plane
  TH2D * _FullZXHist;   ///<- vector of 2D histograms, displays all hits from tracks, ZX plane
  TH2D * _FullZYHist;   ///<- vector of 2D histograms, displays all hits from tracks, ZX plane
  float _xMin;  ///<- minimum value of x range
  float _xMax;  ///<- maximum value of x range
  float _yMin;  ///<- minimum value of y range
  float _yMax;  ///<- maximum value of y range
  float _zMin;  ///<- minimum value of z range
  float _zMax;  ///<- maximum value of z range
  int _xPixel;  ///<- number of pixels in x direction
  int _yPixel;  ///<- number of pixels in y direction
  int _MaxNumberOfHistograms;   ///<- maximum number of created histograms
  int _CurNumberOfHistograms;   ///<- current number of created histograms
  TFile * _file; ///<- file pointer to the output root file
  bool _fileOpen; ///<- flag if file could be opened
};
} // namespace marlintpc
#endif // TIMEPIXTRACKEVENTDISPLAYPROCESSOR_H
