#ifndef TRACKHITEFFPROCESSOR_H
#define TRACKHITEFFPROCESSOR_H

// C++
#include <string>
#include <fstream>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "lccd/ConditionsMap.hh"
//GEAR
#include "gear/TPCParameters.h"

//ROOT
#include "TNtuple.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#endif

namespace marlintpc
{

/**
 *  A processor to calculate the hit efficiency based on hits on tracksper row.
 *  Should be run with the AidaProcessor in the chain to save the histogram/graphs.
 *
 *  <h4>Input - Prerequisites</h4>
 *  a collection of Tracks
 *
 *  <h4>Output</h4>
 *  track hit efficiency stored in tuple in aida file
 *
 * @param InputCollectionName Name of the input collection (default: ClupatraTracks)
 *
 * @author Astrid Muennich, DESY
 */

class TrackHitEfficiencyProcessor : public marlin::Processor
{

    public:

        /* This method will be called by the marlin package
         * It returns a processor of the currend type 
         */
        virtual Processor*  newProcessor()
        {
            return new TrackHitEfficiencyProcessor ;
        }

        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        TrackHitEfficiencyProcessor() ;

        /* Called at the beginning of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        virtual void init() ;

        /* Called for every run.
         * in this processor it is not used
         */
        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        virtual void processEvent(lcio::LCEvent * evt) ;

        /* This method is only called if the check flag is set (default) in the main processor
         * (where the input files are given)
         * It should be used to create check plots which are not needed in a complete analysis
         * (where normaly the check flag is deactivated)
         */
        virtual void check(lcio::LCEvent * evt) ;


        /* Called after data processing for clean up.
         * e.g. saving control histogramms, delete created objects
         * (which are not stored in a collection/event), etc.
         */
        virtual void end() ;




    protected:

    std::string _inputColName ;  ///<- the name of the input collection
    std::map< std::pair<int,int>, int> _hitOnRow; ///<- counter for number of hits on each row
    int _ntracks; ///<- number of tracks processed

    private:
	TNtuple* _tuple;
	const gear::TPCParameters* _gearParameters;
} ;

}

#endif



