#ifndef TRIPLETRESOLUTIONPERFORMANCEPROCESSOR_H
#define TRIPLETRESOLUTIONPERFORMANCEPROCESSOR_H

#include <marlin/Processor.h>
#include "marlin/Global.h"

#include <lcio.h>
#include <string>

//MarlinTPC
#include "ModuleDistortionCalculator.h"
//GEAR
#include "gear/TPCParameters.h"

//ROOT
#include "TH1D.h"
#include "TNtuple.h"


namespace marlintpc
{

/**  Resolution calculator
 *
 *   Determines the resolution via the triplet method.
 *
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A collection of tracks (lcio::Track). <br>
 *  Remember that LCIO::TRBIT_HITS has to be set, so the hits are stored.
 *
 *  <h4>Output</h4>
 *  \li Histograms with residual distribution including all hits (n)
 *  and leaving out the test hit (n-1)
 *
 *
 * @param InputCollectionName: Name of input collection, Track
 * @param xyResidualRange: Histograms will have range (-_histoRange, _histoRange), default _histoRange = 3mm
 * @param xyResidualBins: Histograms will be binned with _histoBin, default _histoBin = 254
 * @param zResidualRange: Histograms will have range (-_histoRange, _histoRange), default _histoRange = 5mm
 * @param zResidualBins: Histograms will be binned with _histoBin, default _histoBin = 254
 * @param meanCalculationMethod: Defines the method of calculating the mean of the histograms. (default _meanMethod = 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails
 * @param ModuleWithBestRow: Module that has the best row, default _bestModule = 0
 * @param RowWithBestPerformance: Row with the best resolution performance, default _bestRow = 0
 * 
 * @author Felix Mueller, DESY
 *
 */


class TripletResolutionPerformanceProcessor : public marlin::Processor
{
 public:
  
  virtual Processor*  newProcessor() { return new TripletResolutionPerformanceProcessor; }
  
  TripletResolutionPerformanceProcessor();
  
  virtual void init();
  
  virtual void processRunHeader(lcio::LCRunHeader* run);
  
  virtual void processEvent(lcio::LCEvent* evt);
  
  virtual void check(lcio::LCEvent* evt);
  
  virtual void end();
  

 protected:
 	/* the place for protected and private member data and functions */
 	std::string _inputColName; ///< Name of the input collection

 private:

 	int _xyTripletsBins; ///< Number of bins in the xy-triplets histograms
 	float _xyTripletsRange; ///< Range of the xy-triplets histograms
 	int _zTripletsBins; ///< Number of bins in the z-triplets histograms
 	float _zTripletsRange; ///< Range of the z-triplets histograms

 	/**
 	 * Histograms containing triplets for each module and row
 	 */
 	std::map< std::pair<int,int>, TH1*> _zTripletsRow;
 	std::map< std::pair<int,int>, TH1*> _xyTripletsRow;


 	/**
 	 * Histograms containing triplets for all modules and rows
 	 */
 	TH1 * _zTripletsAll;
 	TH1 * _xyTripletsAll;


 	TNtuple * _allRowsTuple;
 	TNtuple * _singleRowsTuple;
 	TNtuple * _bestRowTuple;


	int _meanMethodZ; ///< Defines the method of calculating the mean of the z-histograms. (default 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails)
	int _meanMethodXY; ///< Defines the method of calculating the mean of the xy-histograms. (default 2:Gaussian fit within range of 3 sigma, 1:RMS90, 3:rms after 5 sigma iterative cut of tails)
 	int _bestModule; ///< Module with the best row
 	int _bestRow; ///< Row with the best resolution

 	const gear::TPCParameters* _gearParameters;

};
} // namespace marlintpc
#endif // TRIPLETRESOLUTIONPERFORMANCEPROCESSOR_H
