#include <vector>
#include <iostream>
#include "TLatex.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TGraphErrors.h"
#include "TMultiGraph.h"
#include "TFile.h"
#include "TNtuple.h"
#include "TStyle.h"
#include "TROOT.h"
#include "TColor.h"

#include "sstream"
#include <cstdarg>

/*How to call this:
.L $MARLINTPC/analysis/rootscripts/ChargeQuality.C++
ChargeQuality("path for output","path to the aida file", number of runs in the following list, list of the runs)
e.g ChargeQuality("../myplots","../AidaFiles/",3,18850,18851,18860)
*/
void ChargeQuality(char * pathToOutput,char * pathToAida, int nRuns, ...){

//loading the style
gROOT->ProcessLine(".L ${MARLINTPC}/analysis/rootscripts/LCTPCStyle.C");
gROOT->ProcessLine("LCTPCStyle()");

//open a root file to store the graphs in
 TFile *outfile=new TFile(Form("%s%s",pathToOutput,"ChargeQuality.root"),"RECREATE");

//initialize the input list
va_list argList;
va_start(argList,nRuns);


int runNumber;

//create colors
 const int BINNING=nRuns;
 Int_t colors[BINNING];

 const Int_t NRGBs = 5;

 Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
 Double_t red[NRGBs]   = { 0.00, 0.00, 0.87, 1.00, 0.51 };
 Double_t green[NRGBs] = { 0.00, 0.81, 1.00, 0.20, 0.00 };
 Double_t blue[NRGBs]  = { 0.51, 1.00, 0.12, 0.00, 0.00 };
 Int_t FI = TColor::CreateGradientColorTable(NRGBs, stops, red, green, blue, BINNING);

 for (int i=0;i<BINNING;i++)
   colors[i] = FI+i;

// will contain the results from the processors
TMultiGraph *runChargeRowGraph=new TMultiGraph();
std::vector<double> runZPositions;
std::vector<double> runZPositionsError;
std::vector<double> runHitCharge;
std::vector<double> runHitChargeError;


//loop over the given runs
for (int i=0;i<nRuns;i++)
{
	//get run number from input
	runNumber=va_arg(argList,int);

	//create stringstream containing the aida file name
	std::stringstream runNumberString;
	std::cout<<"run number: "<<runNumber<<std::endl;
	runNumberString <<pathToAida<<"aida_run" <<runNumber<< ".root";

	//open the aida file
	TFile *file  = TFile::Open(runNumberString.str().c_str(),"Read");
	//check that the file exists
	if (file==0)
	{
		std::cout<<"This run will be skipped. The aida file does not exist. Please check this!"<<std::endl;
		continue;
	}

	//get the tuples from the file
	TNtuple * zPosTuple=(TNtuple*)file->Get("MyAnalysisCutSelectorProcessor/RunZPositions");
	TNtuple * hitChargeTuple=(TNtuple*)file->Get("MyChargeQualityProcessor/ChargeQualityProcessor");
	//check that the tuples exists
	if ( hitChargeTuple==0||zPosTuple==0 )
	{
		std::cout<<"This run will be skipped. One of the necessary root tuples does not exist. Please check that you ran the corresponding processors and that the tuple is in the aida file!"<<std::endl;
		continue;

	}

	//open the first and only entry of the tuples
	zPosTuple->GetEntry(0);
	hitChargeTuple->GetEntry(0);

	//fill the vectors with the values from the tuples ([0]=mean,[2]=rms)
	//567.6-hitposition is the conversion into drift distance
	runZPositions.push_back( 567.6-zPosTuple->GetArgs()[0] );
	runZPositionsError.push_back( zPosTuple->GetArgs()[2] );
	runHitCharge.push_back( hitChargeTuple->GetArgs()[0] );
	runHitChargeError.push_back( hitChargeTuple->GetArgs()[1] );

	TGraphErrors *chargeRow=(TGraphErrors*)file->Get("MyChargeQualityProcessor/chargeRow");
    if(chargeRow==0 )
	{
	  std::cout<<"This run will be skipped. One of the necessary root graphs does not exist. Please check that you ran the corresponding processors and that the graph is in the aida file!"<<std::endl;
	  continue;
	}
	int zpos= int(567.6-zPosTuple->GetArgs()[0]) ;
    std::ostringstream legEntry;
    legEntry << "z = "<<zpos;
    chargeRow->SetLineColor(colors[i]);
    chargeRow->SetMarkerColor(colors[i]);
    chargeRow->SetTitle(legEntry.str().c_str());
    runChargeRowGraph->Add(chargeRow);

    file->Close();
    delete file;
 }

va_end(argList);

//create Graphs and fill them with the data
TGraphErrors *hitChargeGraph = new TGraphErrors( runZPositions.size(), &runZPositions[0], &runHitCharge[0], &runZPositionsError[0], &runHitChargeError[0] );
hitChargeGraph->SetTitle("");
hitChargeGraph->GetXaxis()->SetTitle("drift distance [mm]");
hitChargeGraph->GetYaxis()->SetTitle("hit charge a.u.");


//save the results
TCanvas *c1 = new TCanvas("HitCharge","hit charge");
hitChargeGraph->Draw("AP");
c1->SaveAs(Form("%s%s",pathToOutput,"HitChargeVsZ.png"));
c1->SaveAs(Form("%s%s",pathToOutput,"HitChargeVsZ.pdf"));

gStyle->SetPadRightMargin(0.18);
TCanvas *c2 = new TCanvas("hitChargeRow","hit charge per row");
runChargeRowGraph->Draw("AP");
runChargeRowGraph->SetTitle(";r_{row} [mm]; hit charge [a.u.]");
gPad->BuildLegend(0.85,0.2,0.99,0.99,"[mm]");
c2->SaveAs(Form("%s%s",pathToOutput,"HitChargePerRow.png"));
c2->SaveAs(Form("%s%s",pathToOutput,"HitChargePerRow.pdf"));




outfile->cd();
hitChargeGraph->SetName("hitCharge");
hitChargeGraph->Write();
runChargeRowGraph->SetName("hitChargePerRow");
runChargeRowGraph->Write();
outfile->Write();
outfile->Close();

delete c1;
}
