#include <vector>
#include <iostream>
#include "TLatex.h"
#include "TCanvas.h"
#include "TH1.h"
#include "TF1.h"
#include "TGraphErrors.h"
#include "TFile.h"
#include "TNtuple.h"
#include "TStyle.h"
#include "TROOT.h"
#include "TText.h"
#include "sstream"
#include <cstdarg>

/*How to call this:
.L $MARLINTPC/analysis/rootscripts/DriftVelocity.C++
DriftVelocity("path to the aida file", number of runs in the following list, list of the runs)
Can be used to calculate the drift velocity. As input on needs to add the run number and the corresponding z position in mm (e.g. which you can get from the table position in the DESY T24 testbeam area)
e.g DriftVelocity("../AidaFiles/",3,18850,10,18851,100,18860,567)
*/
void DriftVelocity(char * pathToOutput, char * pathToAida, int nRuns, ...){

//loading the style
gROOT->ProcessLine(".L ${MARLINTPC}/analysis/rootscripts/LCTPCStyle.C");
gROOT->ProcessLine("LCTPCStyle()");

//open a root file to store the graphs in
TFile *outfile=new TFile(Form("%s%s",pathToOutput,"DriftVelocity.root"),"RECREATE");

//initialize the input list
va_list argList;
va_start(argList,int(nRuns*2));


int runNumber;

// will contain the results from the processors
std::vector<double> runHitTime;
std::vector<double> runHitTimeError;
std::vector<double> runZPos;
std::vector<double> runZPosError;


//loop over the given runs
for (int i=0;i<nRuns;i++)
{
	//get run number from input
	runNumber=va_arg(argList,int);
	double zPos=va_arg(argList,double);
	runZPos.push_back(zPos);
	runZPosError.push_back(sqrt(0.2*0.2+0.0007*zPos*0.0007*zPos));
	//create stringstream containing the aida file name
	std::stringstream runNumberString;
	std::cout<<"run number: "<<runNumber<<std::endl;
	std::cout<<"z Position: "<<zPos<<std::endl;
	runNumberString <<pathToAida<<"aida_run" <<runNumber<< ".root";

	//open the aida file
	TFile *file  = TFile::Open(runNumberString.str().c_str(),"Read");
	//check that the file exists
	if (file==0)
	{
		std::cout<<"This run will be skipped. The aida file does not exist. Please check this!"<<std::endl;
		continue;
	}

	//get the tuples from the file
	TNtuple * hitTimeTuple=(TNtuple*)file->Get("MyAnalysisCutSelectorProcessor/HitTime");
	//check that the tuples exists
	if ( hitTimeTuple==0 )
	{
		std::cout<<"This run will be skipped. One of the necessary root tuples does not exist. Please check that you ran the corresponding processors and that the tuple is in the aida file!"<<std::endl;
		continue;

	}

	//open the first and only entry of the tuples
	hitTimeTuple->GetEntry(0);

	//fill the vectors with the values from the tuples ([0]=mean,[2]=rms)
	//567.6-hitposition is the conversion into drift distance
	std::cout<<"hit time"<<hitTimeTuple->GetArgs()[0]<<std::endl;
	runHitTime.push_back( hitTimeTuple->GetArgs()[0] );
	runHitTimeError.push_back( hitTimeTuple->GetArgs()[1] );
 }

va_end(argList);

//create Graphs and fill them with the data
TGraphErrors *driftVelocity = new TGraphErrors( runHitTime.size(), &runHitTime[0], &runZPos[0], &runHitTimeError[0], &runZPosError[0] );
driftVelocity->SetTitle("");
driftVelocity->SetName("DriftVelocity");
driftVelocity->GetXaxis()->SetTitle("hit time [ns]");
driftVelocity->GetYaxis()->SetTitle("beam position [mm]");


//save the results
TCanvas *c1 = new TCanvas("DriftVelocityCanvas","drift velocity");
driftVelocity->Draw("AP");
driftVelocity->GetXaxis()->SetRangeUser(-100,11000);
driftVelocity->Draw("AP");
c1->Update();
TF1 * straightLine = new TF1("straight","pol1",-100,11000);
straightLine->SetParameters(-100,0.01);
driftVelocity->Fit("straight","MFR");
straightLine->Draw("R");
driftVelocity->Draw("SAMEP");
TText * vdrift = new TText();
vdrift->DrawTextNDC(0.22,0.8,Form("vdrift = %f +- %f", 1000*straightLine->GetParameter(1), 1000*straightLine->GetParError(1)));
c1->SaveAs(Form("%s%s",pathToOutput,"DriftVelocity.png"));
c1->SaveAs(Form("%s%s",pathToOutput,"DriftVelocity.pdf"));
c1->SaveAs(Form("%s%s",pathToOutput,"DriftVelocity.tex"));

outfile->cd();
driftVelocity->Write();

outfile->Write();
outfile->Close();

delete c1;
}
