#include "ChargeQualityProcessor.h"
#include "HistogramWidthExtractionTools.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/Track.h>
#include <IMPL/LCCollectionVec.h>
#include <UTIL/LCTOOLS.h>
#include <IMPL/LCFlagImpl.h>

//marlin
#include <marlin/ConditionsProcessor.h>
#include <marlin/Global.h>

//GEAR
#include "gear/TPCModule.h"

//ROOT
#include "TGraphErrors.h"

//using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;
using namespace std;

namespace marlintpc
{

ChargeQualityProcessor aChargeQualityProcessor ;


ChargeQualityProcessor::ChargeQualityProcessor() : Processor("ChargeQualityProcessor")
,_histoQ(0),_tuple(0),_meanMethod(0),_gearParameters(0)
{
	_description = "ChargeQualityProcessor calculates the mean90 of the hit hcarge spectrum.";


	registerInputCollection(LCIO::TRACK,
			"InputCollectionName" ,
			"Name of the input collection (default: ClupatraTracks)"  ,
			_inputColName ,
			std::string("ClupatraTracks")) ;
	registerProcessorParameter("chargeRange",
			"upper limit for hit charge used for histogram",
			_Qmax,
			int(100000));

	registerProcessorParameter( "meanCalculationMethod",
			"Defines the method of calculating the mean of the charge-histograms. (default 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails)",
			_meanMethod,
			int(1));
}


void ChargeQualityProcessor::init()
{

	// usually a good idea to
	printParameters() ;
	AIDAProcessor::tree(this);
	_gearParameters= &marlin::Global::GEAR->getTPCParameters();
	_tuple=new TNtuple("ChargeQualityProcessor","ChargeQualityProcessor","mean90:mean90Error:rms90:rms90Error:mean90RangeMin:mean90RangeMax");
	_histoQ = new TH1F("hitcharge","hitcharge",1000,0,_Qmax);
	_histoQ->GetXaxis()->SetTitle("hit charge a.u.");
	_histoQ->GetYaxis()->SetTitle("entries");
}

void ChargeQualityProcessor::processRunHeader(lcio::LCRunHeader* run)
{


	run->parameters().setValue(_processorName + "_revision", "$Rev: 2743 $");

	for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
	{
		if (! i->second->isOptional() || i->second->valueSet())
		{
			run->parameters().setValue(_processorName + "_" + i->second->name(),
					i->second->value());
		}
	}


}

void ChargeQualityProcessor::processEvent(lcio::LCEvent * evt)
{
	AIDAProcessor::tree(this);
	// get the input collection from the event
	LCCollection* inputCol = 0;
	try
	{
		inputCol = evt->getCollection(_inputColName) ;
	}

	catch (DataNotAvailableException&)
	{
		// if there is not input collection give a warning to the user
		streamlog_out(MESSAGE2)
            				<< "processEvent: input collection "
            				<< _inputColName << " cannot be found in event " << evt->getEventNumber() << std::endl;
		return;
	}

	// loop the input data
	for (int t=0;t< inputCol->getNumberOfElements();t++)
	{
		lcio::Track* currentTrack = dynamic_cast<lcio::Track*>(inputCol->getElementAt(t));
		const TrackerHitVec hitvec= currentTrack->getTrackerHits();
		for(unsigned int h=0;h<hitvec.size();h++)
		{
			lcio::TrackerHit * hit= dynamic_cast<lcio::TrackerHit*>(hitvec[h]);
			_histoQ->Fill(hit->getEDep());
			const gear::TPCModule& theModule = _gearParameters->getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
			int row=theModule.getRowNumber(theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]));
			int module=theModule.getModuleID();
			std::pair<int, int> key(module,row);
			std::map< std::pair<int,int>, TH1*>::iterator histoiter;
			histoiter = _hitChargeRow.find(key);
			if(histoiter == _hitChargeRow.end())
			{
				TH1* tempHist = new TH1F(Form("hitCharge_%i_%i",module,row),"hitCharge",1000,0,0);
				_hitChargeRow.insert( std::make_pair(key,tempHist));
			}
			histoiter = _hitChargeRow.find(key);
			((*histoiter).second)->Fill( hit->getEDep() );

		}
	}
}


void ChargeQualityProcessor::check(lcio::LCEvent*  /*evt*/)
{ 
	// check what is needed here
}


void ChargeQualityProcessor::end()
{ 
	HistogramWidthExtractionTools *myTool= new HistogramWidthExtractionTools();
	myTool->Calculate(_histoQ,1);

	_tuple->Fill( myTool->GetMean(),myTool->GetMeanError()
			,myTool->GetWidth(),myTool->GetWidthError()
			,myTool->GetRange().first,myTool->GetRange().second,_histoQ->GetEntries());

	std::map< std::pair<int,int>, TH1*>::iterator histoiter = _hitChargeRow.begin();
	std::vector<double> radius,radiusError,charge,chargeError;

	for(; histoiter!=_hitChargeRow.end(); histoiter++)
	{
		std::pair<int,int> key = (*histoiter).first;
		myTool->Calculate(((*histoiter).second),_meanMethod);
		charge.push_back(myTool->GetMean());
		chargeError.push_back(myTool->GetWidth());

		const gear::TPCModule& theModule = _gearParameters->getModule(key.first);
		gear::Vector2D padcenter=theModule.getPadCenter(theModule.getPadIndex(key.second,0));
		radius.push_back(sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]));
		radiusError.push_back((theModule.getRowHeight(key.second))/sqrt(12.));
	}
	AIDAProcessor::tree(this);
	TGraphErrors *graphRow=new TGraphErrors(radius.size(),&radius[0],&charge[0],&radiusError[0],&chargeError[0]);
	graphRow->SetNameTitle("chargeRow","chargeRow");
	graphRow->GetXaxis()->SetTitle("r_{row} [mm]");
	graphRow->GetYaxis()->SetTitle("hit charge [a.u.]");
	graphRow->Write();
	delete myTool;

}



}
