#include "CutChipProcessor.h"

// C++
#include <iostream>
#include <cstdlib>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerRawDataImpl.h>
#include "UTIL/LCTOOLS.h"
#include <IMPL/LCFlagImpl.h>

//Marlin
#include <marlin/Exceptions.h>

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/ITree.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud1D.h>
#include <AIDA/IHistogram2D.h>
#endif


using namespace lcio ;
using namespace marlin ;

namespace marlintpc{


    CutChipProcessor aCutChipProcessor ;
    

    CutChipProcessor::CutChipProcessor() : 
	Processor("CutChipProcessor"),   isFirstCollection(false)
    {
	/* the constructor call first the super constructor which the name 
	 * of the processor
	 * this must by the same as the class name.
	 */
	
	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
	_description = "CutChipProcessor takes the data of the desired ChipIDs" ;
	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * plaese note to get a float you must cast in C++: (float) 1.3 
	 */
	registerInputCollection( LCIO::TRACKERRAWDATA,
				 "InputCollectionName" , 
				 "Name of the input collection"  ,
				 _inputColName ,
				 std::string("TimePixRawData") ) ;
	
	registerOutputCollection( LCIO::TRACKERRAWDATA,
				  "OutputCollectionName" , 
				  "Name of the output collection"  ,
				  _outputColName ,
				  std::string("CutedChipIdRawData") ) ;

	registerProcessorParameter( "SetOutputTransient" , 
				    "if not 0 the output collection is set transient"  ,
				    _outputIsTransient,
				    0);


        registerOptionalParameter ("UsedChipIDs",
				   "A list of chip IDs to use.",
				   _usedChipIDs,
				   _usedChipIDs);
    }
    
    
    void CutChipProcessor::init() { 
	
	// usually a good idea to
	printParameters() ;

	;
	
    }
   
    void CutChipProcessor::processRunHeader(lcio::LCRunHeader* run) { 
	
      
      run->parameters().setValue( _processorName + "_revision", "$Rev: 1659 $");

      for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	if ( ! i->second->isOptional() || i->second->valueSet() ) {
	  run->parameters().setValue( _processorName + "_" + i->second->name(), 
				      i->second->value() );
	}
      }

      
	
    } 
    
    void CutChipProcessor::processEvent(lcio::LCEvent * evt ) { 

      streamlog_out(MESSAGE1)<<"Event "<<evt->getEventNumber()<<std::endl;
      
	
	// get the input collection from the event
      LCCollection* inputCol = 0;

      try{

	inputCol = evt->getCollection( _inputColName ) ;

      } //end of try {

      catch(DataNotAvailableException){
	// if there is not input collection give a warning to the user
	streamlog_out(WARNING0)<<"ProcessEvent:"
		 <<" input collection "
		 <<_inputColName<<" cannot be found in event "
		 <<evt->getEventNumber()<<std::endl; 
	return;
      }
      

      // create the output collection with a string defining the TYPE (see lcio.h)
      LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERRAWDATA ); 

      // set the transient flag as definded by the processor parameter
      //outputCol->setTransient( (_outputIsTransient != 0) );

      // set the flags that cellID1 should be stored
      lcio::LCFlagImpl trkFlag(0) ;
      trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
      outputCol->setFlag( trkFlag.getFlag() );
     
      // get the number of elements in the input collection, 
      // for TimePixRawData usually the number of chips 
      int nRawData = inputCol->getNumberOfElements();


      //and loop over it
      for(int iRawData=0; iRawData< nRawData ; iRawData++)
	{
       
	/* getElement returns only a LCObject
	 * you have to cast it to the get the right type
	 * please use dynamic_cast which will return 0 if the object cannot be casted
	 */
	TrackerRawData* RawData = 
	  dynamic_cast<TrackerRawData*>( inputCol->getElementAt( iRawData ) ) ;
	
	if( RawData == 0 ){
	  streamlog_out(WARNING0) << "Empty object" << std::endl;
	  continue;
	}

	/*loop over the desired chips,
	 and add the RawData to the outputCol*/
	for ( std::vector< int>::const_iterator used_chip = _usedChipIDs.begin() ; used_chip != _usedChipIDs.end(); used_chip++ )
	  {	 
	  if ( RawData->getCellID1() == *used_chip )
	    { 
	    outputCol->addElement(RawData ); 
	    }
	
 
	  } //end of loop over desired chips
	
	}   //end of for(int iRawData=0; iRawData< nRawData ; iRawData++){
      
	        
      /* add the collection to the event and asign a name to it
       * please note again: DO NOT DELETE the collection this will be done by LCIO
       */

 

         if (outputCol->size() > 0)
	{ 
	  outputCol->setSubset( true );
	  evt->addCollection( outputCol, _outputColName );
	}
      else
      {	  // have to delete the collction if not added to the event
	  // the outputCol deletes all the LCObjects it contains
	  delete outputCol;
	 	streamlog_out(WARNING0)<<"outputcollection=0"<<std::endl; 
      }
	
    }
    
    
    void CutChipProcessor::check(lcio::LCEvent * evt )
    { 
      try{
	evt->getCollection( _outputColName );
      }
      catch(lcio::DataNotAvailableException)
	{
	  streamlog_out(WARNING0) << "Empty event: no cuted chip ID TimePix data  in event " 
				  <<evt->getEventNumber()<<std::endl;
	  return;
	}


    }

    
    
    void CutChipProcessor::end(){ 
    
    }
    
}
