#include "DistortionPerformanceProcessor.h"
#include "HistogramWidthExtractionTools.h"
       
#include "marlin/Global.h"

#include "gear/GearMgr.h"
#include "gear/TPCModule.h"


#include "EVENT/LCCollection.h"
#include "EVENT/Track.h"
#include "EVENT/TrackerHit.h"

#include "TGraphErrors.h"


using namespace lcio ;
using namespace marlin ;
using namespace std;

namespace marlintpc
{
  DistortionPerformanceProcessor aDistortionPerformanceProcessor;

  DistortionPerformanceProcessor::DistortionPerformanceProcessor() : Processor("DistortionPerformanceProcessor")
  {
    _description = "Processor to analyze distortions for one run";

    registerInputCollection( LCIO::TRACK,
			     "TrackInputCollectionName",
			     "Name of the input track collection",
			     _inputColName, 
			     std::string("FoundTracks"));

  }

  void DistortionPerformanceProcessor::init()
  {
    printParameters() ;
    AIDAProcessor::tree(this);
    _GearParameters= &marlin::Global::GEAR->getTPCParameters();
    std::vector<gear::TPCModule*> TPCModules = _GearParameters->getModules();

    std::vector<gear::TPCModule*>::const_iterator moduleIter = TPCModules.begin();
    int moduletype0 = (*moduleIter)->getPadLayoutType();

    for( ; moduleIter != TPCModules.end() ; ++moduleIter)
      {
	if((*moduleIter)->getPadLayoutType() != moduletype0)
	  streamlog_out(MESSAGE) << "WARNING: Modules with different pad layouts are defined. Will use first one!"<< std::endl;
      }

	_xyDistortionsTuple = new TNtuple("xy distortions","xyDistortions","module:row:mean_xy:mean_xyError:sigma_xy:sigma_xyError");
	_zDistortionsTuple = new TNtuple("z distortions","zDistortions","module:row:mean_z:mean_zError:sigma_z:sigma_zError");

    _myModuleDistortionCalculator=new ModuleDistortionCalculator( &marlin::Global::GEAR->getTPCParameters());
    _failedIntersections=0;
    _intersectionCalls=0;
  }

  void DistortionPerformanceProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/)
  {
  }

  void DistortionPerformanceProcessor::processEvent(lcio::LCEvent* evt)
  {
    //necessary if histograms are created here instead of in init(), 
    //to get histograms in the correct folder
    AIDAProcessor::tree(this);
    LCCollection* inputCol;

    try
      {
	inputCol = evt->getCollection(_inputColName);
      }
    catch (DataNotAvailableException &e)
      {
	streamlog_out(DEBUG) << "No data available in event for " <<_inputColName << std::endl;
	return;
      }
   
    //loop the tracks
    for(int t=0;t<inputCol->getNumberOfElements();t++)
      {
	lcio::Track *track=dynamic_cast<lcio::Track*>(inputCol->getElementAt(t));
	lcio::TrackerHitVec HitsinTrack = track->getTrackerHits();
	lcio::TrackerHitVec::const_iterator itHit = HitsinTrack.begin();

	//loop the hits
	for(; itHit != HitsinTrack.end(); ++itHit)
	  {
	    TrackerHit* hit = *itHit;
	    //find out module and row of the hit.
	    const gear::TPCModule& theModule = _GearParameters->getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
	    int row=theModule.getRowNumber(theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]));
	    int module=theModule.getModuleID();
	    std::pair<int, int> key(module,row);
	    
	    std::map< std::pair<int,int>, TH1*>::iterator histoiter;
	    histoiter = _HistogramsZ.find(key);
	    
	    if(histoiter != _HistogramsZ.end())
	      {
		((*histoiter).second)->Fill( _myModuleDistortionCalculator->calculateDistanceZ(hit,track) );
	      }
	    else
	      {
		TH1* tempHist = new TH1F(Form("HistZ_%i_%i",module,row),"HistZ",256,0,0);
		tempHist->Fill(_myModuleDistortionCalculator->calculateDistanceZ(hit,track) );
		_HistogramsZ.insert( std::make_pair(key,tempHist));
	      }
	    
	    histoiter = _HistogramsRow.find(key);
	    try{
	      if(histoiter != _HistogramsRow.end())
		{
		  ((*histoiter).second)->Fill( _myModuleDistortionCalculator->calculateDistanceRow(hit,track) );
		  _intersectionCalls++;
		}
	      else
		{
		  TH1* tempHist = new TH1F(Form("HistRow_%i_%i",module,row),"HistRow",256,0,0);
		  tempHist->Fill(_myModuleDistortionCalculator->calculateDistanceRow(hit,track) );
		  _HistogramsRow.insert( std::make_pair(key,tempHist));
		}
	    }
	    catch(std::exception &e) {
	      std::cout<<"calculateDistanceRow failed. Value ignored!"<<std::endl;
	      _failedIntersections++;
	    }
	  }//hit loop
      }//track loop
  }
  
  void DistortionPerformanceProcessor::check(lcio::LCEvent*  /*evt*/)
  {
  }

  void DistortionPerformanceProcessor::end()
  {
    //necessary if histograms are created here instead of in init(), 
    //to get histograms in the correct folder
    AIDAProcessor::tree(this);
    //all histograms are filled
    //now we need to fit them
    HistogramWidthExtractionTools *myTool= new HistogramWidthExtractionTools();	
    std::map< std::pair<int,int>, TH1*>::iterator histoiter = _HistogramsZ.begin();
    for(; histoiter!=_HistogramsZ.end(); histoiter++)    
      {
	std::pair<int,int> key = (*histoiter).first;
	//use rms90 method in z
	myTool->Calculate(((*histoiter).second),1);
	std::pair<double,double> distortions;
	distortions = std::make_pair(myTool->GetMean(),myTool->GetMeanError());
	_ResultsZ.insert(std::make_pair(key,distortions));
  	_zDistortionsTuple->Fill(key.first,key.second,myTool->GetMean(),myTool->GetMeanError(),
  			myTool->GetWidth(),myTool->GetWidthError());
      }
    
    histoiter = _HistogramsRow.begin();
    for(; histoiter!=_HistogramsRow.end(); histoiter++)    
      {
	std::pair<int,int> key = (*histoiter).first;
	//use gaussian 3 sigma fit in rphi
	myTool->Calculate(((*histoiter).second),2);
	std::pair<double,double> distortions;
	distortions = std::make_pair(myTool->GetMean(),myTool->GetMeanError());
	_ResultsRow.insert(std::make_pair(key,distortions));
  	_xyDistortionsTuple->Fill(key.first,key.second,myTool->GetMean(),myTool->GetMeanError(),
  			myTool->GetWidth(),myTool->GetWidthError());
      }
    
    //now put data in graph
    //ROW
    std::vector<double> x,y,xerr,yerr;

    std::map< std::pair<int,int>, std::pair<double,double> >::iterator distiter =  _ResultsRow.begin();
    for(; distiter != _ResultsRow.end(); ++distiter)
      {
	int module = ((*distiter).first).first;
	int row = ((*distiter).first).second;
	double distortion = ((*distiter).second).first;
	double distortionerr = ((*distiter).second).second;
	//get from to corresponding radius at phi=0
	const gear::TPCModule& theModule = _GearParameters->getModule(module);
	gear::Vector2D padcenter=theModule.getPadCenter(theModule.getPadIndex(row,0));
	double radius=sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]);

	x.push_back(radius);
	xerr.push_back((theModule.getRowHeight(row))/sqrt(12.));
	y.push_back(distortion);
  	yerr.push_back(distortionerr);
      }
    TGraphErrors *graph=new TGraphErrors(x.size(),&x[0],&y[0],&xerr[0],&yerr[0]);
    graph->SetNameTitle("XYDistortionRow","XYDistortionRow");
    graph->GetXaxis()->SetTitle("row radius [mm]");
    graph->GetYaxis()->SetTitle("distortion in rphi [mm]");
    //for some reason graphs need a dedicated write call to appear in the aida file
    graph->Write();


    //Z
    std::vector<double> xZ,yZ,xerrZ,yerrZ;

    distiter =  _ResultsZ.begin();
 
    for(; distiter != _ResultsZ.end(); ++distiter)
      {
	int module = ((*distiter).first).first;
	int row = ((*distiter).first).second;
	double distortion = ((*distiter).second).first;
	double distortionerr = ((*distiter).second).second;
	//get from to corresponding radius at phi=0
	const gear::TPCModule& theModule = _GearParameters->getModule(module);
	gear::Vector2D padcenter=theModule.getPadCenter(theModule.getPadIndex(row,0));
	double radius=sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]);

	xZ.push_back(radius);
	xerrZ.push_back((theModule.getRowHeight(row))/sqrt(12.));
	yZ.push_back(distortion);
  	yerrZ.push_back(distortionerr);

       }
    TGraphErrors *graphZ=new TGraphErrors(xZ.size(),&xZ[0],&yZ[0],&xerrZ[0],&yerrZ[0]);
    graphZ->SetNameTitle("ZDistortionRow","ZDistortionRow");
    graphZ->GetXaxis()->SetTitle("row radius [mm]");
    graphZ->GetYaxis()->SetTitle("distortion in z [mm]");
    graphZ->Write();
    
    std::cout<<"Failed intersections: "<< _failedIntersections<<" in %: "<<double(_failedIntersections)/_intersectionCalls*100.<<std::endl;

	delete myTool;
	delete graph;
	delete graphZ;
  }

 
  

}//end namespace
