#include "HistogramWidthExtractionTools.h"

#include "marlin/VerbosityLevels.h"

#include "TF1.h"
#include <cmath>


namespace marlintpc
{

  HistogramWidthExtractionTools:: HistogramWidthExtractionTools():
    _width(0),_width_err(0),_mean(0),_mean_err(0),_height(0),_height_err(0)
  {
  }

  HistogramWidthExtractionTools::~HistogramWidthExtractionTools(){}

  
  void HistogramWidthExtractionTools::Calculate(TH1* histogram, int method)
  {
    switch(method)
      {
      case 1:
	CalculateRMS90(histogram);
 	break;
      case 2:
	Fit3SigmaRange(histogram);
 	break;
      case 3:
	CalculateRMS5SigmaRange(histogram);
 	break;
      case 4:
	FitLowerHalf3SigmaRange(histogram);
 	break;
      }
  }


  void HistogramWidthExtractionTools::CalculateRMS90(TH1* histogram)
  {
    // std::cout<<"CALLING: HistogramWidthExtractionTools::CalculateRMS90()"<<std::endl;
    if ( histogram== NULL) return;
        
    float sum   = 0;
    double sx   = 0.;
    double sxx  = 0.;
    
    int nbins = histogram->GetNbinsX();
    
    float total = 0;
    
    for(int i = 1; i <= nbins; i++)
      {
	float binx = histogram->GetBinLowEdge(i)+histogram->GetBinWidth(i)/2.0;
	float yi   = histogram->GetBinContent(i);
	sx        += yi*binx;
	sxx       += yi*binx*binx;
	total     +=yi;
      }

    int is0 = 0;
    
    for(int i = 1; i <= nbins && sum < total/10.; i++)
      {
	float yi = histogram->GetBinContent(i);
	sum      += yi;
	is0      = i;
      }
  
    float xmean = 0;
    /** TODO: warning: variable 'xrms' set but not used
     * float xrms  = 0;
     */

    for(int istart = 1; istart <= is0; istart++)
      {
	double sumn   = 0.;
	double csum   = 0.;
	double sumx   = 0.;
	double sumxx  = 0.;
	int iend      = 0;

	for(int i = istart; i <= nbins && csum <0.9*total; i++)
	  {
	    float binx = histogram->GetBinLowEdge(i)+histogram->GetBinWidth(i)/2.0;
	    float yi   = histogram->GetBinContent(i); 
	    csum += yi;

	    if(sumn < 0.9*total)
	      {
		sumn   += yi;
		sumx   += yi*binx;
		sumxx  += yi*binx*binx;
		iend   = i;
	      }
	  }
      
	float mean   = sumx/sumn;
	float meansq = sumxx/sumn;
	float rms    = sqrt(meansq - mean*mean);

	if(istart == 1 || rms < _width)
	  {
	    xmean = mean;
	    /** TODO: warning: variable 'xrms' set but not used, see decl. above
	     * xrms  = rms;
	     */
	    _range.first  = histogram->GetBinLowEdge(istart);
	    _range.second = histogram->GetBinLowEdge(iend);
	    
	    _width     = rms;
	    _mean    = xmean;
	  }
      }
    //converting from rms90 to something resembling a sigma
    _width=_width/0.8;
    //first shot at getting an error for rms90 and mean90 value. Needs further work  
    histogram->GetXaxis()->SetRangeUser(_range.first,_range.second);
    _width_err=histogram->GetRMSError();
    _mean_err=histogram->GetMeanError();
    histogram->GetXaxis()->SetRangeUser(0,-1);

  }

  void HistogramWidthExtractionTools::Fit3SigmaRange(TH1* histogram)
  {
    // std::cout<<"CALLING: HistogramWidthExtractionTools::Fit3SigmaRange()"<<std::endl;
    TF1 *fit=new TF1("fit","gaus");
    histogram->Fit(fit,"Q");
    TF1 *fit_range=new TF1("fit_range","gaus",fit->GetParameter(1)-3*fit->GetParameter(2),fit->GetParameter(1)+3*fit->GetParameter(2));
    histogram->Fit(fit_range,"QR");
    _width=fit_range->GetParameter(2);
    _mean=fit_range->GetParameter(1);
    _width_err=fit_range->GetParError(2);
    _mean_err=fit_range->GetParError(1);
    fit_range->GetRange(_range.first, _range.second);
    //_range.first  = fit->GetParameter(1)-3*fit->GetParameter(2);
    //_range.second = fit->GetParameter(1)+3*fit->GetParameter(2);
    _height = fit->GetParameter(0);
    _height_err = fit->GetParError(0);

	delete fit;
	delete fit_range;
  }

  void HistogramWidthExtractionTools::CalculateRMS5SigmaRange(TH1* histogram)
  {
    //  std::cout<<"CALLING: HistogramWidthExtractionTools::CalculateRMS5SigmaRange()"<<std::endl;
    Double_t rms = 0;
    
    // do max 10 iterations to avaoid endless loop
    
    for (int n=0 ; n < 10 ; n++ )
      {    
	// Store the rms of the current Histogram
	rms = static_cast<float>( histogram->GetRMS() );
	// Set the range of the histogram to mean +- 5 rms 
	// This results in a smaller rms of the histo
	//   cout << rms << endl;
	histogram->GetXaxis()->SetRangeUser(histogram->GetMean()-5*rms, histogram->GetMean()+5*rms);
	// quit the loop if the procedure has converged
	if ( fabs(rms - histogram->GetRMS()) < 1e-6 )
	  {
	    break;
	  }
	if ( n == 10 )
	  {
	    streamlog_out(WARNING) << "5 sigma cut did not converge after 10 iterations in"<< std::endl;
	  }  
	
      }
     _width=histogram->GetRMS();
     _mean=histogram->GetMean();
     _width_err=histogram->GetRMSError();
     _mean_err=histogram->GetMeanError();
     _range.first  = histogram->GetMean()-5*rms;
     _range.second = histogram->GetMean()+5*rms;

     histogram->GetXaxis()->SetRangeUser(0,-1);
  }

	void HistogramWidthExtractionTools::FitLowerHalf3SigmaRange(TH1* histogram)
	{
		TF1 *fit=new TF1("fit","gaus");
		double lower_bound = 0;
		double upper_bound = 0;
		bool was_successful = false;

		histogram->Fit(fit,"Q");

		for(int n=0 ; n<10 ; ++n)
		{
			lower_bound = fit->GetParameter(1) - 3*fit->GetParameter(2);
			upper_bound = fit->GetParameter(1) + 0.5*fit->GetParameter(2);

			histogram->Fit(fit,"Q","",lower_bound,upper_bound);

			if( fabs( lower_bound / (fit->GetParameter(1)-3*fit->GetParameter(2)) - 1 ) < 1e-6
			&& fabs( upper_bound / (fit->GetParameter(1)+0.5*fit->GetParameter(2)) - 1 ) < 1e-6 )
			{
				was_successful = true;
				break;
			}
		}

		if(!was_successful) streamlog_out(WARNING) << "Lower half 3 sigma fit did not converge after 10 iterations for " << histogram->GetName() << std::endl;

		_width=fit->GetParameter(2);
		_mean=fit->GetParameter(1);
		_width_err=fit->GetParError(2);
		_mean_err=fit->GetParError(1);
		fit->GetRange(_range.first, _range.second);
		_height = fit->GetParameter(0);
		_height_err = fit->GetParError(0);

		delete fit;
	}

}
