#include "HitAndTrackCounterProcessor.h"
//#include "ZBinTemplateProcessor.cc"
#include <CLHEP/Vector/ThreeVector.h>
//C++
#include <iostream>
#include <cmath>
#include <algorithm>
using namespace std;

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
// Aida
//#ifdef MARLIN_USE_AIDA // the whole processor does not make sense without aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
// endif

#include <marlin/Global.h>

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

// define the indices for the template parameter
// they are for convenience and better code readability
const unsigned int HitAndTrackCounterProcessor::_hitsPerTrackHistosIndex   = 0;
const unsigned int HitAndTrackCounterProcessor::_hitsPerTrackLengthHistosIndex   = 1;
//const unsigned int HitAndTrackCounterProcessor::_hitsPerEventHistosIndex   = 1;

HitAndTrackCounterProcessor aHitAndTrackCounterProcessor;
    
    HitAndTrackCounterProcessor::HitAndTrackCounterProcessor(): 
	ZBinTemplateProcessor<N_1D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR ,
			      N_2D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR>("HitAndTrackCounterProcessor"),
	_hitsPerEventHisto(0),
	_tracksPerEventHisto(0)
{
    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(    LCIO::TRACK,
				"InputTracks" , 
				"Name of the input tracks collection (default: TPCTracks)"  ,
				_tracksColName ,
				std::string("TPCTracks") ) ;
    registerInputCollection(    LCIO::TRACKERHIT,
				"InputHits" , 
				"Name of the input hits collection (default: TPCHits)"  ,
				_hitsColName ,
				std::string("TPCHits") ) ;
    // modify processor description
    _description = "Count hits in events and tracks, and track per event" ;

    registerProcessorParameter( "MaxHitsPerTrack" , 
				"Maximum value in HitsPerTrack  histogram (default: 20 )"  ,
				_maxHitsPerTrack,
				double(20) ) ;

  registerProcessorParameter( "MaxHitsPerEvent" , 
			      "Maximum value in HitsPerEvent  histogram (default: 50 )"  ,
			      _maxHitsPerEvent,
			      double(50) ) ;

  registerProcessorParameter( "MaxTracksPerEvent" , 
			      "Maximum value in TracksPerEvent  histogram (default: 10 )"  ,
			      _maxTracksPerEvent ,
			      double(10) ) ;
  registerProcessorParameter( "MaxHitsPerTrackLength" , 
			      "Maximum value in HitsPerTrackLength  histogram (default: 60 )"  ,
			      _maxHitsPerTrackLength ,
			      double(20) ) ;


  // set correct names for the z binned histograms
  _histo1DBaseNames[_hitsPerTrackHistosIndex]	   = "hitsPerTrackHisto";
  _histo1DTitles[_hitsPerTrackHistosIndex] = "Number of hits per track";
  
  _histo1DBaseNames[_hitsPerTrackLengthHistosIndex]	   = "hitsPerTrackLengthHisto";
  _histo1DTitles[_hitsPerTrackLengthHistosIndex] = "Number of hits per tracklength";
}

HitAndTrackCounterProcessor::~HitAndTrackCounterProcessor()
{
}

void HitAndTrackCounterProcessor::init()
{ 
	// usually a good idea to
	printParameters() ;

	// get the not-z-binned histograms from the AIDA processor
	_hitsPerEventHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "hitsPerEventHisto", 
			     "Number of hits per event", 
			     (int)(_maxHitsPerEvent+1), -0.5, _maxHitsPerEvent);

	_tracksPerEventHisto =   AIDAProcessor::histogramFactory(this)->
	  createHistogram1D( "tracksPerEventHisto", 
			     "Number of tracks per event", 
			     (int)(_maxTracksPerEvent+1), -0.5, _maxTracksPerEvent);

	// set histo range and one bin per int for templated histograms
	// according to Processor parameters
	_histoRanges1D[_hitsPerTrackHistosIndex].max = _maxHitsPerTrack;
	_nBins1D[_hitsPerTrackHistosIndex] = (int)_maxHitsPerTrack;
	
	_histoRanges1D[_hitsPerTrackLengthHistosIndex].max = _maxHitsPerTrackLength;
	_nBins1D[_hitsPerTrackLengthHistosIndex] = (int)(_maxHitsPerTrackLength*50);

	// now call ZBinTemplateProcessor's init which initialises the histograms for us
	ZBinTemplateProcessor<N_1D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR,
	                      N_2D_HISTOS_HIT_TRACK_COUNTER_PROCESSOR>::init();
}

void HitAndTrackCounterProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/ )
{
}
    
void HitAndTrackCounterProcessor::processEvent(lcio::LCEvent * evt ) 
{ 
    //read input data

    // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
    LCCollectionVec* tracksCollection=0;
    try{
	tracksCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _tracksColName )) ;
    }
    catch(DataNotAvailableException &e)
    {
	// there are no tracks
	// nevermind, just catch the exception and continue 
	streamlog_out(DEBUG0) << "no input data returning"<<std::endl;
    }

    typedef std::vector< std::pair<TrackerHit*, double> > HitDouble;
	
    // if there are tracks count them and the hits per track
    if( tracksCollection != 0 )
    {
	for(LCCollectionVec::iterator it = tracksCollection->begin() ; 
	    it < tracksCollection->end(); it ++ )
	{
	    Track * track = dynamic_cast<Track *>(*it) ;
	    
	    TrackerHitVec hitVector = track->getTrackerHits();
	    float sumZ=0;
//	    cout << "nr of hits " << hitVector.size()<< endl;
	    HitDouble HitPairVec;
//	    int hitNr=0;
	  
	    // first calculate some track parameters from track:
	    double phi =  track->getPhi(); // just to improve readability

	    // unit vector from reference point to point of closest approach (pca) (xy plane only)
	    CLHEP::Hep3Vector n_pca( -sin( phi ), cos ( phi ) );

	    // reference point in xy projection
	    CLHEP::Hep3Vector reference_point( track->getReferencePoint ()[0], 
					   track->getReferencePoint ()[1],
					   track->getReferencePoint ()[2]);

	    // point of closest approach (3D point which is closest in the xy plane)
	    CLHEP::Hep3Vector pca =  reference_point 
	    +  track->getD0() *  n_pca  // the xy part
	    +  track->getZ0() * CLHEP::Hep3Vector(0,0,1); // the z part
	    
	    streamlog_out(DEBUG4) << "PCA is " << pca[0] << " , " << pca[1] << " , " << pca[2] << std::endl;
	    
	    // psi is the angle of the pca wrt the helix center
	    // (Note: in the paper from Lasiuk et al. from 1998 this is Phi_0,
	    //  while Psi is the angle in track direktion.
	    //  As Phi is the angle in track direction in LCIO,
	    //  Phi and psi are exchanged wrt. the Lasiuk paper.)
	    double psi = track->getPhi() + ( track->getOmega()<0 ? -1 : 1 ) * M_PI_2;
	    
	    // wrap psi to -2pi .. 2pi to prevent the while loop running for ages 
	    // in case there are nonsense values for the track's phi
	    if ( std::fabs(psi) > 20 ) 
	    {
	      streamlog_out(WARNING3) << "Psi is " << psi << ", check phi of the input tracks!"
	      << std::endl;
	    }
	    psi = std::fmod(psi, 2*M_PI);

	
	
	
//	    cout << "\t\t hit\t x \t y \t z \t residual_x/r*phi" << endl;
	    for (TrackerHitVec::iterator hitIter = hitVector.begin();
		 hitIter <  hitVector.end(); hitIter ++)
	    {
	      TrackerHit* trackerHit = *hitIter;
	      // A Hep3Vector for the test hit
	      CLHEP::Hep3Vector hit(trackerHit->getPosition()[0],
			      trackerHit->getPosition()[1],
			      trackerHit->getPosition()[2] );
	      sumZ+=hit[2];
	      double s; // the track length in the xy-plane (projection)
	      if (  track->getOmega() != 0. ) 
	      {
		double co = std::fabs(track->getOmega())*(hit[0] - pca[0]) + cos(psi);
		double si = std::fabs(track->getOmega())*(hit[1] - pca[1]) + sin(psi);
		streamlog_out(DEBUG2)<< "TrackFitterBase: co "<<co <<" : si " << si 
		<< " : psi " << psi << std::endl;
		
		
		double phi =  ( co == 0 ? ( si < 0 ? - M_PI_2    : M_PI_2 ) 
		: ( co > 0 ? - atan(si/co) : M_PI - atan(si/co) ) );
		
		// wrap phi to be in +- pi wrt. -psi (minus because the pca is at -psi wrt helix centre)
		while ( -psi - phi >  M_PI ) phi+=2*M_PI ;
		while ( -psi - phi < -M_PI ) phi-=2*M_PI ;
		
		s = (-psi - phi) / track->getOmega();
		//	  s = ( atan(si/co) + ( co < 0 ? M_PI : 0 ) - psi ) / track->getOmega();
		
		//	  //This is the method from Lasiuk et al. It does not work because 
		//	  // the atan only works for +- pi/2, but we need a full circle
		//	  s = atan( ( (hit[1] - pca[1] ) * sin(phi) + (hit[0] -pca[0]) * cos(phi) ) /
		//		    ( 1/std::fabs( track->getOmega() ) +
		//		      (hit[0] - pca[0]) * sin(phi) -
		//		      (hit[1] - pca[1]) * cos(phi)                          
		//		    )
		//		  )
		//	      /  track->getOmega() ;
		streamlog_out(DEBUG3)<< "TrackFitterBase: Hit "<<hit[0] <<" : " <<hit[1] <<" : "<<hit[2] 
		<<" ; s is " << s << std::endl;
	      }
	      else
	      {
		s =  (hit[1] - pca[1] ) * sin(phi) + (hit[0] -pca[0]) * cos(phi);	// tracklength in xy projection
		s = s * std::sqrt( 1 + track->getTanLambda()*track->getTanLambda()); 	// here comes the third dimension
		streamlog_out(DEBUG3)<< "TrackFitterBase: Hit "<<hit[0] <<" : " <<hit[1] <<" : "<<hit[2] 
		<<" ; s is " << s << std::endl;
	      }
	       
	      std::pair<TrackerHit*, double> HitPair( trackerHit , s);
	      HitPairVec.push_back( HitPair );
	      
	    }
	    sort( HitPairVec.begin(), HitPairVec.end(), compareHit );
	    
	    //effective length of track is distance between outer hits times n/(n-1):
	    //distance between outer points on track
	    double l = fabs( (*HitPairVec.begin()).second - (*(HitPairVec.end()-1)).second );
	    //effective length
	    double l_eff = l*hitVector.size()/(hitVector.size()-1);
	    
	    // fill the number if hits per track length into the histogram, use z0 of the track as z position
	    fillHistogram( _hitsPerTrackLengthHistosIndex, (track->getTrackerHits().size())/l_eff , sumZ/hitVector.size() );
	    // fill the number if hits per track into the histogram, use z0 of the track as z position
	    fillHistogram( _hitsPerTrackHistosIndex, track->getTrackerHits().size() , sumZ/hitVector.size() );
	}

	_tracksPerEventHisto->fill(tracksCollection->getNumberOfElements());
    }
    else
    {
	_tracksPerEventHisto->fill( 0 );	
    }

    try{
	_hitsPerEventHisto->fill( evt->getCollection( _hitsColName )->getNumberOfElements() ) ;
    }
    catch(DataNotAvailableException &e)
    {
	// there are no hits
	// just continue
    }
}


}// namespace 
