#include "HitQualityProcessor.h"

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <Exceptions.h>

// MarlinTPC
#include "FlagwordDefinitions.h"

using namespace lcio;
using namespace marlin;
using namespace std;





namespace marlintpc
{

HitQualityProcessor aHitQualityProcessor;


HitQualityProcessor::HitQualityProcessor() : Processor("HitQualityProcessor")
,_overRangeHitsHisto(0),_deadChannelsHitsHisto(0)
{
	// the processor description
	_description = "HitQualityProcessor creates histograms for the number of hits in over range and with dead channel";

	/* register steering parameters: name, description, class-variable, default value
	 */

	registerInputCollection(    LCIO::TRACK,
			"InputCollectionName" ,
			"Name of the input Track collection, default: TPCTracks"  ,
			_inputColName ,
			string("TPCTracks")) ;

	registerProcessorParameter( "OverRangeHistoMin",
			"Minimum of histogram of the over range hits, default: 0. %",
			_overRangeHistoMin,
			float(0.));

	registerProcessorParameter( "OverRangeHistoMax" ,
			"Minimum of histogram of the over range hits, default: 100. %",
			_overRangeHistoMax,
			float(100.));

	registerProcessorParameter( "DeadChannelHistoMin",
			"Minimum of histogram of the hits with dead channels, default: 0. %",
			_deadChannelHistoMin,
			float(0.));

	registerProcessorParameter( "DeadChannelHistoMax" ,
			"Minimum of histogram of the hits with dead channels, default: 100. %",
			_deadChannelHistoMax,
			float(100.));
}


void HitQualityProcessor::init()
{ 
	streamlog_out(DEBUG) << " init() called  " << endl;

	// you might want to list the used parameters by invoking
	printParameters();
	AIDAProcessor::tree(this);
	_overRangeHitsHisto =  new TH1D( "OverRangeHitsHisto",
			"Hits in over range",
			1024, _overRangeHistoMin , _overRangeHistoMax);
	_overRangeHitsHisto->GetXaxis()->SetTitle("hits in over range [%]");
	_overRangeHitsHisto->GetYaxis()->SetTitle("# events");

	_deadChannelsHitsHisto =  new TH1D( "DeadChannelsHitsHisto",
			"Hits with dead channels",
			1024, _deadChannelHistoMin , _deadChannelHistoMax);
	_deadChannelsHitsHisto->GetXaxis()->SetTitle("hits with dead channels [%]");
	_deadChannelsHitsHisto->GetYaxis()->SetTitle("# events");

	_overRangeTuple=new TNtuple("OverRangeHits","OverRangeHits","mean90:mean90Error:rms90:rms90Error:mean90RangeMin:mean90RangeMax:nEntries");

	_deadChannelsTuple=new TNtuple("DeadChannelsHits","DeadChannelsHits","mean90:mean90Error:rms90:rms90Error:mean90RangeMin:mean90RangeMax:nEntries");

}


void HitQualityProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
	// do something for each run
} 


void HitQualityProcessor::processEvent(LCEvent* evt)
{ 
	// do something for each event, usually access the input collection:

	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		return;
	}

	// loop the data collection and do something
	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
			inputObjectIter < inputColVec->end() ;
			inputObjectIter++ )
	{
		// dynamic cast the input object (LCObject) to what it really is e.g.
		Track * inputTrack = dynamic_cast<Track *>( *inputObjectIter );

		TrackerHitVec hitVec = inputTrack->getTrackerHits();
		streamlog_out(DEBUG4)<<"number of hits " << hitVec.size()<<std::endl;
		int nDeadChannelHits=0;
		int nOverRangeHits=0;
		for ( TrackerHitVec::iterator hitIter = hitVec.begin(); hitIter < hitVec.end(); hitIter++ )
		{
			//check the quality bits of the hits whether you have dead channels or over flow pulses
			int quality=(*hitIter)->getQuality();
			streamlog_out(DEBUG4) << "Quality: " << quality << std::endl;
			streamlog_out(DEBUG4) << "hitflag::hasDeadChannel(quality): " << hitflag::hasDeadChannel(quality) << std::endl;
			streamlog_out(DEBUG4) << "hitflag::hasOverflowPulse(quality): " << hitflag::hasOverflowPulse(quality) << std::endl;
			if (hitflag::hasDeadChannel(quality)) nDeadChannelHits++;
			if (hitflag::hasOverflowPulse(quality)) nOverRangeHits++;
		}

		_overRangeHitsHisto->Fill(nOverRangeHits/double(hitVec.size())*100);
		_deadChannelsHitsHisto->Fill(nDeadChannelHits/double(hitVec.size())*100);

	}

}


void HitQualityProcessor::check(LCEvent*  /*evt*/)
{ 
	// check what is needed here
}


void HitQualityProcessor::end()
{ 

	_overRangeTuple->Fill( _overRangeHitsHisto->GetMean(),_overRangeHitsHisto->GetMeanError()
			,_overRangeHitsHisto->GetRMS(),_overRangeHitsHisto->GetRMSError()
			,_overRangeHitsHisto->GetXaxis()->GetXmin(),_overRangeHitsHisto->GetXaxis()->GetXmax(),_overRangeHitsHisto->GetEntries() );

	_deadChannelsTuple->Fill( _deadChannelsHitsHisto->GetMean(),_deadChannelsHitsHisto->GetMeanError()
			,_deadChannelsHitsHisto->GetRMS(),_deadChannelsHitsHisto->GetRMSError()
			,_deadChannelsHitsHisto->GetXaxis()->GetXmin(),_deadChannelsHitsHisto->GetXaxis()->GetXmax(),_deadChannelsHitsHisto->GetEntries());

}
} // close namespace brackets
