#include "NoiseHitAndPulseCounterProcessor.h"
#include <iostream>
#include <string>
#include <algorithm>

#include "marlin/Global.h"

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/TrackerPulse.h>
#include <Exceptions.h>

// GEAR
#include <gear/TPCModule.h>

//ROOT
#include "TH1.h"
#include "TNtuple.h"

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{

NoiseHitAndPulseCounterProcessor aNoiseHitAndPulseCounterProcessor;


NoiseHitAndPulseCounterProcessor::NoiseHitAndPulseCounterProcessor() : Processor("NoiseHitAndPulseCounterProcessor")
{
	// the processor description
	_description = "NoiseHitAndPulseCounterProcessor counts the number of hits and pulses, which are not included in tracks and separately also those included in tracks for comparison. It also maps the location of noise pulses.";
	
	//register steering parameters: name, description, class-variable, default value
	registerInputCollection( LCIO::TRACK,
			"InputTracks",
			"Name of the input track collection, default: TPCTracks",
			_inputTrackColName,
			string("TPCTracks") );

	//TODO? Make optional if possible
	registerInputCollection( LCIO::TRACKERHIT,
			"InputHits",
			"Name of the input hit collection; make sure that this is the collection used to form the input tracks; default: TPCHits",
			_inputHitColName,
			string("TPCHits") );

	//TODO? Make optional if possible
	registerInputCollection( LCIO::TRACKERPULSE,
			"InputPulsesMapped",
			"Name of the mapped input pulse collection, default: TPCPulsesMapped",
			_inputPulseColName,
			string("TPCPulsesMapped") );

	registerProcessorParameter( "NoiseHitsHistoMax",
			"Maximum of NoiseHitsHisto, default: 100",
			_noiseHitsHistoMax,
			int(100) );

	registerProcessorParameter( "NoisePulsesHistoMax",
			"Maximum of NoisePulsesHisto, default: 1000",
			_noisePulsesHistoMax,
			int(1000) );

	registerProcessorParameter( "NotInHitPulsesHistoMax",
			"Maximum of NotInHitPulsesHisto, default: 100",
			_notInHitPulsesHistoMax,
			int(100) );

	registerProcessorParameter( "TrackPulsesHistoMax",
			"Maximum of TrackPulsesHisto, default: 840",
			_trackPulsesHistoMax,
			int(840) );
}


void NoiseHitAndPulseCounterProcessor::init()
{ 
	streamlog_out(DEBUG) << " init() called	" << endl;
	
	// you might want to list the used parameters by invoking
	printParameters();

	AIDAProcessor::tree(this);

	_gearParameters = &marlin::Global::GEAR->getTPCParameters();
	DoubleVec planeExtent = _gearParameters->getPlaneExtent();
	streamlog_out(DEBUG4) << "plane extent: x=" << planeExtent[0] << ".." << planeExtent[1] << "; y=" << planeExtent[2] << ".." << planeExtent[3] << endl;

	_noiseHitsHisto = new TH1I( "NoiseHitsHisto", 
		"Noise Hits;# hits not in track;# events",
		 _noiseHitsHistoMax+1, -0.5, _noiseHitsHistoMax+0.5 );

	_noisePulsesHisto = new TH1I( "NoisePulsesHisto",
		 "Noise Pulses;# pulses not in track;# events",
		 _noisePulsesHistoMax+1, -0.5, _noisePulsesHistoMax+0.5 );

	_notInHitPulsesHisto = new TH1I( "NotInHitPulsesHisto",
		 "PulsesNotInHit;# pulses not in hit;# events",
		 _notInHitPulsesHistoMax+1, -0.5, _notInHitPulsesHistoMax+0.5 );

	_noiseHitsNormHisto = new TH1I( "NoiseHitsNormHisto", 
		"Noise Hits;# hits not in track / # hits;# events",
		 1001 , -0.0005, 1.0005 );

	_noisePulsesNormHisto = new TH1I( "NoisePulsesNormHisto",
		 "Noise Pulses;# pulses not in track / # pulses;# events",
		 1001 , -0.0005, 1.0005 );

	_notInHitPulsesNormHisto = new TH1I( "NotInHitPulsesNormHisto",
		 "Pulses Not in Hit;# pulses not in hit / # pulses;# events",
		 1001 , -0.0005, 1.0005 );

	_trackHitsHisto = new TH1I( "TrackHitsHisto", 
		"Track Hits;# hits in track;# events",
		 85, -0.5, 84.5 );

	_trackPulsesHisto = new TH1I( "TrackPulsesHisto",
		 "Track Pulses;# pulses in track;# events",
		 _trackPulsesHistoMax+1, -0.5, _trackPulsesHistoMax+0.5 );

	_trackHitsNormHisto = new TH1I( "TrackHitsNormHisto", 
		"Track Hits;# hits in track / # hits;# events",
		 1001 , -0.0005, 1.0005 );

	_trackPulsesNormHisto = new TH1I( "TrackPulsesNormHisto",
		 "Track Pulses;# pulses in track / # pulses;# events",
		 1001 , -0.0005, 1.0005 );

	_noisePulsesDistribution = new TH2I( "NoisePulsesSpatialDistribution",
		"Spatial Distribution of Noise Pulses;x [mm];y [mm];# events",
		(int)floor((planeExtent[1]-planeExtent[0])/6.), planeExtent[0], planeExtent[1], 
		(int)floor((planeExtent[3]-planeExtent[2])/2.5), planeExtent[2], planeExtent[3] ); 

	_notInHitPulsesDistribution = new TH2I( "NotInHitPulsesSpatialDistribution",
		"Spatial Distribution of Pulses not in Hit;x [mm];y [mm];# events",
		(int)floor((planeExtent[1]-planeExtent[0])/6.), planeExtent[0], planeExtent[1], 
		(int)floor((planeExtent[3]-planeExtent[2])/2.5), planeExtent[2], planeExtent[3] ); 
}


void NoiseHitAndPulseCounterProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
	// do something for each run
} 


void NoiseHitAndPulseCounterProcessor::processEvent(LCEvent* evt)
{ 
	// do something for each event, usually access the input collection:

	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputTrackColVec = 0;
	LCCollectionVec* inputHitColVec = 0;
	LCCollectionVec* inputPulseColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputTrackColVec = dynamic_cast<LCCollectionVec*>( evt->getCollection(_inputTrackColName) );
		inputHitColVec = dynamic_cast<LCCollectionVec*>( evt->getCollection(_inputHitColName) );
		inputPulseColVec = dynamic_cast<LCCollectionVec*>( evt->getCollection(_inputPulseColName) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		return;
	}

	streamlog_out(DEBUG4) << "event: " << evt->getEventNumber() << std::endl;
	streamlog_out(DEBUG4) << "number of tracks in event: " << inputTrackColVec->getNumberOfElements() << std::endl;
	streamlog_out(DEBUG4) << "number of hits in event: " << inputHitColVec->getNumberOfElements() << std::endl;
	streamlog_out(DEBUG4) << "number of pulses in event: " << inputPulseColVec->getNumberOfElements() << std::endl;

	// loop the track collection and count hits and pulses 
	const int totalHitsInEvent = inputHitColVec->getNumberOfElements();
	const int totalPulsesInEvent = inputPulseColVec->getNumberOfElements();
	int totalHitsInTracks = 0;
	int totalPulsesInTracks = 0;
	for( LCCollectionVec::iterator inputTrackIter = inputTrackColVec->begin() ; inputTrackIter < inputTrackColVec->end() ; ++inputTrackIter )
	{
		// dynamic cast the input object (LCObject) to what it really is e.g.
		lcio::Track* inputTrack = dynamic_cast<lcio::Track*>(*inputTrackIter);
		lcio::TrackerHitVec trackHitsVec = inputTrack->getTrackerHits();
		totalHitsInTracks += trackHitsVec.size();
		streamlog_out(DEBUG3) << "number of hits in track: " << trackHitsVec.size() << std::endl;

		//add up pulses in the hits of this track
		int totalPulsesInThisTrack = 0;
		for( lcio::TrackerHitVec::iterator hitIter = trackHitsVec.begin() ; hitIter<trackHitsVec.end() ; ++hitIter)
		{
			lcio::TrackerHit* trackHit = *hitIter;
			totalPulsesInThisTrack += (trackHit->getRawHits()).size();
		}
		totalPulsesInTracks += totalPulsesInThisTrack;
		streamlog_out(DEBUG3) << "number of pulses in hits in track: " << totalPulsesInThisTrack << std::endl;
	}
	_noiseHitsHisto->Fill( totalHitsInEvent-totalHitsInTracks );
	_noisePulsesHisto->Fill( totalPulsesInEvent-totalPulsesInTracks );
	_noiseHitsNormHisto->Fill( double(totalHitsInEvent-totalHitsInTracks) / double(totalHitsInEvent) );
	_noisePulsesNormHisto->Fill( double(totalPulsesInEvent-totalPulsesInTracks) / double(totalPulsesInEvent) );
	_trackHitsHisto->Fill(totalHitsInTracks);
	_trackPulsesHisto->Fill(totalPulsesInTracks);
	_trackHitsNormHisto->Fill( double(totalHitsInTracks) / double(totalHitsInEvent) );
	_trackPulsesNormHisto->Fill( double(totalPulsesInTracks) / double(totalPulsesInEvent) );
	streamlog_out(DEBUG4) << "number of noise hits: " << totalHitsInEvent-totalHitsInTracks << std::endl;
	streamlog_out(DEBUG4) << "number of noise pulses: " << totalPulsesInEvent-totalPulsesInTracks << std::endl;

	//count pulses not included in any hit
	int totalPulsesInHits = 0;
	for( LCCollectionVec::iterator inputHitIter = inputHitColVec->begin() ; inputHitIter < inputHitColVec->end() ; ++inputHitIter )
	{
		// dynamic cast the input object (LCObject) to what it really is e.g.
		lcio::TrackerHit* inputHit = dynamic_cast<lcio::TrackerHit*>(*inputHitIter);
		LCObjectVec hitPulsesVec = inputHit->getRawHits();
		totalPulsesInHits += hitPulsesVec.size();
	}
	_notInHitPulsesHisto->Fill( totalPulsesInEvent-totalPulsesInHits );
	_notInHitPulsesNormHisto->Fill( double(totalPulsesInEvent-totalPulsesInHits) / double(totalPulsesInEvent) );
	streamlog_out(DEBUG4) << "number of pulses not in hits: " << totalPulsesInEvent-totalPulsesInHits << std::endl;

	//get the location of the noise pulses
	int numberOfMappedNoisePulses = 0;
	int numberOfMappedNotInHitPulses = 0;
	for( LCCollectionVec::iterator inputPulseIter = inputPulseColVec->begin() ; inputPulseIter < inputPulseColVec->end() ; ++inputPulseIter )
	{
		const lcio::TrackerPulse* inputPulse = dynamic_cast<const lcio::TrackerPulse*>(*inputPulseIter);
		bool pulseIsInHit = false;
		bool pulseIsInTrack = false;

		//try and find the pulse in each hit of the track
		for( LCCollectionVec::iterator inputHitIter = inputHitColVec->begin() ; inputHitIter < inputHitColVec->end() ; ++inputHitIter )
		{
			lcio::TrackerHit* inputHit = dynamic_cast<lcio::TrackerHit*>(*inputHitIter);
			LCObjectVec allPulsesInThisHit = inputHit->getRawHits();
			LCObjectVec::iterator findPulseIter = find( allPulsesInThisHit.begin(), allPulsesInThisHit.end(), inputPulse );

			//if pulse is not found in this hit, try the next one
			if(findPulseIter != allPulsesInThisHit.end())
			{
				pulseIsInHit = true;
			}
			if(!pulseIsInHit) continue;

			//if the pulse is in a hit, try and find the hit in the track(s)
			for( LCCollectionVec::iterator inputTrackIter = inputTrackColVec->begin() ; inputTrackIter < inputTrackColVec->end() ; ++inputTrackIter )
			{
				lcio::Track* inputTrack = dynamic_cast<lcio::Track*>(*inputTrackIter);
				lcio::TrackerHitVec trackHitsVec = inputTrack->getTrackerHits();
				lcio::TrackerHitVec::iterator findHitIter = find( trackHitsVec.begin(), trackHitsVec.end(), inputHit );
				//stop searching, if the hit is found
				if(findHitIter != trackHitsVec.end())
				{
					pulseIsInTrack = true;
					break;
				}
			}
			//break since the pulse was already found in a hit
			break;
		}
		//skip filling the histograms, if pulse is included in a track
		if(pulseIsInTrack) continue;
	
		//get the position of the pulse and fill the noise histogramm
		const gear::TPCModule& myModule = _gearParameters->getModule( inputPulse->getCellID1() );
		gear::Vector2D padcenter = myModule.getPadCenter( inputPulse->getCellID0() );
		_noisePulsesDistribution->Fill(padcenter[0],padcenter[1]);
		++numberOfMappedNoisePulses;

		//skip filling the not-in-hit histogram, if the pulse is in a hit
		if(pulseIsInHit) continue;
		_notInHitPulsesDistribution->Fill(padcenter[0],padcenter[1]);
		++numberOfMappedNotInHitPulses; 
		
	}
	streamlog_out(DEBUG4) << "number of mapped noise pulses: " << numberOfMappedNoisePulses << std::endl;
	streamlog_out(DEBUG4) << "number of mapped pulses not in hits: " << numberOfMappedNotInHitPulses << std::endl;
	
}


void NoiseHitAndPulseCounterProcessor::check(LCEvent*  /*evt*/)
{ 
	// check what is needed here 
}


void NoiseHitAndPulseCounterProcessor::end()
{ 
	AIDAProcessor::tree(this);

	TNtuple* noiseHitsTuple = new TNtuple("NoiseHitsTuple","NoiseHitsStats","mean90:mean90err:rms:rmserr");
	TNtuple* noiseHitsNormTuple = new TNtuple("NoiseHitsNormTuple","NoiseHitsStats","mean90:mean90err:rms:rmserr");

	TNtuple* noisePulsesTuple = new TNtuple("NoisePulsesTuple","NoisePulsesStats","mean90:mean90err:rms:rmserr");
	TNtuple* noisePulsesNormTuple = new TNtuple("NoisePulsesNormTuple","NoisePulsesStats","mean90:mean90err:rms:rmserr");

	TNtuple* notInHitPulsesTuple = new TNtuple("NotInHitPulsesTuple","NotInHitPulsesStats","mean90:mean90err:rms:rmserr");
	TNtuple* notInHitPulsesNormTuple = new TNtuple("NotInHitPulsesNormTuple","NotInHitPulsesStats","mean90:mean90err:rms:rmserr");

	TNtuple* trackHitsTuple = new TNtuple("TrackHitsTuple","TrackHitsStats","mean90:mean90err:rms:rmserr");
	TNtuple* trackHitsNormTuple = new TNtuple("TrackHitsNormTuple","TrackHitsStats","mean90:mean90err:rms:rmserr");

	TNtuple* trackPulsesTuple = new TNtuple("TrackPulsesTuple","TrackPulsesStats","mean90:mean90err:rms:rmserr");
	TNtuple* trackPulsesNormTuple = new TNtuple("TrackPulsesNormTuple","TrackPulsesStats","mean90:mean90err:rms90:rms90err");

	//fill tuples with mean, RMS and the respective errors of the distributions
	//TODO: find values with more meaning to compare distributions

	double noiseHitsStats[4];
	double noiseHitsNormStats[4];
	double noisePulsesStats[4];
	double noisePulsesNormStats[4];
	double notInHitPulsesStats[4];
	double notInHitPulsesNormStats[4];
	double trackHitsStats[4];
	double trackHitsNormStats[4];
	double trackPulsesStats[4];
	double trackPulsesNormStats[4];

	_getMeanRMS90(_noiseHitsHisto, noiseHitsStats);
	_getMeanRMS90(_noiseHitsNormHisto, noiseHitsNormStats);
	_getMeanRMS90(_noisePulsesHisto, noisePulsesStats);
	_getMeanRMS90(_noisePulsesNormHisto, noisePulsesNormStats);
	_getMeanRMS90(_notInHitPulsesHisto, notInHitPulsesStats);
	_getMeanRMS90(_notInHitPulsesNormHisto, notInHitPulsesNormStats);
	_getMeanRMS90(_trackHitsHisto, trackHitsStats);
	_getMeanRMS90(_trackHitsNormHisto, trackHitsNormStats);
	_getMeanRMS90(_trackPulsesHisto, trackPulsesStats);
	_getMeanRMS90(_trackPulsesNormHisto, trackPulsesNormStats);

	noiseHitsTuple->Fill(noiseHitsStats[0],noiseHitsStats[1],noiseHitsStats[2],noiseHitsStats[3]);
	noiseHitsNormTuple->Fill(noiseHitsNormStats[0],noiseHitsNormStats[1],noiseHitsNormStats[2],noiseHitsNormStats[3]);

	noisePulsesTuple->Fill(noisePulsesStats[0],noisePulsesStats[1],noisePulsesStats[2],noisePulsesStats[3]);
	noisePulsesNormTuple->Fill(noisePulsesNormStats[0],noisePulsesNormStats[1],noisePulsesNormStats[2],noisePulsesNormStats[3]);

	notInHitPulsesTuple->Fill(notInHitPulsesStats[0],notInHitPulsesStats[1],notInHitPulsesStats[2],notInHitPulsesStats[3]);
	notInHitPulsesNormTuple->Fill(notInHitPulsesNormStats[0],notInHitPulsesNormStats[1],notInHitPulsesNormStats[2],notInHitPulsesNormStats[3]);

	trackHitsTuple->Fill(trackHitsStats[0],trackHitsStats[1],trackHitsStats[2],trackHitsStats[3]);
	trackHitsNormTuple->Fill(trackHitsNormStats[0],trackHitsNormStats[1],trackHitsNormStats[2],trackHitsNormStats[3]);

	trackPulsesTuple->Fill(trackPulsesStats[0],trackPulsesStats[1],trackPulsesStats[2],trackPulsesStats[3]);
	trackPulsesNormTuple->Fill(trackPulsesNormStats[0],trackPulsesNormStats[1],trackPulsesNormStats[2],trackPulsesNormStats[3]);

}

void NoiseHitAndPulseCounterProcessor::_getMeanRMS90(TH1* histo, double returnValues[])
{
	const int totalEntries = histo->GetEntries();
	int lowBin=1, upBin=histo->GetNbinsX();
	while(histo->Integral(lowBin, upBin)>0.9*totalEntries)
	{
		if( histo->GetBinContent(lowBin) >= histo->GetBinContent(upBin) ){ --upBin; }
		else ++lowBin;
	}
	histo->GetXaxis()->SetRange(lowBin, upBin);
	returnValues[0] = histo->GetMean();
	returnValues[1] = histo->GetMeanError();
	returnValues[2] = histo->GetRMS();
	returnValues[3] = histo->GetRMSError();
	histo->GetXaxis()->SetRange();
	return;
}

} // close namespace brackets
