#include "ThetaResolutionPerformanceProcessor.h"

#include "HistogramWidthExtractionTools.h"
#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"
#include "LCObjectCopier.h"

#include <iostream>
#include <string>
#include <algorithm>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>
#include "TrackFitterBase.h"
#include <EVENT/TrackerData.h>

#include <Exceptions.h>

//GEAR
#include "gear/TPCModule.h"
#include <gear/BField.h>
#include <gearimpl/Vector3D.h>

//ROOT
#include "TGraphErrors.h"
#include "TPad.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#endif

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{

ThetaResolutionPerformanceProcessor aThetaResolutionPerformanceProcessor;


ThetaResolutionPerformanceProcessor::ThetaResolutionPerformanceProcessor() : Processor("ThetaResolutionPerformanceProcessor")
{
	// the processor description
	_description = "The processor is used to determine the dependence of the resolution on the theta track angel. The resolution is determined via the geometric mean method. ";



	registerInputCollection(    LCIO::TRACK,
			"InputCollectionName" ,
			"Name of the input Track collection, default: TPCTracks"  ,
			_inputColName ,
			string("TPCTracks")) ;

	registerProcessorParameter( "xyResolutionBins" ,
			"Number of bins in the xy-residual histogram, default: 256",
			_xyResolutionBins,
			int(256));

	registerProcessorParameter( "xyResolutionRange",
			"Range of xy-residual histogram, default: 3.mm",
			_xyResolutionRange,
			float(3));

	registerProcessorParameter( "zResolutionBins" ,
			"Number of bins in the z-residual histogram, default: 256",
			_zResolutionBins,
			int(256));

	registerProcessorParameter( "zResolutionRange",
			"Range of z-residual histogram, default: 5.mm",
			_zResolutionRange,
			float(5));

	registerProcessorParameter( "meanCalculationMethodZ",
			"Defines the method of calculating the mean of the z-histograms. (default 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails)",
			_meanMethodZ,
			int(1));

	registerProcessorParameter( "meanCalculationMethodXY",
			"Defines the method of calculating the mean of the xy-histograms. (default 2:Gaussian fit within range of 3 sigma, 1:RMS90, 3:rms after 5 sigma iterative cut of tails)",
			_meanMethodXY,
			int(2));

	registerProcessorParameter( "ModuleWithBestRow",
			"Module that has the best row (default:0)",
			_bestModule,
			int(0));

	registerProcessorParameter( "RowWithBestPerformance",
			"Row with the best resolution performance (default:0)",
			_bestRow,
			int(0));

	registerProcessorParameter("BFieldScaleFactor",
			"Scales magnetic field (map), use 1.0 (default) for field ON or 0.0 for field OFF",
			_bfieldScaleFactor,
			double(1.0));
	
	registerProcessorParameter( "zRangeLow",
			"cut on the lowest z position for the hits, default 160mm",
			_zRangeLow,
			double(160));

	registerProcessorParameter("zRangeHigh",
			"cut on the highest z position for the hits, default 170mm",
			_zRangeHigh,
			double(170));	
	

	registerOptionalParameter("MaximumPulseCharge",
                              "Optional: Maximum charge of max pulse in hit to determine whether a hit is evaluated for resolution.",
                              _maxChargeCut,
                              float(1024.));

	registerOptionalParameter("MinimumPulseCharge",
                              "Optional: Minimum charge of max pulse in hit to determine whether a hit is evaluated for resolution.",
                              _minChargeCut,
                              float(0.));

}


void ThetaResolutionPerformanceProcessor::init()
{
	streamlog_out(DEBUG) << " init() called  " << endl;

	// you might want to list the used parameters by invoking
	printParameters();
	AIDAProcessor::tree(this);
	const gear::TPCParameters* tpcParameters = &marlin::Global::GEAR->getTPCParameters();
	_gearParameters= &marlin::Global::GEAR->getTPCParameters();
	std::vector<gear::TPCModule*> TPCModules = tpcParameters->getModules();

	std::vector<gear::TPCModule*>::const_iterator moduleIter = TPCModules.begin();
	int moduletype0 = (*moduleIter)->getPadLayoutType();

	for( ; moduleIter != TPCModules.end() ; ++moduleIter)
	{
		if((*moduleIter)->getPadLayoutType() != moduletype0)
			streamlog_out(MESSAGE) << "WARNING: Modules with different pad layouts are defined. Will use first one!"<< std::endl;
	}

	_myModuleDistortionCalculator=new ModuleDistortionCalculator( &marlin::Global::GEAR->getTPCParameters());

	_allRowsTuple = new TNtuple("allRowsResolution","allRowsResolution","module:row:theta:dist_x:dist_z:res_x:res_z");
	_bestRowTuple = new TNtuple("bestRowResolution","bestRowResolution","module:row:theta:thetaError:geoMean_xy:geoMean_xyError:geoMean_z:geoMean_zError");

	gear::Vector3D theTPCCenter(_gearParameters->getDoubleVal("TPC_cylinder_centre_c0"),
			_gearParameters->getDoubleVal("TPC_cylinder_centre_c1"),
			0.);

	gear::Vector3D field = marlin::Global::GEAR->getBField().at(theTPCCenter);

	// const bool bOff ( gear_field[2] == 0.);
	const bool bOff ( field.z()*_bfieldScaleFactor == 0. );
	if (bOff)
	{	streamlog_out(DEBUG) << "BField(GEAR) is OFF" << std::endl;}
	else
	{	streamlog_out(DEBUG) << "BField(GEAR) is ON" << std::endl;}
	// constant magnetic field (in z direction)
	// double Bz = gear_field[2];
	double Bz = field.z()*_bfieldScaleFactor;// rho component is accessible with field.rho()

	unsigned char fitterType= TrackFitterBase::FitterTypes::SIMPLEHELIXFIT;
	if(Bz==0)
	{
		fitterType=TrackFitterBase::FitterTypes::LINEARREGRESSION;
	}
	_trackFitter  = TrackFitterFactory::getTrackFitter(fitterType,0);

	_chargeCutSet=false;
	if(_parameters->isParameterSet("MaximumPulseCharge"))
	  _chargeCutSet=true;
	if(_parameters->isParameterSet("MinimumPulseCharge"))
	  _chargeCutSet=true;
}


void ThetaResolutionPerformanceProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
	// do something for each run
}


void ThetaResolutionPerformanceProcessor::processEvent(LCEvent* evt)
{

	AIDAProcessor::tree(this);
	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		return;
	}

	// loop the track collection and do something
	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
			inputObjectIter < inputColVec->end() ;
			inputObjectIter++ )
	{
		// dynamic cast the input track object
		Track * inputTrack = dynamic_cast<Track *>( *inputObjectIter );

		if ( inputTrack == 0 )
		{
			throw DataNotAvailableException("No input tracks found.");
		}

		TrackImpl * refittedTrack = _trackFitter->fitTrack(inputTrack);
		//TrackImpl * refittedTrack = LCObjectCopier::copy(inputTrack);

		TrackerHitVec hitVec = inputTrack->getTrackerHits();

		for ( TrackerHitVec::iterator itHit = hitVec.begin(); itHit < hitVec.end(); itHit++ )
		{

			TrackerHit* hit = *itHit;
			bool skipHit=false;
			if(_chargeCutSet)
			  {
			double maxcharge=0;
			//find pulse with maximum charge
			for ( LCObjectVec::const_iterator pulseIter = hit->getRawHits().begin() ;
			      pulseIter < hit->getRawHits().end() ; pulseIter++ )
			  {
			    TrackerPulse const * pulse = dynamic_cast<TrackerPulse const *>(*pulseIter);
			    EVENT::TrackerData*  trackerData = pulse->getTrackerData();
			    vector<float> theADCSpectrum = trackerData->getChargeValues();
			    if(*max_element(theADCSpectrum.begin(), theADCSpectrum.end()) >maxcharge)
			      {
				maxcharge=*max_element(theADCSpectrum.begin(), theADCSpectrum.end());
			      }
			  }
			    if( maxcharge<=_minChargeCut || maxcharge>=_maxChargeCut)
			      {
				skipHit=true;
			      }

			  }
		    if(hit->getPosition()[2]>_zRangeHigh || hit->getPosition()[2]<_zRangeLow)
		    {
		    	skipHit=true;
		    }
		    if(skipHit)
		    	continue;

			TrackImpl * trackWOHit = LCObjectCopier::copyTrackWithoutHits(refittedTrack);
			for (EVENT::TrackerHitVec::const_iterator hitIter = hitVec.begin();
					hitIter < hitVec.end(); hitIter++)
			{
				if(hit!=*hitIter)
					trackWOHit->addHit(*hitIter);
			}

			TrackImpl * refittedHitTrack = _trackFitter->fitTrack(trackWOHit);

			//find out module and row of the hit.
			const gear::TPCModule& theModule = _gearParameters->getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
			int row=theModule.getRowNumber(theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]));
			int module=theModule.getModuleID();
			double distValue [4]={10,10,10,10};
			//fill the histograms with the distance and residual values
			try{
				distValue[0] = _myModuleDistortionCalculator->calculateDistanceRow(hit,refittedTrack);
			}
			catch(std::exception &e)
			{
				std::cout<<"calculateDistanceRow for distances failed. Value ignored!"<<std::endl;
			}

			distValue[1] = _myModuleDistortionCalculator->calculateDistanceZ(hit,refittedTrack);
			
			//TODO: new track fit and change input track to fittrack
			try{
				distValue[2] = _myModuleDistortionCalculator->calculateDistanceRow(hit,refittedHitTrack);
			}
			catch(std::exception &e)
			{
				std::cout<<"calculateDistanceRow for residuals failed. Value ignored!"<<std::endl;
			}

			distValue[3] = _myModuleDistortionCalculator->calculateDistanceZ(hit,refittedHitTrack);
			_allRowsTuple->Fill(module,row,inputTrack->getTanLambda(),distValue[0],distValue[1],distValue[2],distValue[3]);
			delete trackWOHit;
			delete refittedHitTrack;
		}//hit loop

		delete refittedTrack;
	}// track loop
}


void ThetaResolutionPerformanceProcessor::check(LCEvent*  /*evt*/)
{
	// check what is needed here
}


void ThetaResolutionPerformanceProcessor::end()
{
	AIDAProcessor::tree(this);
	_zDistanceAll = new TH1F("zDistance","zDistance",_zResolutionBins,-_zResolutionRange,_zResolutionRange);
	_zResidualAll = new TH1F("zResidual","zResidual",_zResolutionBins,-_zResolutionRange,_zResolutionRange);
	_xyDistanceAll = new TH1F("xyDistance","xyDistance",_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);
	_xyResidualAll = new TH1F("xyResidual","xyResidual",_xyResolutionBins,-_xyResolutionRange,_xyResolutionRange);
	_theta = new TH1F("Theta","Theta",1028,-0.35,0.35);
	_allRowsTuple->Draw("module*100+row>>htemp(550,0,550)","row==10 && module==3");
//	_allRowsTuple->Draw("module*100+row>>htemp(550,0,550)","row!=18 && row!=14 &&row>6 && row<21");
	//_allRowsTuple->Draw("module*100+row>>htemp(550,0,550)","row!=13 && row!=14 &&row>4 && row<23");
	TH1F *htemp = (TH1F*)gPad->GetPrimitive("htemp");
	int binmax = htemp->GetMaximumBin();
	double x = htemp->GetXaxis()->GetBinCenter(binmax);
	int module= (int)(x/100);
	int row= ((int)x) % 100;
//		int module= 3;
//		int row= 11;
	_allRowsTuple->Draw("dist_x>>xyDistance",Form("module==%i && row==%i",module,row));
	_allRowsTuple->Draw("res_x>>xyResidual",Form("module==%i && row==%i",module,row));
	_allRowsTuple->Draw("dist_z>>zDistance",Form("module==%i && row==%i",module,row));
	_allRowsTuple->Draw("res_z>>zResidual",Form("module==%i && row==%i",module,row));
	_allRowsTuple->Draw("theta>>Theta",Form("module==%i && row==%i",module,row));

	//necessary if histograms are created here instead of in init(),
	//to get histograms in the correct folder
	
	//all histograms are filled
	//now we need to fit them
	HistogramWidthExtractionTools *myTool= new HistogramWidthExtractionTools();

	myTool->Calculate(_zDistanceAll,_meanMethodZ);
	double z_widthDist=myTool->GetWidth();
	double z_widthDist_err=myTool->GetWidthError();
	myTool->Calculate(_zResidualAll,_meanMethodZ);
	double z_widthRes=myTool->GetWidth();
	double z_widthRes_err=myTool->GetWidthError();

	double resZ = sqrt(z_widthDist*z_widthRes);
	double resZError = 0.5/resZ*sqrt(z_widthRes*z_widthRes*(z_widthDist_err*z_widthDist_err)+z_widthDist*z_widthDist*(z_widthRes_err*z_widthRes_err));

	myTool->Calculate(_xyDistanceAll,_meanMethodXY);
	double xy_widthDist=myTool->GetWidth();
	double xy_widthDist_err=myTool->GetWidthError();
	myTool->Calculate(_xyResidualAll,_meanMethodXY);
	double xy_widthRes=myTool->GetWidth();
	double xy_widthRes_err=myTool->GetWidthError();

	double resXY = sqrt(xy_widthDist*xy_widthRes);
	double resXYError = 0.5/resXY*sqrt(xy_widthRes*xy_widthRes*(xy_widthDist_err*xy_widthDist_err)+xy_widthDist*xy_widthDist*( xy_widthRes_err* xy_widthRes_err));
	myTool->Calculate(_theta,2);

	double theta = myTool->GetMean();
	double thetaError = myTool->GetWidth();
	
	_bestRowTuple->Fill(module, row, theta, thetaError, resXY,resXYError,resZ,resZError);


    delete myTool;

}
} // close namespace brackets
