#include "TimePixTrackEventDisplayProcessor.h"

#include <algorithm>

using namespace lcio;
using namespace marlin;

namespace marlintpc
{

TimePixTrackEventDisplayProcessor aTimePixTrackEventDisplayProcessor;


TimePixTrackEventDisplayProcessor::TimePixTrackEventDisplayProcessor() : Processor("TimePixTrackEventDisplayProcessor")
{
  // the processor description
  _description = "TimePixTrackEventDisplayProcessor draws all TrackerHits in an event as black dots (kFullDotMedium). "
                 "Also Tracks are draw on top of their corresponding Hits where the tracks and hits have the same color. "
                 "Tracks are only drawn in the xy plane.";

        registerInputCollection( LCIO::TRACK,
            "InputCollectionNameTracks" ,
            "Name of the input Track collection (default TPCTracks)"  ,
            _inputColNameTracks ,
            std::string("TPCTracks"));

        registerInputCollection( LCIO::TRACKERHIT,
             "TrackerDataColVec" ,
             "the name of an collection that contains all TrackerHits (default: TimePixCorrectedHits)"  ,
             _TrackerHitColName,
             std::string("TimePixCorrectedHits"));

        registerProcessorParameter("SetOutputTransient" ,
          "if not 0 the output collection is set transient (default: 0)"  ,
          _outputIsTransient,
          int(0));

        registerProcessorParameter("MaxNumberOfHistograms" ,
          "maximum number of created histograms (default: 500)"  ,
          _MaxNumberOfHistograms,
          (int)200 ) ;

        registerProcessorParameter("xMax" ,
          "maximum value of x range (default: 65.0 mm)"  ,
          _xMax,
          float(65.0)) ;

        registerProcessorParameter("xMin" ,
          "minimum value of x range (default: 0.0 mm)"  ,
          _xMin,
          float(0.0)) ;

        registerProcessorParameter("yMax" ,
          "maximum value of y range (default: 45.0 mm)"  ,
          _yMax,
          float(45.0)) ;

        registerProcessorParameter("yMin" ,
          "minimum value of y range (default: 0.0 mm)"  ,
          _yMin,
          float(0.0)) ;

        registerProcessorParameter("zMax" ,
          "maximum value of z range (default: 600.0 mm)"  ,
          _zMax,
          float(600.0)) ;

        registerProcessorParameter("zMin" ,
          "minimum value of z range (default: 0.0 mm)"  ,
          _zMin,
          float(-5.0)) ;

        registerProcessorParameter("xPixel" ,
          "number of pixels in x direction (default: 256)"  ,
          _xPixel,
          (int)1024) ;

        registerProcessorParameter("yPixel" ,
          "number of pixels in y direction (default: 256)"  ,
          _yPixel,
          (int)512) ;

        _CurNumberOfHistograms = 0;
}


void TimePixTrackEventDisplayProcessor::init()
{
    streamlog_out(DEBUG0) << " init() called  " << std::endl;

    //open file to save histograms
    _file = new TFile("TrackHistos.root", "recreate");
    if( !_file->IsOpen() )
    {
        streamlog_out(ERROR4) << "file could not be opened" << std::endl;
        _fileOpen = false;
    }
    else
    {
        _fileOpen = true;
    }

    // you might want to list the used parameters by invoking
    printParameters();
}


void TimePixTrackEventDisplayProcessor::processRunHeader(LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev:1$");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
    {
        if (! i->second->isOptional() || i->second->valueSet())
        {
            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
        }
    }
}

void TimePixTrackEventDisplayProcessor::processEvent(LCEvent* evt)
{
    streamlog_out(DEBUG4) << "starting processor" << std::endl;

    // give number of event
    if( (evt->getEventNumber() % 100) == 0 )
    {
        streamlog_out(MESSAGE4) << "EventNumber " << evt->getEventNumber() << std::endl;
    }

    if( _fileOpen == false )
    {
        streamlog_out(ERROR4) << "file could not be opened" << std::endl;
        return;
    }

    // only process, if still not all histograms filled
    if( _CurNumberOfHistograms == _MaxNumberOfHistograms )
    {
        return;
    }


    // get input collection

        // use an LCCollectionVec instead of LCCollection so we can use iterators later
        bool bDrawTracks( true );
        LCCollectionVec* inputColVecTracks = 0;
        try
        {
            inputColVecTracks = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColNameTracks ) );
        }
        catch (DataNotAvailableException &e)
        {
            streamlog_out(WARNING0) << "evt number: " << evt->getEventNumber() << ", e.what(): " << e.what() << std::endl;
            bDrawTracks = false;
        }


        // cast did not succeed
        if( bDrawTracks && inputColVecTracks == NULL )
        {
            streamlog_out(WARNING0) << "InputCollection vector is 0" << std::endl;
            bDrawTracks = false;
        }


        // no in collection
        if( bDrawTracks && inputColVecTracks->empty() )
        {
            streamlog_out(MESSAGE4) << "InputCollection vector is empty" << std::endl;
            bDrawTracks = false;
        }


    // get collection of all TrackerHits

        // use an LCCollectionVec instead of LCCollection so we can use iterators later
        bool bDrawHits( true );
        LCCollectionVec* inputColVecHits = 0;
        try
        {
          inputColVecHits = dynamic_cast<LCCollectionVec *>(evt->getCollection( _TrackerHitColName ) );
        }
        catch (DataNotAvailableException &e)
        {
          streamlog_out(WARNING0) << "evt number: " << evt->getEventNumber() << ", e.what(): " << e.what() << std::endl;
          bDrawHits = false;
        }

        // cast did not succeed
        if( bDrawHits && inputColVecHits == NULL )
        {
          streamlog_out(WARNING0) << "InputCollection vector is 0" << std::endl;
          bDrawHits = false;
        }

        // no in collection
        if( bDrawHits && inputColVecHits->empty() )
        {
          streamlog_out(MESSAGE4) << "InputCollection vector is empty" << std::endl;
          bDrawHits = false;
        }

    if( not bDrawHits && not bDrawTracks ) { return; }


    // push_back TH2D
    streamlog_out(DEBUG0) << "push_back histos" << std::endl;
    std::ostringstream osFullXY;
    osFullXY << "XY_full_Evt_" << evt->getEventNumber();
    std::ostringstream osName;
    osName << "Drawing";
    if( bDrawTracks ) osName << "_" << inputColVecTracks->size() << "_Tracks";
    else osName << "_no_Tracks";
    if( bDrawHits   ) osName << "_" << inputColVecHits->size()   << "_Hits";
    else osName << "_no_Hits";
    _FullXYHist = new TH2D( osFullXY.str().c_str(), osName.str().c_str(), _xPixel, _xMin, _xMax, _yPixel, _yMin, _yMax );
    _FullXYHist->GetXaxis()->SetTitle("x [mm]");
    _FullXYHist->GetYaxis()->SetTitle("y [mm]");
    _FullXYHist->SetStats(0);
    std::ostringstream osZX;
    osZX << "ZX_full_Evt_" << evt->getEventNumber();
    _FullZXHist = new TH2D( osZX.str().c_str(), osName.str().c_str(), _xPixel, _zMin, _zMax, _xPixel, _xMin, _xMax );
    _FullZXHist->GetXaxis()->SetTitle("z [mm]");
    _FullZXHist->GetYaxis()->SetTitle("x [mm]");
    _FullZXHist->SetStats(0);
    std::ostringstream osZY;
    osZY << "ZY_full_Evt_" << evt->getEventNumber();
    _FullZYHist = new TH2D( osZY.str().c_str(), osName.str().c_str(), _xPixel, _zMin, _zMax, _yPixel, _yMin, _yMax );
    _FullZYHist->GetXaxis()->SetTitle("z [mm]");
    _FullZYHist->GetYaxis()->SetTitle("y [mm]");
    _FullZYHist->SetStats(0);

    streamlog_out(DEBUG0) << "loop through collection of all hits" << std::endl;

    // fill FullHistogram with all hits
    if( bDrawHits )
    for( LCCollectionVec::iterator itFullColVec = inputColVecHits->begin()
            ; itFullColVec != inputColVecHits->end()
            ; itFullColVec++
       )
    {
        streamlog_out(DEBUG0) << "dynamic cast" << std::endl;

        // get Track and TrackerHitVec from inputCol
        TrackerHit * myTrackerHit = dynamic_cast< TrackerHit *>( *itFullColVec );

        // cast did not succeed
        if( myTrackerHit == NULL )
        {
            streamlog_out(WARNING0) << "InputCollection vector is 0" << std::endl;
            return;
        }
        streamlog_out(DEBUG0) << "get position" << std::endl;
        const double * pos = myTrackerHit->getPosition();
        streamlog_out(DEBUG0) << "fill histogram" << std::endl;

        Color_t color = kBlack;
        Marker_t marker = kFullDotMedium;
        float size = 0.3;
        // fill histogram with hits
        TMarker * markerXY = new TMarker( pos[0], pos[1], marker );
        markerXY->SetMarkerColor( color );
        markerXY->SetMarkerSize( size );
        _FullXYHist->Fill( pos[0], pos[1] );
        _FullXYHist->GetListOfFunctions()->Add( markerXY );

        TMarker * markerZX = new TMarker( pos[2], pos[0], marker );
        markerZX->SetMarkerColor( color );
        markerZX->SetMarkerSize( size );
        _FullZXHist->Fill( pos[2], pos[0] );
        _FullZXHist->GetListOfFunctions()->Add( markerZX );

        TMarker * markerZY = new TMarker( pos[2], pos[1], marker );
        markerZY->SetMarkerColor( color );
        markerZY->SetMarkerSize( size );
        _FullZYHist->Fill( pos[2], pos[1] );
        _FullZYHist->GetListOfFunctions()->Add( markerZY );

    }

    streamlog_out(DEBUG0) << "loop through collection of tracks" << std::endl;

    // loop the InputCollection vector
    if( bDrawTracks )
    {
        Color_t color = kRed;
        Marker_t marker = kFullDotMedium;
        float size = 0.3;
        for( LCCollectionVec::iterator itInputColVec = inputColVecTracks->begin()
                ; itInputColVec != inputColVecTracks->end()
                ; itInputColVec++ )
        {

            // get Track and TrackerHitVec from inputCol
            Track * myTrack = dynamic_cast< Track *>( *itInputColVec );
            TrackerHitVec myHitVec = myTrack->getTrackerHits();

            // vectors to store x-, y- and z-coordinates of all hits
            std::vector< double > xCoord, yCoord, zCoord;
            xCoord.reserve( myHitVec.size() ); yCoord.reserve( myHitVec.size() ); zCoord.reserve( myHitVec.size() );

            // minimum and maximum x values for this track
            // minimum predefined as _xMax since this value should be decreased (maximum as _xMin analogous)
            double xMinTrack( _xMax ), xMaxTrack( _xMin );
            // same for z values
            double zMinTrack( _zMax ), zMaxTrack( _zMin );

            // write pixels on track to histograms
            for (TrackerHitVec::iterator hit = myHitVec.begin()
                     ; hit !=  myHitVec.end()
                     ; hit ++)
             {
                 const double *pos = (*hit)->getPosition();
                 // draw hit
                 xCoord.push_back( pos[0] ); yCoord.push_back( pos[1] ); zCoord.push_back( pos[2] );

                 // fill histograms with hits on teh track
                 TMarker * markerXY = new TMarker( pos[0], pos[1], marker );
                 markerXY->SetMarkerColor( color );
                 markerXY->SetMarkerSize( size );
                 _FullXYHist->GetListOfFunctions()->Add( markerXY );

                 TMarker * markerZX = new TMarker( pos[2], pos[0], marker );
                 markerZX->SetMarkerColor( color );
                 markerZX->SetMarkerSize( size );
                 _FullZXHist->GetListOfFunctions()->Add( markerZX );

                 TMarker * markerZY = new TMarker( pos[2], pos[1], marker );
                 markerZY->SetMarkerColor( color );
                 markerZY->SetMarkerSize( size );
                 _FullZYHist->GetListOfFunctions()->Add( markerZY );

                 // update minimum and maximum x values
                 if( pos[0] < xMinTrack )
                 {
                     xMinTrack = pos[0];
                 }
                 else if( pos[0] > xMaxTrack )
                 {
                     xMaxTrack = pos[0];
                 }

                 // update minimum and maximum z values
                 if( pos[2] < zMinTrack )
                 {
                     zMinTrack = pos[2];
                 }
                 else if( pos[2] > zMaxTrack )
                 {
                     zMaxTrack = pos[2];
                 }
             }
            // -------- XY plane --------
                TGraph * grHelixXY = DrawHelixXY( myTrack );
                grHelixXY->SetLineColor( color );
                _FullXYHist->GetListOfFunctions()->Add( grHelixXY );
            // change color
                color++;
        }// for( LCCollectionVec::iterator itInputColVec = inputColVec->begin()
    }

    _FullXYHist->Write();
    _FullXYHist->Delete();

    _FullZXHist->Write();
    _FullZXHist->Delete();

    _FullZYHist->Write();
    _FullZYHist->Delete();

    // increment counter for histograms
    streamlog_out(DEBUG0) << "Created Histogram no " << _CurNumberOfHistograms << std::endl;
    _CurNumberOfHistograms += 1;

    return;
}

void TimePixTrackEventDisplayProcessor::check(LCEvent*  /*evt*/)
{
    // check what is needed here
    return;
}


void TimePixTrackEventDisplayProcessor::end()
{
    streamlog_out(DEBUG0) << "end()" << std::endl;

    // closing file
    _file->Close();

    return;
}



TGraph * TimePixTrackEventDisplayProcessor::DrawStraightLineXY( const Track * const myTrack )
{
    TrackerHitVec myHitVec = myTrack->getTrackerHits();

    // minimum and maximum x values for this track
    double xMinTrack = 0., xMaxTrack = 0.;
    bool bFirstIteration = true;

    // write pixels on track to histograms
    for (TrackerHitVec::iterator hit = myHitVec.begin()
            ; hit !=  myHitVec.end()
            ; hit ++)
    {
        const double * pos = (*hit)->getPosition();

        if( bFirstIteration )
        {
            bFirstIteration = false;
            xMinTrack = pos[0];
            xMaxTrack = pos[0];
        }
        // update minimum and maximum x values
        else if( pos[0] < xMinTrack )
        {
            xMinTrack = pos[0];
        }
        else if( pos[0] > xMaxTrack )
        {
            xMaxTrack = pos[0];
        }
    }

    // get track parameters
    double slope = TMath::Tan( myTrack->getPhi() );
    double ordin = myTrack->getD0() / TMath::Cos( myTrack->getPhi() );

    TVectorF x, y;
    x.ResizeTo( 2 ); y.ResizeTo( 2 );

    x[0] = xMinTrack; y[0] = ordin + x[0] * slope;
    x[1] = xMaxTrack; y[1] = ordin + x[1] * slope;

    streamlog_out(DEBUG8) << "(x,y) = " << "( " << x[0] << " , " << y[0] << " )" << std::endl;
    streamlog_out(DEBUG8) << "(x,y) = " << "( " << x[1] << " , " << y[1] << " )" << std::endl;

    return new TGraph( x, y );
}




// fixme right now drawing full circle
TGraph * TimePixTrackEventDisplayProcessor::DrawHelixXY( const Track * const myTrack )
{
    // Omega is too small => displayed Track would not represent data
    if( TMath::Abs( myTrack->getOmega() ) < 1.e-9 )
    {
        // draw straight line instead
        return DrawStraightLineXY( myTrack );
    }


    TrackerHitVec myHitVec = myTrack->getTrackerHits();

    // reference point
    const float * RefPoint = myTrack->getReferencePoint();
    // Omega ( Radius = 1/Omega )
    const float Omega = myTrack->getOmega();
    // D0
    const float D0 = myTrack->getD0();
    // Phi
    const float Phi = myTrack->getPhi();

    // get center of circle
    const float Px = RefPoint[0] + ( 1./Omega - D0 ) * TMath::Sin( Phi );
    const float Py = RefPoint[1] - ( 1./Omega - D0 ) * TMath::Cos( Phi );


    // number of points for TGraph; 1 point every pi mm, but at least 500 points
    int N = ( 1./( TMath::Abs( Omega )) > 500. ) ? static_cast<int>(1./( TMath::Abs( Omega ))) : 500;

    streamlog_out(DEBUG0) << "Using " << N << " points to approximate helix." << std::endl;
    TVectorF x, y;
    x.ResizeTo( N ); y.ResizeTo( N );

    // definition of constant: TwoPi = 2*Pi
    const float TwoPi = 2. * 4. * TMath::ATan( 1. );

    for( int i = 0; i < N; i++ )
    {
        x[i] = Px + TMath::Cos( i * TwoPi / (N-1) ) / Omega;
        y[i] = Py + TMath::Sin( i * TwoPi / (N-1) ) / Omega;
        streamlog_out(DEBUG8) << "(x,y) = " << "( " << x[i] << " , " << y[i] << " )" << std::endl;
    }

    return new TGraph( x, y );
}


// fixme not correct formula
TGraph * TimePixTrackEventDisplayProcessor::DrawStraightLineZX( const Track * const myTrack )
{
    // get track parameters
    double slope = 1. / ( ( myTrack->getTanLambda() ) * TMath::Sqrt( 1. + TMath::Power(TMath::Tan( myTrack->getPhi()),2) ) );
    double ordin = -( myTrack->getZ0() ) * slope - ( (myTrack->getD0() ) * TMath::Sin( myTrack->getPhi() ) );

    // minimum and maximum x values for this track
    double xMinTrack = 0., xMaxTrack = 0.;
    bool bFirstIteration = true;

    // write pixels on track to histograms
    TrackerHitVec myHitVec = myTrack->getTrackerHits();
    for (TrackerHitVec::iterator hit = myHitVec.begin()
            ; hit !=  myHitVec.end()
            ; hit ++)
    {
        const double * pos = (*hit)->getPosition();

        if( bFirstIteration )
        {
            bFirstIteration = false;
            xMinTrack = pos[0];
            xMaxTrack = pos[0];
        }
        // update minimum and maximum x values
        else if( pos[0] < xMinTrack )
        {
            xMinTrack = pos[0];
        }
        else if( pos[0] > xMaxTrack )
        {
            xMaxTrack = pos[0];
        }
    }

    // initalize coordinates
    TVectorF x, y;
    x.ResizeTo( 2 ); y.ResizeTo( 2 );

    x[0] = xMinTrack; y[0] = ordin + slope * x[0];
    x[1] = xMaxTrack; y[1] = ordin + slope * x[1];

    streamlog_out(DEBUG8) << "(x,y) = " << "( " << x[0] << " , " << y[0] << " )" << std::endl;
    streamlog_out(DEBUG8) << "(x,y) = " << "( " << x[1] << " , " << y[1] << " )" << std::endl;

    return new TGraph( x, y );
}


// fixme not correct formula
TGraph * TimePixTrackEventDisplayProcessor::DrawHelixZX( const Track * const myTrack )
{
    return DrawStraightLineZX( myTrack );
    return new TGraph();
}



} // close namespace brackets
