#include "TripletResolutionPerformanceProcessor.h"

#include "HistogramWidthExtractionTools.h"
#include "LCObjectCopier.h"
#include "TrackRecoTools.h"
#include "TripletCalculationTool.h"

#include <iostream>
#include <string>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>

#include <Exceptions.h>

//GEAR
#include "gear/TPCModule.h"

//ROOT
#include "TGraphErrors.h"

// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <marlin/AIDAProcessor.h>
#endif

using namespace lcio;
using namespace marlin;
using namespace std;



namespace marlintpc
{

TripletResolutionPerformanceProcessor aTripletResolutionPerformanceProcessor;


TripletResolutionPerformanceProcessor::TripletResolutionPerformanceProcessor() : Processor("TripletResolutionPerformanceProcessor")
{
	// the processor description
	_description = "Determines the resolution via the triplet method.";


	registerInputCollection(    LCIO::TRACK,
			"InputCollectionName" ,
			"Name of the input Track collection, default: TPCTracks"  ,
			_inputColName ,
			string("TPCTracks")) ;

	registerProcessorParameter( "xyTripletsBins" ,
			"Number of bins in the xy-triplets histogram, default: 254",
			_xyTripletsBins,
			int(254));

	registerProcessorParameter( "xyTripletsRange",
			"Range of xy-triplets histogram, default: 3.",
			_xyTripletsRange,
			float(3));

	registerProcessorParameter( "zTripletsBins" ,
			"Number of bins in the z-triplets histogram, default: 254",
			_zTripletsBins,
			int(254));

	registerProcessorParameter( "zTripletsRange",
			"Range of z-triplets histogram, default: 5.",
			_zTripletsRange,
			float(5));

	registerProcessorParameter( "meanCalculationMethodZ",
			"Defines the method of calculating the mean of the z-histograms. (default 1:RMS90, 2:Gaussian fit within range of 3 sigma, 3:rms after 5 sigma iterative cut of tails)",
			_meanMethodZ,
			int(1));

	registerProcessorParameter( "meanCalculationMethodXY",
			"Defines the method of calculating the mean of the xy-histograms. (default 2:Gaussian fit within range of 3 sigma, 1:RMS90, 3:rms after 5 sigma iterative cut of tails)",
			_meanMethodXY,
			int(2));

	registerProcessorParameter( "ModuleWithBestRow",
			"Module that has the best row (default:0)",
			_bestModule,
			int(0));

	registerProcessorParameter( "RowWithBestPerformance",
			"Row with the best resolution performance (default:0)",
			_bestRow,
			int(0));


}


void TripletResolutionPerformanceProcessor::init()
{ 
	streamlog_out(DEBUG) << " init() called  " << endl;

	// you might want to list the used parameters by invoking
	printParameters();
	AIDAProcessor::tree(this);
	const gear::TPCParameters* tpcParameters = &marlin::Global::GEAR->getTPCParameters();
	_gearParameters= &marlin::Global::GEAR->getTPCParameters();
	std::vector<gear::TPCModule*> TPCModules = tpcParameters->getModules();

	std::vector<gear::TPCModule*>::const_iterator moduleIter = TPCModules.begin();
	int moduletype0 = (*moduleIter)->getPadLayoutType();

	for( ; moduleIter != TPCModules.end() ; ++moduleIter)
	{
		if((*moduleIter)->getPadLayoutType() != moduletype0)
			streamlog_out(MESSAGE) << "WARNING: Modules with different pad layouts are defined. Will use first one!"<< std::endl;
	}


	_zTripletsAll = new TH1F("zTriplets","zTriplets",_zTripletsBins,-_zTripletsRange,_zTripletsRange);
	_xyTripletsAll = new TH1F("xyTriplets","xyTriplets",_xyTripletsBins,-_xyTripletsRange,_xyTripletsRange);

	_allRowsTuple = new TNtuple("allRowsTripletResolution","allRowsTripletResolution","geoMean_xy:geoMean_xyError:geoMean_z:geoMean_zError");
	_singleRowsTuple = new TNtuple("singleRowsTripletResolution","singleRowsTripletResolution","module:row:geoMean_xy:geoMean_xyError:geoMean_z:geoMean_zError");;
	_bestRowTuple = new TNtuple("bestRowTripletResolution","bestRowTripletResolution","module:row:geoMean_xy:geoMean_xyError:geoMean_z:geoMean_zError");;

}


void TripletResolutionPerformanceProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
	// do something for each run
} 


void TripletResolutionPerformanceProcessor::processEvent(LCEvent* evt)
{ 
	AIDAProcessor::tree(this);
	// use an LCCollectionVec instead of LCCollection so we can use iterators later
	LCCollectionVec* inputColVec = 0;

	// do a try / catch in case there is an empty event
	try
	{
		inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
	}
	catch (DataNotAvailableException &e)
	{
		// no input data, just return and process the next event
		return;
	}


	for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
			inputObjectIter < inputColVec->end() ;
			inputObjectIter++ )
	{
		// dynamic cast the input track object
		Track * inputTrack = dynamic_cast<Track *>( *inputObjectIter );

		if ( inputTrack == 0 )
		{
			throw DataNotAvailableException("No input tracks found.");
		}

		TrackerHitVec hitVec = inputTrack->getTrackerHits();
		std::vector< std::pair<TrackerHit*, double> > HitPairVec;
		for ( TrackerHitVec::iterator itHit = hitVec.begin(); itHit < hitVec.end(); itHit++ )
		{

			TrackerHit* hit = *itHit;
			double s = tracktools::CalculateS(inputTrack, hit);
			std::pair<TrackerHit*, double> HitPair( hit , s);
			HitPairVec.push_back( HitPair );

		}

		streamlog_out(DEBUG0)<<"SORT"<<std::endl;

		sort( HitPairVec.begin(), HitPairVec.end(), tracktools::compareHit );

		streamlog_out(DEBUG0)<<"sorted!"<<std::endl;

		for ( std::vector< std::pair<TrackerHit*, double> >::iterator iterHit = HitPairVec.begin(); iterHit < HitPairVec.end(); iterHit++ ) {

			//std::pair<double,double> tripletDist=std::make_pair(0,0);
			if ( iterHit < HitPairVec.end() - 1 && iterHit > HitPairVec.begin()) {

				TripletCalculationTool tripletTool(_gearParameters,iterHit->first,(iterHit-1)->first,(iterHit+1)->first);


				const gear::TPCModule& theModule = _gearParameters->getNearestModule(iterHit->first->getPosition()[0],iterHit->first->getPosition()[1]);
				int row=theModule.getRowNumber(theModule.getNearestPad(iterHit->first->getPosition()[0],iterHit->first->getPosition()[1]));
				int module=theModule.getModuleID();
				std::pair<int, int> key(module,row);
				std::map< std::pair<int,int>, TH1*>::iterator histoiter;
				histoiter = _xyTripletsRow.find(key);
				if(histoiter == _xyTripletsRow.end())
				{
					TH1* tempHist = new TH1F(Form("xyTriplets_%i_%i",module,row),"xyTriplets",_xyTripletsBins,-_xyTripletsRange,_xyTripletsRange);
					_xyTripletsRow.insert( std::make_pair(key,tempHist));

					TH1* tempHist2 = new TH1F(Form("zTriplets_%i_%i",module,row),"zTriplets",_zTripletsBins,-_zTripletsRange,_zTripletsRange);
					_zTripletsRow.insert( std::make_pair(key,tempHist2));

				}

				((*histoiter).second)->Fill(tripletTool.getXYTriplet());
				_xyTripletsAll->Fill(tripletTool.getXYTriplet());

				histoiter = _zTripletsRow.find(key);
				((*histoiter).second)->Fill(tripletTool.getZTriplet());
				_zTripletsAll->Fill(tripletTool.getZTriplet());
			}

		}


	}// track loop

}


void TripletResolutionPerformanceProcessor::check(LCEvent*  /*evt*/)
{ 
	// check what is needed here
}


void TripletResolutionPerformanceProcessor::end()
{ 
	//necessary if histograms are created here instead of in init(),
	//to get histograms in the correct folder
	AIDAProcessor::tree(this);
	//all histograms are filled
	//now we need to fit them
	HistogramWidthExtractionTools *myTool= new HistogramWidthExtractionTools();
	myTool->Calculate(_xyTripletsAll,_meanMethodXY);
	double xy_width=myTool->GetWidth();
	double xy_width_err=myTool->GetWidthError();
	myTool->Calculate(_zTripletsAll,_meanMethodZ);
	double z_width=myTool->GetWidth();
	double z_width_err=myTool->GetWidthError();
	
	_allRowsTuple->Fill(xy_width,xy_width_err,z_width,z_width_err);


	std::vector<double> radius,radiusError,sigmaXY,sigmaXYError,sigmaZ,sigmaZError;

	std::map< std::pair<int,int>, TH1*>::iterator histoiter = _xyTripletsRow.begin();
	for(; histoiter!=_xyTripletsRow.end(); histoiter++)
	{
		std::pair<int,int> key = (*histoiter).first;
		myTool->Calculate(((*histoiter).second),_meanMethodXY);
		double xy_width=myTool->GetWidth();
		double xy_width_err=myTool->GetWidthError();
		myTool->Calculate(_zTripletsRow[key],_meanMethodZ);
		double z_width=myTool->GetWidth();
		double z_width_err=myTool->GetWidthError();

		_singleRowsTuple->Fill(key.first, key.second, xy_width,xy_width_err, z_width,z_width_err);
		if (key.first==_bestModule && key.second==_bestRow)
			_bestRowTuple->Fill(key.first, key.second, xy_width,xy_width_err, z_width,z_width_err);

		sigmaXY.push_back(xy_width);
		sigmaXYError.push_back(xy_width_err);
		sigmaZ.push_back(z_width);
		sigmaZError.push_back(z_width_err);
	
		//get from to corresponding radius at phi=0
		const gear::TPCModule& theModule = _gearParameters->getModule(key.first);
		gear::Vector2D padcenter=theModule.getPadCenter(theModule.getPadIndex(key.second,0));
		radius.push_back(sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]));
		radiusError.push_back((theModule.getRowHeight(key.second))/sqrt(12.));
	}

	TGraphErrors *graphRow=new TGraphErrors(radius.size(),&radius[0],&sigmaXY[0],&radiusError[0],&sigmaXYError[0]);
	graphRow->SetNameTitle("XYTripletResolutionRow","XYTripletResolutionRow");
	graphRow->GetXaxis()->SetTitle("r_{row} [mm]");
	graphRow->GetYaxis()->SetTitle("#sigma_{row} [mm]");
	graphRow->Write();

	TGraphErrors *graphZ=new TGraphErrors(radius.size(),&radius[0],&sigmaZ[0],&radiusError[0],&sigmaZError[0]);
	graphZ->SetNameTitle("ZTripletResolutionRow","ZTripletResolutionRow");
	graphZ->GetXaxis()->SetTitle("r_{row} [mm]");
	graphZ->GetYaxis()->SetTitle("#sigma_{z} [mm]");
	graphZ->Write();
	delete graphZ;
	delete graphRow;
	delete myTool;

}
} // close namespace brackets
