#include "XrayGeometryAnalysisProcessor.h"

//c++
#include <cmath>

//lcio
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerHit.h>

//marlin
#include <marlin/Global.h>

//aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

namespace marlintpc{
  
  XrayGeometryAnalysisProcessor aXrayGeometryAnalysisProcessor;
  
  XrayGeometryAnalysisProcessor::XrayGeometryAnalysisProcessor() : Processor("XrayGeometryAnalysisProcessor"){
    
    _description = "Analysis of xray geometry";
    
    registerInputCollection(LCIO::LCRELATION,"InputCollection","InputCollection containing XrayObjects",_inputCol,std::string("TimepixXrayObjects"));
    
    //registerInputCollection(LCIO::LCGENERICOBJECT,"InputXrayCollection","InputCollection containing GridPixXrayObjects",_inputXrays,std::string("GridPixXrayObjects"));
    //registerInputCollection(LCIO::TRACKERHIT,"InputHitsCollection","InputCollection containing TrackerHits",_inputHits,std::string("XrayHits"));
    
  }
  
  XrayGeometryAnalysisProcessor::~XrayGeometryAnalysisProcessor(){}
  
  void XrayGeometryAnalysisProcessor::init(){
    
    printParameters();
    
    _geoService = new XrayGeometryService();
    
    _histoVec.push_back( new TH1D("excentricity","Excentricity",150,1.,10.) );//0
    _histoVec.push_back( new TH1D("rmsx","RmsX",150,0.,5.) );//1
    _histoVec.push_back( new TH1D("rmsy","RmsY",150,0.,5.) );//2
    _histoVec.push_back( new TH1D("skewnessl","SkewnessL",100,-5.,5.) );//3
    _histoVec.push_back( new TH1D("skewnesst","SkewnessT",100,-5.,5.) );//4
    _histoVec.push_back( new TH1D("kurtosisl","KurtosisL",100,-5.,5.) );//5
    _histoVec.push_back( new TH1D("kurtosist","KurtosisT",100,-5.,5.) );//6
    _histoVec.push_back( new TH1D("length","Length",200,0.,20.) );//7
    _histoVec.push_back( new TH1D("width","Width",100,0.,10.) );//8
    _histoVec.push_back( new TH1D("radius","Radius",100,0.,5.) );//9
    _histoVec.push_back( new TH1D("rotAngle","Rotation angle",100,0.,4.*std::atan(1.)));//10
    _histoVec.push_back( new TH1D("balance","Balance",500,0,0.4));//11
    _histoVec.push_back( new TH1D("fractionwithin0.5radius","Fraction of pixels within 0.5 radius",100,0.,1.));//12
    _histoVec.push_back( new TH1D("fractionwithinrmsy","Fraction of pixels within rmsY",100,0.,1.));//13
    _histoVec.push_back( new TH1D("length/radius","Length divided by radius",100,1.,2.));//14
    _histoVec.push_back( new TH1D("radius/rmsy","Radius divided by rmsY",100,0.,10.));//15
    _histoVec.push_back( new TH1D("lengthdivbyrmsy","Length divided by rmsy",150,0.,30.));//16
    
    _scatterVec.push_back( new TH2D("rmsxrmsy","RmsX vs RmsY",100,0.,2.,100,0.,2.) );//0
    _scatterVec.push_back( new TH2D("chargeexcentricity","Charge vs Excentricity",200,0.,4.e6,100,1.,2.5) );//1
    _scatterVec.push_back( new TH2D("rotangleexcentricity","RotAngle vs Excentricity",100,0.,4*std::atan(1.),100,1.,2.5) );//2
    _scatterVec.push_back( new TH2D("rotanglermsy","RotAngle vs RmsY",100,0.,4*std::atan(1.),100,0.,2.) );//3
    _scatterVec.push_back( new TH2D("rmsyvsbalance","RmsY vs Balance",100,0,1.,500,0,0.4));//4
    _scatterVec.push_back( new TH2D("energyvspixels","Energy vs Pixels",100,0,10,250,0,500));//5
    _scatterVec.push_back( new TH2D("distancefromchipcentervsskewnesst","Distance from Chip center vs SkewnessT",100,0.,8.,100,-5.,5.) );//6
    _scatterVec.push_back( new TH2D("distancefromchipcentervskurtosist","Distance from Chip center vs KurtosisT",100,0.,8.,100,-5.,5.) );//7
    _scatterVec.push_back( new TH2D("pixelsvsrmsy","Pixels vs RmsY",250,0,500,100,0.,1.));//8
    _scatterVec.push_back( new TH2D("chargesvsrmsy","Charge vs RmsY",200,0.,4.e6,100,0.,1.));//9

  }
  
  void XrayGeometryAnalysisProcessor::processEvent( lcio::LCEvent* evt ){
    
    //lcio::LCCollectionVec* inputXrays(NULL);
    //lcio::LCCollectionVec* inputHits(NULL);
    
    lcio::LCCollectionVec* inputCol(NULL);
    
    try{
      inputCol = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCol));
      //inputHits = dynamic_cast<LCCollectionVec*>(evt->getCollection(_inputHits));
    }
    catch(DataNotAvailableException &e){
      streamlog_out(MESSAGE3) << "Input collection is unavailable" << std::endl;
      return;
    }
    
    if(inputCol == NULL){
      
      streamlog_out(ERROR2) << "Input collection vector is 0" << std::endl;
      return;
      
    }
    
    for(lcio::LCCollectionVec::iterator iterCol = inputCol->begin(); iterCol != inputCol->end(); iterCol++)
    {
    
      lcio::LCRelation* xrayObject = dynamic_cast<lcio::LCRelation*>(*iterCol);
      
      lcio::TrackerHit* hit = dynamic_cast<lcio::TrackerHit*>(xrayObject->getTo());
      GridPixXrayObject* xray = new GridPixXrayObject(dynamic_cast<lcio::LCGenericObject*>(xrayObject->getFrom()));
      
      if(xray->getRotAngle() == -1){
	delete xray;
	continue;
      }
      
      _geoService->calculateGeometry(hit, xray->getRotAngle());
      
      _histoVec[0]->Fill(_geoService->getExcentricity());
      _histoVec[1]->Fill(_geoService->getRmsX());
      _histoVec[2]->Fill(_geoService->getRmsY());
      _histoVec[3]->Fill(_geoService->getSkewnessL());
      _histoVec[4]->Fill(_geoService->getSkewnessT());
      _histoVec[5]->Fill(_geoService->getKurtosisL());
      _histoVec[6]->Fill(_geoService->getKurtosisT());
      _histoVec[7]->Fill(_geoService->getLength());
      _histoVec[8]->Fill(_geoService->getWidth());
      _histoVec[9]->Fill(_geoService->getRadius());
      _histoVec[10]->Fill(xray->getRotAngle());
      
      double balance = xray->getEnergyFromCharge() / ((double)xray->getPixels());

      _histoVec[11]->Fill(balance);

      double fraction = (double)_geoService->numberPixelsWithinRadius(0.5*_geoService->getRadius()) / (double)xray->getPixels();

      _histoVec[12]->Fill(fraction);

      double fraction2 = (double)_geoService->numberPixelsWithinRadius(_geoService->getRmsY()) / (double)xray->getPixels();

      _histoVec[13]->Fill(fraction2);

      double lengthDivByRadius = _geoService->getLength()/_geoService->getRadius();

      _histoVec[14]->Fill(lengthDivByRadius);

      double radiusDivByRmsY = _geoService->getRadius()/_geoService->getRmsY();

      _histoVec[15]->Fill(radiusDivByRmsY);

      double lengthDivByRmsY = _geoService->getLength()/_geoService->getRmsY();

      _histoVec[16]->Fill(lengthDivByRmsY);

      _scatterVec[0]->Fill(_geoService->getRmsX(),_geoService->getRmsY());
      _scatterVec[1]->Fill(xray->getCharge(),_geoService->getExcentricity());
      _scatterVec[2]->Fill(xray->getRotAngle(),_geoService->getExcentricity());
      _scatterVec[3]->Fill(xray->getRotAngle(),_geoService->getRmsY());
      _scatterVec[4]->Fill(_geoService->getRmsY(),balance);
      _scatterVec[5]->Fill(xray->getEnergyFromCharge(),xray->getPixels());

      double distanceFromChipCenter = std::sqrt( (xray->getCenterX() - 7.)*(xray->getCenterX() - 7.) + (xray->getCenterY() - 7.)*(xray->getCenterY() - 7.) );

      _scatterVec[6]->Fill(distanceFromChipCenter,_geoService->getSkewnessT());
      _scatterVec[7]->Fill(distanceFromChipCenter,_geoService->getKurtosisT());
      
      _scatterVec[8]->Fill(xray->getPixels(),_geoService->getRmsY());
      _scatterVec[9]->Fill(xray->getCharge(),_geoService->getRmsY());

      delete xray;

    }
    
  }
  
  
  void XrayGeometryAnalysisProcessor::end(){
    
    delete _geoService;
    
    marlin::AIDAProcessor::tree(this);
    
    for(int i = 0; i < (int)_histoVec.size(); i++){
      _histoVec[i]->Write();
      delete _histoVec[i];
    }
    
    for(int i = 0; i < (int)_scatterVec.size(); i++){
      _scatterVec[i]->Write();
      delete _scatterVec[i];
    }
    
  }
  
  
  
}//namespace marlintpc
