#include "MMRefitTool.h"
#include <cmath>
#include <iostream>

//C++
#include <vector>

//LCIO
#include "lcio.h"
#include <EVENT/LCFloatVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackerHitImpl.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>

// ROOT
#include <Minuit2/FCNBase.h>
#include <TMath.h>
#include "Math/Functor.h"
#include <Math/Minimizer.h>
#include <Math/Factory.h>

// Minuit2 headers used to find seed track.
#include <Minuit2/MnUserParameters.h>
#include <Minuit2/MnMigrad.h>
#include <Minuit2/FunctionMinimum.h>
#include <Minuit2/Minuit2Minimizer.h>

//MARLIN
#include "marlin/VerbosityLevels.h"
#include <marlin/Exceptions.h>
#include <marlin/Global.h>
#include "IntersectionCalculator.h"
#include "MMHitFinderProcessor.h"

namespace marlintpc
{

  MMRefitTool::MMRefitTool(int fitType, int nPar, std::string prfType, bool rowDep, const IMPL::TrackImpl* seed, bool fitTrack, int normType, bool padHits) : 
    _fit_type(fitType), _nPar(nPar), _prfType(prfType), _rowDep(rowDep), _moduleDep(false), _seedTrack(seed), _fitTrack(fitTrack), _normType(normType), 
    _padHits(padHits)
  {

    //Fill the PRF vector
    const EVENT::TrackerHitVec & trackHits = _seedTrack->getTrackerHits();
    const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

    for (EVENT::TrackerHitVec::const_iterator hitIt = trackHits.begin(); hitIt != trackHits.end(); ++hitIt) 
    {
      const IMPL::TrackerHitImpl * aHit = static_cast<const IMPL::TrackerHitImpl*>(*hitIt);

      const EVENT::LCObjectVec& pulses = aHit->getRawHits();

      const gear::TPCModule& module = tpcParameters.getModule(static_cast<const EVENT::TrackerPulse *>(pulses[0])->getCellID1());

      //Temporary local variables
      float  padAmp = 0.;
      float  maxAmp = 0.;
      double position = 0.;
      LCObjectVec::const_iterator max;
      gear::Vector2D padCenterLocal;
      std::vector< std::pair <double, float> > fitData;

      for(LCObjectVec::const_iterator pulseIt = pulses.begin(); pulseIt != pulses.end(); ++pulseIt)
      {
        const EVENT::TrackerPulse * tmp = static_cast<const EVENT::TrackerPulse *>(*pulseIt);

        padAmp = tmp->getCharge();
        if(padAmp > maxAmp)
        {
          maxAmp = padAmp;
          max = pulseIt;
        }
        padCenterLocal = module.getLocalPadLayout().getPadCenter( tmp->getCellID0() );
        position = padCenterLocal[0]*padCenterLocal[1]; // convert to arclength.

        fitData.push_back(std::pair<double, float>(position, padAmp));
      }

      if(_padHits)
      {
        padData(static_cast<const EVENT::TrackerPulse *>(*max), pulses, fitData);
      }

      MMPadResponseFunction aPRF(_prfType);
      aPRF.SetData(fitData);

      switch(_normType)
      {
        case 1:
          aPRF.SetFit(MMPadResponseFunction::SUM_FIT); break;
        case 2:
          aPRF.SetFit(MMPadResponseFunction::GAUS_FIT); break;
        case 3:
          aPRF.SetFit(MMPadResponseFunction::CHI2_MIN); break;
        default:
          aPRF.SetFit(MMPadResponseFunction::FIT_BOTH); break;
      }

      _prfs.push_back(aPRF);
    }

    if(_fit_type == 0)
      trackFitter = TrackFitterFactory::getTrackFitter( 
          TrackFitterBase::FitterTypes::LINEARREGRESSION,0);
    else if(_fit_type == 1)
      trackFitter = TrackFitterFactory::getTrackFitter( 
          TrackFitterBase::FitterTypes::SIMPLEHELIXFIT,0);


  }

  double MMRefitTool::operator()(const double* x) const 
  {
    //DEBUG
    streamlog_out(DEBUG4) << "Start of function call." << std::endl;

    IMPL::TrackImpl * simpleCopy = new IMPL::TrackImpl();

    int ndf = 0;
    int nHits = 0;
    double chi2 = 0.;

    const EVENT::TrackerHitVec & trackHits = _seedTrack->getTrackerHits();
    const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

    std::vector<IMPL::TrackerHitImpl *> copyHits(trackHits.size());

    for (unsigned int i = 0; i < trackHits.size(); i++) 
    {
      const IMPL::TrackerHitImpl * tmp = static_cast<const IMPL::TrackerHitImpl*>(trackHits[i]);
      const gear::TPCModule& module = tpcParameters.getModule(static_cast<const EVENT::TrackerPulse *>(tmp->getRawHits()[0])->getCellID1());

      //Fetch the PRF parameters
      int modIndex = -1;
      int rowIndex = -1;
      if(_moduleDep)
      {
        std::map<int, int>::const_iterator modIt = _moduleList.find(module.getModuleID());
        if(modIt == _moduleList.end())
        {
          continue;
        }
        else
          modIndex = modIt->second;
      }
      if(_rowDep) rowIndex = module.getRowNumber( static_cast<const EVENT::TrackerPulse *>(tmp->getRawHits()[0])->getCellID0() );
      const std::vector<double> tmpPar = MMHitFinderProcessor::calcParArray(x, _nPar, (int)_moduleList.size(), modIndex, rowIndex);

      //Assemble full parameter variable (i.e., x0 and PRF parameters)
      std::vector<double> eval(_nPar + 1);
      int totalPar = _nPar;
      if(_moduleDep) totalPar *= _moduleList.size();
      if(_rowDep) totalPar *= 3; //This should be retrievable from one central location
      eval[0] = x[totalPar + i];

      for(int j = 0; j < _nPar; j++)
      {
        eval[1+j] = tmpPar[j];
      }

      double hitChi2 = _prfs[i](&(eval[0]));

      if(!_fitTrack && hitChi2 != 9E9)
      {
        nHits++;
        ndf += 1;
        chi2 += hitChi2/(double)tmp->getRawHits().size();
        continue;
      }

      if(hitChi2 != 9E9)
      {
        gear::Vector2D copyPosLocal = module.getLocalPadLayout().getPadCenter( static_cast<const EVENT::TrackerPulse *>(tmp->getRawHits()[0])->getCellID0() );
        copyPosLocal[1] = eval[0]/copyPosLocal[0];
        gear::Vector2D copyPosGlobal = module.localToGlobal(copyPosLocal[0], copyPosLocal[1]);
        std::vector<double> copyPos(3); 
        copyPos[0] = copyPosGlobal[0];
        copyPos[1] = copyPosGlobal[1]; 
        copyPos[2] = tmp->getPosition()[2];

        IMPL::TrackerHitImpl * copyHit = new IMPL::TrackerHitImpl();
        copyHit->setPosition(&copyPos[0]);
        copyHit->setCovMatrix(tmp->getCovMatrix());

        simpleCopy->addHit(copyHit);
        nHits++;
        ndf += 1;
        chi2 += hitChi2/(double)tmp->getRawHits().size();

        copyHits[i] = copyHit;
      }

      //More Debug (This class took a lot of debugging...)
      if(!_rowDep)
      {
        streamlog_out(DEBUG4) << "Hit Chi2 for hit " << i << " = " << hitChi2 << ",\t x0 = " << x[i+_nPar] << std::endl;
      }
      else
      {
        streamlog_out(DEBUG4) << "Hit Chi2 for hit " << i << " = " << hitChi2 << ",\t x0 = " << x[i+2*_nPar] << std::endl;
      }
    }

    streamlog_out(DEBUG4) << "ndf before fit:  " << ndf << std::endl;

    if(_fitTrack)
    {
      float ref[3] = {_seedTrack->getReferencePoint()[0], _seedTrack->getReferencePoint()[1], _seedTrack->getReferencePoint()[2]};
      simpleCopy->setReferencePoint(ref);

      //Fit the copied track
      IMPL::TrackImpl * refitTrack = trackFitter->fitTrack(simpleCopy);

      chi2 += refitTrack->getChi2()/(3.*(double)nHits - 4. - (double)_fit_type); //3 dof per hit, 4 or 5 track parameters, depending on magnetic field.

      delete refitTrack;
    }

    for(unsigned int i = 0; i < copyHits.size(); i++)
      delete copyHits[i];
    delete simpleCopy;

    if(nHits < 5) return 9E9;

    else
    {

      if(_rowDep)
      {
        streamlog_out(DEBUG3) << "r = " << x[1] << ", w = " << x[3] << std::endl;
      }
      else
      {
        streamlog_out(DEBUG3) << "r = " << x[0] << ", w = " << x[1] << std::endl;
      }
      streamlog_out(DEBUG3) << "Chi2: " << chi2 << std::endl;
      streamlog_out(DEBUG3) << "ndf:  " << ndf << std::endl;

      return chi2;
    }

  }

  std::vector<double> MMRefitTool::retrievePar(const double* x)
  {

    IMPL::TrackImpl * simpleCopy = new IMPL::TrackImpl();

    const EVENT::TrackerHitVec & trackHits = _seedTrack->getTrackerHits();
    const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

    std::vector<IMPL::TrackerHitImpl *> copyHits(trackHits.size());

    double chi2 = 0.;

    for (unsigned int i = 0; i < trackHits.size(); i++) 
    {
      const IMPL::TrackerHitImpl * tmp = static_cast<const IMPL::TrackerHitImpl*>(trackHits[i]);

      const gear::TPCModule& module = tpcParameters.getModule(static_cast<const EVENT::TrackerPulse *>(tmp->getRawHits()[0])->getCellID1());

      //Evaluate the PRF again so that we report only the hits that were actually used in the fit at the minimum
      double * eval = new double[_nPar + 1];
      eval[0] = x[_nPar + i];

      for(int j = 0; j < _nPar; j++)
      {
        int index = j;

        if(_moduleDep)
        {
          std::map<int, int>::const_iterator modIndex = _moduleList.find(module.getModuleID());
          index = j*_moduleList.size() + modIndex->second;
        }

        if(_rowDep)
        {
          int row = module.getRowNumber( static_cast<const EVENT::TrackerPulse *>(tmp->getRawHits()[0])->getCellID0() );
          eval[1+j] = MMHitFinderProcessor::calcRowDep(x, index, row);
        }
        else
        {
          eval[1+j] = x[index];
        }
      }

      _prfs[i].SetFit(MMPadResponseFunction::FIT_BOTH);
      double hitChi2 = _prfs[i](eval);

      if(hitChi2 != 9E9)
      {
        gear::Vector2D copyPosLocal = module.getLocalPadLayout().getPadCenter( static_cast<const EVENT::TrackerPulse *>(tmp->getRawHits()[0])->getCellID0() );
        copyPosLocal[1] = eval[0]/copyPosLocal[0];

        gear::Vector2D copyPosGlobal = module.localToGlobal(copyPosLocal[0], copyPosLocal[1]);
        double copyPos[3] = {copyPosGlobal[0], copyPosGlobal[1], tmp->getPosition()[2]};

        IMPL::TrackerHitImpl * copyHit = new IMPL::TrackerHitImpl();
        copyHit->setPosition(copyPos);
        copyHit->setCovMatrix(tmp->getCovMatrix());

        simpleCopy->addHit(copyHit);
        copyHits[i] = copyHit;

        chi2 += hitChi2/(double)tmp->getRawHits().size();
      }
    }

    float ref[3] = {_seedTrack->getReferencePoint()[0], _seedTrack->getReferencePoint()[1], _seedTrack->getReferencePoint()[2]};
    simpleCopy->setReferencePoint(ref);

    //Fit the copied track
    IMPL::TrackImpl * refitTrack = trackFitter->fitTrack(simpleCopy);

    //		chi2 += refitTrack->getChi2()/(3.*(double)copyHits.size() - 4. - (double)_fit_type);

    std::vector<double> trackPar(8);
    trackPar[0] = (double)refitTrack->getD0();
    trackPar[1] = (double)refitTrack->getPhi();
    if(_fit_type == 1)
      trackPar[2] = (double)refitTrack->getOmega();
    else
      trackPar[2] = 0.;
    trackPar[3] = (double)refitTrack->getZ0();
    trackPar[4] = (double)refitTrack->getTanLambda();
    trackPar[5] = (double)refitTrack->getChi2();
    trackPar[6] = chi2;
    trackPar[7] = (double)copyHits.size();

    delete refitTrack;
    delete simpleCopy;

    for(unsigned int i = 0; i < copyHits.size(); i++)
      delete copyHits[i];

    return trackPar;
  }

  std::vector<double> MMRefitTool::retrieveChi2(const double* x)
  {

    const EVENT::TrackerHitVec & trackHits = _seedTrack->getTrackerHits();
    const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

    std::vector<double> retVal(trackHits.size());

    for (unsigned int i = 0; i < trackHits.size(); i++) 
    {
      const IMPL::TrackerHitImpl * tmp = static_cast<const IMPL::TrackerHitImpl*>(trackHits[i]);

      const gear::TPCModule& module = tpcParameters.getModule(static_cast<const EVENT::TrackerPulse *>(tmp->getRawHits()[0])->getCellID1());

      //Evaluate the PRF again so that we report only the hits that were actually used in the fit at the minimum
      double * eval = new double[_nPar + 1];
      eval[0] = x[_nPar + i];

      for(int j = 0; j < _nPar; j++)
      {
        int index = j;

        if(_moduleDep)
        {
          std::map<int, int>::const_iterator modIndex = _moduleList.find(module.getModuleID());
          index = j*_moduleList.size() + modIndex->second;
        }

        if(_rowDep)
        {
          int row = module.getRowNumber( static_cast<const EVENT::TrackerPulse *>(tmp->getRawHits()[0])->getCellID0() );
          eval[1+j] = MMHitFinderProcessor::calcRowDep(x, index, row);
        }
        else
        {
          eval[1+j] = x[index];
        }

      }

      _prfs[i].SetFit(MMPadResponseFunction::FIT_BOTH);
      double hitChi2 = _prfs[i](eval);

      if(hitChi2 != 9E9)
      {
        retVal[i] = hitChi2;
      }
    }

    return retVal;

  }


  void MMRefitTool::padData(const EVENT::TrackerPulse * maxPulse, const EVENT::LCObjectVec& pulses, std::vector< std::pair <double, float> >& fitData) const
  {
    const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

    int nLeft = 0;
    int nRight = 0;

    for(unsigned int i = 0; i < pulses.size(); i++)
    {
      if(pulses[i] == maxPulse) break;
      nLeft++;
    }

    nRight = pulses.size() - nLeft - 1;

    int prevPadID = static_cast<const EVENT::TrackerPulse *>(pulses[0])->getCellID0();
    const gear::TPCModule& module = tpcParameters.getModule(static_cast<const EVENT::TrackerPulse *>(pulses[0])->getCellID1());

    gear::Vector2D padCenterLocal;

    for(int i = 0; i < 2 - nLeft; i++)
    {
      try
      {
        int leftNeighbour = module.getLeftNeighbour(prevPadID);

        // Get the location of the pad center.
        padCenterLocal = module.getPadCenter( leftNeighbour );

        // Add the point to the vector of fit data.
        fitData.push_back(std::pair<double, float>(padCenterLocal[0]*padCenterLocal[1], 0));

        prevPadID = leftNeighbour;

      }	
      catch(std::exception &e)
      {
        break;
      }

    }

    prevPadID = static_cast<const EVENT::TrackerPulse *>(pulses.back())->getCellID0();

    for(int i = 0; i < 2 - nRight; i++)
    {
      try
      {
        int rightNeighbour = module.getRightNeighbour(prevPadID);

        // Get the location of the pad center.
        padCenterLocal = module.getPadCenter( rightNeighbour );

        // Add the point to the vector of fit data.
        fitData.push_back(std::pair<double, float>(padCenterLocal[0]*padCenterLocal[1], 0));

        prevPadID = rightNeighbour;

      }	
      catch(std::exception &e)
      {
        break;
      }

    }
  }//End padData()

}//End namespace marlintpc

