#ifndef TPCGAS_H
#define TPCGAS_H 1

#include <string>

namespace marlintpc
{
  
  class TPCGas;
  
  /** The TPCGas is a helper class containing a parametrisation of some gas mixtures
   *  It can be used by processors thet need to know drift velocity or diffusion properties
   *  Usage: create a gas with a name, set E and B field, then values are available through the get functions.
   *  @author Astrid Muennich
   */
  
  
  class TPCGas
  {
  public:
    ///Constructor, gases are identified by names, e.g. "TDR","P5"
    TPCGas(std::string gasname);
    ///Destructor
    ~TPCGas();
    ///Set the electric field, if field is changed call the CalculateGasParameters method again to obtain new values
    void SetEField(double EField){_EField=EField;}
    double GetEField(){return _EField;}
    ///Set the magnetic field, if field is changed call the CalculateGasParameters method again to obtain new values
    void SetBField(double BField){_BField=BField;}
    double GetBField(){return _BField;}
    ///returns the drift velocity in [mm/us]
    double GetDriftVelocity(){return _vdrift;}
    ///returns transverse diffusion in [um/sqrt(cm)]
    double GetTransverseDiffusion(){return _diffTrans;}
    ///returns longitudinal diffusion in [um/sqrt(cm)]
    double GetLongitudinalDiffusion(){return _diffLong;}
    ///calculate attachment over a given distance in cm
    double GetAttachment(double EField, double dist);
   ///After name, E and B field are set call this function to calcuate the gas parameters
    void CalculateGasParameters();
    std::string GetGasName(){return _gasName;}
 
  protected:
    void CalculateTransverseDiffusion();
    ///used for v drift and diffusion and all gases, just the parameters change
    double EvaluateEFieldDependence(double a, double b ,double c, double d,
				    double e, double f, double g, double h,
				    double i, double j, double k);
    ///used for transverse diffusion dependence on magnetic field
    double EvaluateBFieldDependence(double a, double b, double c, double d,double e, double f);
    
    void initTDRGas();
    void initP5Gas();
    void initP10Gas();
    void initT2KGas();

    double _BField;
    double _EField;
    double _vdrift;
    double _diffLong;
    double _diffTrans;
    double _attachment;

    ///container for the parametrisation of longitudinal diffusion
    double _dL_params[11];
   ///container for the parametrisation of drift velocity
    double _vd_params[11];
    ///container for the parametrisation of transverse diffusion
    double _dT_params[10][6];
    ///container for parametrisation of attachment
    double _att[7];

    std::string _gasName;

   };
  
} //namespace marlintpc 

#endif // TPCGAS_H
