#include <algorithm>
//#include <vector>
//#include <string>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <Exceptions.h>
#include <streamlog/streamlog.h>
#include <lcio.h>

#include <IMPL/TrackerRawDataImpl.h>
#include "ALTROADCConvolutedPulse.h"
#include "TMath.h"


namespace marlintpc
{

ALTROADCConvolutedPulse::ALTROADCConvolutedPulse(double frequencyMHz, double peakingTime, double nElectronsDynamicRange, int ADCMaxValue, int threshold)
    :  ADCPulse(frequencyMHz, ADCMaxValue, threshold), _peakingTime(peakingTime)
{

    _ADCChargeFactor = _ADCMaxValue / nElectronsDynamicRange ;
    _g4Par1 = 4.0;
    _g4expPar1 = exp(_g4Par1);

    _useJitter = false;
}

void ALTROADCConvolutedPulse::convoluteSignal() {

	// Set the range where the transfer function is valid.
	double cnvMin = 0.;
	double cnvMax = 1.e10;
	double m_tStep = 1.;
	//make vector and fill with gamma4 function. vector should be of relative large size and is chosen here arbitrarily
	std::vector<double> cnvTab;
	unsigned int minimalLengthOfVector = 28;
	if (_rawData.size()*2>minimalLengthOfVector-1) cnvTab.resize(_rawData.size()*2);
	else cnvTab.resize(minimalLengthOfVector);
	unsigned int iOffset = 0;
	// Evaluate the transfer function.
	for (unsigned int i = 0; i < cnvTab.size(); ++i) {
		if (i<iOffset) {cnvTab[i]=0;}
		else
		{
			double t = (i-iOffset) * m_tStep;
			if (t < cnvMin || t > cnvMax)
			{
				cnvTab[i] = 0.;
			}
			else
			{
				cnvTab[i] = _gamma4(t);
			}
		}
	}
//	for (int i =0 ; i< _rawData.size();i++)
//		std::cout<< "_rawData[" <<i<<"]="<<_rawData[i]<<std::endl;
//	for (int i =0 ; i< cnvTab.size();i++)
//		std::cout<< "cnvTab[" <<i<<"]="<<cnvTab[i]<<std::endl;
	// increase size of rawdata by adding zeros at the front and back
	std::vector<double> zerovector (cnvTab.size()-_rawData.size()-5,0);
	_rawData.insert(_rawData.end(),zerovector.begin(),zerovector.end());
	std::vector<double> zerovector2 (5,0);
	_rawData.insert(_rawData.begin(),zerovector2.begin(),zerovector2.end());

	//Evaluate the convolution of rawdata and electronics response
	std::vector<double> tmpSignal;
	tmpSignal.resize(_rawData.size());
	for (unsigned int j = 0; j < _rawData.size(); ++j) {
		tmpSignal[j] = 0.;
		for (unsigned int k = 0; k <= j; ++k) {
			tmpSignal[j] +=
					m_tStep * cnvTab[j - k] * _rawData[k];
		}
	}
	// fill convoluted signal back into _rawdata
    for (unsigned int j = 0; j < _rawData.size(); ++j) {
    	_rawData[j] = tmpSignal[j];
    }
//	for (int i =0 ; i< _rawData.size();i++)
//		std::cout<< "_rawData[" <<i<<"]="<<_rawData[i]<<std::endl;
}


// implementation as a simple function is much faster than TF1::Eval
double ALTROADCConvolutedPulse::_gamma4(double x)
{
	double ADCChargeValue=_ADCChargeFactor;
	double startTime = 0;
    // parameters of the gamma4 function:
    // charge, order, start time, peaking time
	// not normalized
    // return ((ADCChargeValue * _g4expPar1 * pow(((x - startTime) / _peakingTime), _g4Par1) * exp(-_g4Par1 * ((x - startTime) / _peakingTime))) * (x > startTime));
    // correctly normalized
    return (ADCChargeValue* (pow(((x - startTime) / (_peakingTime/(1/_frequencyGHz))), _g4Par1) * exp(-_g4Par1 * ((x - startTime) / (_peakingTime/(1/_frequencyGHz)))) / (pow(1./(_peakingTime/(1/_frequencyGHz)),_g4Par1-1) * pow(_g4Par1/(_peakingTime/(1/_frequencyGHz)),-_g4Par1) * TMath::Gamma(_g4Par1))) * (x > startTime));
}



void ALTROADCConvolutedPulse::add(double startTime, int nElectrons)
{
    // Check that the pulses really overlap
    if((startTime > getEndTime()) && (_startBin >= 0))
    {
        //    throw Exception("ADCPulse: ADCPulses do not overlap. Check the calling code!") ;
        throw lcio::Exception("ADCPulse: ADCPulses do not overlap. Check the calling code!") ;
    }

    // Check that the start time of the new charge is not before the start time of the existing pulse.
    // This cannot be handles by this algorithm, the calling code has to take care of this.
    if(startTime < _startBin / _frequencyGHz)
    {
        throw lcio::Exception("ALTROADCConvolutedPulse: Error: Start time of new charge is before the start of the existing pulse. Check the calling code!") ;
    }

    double ADCChargeValue = nElectrons * _ADCChargeFactor;

    //shaping function starts at t0=startTime.

    int bin = static_cast<int>(startTime * _frequencyGHz); // bin number for the first bin with charge
    if(_startBin < 0) // if there is no charge in the pulse yet the pulse starts at "bin"
    {
        _startBin = bin;
    }
    streamlog_out(DEBUG2) << "GAMMA4: st " << startTime << ", sb " << _startBin << ", pt " << _peakingTime << ", ne " << nElectrons << ", adc " << ADCChargeValue << ", f" << _frequencyGHz << std::endl;
    double   ADCValue ;
//    	  double binTime = (_startBin + 0.5) / _frequencyGHz;
// 		  std::map<int, double>::iterator it;
//        // calculate the ADC value for this bin by evaluating gamma4 function in bin center.
//        if(_useJitter)
//        {
//            it = _timeJitter.find(bin);
//            if(it == _timeJitter.end())
//                throw lcio::Exception("ALTROADCConvolutedPulse: Error: Jitter map does not contain a value for this time bin!") ;
//            else
//	      ADCValue = _gamma4(binTime + it->second, ADCChargeValue, startTime);
//            streamlog_out(DEBUG1) << "bin " << bin << ", binTime " << binTime << ", ADCValue " << ADCValue << " using jitter of " << it->second << " (jitter map size " << _timeJitter.size() << ")" << std::endl;
//        }
//        else
//        {
//	    ADCValue = _gamma4(binTime, ADCChargeValue, startTime);
//        }
//         debug: check for 0
//
//        if(ADCValue == 0 || ADCValue < 0)
//        {
//            streamlog_out(DEBUG2) << "evaluation of gamma4 function at " << binTime
//                                  << " resulted in zero" << std::endl;
//        }

        // check if the bin already exists in the vector, else append it
        if((bin - _startBin) < static_cast<int>(_rawData.size()))
        {
            // time bin is already there, add the charge
            _rawData[bin - _startBin] += nElectrons;
            streamlog_out(DEBUG1) << "ADCValue " << ADCValue << ", _rawData for bin " << bin - _startBin << " is now " << _rawData[bin - _startBin] << std::endl;

        }
        else // the time bin did not exist yet
        {
            _rawData.push_back(nElectrons);
        }

}



}// namespace marlintpc
