/** Processor to sort amplified electrons into voxel
 */
#include "ChargeDistributionProcessor.h"
#include "TPCGas.h"

//MarlinTPC
#include "libgem.h"
#include "LCObjectCopier.h"

//tpcconddata
#include "TPCConditions.h"

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd/DataFileHandler.hh>



//C++
#include <iostream>
#include <cmath>
#include <cstring>

//LCIO
#include <IO/LCWriter.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimTrackerHitImpl.h>
#include <EVENT/LCCollection.h>

//GEAR
#include <gear/TPCParameters.h>
#include <gearimpl/Vector3D.h>
#include <gear/BField.h>

//Marlin
#include <marlin/Global.h>



using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;

namespace marlintpc
{

std::list< std::pair< VoxelTPC::VoxelIndex, int > >* ChargeDistributionProcessor::_fetchedList = 0;


ChargeDistributionProcessor aChargeDistributionProcessor ;


ChargeDistributionProcessor::ChargeDistributionProcessor() : Processor("ChargeDistributionProcessor"),
    _trackerData(0), _isFirstEvent(true)
{

    _description = "ChargeDistributionProcessor maps charge to padplane ..." ;


    // register steering parameters: name, description, class-variable, default value


    registerInputCollection(LCIO::SIMTRACKERHIT,
                            "InputCollectionName" ,
                            "Name of the Input collection with drifted electrons (default: AmplifiedElectrons)"  ,
                            _inputCollectionName ,
                            string("AmplifiedElectrons")) ;

    registerOutputCollection(LCIO::LCGENERICOBJECT ,
                             "OutputCollectionName" ,
                             "Name of the Output GEM collection (default: DistributedElectrons)" ,
                             _outputCollectionName ,
                             string("DistributedElectrons"));

    registerProcessorParameter("ETrans1" ,
                               "Value of Transfer Field 1 in V/cm (default: 2500)"  ,
                               _ETrans1 ,
                               float(2500)) ;

    registerProcessorParameter("ETrans2" ,
                               "Value of Transfer Field 2 in V/cm (default: 2500)"  ,
                               _ETrans2 ,
                               (float)2500) ;

    registerProcessorParameter("EInd" ,
                               "Value of Induction Field in V/cm (default: 5000)"  ,
                               _EInd ,
                               (float)5000) ;

    registerProcessorParameter("Gas" ,
                               "DriftGas (default: TDR)"  ,
                               _gas ,
                               std::string("TDR")) ;

    registerProcessorParameter("BunchIntervall",
                               "Intervall between two bunch crossings in ns (default: 369)",
                               _bunchIntervall,
                               (double) 369);

    registerProcessorParameter("WriteOutputToStorage",
                               "If true the output collection is written (default: true)",
                               _outputIsPersistent,
                               true);

    registerProcessorParameter("zBinLength" ,
                               "Length of a zBin in ns (default: 1)"  ,
                               _zBinLength,
                               (double) 1) ;

    registerOptionalParameter("TPCConditions" ,
                              "Name of collection with TPCConditions" ,
                              _tpcConditionsColName,
                              tpcconddata::TPCConditions::getDefaultColName());

    registerOptionalParameter("AlternativTPCConditionsName",
                              "This is the name of newly created TPCConditions, if the original file specified by \"TPCConditions\" is not writeable (default: TPCConditionsChargeDistr)" ,
                              _altTPCConditionsName,
                              string("TPCConditionsChargeDistr")) ;

    registerProcessorParameter("IonisationEnergy" ,
                               "The energy in GeV needed to ionize a gas particle (default: 26.E-9 = Argon) ",
                               _ionisationEnergy,
                               26.E-9); //default value is for argon

    registerOptionalParameter("WriteAfterNEvents" ,
                              "Force readout of the full TPC after overlaying n events (optional)" ,
                              _writeAfterNEvents,
                              312);

    registerProcessorParameter("ImproveMemoryFootprint",
                               "Flag whether to use fetchVoxelCollection instead of the output collection (default:false)",
                               _improveMemoryFootprint,
                               false);

    registerOptionalParameter("SigmaOverride",
                              "Override the sigma which usually is calculated from the diffusion in the GEM stack (optional)",
                              _sigmaOverride,
                              0.0);
    registerProcessorParameter("InductionFactor",
                               "Amount of induction on neighbour pads in % of the charge on the central pad",
                               _inductionFactor,
                               0.0);
    registerOptionalParameter("BFieldOverride",
                              "Override the magnetic field taken from gear.",
                              _BOverride,
                              0.0);
}

void ChargeDistributionProcessor::init()
{

    streamlog_out(MESSAGE2) << _description << endl;

    _tpcParameters = &(Global::GEAR->getTPCParameters());

    if(parameterSet("BFieldOverride"))
    {
        _BField = _BOverride;
        streamlog_out(WARNING) << "Using override parameter " << _BOverride << "for the magnetic field." << std::endl;

    }
    else
    {
        //TODO: Check if maybe also another component of B is used as drift direction by someone else!
        try
        {
            // try to get the magnetic field from the global B field
            //REMARK: Here the z component is considered as drift direction
            _BField = (Global::GEAR->getBField().at(gear::Vector3D(0, 0, 0))).z();
        }
        catch(gear::UnknownParameterException&)
        {
            streamlog_out(WARNING) << "Failed to load the b field from the global gear definition...you should add something like"
                                   << "<BField type=\"ConstantBField\" x=\"0.0\" y=\"0.0\" z=\"1.0\"/> to your gear file." << std::endl;
            streamlog_out(WARNING) << "Now I will try to get the B field from TPCParameters..." << std::endl;
            // try to get the magnetic field in the deprecated style
            _BField = Global::GEAR->getTPCParameters().getDoubleVal("BField");
        }
    }

    if(parameterSet("SigmaOverride"))
    {
        _sigma = _sigmaOverride;
        _sigmaLong = 0;
        streamlog_out(WARNING) << "Using override parameter " << _sigmaOverride << "for sigma."
                               << "GEM parameters will be ignored." << std::endl;

    }
    else
    {
        //this is for a GEM setup with 2 mm between the GEMs and 3 mm below
        //NOTE: (this should match with setup used in GEMProcessor to get the attachment correct)
        TPCGas* myGas = new TPCGas(_gas);
        myGas->SetBField(_BField);
        myGas->SetEField(_ETrans1);
        myGas->CalculateGasParameters();
        double diff_trans1 = myGas->GetTransverseDiffusion();
        double diff_long1 = myGas->GetLongitudinalDiffusion();
        myGas->SetEField(_ETrans2);
        myGas->CalculateGasParameters();
        double diff_trans2 = myGas->GetTransverseDiffusion();
        double diff_long2 = myGas->GetLongitudinalDiffusion();
        myGas->SetEField(_EInd);
        myGas->CalculateGasParameters();
        double diff_ind = myGas->GetTransverseDiffusion();
        double diff_longind = myGas->GetLongitudinalDiffusion();

        _sigma = 1. / 1000 * sqrt((0.2 * pow(diff_trans1 , 2)
                                   + 0.2 * pow(diff_trans2 , 2)
                                   + 0.3 * pow(diff_ind    , 2))); // in mm

        _sigmaLong = 1. / 1000 * sqrt((0.2 * pow(diff_long1 , 2)
                                       + 0.2 * pow(diff_long2 , 2)
                                       + 0.3 * pow(diff_longind    , 2))); // in mm


        streamlog_out(MESSAGE1) << "Diffusion (transversal): " << _sigma << endl;
        streamlog_out(MESSAGE1) << "Diffusion (longitudinal): " << _sigmaLong << endl;
    }

    //Get a VoxelTPC
    if(_trackerData) delete _trackerData;
    _trackerData = new marlintpc::VoxelTPC(_tpcParameters, _zBinLength, _inductionFactor);

    _isFirstEvent = true;

    printParameters();
}



void ChargeDistributionProcessor::processRunHeader(lcio::LCRunHeader* run)
{

    run->parameters().setValue(_processorName + "_revision", "$Rev: 4976 $");

    for(ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
    {
        if(! i->second->isOptional() || i->second->valueSet())
        {
            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
        }
    }

}


void ChargeDistributionProcessor::processEvent(lcio::LCEvent* evt)
{

    if(_isFirstEvent)
    {

        LCCollectionVec* conditionsCol = 0;

        try
        {
            //First try to get the conditions collection with TPCConditions from
            //the current event.
            conditionsCol = dynamic_cast<LCCollectionVec*>(evt->getCollection(_tpcConditionsColName));
            TPCConditions tpcCond(conditionsCol->getElementAt(0));

            //Check whether object is writeable
            try
            {
                //This thing throws an exception if the object is write protected
                //tpcCond.obj()->checkAccess();

                //Add amplification to TPCCondtions if collection is writeable
                tpcCond.setTransDefocussing(_sigma);

            }
            catch(EVENT::ReadOnlyException&)
            {
                //If collection is not writeable an new writeable collection has
                //to be created with a new also writeable conditions object. The
                //new conditions object needs to be a deep copy of the original
                //conditions object

                conditionsCol  = new LCCollectionVec(LCIO::LCGENERICOBJECT);
                TPCConditions* newtpcCond = new TPCConditions;
                newtpcCond->setTransDiffusionCoef(tpcCond.getTransDiffusionCoef());
                newtpcCond->setLongDiffusionCoef(tpcCond.getTransDiffusionCoef());
                newtpcCond->setDriftVelocity(tpcCond.getDriftVelocity());

                //Add mean amplification to the new conditions object
                newtpcCond->setAmplification(_sigma);

                //Add conditions object to collection and collection to event
                conditionsCol->addElement(newtpcCond);
                evt->addCollection(conditionsCol, _altTPCConditionsName);

                //Create a lccd::DataFileHandler
                lccd::LCConditionsMgr::instance()->registerHandler(_altTPCConditionsName , new lccd::DataFileHandler(_altTPCConditionsName));

                // call updateEvent only for the TPCConditions handler! Don't run it on LCConditionsMgr!!!
                lccd::LCConditionsMgr::instance()->getHandler(_altTPCConditionsName)->updateEvent(evt) ;

                // The updateEvent performs an LCEvent::takeCollection, which sets the collection to transient
                // In this special case we want to write it to the file, so we have to remove the transient flag
                conditionsCol->setTransient(false);

            }// catch ReadOnlyException

        }
        catch(DataNotAvailableException& e)
        {
            //If there is no TPCConditions collection create a new one and register it
            //with lccd.

            conditionsCol  = new LCCollectionVec(LCIO::LCGENERICOBJECT);

            TPCConditions* tpcCond = new TPCConditions;
            tpcCond->setAmplification(_sigma);

            conditionsCol->addElement(tpcCond);
            evt->addCollection(conditionsCol, _tpcConditionsColName);

            //Create a lccd::DataFileHandler
            lccd::LCConditionsMgr::instance()->registerHandler(_tpcConditionsColName , new lccd::DataFileHandler(_tpcConditionsColName));

            // call updateEvent only for the TPCConditions handler! Don't run it on LCConditionsMgr!!!
            lccd::LCConditionsMgr::instance()->getHandler(_tpcConditionsColName)->updateEvent(evt) ;


            // The updateEvent performs an LCEvent::takeCollection, which sets the collection to transient
            // In this special case we want to write it to the file, so we have to remove the transient flag
            conditionsCol->setTransient(false);

        }// catch DataNotAvailableException

        _isFirstEvent = false;
    }// if isFirstEvent


    LCCollection* inputCol = 0;
    int objects = 1;

    try
    {
        inputCol = evt->getCollection(_inputCollectionName);
        objects = inputCol->getNumberOfElements();

        streamlog_out(DEBUG4) << "object number: " << objects << std::endl;
    }
    catch(EventException& exception)
    {
        streamlog_out(DEBUG4) << "Exception: " << exception.what() << endl;
        return;
    }

    //time of this event in ns
    double eventTime = evt->getEventNumber() * _bunchIntervall;

    if(_bunchIntervall < 0) eventTime = 0;

    for(int i = 0; i < objects; i++)
    {
        // prepare the sim hit for the voxelTPC
        // copy the hit before modification, the input collection should not be changed
        SimTrackerHitImpl* hit_in = LCObjectCopier::copy(dynamic_cast<SimTrackerHit*>(inputCol->getElementAt(i)));

        if(hit_in->getPosition()[2] < 0)
        {
            streamlog_out(DEBUG2) << "negative half tpc, skipping " << std::endl;
            delete hit_in;
            continue;
        }

        //The time of the hit is its drifttime [ns].
        hit_in->setTime(hit_in->getTime() + eventTime);

        streamlog_out(DEBUG4) << "hit has EDep of " << hit_in->getEDep() << std::endl;

        // check whether charge is in GeV or electrons
        if(hit_in->getEDep() < 1)  // obviously not in electrons
        {
            hit_in->setEDep(round(hit_in->getEDep() / _ionisationEnergy));
        }
        // else: everything above 1 is assumed to be number of electrons. 1 GeV per hit is too much

        // does the whole smearing and management of the charges
        streamlog_out(DEBUG4) << "now put charge In Voxels..." << std::endl;
        _trackerData->putChargeInVoxels(hit_in, _sigma, _sigmaLong);
        //    _trackerData->putChargeInVoxels(hit_in, _sigma, 0);

        // delete the temporary hit we created
        delete hit_in;
    }

    // filling of the map complete, give debug output
    streamlog_out(DEBUG4) << "Event " << evt->getEventNumber()
                          << " : Size of voxel map is " << _trackerData->getSize() << std::endl;

    // check if writing out after a fixed number of events is activated
    if(parameterSet("WriteAfterNEvents"))
    {
        // if the reqested number of events is reached read out the full TPC
        if(evt->getEventNumber() % _writeAfterNEvents == (_writeAfterNEvents - 1))
        {
            eventTime = -1; // negative times read out the whole map
        }
        else
        {
            return; // continue with next event, do not read out anything
        }
    }

    if(_bunchIntervall < 0)
    {
        eventTime = -1;
    }

    if(_improveMemoryFootprint)
    {
        // It is safe to do a delete since fetchVoxelList sets this to 0.
        // In case the list is not fetched this avoids a memory leak
        delete _fetchedList;
        _fetchedList = _trackerData->fetchVoxelList(eventTime);
    }
    else // no memory saving, write the output collection
    {
        LCCollectionVec* voxelVec = _trackerData->fetchVoxelCollection(eventTime);

        if(!voxelVec)
        {
            streamlog_out(DEBUG2) << "Empty voxel collection while input was not empty." << std::endl;
            return;
        }

        streamlog_out(DEBUG2) << "voxel collection has size " << voxelVec->size() << std::endl;

        if(_outputIsPersistent)
        {
            voxelVec->setTransient(false);
        }
        else
        {
            voxelVec->setTransient(true);
        }

        voxelVec->parameters().setValue("zBinLength", static_cast<float>(_zBinLength));
        evt->addCollection(voxelVec , _outputCollectionName);
    }
}




std::list< std::pair< VoxelTPC::VoxelIndex, int > >* ChargeDistributionProcessor::fetchVoxelList()
{
    std::list< std::pair<VoxelTPC::VoxelIndex, int > >* retval = _fetchedList;
    _fetchedList = 0;
    return retval;
}

}//end of namespace marlintpc
