//*****************************************************
// Part of program to simulate a TPC with GEM readout *
//                                                    *
// Author: Astrid Muennich                            *
//                                                    *
//                                                    *
//*****************************************************
/*! Programm to drift electrons in specified gas over certain
   distance.
*/

#include "DriftProcessor.h"


//tpcconddata
#include "TPCConditions.h"

//C++
#include <iostream>
#include <cmath>
#include <cstring>

//LCIO
#include <IO/LCWriter.h>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <EVENT/SimTrackerHit.h>
#include <IMPL/SimTrackerHitImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <Exceptions.h>

//GEAR
#include <gear/TPCParameters.h>

//Marlin
#include <marlin/Global.h>

//LCCD
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include "lccd/DataFileHandler.hh"

//CLHEP
#include <CLHEP/Random/RandBinomial.h> 



using namespace lcio ;
using namespace marlin ;
using namespace std;

namespace marlintpc{

DriftProcessor aDriftProcessor ;


  DriftProcessor::DriftProcessor() : Processor("DriftProcessor") ,
				     _isFirstEvent(true)
  {
    
    _description = "DriftProcessor drifts electrons to GEM structure ..." ;
    
    
   // register steering parameters: name, description, class-variable, default value

  registerInputCollection(    LCIO::SIMTRACKERHIT,
                              "InputCollectionName" , 
			      "Name of the input collection with primary electrons"  ,
			      _inputCollectionName ,
			      string("PrimaryIonisation") ) ;

  registerOutputCollection(   LCIO::SIMTRACKERHIT,
                              "OutputCollectionName" ,
                              "Name of the output collection with drifted electrons" ,
                              _outputCollectionName ,
                              string("DriftedElectrons") ) ;

  registerProcessorParameter( "Gas" , 
			      "DriftGas [TDR, P10, P5, T2K] (default: TDR)"  ,
			      _gas ,
			      std::string("TDR")) ;
  registerProcessorParameter( "DriftField" , 
			      "Drift Field ( default 260 V/cm )"  ,
			      _EField ,
			      (float)260.) ;
  registerProcessorParameter( "Transient" , 
			      "switch off writing of output collection"  ,
			      _transient,
			      (int) 0) ;

  registerProcessorParameter( "TPCConditions",
			      "Name of collection to write the TPCConditions in (default: TPCConditions)",
			      _tpcConditionsColName,
			      tpcconddata::TPCConditions::getDefaultColName() );

  registerProcessorParameter( "IonisationEnergy" ,
			      "The energy in GeV needed to ionize a gas particle (default: 26.E-9 = Argon) ",
			      _ionisationEnergy,
			      26.E-9); //default value is for argon

  registerProcessorParameter( "FullCurlerBackwardCompatibility" ,
			      "NOT IMPLEMENTED YET: Turn on for data from PrimaryIonisation < r1876 and FullCurler turned off (default: false)",
			      _fullCurlerBackwardCompatibility,
			      false);

  registerProcessorParameter( "Attachment" ,
 			      "Possible attachment per cm due to oxygen or water in gas (default: 0)",
 			      _attachment,
 			      (double) 0.0);
}





void DriftProcessor::init() {
 
  streamlog_out(MESSAGE2) << _description << endl;
 
  printParameters() ;

  //read TPC parameters from gear
  _TPCOuterRadius      = Global::GEAR->getTPCParameters().getDoubleVal("TPCRadius");
  _TPCInnerRadius = Global::GEAR->getTPCParameters().getDoubleVal("TPCInnerRadius");
  _BField         = Global::GEAR->getTPCParameters().getDoubleVal("BField");
  _TPCLength      = Global::GEAR->getTPCParameters().getMaxDriftLength();

  r = gsl_rng_alloc( gsl_rng_mt19937 );

  _distance = 0.;

  x_e = 0. ;
  y_e = 0. ;
  z_e = 0. ;
  t_e = 0. ; 
   
  xdrift = 0. ;
  ydrift = 0. ;
  zdrift = 0. ;
  tdrift = 0. ;

  //Calculate drift velocity, longitudinal and transverse diffusion
  _gasClass= new TPCGas(_gas);
  _gasClass->SetEField(_EField);
  _gasClass->SetBField(_BField);
  streamlog_out(MESSAGE0) <<"Drifting for "<< _gasClass->GetGasName()    <<" gas with E=" << _gasClass->GetEField()
			  <<" V/cm, B="    <<_gasClass->GetBField() <<" T"           << endl;
 
  _gasClass->CalculateGasParameters();
  _difflong= _gasClass->GetLongitudinalDiffusion();
  _difftrans=_gasClass->GetTransverseDiffusion();
  _vdrift= _gasClass->GetDriftVelocity() ;
  streamlog_out(MESSAGE0) <<"Longitudinal diffusion: "<< _difflong <<" [um/sqrt(cm)]"<<endl;
  streamlog_out(MESSAGE0)<<"Drift velocity: "<< _vdrift <<" [mm/us]"<<endl;
  streamlog_out(MESSAGE0) <<"Transverse diffusion: "<< _difftrans <<" [um/sqrt(cm)]"<<endl;

  _isFirstEvent = true;

}




void DriftProcessor::processRunHeader(lcio::LCRunHeader* run) 
{ 

  //Revision logging
  run->parameters().setValue( _processorName + "_revision", "$Rev$" );

  //Parameter logging
  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

    if ( ! i->second->isOptional() || i->second->valueSet() ) 
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
  }
} 





void DriftProcessor::processEvent(lcio::LCEvent* evt ) { 

  if( _isFirstEvent )
  {
    // only register the conditions data once
    _isFirstEvent = false;

    //Create a TPCConditions object and add it to the conditions data
    tpcconddata::TPCConditions* myTPCCondPtr = new tpcconddata::TPCConditions;
    myTPCCondPtr->setDriftVelocity( _vdrift );
    myTPCCondPtr->setTransDiffusionCoef( _difftrans );
    myTPCCondPtr->setLongDiffusionCoef( _difflong );
    
    LCCollectionVec* condColVec = new LCCollectionVec( LCIO::LCGENERICOBJECT );
    condColVec->addElement( myTPCCondPtr );

    evt->addCollection( condColVec, _tpcConditionsColName );

    //Create a lccd::DataFileHandler 
    lccd::LCConditionsMgr::instance()->registerHandler( _tpcConditionsColName , new lccd::DataFileHandler( _tpcConditionsColName ) );

    // call updateEvent only for the TPCConditions handler! Don't run it on LCConditionsMgr!!!
    lccd::LCConditionsMgr::instance()->getHandler( _tpcConditionsColName )->updateEvent(  evt ) ;
    
    // The updateEvent performs an LCEvent::takeCollection, which sets the collection to transient
    // In this special case we want to write it to the file, so we have to remove the transient flag
    condColVec->setTransient(false);

  }

  LCCollectionVec* inputCol = 0;

  try {
    inputCol = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputCollectionName  )) ;
  }catch(EventException &exception){
    return;
  }
 
  //Declare output collection and set transient
  LCCollectionVec* outputColVec = new LCCollectionVec( LCIO::SIMTRACKERHIT )  ;
  if( _transient )
    outputColVec->setTransient( true );

  for( LCCollectionVec::const_iterator inputClusterIter = inputCol->begin();
       inputClusterIter < inputCol->end(); ++inputClusterIter)
  {

    SimTrackerHit *hit_in=dynamic_cast<SimTrackerHit*>( *inputClusterIter );
    
    int nElectrons = static_cast<int>(round( hit_in->getEDep() / _ionisationEnergy ));

    if (_fullCurlerBackwardCompatibility)
    {
      // patch for files that have 10 electrons grouped to 1 simtrackerhit, but 
      // charge only set to 1*26eV
      MCParticle    *mcpart=dynamic_cast<MCParticle*>( hit_in->getMCParticle() );

      if( mcpart != 0 )
      {
	//Get momentum of the SimTrackerHits parent particle 
	const double * parray=mcpart->getMomentum();
   
	if( sqrt( pow(parray[0],2) + pow(parray[1],2) ) < 1 && 
	    atan( sqrt( pow(parray[0] , 2) + pow(parray[1] , 2) )/(fabs(parray[2]))) * 180/M_PI > 88)
	{ // it is a curler
	  if (nElectrons != 1)
          {
	    streamlog_out(WARNING) << "Curler backward compatibility requested but more than one electron "
				   << " in the SimTrackerHit. Skipping correction" << std::endl;
	  }
	  else
	  {
	    // correct the number of electrons
	    nElectrons = 10;
	  }
	}
      }
    }//  if (_fullCurlerBackwardCompatibility)

    for ( int ele = 0 ; ele < nElectrons; ele++)
    {
      GetElectronCoordinates(hit_in);

      if( z_e < 0 )
	_distance=-_TPCLength; // decide witch readoutplane
      else 
	_distance=_TPCLength;

      //check if electron gets attached during drift
      double driftdistance = _distance-z_e;
      if(_attachment>0)
	{
	  int attached = static_cast<int>(  CLHEP::RandBinomial::shoot( 1 ,exp( -(fabs(driftdistance)/10.*_attachment)) ));
	  if ( attached > 1 || attached < 0 )
	    throw std::out_of_range( "DriftProcessor::Attachment probability out of range [0,1]" );
	  if(attached==0)
	    continue;
	}
      MakeElectronDriftCoordinates();

      //check if e- driftet out ouf TPC volume
      if( xdrift*xdrift+ydrift*ydrift <= _TPCOuterRadius*_TPCOuterRadius &&
	  xdrift*xdrift+ydrift*ydrift >= _TPCInnerRadius*_TPCInnerRadius ){

      SimTrackerHitImpl* driftedElectron = new SimTrackerHitImpl ;
      double pos[3]={xdrift,ydrift,zdrift};
      driftedElectron->setPosition(pos);
      driftedElectron->setTime(tdrift);
    
      driftedElectron->setMCParticle( const_cast<MCParticle*>( hit_in->getMCParticle() ) );
 
      //Set energy, it's the ionisation energy of a single electron
      driftedElectron->setEDep(  _ionisationEnergy );

      //Add only electrons with z >0 to output collection
      //      if( hit_in->getPosition()[2] >= 0 )
      outputColVec->addElement( driftedElectron ) ;

      }// if in drift volume
    }//for ele
  }// for inputClusterIter

  //Write output collection
  if ( outputColVec->empty() ) 
    delete outputColVec;
  else
    evt->addCollection( outputColVec , _outputCollectionName ) ;
  
}


void DriftProcessor::GetElectronCoordinates( SimTrackerHit *hit ){

  const double *pos;
  pos=hit->getPosition();
  x_e=pos[0]; //input in mm 
  y_e=pos[1];
  z_e=pos[2];
  t_e=hit->getTime();
}

void DriftProcessor::MakeElectronDriftCoordinates(){

   //drift electron
  double driftdistance = _distance-z_e;	
  double meandrift = sqrt(fabs(driftdistance/10.))/1000.;
  double gaus_transx=gsl_ran_gaussian(r,_difftrans*meandrift);
  double gaus_transy=gsl_ran_gaussian(r,_difftrans*meandrift);
  double gaus_long=gsl_ran_gaussian(r,_difflong*sqrt(fabs(driftdistance/10.))/1000.); 
  /*For diffusion full tpc length, t=0 at fist GEMS, rest of transverse diff is 
   *done between GEMS in ChargeDistributionProcessor
   */ 
  tdrift = t_e + (fabs(driftdistance)+gaus_long)/(_vdrift*0.001); // [ns]
  xdrift = x_e + gaus_transx; // [mm]
  ydrift = y_e + gaus_transy;
  zdrift = z_e + driftdistance;

}







  void DriftProcessor::end(){
    
    lccd::LCConditionsMgr::instance()->removeHandler( "tpcCondHandler" );

  }

}//end of namespace marlintpc
