#include <algorithm>
//#include <vector>
//#include <string>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <Exceptions.h>
#include <streamlog/streamlog.h>
#include <lcio.h>

#include <IMPL/TrackerRawDataImpl.h>
#include "GaussianADCPulse.h"


namespace marlintpc
{

GaussianADCPulse::GaussianADCPulse(double frequencyMHz, double riseTime, double asym,
                                   double nElectronsDynamicRange, int ADCMaxValue, int threshold)
    :  ADCPulse(frequencyMHz, ADCMaxValue, threshold),
       _riseTime(riseTime), _asym(asym),
       _sqrt2_sigma(riseTime / 3. *sqrt(2))
{
    // Calculate the maximal possible charge in one time bin originating from
    // nElectronsDynamicRange electrons.
    // This value should correspond to ADCMaxValue.

    // The maximum charge is in one bin if the centre of the bin is the centre of the gaussian.
    // So the bin from -halfTimeBin to +halfTimeBin (if the centre is 0 like in the standard erf).
    double halfTimeBin = 0.5 / _frequencyGHz;
    double maxPossibleChargeIntegral = nElectronsDynamicRange / 2. *
                                       (erf(halfTimeBin / _sqrt2_sigma)
                                        - erf(-halfTimeBin / (_asym * _sqrt2_sigma)));
    _ADCChargeFactor = _ADCMaxValue / maxPossibleChargeIntegral;

    _useJitter = false;
}



void GaussianADCPulse::add(double startTime, int nElectrons)
{
    // Check that the pulses really overlap
    if((startTime > getEndTime()) && (_startBin >= 0))
    {
        //    throw Exception("ADCPulse: ADCPulses do not overlap. Check the calling code!") ;
        throw lcio::Exception("ADCPulse: ADCPulses do not overlap. Check the calling code!") ;
    }

    // Check that the start time of the new charge is not before the start time of the existing pulse.
    // This cannot be handles by this algorithm, the calling code has to take care of this.
    if(startTime < _startBin / _frequencyGHz)
    {
        throw lcio::Exception("GaussianADCPulse: Error: Start time of new charge is before the start of the existing pulse. Check the calling code!") ;
    }

    // the end time is start time + the width of the ADCPulse, currently 2*risetime
    double endTime = startTime + _riseTime + _asym * _riseTime;

    // the centre fo the Gaussian is start time plus rise time
    double centreTime = startTime + _riseTime;

    // the ADC values are calculated by intergating the gaussian inside the time bin and multiplying
    // with the ADCChargeFactor

    streamlog_out(DEBUG3) << "GaussianADCPulse::add st " << startTime << ", et " << endTime << ", ne " << nElectrons
                          << ", adc " << nElectrons* _ADCChargeFactor <<  ", f "  <<  _frequencyGHz  <<  std::endl;

    // start integration at start time for the first bin, the following bins start at the edge of the bin
    double integrationStartTime =  startTime;

    int bin = static_cast<int>(startTime * _frequencyGHz); // bin number fo the first bin with charge
    if(_startBin < 0) // if there is no charge in the pulse yet the pulse starts at "bin"
    {
        _startBin = bin;
    }

    // The stop time for the integration is the end of the bin.
    // If the end time is within the bin only integrate up to this point
    double integrationStopTime = std::min(endTime, (bin + 1) / _frequencyGHz);

    std::map<int, double>::iterator it;
    if(_useJitter)
    {
        it = _timeJitter.find(bin);
        if(it == _timeJitter.end())
            throw lcio::Exception("GaussianADCPulse: Error: Jitter map does not contain a value for this time bin!") ;
        else
        {
            integrationStopTime += it->second;
            streamlog_out(DEBUG1) << "bin " << bin << ", jitter " << it->second << ", integrationStopTime " << integrationStopTime << std::endl;
        }
    }
    else
        streamlog_out(DEBUG1) << "bin " << bin << ", integrationStopTime " << integrationStopTime << std::endl;


    double ADCChargeValue = nElectrons * _ADCChargeFactor;

    do
    {
        // calculate the ADC value for this bin by integrating the gaussian from
        // integrationStartTime to integrationStopTime with a centre of the gaussian at centreTime

        double ADCValue = ADCChargeValue / 2.*(erf((integrationStopTime  - centreTime) / _sqrt2_sigma)
                                               - erf((integrationStartTime - centreTime) / (_asym * _sqrt2_sigma)));

        // debug: check for 0
        if(ADCValue == 0)
        {
            streamlog_out(DEBUG3) << "integration from " << integrationStartTime
                                  << " to " << integrationStopTime
                                  << " resulted in zero" << std::endl;
        }

        // check if the bin already exists in the vector, else append it
        if((bin - _startBin) < static_cast<int>(_rawData.size()))
        {
            // time bin is already there, add the charge
            _rawData[bin - _startBin] += ADCValue;
        }
        else // the time bin did not exist yet
        {
            _rawData.push_back(ADCValue);
        }

        // this is done later to get the full info when debugging
        //   // perform ADC out of range check
        //   if (_rawData[bin-_startBin] > _ADCMaxValue)
        //   {
        //     _rawData[bin-_startBin] = _ADCMaxValue;
        //   }

        // move to the next time bin
        integrationStartTime = integrationStopTime;
        bin++;
        integrationStopTime = std::min(endTime, (bin + 1) / _frequencyGHz);

        // take clock jitter into account if chosen
        //    if (_useJitter){
        // adding negative jitter to the end time could result in running the while-loop once too often
        // actually, should integrationStopTime be set to the endTime - what happens to the rest of that bin?
        if(_useJitter && integrationStopTime != endTime)
        {
            it = _timeJitter.find(bin);
            if(it == _timeJitter.end())
                throw lcio::Exception("GaussianADCPulse: Error: Jitter map does not contain a value for this time bin!") ;
            else
            {
                streamlog_out(DEBUG1) << "bin " << bin << ", jitter " << it->second << ", integrationStopTime " << integrationStopTime << std::endl;
                integrationStopTime += it->second;
            }
        }
        else
        {
            streamlog_out(DEBUG1) << "bin "  << bin << ", integrationStopTime " << integrationStopTime << std::endl;
        }

    }
    while(integrationStartTime < endTime);

}



}// namespace marlintpc
