#ifndef ADCPulseConverterProcessor_h
#define ADCPulseConverterProcessor_h 1

//C++
#include <string>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerData.h"
#include "EVENT/TrackerPulse.h"
#include "IMPL/LCCollectionVec.h"
//Marlin
#include "marlin/Processor.h"


namespace marlintpc
{

/** Convert zero-suppressed ADC raw data to pulses.
 *
 *  Runs on all channels without applying an additional threshold, no pedestal information needed.
 *  No pre- and afterpulses are considered as the threshold is 0,
 *  so everything that is there is put into the pulse.
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li Collection of zero suppressed TrackerData
 *  \li The correct readoutfrequency set in the GEAR TPCParameters
 *
 *  <h4>Output</h4>
 *  Collection of TrackerPulses
 *  //Collection of TrackerData which belong to a pulse
 *
 *  @param InputCollectionName name of the input TrackerData collection
 *  @param OutputTrackerPulseCollectionName name of the output TrackerPulse collection
 *  @param // OutputTrackerDataCollectionName  name of the output TrackerData collection
 *  @param ForceSpectrumSave if not 0 the ADC spectrum will be stored for all pulses
 *
 *  @author Martin Killenberg, University of Bonn
 */
class ADCPulseConverterProcessor : public marlin::Processor
{

    public:

        virtual Processor*  newProcessor()
        {
            return new ADCPulseConverterProcessor ;
        }

        /// constructor
        ADCPulseConverterProcessor() ;

        virtual void init() ;

        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        virtual void processEvent(lcio::LCEvent * evt) ;

      // FIXME: not needed:  virtual void check(lcio::LCEvent * evt) ;

        virtual void end() ;


    protected:

        //  int _nRun ; ///< run counter
        //  int _nEvt ; ///< event counter

        std::string  _inputColName;         ///< name of the collection, where input data are stored
        //  std::string  _outputDataColName;    ///< name of the collection, where TrackerData are stored
        std::string  _outputPulseColName;   ///< name of the collection, where TrackerPulse are stored

        /** set the status which is set to the transient flag of the output collection
         *  it is an int instead of a boolean, because the processor parameter cannot be a boolean
         */
        int _outputIsTransient;

        /** if not 0 the ADC spectrum will be saved for all pulses
         *  instead of only for pulses with quality !=0
         *  it is an int instead of a boolean, bcause the processor parameter cannot be a boolean
         */
        int _forceSpectrumSave;

        /* the minmal bin per pulse
        * = _saveNBinsBeforeStart + _saveNBinsAfterStart
        */
        int   _minPulseDataSize;

        // this has to be taken in dependence on the module, cannot be a global parameter
        //  double _ReadoutFrequency;    ///< readout frequency in Hz (node: the time will be stored in ns)

        /** find pulses by threshold in the ADC information
         *  @param thisData 'raw' data (non zero supressed, can also handle zero supressed data)
         */
        int   findPulses(lcio::TrackerData* thisData, lcio::LCCollectionVec* thisCollection,
                         float StartThreshold);

        /** The time is just the centre of gravity of the pulse.
         */
        float calcPulseTime(lcio::TrackerData* thisPulse);

        /** calculate the integrated charge of the given pulse
         *  by sum up all bins of the ADC spectrum
         */
        float calcPulseCharge(lcio::TrackerData* thisPulse);

};

}// namespace marlintpc
#endif //ADCPulseConverterProcessor_h
