#ifndef PedestalSubtractorProcessor_h
#define PedestalSubtractorProcessor_h 1

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "marlin/Exceptions.h"

// MarlinTPC
#include "Pedestal.h"
#include "PedestalListener.h"
#include "ADCElectronicsListener.h"

// C++
#include <string>



namespace marlintpc
{

/** Pedestal subtracting processor of the MarlinTPC package.
 *
 *  This processor subtracts pedestals provided via LCCD (simple file, database file or database).
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TrackerData
 *
 *  <h4>Output</h4>
 *  Collection of TrackerData.
 *
 * @param InputCollectionName Name of input data collection
 * @param OutputCollectionName Name of output data collection
 * @param WriteOutputToStorage:bool output collection will be written if set to true (which is the default)
 * @param PedestalCollectionName Name of the collection containing the pedestals
 * @param MaximumADCValueOverride Maximum ADC Value of the read-out electronic to detect overflow samples, if this parameter is set it overides condition information and no condition listener is used (optional parameter, if not set conditions data is used)
 * @param PedestalValueOverride Pedestal of the read-out electronic to detect overflow samples, if this parameter is set it overides condition information and no condition listener is used (optional parameter, if not set, conditions data is used)
 * @param MissingConditionsDataOverrideTimespan Sets the time in seconds that missing conditions data will be ignored by using the last valid conditions data collection (optional parameter, if not set, processing stops if no conditions data available for event time)
 * @param ElectronicsParametersObjectName Name of the conditions object in which the parameters of the read out electronics is stored. For example "GenericADCElectronicsParameters".
 * @param ElectronicsParametersCollectionName Name of the conditions collection in which the parameters of the read out electronics is stored
**/
/**
*
* @author M. E. Janssen, DESY; R. Diener, DESY
*/

class PedestalSubtractorProcessor : public marlin::Processor
{

    public:

        virtual Processor*  newProcessor()
        {
            return new PedestalSubtractorProcessor ;
        }

        PedestalSubtractorProcessor() ;

        virtual void init() ;

        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        virtual void processEvent(lcio::LCEvent * evt) ;

        virtual void check(lcio::LCEvent * evt) ;

        virtual void end() ;


    protected:

        /// the name of the input collection
        std::string _inputColName ;
        /// the name of the output collection
        std::string _outputColName ;
        /// the name of the pedestal collection
        std::string _pedestalColName ;
        /// the name of the collection with the read out electronics parameters
        std::string _electrParaObjName ;
        /// the name of the collection with the read out electronics parameters
        std::string _electrParaColName ;

        /// set the persistency flag of the output collection
        bool _outputIsPersistent;

        /** The maximum adc value, needed to determine if a pulse is maxed out. This overrides any information provided by LCCD
         */
        int _maximum_ADC_value_override;

        /** The pedestal value to be used for every hardware channel in every readout group. This overrides any information provided by LCCD
         */
        float _pedestal_value_override;

        /** The timepan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.
         */
        int _missingConditionsDataOverrideTimespan;

    private:

        PedestalListener* _pedestalListener;
        ADCElectronicsListener* _adcElectronicsListener;

        /// Internal member to hold the time for the missing conditions overide. Unit: nanoseconds.
        lcio::long64 _missingConditionsDataOverrideLCIOTime;
} ;
}
#endif
