#include "FieldSettingListener.h"

//C++
#include <utility>
#include <map>
#include <iostream>
#include <string.h>
#include <exception>
#include <sstream>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd_exceptions.h>

//lcio
#include "lcio.h"
#include "EVENT/LCIO.h"
#include "UTIL/LCFixedObject.h"
#include "UTIL/LCTime.h"

//MarlinTPC
#include "TPCCondDataExeptions.h"

namespace marlintpc
{

FieldSettingListener::FieldSettingListener(std::string ColName)
{
    //initialize first the members, else it goes wrong
    _collectionName = ColName;
    _thisEmptyCollection = new LCCollectionVec(LCIO::LCGENERICOBJECT);
    _thisEmptyCollection->parameters().setValue("CollectionName", "EmptyCollectionFieldSettingListener");
    _timeStampSince = 0;
    _timeStampTill = 0;
    _fieldSettings = NULL;
    _useLastValidIfNoDataAvailable = false;
    _noDataAvailable = true;
    _condHandler = NULL;
    // try to register with the Marlin conditions processor
    if (!(marlin::ConditionsProcessor::registerChangeListener(this, ColName)))
        {
            std::stringstream message;
            message << " FieldSettingListener couldn't register collection " << ColName << " with the ConditionsProcessor.";
            throw lccd::LCCDException(message.str());
        }
}



FieldSettingListener::~FieldSettingListener()
{
    delete _thisEmptyCollection;
}



void FieldSettingListener::conditionsChanged(EVENT::LCCollection* col)
{
    streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged()" << std::endl;

    // get collection time parameters and set internal members
    std::stringstream sstrSince(col->getParameters().getStringVal(lccd::DBSINCE));
    sstrSince >> _timeStampSince;
    lcio::LCTime timeSince(_timeStampSince) ;
    std::stringstream sstrTill(col->getParameters().getStringVal(lccd::DBTILL));
    sstrTill >> _timeStampTill;
    lcio::LCTime timeTill(_timeStampTill) ;
    streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged(): _timeStampSince=" << _timeStampSince << " _timeStampTill=" << _timeStampTill << std::endl;

    // look into the map to see if we have accepted this collection as a default
    std::map<lcio::LCCollection* , lccd::IConditionsHandler* >::iterator itDefaultCollectionMap = _handlerDefaultCollectionMap.find(col);

    // check if the collection is our default collection
    if (itDefaultCollectionMap != _handlerDefaultCollectionMap.end())
        {
            _noDataAvailable = true;

            streamlog_out(MESSAGE) << "FieldSettingListener::conditionsChanged(): default collection " << col->getParameters().getStringVal("CollectionName") << " sent " << std::endl;
            streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged(): address of default collection: " << col << std::endl;

            //give some information on the missing conditions data
            streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged(): CollectionName: " << _collectionName << " not available at this timestamp." << std::endl;
            streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged(): No data in database from: " << timeSince.timeStamp() << " (" << timeSince.year() << " " << timeSince.month() << " " << timeSince.day() << " " << timeSince.hour() << " " << timeSince.min() << " " << timeSince.sec() << " " << timeSince.ns() << ")" << std::endl;
            streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged(): No data in database till: " << timeTill.timeStamp() << " (" << timeTill.year() << " " << timeTill.month() << " " << timeTill.day() << " " << timeTill.hour() << " " << timeTill.min() << " " << timeTill.sec() << " " << timeTill.ns() << ")" << std::endl;

            //check what to do if there are not data available
            if (_useLastValidIfNoDataAvailable)
                {
                    // check if there has been a last valid collection
                    if (_condHandler->lastValidCollection() != NULL)
                        {
                            // there should be only one object in the collection:
                            if (col->getNumberOfElements() > 1)
                                {
                                    std::stringstream out;
                                    out << "FieldSettingListener::conditionsChanged(): Found more than one object (#=" << col->getNumberOfElements() << ") in the conditions collection, although it should be only one.";
                                    throw lccd::InconsistencyException(out.str());
                                }
                            else
                                {
                                    _fieldSettings = (dynamic_cast<tpcconddata::FieldSetting*>((_condHandler->lastValidCollection())->getElementAt(0))) ;
                                }
                        }
                    else
                        {
                            // no last valid collection available, so we bail out by throwing an exception
                            std::stringstream out;
                            out << "FieldSettingListener::conditionsChanged(): Tried to use last valid collection but found none. Exiting processing";
                            throw lccd::DataNotAvailableException(out.str());
                        }
                }
            else
                {
                    // no conditions data available, so we bail out by throwing an exception
                    std::stringstream out;
                    out << "FieldSettingListener::conditionsChanged(): No conditions data stored for current time stamp. Exiting processing";
                    throw lccd::DataNotAvailableException(out.str());
                }
        }
    else                     // it is not a default so we can do the updating
        {
            _noDataAvailable = false;

            streamlog_out(MESSAGE) << "FieldSettingListener::conditionsChanged(): new values available. Updating parameter map." << std::endl;
            streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged(): CollectionName: " << col->getParameters().getStringVal("CollectionName") << std::endl;
            streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged(): Valid Since: " << col->getParameters().getStringVal("DBSince") << std::endl;
            streamlog_out(DEBUG) << "FieldSettingListener::conditionsChanged(): Valid Till: " << col->getParameters().getStringVal("DBTill") << std::endl;

            // there should be only one object stored in the collection
            if (col->getNumberOfElements() > 1)
                {
                    std::stringstream out;
                    out << "FieldSettingListener::conditionsChanged(): Found more than one object (#=" << col->getNumberOfElements() << ") in the conditions collection, although it should be only one.";
                    throw lccd::InconsistencyException(out.str());
                }
            else
                {
                    _fieldSettings = new tpcconddata::FieldSetting(dynamic_cast<LCGenericObject*>(col->getElementAt(0)));
                }
        }
}



tpcconddata::FieldSetting* FieldSettingListener::getFieldSetting()
{
    return _fieldSettings;
}

int FieldSettingListener::getNGEMs() const
{
    return _fieldSettings->getNGEMs();
}

int FieldSettingListener::getGEMStatus() const
{
	return _fieldSettings->getGEMStatus();
}

int FieldSettingListener::getDriftFieldStatus() const
{
	return _fieldSettings->getDriftFieldStatus();
}

int FieldSettingListener::getMagneticFieldStatus() const
{
	return _fieldSettings->getMagneticFieldStatus();
}

float FieldSettingListener::getGEMVoltage( int GEMID ) const
{
	return _fieldSettings->getGEMVoltage( GEMID);
}

float FieldSettingListener::getTransferField( int GEMID ) const
{
	return _fieldSettings->getTransferField( GEMID );
}

float FieldSettingListener::getInductionField() const
{
	return _fieldSettings->getInductionField();
}

float FieldSettingListener::getGuardRingVoltage() const
{
	return _fieldSettings->getGuardRingVoltage();
}

float FieldSettingListener::getCathodeVoltage() const
{
	return _fieldSettings->getCathodeVoltage();
}

float FieldSettingListener::getAnodeVoltage() const
{
	return _fieldSettings->getAnodeVoltage();
}

float FieldSettingListener::getDriftField( float maxDriftLegth ) const
{
	return _fieldSettings->getDriftField( maxDriftLegth );
}

float FieldSettingListener::getMagneticField() const
{
	return _fieldSettings->getMagneticField();
}


void FieldSettingListener::registeredWithHandler(lccd::IConditionsHandler* condHandler)
{
    _condHandler = condHandler;
    streamlog_out(MESSAGE) << "FieldSettingListener::registeredWithHandler(): registered: " << condHandler->name() << std::endl;
    streamlog_out(DEBUG) << "FieldSettingListener::registeredWithHandler(): register default collection:" << std::endl;
    // try to get the default collection
    try
        {
            LCCollection* col = condHandler->defaultCollection();

            if (! col)           // it will be null if none has so far been registered. So let's register ours
                {
                    condHandler->registerDefaultCollection(_thisEmptyCollection);
                    streamlog_out(DEBUG) << "FieldSettingListener::registeredWithHandler(): default collection registered; address: " << _thisEmptyCollection << std::endl;
                    // and put in the map for this handler
                    _handlerDefaultCollectionMap[_thisEmptyCollection] = condHandler;
                }
            //then the default handler was already registered, that's odd ... ;)
            else if (col == _thisEmptyCollection)
                {
                    streamlog_out(DEBUG) << "FieldSettingListener::registeredWithHandler(): default collection is already set to thisEmptyCollection:" << std::endl;
                }
            else                 // somebody has got there before us, let's see if we like the default
                {
                    // if it has the same CollectionName, we can use it, no problem ...
                    if (col->getParameters().getStringVal("CollectionName") == "EmptyCollectionFieldSettingListener")
                        {
                            streamlog_out(DEBUG) << "FieldSettingListener::registeredWithHandler(): I like your default ;)" << std::endl;
                            _handlerDefaultCollectionMap[col] = condHandler;
                        }
                    // if not, an exception is thrown because things should be handled in a uniform standardized way
                    else
                        {
                            streamlog_out(ERROR) << "FieldSettingListener::registeredWithHandler(): a non-standard empty default collection has already been registered with the conditions handler" << std::endl;
                            throw lccd::InconsistencyException("FieldSettingListener::registeredWithHandler() found a default collection that differs from the standard one has already been registered with the handler");
                        }
                }
        }
    // catch MemberNotImplemented exception, which happens, if a simple file is used as input
    catch (lccd::MemberNotImplementedException&)
        {
            streamlog_out(MESSAGE) << "FieldSettingListener::registeredWithHandler: Could not register default collection. Seems simple file input is used for pedestal conditions." << std::endl;
        }
}



void FieldSettingListener::deRegisteredWithHandler(lccd::IConditionsHandler* condHandler)
{
    streamlog_out(WARNING) << "FieldSettingListener::deRegisteredWithHandler() deregistered with:" << condHandler->name() << std::endl;
    _condHandler = NULL;
}



lcio::long64 FieldSettingListener::collectionValidSince()
{
    return _timeStampSince;
}



lcio::long64 FieldSettingListener::collectionValidTill()
{
    return _timeStampTill;
}



void FieldSettingListener::useLastValidIfNoDataAvailable(bool useLastValid)
{
    _useLastValidIfNoDataAvailable = useLastValid;
}



bool FieldSettingListener::noConditionsDataAvailable()
{
    return _noDataAvailable;
}


}
