/*
 * ModuleDistortionCalculator.cc
 *
 *  Created on: June 4, 2013
 *      Author: , amunnich
 */


#include "ModuleDistortionCalculator.h"
#include <cmath>
#include "EVENT/TrackerPulse.h"
#include "gear/TPCParameters.h"
#include "gear/TPCModule.h"
#include "IntersectionCalculator.h"
#include "streamlog/streamlog.h"

namespace  marlintpc
{
  ModuleDistortionCalculator::ModuleDistortionCalculator():
		  _GearParameters(nullptr)
  {
  }

  ModuleDistortionCalculator::ModuleDistortionCalculator(const gear::TPCParameters* tpcParameters):
    _GearParameters(tpcParameters)
  {
  }
  
  double ModuleDistortionCalculator::calculateDistanceRow(lcio::TrackerHit *hit, lcio::Track *track)
  {
    const gear::TPCModule& theModule = _GearParameters->getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
    gear::Vector2D hitLocal = theModule.globalToLocal(hit->getPosition()[0],hit->getPosition()[1]);
    double distrow=0.;
    if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)//if circular pad layout
      {
	//get radius and center of measurement layer in global coordinates
	gear::Vector2D padCenterGlobal = theModule.getPadCenter(theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]));
	gear::Vector2D padCenterLocal = theModule.globalToLocal(padCenterGlobal[0],padCenterGlobal[1]);
	
	streamlog_out(DEBUG)<<"Polar Coordinates"<<std::endl;
	double Radius = padCenterLocal[0];
	
	//distinguish between straight track and helix track
	std::pair<double,double> phi_intersec(0,0); 
	if(fabs(track->getOmega())<0.000000001)
	  {
	    try{
	      //calculate two points on the track and transform them into local padplane coordinates to obtain a track parametrisation
	      // in local coordionates
	      //point1=point of closest approach
	      double pca_x = ( -1.0 * track->getD0() ) * std::sin(track->getPhi())+track->getReferencePoint()[0];
	      double pca_y = ( track->getD0() ) * std::cos(track->getPhi())+track->getReferencePoint()[1];
	      //point 2, if we are on a straight line just move one step along the line
	      double p2x=pca_x+std::cos(track->getPhi());
	      double p2y=pca_y+std::sin(track->getPhi());
	      //transform both points to local coordinate system
	      //this gives back r and phi, not x and y for polar padlayout!
	      gear::Vector2D pca_local = theModule.globalToLocal(pca_x,pca_y);
	      gear::Vector2D p2_local = theModule.globalToLocal(p2x,p2y);
	      double pca_local_x=pca_local[0]*cos(pca_local[1]);
	      double pca_local_y=pca_local[0]*sin(pca_local[1]);
	      double p2_local_x=p2_local[0]*cos(p2_local[1]);
	      double p2_local_y=p2_local[0]*sin(p2_local[1]);
	      double phi=atan2(p2_local_y-pca_local_y,p2_local_x-pca_local_x);
	      phi_intersec= marlintpc::IntersectionCalculator::circle_straightLine(Radius, pca_local_x,pca_local_y,phi);
	    }
	    catch(std::exception &e) {
	      std::cout<<"NO INTERSECTION STRAIGHT LINE: " <<Radius<<" "
		       <<track->getOmega()<<" "<< track->getD0()<<" "<<track->getPhi()<<std::endl;
	      //throw it again to be availbale in code that uses function
	      throw e;
	    }
	  }
	else
	  {
	    //center of helix in global coordinates
	    double xm, ym;
	    xm = track->getReferencePoint()[0]+(1. / track->getOmega() - track->getD0()) * sin(track->getPhi());
	    ym = track->getReferencePoint()[1]-(1. / track->getOmega()  - track->getD0()) * cos(track->getPhi());
	    //transform helix center into local coordinate system
	    gear::Vector2D helixCenterLocal = theModule.globalToLocal(xm,ym);
	    //local comes on in polar, go back to cartesian
	    double xm_l=helixCenterLocal[0]*cos(helixCenterLocal[1]);
	    double ym_l=helixCenterLocal[0]*sin(helixCenterLocal[1]);
	    
	    try{
	      phi_intersec= marlintpc::IntersectionCalculator::circle_circle(Radius, xm_l, ym_l,fabs(1. / track->getOmega()));
	    }
	    catch(std::exception &e) {
	      std::cout<<"NO INTERSECTION CIRCLE: " <<Radius<<" "<<xm_l<<" "<< ym_l<<" : "
		       <<track->getOmega()<<" "<< track->getD0()<<" "<<track->getPhi()<<std::endl;
	      //throw it again to be availbale in code that uses function
	      throw e;
	    }
	  }
	distrow=(hitLocal[1]-phi_intersec.first)*Radius;
	if(fabs((hitLocal[1]-phi_intersec.second))*Radius<fabs(distrow))
	  distrow=(hitLocal[1]-phi_intersec.second)*Radius;
        }
    else if((theModule.getLocalPadLayout()).getPadLayoutType() == 1)//rectangular
      {
	streamlog_out(DEBUG)<<"Rectangular pad layout"<<std::endl;
	std::pair<double,double> intersec(0,0); 
	//get y and center of measurement layer in global coordinates
	gear::Vector2D padCenterGlobal = theModule.getPadCenter(theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]));
	gear::Vector2D padCenterLocal = theModule.globalToLocal(padCenterGlobal[0],padCenterGlobal[1]);
	
	double Ylocal = padCenterLocal[1];   //center of pad row
	//distinguish between straight track and helix track
	if(fabs(track->getOmega())<0.000000001)
	  {
	    // 2 points on the track in local coordinates
	    //point1=point of closest approach
	    double pca_x = ( -1.0 * track->getD0() ) * std::sin(track->getPhi())+track->getReferencePoint()[0];
	    double pca_y = ( track->getD0() ) * std::cos(track->getPhi())+track->getReferencePoint()[1];
	    //point 2, if we are on a straight line just move one step along the line
	    double p2x=pca_x+std::cos(track->getPhi());
	    double p2y=pca_y+std::sin(track->getPhi());
	    //transform both points to local coordinate system
	    gear::Vector2D p2_local = theModule.globalToLocal(p2x,p2y);
	    gear::Vector2D pca_local = theModule.globalToLocal(pca_x,pca_y);
	    //double inclination=atan2(p2_local_y-pca_local_y,p2_local_x-pca_local_x);
	    double inclination=(p2_local[0]-pca_local[0])/(p2_local[1]-pca_local[1]);	    
	    double xAxisIntersect=pca_local[0]-inclination*pca_local[1];
	    
	    double x_intersec= marlintpc::IntersectionCalculator::horizontalLine_straightLine(Ylocal,inclination, xAxisIntersect);
	    distrow= hitLocal[0]-x_intersec;
	  }	
	else
	  {
	    double xm, ym;
	    xm = track->getReferencePoint()[0]+(1. / track->getOmega() - track->getD0()) * sin(track->getPhi());
	    ym = track->getReferencePoint()[1]-(1. / track->getOmega()  - track->getD0()) * cos(track->getPhi());
	    //transform helix center into local coordinate system
	    gear::Vector2D helixCenterLocal = theModule.globalToLocal(xm,ym);
	    intersec= marlintpc::IntersectionCalculator::horizonalLine_circle(Ylocal, helixCenterLocal[0],helixCenterLocal[1] , fabs(1. / track->getOmega()));
	    //intersection give the two x coordinates where circle intesects line. Coose the minimal distance.
	    distrow=hitLocal[0]-intersec.first;
	    if(fabs(hitLocal[0]-intersec.first)>fabs(hitLocal[0]-intersec.second))
	      distrow=hitLocal[0]-intersec.second;
	  }
      }
    else
      {
	streamlog_out(ERROR)<<"Error: Invalid pad layout type!"<<std::endl;
	throw gear::Exception("Unknown pad layout type");
      }

    return distrow;
  }
  
  double ModuleDistortionCalculator::calculateDistanceXY(lcio::TrackerHit *hit, lcio::Track *track)
  {
    Double_t distxy = 0.;
    if(fabs(track->getOmega())<0.000000001)
      {
	//added a minus to mkae it consistent with value from curved tracks and row result
	distxy =-( hit->getPosition()[0] * std::sin(track->getPhi())
		   - hit->getPosition()[1] * std::cos(track->getPhi()) + track->getD0());
      }
    else
      {
	double xm, ym;
	xm = (1. / track->getOmega() - track->getD0()) * sin(track->getPhi())+track->getReferencePoint()[0];
	ym = -(1. / track->getOmega()  - track->getD0()) * cos(track->getPhi())+track->getReferencePoint()[1];
	
	double d_p =   sqrt(
			    (hit->getPosition()[0] - xm) *
			    (hit->getPosition()[0] - xm) +
			    (hit->getPosition()[1] - ym) *
			    (hit->getPosition()[1] - ym)
			    )
	  - 1. / track->getOmega();
	double d_m =  (-1.) * sqrt(
				   (hit->getPosition()[0] - xm) *
				   (hit->getPosition()[0] - xm) +
				   (hit->getPosition()[1] - ym) *
				   (hit->getPosition()[1] - ym)
				   )
	  - 1. / track->getOmega();
	
	if (fabs(d_p) <= fabs(d_m)) distxy = d_p;
	else distxy = d_m;
	
      }
    
    return distxy;
    
  }  
  
  double ModuleDistortionCalculator::calculateDistanceZ(lcio::TrackerHit *hit, lcio::Track *track)
  {
    const Double_t PI = 3.14159265358979323846;
         
    double pca_x = ( -1.0 * track->getD0() ) * std::sin(track->getPhi())+track->getReferencePoint()[0];
    double pca_y = ( track->getD0() ) * std::cos(track->getPhi())+track->getReferencePoint()[1];

    Double_t s = 0.;
    
    if(fabs(track->getOmega()) < 0.000000001)
      {
	s = std::sqrt(
		      static_cast<long double>(( hit->getPosition()[0] - pca_x ) * ( hit->getPosition()[0] - pca_x )
					       + ( hit->getPosition()[1] - pca_y ) * ( hit->getPosition()[1] - pca_y )));
	
	//make sure the sign of s is right
	if(( ( ( -1.25 * PI ) <= track->getPhi() ) && ( track->getPhi() <= ( -0.75 * PI ) ) )
	   || ( ( ( -0.25 * PI ) <= track->getPhi() ) && ( track->getPhi() <= ( 0.25 * M_PI ) ) )
	   || ( ( ( 0.75 * PI ) <= track->getPhi() ) && track->getPhi() <= ( 1.25 * M_PI ) ))
	  {
	    if(pca_x <  hit->getPosition()[0])
	      {
		s = ( -1.0 ) * s;
	      }
	  }
	else
	  {
	    if(pca_y <  hit->getPosition()[1])
	      {
		s = ( -1.0 ) * s;
	      }
	  }
	s = (-1.)*s; //Needed to make tracks look correctly in CEDViewer 
      }
    else
      {
	Double_t r = 1. / track->getOmega();
	//center of circle
	Double_t xm, ym;
	xm = (r - track->getD0()) * sin(track->getPhi())+track->getReferencePoint()[0];
	ym = -(r - track->getD0()) * cos(track->getPhi())+track->getReferencePoint()[1];

	s = r * ( atan2(pca_y - ym , pca_x - xm)
		  -
		  atan2( hit->getPosition()[1]- ym ,  hit->getPosition()[0] - xm)
		  );
	
	double s_1, s_2;
	s_1 = s;
	if(s < 0)
	  {
	    s_2 = s + fabs(2.*M_PI/track->getOmega());
	  }
	else
	  {
	    s_2 = s - fabs(2.*M_PI/track->getOmega());
	  }
	
	if(fabs(s_1) <= fabs(s_2)) s = s_1;
	else s = s_2;
      }
    
    double distz =  track->getTanLambda() * s +  track->getZ0()+track->getReferencePoint()[2] - hit->getPosition()[2];
//    double distz =  track->getTanLambda() * s +  track->getZ0() - hit->getPosition()[2];




//std::cout<<"distz="<<distz<<std::endl;

//std::cout<<"track->getZ0()="<<track->getZ0()<<std::endl;
//std::cout<<"track->getReferencePoint()[2]="<<track->getReferencePoint()[2]<<std::endl;




    return -1. * distz;
    
  }  



}

