/*
 * ModuleDistortionCorrectionProcessor.cpp
 *
 *  Created on: July 11, 2013
 *      Author: amunnich
 */

#include "ModuleDistortionCorrectionProcessor.h"


//LCIO
#include "lcio.h"
#include "LCIOSTLTypes.h"
#include <IMPL/LCCollectionVec.h>
#include <EVENT/LCCollection.h>
#include "EVENT/TrackerPulse.h"
#include "IMPL/TrackerHitImpl.h"

//MARLIN
#include "marlin/Global.h"
#include "marlin/Exceptions.h"

//GEAR
#include "gear/TPCModule.h"
#include "gear/TPCParameters.h"

//ilcutil
#include "streamlog/streamlog.h"

//C++
#include <fstream>
#include <sstream>

using namespace std;
using namespace lcio;
using namespace gear;
using namespace marlin;

namespace marlintpc
{


ModuleDistortionCorrectionProcessor aModuleDistortionCorrectionProcessor;

ModuleDistortionCorrectionProcessor::ModuleDistortionCorrectionProcessor() :
	marlin::Processor( "ModuleDistortionCorrectionProcessor" )
{
	_description = "The ModuleDistortionCorrectionProcessor processor applies the correction of field distortion to all hits"
		"as calculated for example from ModuleDistortionCalculatorProcessor.";


	const std::string HitDataCollNameDesc_OUT =
		"Name of the output collection where the shifted hits are stored. "
			"Default is \"Hits_shifted\"";
	const std::string HitDataCollNameDefault_OUT = "Hits_shifted";
	registerOutputCollection( LCIO::TRACKERHIT,
				"HitDataCollName_OUT",
				HitDataCollNameDesc_OUT,
				_HitDataCollName_OUT,
				HitDataCollNameDefault_OUT );

	const std::string HitDataCollNameDesc =
		"Name of the input collection where the hits are stored. "
			"Default is \"TPCHits\"";
	const std::string HitDataCollNameDefault = "TPCHits";
	registerInputCollection( LCIO::TRACKERHIT,
				"InputCollections",
				HitDataCollNameDesc,
				_HitDataCollName,
				HitDataCollNameDefault );


	registerOptionalParameter( "MillepedeDistortionFileName",
				"Path to file containing distortions in row and in z and charge dependent calculated with Millepede; either this or at least one of the other distortion files must be given",
				_MillepedeDistortionFileName,
				std::string( "" ) );


	const std::string DistortionZFileNameDesc = "Path to file containing amount of distortion in z";
	const std::string DistortionZFileNameDefault = "";
	registerOptionalParameter( "DistortionZFileName",
				DistortionZFileNameDesc,
				_DistortionZFileName,
				DistortionZFileNameDefault );

	const std::string DistortionRowFileNameDesc = "Path to file containing amount of distortion in row";
	const std::string DistortionRowFileNameDefault = "";
	registerOptionalParameter( "DistortionRowFileName",
				DistortionRowFileNameDesc,
				_DistortionRowFileName,
				DistortionRowFileNameDefault );
}


ModuleDistortionCorrectionProcessor::~ModuleDistortionCorrectionProcessor()
{
}

void
ModuleDistortionCorrectionProcessor::init()
{
	printParameters();

	if( !( parameterSet( "MillepedeDistortionFileName" ) || parameterSet( "DistortionZFileName" ) || parameterSet( "DistortionRowFileName" ) ) )
	{
		streamlog_out( ERROR ) << "No distortion file given!" << endl;
		throw StopProcessingException(this);
	}

	_GearParameters = &marlin::Global::GEAR->getTPCParameters();

	//vectors by which points should be shifted, read in from files
	if( parameterSet( "MillepedeDistortionFileName" ) )
	{
		ifstream distIn;
		string lineIn;
		distIn.open( _MillepedeDistortionFileName.c_str() );
		if( !distIn.good() )
		{
			streamlog_out( ERROR ) << "Distortion file \"" << _MillepedeDistortionFileName << "\" could not be opened!" << endl;
			throw StopProcessingException(this);
		}
		int nModules = _GearParameters->getNModules();
		int label, module, row;
		double distortion, presigma, diff, distortionerr;
		int line = 0;
		streamlog_out( DEBUG4 ) << "Reading Millepede distortions." << endl;
		while( distIn.good() )
		{
			getline( distIn, lineIn );
			stringstream lineInStream ( lineIn );
			if( !( lineInStream >> label >> distortion >> presigma >> diff >> distortionerr ) )
			{
				streamlog_out( DEBUG4 ) << "Skipping line" << endl;
				continue;
			}
			streamlog_out( DEBUG4 ) << "Reading Line " << line << ": " << label << " ";
			if( label<=100*nModules )
			{
				row = label%100 - 1;
				module = ( label - row - 1 ) / 100;
				streamlog_out( DEBUG4 ) << module << " " << row << " ";
				std::pair<int, int> key = std::make_pair( module, row );
				std::pair<float, float> valueRow = std::make_pair( distortion, distortionerr );
				_RowDistortionCorrection.insert( std::make_pair( key, valueRow ) );
			}
			else if( label<=2*100*nModules )
			{
				row = label%100 - 1;
				module = ( label - 100*nModules - row - 1 ) / 100;
				streamlog_out( DEBUG4 ) << module << " " << row << " ";
				std::pair<int, int> key = std::make_pair( module, row );
				std::pair<float, float> valueZ = std::make_pair( distortion, distortionerr );
				_ZDistortionCorrection.insert( std::make_pair( key, valueZ ) );
			}
			else if( label<=2*100*nModules+100 )
			{
				int bin = label - 2*100*nModules;
				streamlog_out( DEBUG4 ) << bin << " ";
				std::pair<float, float> valueRow = std::make_pair( distortion, distortionerr );
				_RowChargeCorrection.push_back( valueRow );
			}
			else
			{
				int bin = label - 2*100*nModules - 100;
				streamlog_out( DEBUG4 ) << bin << " ";
				std::pair<float, float> valueZ = std::make_pair( distortion, distortionerr );
				_ZChargeCorrection.push_back( valueZ );
			}
			streamlog_out( DEBUG4 ) << distortion << " " << distortionerr << std::endl;
			++line;
		}
		if( distIn.bad() || line==0 )
		{
			streamlog_out( ERROR ) << "Error while reading the distortions file \"" << _MillepedeDistortionFileName << "\"!" << endl;
			throw StopProcessingException(this);
		}
		distIn.close();
		streamlog_out( DEBUG4 ) << "Read " << line << " lines of distortion data containing:" << endl;
		streamlog_out( DEBUG4 ) << "row distortions for " << _RowDistortionCorrection.size() << " rows" << endl;
		streamlog_out( DEBUG4 ) << "z distortions for " << _ZDistortionCorrection.size() << " rows" << endl;
		streamlog_out( DEBUG4 ) << _RowChargeCorrection.size() << " values for row charge corrections" << endl;
		streamlog_out( DEBUG4 ) << _ZChargeCorrection.size() << " values for z charge corrections" << endl;
	}

	if( parameterSet( "DistortionZFileName" ) && !parameterSet( "MillepedeDistortionFileName" ) )
	{
		std::ifstream inz;
		inz.open( _DistortionZFileName.c_str() );
		if( !inz.good() )
		{
			streamlog_out( ERROR ) << "Distortion file \"" << _DistortionZFileName << "\" could not be opened!" << endl;
			throw StopProcessingException(this);
		}
		int module;
		int row;
		float distortion;
		float distortionerr;
		int linez = 0;
		streamlog_out( DEBUG4 ) << "Reading z distortions: " << std::endl;
		while( 1 )
		{
			if ( !inz.good() ) break;
			inz >> module >> row >> distortion >> distortionerr;
			streamlog_out( DEBUG4 ) << "Reading Line " << linez << ": " << module << " "
				<< row << " " << distortion << " " << distortionerr << std::endl;

			std::pair<int, int> key = std::make_pair( module, row );
			std::pair<float, float> valuez = std::make_pair( distortion, distortionerr );
			_ZDistortionCorrection.insert( std::make_pair( key, valuez ) );
			linez++;
		}
		if( inz.bad() )
		{
			streamlog_out( ERROR ) << "Error while reading the distortions file \"" << _DistortionZFileName << "\". Aborting!" << endl;
			throw StopProcessingException(this);
		}
		inz.close();
	}
	else if( parameterSet( "DistortionZFileName" ) )
		streamlog_out( DEBUG4 ) << "Ignoring DistortionZFileName since MillepedeDistortionFileName is given." << endl;



	if( parameterSet( "DistortionRowFileName" ) && !parameterSet( "MillepedeDistortionFileName" ) )
	{
		std::ifstream inrow;
		inrow.open( _DistortionRowFileName.c_str() );
		if( !inrow.good() )
		{
			streamlog_out( ERROR ) << "Distortion file \"" << _DistortionRowFileName << "\" could not be opened!" << endl;
			throw StopProcessingException(this);
		}
		int module;
		int row;
		float distortion;
		float distortionerr;
		int linerow = 0;
		streamlog_out( DEBUG4 ) << "Reading row distortions: " << std::endl;
		while( 1 )
		{
			if ( !inrow.good() ) break;
			inrow >> module >> row >> distortion >> distortionerr;
			streamlog_out( DEBUG4 ) << "Reading Line " << linerow << ": " << module << " "
				<< row << " " << distortion << " " << distortionerr << std::endl;

			std::pair<int, int> key = std::make_pair( module, row );
			std::pair<float, float> valuerow = std::make_pair( distortion, distortionerr );
			_RowDistortionCorrection.insert( std::make_pair( key, valuerow ) );
			linerow++;
		}
		if( inrow.bad() )
		{
			streamlog_out( ERROR ) << "Error while reading the distortions file \"" << _DistortionRowFileName << "\". Aborting!" << endl;
			throw StopProcessingException(this);
		}
		inrow.close();
	}
	else if( parameterSet( "DistortionRowFileName" ) )
		streamlog_out( DEBUG4 ) << "Ignoring DistortionRowFileName since MillepedeDistortionFileName is given." << endl;
}

void
ModuleDistortionCorrectionProcessor::processRunHeader( LCRunHeader* run )
{
	run->parameters().setValue( _processorName + "_revision", "$Rev: 170 $" );

	for ( marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ )
	{
		if ( !i->second->isOptional() || i->second->valueSet() )
			run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
	}
	streamlog_out( MESSAGE ) << "Run Header processed" << std::endl;
}

void
ModuleDistortionCorrectionProcessor::processEvent( LCEvent * evt )
{
 LCCollection* hitCollection;

 try{
	hitCollection = evt->getCollection( _HitDataCollName );
 }
 catch ( DataNotAvailableException &e )
	{
	streamlog_out( DEBUG ) << "No data available in event for " << _HitDataCollName << std::endl;
	return;
	}
 _AllShiftedHits.clear();
 int EventNumber = evt->getEventNumber();
 int hitNum = hitCollection->getNumberOfElements();
 //loop the hits
 for( int i = 0; i < hitNum; ++i )
	{
	//find out the pad and module number
	lcio::TrackerHit* theHit = dynamic_cast<lcio::TrackerHit*>( hitCollection->getElementAt( i ) );
	const double* HitPosition = theHit->getPosition();
	lcio::LCObjectVec PulseVect = theHit->getRawHits();
	if( PulseVect.size() == 0 )
	{
		streamlog_out( WARNING ) << "WARNING: In Event " << EventNumber << " for Hit "
					<< i
					<< ": The vector that should contain the pulse list is empty."
					<< "Impossible to decode row and module number " << std::endl;
		_AllShiftedHits.push_back( theHit );
		continue;
	}

	lcio::TrackerPulse const * thePulse = dynamic_cast<lcio::TrackerPulse const *>( *( PulseVect.begin() ) );
	if( thePulse == 0 )
	{
		streamlog_out( WARNING ) << "WARNING: In Event " << EventNumber << " for Hit on track "
					<< i
					<< ": The raw hits do not contain valid TrackerPulse objects."
					<< "Impossible to decode row and module number " << std::endl;
		_AllShiftedHits.push_back( theHit );
		continue;
	}

	double maxPulseCharge = -9e9;
	for( lcio::LCObjectVec::iterator pulseIter=PulseVect.begin() ; pulseIter!=PulseVect.end() ; ++pulseIter )
	{
		lcio::TrackerPulse const * aPulse = dynamic_cast<lcio::TrackerPulse const *>( *pulseIter );
		if( maxPulseCharge < aPulse->getCharge() )
		{
			maxPulseCharge = aPulse->getCharge();
			thePulse = aPulse;
		}
	}

	int Module = thePulse->getCellID1();
	std::pair<int, int> RowModuleID = std::make_pair( -1, Module );
	double Radius = 0.;
	double NewHitPosition[3];
	if( _GearParameters == 0 )
	{
		RowModuleID.first = ( 0xffff0000 & ( thePulse->getCellID0() ) ) >> 16;
	}
	else
	{
		const gear::TPCModule& theModule = _GearParameters->getModule( thePulse->getCellID1() );
		RowModuleID.first = theModule.getRowNumber( thePulse->getCellID0() );
		gear::Vector2D padCenterGlobal = theModule.getPadCenter( thePulse->getCellID0() );
		streamlog_out( DEBUG ) << "pad center global: " << padCenterGlobal[0] << " " << padCenterGlobal[1] << std::endl;

		//transform pad centers from local to global coordinates
		gear::Vector2D padCenterLocal = theModule.globalToLocal( padCenterGlobal[0], padCenterGlobal[1] );
		streamlog_out( DEBUG ) << "pad center local: " << padCenterLocal[0] << " " << padCenterLocal[1] << std::endl;

		//check if local coordinate system is cartesian or polar
		if( ( theModule.getLocalPadLayout() ).getPadLayoutType() == 1 )
		{
			//cartesian
			streamlog_out( DEBUG ) << "Cartesian Coordinates" << std::endl;
			Radius = sqrt( padCenterLocal[0]*padCenterLocal[0] + padCenterLocal[1]*padCenterLocal[1] );
		}
		else if( ( theModule.getLocalPadLayout() ).getPadLayoutType() == 2 )
		{
			//polar
			streamlog_out( DEBUG ) << "Polar Coordinates" << std::endl;
			Radius = padCenterLocal[0];
		}
		else
		{
			streamlog_out( ERROR ) << "Error: Invalid pad layout type!" << std::endl;
		}


		streamlog_out( DEBUG ) << "Radius: " << Radius << std::endl;

		ShiftHits( HitPosition, NewHitPosition, RowModuleID.second, RowModuleID.first, Radius, maxPulseCharge, ( theModule.getLocalPadLayout() ).getPadLayoutType() );

		streamlog_out( DEBUG ) << "Original Hit: " << HitPosition[0] << " " << HitPosition[1] << " " << HitPosition[2] << std::endl;
		streamlog_out( DEBUG ) << "Shifted Hit: " << NewHitPosition[0] << " " << NewHitPosition[1] << " " << NewHitPosition[2] << std::endl;
	}

	lcio::TrackerHitImpl* theNewHit = new lcio::TrackerHitImpl();
	theNewHit->setPosition( NewHitPosition );
	theNewHit->setCellID0( theHit->getCellID0() );
	theNewHit->setCellID1( theHit->getCellID1() );
	theNewHit->setCovMatrix( theHit->getCovMatrix() );
	theNewHit->setEDep( theHit->getEDep() );
	theNewHit->setEDepError( theHit->getEDepError() );
	theNewHit->setTime( theHit->getTime() );
	theNewHit->setType( theHit->getType() );
	theNewHit->setQuality( theHit->getQuality() );
	theNewHit->rawHits() = PulseVect;
	_AllShiftedHits.push_back( theNewHit );

	}// hit loop
 streamlog_out( DEBUG ) << "Shifted Hits: " << _AllShiftedHits.size() << " in:" << hitNum << std::endl;
 IMPL::LCCollectionVec* outputCol = new IMPL::LCCollectionVec( LCIO::TRACKERHIT );
 //loop all shifted hits and add to collection
 for ( unsigned int h=0;h<_AllShiftedHits.size();h++ )
	{
	outputCol->addElement( _AllShiftedHits[h] );
	}
 evt->addCollection( outputCol, _HitDataCollName_OUT );
 streamlog_out( DEBUG ) << "Adding " << _HitDataCollName_OUT << " with " << _AllShiftedHits.size() << " hits." << std::endl;
}

 void ModuleDistortionCorrectionProcessor::ShiftHits( const double* hitPos, double* newPos, int module, int row, double radius, double charge, int coordinate )
 {
	//initialize with current position
	newPos[0] = hitPos[0];
	newPos[1] = hitPos[1];
	newPos[2] = hitPos[2];

	std::map< std::pair<int, int>, std::pair<double, double> >::const_iterator iterz =
	( _ZDistortionCorrection ).find( std::make_pair( module, row ) );
	std::map< std::pair<int, int>, std::pair<double, double> >::const_iterator iterrow =
	( _RowDistortionCorrection ).find( std::make_pair( module, row ) );

	double lgQ = log10( charge );
	unsigned int bin;
	if( lgQ>=4 )
		bin = 40;
	else if( lgQ<2 )
		bin = 1;
	else
		bin = static_cast<unsigned int>((lgQ - 2) / 2 * 40) + 1;
	streamlog_out( DEBUG ) << "charge=" << charge << " --> bin=" << bin << endl;

	//shift in row
	if( iterrow != ( _RowDistortionCorrection ).end() )
	{
		double alpha = 0.;
		const gear::TPCModule& theModule = _GearParameters->getModule( module );
		double distortionRow = ( ( *iterrow ).second ).first;

		streamlog_out( DEBUG ) << "Shifting along row" << std::endl;
		if( ( theModule.getLocalPadLayout() ).getPadLayoutType() == 2 )
		{
			gear::Vector2D HitPosLocalTemp = theModule.globalToLocal( hitPos[0], hitPos[1] );
			double HitPositionLocal[2] = {HitPosLocalTemp[0], HitPosLocalTemp[1]};//Attention: These are now cylinder coordinates!
			double NewHitPositionLocal[2] = {0., 0.};

			if( coordinate == 2 )//polar
			{
				alpha = asin( distortionRow/radius );
				NewHitPositionLocal[0] = HitPositionLocal[0];
				NewHitPositionLocal[1] = HitPositionLocal[1] - alpha;
			}
			else if( coordinate == 1 )//cartesian
			{
				alpha = asin( distortionRow/radius );
				NewHitPositionLocal[0] = HitPositionLocal[0]*cos( alpha ) - HitPositionLocal[1]*sin( alpha );
				NewHitPositionLocal[1] = HitPositionLocal[0]*sin( alpha ) + HitPositionLocal[1]*cos( alpha );
			}
			else
			{
				streamlog_out( ERROR ) << "Invalid coordinate type!" << std::endl;
				NewHitPositionLocal[0] = HitPositionLocal[0];
				NewHitPositionLocal[1] = HitPositionLocal[1];
			}
			gear::Vector2D HitPosGlobalTemp = theModule.localToGlobal( NewHitPositionLocal[0], NewHitPositionLocal[1] );
			newPos[0] = HitPosGlobalTemp[0];
			newPos[1] = HitPosGlobalTemp[1];
		}
		else if( ( theModule.getLocalPadLayout() ).getPadLayoutType() == 1 )
		{
			streamlog_out( MESSAGE ) << "ATTENTION! This was not yet implemented for this pad layout type. Good luck doing it..." << std::endl;
			newPos[0] = hitPos[0];
			newPos[1] = hitPos[1];
		}
		else
		{
			streamlog_out( ERROR ) << "Error: Invalid pad layout type!" << std::endl;
		}

	}


	//shift in z
	double distortionZ = 0.;
	if( iterz != ( _ZDistortionCorrection ).end() )
	{
		distortionZ += ( ( *iterz ).second ).first;
		streamlog_out( DEBUG ) << "Correction along z for module " << ( ( *iterz ).first ).first << "/row " << ( ( *iterz ).first ).second << ": " << distortionZ << endl;
	}
	if( bin<=_ZChargeCorrection.size() )
	{
		distortionZ += ( _ZChargeCorrection.at( bin-1 ) ).first;
		streamlog_out( DEBUG ) << "Correction along z from charge bin " << bin << ":"<< ( _ZChargeCorrection.at( bin-1 ) ).first  << endl;
	}
	if( distortionZ )
	{
		streamlog_out( DEBUG ) << "Shifting along z by " << -distortionZ << " mm" << std::endl;
		newPos[2] = hitPos[2] - distortionZ;
	}

 }

void
ModuleDistortionCorrectionProcessor::check( LCEvent *  /*evt*/ )
{

}

void
ModuleDistortionCorrectionProcessor::end()
{

}



}

