/*
 * RawDataCutterProcessor.cc
 *
 *  Created on: Oct 13, 2011
 *      Author: caiazza
 */

#include "RawDataCutterProcessor.h"

//Standard inclusions
#include <iostream>
#include <sstream>
#include <exception>
#include <cmath>
#include <cassert>

//LCIO Inclusion
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "EVENT/TrackerRawData.h"
#include "IMPL/TrackerRawDataImpl.h"

//Marlin inclusion
#include "marlin/Global.h"
#include "marlin/Exceptions.h"
#include "marlin/ConditionsProcessor.h"

//GEAR inclusions
#include "gear/TPCParameters.h"

//MarlinTPC Inclusions
#include "TPCCondDataExeptions.h"

namespace marlintpc
{

RawDataCutterProcessor aRawDataCutterProcessor;

RawDataCutterProcessor::RawDataCutterProcessor()
        : marlin::Processor("RawDataCutterProcessor")
{

    // Processor description
    _description = "This processor cuts all the data arriving later than the maximum drift time";
    _TPCConditionsListener = 0;
    _ADCElectronicsListener = 0;

    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(LCIO::TRACKERRAWDATA, "InputCollectionName",
            "Name of the input collection containing the raw data", _inputcolName,
            std::string("TPCRawData"));

    registerOutputCollection(LCIO::TRACKERRAWDATA, "OutputCollectionName",
            "Name of the collection where to store the corrected data", _outcolName,
            std::string("CorrectedTPCRawData"));

    registerProcessorParameter("WriteOutputToStorage",
            "If true the output collection is written to file", _outputIsPersistent, true);

    registerOptionalParameter(
            "MaxTimeBinOverride",
            "Set this parameter if you want to define manually the maximum time bin length. No calculation will be executed",
            _maxTimeBin, 0);

    registerOptionalParameter("StartOffset", "Defines the time bin of the anode position",
            _startOffset, 0);

    registerOptionalParameter(
            "DriftVelocityOverride",
            "Defines the drift velocity in mm/us. If set the data in the condition database will not be considered",
            _driftVelocity, static_cast<float>(80.0));

    registerOptionalParameter(
            "DriftLengthOverride",
            "Defines the maximum drift length in mm of the TPC. If set the data in the GEAR file will not be considered",
            _driftLength, static_cast<float>(569.37));

    registerOptionalParameter(
            "ReadoutFrequencyOverride",
            "Defines the electronics readout frequency in MHz. If set the data in the condition database will not be considered",
            _readoutFreq, static_cast<float>(20.0));

    registerOptionalParameter("EndOffset",
            "Defines the time bin to add to the calculated maximum as a safety factor", _endOffset,
            20);

    registerProcessorParameter("TPCConditionsCollName",
            "Defines the name of the condition collection containing the drift velocity",
            _TPCConditionsCollName, tpcconddata::TPCConditions::getDefaultColName());

    registerProcessorParameter("ADCElectronicsCollName",
            "Defines the name of the condition collection containing the readout frequency",
            _ADCElectronicsCollName,
            tpcconddata::GenericADCElectronicsParameters::getDefaultColName());

    registerOptionalParameter(
            "ShortenPulse",
            "Boolean selector that specifies whether to shorten the pulses started before the filter time but continuing afterwards",
            _shortenPulse, false);

    streamlog_out(DEBUG) << "Construction complete" << std::endl;
}

void
RawDataCutterProcessor::init()
{
    //TODO Maybe generalize the electronics type
    const std::string ELECTRONICSTYPE = std::string("GenericADCElectronicsParameters");

    streamlog_out(DEBUG) << "RawDataCutterProcessor initialization." << std::endl;

    // usually a good idea to
    printParameters();

    //If the MaxTimeBinOverride is not set calculate it, otherwise use the one set by the user
    if (!parameterSet("MaxTimeBinOverride"))
    {
        //Get the drift length from the GEAR file only if the override parameter is not set
        if (!parameterSet("DriftLengthOverride"))
        {
            try
            {
                streamlog_out(DEBUG)
                            << "The drift length was not set by the user. Trying to recover it from the GEAR file"
                            << std::endl;
                _driftLength = marlin::Global::GEAR->getTPCParameters().getMaxDriftLength();
                streamlog_out(MESSAGE) << "Drift length set to " << _driftLength << std::endl;
            }
            catch (std::exception &e)
            {
                streamlog_out(ERROR) << "Impossible to access the drift length from the GEAR file"
                            << std::endl;
                throw lcio::DataNotAvailableException(e.what());
            }
        }

        //If the condition parameter are overridden calculate the maximum time bin.
        if (parameterSet("DriftVelocityOverride") && parameterSet("ReadoutFrequencyOverride"))
        {
            _maxTimeBin = _startOffset
                    + static_cast<int>(std::ceil((_driftLength / _driftVelocity * _readoutFreq)))+_endOffset;
        }
        else
        {
            if(!parameterSet("DriftVelocityOverride"))
            {
                try
                {
                    _TPCConditionsListener = new TPCConditionsListener(_TPCConditionsCollName);
                    streamlog_out(MESSAGE) << "Listener created for the TPC Conditions object named: "<< _TPCConditionsCollName << std::endl;
                }
                catch(lccd::LCCDException &le)
                {
                    streamlog_out(ERROR) << "TPC Conditions collection with name " << _TPCConditionsCollName << " was not found." << std::endl;
                    throw le;
                }
            }

            if(!parameterSet("ReadoutFrequencyOverride"))
            {
                try
                {
                    _ADCElectronicsListener = new ADCElectronicsListener(ELECTRONICSTYPE, _ADCElectronicsCollName);
                    //FIXME Remove the use last valid as soon as I filled the database and prepare an algorithm of checks and warnings for no data available
                    _ADCElectronicsListener->useLastValidIfNoDataAvailable(true);
                    streamlog_out(MESSAGE) << "Listener created for the Electronics Conditions object named: "<< _ADCElectronicsCollName << std::endl;
                }
                catch(lccd::LCCDException &le)
                {
                    streamlog_out(ERROR) << "ADC Electronics conditions collection with name " << _ADCElectronicsCollName << "not found." << std::endl;
                    throw le;
                }
            }
        }
    }

    _nRun = 0;
    _nEvt = 0;
}

void
RawDataCutterProcessor::processRunHeader(lcio::LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2806 $");
    for (marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
    {
        if (!i->second->isOptional() || i->second->valueSet())
        {
            run->parameters().setValue(_processorName + "_" + i->second->name(),
                    i->second->value());
        }
    }

    _nRun++;
}

void
RawDataCutterProcessor::processEvent(lcio::LCEvent * evt)
{
    //TODO Check for the presence of multiple readout group and calculate the value for each of them separately
    const int STANDARDREADOUTGROUP = 0;
    LCCollection* inputCol = 0;
    LCCollectionVec* outputCol = 0;

    //Computes the cut variable only if is the first event
    if (isFirstEvent())
    {
        streamlog_out(DEBUG) << "First event variable calculation." << std::endl << "Event time stamp: " << evt->getTimeStamp() << std::endl;
        /*try
        {*/
            if (!parameterSet("MaxTimeBinOverride"))
            {
                streamlog_out(DEBUG)
                            << "The maximum number of time bins was not set by the user. Calculating it..."
                            << std::endl;
                if (!parameterSet("DriftVelocityOverride"))
                {
                    if (_TPCConditionsListener->noConditionsDataAvailable())
                    {
                        std::stringstream out;
                        out << _processorName << "::processEvent: No TPCCondition data available"
                                << std::endl;
                        throw tpcconddata::CondDataNotAvailableException(out.str());
                    }assert(_TPCConditionsListener!=0);

                    streamlog_out(DEBUG)
                                << "The Drift Velocity was not set by the user. Attempting to recover it from the database"
                                << std::endl;

                    assert(_TPCConditionsListener->getTPCConditions() !=0);
                    _driftVelocity = _TPCConditionsListener->getDriftVelocity();
                    streamlog_out(MESSAGE) << "Drift velocity set to " << _driftVelocity
                                << "[mum/ns] using the condition data from the database" << std::endl;
                }

                if (!parameterSet("ReadoutFrequencyOverride"))
                {
                    streamlog_out(DEBUG)
                                << "The Electronic readout frequency was not set by the user. Attempting to recover it from the database"
                                << std::endl;
                    if (_ADCElectronicsListener->noConditionsDataAvailable())
                    {
                        std::stringstream out;
                        out << _processorName
                                << "::processEvent: No ADCElectronics Conditions data available"
                                << std::endl;
                        throw tpcconddata::CondDataNotAvailableException(out.str());
                    }

                    _readoutFreq = _ADCElectronicsListener->getFrequency(STANDARDREADOUTGROUP);
                    streamlog_out(MESSAGE) << "Readout Frequency set to " << _readoutFreq
                                << "[MHz] using the condition data from the database" << std::endl;
                }

                _maxTimeBin = _startOffset
                        + static_cast<int>(std::ceil(
                                (_driftLength / _driftVelocity * _readoutFreq)))+ _endOffset;
                streamlog_out(MESSAGE) << "The maximum time bin for this run is calculated to be "
                            << _maxTimeBin << " timeslices" << std::endl;
            }
        /*}
        catch (std::exception &e)
        {
            streamlog_out(ERROR) << e.what();
            std::cerr << e.what();
        }*/
    }

    try
    {
        streamlog_out(DEBUG) << "Starting event " << _nEvt << std::endl;
        inputCol = evt->getCollection(_inputcolName);
        outputCol = new LCCollectionVec(LCIO::TRACKERRAWDATA);
        outputCol->setTransient(!_outputIsPersistent);
        //Only set collection to subset if the raw data are not altered
        if (_shortenPulse == false)
        {
            outputCol->setSubset(true);
        }

        for (int idPulse = 0; idPulse < inputCol->getNumberOfElements(); idPulse++)
        {
            //Retrieve the Raw Data from the input collection
            TrackerRawData* inputData = dynamic_cast<TrackerRawData*>(inputCol->getElementAt(
                    idPulse));
            //Measure the length of the pulse
if                    (inputData->getTime() <= _maxTimeBin)
                    {
                        //Check if the pulse can be written as a copy of the original one
                        //This happens if I ask never to shorten pulses or if the pulse is fully contained in my time window
                        if((!_shortenPulse) || (inputData->getTime() + static_cast<int>(inputData->getADCValues().size()) <= _maxTimeBin))
                        {
                            if(outputCol->isSubset())
                            {
                                outputCol->addElement(inputData);
                            }
                            else
                            {
                                TrackerRawDataImpl* newPulse = new TrackerRawDataImpl();
                                newPulse->setTime(inputData->getTime());
                                newPulse->setCellID0(inputData->getCellID0());
                                newPulse->setCellID1(inputData->getCellID1());
                                newPulse->setADCValues(inputData->getADCValues());
                                outputCol->addElement(newPulse);
                            }
                        }
                        //If not copy the pulse and change the adcspectrum
                        else
                        {
                            streamlog_out(DEBUG) << "Shortening pulse " << idPulse << std::endl;
                            //TrackerRawDataImpl* newPulse = dynamic_cast<TrackerRawDataImpl*>(inputData->clone());
                            TrackerRawDataImpl* newPulse = new TrackerRawDataImpl();
                            newPulse->setTime(inputData->getTime());
                            newPulse->setCellID0(inputData->getCellID0());
                            newPulse->setCellID1(inputData->getCellID1());
                            int maxLength = _maxTimeBin - inputData->getTime();
                            EVENT::ShortVec newSpectrum = EVENT::ShortVec();
                            newSpectrum.reserve(maxLength);
                            newSpectrum.clear();
                            for(EVENT::ShortVec::const_iterator it = inputData->getADCValues().begin(); it <= inputData->getADCValues().begin() + maxLength; it++)
                            {
                                newSpectrum.push_back(*it);
                            }
                            newPulse->setADCValues(newSpectrum);
                            outputCol->addElement(newPulse);
                        }
                    }
                }

        streamlog_out(DEBUG) << "Output collection type is " << outputCol->getTypeName() << std::endl;
        evt->addCollection(outputCol, _outcolName);
        streamlog_out(DEBUG) << "Pulse Selection in event " << evt->getEventNumber() << " completed."
                    << std::endl;
    }
    catch (lcio::DataNotAvailableException &e)
    {
        streamlog_out(WARNING) << _processorName << " : Empty event " << evt->getEventNumber()
                    << std::endl;
        throw marlin::SkipEventException(this);
    }
    catch (std::exception &e)
    {
        streamlog_out(ERROR) << e.what() << std::endl;
        throw marlin::StopProcessingException(this);
    }

    _nEvt++;
}

void
RawDataCutterProcessor::check(lcio::LCEvent *  /*evt*/)
{
    // nothing to check here - could be used to fill checkplots in reconstruction processor
}

void
RawDataCutterProcessor::end()
{
    //Some cleanup
    if (_TPCConditionsListener != 0)
    {
        delete _TPCConditionsListener;
        _TPCConditionsListener = 0;
    }

    if (_ADCElectronicsListener != 0)
    {
        delete _ADCElectronicsListener;
        _ADCElectronicsListener = 0;
    }

}

}

