#include "TrackerRawDataToDataConverterProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCFlagImpl.h>

// Aida
#ifdef MARLIN_USE_AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud1D.h>
#endif

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{
  TrackerRawDataToDataConverterProcessor aTrackerRawDataToDataConverterProcessor;

  TrackerRawDataToDataConverterProcessor::TrackerRawDataToDataConverterProcessor() : Processor("TrackerRawDataToDataConverterProcessor")
  {
    /* the constructor call first the super constructor which the name
     * of the processor
     * this must by the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file automatically generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "TrackerRawDataToDataConverterProcessor converts the TrackerRawData to TrackerData without processing the values";

    /* register steering parameters: name, description, class-variable, default value
     * the type will by defined by the type of the default value
     * string, double, float and int are possible
     * please note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERRAWDATA,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TPCRawData"));

    registerOutputCollection(LCIO::TRACKERDATA,
                             "OutputCollectionName" ,
                             "Name of the output collection"  ,
                             _outputColName ,
                             std::string("TPCConvertedRawData"));

    registerProcessorParameter("WriteOutputToStorage",
                               "if true the output collection is written",
                               _outputIsPersistent,
                               true);
    registerOptionalParameter("SetEntriesToZeroAfterBin" ,
                              "The last bin number, after which all bins will be set to zero",
                              _setToZeroAfterBin,
                              0);

    registerOptionalParameter("SetEntriesToZeroBeforeBin" ,
                              "The first bin number which is taken into account; bins before will be set to zero",
                              _setToZeroBeforeBin,
                              0);
  }



  void TrackerRawDataToDataConverterProcessor::init()
  {
    printParameters();

    _changeBinsAfter  = parameterSet("SetEntriesToZeroAfterBin");
    _changeBinsBefore = parameterSet("SetEntriesToZeroBeforeBin");
  }



  void TrackerRawDataToDataConverterProcessor::processRunHeader(lcio::LCRunHeader* run)
  {
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2711 $");
    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
      {
        if (! i->second->isOptional() || i->second->valueSet())
          {
            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
          }
      }
  }



  void TrackerRawDataToDataConverterProcessor::processEvent(lcio::LCEvent* evt)
  {
    // get the input collection from the event
    try
      {
        LCCollection* inputCol = evt->getCollection(_inputColName);

        streamlog_out(DEBUG4) << "Processing " << inputCol->getNumberOfElements() << " TrackerRawData elements from collection " << _inputColName << "." << endl;

        // create the output collection with a string defining the TYPE (see lcio.h)
        LCCollectionVec* outputCol = new LCCollectionVec(LCIO::TRACKERDATA);
        outputCol->reserve(inputCol->getNumberOfElements());  // to make the final storage call cheaper

	// set the flags that cellID1 should be stored
	lcio::LCFlagImpl trkFlag(0) ;
	trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
	outputCol->setFlag(trkFlag.getFlag());

        // set the transient flag as defined by the processor parameter
        outputCol->setTransient(!_outputIsPersistent);   // transience is the opposite of persistency

        // get the number of elements in the input collection and loop over it
        for (unsigned int iRawData = 0, lastElement = inputCol->getNumberOfElements(); iRawData < lastElement; ++iRawData)
          {
            /* getElement return only a LCObject
             * you have to cast it the get the right type
             * please use dynamic_cast which will return 0 if the object can not be casted
             */
            TrackerRawData* RawData = static_cast<TrackerRawData*>(inputCol->getElementAt(iRawData));

            if (RawData != 0)
              {
                //only proceed if the cast was successful

                /* create a new object of the TrackerData type
                * you have to use the IMPL, otherwise you can not set values
                */
                TrackerDataImpl* Data = new TrackerDataImpl();

                // transfer the values

                //first the simple
                Data->setCellID0(RawData->getCellID0());
                Data->setCellID1(RawData->getCellID1());  // normally 0 for TPC data

                Data->setTime((float)RawData->getTime());

                streamlog_out(DEBUG0) << "Converting data from pad " << RawData->getCellID0() << " and module " <<  RawData->getCellID1() << endl;
                streamlog_out(DEBUG0) << "Time for ADC data is " << (float)RawData->getTime() << endl;
                streamlog_out(DEBUG0) << "ADC data is: [";
                if (parameterSet("SetEntriesToZeroBeforeBin"))
                  streamlog_out(DEBUG0) << "Skip the first " << _setToZeroBeforeBin << " bins! " << endl;
                if (parameterSet("SetEntriesToZeroAfterBin"))
                  streamlog_out(DEBUG0) << "Skip the first " << _setToZeroAfterBin << " bins! " << endl;

                // and now the vector of shorts
                ShortVec RawDataADCVal = RawData->getADCValues();
                // create a handle to reserve the float vector for storage
                EVENT::FloatVec& chargeVectorHandle = Data->chargeValues();
                chargeVectorHandle.reserve(RawDataADCVal.size());

                for (ShortVec::iterator iSlice = RawDataADCVal.begin(), endIter = RawDataADCVal.end(); iSlice != endIter; ++iSlice)
                  {
                    if (_changeBinsAfter)
                      {
                        int index = iSlice - RawDataADCVal.begin();
                        if (index > _setToZeroAfterBin)
                          {
                            chargeVectorHandle.push_back(0.);
                            continue;
                          }
                      }
                    if (_changeBinsBefore)
                      {
                        int index = iSlice - RawDataADCVal.begin();
                        if (index < _setToZeroBeforeBin)
                          {
                            chargeVectorHandle.push_back(0.);
                            continue;
                          }
                      }

                    // else fill the data vector with converted adc values
                    chargeVectorHandle.push_back(static_cast<float>(*iSlice));
                    streamlog_out(DEBUG) << (float)(*iSlice) << ",";
                  }
                streamlog_out(DEBUG) << "]" << endl;

                /* fill the TrackerData in the collection
                * please note: DO NOT DELETE the object this will be done by LCIO
                */
                outputCol->addElement(Data);
              }
          }

        /* after converting all raw data objects
         * add the collection to the eventand asigning a name to it
         * please note again: DO NOT DELETE the collection this will be done by LCIO
         */
        evt->addCollection(outputCol, _outputColName);

      }
    catch (DataNotAvailableException)
      {
        // if there is not input collection give a warning to the user
        // note: this is not a graceful way to handle this exception
        streamlog_out(WARNING) << "Warning of TrackerRawDataToDataConverterProcessor::processEvent: input collection "
        << _inputColName << " cannot be found in event " << evt->getEventNumber() << std::endl;
      }
  }



  void TrackerRawDataToDataConverterProcessor::check(lcio::LCEvent*  /*evt*/)
  {
    // nothing to check here - could be used to fill checkplots in reconstruction processor
  }



  void TrackerRawDataToDataConverterProcessor::end()
  {
    //   nothing done here
  }
}
