#ifndef BIASCALCULATORPROCESSOR_H
#define BIASCALCULATORPROCESSOR_H

#include <marlin/Processor.h>
#include <lcio.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>

#include <TProfile.h>
#include <TDirectoryFile.h>
#include <TGraphErrors.h>
#include <TNtuple.h>

namespace marlintpc
{
  /* Processor to output bias data (i.e., residuals as a function of hit-position on pad).
   * 
   * Parameters:
   * 
   *    TrackInputCollectionName  =  Input track collection  [string]
   *    BiasOutputFileName_RPhi   =  R-Phi output file name  [string]
   *    BiasOutputFileName_Z      =  Z output file name  [string] 
   *    RootOutput                =  Whether or not to save the TProfiles to file  [boolean]
   *    SingleTrack               =  Restrict to single-track events  [boolean]
   *    SmallAngle                =  Restrict to hits with a local track angle between -5 and 5 degrees  [boolean]
   *    SeparateDistortions       =  Whether or not to remove the mean of the bias distributions (i.e., the distortions)  [boolean] 
   *    RelativeBias              =  Whether or not to use x_relative (i.e. (x_hit - x_pad) / pad_width )  [boolean] 
   *    NPulses                   =  Restrict to hits with this many pulses [integer]
   * 
   * Author: P. Hayman, Carleton University
   */

  class BiasCalculatorProcessor : public marlin::Processor
  {

    public:

      virtual Processor*  newProcessor() { return new BiasCalculatorProcessor; }

      BiasCalculatorProcessor();

      virtual void init();

//FIXME: not used:      virtual void processRunHeader(lcio::LCRunHeader* run);

      virtual void processEvent(lcio::LCEvent* evt);

//FIXME: not used:      virtual void check(lcio::LCEvent* evt);

      virtual void end();

    protected:
      std::string _inputTracks; ///< Name of the input track collection 

    private:

      //Class Variables
      int  _nPulses;
      bool _rootOutput;
      bool _singleTrack;
      bool _smallAngle;
      bool _selectPulses;
      bool _separateDistortions;
      bool _relative;

      TDirectoryFile * _dir_rphi;
      TDirectoryFile * _dir_z;

      std::string _biasOutputFileName_rphi;
      std::string _biasOutputFileName_z;
      std::map< std::pair<int,int>, TProfile*> _histograms_rphi;
      std::map< std::pair<int,int>, TProfile*> _histograms_z;

      TGraphErrors * _rphiGr;
      TGraphErrors * _zGr;
      TNtuple *      _biasTuple;  //This stores some statistical information about the bias

      //Private Functions
      std::pair<double, double> calculateBias_rphi(const EVENT::TrackerHit *, const EVENT::Track *);
      std::pair<double, double> calculateBias_z(const EVENT::TrackerHit *, const EVENT::Track *);
      void writeBias();
      void calculateAverageBias();
      void writeBiasOutput(std::string filename, std::map< std::pair<int,int>, std::pair<double,double> > results);

  };//End BiasCalculatorProcessor class

}//End namespace marlintpc

#endif //BIASCALCULATORPROCESSOR_H
