#ifndef HITTIMERECALCULATIONPROCESSOR_H
#define HITTIMERECALCULATIONPROCESSOR_H

//LCIO
#include <lcio.h>
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/TrackerDataImpl.h>
#include <EVENT/TrackerPulse.h>

//Marlin
#include <marlin/Processor.h>
#include "marlin/Global.h"
//GEAR
#include "gear/TPCParameters.h"
//LCCD
#include <lccd/ConditionsMap.hh>
#include "Pedestal.h"
#include "GenericADCElectronicsParameters.h"
// MARLINTPC
#include "TPCConditionsListener.h"
//C++
#include <string>

namespace marlintpc
{
  class PulseFinder; // forward declaration of the PulseFinder class. 

  /** This processor recalculates the time of the hit based on the highest pulse in the hit according to the
   *  choosen time evaluation method
   *
   *  <h4>Input - Prerequisites</h4>
   *  Collection of TrackerHits
   *
   *  <h4>Output</h4>
   *  \li Collection of TrackerHits
   *
   *  @param InputCollectionName:string Name of the input TrackerHit collection
   *  @param OutputCollectionName:string Name of the output TrackerHit collection
   *  @param PulseStartPedestalWidthFactor:float A pulse starts if the ADC value is this factor 
   *         times width of the pedestal (default: : 4.0)
   *  @param NoiseValuePedestalWidthFactor:float The noise (cut) value is calculated as this factor times pedestal width (default: 3.0)
   *  @param ElectronicsParameterObjectName:string Name of the conditions object in which the parameters of the read out electronics is stored
   *  @param ElectronicsParametersCollectionName:string Name of the conditions collection in which the parameters of the read out electronics is stored
   *  @param PedestalCollectionName:string Name of the conditions collection in which the pedestals are stored
   *  @param TimeCalculationMode:int Specifies the mode for time calculation. Valid values:<br>
   *                                 INFLEXION_POINT = 0 (default) <br>
   *                                 FIT_PARABOLA = 1 <br>
   *                                 CENTRE_OF_GRAVITY = 2 <br>
   *                                 FIRST_ABOVE_THRESHOLD = 3 <br>
   *                                 AT_MAXIMUM = 4 <br>
   *                                 CONSTANT_FRACTION = 5 <br>
   *                                 GAMMA_FIT = 6<br>
   *				     COG_ATMAX = 7<br>
   *
   *  @param PedestalValueOverride \c OPTIONAL: Manually set the pedestal value (in ADC counts)
   *  @param PedestalWidthOverride \c OPTIONAL: Manually set value of the pedestal width (in ADC coutns)
   *  @param IsPedestalSubtractedOverride \c OPTIONAL: Manually set whether the pedestal has been subtracted (true or false)
   *  @param SignalPolarityOverride \c OPTIONAL: Manually set value of the electronics polarity (+-1)
   *  @param ReadoutFrequencyOverride \c OPTIONAL: Manually set value of the readout frequency (in MHz)
   *  @param PlateauLengthCutOff \c OPTIONAL: The maximum length of a plateau within a pulse in time bins. This used to cut off strange pulses that go into a plateau.
   *
   * As the interpretation of the data changes with the charge calculation mode 
   * (all charge or max charge) or the time calculation mode (different offsets), the parameters
   * \c TimeCalculationMode and \c ChargeCalculationMode are stored as collection parameters of the
   * pulse collection (in addition to being written to the run header together with all other 
   * processor parameters).
   *
   *
   * @author ChA. Muennich, DESY
   */
  class HitTimeRecalculationProcessor : public marlin::Processor
  {

    public:

      virtual Processor*  newProcessor()
      {
        return new HitTimeRecalculationProcessor;
      }

      HitTimeRecalculationProcessor();

      virtual void init();

      virtual void processRunHeader(lcio::LCRunHeader* run);

      virtual void processEvent(lcio::LCEvent* evt);

      virtual void check(lcio::LCEvent* evt);

      virtual void end();

    protected:

      /// name of the collection, where input TrackerHits are stored
      std::string  _inputColName;
      /// name of the collection, where TrackerHits are stored
      std::string  _outputColName;
      /// the name of the pedestal collection
      std::string  _pedestalColName;

      /// Name of the conditions object in which the parameters of the read out electronics is stored
      std::string  _electrParaObjName;

      /// Name of the conditions collection in which the parameters of the read out electronics is stored
      std::string _electrParaColName;

      /** the name of the collection the TPC conditions are stored with
       */
      std::string _inputTPCConditionsCollectionName;
      /** the standard way to access TPC conditions: the TPCConditions object via listening
       */
      TPCConditionsListener* _tpcConditionsListener;
      
      int _timeCalculationMode; ///< the mode to calulate the time      
      double _timeCalculationParameter; ///<parameter for time methods (different meaning for different methods)

      /// factor to calculate the start threshold
      float _pulseStartPedestalWidthFactor;
       /// factor to calculate the noise cut
      float _noiseValuePedestalWidthFactor;

      // override variables

      /** The readout frequency of every readout group (in MHz). This overrides any information provided by LCCD
       */
      float _readoutFrequencyOverride;

      /** The pedestal value to be used for every hardware channel in every readout group. This overrides any information provided by LCCD
       */
      float _pedestalValueOverride;

      /** The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD
       */
      float _pedestalWidthOverride;

      /** Sets, if the data is already pedestal subtracted in the read-out electronics. This overrides any information provided by LCCD
       */
      bool _isPedestalSubtractedOverride;

      /** the polarity of all readout groups, overrides LCCD
         */
      int _polarityOverride;

      /** drift velocity in [mm/mu s]
       */
      float _driftVelocityOverride;
      double _getDriftVelocity();

      PulseFinder* _pulseFinder; ///< The instance of the pulse finder object.

  };
}
#endif // HITTIMERECALCUILATIONPROCESSOR_H

