#ifndef MMHITTIMECORRECTIONPROCESSOR_H
#define MMHITTIMECORRECTIONPROCESSOR_H

#include "marlin/Processor.h"
#include "lcio.h"
#include <string>

#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include "PedestalListener.h"
#include "ADCElectronicsListener.h"
#include "TPCConditionsListener.h"

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

  /** Processor to calculate a more accurate time estimate for Micromegas hits. 
   * 
   *  MM Pulses are given a simple time estimate from the time of the max bin. This is
   *  used solely for grouping pulses in the hit finder. This processor revisits the max 
   *  pulses in the hits and calculates their times with a more careful method.
   * 
   *  Time Calculation Options:
   *  -Inflexion point (0, default)
   *  -Max bin (1)
   *  -Box method (2)
   *  -Half-Max Bin (3)
   *  -Fixed-Value (4)
   *  -Weighted-Mean (5)
   *  -Gaussian Inflexion Point (6)
   *  -Gaussian Mean (7)
   *  -Gamma5 Mean (8)
   *
   *  <h4>Input - Prerequisites</h4>
   *  Needs the collection of TPCHits.
   *
   *  <h4>Output</h4> 
   *  The corrected collection of hits. 
   * 
   * @param InputTrackerHits Name of the input TrackerHit collection.
   * @param DriftVelocityOverride User-supplied drift velocity for z-coordinate calculation (LCCD should be implemented soon). 
   * @param ReadoutFrequency User-supplied readout frequency (LCCD should be implemented soon). 
   * @param OutputTrackerHits Name of the output TrackerHit collection.
   * @param TimeCalculationMethod Optional: Select which method to use to calculate time (DEFAULT: 0)
   * @param NNeighbours Optional: number of neighbour pads to be used in the time calculation on each side of the max (DEFAULT: 0)
   * 
   * @author P. Hayman, Carleton University
   * @version $Id: MMHitTimeCorrectionProcessor.h,v 1.0 2013-08-14 $ 
   */

  class MMHitTimeCorrectionProcessor : public Processor 
  {

    public:

      virtual Processor*  newProcessor() { return new MMHitTimeCorrectionProcessor; }


      MMHitTimeCorrectionProcessor() ;

      /** Called at the beginning of the job before anything is read.
       * Use to initialize the processor, e.g. book histograms.
       */
      virtual void init() ;

      /** Called for every run.
       */
      virtual void processRunHeader( LCRunHeader* run ) ;

      /** Called for every event - the working horse.
       */
      virtual void processEvent( LCEvent * evt ) ; 


      virtual void check( LCEvent * evt ) ; 


      /** Called after data processing for clean up.
       */
      virtual void end() ;


    protected:

      /** Input collection name.
       */
      std::string _inputHitCollectionName ;
      std::string _outputHitCollectionName ;
      std::string _electrParaColName;
      std::string _electrParaObjName;
      std::string _tpcCondColName;
      /// the name of the pedestal collection
      std::string  _pedestalColName;
      float _driftVelocityOverride;
      float _readoutFrequencyOverride;
      float _timeOffsetOverride;
      int _timeCalculationMethod;
      bool _write_new_collection;
      bool _outputHitsPersistent;
      bool _useOverflowPulses;
      bool _discardHitsWithOverflow;
      /** The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD
       */
      float _pedestalWidthOverride;
      float _driftVelocity;
      float _readoutFrequency;
      float _timeOffset;

      int _nNeighbours;///< number of neighbour pads to be used in the time calculation on each side of the max

      int _nRun ;
      int _nEvt ;

      ADCElectronicsListener * _adcElectronicsListener;
      TPCConditionsListener * _tpcConditionsListener;
      PedestalListener * _pedestalListener;

      //Method to iterate over hits and correct the time estimate
      void _correctTimes(LCCollectionVec * inputCol, LCCollectionVec * outputCol) ;

      //General method calling one of the individual methods to carry out the desired method of time calculation
      std::pair<double, double> _calculateCorrectedTime(const TrackerPulse * pulse) ;

      //Individual methods to carry out the desired method of time calculation
      std::pair<double, double> _calculateInflexionPoint(const TrackerPulse * pulse) ;
      std::pair<double, double> _calculateMaxPoint(const TrackerPulse * pulse) ;
      std::pair<double, double> _calculateBoxMethod(const TrackerPulse * pulse) ;
      std::pair<double, double> _calculateHalfMaxPoint(const TrackerPulse * pulse) ;
      std::pair<double, double> _calculateWeightedMean(const TrackerPulse * pulse) ;
      std::pair<double, double> _calculateGaussianInflexion(const TrackerPulse * pulse) ;
      std::pair<double, double> _calculateGaussianMean(const TrackerPulse * pulse) ;
      std::pair<double, double> _calcPulseFitWithFunction(const EVENT::TrackerPulse * thePulse);
      std::pair<double, double> _calculateWeightedMeanAtMax(const EVENT::TrackerPulse * thePulse);

  };//End MMHitTimeCorrectionProcessor class

}//End namespace marlintpc
#endif
