#ifndef MMPadResponseFunction_H
#define MMPadResponseFunction_H 

// C++
#include <iostream>
#include <string>
#include <vector>
#include <map>

// ROOT
#include "Math/ParamFunctor.h"
#include <Minuit2/FCNBase.h>
#include <TMath.h>

namespace marlintpc
{ 

  class MMPadResponseFunction : public ROOT::Minuit2::FCNBase  //Make this object a subclass of FCNBase so it can be minimized
  {

    /*  Class for the Micromegas Pad Response Function. This class holds general PRF data, and has a chi2 function that can be used to 
     *  fit a range of PRF functional forms to that data, using a range of normalization options. It also has functionality to evaluate a given PRF function
     *  on demand, and normalize a given set of data.
     *
     *
     *  Author: P. Hayman, Carleton University   
     */

    public:

      //Various fit options
      static const int FIT_MEAN = 0;   //Fit just the mean
      static const int FIT_PARAMS = 1; //Fit just the parameters (obsolete)
      static const int FIT_BOTH = 2;   //Fit both the mean and the parameters
      static const int SUM_FIT = 3;    //For calibration: fits both mean and parameters, and uses the sum of charges as the normalization
      static const int GAUS_FIT = 4;   //For calibration: fits both mean and parameters, and uses the amplitude of a gaussian fit as the normalization
      static const int CHI2_MIN = 5;   //For calibration: fits both mean and parameters, and uses the old min-chi2 definition of the normalization

      //This is the format in which the PRF stores its data. This format allows the PRF to hold data ranging from a single hit to an entire detector.
      typedef std::vector<std::vector<std::pair<double, float> > > PointsByRow;

      //The default settings are the classic "product" form for the PRF function, and the option to only fit the mean, as in the hit finder.
      MMPadResponseFunction(std::string prfType = "product", int fitType = MMPadResponseFunction::FIT_MEAN)
      {
        _PRFType = prfType;
        _fitType = fitType;
        x0 = 0;
        _restrict = false;

        //Always initialize the params array
        int nPar = getNPar(_PRFType);
        _params = new double[nPar];

        for(int i = 0; i < nPar; i++)
        {
          _params[i] = 0.;
        }
      }

      ~MMPadResponseFunction()
      { 
        if(_params != 0) 
        { 
          delete _params; 
          _params = 0; 
        }
      }
      MMPadResponseFunction( const MMPadResponseFunction & other)
      {
        _theErrorDef = other._theErrorDef;
        x0 = other.x0;
        _fitType = other._fitType; 
        _PRFType = other._PRFType;
        _restrict = other._restrict;
        _dataByRow = other._dataByRow;

        int nPar = getNPar(_PRFType);

        _params = new double[nPar];

        for(int i = 0; i < nPar; i++)
        {
          _params[i] = other._params[i];
        }
      }

      MMPadResponseFunction &operator=( const MMPadResponseFunction & other )
      {
        _theErrorDef = other._theErrorDef;
        x0 = other.x0;
        _fitType = other._fitType; 
        _PRFType = other._PRFType;
        _restrict = other._restrict;
        _dataByRow = other._dataByRow;

        int nPar = getNPar(_PRFType);

        delete [] _params;
        _params = new double[nPar];

        for(int i = 0; i < nPar; i++)
        {
          _params[i] = other._params[i];
        }

        return *this;
      }

      void SetData(std::vector< std::pair<double, float> >& aVec);  //This will convert the given vector to a PointsByRow variable
      void SetData(PointsByRow& aVec);

      void SetType(std::string prfType)
      {   
        _PRFType = prfType;
      }

      void SetFit(int fitType)
      {   
        _fitType = fitType;
      }

      //This option allows the user to restrict minimization of the PRF to +/- 5mm of x0. Could be useful.
      void SetRestrictDomain(bool rest)
      {   
        _restrict = rest;
      }

      //These functions allow for the normalization of any given data. C.f. MMScatterProcessor
      void normalizeData(PointsByRow& data);
      void normalizeData(std::vector< std::pair<double, float> >& data);

      //This form is necessary for cases where the parameters are being provided by Minuit. C.f. MMRefitTool.
      void SetParameters(const double * x, int nPar)
      {
        _params = new double[nPar];

        for(int i = 0; i < nPar; i++)
        {
          _params[i] = x[i];
        }

      }

      // Required functions from inheritance ///////////////////////////////////////////////////
      virtual double Up() const
      {
        return _theErrorDef;
      }

      double operator()(const std::vector<double>& x) const 
      {
        double * xx = new double[x.size()];
        for(unsigned int i = 0; i < x.size(); i++)  xx[i] = x[i];
        return (*this)(xx); 
      };

      //The main chi2 function
      double operator()(const double* x) const ;   

      //For a 1D minimization
      double operator()(const double xx) const
      {
        return (*this)(&xx);
      };

      void SetErrorDef(double def)
      {
        _theErrorDef = def;
      }
      //////////////////////////////////////////////////////////////////////////////////////////

      // Important functions related to the PRF functional form ////////////////////////////////

      //This allows other classes to check the number of parameters required for a given PRF form
      static int getNPar(const std::string & type)
      {

        std::map<std::string, std::pair<int, int> >::const_iterator it = MMPadResponseFunction::_validPRFs.find(type);
        if(it != _validPRFs.end())
        {
          return it->second.second;
        }
        else
          return -1;

      }

      //This is a map of the defined PRF functional forms. The map includes an index (for faster retrieval in the prf function call) 
      //and the number of parameters required
      static std::map<std::string, std::pair<int, int> > _validPRFs;       //Variable
      static std::map<std::string, std::pair<int, int> > generatePRFMap()  //Function
      {
        std::map<std::string, std::pair<int, int> > retVal;

        int index = 0;

        retVal["product"] = std::pair<int, int>(index, 2);     //0
        retVal["gaus*lor"] = std::pair<int, int>(index++, 2);
        retVal["product2"] = std::pair<int, int>(index, 3);    //1
        retVal["gaus*lor2"] = std::pair<int, int>(index++, 3);
        retVal["gaus+lor2"] = std::pair<int, int>(index++, 2); //2
        retVal["gaus"] = std::pair<int, int>(index++, 1);      //3
        retVal["gaus2"] = std::pair<int, int>(index++, 1);     //4
        retVal["lor"] = std::pair<int, int>(index++, 1);       //5
        retVal["sech"] = std::pair<int, int>(index++, 1);      //6
        retVal["sec"] = std::pair<int, int>(index++, 1);       //7
        retVal["sech+sec"] = std::pair<int, int>(index++, 2);  //8
        retVal["sech+sec2"] = std::pair<int, int>(index++, 2); //9
        retVal["sec*lor"] = std::pair<int, int>(index++, 2);   //10
        retVal["sec*gaus"] = std::pair<int, int>(index++, 2);  //11
        retVal["sec+lor"] = std::pair<int, int>(index++, 2);   //12
        retVal["gaus+lor"] = std::pair<int, int>(index++, 2);  //13
        retVal["gaus+lor3"] = std::pair<int, int>(index++, 3); //14
        retVal["gaus+gaus"] = std::pair<int, int>(index++, 2); //15
        retVal["gaus+sech"] = std::pair<int, int>(index++, 2); //16
        retVal["gaus+sec2"] = std::pair<int, int>(index++, 2); //17
        retVal["gaus+sec3"] = std::pair<int, int>(index++, 3); //18
        retVal["lor+lor"] = std::pair<int, int>(index++, 2);   //19
        retVal["sec+sec"] = std::pair<int, int>(index++, 2);   //20

        return retVal;
      }

      //The PRF function itself. 
      double prf(const double c, const double * x) const
      {
        //This retrieves the index of the desired PRF
        static std::map<std::string, std::pair<int, int> >::const_iterator it = MMPadResponseFunction::_validPRFs.find(_PRFType);

        if(it != _validPRFs.end())
        {
          //The index is then compared against the known options to retrieve the correct function. The integer index comparison is 
          //much faster than comparing against a string
          switch(it->second.first)
          {
            case 0:
              {
                static const double log2 = log(2.);
                return exp(-4.*log2*(1.-x[0])*c*c/(x[1]*x[1]))/ (1. + 4. * x[0] * c*c / (x[1]*x[1]) ); 
              }

            case 1:
              {
                static const double log2 = log(2.);
                return exp(-4.*log2*(1.-x[0])*c*c/(x[1]*x[1]))/ (1. + 4. * x[0] * c*c / (x[2]*x[2]) );
              }

            case 2:
              return ( exp(-(c*c)/(x[1]*x[1]) )*(1. - x[0]) + x[0]/(1.+(c*c)/(x[1]*x[1]) ) ); 

            case 3:
              return exp(-c*c/(x[0]*x[0]));

            case 4:
              {
                double tmp = exp(-c*c/(x[0]*x[0]));
                return  2.*tmp/(tmp + 1.);
              }

            case 5:
              return 1./(1. + c*c/(x[0]*x[0]));

            case 6:
              {
                double tmp = exp(c*x[0]);
                return  2.*tmp/(tmp*tmp + 1.);
              }

            case 7:
              {
                double tmp = exp(c*x[0]);
                return  4.*tmp/((tmp + 1.)*(tmp + 1.));
              }

            case 8:
              {
                double tmp1 = exp(c*x[0]);
                double tmp2 = exp(c*x[1]);
                return (tmp1/(tmp1*tmp1 + 1.) + 2.*tmp2/((tmp2 + 1.)*(tmp2 + 1.)));
              }

            case 9:
              {
                double tmp1 = exp(c/x[1]);
                double tmp2 = exp(c*x[1]);
                return (2.*tmp1/(tmp1*tmp1 + 1.)*(1. - x[0]) + 4.*x[0]*tmp2/((tmp2 + 1.)*(tmp2 + 1.)));
              }

            case 10:
              {
                double tmp = exp(-c/x[0]);
                return  4.*tmp/((tmp + 1.)*(tmp + 1.)*(1. + c*c/x[1]));
              }

            case 11:
              {
                double tmp = exp(-c*x[0]);
                return  4.*tmp/((tmp + 1.)*(tmp + 1.))*(exp(-c*c/(x[1]*x[1])));
              }

            case 12:
              {
                double tmp = exp(-c*x[0]);
                return  0.5*(4.*tmp/((tmp + 1.)*(tmp + 1.)) + 1./(1. + c*c/(x[1]*x[1])));
              }

            case 13:
              return 0.5 * ( TMath::Exp(-(c*c)/(x[0]*x[0]) ) + 1./(1.+(c*c)/(x[1]*x[1]) ) ); 

            case 14:
              return ( TMath::Exp(-(c*c)/(x[1]*x[1]) )*(1. - x[0]) + x[0]/(1.+(c*c)/(x[2]*x[2]) ) ); 

            case 15:
              return 0.5 * ( TMath::Exp(-(c*c)/(x[0]*x[0]) ) + TMath::Exp(-(c*c)/(x[1]*x[1]) ) ); 

            case 16:
              {
                double tmp = exp(-c*x[1]);
                return  0.5*(exp(-(c*c)/(x[0]*x[0])) + 2.*tmp/(tmp*tmp + 1.));
              }

            case 17:
              {
                double tmp = exp(-c*x[1]);
                return  0.5*(exp(-(c*c)/(x[0]*x[0])) + 4.*tmp/((tmp + 1.)*(tmp + 1.)));
              }

            case 18:
              {
                double tmp = exp(-c*x[2]);
                return  (TMath::Exp(-(c*c)/(x[1]*x[1]))*(1. - x[0]) + 4.*x[0]*tmp/((tmp + 1.)*(tmp + 1.)));
              }

            case 19:
              return 0.5 * ( 1./(1.+(c*c)*(x[0]*x[0])) + 1./(1.+(c*c)/(x[1]*x[1])) ); 

            case 20:
              {
                double tmp1 = exp(c*x[0]);
                double tmp2 = exp(c*x[1]);
                return  2.*tmp1/((tmp1 + 1.)*(tmp1 + 1.)) + 2.*tmp2/((tmp2 + 1.)*(tmp2 + 1.));
              }
            default:
              return -9E9;
          }
        }
        else 
        {
          std::cout << "PRF NOT DEFINED!!!" << std::endl;
          return -99999999.;
        }

      }//End PRF Function

    private:

      //Needed for FCN_Base inheritance
      double          _theErrorDef;

      //Mean of the prf
      double          x0;

      //PRF Parameters
      double*         _params;

      //Fit type
      int             _fitType; 

      //PRF Type
      std::string     _PRFType;

      //Restrict domain in fitting for parameters.
      bool            _restrict;

      //The data to minimize
      PointsByRow     _dataByRow;

      //Helper function to normalize the amplitudes    
      double calculateRowAmp(int row, double x0, const double *) const;
      double calculateRowAmp(int row, double x0, const double *, const PointsByRow& data) const;
      double calculateRowAmp(double x0, const double * x, const std::vector< std::pair<double, float> >& data) const;


  };//End PRF class definition

}//End namespace marlintpc

#endif
