#ifndef PULSEFINDERPROCESSOR_H
#define PULSEFINDERPROCESSOR_H

//LCIO
#include <lcio.h>
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/TrackerDataImpl.h>
#include <EVENT/TrackerPulse.h>

//Marlin
#include <marlin/Processor.h>

//LCCD
#include <lccd/ConditionsMap.hh>
#include "Pedestal.h"
#include "GenericADCElectronicsParameters.h"


//C++
#include <string>

namespace marlintpc
{
  class PulseFinder; // forward declaration of the PulseFinder class. 

  /** This processor creates lcio::TrackerPulse objects from lcio::TrackerData. It is the 
   *  modular version of the PulseFinderProcessor in the sense that the functionality has been
   *  extracted into a separate class (PulseFinder) to allow unit testing. See the documentation
   * of PulseFinder for the description of the algorithm.
   *
   *  <h4>Input - Prerequisites</h4>
   *  Collection of TrackerData (can be zero supressed and/ or pedestal subtracted)
   *
   *  <h4>Output</h4>
   *  \li Collection of TrackerPulses
   *  \li Collection of TrackerData which belong to a pulse. As the pulses might have been split
   *  these TrackerData are not identical to the input collection.
   *
   *  @param InputCollectionName:string Name of the input TrackerData collection
   *  @param OutputTrackerPulseCollectionName:string Name of the output TrackerPulse collection
   *  @param OutputTrackerDataCollectionName:string  Name of the output TrackerData collection
   *  @param WriteOutputToStorage:bool Output collection will be written if set to true (which is the default)
   *  @param ForceSpectrumSave:bool If true the ADC spectrum will be stored for all pulses,
   *         otherwise spectra are only stored if the quality word is not 0 (default: true)
   *  @param SwitchOffSplitting:bool If true pulses will not be splitted, e.g. if double pulse is detected
   *  @param PulseStartPedestalWidthFactor:float A pulse starts if the ADC value is this factor 
   *         times width of the pedestal (default: : 7.0)
   *  @param PulseEndPedestalWidthFactor:float The pulse ends if the ADC value drops below 
   *         this factor timed the width of the pedestal (default: 3.0)
   *  @param MinimumPulseHeight:float To accept a pulse its maximal ADC entry must exceed this cut
   *         (default: 12.0 ADC counts)
   *  @param MinumumPulseLength:int To accept a pulse its length in time bins must be larger 
   *         than this value without Start/End bins (default: 5)
   *  @param SaveNBinsBeforeStart:int # bins saved before the start bin
   *         (exept there is another pulse, default:2)
   *  @param SaveNBinsAfterStart:int # bins saved after the start bin
   *         (exept there is another pulse, default 8)
   *  @param NoiseValuePedestalWidthFactor:float The noise (cut) value is calculated as this factor times pedestal width (default: 3.0)
   *  @param ElectronicsParameterObjectName:string Name of the conditions object in which the parameters of the read out electronics is stored
   *  @param ElectronicsParametersCollectionName:string Name of the conditions collection in which the parameters of the read out electronics is stored
   *  @param PedestalCollectionName:string Name of the conditions collection in which the pedestals are stored
   *  @param TimeCalculationMode:int Specifies the mode for time calculation. Valid values:<br>
   *                                 INFLEXION_POINT = 0 (default) <br>
   *                                 FIT_PARABOLA = 1 <br>
   *                                 CENTRE_OF_GRAVITY = 2 <br>
   *                                 FIRST_ABOVE_THRESHOLD = 3 <br>
   *                                 AT_MAXIMUM = 4 <br>
   *                                 CONSTANT_FRACTION = 5 <br>
   *                                 GAMMA_FIT = 6<br>
   *				     COG_ATMAX = 7<br>
   *
   *  @param ChargeCalculationMode:int Specifies the mode for charge calculation. Valid values:<br>
   *                                   SUMQ = 0 (default) <br>
   *                                   FIT_PARABOLA = 1 <br>
   *                                   MAXQ = 2
   *  @param PedestalValueOverride \c OPTIONAL: Manually set the pedestal value (in ADC counts)
   *  @param PedestalWidthOverride \c OPTIONAL: Manually set value of the pedestal width (in ADC coutns)
   *  @param MaximumADCValueOverride \c OPTIONAL: Manually set value of the maximum adc value
   *  @param IsPedestalSubtractedOverride \c OPTIONAL: Manually set whether the pedestal has been subtracted (true or false)
   *  @param SignalPolarityOverride \c OPTIONAL: Manually set value of the electronics polarity (+-1)
   *  @param ReadoutFrequencyOverride \c OPTIONAL: Manually set value of the readout frequency (in MHz)
   *  @param PlateauLengthCutOff \c OPTIONAL: The maximum length of a plateau within a pulse in time bins. This used to cut off strange pulses that go into a plateau.
   *
   * As the interpretation of the data changes with the charge calculation mode 
   * (all charge or max charge) or the time calculation mode (different offsets), the parameters
   * \c TimeCalculationMode and \c ChargeCalculationMode are stored as collection parameters of the
   * pulse collection (in addition to being written to the run header together with all other 
   * processor parameters).
   *
   * \attention The following parameters have been renamed:
   * \li IsPedestalSubstractedOverride (Was a typo, there should ne no 's' after the 'b')
   * \li PulseStartThreshold (Is not a threshold, new name is PulseStartPedestalWidthFactor)
   * \li PulseEndThreshold (Is not a threshold, new name is PulseEndPedestalWidthFactor)
   * \li PedestalWidthScale (New name is NoiseValuePedestalWidthFactor)
   *
   * \attention The following parameters have been replaced (versions with and without question mark):
   * \li UseParabolicFitAmplitudeAsCharge? (Use ChargeCalculationMode instead)
   * \li UseParabolicFitAmplitudeAsCharge (Use ChargeCalculationMode instead)
   * \li UseParabolicFitMaxAsTime? (Use TimeCalculationMode instead) 
   * \li UseParabolicFitMaxAsTime  (Use TimeCalculationMode instead) 
   *
   * The processor will not run if any of the old parameters is set in the steering file.
   * This avoids the old parameter being ignored and
   * silently being replaced by the default of the new parameter.
   * This would lead to unwanted behaviour and would be very difficult to identify.
   * <b>You can use the script tools/scripts/updatePulseFinderParameters.sh to automatically 
   *    update your steering files.</b>.
   *
   * \attention The ReadoutFrequencyOverride value has changed from Hz to MHz to be 
   * consistent with the conditions data. The processor will refuse to 
   * work if values above 1e5 (50 GHz) are detected in case old values in Hz have not been updated.
   *
   * @author Ch. Rosemann, DESY (original author)
   * @author Martin Killenberg
   */
  class PulseFinderProcessor : public marlin::Processor
  {

    public:

      virtual Processor*  newProcessor()
      {
        return new PulseFinderProcessor;
      }

      PulseFinderProcessor();

      virtual void init();

      virtual void processRunHeader(lcio::LCRunHeader* run);

      virtual void processEvent(lcio::LCEvent* evt);

      virtual void check(lcio::LCEvent* evt);

      virtual void end();

    protected:

      /// name of the collection, where input data are stored
      std::string  _inputColName;
      /// name of the collection, where TrackerData are stored
      std::string  _outputDataColName;
      /// name of the collection, where TrackerPulse are stored
      std::string  _outputPulseColName;
      /// the name of the pedestal collection
      std::string  _pedestalColName;

      /// Name of the conditions object in which the parameters of the read out electronics is stored
      std::string  _electrParaObjName;

      /// Name of the conditions collection in which the parameters of the read out electronics is stored
      std::string _electrParaColName;

      /// Set this flag to use the amplitude (maximum bin value) of the pulse as the charge
      //bool _use_amplitude_as_charge; // currently not implemented, but planned

      /// set the persistency flag of the output collection
      bool _outputIsPersistent;

      /// if not 0 the ADC spectrum will be saved for all pulses
      bool _forceSpectrumSave;
      /// if not 0 pulses will not be split
      bool _switchOffSplitting;
      int _chargeCalculationMode; ///< the mode to calulate the charge
      int _timeCalculationMode; ///< the mode to calulate the time      
      double _timeCalculationParameter; ///<parameter for time methods (different meaning for different methods)

      /// factor to calculate the start threshold
      float _pulseStartPedestalWidthFactor;
      /// factor to calculate the start threshold
      float _pulseEndPedestalWidthFactor;
      /// minimal value for the maximum adc value
      float _minimumPulseHeight;
      /// minimal length for a pulse in number of time bins
      int _minimumPulseLength;
      /// # of bins stored before the start threshold
      int _saveNBinsBeforeStart;
      /// # of bins stored after the end threshold
      int _saveNBinsAfterEnd;
      /// factor to calculate the noise cut
      float _noiseValuePedestalWidthFactor;

      // override variables

      /** maximum adc value, needed to determine if a pulse is maxed out
       */
      float _maximumADCValueOverride;

      /** The readout frequency of every readout group (in MHz). This overrides any information provided by LCCD
       */
      float _readoutFrequencyOverride;

      /** The pedestal value to be used for every hardware channel in every readout group. This overrides any information provided by LCCD
       */
      float _pedestalValueOverride;

      /** The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD
       */
      float _pedestalWidthOverride;

      /** Sets, if the data is already pedestal subtracted in the read-out electronics. This overrides any information provided by LCCD
       */
      bool _isPedestalSubtractedOverride;

      /** the polarity of all readout groups, overrides LCCD
         */
      int _polarityOverride;

      /**The maximum length of a plateau within a pulse in time bins. 
	 This used to cut off strange pulses that go into a plateau.
      */
      int _maxPlateauLength;

      PulseFinder * _pulseFinder; ///< The instance of the pulse finder object.

  };
}
#endif // PULSEFINDERPROCESSOR_H

