#ifndef ROWBASEDPADPULSEROADSEARCHPROCESSOR_H
#define ROWBASEDPADPULSEROADSEARCHPROCESSOR_H

#include "RowTripletBasedTrackFinderProcessor.h"

#include <marlin/Processor.h>
#include <marlin/Global.h>
#include <lcio.h>
#include <string>

#include <gear/TPCModule.h>

// LCIO:
#include "EVENT/TrackerPulse.h"

#include "TFile.h"
#include "TTree.h"
#include "TMath.h"
#include "TVectorD.h"
#include "TMatrixD.h"
#include "TMatrixDSym.h"

namespace marlintpc {
/// Simultaneous hit and track finding (pad row based).

/**
 * Input are the (position and measurement direction) of the TPC pad pulses,
 * a parametrization of the hit measurement errors
 * and module parameters (offset and extend) from GEAR.
 *
 * In each module (with pulses):
 *  - For all pulses the charge is normalized to the average pad charge in the neighbourhood (adjacent columns).
 *  - Pulses with a relative charge larger than 1. are called leading, the others trailing.
 *  - From pairs of leading pulses search roads are defined to build (segment) seeds which may share pulses.
 *  - Pulse association ambiguities are resolved (nearest seed wins).
 *  - Trailing pulses are added to the seeds (by road search) and ambiguties are resolved.
 *  - Hits are build (in rows similar to the RowBasedHitFinderProcessor) exploring the seed association information.
 *  - From the hits and seeds corresponding (track) segments are constructed.
 *
 * Segments from each module are combined with compatible segments from
 * the corresponding neighbouring modules to build track candidates (using only unique matches). As
 * reference point for the track candidates the first hit is used (optional: PCA).
 *
 * More details have been presented at the DESY FLC/TPC meeting on the 9th of February 2017.
 *
 * The hits on the track candidates are ordered by row number, no multiple
 * hits in a row are allowed. All matching criteria use chi2 cuts based on the
 * hit measurement errors.
 * The calculation of the measurement (pad) direction in XY uses
 * "module.getLocalPadLayout().getCoordinateType()" to select polar or cartesian pad row geometry.
 *
 * The default parameters value in '()' below correspond to B=1.0 T data, alternative settings for B=0 are given in '[]'.
 *
 * \param  "InputTrackerPulses":string  The name of the input collection of TPC (pad) pulses ("TPCPulses")
 * \param  "OutputTrackerHits":string   The name of the output collection with the found hits ("TPCHits")
 * \param  "OutputTracks":string        The name of the output collection with the found tracks ("TPCTracks")
 * \param  "BFieldScaleFactor":double   \c Optional parameter, scales magnetic field (map), use 1.0 (default) for field ON or 0.0 for field OFF (1.) [0.]
 * \param  "DriftVelocity":double  \c Optional parameter, drift velocity (74.)
 * \param  "MaxColDiffNeighbourhood":int  \c Optional parameter, maximal column difference to pulse in neighbourhood (1) [2]
 * \param  "MaxTimeDiff":double  \c Optional parameter, maximum drift time difference in neighboorhood/hit (80.)
 * \param  "MaxSeedsInRow":int  \c Optional parameter, maximum number of seeding pulses in row (10)
 * \param  "MinRowDiff":int  \c Optional parameter, minimum row difference for pulse pair seeding road (5)
 * \param  "MaxDxyRoad":double  \c Optional parameter, maximum interpolation deviation in XY (road search) (2.) [4.]
 * \param  "MaxDzRoad":double  \c Optional parameter, maximum interpolation deviation in Z (road search) (3.)
 * \param  "MinRowNum":int  \c Optional parameter, minimum number of rows for segment seed (6)
 * \param  "MinRowDensity":double  \c Optional parameter, minimum row 'density' for segment seed (0.8)
 * \param  "MaxVarXY":double  \c Optional parameter, maximum (average) XY variance (unweighted Chi2/ndf) for segment seed (0.25) [2.]
 * \param  "MaxOverlapFraction":double  \c Optional parameter, maximum overlap fraction of segment seeds (0.5)
 * \param  "MaxColDiffHit":int  \c Optional parameter, maximum distance of pad to seed in hit (if > 0) (0)
 * \param  "MinNumPulses":double  \c Optional parameter, minimum number of pulses in hit (2)
 * \param  "MaxColGap":double  \c Optional parameter, maximum column gap in hit (1)
 * \param  "OffsetCol":double   \c Optional parameter, hit variance: sigma0^2 in column direction (0.0085)
 * \param  "SlopeCol":double    \c Optional parameter, hit variance: diffusion in column direction (0.00011) [0.000397]
 * \param  "OffsetDrift":double   \c Optional parameter, hit variance: sigma0^2 in drift direction (0.25)
 * \param  "SlopeDrift":double    \c Optional parameter, hit variance: diffusion in drift direction (0.0)
 * \param  "OffsetRow":double   \c Optional parameter, hit variance: sigma0^2 in row direction (0.0)
 * \param  "SlopeRow":double    \c Optional parameter, hit variance: diffusion in row direction (0.0)
 * \param  "ChargeConversionFactor":double    \c Optional parameter, charge conversion factor from ADC values to primary electrons (1.)
 * \param  "SegmentMatchingChi2Cut":double    \c Optional parameter, Chi2/Ndf cut for segment matching (Ndf=4 for B off, 5 for B on) (30.)
 * \param  "ReferencePointAtPca":bool \c Optional parameter,  use PCA as reference point, else 1. hit (false)
 * \param  "SkipRoadSearch":bool \c Optional parameter, skip road search, hit finding only (false)
 * \param  "SkipMultiplePulses":bool \c Optional parameter, skip multiple pulse candidates as road search seeds (true)
 *
 * @author  C. Kleinwort (170310)
 * 
 * \par Credits:
 * The processor skeleton was generated by the script \c createProcessor.py
 *
 * \todo for track finding
 *   - definition of module neighbourhood from POLAR module extend ?
 *   - better resolve equivalent (indirectly matching) overlapping segments ?
 *
 * \todo for hit finding
 *   - implement pulse (charge, time and Z) calibration (from database)
 *   - implement dead and noisy pads (from database)
 *   - implement more sophisticated resolution of ambiguities and hit splitting
 *
 */
class RowBasedPadPulseRoadSearchProcessor: public marlin::Processor {
	public:

		virtual Processor* newProcessor() {
			return new RowBasedPadPulseRoadSearchProcessor;
		}

		RowBasedPadPulseRoadSearchProcessor();

		virtual void init();

		virtual void processRunHeader(EVENT::LCRunHeader* run);

		virtual void processEvent(EVENT::LCEvent* evt);

		virtual void check(EVENT::LCEvent* evt);

		virtual void end();

	protected:
		/* the place for protected and private member data and functions */
		std::string _inputTrackerPulsesColName; ///< Name of the TPc pulse input collection
		std::string _outputTrackerHitsColName;  ///< Name of the TPC hit output collection
		std::string _outputTrackColName;  ///< Name of the track output collection
		double _bfieldScaleFactor; ///< scale factor for magnetic field (default: 1.0)
		double _driftVelocity; ///< drift velocity
		int _maxColDiff; ///< maximal column difference to pulse in neighbourhood
		double _dtMax; ///< maximum drift time difference in neighboorhood/hit
		int _maxSeedsInRow; ///< maximum number of seeding pulses in row
		int _minRowDiff; ///< minimum row difference for pulse pair seeding road
		double _maxDxyRoad; ///< maximum interpolation deviation in XY (road search)
		double _maxDzRoad; ///< maximum interpolation deviation in Z (road search)
		int _minRows; ///< minimum number of rows for segment seed
		double _rowDensityCut; ///< minimum row 'density' for segment seed
		double _maxVarXY; ///< maximum (average) XY variance (unweighted Chi2/ndf) for segment seed
		double _maxOverlapFraction; ///< maximum overlap fraction of segment seeds
		int _maxColDiffHit; ///< maximum distance of pad to seed in hit (if > 0)
		int _minNumPulses; ///< minimum number of pulses in hit 
		int _maxColGap; ///< maximum column gap in hit
		double _offsetCol; ///< measurement variance: sigma0^2 in column direction
		double _slopeCol; ///< measurement variance: diffusion term in column direction
		double _offsetDrift; ///< measurement variance: sigma0^2 in drift direction
		double _slopeDrift; ///< measurement variance: diffusion term in drift direction
		double _offsetRow; ///< measurement variance: sigma0^2 in row direction (use rowHeight^2/12. if < 0.)
		double _slopeRow; ///< measurement variance: diffusion term in row direction
		double _chargeConversionFactor; ///< charge conversion factor from ADC values to primary electrons
		double _segCut; ///< Chi2/Ndf cut for segment matching (Ndf=4 for B off, 5 for B on)
		bool _refAtPCA; ///< Use Pca as reference point (else 1. hit)
		bool _skipRoadSearch; ///< skip road search, hit finding only
		bool _skipMultiplePulses; ///< skip multiple pulse candidates as road search seeds

	private:

		double _Bzc; ///< magnetic field strength (Bz*c)
		std::map<int, std::vector<int> > _modNeighbours; ///< neighbour modules
		double _Xcenter; ///< TPC center, X coordinate
		double _Ycenter; ///< TPC center, Y coordinate
		std::map<int, int> _moduleRowOffset; ///< row offsets per module layer

		bool areNeighbourModules(gear::TPCModule*, gear::TPCModule*, bool);
};

typedef std::list<int> seedIdListType;

/// Pad based pulse.
/**
 * Has identification and measurements with directions and errors.
 */
class pb_Pulse {
	public:
		pb_Pulse(const int iPulse, const EVENT::TrackerPulse& aPulse, const gear::TPCModule& aModule, const double zPos);
		void print() const;
		int getPulseNum() const;
		int getModule() const;
		int getPadId() const;
		int getRow() const;
		int getColumn() const;
		int getQuality() const;
		double getCharge() const;
		double getRelCharge() const;
		double getTime() const;
		double getX() const;
		double getY() const;
		double getZ() const;
		void getPos(double*) const;
		void getCOG(double*) const;
		void getMeasDir(double*) const;
		double getPhiMeas() const;
		bool isMaxPulse() const;
		bool inSeed() const;
		int getNumSeeds() const;
		void addSeed(int);
		void removeSeed(int);
		void clearSeeds();
		const seedIdListType& getSeedIdList() const;
		int getSeedId() const;
		void addNeighbour(pb_Pulse*);
		bool inHit() const;
		void addToHit();
		int getIndex() const;
		void setIndex(int);
		int getHitNum() const;
		void setHitNum(int);

	private:
		const int _pulseNum; ///< input pulse collection index
		const int _pulseModule; ///< module number
		const int _pulsePadID; ///< pad ID
		const int _pulseRow; ///< row number
		const int _pulseCol; ///< column number
		const int _pulseQuality; ///< quality
		const double _pulseCharge; ///< charge
		const double _pulseTime; ///< drift time
		const double _pulseX; ///< X position (local)
		const double _pulseY; ///< Y position (local)
		const double _pulseZ; ///< Z position
		const double _pulsePhiMeas; ///< measurement direction in XY (tangential to row)
		const double _pulseDirX; ///< measurement direction in X
		const double _pulseDirY; ///< measurement direction in Y
		int _firstCol; ///< start (column) of local neigbourhood
		int _lastCol; ///< end (column) of local neigbourhood
		double _maxCharge; ///< maximum charge in local neigbourhood
		double _sumCharge; ///< sum of charge in local neigbourhood
		double _sumChargeDist; ///< sum of charge*distance (to pulse) in local neigbourhood
		seedIdListType _seeds; ///< list of seeds (using pulse)
		bool _inHit; ///< flag for being part of a hit
		int _indexCol; ///< index in list of pulses per row  
		int _hitNum; ///< index of seeded hit
};

typedef std::vector<pb_Pulse*> pulseListType;
typedef std::map<int, pulseListType> rowPulseMapType;
typedef std::map<int, rowPulseMapType> modPulseMapType;

/// Pad based (track segment) seed.
/**
 * Has identification and measurements with directions and errors.
 */
class pb_Seed {
	public:
		pb_Seed(const int seedID, const double Bzc, const pulseListType& pulses, const int minRows, const double rowDensityCut,
				const double maxVarXY);
		pb_Seed(const int seedID, const double Bzc, const pulseListType& pulses);
		void print() const;
		int getSeedId() const;
		bool getValid() const;
		void setInvalid();
		int getNumSeedPulses() const;
		int getNumUniqueSeedPulses() const;
		bool hasAddedPulses() const;
		double getVarXY() const;
		void getRefPoint(double*) const;
		void getPar(double *) const;
		void markPulses();
		void freePulses();
		void addPulse(pb_Pulse*);
		const pulseListType& getPulseList() const;
		void roadSearch(pulseListType& pulses, pulseListType& insidePulses, const double maxDxy, const double maxDz) const;

	private:
		const int _seedId; ///< seed ID
		const double _bzc; ///< magnetic field strength (Bz*c)
		const bool _checked; ///< checked flag
		const unsigned int _numPulses; ///< number of pulses 
		pulseListType _pulseList; ///< pointers to pulses
		bool _valid; ///< validity flag
		double _refX; ///< X of reference point
		double _refY; ///< Y of reference point
		double _refZ; ///< Z of reference point
		int _npar; ///< number of parameters (5: helix, 4: line)
		double _varXY; ///< chi2/ndf from (unweighted) XY fit
		double _varZS; ///< chi2/ndf from (unweighted) ZS fit
		TVectorD _parameters; ///< parameter vector
		void _fitSeed();

};

typedef std::vector<pb_Seed*> seedListType;

/// Row based segment match.
/**
 * Has segment ID pair and row ranges.
 */
class rb_SegmentMatch {
	public:
		rb_SegmentMatch(const rb_Segment* segment1, const rb_Segment* segment2);
		void print() const;
		int getSegId1() const;
		int getSegId2() const;
		int getFirstRow1() const;
		int getFirstRow2() const;
		int getLastRow1() const;
		int getLastRow2() const;
		bool getValid() const;
		bool overlap(rb_SegmentMatch*) const;
		void setInvalid();

	private:
		const int _segId1; ///< first segment ID
		const int _segId2; ///< second segment ID
		const int _firstRow1; ///< first row of first segment
		const int _firstRow2; ///< first row of second segment
		const int _lastRow1; ///< last row of first segment
		const int _lastRow2; ///< last row of second segment
		bool _valid; ///< validity flag 
};

typedef std::vector<rb_SegmentMatch*> segMatchListType;
typedef std::map<int, segMatchListType> segMatchMapType;

} // namespace marlintpc
#endif // ROWBASEDPADPULSEROADSEARCHPROCESSOR_H
