#ifndef TrackFinderHoughTrafoProcessor_h
#define TrackFinderHoughTrafoProcessor_h 1

// C++
#include <string>
#include <map>
#include <vector>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "lccd/ConditionsMap.hh"

namespace AIDA
{

class IHistogram1D;
class IHistogram2D;


}

namespace marlintpc
{

/* the following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash
 */
/**
 *  uses the hough transformation to find hits on straight tracks.
 *  track parameters can be used as starting values for the TrackFitter
 *
 *  handling of conditions data is done by the ConditionsProcessor
 *
 *  <h4>Input - Prerequisites</h4>
 *  a collection of TrackerHits
 *
 *  <h4>Output</h4>
 *  collection of Tracks
 *
 * @param InputCollectionName Name of the input collection (default: TimePixHits)
 * @param OutputCollectionName Name of the output collection (default: TimePixTrackCandidates)
 * @param SetOutputTransient If not 0 the output collection is set transient (default: 0)
 *
 * @author Simone Zimmermann
 */

class TrackFinderHoughTrafoProcessor : public marlin::Processor
{

    public:

        /* This method will be called by the marlin package
         * It returns a processor of the currend type (here TrackFinderHoughTrafoProcessor)
         */
        virtual Processor*  newProcessor()
        {
            return new TrackFinderHoughTrafoProcessor ;
        }

        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        TrackFinderHoughTrafoProcessor() ;

        /* Called at the beginning of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        virtual void init() ;

        /* Called for every run.
         * in this processor it is not used
         */
        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        virtual void processEvent(lcio::LCEvent * evt) ;

        /* This method is only called if the check flag is set (default) in the main processor
         * (where the input files are given)
         * It should be used to create check plots which are not needed in a complete analysis
         * (where normaly the check flag is deactivated)
         */
        virtual void check(lcio::LCEvent * evt) ;


        /* Called after data processing for clean up.
         * e.g. saving control histogramms, delete created objects
         * (which are not stored in a collection/event), etc.
         */
        virtual void end() ;



        virtual double distanceToTrackFinder(const double *pos,
                                             DoubleVec par,
                                             std::string _trackType);
        /** nested helper class that stores the properties of a straight track.
         */
        struct TrackProperties
        {

            double a; ///< slope of track
            double b; ///< intersection of track on axis set by orientation
            int hitsontrack; ///< number of hits on the track
            int orientation; ///< orientation of the track: \n 0: y = ax + b \n 1: x = ya + b

            TrackProperties(double avalue = 0, double bvalue = 0,
                            int hitsontrackvalue = 0, int orientationvalue = 0)
            {

                a = avalue;
                b = bvalue;
                hitsontrack = hitsontrackvalue;
                orientation = orientationvalue;

            };

            bool operator<(TrackProperties const trackproperties) const
            {
                return (hitsontrack < trackproperties.hitsontrack);
            }

            bool operator>(TrackProperties const trackproperties) const
            {
                return (hitsontrack > trackproperties.hitsontrack);
            }

            bool operator==(TrackProperties const trackproperties) const
            {
                return(a == trackproperties.a &&
                       b == trackproperties.b &&
                       hitsontrack == trackproperties.hitsontrack);
            }




        };

        static bool sortPredicate(const TrackProperties& lhs, const TrackProperties& rhs)
        {

            return lhs.hitsontrack > rhs.hitsontrack;

        }

        virtual TrackProperties maxBinFinder(AIDA::IHistogram2D * histo);

        typedef std::vector<TrackProperties> TPVec;



    protected:

        /* normally the doxygen documentation is palced before the method/variable
         * but for one line comments in case of variables it may be more convenied for you
         * to place it after the variable. In this case you use an arrow to tell doxygen
         * to which variable the commet belongs to.
         */

        std::string _inputColName ;  ///<- the name of the input collection
        std::string _outputColName ; ///<- the name of the output collection
        std::string _trackType; ///<- the type of track, that is looked for
//  double _pitch; ///<- the pitch of the chip in mm
        double _maxDistanceToTrack; ///<- Maximum distance of a hit to the track in mm
        int _nHitsOnTrack; ///<- minimum nuber if hits on track
        int _chipID; ///<- number of current chip, 0 if single chip

        /** give the status which is set to the transient flag of the output collection
         * it is an int instead of a boolean, because the processor parameter cannot be a boolean
         */
        int _outputIsTransient;

        int _nRun ; ///<- counts the processed runs

        double a; ///<- parameter of hough transformation
        double b; ///<- parameter of hough transformation

        int _houghSpaceNBinsA; ///< The number of bins in the hough space histogram in 'a' direction
        int _houghSpaceNBinsB; ///< The number of bins in the hough space histogram in 'b' direction
        float _houghSpaceMaxA; ///< The half range (-max..max) of the hough space histogram in 'a' direction
        float _houghSpaceMaxB; ///< The half range (-max--max) of the hough space histogram in 'b' direction

    private:
        AIDA::IHistogram1D * histod;
        AIDA::IHistogram2D * HistoEvent;
        std::vector<AIDA::IHistogram2D *> HistoLineVec;
        std::vector<AIDA::IHistogram2D *> HistoTrackVec;
        std::vector<AIDA::IHistogram2D *> HistoHoughVec;


} ;

}

#endif



