#include "BiasCalculatorProcessor.h"
        
// MARLIN
#include <marlin/Exceptions.h>
#include "marlin/Global.h"
#include "IntersectionCalculator.h"
#include "streamlog/streamlog.h"

// LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "EVENT/Track.h"
#include "EVENT/TrackerHit.h"
#include "IMPL/TrackerHitImpl.h"
#include "EVENT/TrackerPulse.h"
#include "IMPL/LCCollectionVec.h"

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gear/PadRowLayout2D.h>
#include <gearimpl/GlobalPadIndex.h>

// ROOT 
#include <TDirectoryFile.h>

// STD
#include <Exceptions.h>
#include <iostream>
#include <fstream>
#include <cmath>

namespace marlintpc
{
  BiasCalculatorProcessor aBiasCalculatorProcessor;

  BiasCalculatorProcessor::BiasCalculatorProcessor() : Processor("BiasCalculatorProcessor")
  {
    _description = "Processor to analyze bias";

    registerInputCollection( LCIO::TRACK,
			     "TrackInputCollectionName",
			     "Name of the input track collection of final fit, default: FoundTracks",
			     _inputTracks, 
			     std::string("FoundTracks"));

    registerProcessorParameter("BiasOutputFileName_RPhi",
                              "Output file name for the rphi bias data",
                              _biasOutputFileName_rphi,
                              std::string("bias_rphi.dat"));

    registerProcessorParameter("BiasOutputFileName_Z",
                              "Output file name for the z bias data",
                              _biasOutputFileName_z,
                              std::string("bias_z.dat"));

    registerOptionalParameter("RootOutput",
                              "True if TProfiles should be saved.",
                              _rootOutput,
                              false);

    registerOptionalParameter("SingleTrack",
                              "True if only single-track events should be used.",
                              _singleTrack,
                              false);

    registerOptionalParameter("SmallAngle",
                              "True if only hits with local track angles < 5 degrees should be used.",
                              _smallAngle,
                              false);

    registerOptionalParameter("SeparateDistortions",
                              "True if bias calculation should remove the mean (i.e., the distortion) to separate the two effects (default: true).",
                              _separateDistortions,
                              true);

    registerOptionalParameter("RelativeBias",
                              "True if bias calculation should use x_relative as opposed to a simple local position (default: true).",
                              _relative,
                              true);

    registerOptionalParameter("NPulses",
                              "Optional: Restrict calculation to hits with this many pulses.",
                              _nPulses,
                              -1);
  }

  void BiasCalculatorProcessor::init()
  {
		if(parameterSet("NPulses"))
		{
			_selectPulses = true;
		}
		else
		{
			_selectPulses = false;
		}

		if(_rootOutput) 
		{
			marlin::AIDAProcessor::tree(this);

			//Create directories for the bias plots 
			_dir_rphi = new TDirectoryFile("BiasData_rphi", "BiasData_rphi");
			_dir_z = new TDirectoryFile("BiasData_z", "BiasData_z");

		}

  }

// FIXME: not used:  void BiasCalculatorProcessor::processRunHeader(lcio::LCRunHeader* run)
//  {
//  }

  void BiasCalculatorProcessor::processEvent(lcio::LCEvent* evt)
  {

    LCCollectionVec* trackCol;
		try
		{
			trackCol = static_cast<LCCollectionVec *>(evt->getCollection(_inputTracks));
		}
		catch (DataNotAvailableException &e)
		{
			streamlog_out(DEBUG) << "No data available in event for " <<_inputTracks << std::endl;
			return;
		}

    static const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

		//Cut on NTracks
		if(_singleTrack && trackCol->size() > 1) return;

		//Loop over tracks
		for(IMPL::LCCollectionVec::iterator trackIt = trackCol->begin(); trackIt != trackCol->end(); ++trackIt)
		{

			//Get the track
			EVENT::Track *track 		=		static_cast<EVENT::Track*>(*trackIt);

			//If this case fails, they should all fail, so give up right here.
			if(!track)
			{
				streamlog_out (ERROR) << "Error: The given collection " << _inputTracks << " appears not to contain EVENT::Track objects! " << std::endl;
				throw marlin::StopProcessingException(this);
			}

			//Get the Track Hits
			const EVENT::TrackerHitVec & trackHits = track->getTrackerHits();

			for(TrackerHitVec::const_iterator hitIt = trackHits.begin(); hitIt != trackHits.end(); ++hitIt)
			{
				const EVENT::TrackerHit* aHit = static_cast<const EVENT::TrackerHit*>(*hitIt);

				const LCObjectVec& pulses = aHit->getRawHits();

				//If the pulses aren't there, we can't calculate the bias. (Well we can, but for consistency, we won't).
				if(pulses.size() == 0 || !pulses[0])
				{
					std::cout << "WARNING: In Event " << evt->getEventNumber() << " for Hit " << ( hitIt - trackHits.begin() )
						<< ": The vector that should contain the pulse list is empty. Pulses are needed for the bias calculation." << std::endl;
					throw marlin::StopProcessingException(this);
				}

				//Cut on number of pulses, if requested. This is good for examining possible systematic effects due to hit size
				if(_selectPulses && pulses.size() != (unsigned int)_nPulses)
					continue;

				EVENT::TrackerPulse* aPulse = static_cast<EVENT::TrackerPulse*>(*( pulses.begin() ));

				int moduleID = aPulse->getCellID1();
				int rowID = 0;

				const gear::TPCModule& theModule = tpcParameters.getModule(moduleID);
				rowID = theModule.getRowNumber(aPulse->getCellID0());


				//Cut on local angle
				if(_smallAngle)
				{
					//calculate angle between hit and track
					int padIndex = theModule.getNearestPad(aHit->getPosition()[0],aHit->getPosition()[1]);
					gear::Vector2D padCenterGlobal = theModule.getPadCenter(padIndex);
					gear::Vector2D e_pad (padCenterGlobal[0]-theModule.getOffset()[0],padCenterGlobal[1]-theModule.getOffset()[1]);
					double phi_s=atan2(sin(track->getPhi())-track->getOmega()*(aHit->getPosition()[0]-(-track->getD0()*sin(track->getPhi())+track->getReferencePoint()[0])),
							cos(track->getPhi())+track->getOmega()*(aHit->getPosition()[1]-(track->getD0()*cos(track->getPhi())+track->getReferencePoint()[1])));

					double  phiHit=phi_s-atan(e_pad[1]/e_pad[0]);

					if(phiHit > 5.*M_PI/180. || phiHit < -5.*M_PI/180.) continue;
				}

				std::pair<int, int> key(moduleID, rowID);

				//// R-PHI ///
				if(_rootOutput) _dir_rphi->cd();
				std::map< std::pair<int,int>, TProfile*>::iterator histIter;
				histIter = _histograms_rphi.find(key);
				TProfile* tempHist = NULL;

				//If there isn't yet an entry for this module and row, make one;
				if(histIter == _histograms_rphi.end())
				{
					std::stringstream title;
					title << "rphi Bias for Module " << moduleID << " and Row " << rowID;
					if(_relative)
					{
						tempHist = new TProfile(title.str().c_str(), title.str().c_str(), 50, -0.5, 0.5);
						tempHist->SetXTitle("x_relative [ (x_hit - x_pad)/distance_between_pads ]");
					}
					else
					{
						//tempHist = new TProfile(title.str().c_str(), title.str().c_str(), 500, theModule.getLocalModuleExtent()[2], theModule.getLocalModuleExtent()[3]);
						tempHist = new TProfile(title.str().c_str(), title.str().c_str(), 50, 0., 0.);
						tempHist->SetXTitle("x_local");
					}
					tempHist->SetYTitle("Residual (mm)");
					_histograms_rphi.insert(std::make_pair(key,tempHist));
				}
				//Otherwise, retrieve the histogram
				else
				{
					tempHist = (*histIter).second;	
				}
				std::pair<double, double> bias;
			 	try
				{
					bias = calculateBias_rphi(aHit, track);
				}
				catch(std::exception &e)
				{
					continue;
				}
				tempHist->Fill( bias.first, bias.second );
				////////////

				//// Z ///
				if(_rootOutput) _dir_z->cd();
				histIter = _histograms_z.find(key);
				tempHist = NULL;

				//If there isn't yet an entry for this module and row, make one;
				if(histIter == _histograms_z.end())
				{
					std::stringstream title;
					title << "z Bias for Module " << moduleID << " and Row " << rowID;
					if(_relative)
					{
						tempHist = new TProfile(title.str().c_str(), title.str().c_str(), 50, -0.5, 0.5);
						tempHist->SetXTitle("x_relative [ (x_hit - x_pad)/distance_between_pads ]");
					}
					else
					{
						tempHist = new TProfile(title.str().c_str(), title.str().c_str(), 50, 0., 0.);
						tempHist->SetXTitle("x_local");
					}
					tempHist->SetYTitle("Residual (mm)");
					_histograms_z.insert(std::make_pair(key,tempHist));
				}
				//Otherwise, retrieve the histogram
				else
				{
					tempHist = (*histIter).second;	
				}
			 	try
				{
					bias = calculateBias_z(aHit, track);
				}
				catch(std::exception &e)
				{
					continue;
				}
				tempHist->Fill( bias.first, bias.second );
				////////////

			}//hit loop

		}//track loop

  }//processEvent()
  
//FIXME: not used:  void BiasCalculatorProcessor::check(lcio::LCEvent* evt)
//  {
//  }

  void BiasCalculatorProcessor::end()
  {

		//If desired, remove the distortion so we only consider the actual bias shape
		if(_separateDistortions)
		{
			//// R-PHI ////
			for(std::map<std::pair<int, int>, TProfile*>::iterator histIt = _histograms_rphi.begin(); 
					histIt != _histograms_rphi.end(); ++histIt)
			{
				double distortion = (*histIt).second->GetMean(2);

				for(int i = 1; i <= (*histIt).second->GetNbinsX(); i++)
					(*histIt).second->GetArray()[i] -= distortion*(*histIt).second->GetBinEntries(i);

				(*histIt).second->SetMinimum(-0.5);
				(*histIt).second->SetMaximum(0.5);

			}
			///////////////
			////// Z //////
			for(std::map<std::pair<int, int>, TProfile*>::iterator histIt = _histograms_z.begin(); 
					histIt != _histograms_z.end(); ++histIt)
			{
				double distortion = (*histIt).second->GetMean(2);

				for(int i = 1; i <= (*histIt).second->GetNbinsX(); i++)
					(*histIt).second->GetArray()[i] -= distortion*(*histIt).second->GetBinEntries(i);

				(*histIt).second->SetMinimum(-0.5);
				(*histIt).second->SetMaximum(0.5);

			}
			///////////////
		}

		//Write the bias data to file
		writeBias();
		
		calculateAverageBias();

		//If no profiles are desired, delete the TProfiles
		if(!_rootOutput)
		{	
			for(std::map<std::pair<int, int>, TProfile*>::iterator histIt = _histograms_rphi.begin(); 
					histIt != _histograms_rphi.end(); ++histIt)
			{
				if((*histIt).second)
				{
					delete (*histIt).second;
					(*histIt).second = NULL; 
				}
			}
			for(std::map<std::pair<int, int>, TProfile*>::iterator histIt = _histograms_z.begin(); 
					histIt != _histograms_z.end(); ++histIt)
			{
				if((*histIt).second)
				{
					delete (*histIt).second;
					(*histIt).second = NULL; 
				}
			}
		}

  }
  void BiasCalculatorProcessor::calculateAverageBias()
	{
		//This method calculates a measure of the average bias as a function of row radius, and produces a total figure.
		static const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

		//Change directory
		marlin::AIDAProcessor::tree(this);
    _biasTuple = new TNtuple("Total Bias Tuple","Total Bias Tuple","bias_rphi:bias_err_rphi:bias_z:bias_err_z");

		//Return values
		double sum_rphi = 0., sum_rphi2 = 0., sum_rphi_w = 0., sum_rphi_w2 = 0.;
		double sum_z = 0., sum_z2 = 0., sum_z_w = 0., sum_z_w2 = 0.;
		std::vector<double> rows;
		std::vector<double> rows_err;
		std::vector<double> average_rphi;
		std::vector<double> average_rphi_err;
		std::vector<double> average_z;
		std::vector<double> average_z_err;

		//Calculate rphi
		for(std::map<std::pair<int, int>, TProfile*>::const_iterator histIt = _histograms_rphi.begin();
				histIt != _histograms_rphi.end(); ++histIt)
		{
			int module = ((*histIt).first).first;
			int row = ((*histIt).first).second;

			const gear::TPCModule& theModule = tpcParameters.getModule(module);
			gear::Vector2D padcenter = theModule.getPadCenter(theModule.getPadIndex(row, 0));

			double radius = sqrt(padcenter[0]*padcenter[0]+padcenter[1]*padcenter[1]);
			rows.push_back(radius);
			rows_err.push_back(0.);

			double sum_x = 0., sum_x2 = 0., sum_w = 0., sum_w2 = 0.;
			double sum_plus = 0., sum_minus = 0.; //Used for determining the sign 

			for(int i = 1; i <= ((*histIt).second)->GetNbinsX(); i++)
			{
				double x_rel = ((*histIt).second)->GetBinLowEdge(i);
				x_rel += ((*histIt).second)->GetBinWidth(i)/2.;
				double bias = ((*histIt).second)->GetBinContent(i);
				double biasErr = ((*histIt).second)->GetBinError(i);
				if(biasErr == 0) continue;
				double weight = 1./(biasErr*biasErr);

				if(x_rel < 0) sum_minus -= bias; //I define a positive average bias as one mostly negative when x < 0, 
				else sum_plus += bias;					 //and mostly positive when x > 0, as with the sine function.

				sum_x += fabs(bias)*weight;
				sum_x2 += bias*bias*weight;
				sum_w += weight;
				sum_w2 += weight*weight;
			}

			double av_x = sum_x/sum_w;
			if(sum_plus + sum_minus == 0) average_z.push_back(av_x);
			else average_rphi.push_back(av_x*(sum_plus + sum_minus)/fabs(sum_plus + sum_minus));
			double sigma2_x = sum_w*(sum_x2 - sum_x*av_x)/(sum_w*sum_w - sum_w2);
			average_rphi_err.push_back(sqrt(sigma2_x));

			if(sigma2_x == 0) { sum_z_w += 1; sum_z_w2 += 1; continue; }
			sum_rphi += av_x/sigma2_x;
			sum_rphi2 += av_x*av_x/sigma2_x;
			sum_rphi_w += 1./sigma2_x;
			sum_rphi_w2 += 1./(sigma2_x*sigma2_x);
		}

		//Calculate z 
		for(std::map<std::pair<int, int>, TProfile*>::const_iterator histIt = _histograms_z.begin();
				histIt != _histograms_z.end(); ++histIt)
		{

			double sum_x = 0., sum_x2 = 0., sum_w = 0., sum_w2 = 0.;
			double sum_plus = 0., sum_minus = 0.; //Used for determining the sign 

			for(int i = 1; i <= ((*histIt).second)->GetNbinsX(); i++)
			{
				double x_rel = ((*histIt).second)->GetBinLowEdge(i);
				x_rel += ((*histIt).second)->GetBinWidth(i)/2.;
				double bias = ((*histIt).second)->GetBinContent(i);
				double biasErr = ((*histIt).second)->GetBinError(i);
				if(biasErr == 0) continue;
				double weight = 1./(biasErr*biasErr);

				if(x_rel < 0) sum_minus -= bias;
				else sum_plus += bias;

				sum_x += fabs(bias)*weight;
				sum_x2 += bias*bias*weight;
				sum_w += weight;
				sum_w2 += weight*weight;
			}

			double av_x = sum_x/sum_w;
			if(sum_plus + sum_minus == 0) average_z.push_back(av_x);
			else average_z.push_back(av_x*(sum_plus + sum_minus)/fabs(sum_plus + sum_minus));
			double sigma2_x = sum_w*(sum_x2 - sum_x*av_x)/(sum_w*sum_w - sum_w2);
			average_z_err.push_back(sqrt(sigma2_x));

			if(sigma2_x == 0) { sum_z_w += 1; sum_z_w2 += 1; continue; }
			sum_z += av_x/sigma2_x;
			sum_z2 += av_x*av_x/sigma2_x;
			sum_z_w += 1./sigma2_x;
			sum_z_w2 += 1./(sigma2_x*sigma2_x);
		}

	  double total_rphi = sum_rphi/sum_rphi_w;
		double total_rphi_err2 = sum_rphi_w*(sum_rphi2 - sum_rphi*total_rphi)/(sum_rphi_w*sum_rphi_w - sum_rphi_w2);
	  double total_z = sum_z/sum_z_w;
		double total_z_err2 = sum_z_w*(sum_z2 - sum_z*total_z)/(sum_z_w*sum_z_w - sum_z_w2);

		_biasTuple->Fill(total_rphi, sqrt(total_rphi_err2), total_z, sqrt(total_z_err2));

		_rphiGr = new TGraphErrors(rows.size(), &(rows[0]), &(average_rphi[0]), &(rows_err[0]), &(average_rphi_err[0]));
		_rphiGr->SetNameTitle("Signed Average Absolute Rphi Bias by Row", "Signed Average Absolute RPhi Bias by Row");
		_rphiGr->GetXaxis()->SetTitle("Row Radius (mm)");
		_rphiGr->GetYaxis()->SetTitle("Signed Average Absolute Bias (mm)");
		_rphiGr->Write();

		_zGr = new TGraphErrors(rows.size(), &(rows[0]), &(average_z[0]), &(rows_err[0]), &(average_z_err[0]));
		_zGr->SetNameTitle("Signed Average Absolute Z Bias by Row", "Signed Average Absolute Z Bias by Row");
		_zGr->GetXaxis()->SetTitle("Row Radius (mm)");
		_zGr->GetYaxis()->SetTitle("Signed Average Absolute Bias (mm)");
		_zGr->Write();

	}
  void BiasCalculatorProcessor::writeBias()
  {
    
		//Open the output file -- rphi
    std::ofstream outfile;
    outfile.open(_biasOutputFileName_rphi.c_str());

		outfile << "-- Bias Data RPhi --" << std::endl << std::endl;
    
		for(std::map<std::pair<int, int>, TProfile*>::const_iterator histIt = _histograms_rphi.begin();
				histIt != _histograms_rphi.end(); ++histIt)
		{
			int module = ((*histIt).first).first;
			int row = ((*histIt).first).second;

			outfile << module << " " << row << std::endl << std::endl;

			std::string delim = "";

			for(int i = 1; i <= ((*histIt).second)->GetNbinsX(); i++)
			{
				double x_rel = ((*histIt).second)->GetBinLowEdge(i);
				x_rel += ((*histIt).second)->GetBinWidth(i)/2.;
				double bias = ((*histIt).second)->GetBinContent(i);
				double biasErr = ((*histIt).second)->GetBinError(i);

				outfile << delim << x_rel << " " << bias << " " << biasErr;

				delim = ", ";
			}

			outfile << std::endl << std::endl;
		}

		outfile.close();

		////////////////////////////////////////

		//Open the output file -- z 
    outfile.open(_biasOutputFileName_z.c_str());

		outfile << "-- Bias Data Z --" << std::endl << std::endl;
    
		for(std::map<std::pair<int, int>, TProfile*>::const_iterator histIt = _histograms_z.begin();
				histIt != _histograms_z.end(); ++histIt)
		{
			int module = ((*histIt).first).first;
			int row = ((*histIt).first).second;

			outfile << module << " " << row << std::endl << std::endl;

			std::string delim = "";

			for(int i = 1; i <= ((*histIt).second)->GetNbinsX(); i++)
			{
				double x_rel = ((*histIt).second)->GetBinLowEdge(i);
				x_rel += ((*histIt).second)->GetBinWidth(i)/2.;
				double bias = ((*histIt).second)->GetBinContent(i);
				double biasErr = ((*histIt).second)->GetBinError(i);

				outfile << delim << x_rel << " " << bias << " " << biasErr;

				delim = ", ";
			}

			outfile << std::endl << std::endl;
		}

		outfile.close();
	}
  
	std::pair<double, double> BiasCalculatorProcessor::calculateBias_rphi(const EVENT::TrackerHit * hit, const EVENT::Track * track)
	{

    static const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();
		std::pair<double, double> retVal;

		//The following is taken almost verbatim out of the ModuleDistortionCalculator, just with the addition of the x_relative calculation
		
    const gear::TPCModule& theModule = tpcParameters.getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
    gear::Vector2D hitLocal = theModule.globalToLocal(hit->getPosition()[0],hit->getPosition()[1]);
		int nearestPad = theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]); 
		gear::Vector2D padCenterGlobal = theModule.getPadCenter(nearestPad); //Moved these outside the if-statments
		gear::Vector2D padCenterLocal = theModule.globalToLocal(padCenterGlobal[0],padCenterGlobal[1]);
    double distrow = 0.;
		double x_rel = 0.; //Adding x_rel

		if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)//if circular pad layout
		{

			////Calculate x_rel for circular layout
			if(_relative)
			{
				double x_width = theModule.getPadWidth(nearestPad);
				x_rel = (hitLocal[1] - padCenterLocal[1])/x_width;
			}
			else
			{
				x_rel = hitLocal[1];
			}
			////

			streamlog_out(DEBUG)<<"Polar Coordinates"<<std::endl;
			double Radius = padCenterLocal[0];

			//distinguish between straight track and helix track
			std::pair<double,double> phi_intersec(0,0); 
			if(fabs(track->getOmega())<0.000000001)
			{
				try{
					//calculate two points on the track and transform them into local padplane coordinates to obtain a track parametrisation
					// in local coordionates
					//point1=point of closest approach
					double pca_x = ( -1.0 * track->getD0() ) * std::sin(track->getPhi())+track->getReferencePoint()[0];
					double pca_y = ( track->getD0() ) * std::cos(track->getPhi())+track->getReferencePoint()[1];
					//point 2, if we are on a straight line just move one step along the line
					double p2x=pca_x+std::cos(track->getPhi());
					double p2y=pca_y+std::sin(track->getPhi());
					//transform both points to local coordinate system
					//this gives back r and phi, not x and y for polar padlayout!
					gear::Vector2D pca_local = theModule.globalToLocal(pca_x,pca_y);
					gear::Vector2D p2_local = theModule.globalToLocal(p2x,p2y);
					double pca_local_x=pca_local[0]*cos(pca_local[1]);
					double pca_local_y=pca_local[0]*sin(pca_local[1]);
					double p2_local_x=p2_local[0]*cos(p2_local[1]);
					double p2_local_y=p2_local[0]*sin(p2_local[1]);
					double phi=atan2(p2_local_y-pca_local_y,p2_local_x-pca_local_x);
					phi_intersec= marlintpc::IntersectionCalculator::circle_straightLine(Radius, pca_local_x,pca_local_y,phi);
				}
				catch(std::exception &e) {
					std::cout<<"NO INTERSECTION STRAIGHT LINE: " <<Radius<<" "
						<<track->getOmega()<<" "<< track->getD0()<<" "<<track->getPhi()<<std::endl;
					//throw it again to be availbale in code that uses function
					throw e;
				}
			}
			else
			{
				//center of helix in global coordinates
				double xm, ym;
				xm = track->getReferencePoint()[0]+(1. / track->getOmega() - track->getD0()) * sin(track->getPhi());
				ym = track->getReferencePoint()[1]-(1. / track->getOmega()  - track->getD0()) * cos(track->getPhi());
				//transform helix center into local coordinate system
				gear::Vector2D helixCenterLocal = theModule.globalToLocal(xm,ym);
				//local comes on in polar, go back to cartesian
				double xm_l=helixCenterLocal[0]*cos(helixCenterLocal[1]);
				double ym_l=helixCenterLocal[0]*sin(helixCenterLocal[1]);

				try{
					phi_intersec= marlintpc::IntersectionCalculator::circle_circle(Radius, xm_l, ym_l,fabs(1. / track->getOmega()));
				}
				catch(std::exception &e) {
					std::cout<<"NO INTERSECTION CIRCLE: " <<Radius<<" "<<xm_l<<" "<< ym_l<<" : "
						<<track->getOmega()<<" "<< track->getD0()<<" "<<track->getPhi()<<std::endl;
					//throw it again to be availbale in code that uses function
					throw e;
				}
			}
			distrow=(hitLocal[1]-phi_intersec.first)*Radius;
			if(fabs((hitLocal[1]-phi_intersec.second))*Radius<fabs(distrow))
				distrow=(hitLocal[1]-phi_intersec.second)*Radius;
		}
		else if((theModule.getLocalPadLayout()).getPadLayoutType() == 1)//rectangular
		{
			streamlog_out(DEBUG)<<"Rectangular pad layout"<<std::endl;
			std::pair<double,double> intersec(0,0); 
			//get y and center of measurement layer in global coordinates

			////Calculate x_rel for circular layout
			if(_relative)
			{
				double x_width = theModule.getPadWidth(nearestPad);
				x_rel = (hitLocal[0] - padCenterLocal[0])/x_width;
			}
			else
			{
				x_rel = hitLocal[0];
			}
			////

			double Ylocal = padCenterLocal[1];   //center of pad row
			//distinguish between straight track and helix track
			if(fabs(track->getOmega())<0.000000001)
			{
				// 2 points on the track in local coordinates
				//point1=point of closest approach
				double pca_x = ( -1.0 * track->getD0() ) * std::sin(track->getPhi())+track->getReferencePoint()[0];
				double pca_y = ( track->getD0() ) * std::cos(track->getPhi())+track->getReferencePoint()[1];
				//point 2, if we are on a straight line just move one step along the line
				double p2x=pca_x+std::cos(track->getPhi());
				double p2y=pca_y+std::sin(track->getPhi());
				//transform both points to local coordinate system
				gear::Vector2D p2_local = theModule.globalToLocal(p2x,p2y);
				gear::Vector2D pca_local = theModule.globalToLocal(pca_x,pca_y);
				//double inclination=atan2(p2_local_y-pca_local_y,p2_local_x-pca_local_x);
				double inclination=(p2_local[0]-pca_local[0])/(p2_local[1]-pca_local[1]);	    
				double xAxisIntersect=pca_local[0]-inclination*pca_local[1];

				double x_intersec= marlintpc::IntersectionCalculator::horizontalLine_straightLine(Ylocal,inclination, xAxisIntersect);
				distrow= hitLocal[0]-x_intersec;
			}	
			else
			{
				double xm, ym;
				xm = track->getReferencePoint()[0]+(1. / track->getOmega() - track->getD0()) * sin(track->getPhi());
				ym = track->getReferencePoint()[1]-(1. / track->getOmega()  - track->getD0()) * cos(track->getPhi());
				//transform helix center into local coordinate system
				gear::Vector2D helixCenterLocal = theModule.globalToLocal(xm,ym);
				intersec= marlintpc::IntersectionCalculator::horizonalLine_circle(Ylocal, helixCenterLocal[0],helixCenterLocal[1] , fabs(1. / track->getOmega()));
				//intersection give the two x coordinates where circle intesects line. Coose the minimal distance.
				distrow=hitLocal[0]-intersec.first;
				if(fabs(hitLocal[0]-intersec.first)>fabs(hitLocal[0]-intersec.second))
					distrow=hitLocal[0]-intersec.second;
			}
		}
		else
		{
			streamlog_out(ERROR)<<"Error: Invalid pad layout type!"<<std::endl;
			throw gear::Exception("Unknown pad layout type");
		}

		retVal.first  = x_rel;
		retVal.second = distrow;

		return retVal;
	}

	std::pair<double, double> BiasCalculatorProcessor::calculateBias_z(const EVENT::TrackerHit * hit, const EVENT::Track * track)
	{
		static const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();
		std::pair<double, double> retVal;

		double x_rel = 0.; //Adding x_rel
    const gear::TPCModule& theModule = tpcParameters.getNearestModule(hit->getPosition()[0],hit->getPosition()[1]);
		int nearestPad = theModule.getNearestPad(hit->getPosition()[0],hit->getPosition()[1]); 
		gear::Vector2D padCenterGlobal = theModule.getPadCenter(nearestPad); //Moved these outside the if-statments
		gear::Vector2D padCenterLocal = theModule.globalToLocal(padCenterGlobal[0],padCenterGlobal[1]);
		gear::Vector2D hitLocal = theModule.globalToLocal(hit->getPosition()[0],hit->getPosition()[1]);

		if((theModule.getLocalPadLayout()).getPadLayoutType() == 2)//if circular pad layout
		{

			////Calculate x_rel for circular layout
			if(_relative)
			{
				double x_width = theModule.getPadWidth(nearestPad);
				x_rel = (hitLocal[1] - padCenterLocal[1])/x_width;
			}
			else
			{
				x_rel = hitLocal[1];
			}
			////

		}
		else if((theModule.getLocalPadLayout()).getPadLayoutType() == 1)//rectangular
		{
			////Calculate x_rel for circular layout
			if(_relative)
			{
				double x_width = theModule.getPadWidth(nearestPad);
				x_rel = (hitLocal[0] - padCenterLocal[0])/x_width;
			}
			else
			{
				x_rel = hitLocal[0];
			}
			////
		}


		//Again, the following is taken almost verbatim from the ModuleDistortionCalculator

		double pca_x = ( -1.0 * track->getD0() ) * std::sin(track->getPhi())+track->getReferencePoint()[0];
		double pca_y = ( track->getD0() ) * std::cos(track->getPhi())+track->getReferencePoint()[1];

		Double_t s = 0.;

		if(fabs(track->getOmega()) < 0.000000001)
		{
			s = std::sqrt(
					static_cast<long double>(( hit->getPosition()[0] - pca_x ) * ( hit->getPosition()[0] - pca_x )
						+ ( hit->getPosition()[1] - pca_y ) * ( hit->getPosition()[1] - pca_y )));

			//make sure the sign of s is right
			if(( ( ( -1.25 * M_PI ) <= track->getPhi() ) && ( track->getPhi() <= ( -0.75 * M_PI ) ) )
					|| ( ( ( -0.25 * M_PI ) <= track->getPhi() ) && ( track->getPhi() <= ( 0.25 * M_PI ) ) )
					|| ( ( ( 0.75 * M_PI ) <= track->getPhi() ) && track->getPhi() <= ( 1.25 * M_PI ) ))
			{
				if(pca_x <  hit->getPosition()[0])
				{
					s = ( -1.0 ) * s;
				}
			}
			else
			{
				if(pca_y <  hit->getPosition()[1])
				{
					s = ( -1.0 ) * s;
				}
			}
			s = (-1.)*s; //Needed to make tracks look correctly in CEDViewer 
		}
		else
		{
			Double_t r = 1. / track->getOmega();
			//center of circle
			Double_t xm, ym;
			xm = (r - track->getD0()) * sin(track->getPhi())+track->getReferencePoint()[0];
			ym = -(r - track->getD0()) * cos(track->getPhi())+track->getReferencePoint()[1];

			s = r * ( atan2(pca_y - ym , pca_x - xm)
					-
					atan2( hit->getPosition()[1]- ym ,  hit->getPosition()[0] - xm)
					);

			double s_1, s_2;
			s_1 = s;
			if(s < 0)
			{
				s_2 = s + fabs(2.*M_PI/track->getOmega());
			}
			else
			{
				s_2 = s - fabs(2.*M_PI/track->getOmega());
			}

			if(fabs(s_1) <= fabs(s_2)) s = s_1;
			else s = s_2;
		}

		double distz =  track->getTanLambda() * s +  track->getZ0()+track->getReferencePoint()[2] - hit->getPosition()[2];

		retVal.first = x_rel;
		retVal.second = -1.*distz;

		return retVal; 

	}  

}//end namespace
