#include "BiasCorrectorProcessor.h"
        
// MARLIN
#include <marlin/Exceptions.h>
#include "marlin/Global.h"
#include "IntersectionCalculator.h"
#include "streamlog/streamlog.h"

// LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerHit.h"
#include "IMPL/TrackerHitImpl.h"
#include "IMPL/LCCollectionVec.h"
#include <IMPL/LCFlagImpl.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gear/PadRowLayout2D.h>
#include <gearimpl/GlobalPadIndex.h>

// STD
#include <Exceptions.h>
#include <iostream>
#include <fstream>
#include <cmath>

// GSL
#include <gsl/gsl_spline.h>

// OTHER
#include "LCObjectCopier.h"

namespace marlintpc
{
  BiasCorrectorProcessor aBiasCorrectorProcessor;

  BiasCorrectorProcessor::BiasCorrectorProcessor() : Processor("BiasCorrectorProcessor")
  {
    _description = "Processor to correct bias";

    registerInputCollection( LCIO::TRACKERHIT,
			     "HitInputCollectionName",
			     "Name of the input hit collection to be corrected, default: TPCHits",
			     _inputHits, 
			     std::string("TPCHits"));

    registerOutputCollection( LCIO::TRACKERHIT,
			     "HitOutputCollectionName",
			     "Name of the output hit collection, default: TPCHitsCorrected",
			     _outputHits, 
			     std::string("TPCHitsCorrected"));

    registerProcessorParameter("BiasInputFileName_RPhi",
                              "Input file name for the rphi bias data",
                              _biasInputFileName_rphi,
                              std::string("bias_rphi.dat"));

    registerOptionalParameter("BiasInputFileName_Z",
                              "Input file name for the z bias data",
                              _biasInputFileName_z,
                              std::string("bias_z.dat"));

    registerOptionalParameter("SplineType",
                              "Type of spline to use: 0 = Cubic spline, 1 = Polynomial, 2 = Akima (Default: 0)",
                              _spline,
                              0);

    registerOptionalParameter("RelativeBias",
                              "True if bias calculation should use x_relative as opposed to a simple local position (default: true).",
                              _relative,
                              true);

  }

  void BiasCorrectorProcessor::init()
  {
		//Do rphi first, z is optional
		std::ifstream inputFile(_biasInputFileName_rphi.c_str());

		//Verify that the file is open, and is valid
		std::string headerString;
		if(inputFile.is_open())
		{
			getline(inputFile, headerString);	
		}
		if(!inputFile.is_open() || headerString.compare("-- Bias Data RPhi --") != 0)
		{
			streamlog_out(ERROR) << "Error : Input file " << _biasInputFileName_rphi << " does not appear to contain valid bias data. Stopping execution." << std::endl;
			throw marlin::StopProcessingException(this);
		}

		//Fill the data
		fillData(inputFile, 0);

		inputFile.close();

		//Now do Z, if requested
		if(parameterSet("BiasInputFileName_Z"))
		{
			inputFile.open(_biasInputFileName_z.c_str());

			//Verify that the file is open, and is valid
			std::string check2;
			if(inputFile.is_open())
			{
				getline(inputFile, check2);	
			}
			if(!inputFile.is_open() || check2.compare("-- Bias Data Z --") != 0)
			{
				streamlog_out(ERROR) << "Error : Input file " << _biasInputFileName_z << " does not appear to contain valid bias data. Stopping execution." << std::endl;
				throw marlin::StopProcessingException(this);
			}

			//Fill the data
			fillData(inputFile, 1);

			inputFile.close();
		}	
	}

//FIXME: not used:  void BiasCorrectorProcessor::processRunHeader(lcio::LCRunHeader* run)
//  {
//  }

  void BiasCorrectorProcessor::processEvent(lcio::LCEvent* evt)
  {

		//Get the input hits
    LCCollectionVec* inputHitCol;
		try
		{
			inputHitCol = static_cast<LCCollectionVec *>(evt->getCollection(_inputHits));
		}
		catch (DataNotAvailableException &e)
		{
			streamlog_out(DEBUG) << "No data available in event " << evt->getEventNumber() <<  " for " << _inputHits << std::endl;
			return;
		}

		//Prepare the output hits collection
		LCCollectionVec* outputHitCol = new LCCollectionVec(LCIO::TRACKERHIT);
		//outputHitCol->setTransient(!_outputHitsPersistent);
		outputHitCol->setTransient(false);
		// set the flags that cellID1 should be stored
		lcio::LCFlagImpl trkFlag(0);
		trkFlag.setBit(lcio::LCIO::RTHBIT_ID1);
		outputHitCol->setFlag(trkFlag.getFlag());

		//Get the TPC Parameters, for use later
    static const gear::TPCParameters& tpcParameters = marlin::Global::GEAR->getTPCParameters();

		//Loop over hits 
		for(IMPL::LCCollectionVec::const_iterator hitIt = inputHitCol->begin(); hitIt != inputHitCol->end(); ++hitIt)
		{

			const EVENT::TrackerHit* aHit = static_cast<const EVENT::TrackerHit*>(*hitIt);

			//If this case fails, they should all fail, so give up right here.
			if(!aHit)
			{
				streamlog_out (ERROR) << "Error: The given collection " << _inputHits << " appears not to contain EVENT::TrackerHit objects! " << std::endl;
				throw marlin::StopProcessingException(this);
			}

			const gear::TPCModule& module = tpcParameters.getNearestModule( aHit->getPosition()[0], aHit->getPosition()[1]);
			const int padID = module.getNearestPad( aHit->getPosition()[0], aHit->getPosition()[1]);
			const int moduleID = module.getModuleID();
			const int rowID = module.getRowNumber(padID);

			//Prepare a position array for the corrected hit
			double corPos[3] = {aHit->getPosition()[0], aHit->getPosition()[1], aHit->getPosition()[2]};

			std::pair<int, int> key(moduleID, rowID);


			std::map< std::pair<int,int>, gsl_spline *>::iterator splineIter;
			splineIter = _interp_rphi.find(key);

			//If there is an interpolator for this mod/row, we can correct the hit
			if(splineIter != _interp_rphi.end())
			{
				gear::Vector2D hitLocal = module.globalToLocal(aHit->getPosition()[0],aHit->getPosition()[1]);
				gear::Vector2D padCenterLocal = module.getLocalPadLayout().getPadCenter(padID);
				double x_width = module.getPadWidth(padID);
				double x_rel = 0.;

				if((module.getLocalPadLayout()).getPadLayoutType() == 1)//rectangular
				{
					if(_relative)
					{
						x_rel = (hitLocal[0] - padCenterLocal[0])/x_width;
					}
					else
					{
						x_rel = hitLocal[0];
					}

					//Evaluate the bias
					double bias = gsl_spline_eval (splineIter->second, x_rel, _interp_acc_rphi[key]);
					gear::Vector2D globalCor = module.localToGlobal(hitLocal[0] - bias, hitLocal[1]);
					corPos[0] = globalCor[0];
					corPos[1] = globalCor[1];
				}
				else if((module.getLocalPadLayout()).getPadLayoutType() == 2)//if circular pad layout
				{
					if(_relative)
					{
						x_rel = (hitLocal[1] - padCenterLocal[1])/x_width;
					}
					else
					{
						x_rel = hitLocal[1];
					}

					//Evaluate the bias
//					std::cout << "moduleID = " << moduleID << std::endl;
//					std::cout << "rowID = " << rowID << std::endl;
//					std::cout << "x_rel = " << x_rel << std::endl;
// FIXED: spline does not work at boundaries of x_relative 
					double bias(0.0);
					if (x_rel>-0.49 && x_rel<0.49)
					 bias = gsl_spline_eval (splineIter->second, x_rel, _interp_acc_rphi[key]);
//					std::cout << "bias = " << bias << std::endl;
					gear::Vector2D globalCor = module.localToGlobal(hitLocal[0], hitLocal[1] - bias/hitLocal[0]);
					corPos[0] = globalCor[0];
					corPos[1] = globalCor[1];
				}

				//Now, if there is z-data, apply that correction
				std::map< std::pair<int,int>, gsl_spline *>::iterator splineIter_z;
				splineIter_z = _interp_z.find(key);

				if(splineIter_z != _interp_z.end())
				{
					//Evaluate the bias
//					std::cout << "moduleID = " << moduleID << std::endl;
//					std::cout << "rowID = " << rowID << std::endl;
//					std::cout << "x_rel = " << x_rel << std::endl;
// FIXED: spline does not work at boundaries of x_relative 
					double bias(0.0);
					if (x_rel>-0.49 && x_rel<0.49)
					  bias = gsl_spline_eval (splineIter_z->second, x_rel, _interp_acc_z[key]);
//					std::cout << "bias = " << bias << std::endl;

					//Apply the correction
					corPos[2] -= bias;

				}

			}

			//Create and store the new hit
			TrackerHitImpl* corHit = LCObjectCopier::copy(aHit);
			corHit->setPosition	(	corPos );
			outputHitCol->addElement(corHit);

		}//hit loop

		evt->addCollection(outputHitCol, _outputHits);

  }//processEvent()
  
//FIXME: not used:  void BiasCorrectorProcessor::check(lcio::LCEvent* evt)
//  {
//  }

  void BiasCorrectorProcessor::end()
  {
  }

  void BiasCorrectorProcessor::fillData(std::ifstream& file, int biasType)
	{
		//This function assumes the input file is valid, and the first line has been used

		std::string tmp;
		getline(file, tmp);  //Dump the extra first line

		while(getline(file, tmp)) //Get the mod/row as long as the file still has data
		{
			int mod = 0, row = 0;
			std::stringstream ss(tmp);
			ss >> mod >> row;

			getline(file, tmp);  //Dump the extra newline 
			getline(file, tmp);  //Get the bias data 

			std::stringstream data(tmp);

			std::vector<double> x, y;
			while(data.good())
			{
				double x_rel = 0., bias = 0., bias_e = 0.;
				std::string dump;

				data >> x_rel >> bias >> bias_e;
				
				data >> dump; //Dump the excess comma, and set the end bit if out of data.

				x.push_back(x_rel);
				y.push_back(bias);
			}			

			gsl_interp_accel *acc = gsl_interp_accel_alloc ();
			gsl_spline *spline = NULL;
			switch(_spline)
			{
				case 1:
					spline = gsl_spline_alloc (gsl_interp_polynomial, x.size()); break;
				case 2:
					spline = gsl_spline_alloc (gsl_interp_akima, x.size()); break;
				default:
					spline = gsl_spline_alloc (gsl_interp_cspline, x.size()); break;
			}
			gsl_spline_init (spline, &(x[0]), &(y[0]), x.size());

			//rphi
			if(biasType == 0)
			{
				_interp_acc_rphi[std::pair<int, int>(mod, row)] = acc;
				_interp_rphi[std::pair<int, int>(mod, row)] = spline;
			}
			//z
			else
			{
				_interp_acc_z[std::pair<int, int>(mod, row)] = acc;
				_interp_z[std::pair<int, int>(mod, row)] = spline;
			}

			getline(file, tmp);  //Dump the extra newline 
		}
	}

}//end namespace
