#include "HitTimeRecalculationProcessor.h"

#include <iostream>
#include <string>
#include <cstdlib>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/Track.h>
#include <EVENT/TrackerHit.h>
#include <EVENT/TrackerPulse.h>
#include <EVENT/TrackerData.h>
#include <Exceptions.h>
#include <UTIL/BitField64.h>
#include <UTIL/ILDConf.h>
#include <UTIL/LCTOOLS.h>
#include "LCObjectCopier.h"

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace gear;
using namespace tpcconddata;

// MarlinTPC
#include "PulseFinder.h"
#include "ADCElectronicsListener.h"
#include "PedestalListener.h"
// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
#include "TPCConditions.h"

//MARLIN
#include <marlin/Global.h>
#include <marlin/ConditionsProcessor.h>
#include <marlin/Exceptions.h>

namespace marlintpc
{

  HitTimeRecalculationProcessor aHitTimeRecalculationProcessor;


  HitTimeRecalculationProcessor::HitTimeRecalculationProcessor() : Processor("HitTimeRecalculationProcessor"),_pulseFinder(NULL)
								   
  {
    // the processor description
    _description = "HitTimeRecalculationProcessor creates the PRF plots for each row and for all row integrated.";
    
    
    registerInputCollection(    LCIO::TRACKERHIT,
				"InputCollectionName" ,
				"Name of the input hit collection, default: TPCHits"  ,
				_inputColName ,
				string("TPCHits")) ;

    registerOutputCollection(    LCIO::TRACKERHIT,
				 "OutputCollectionName" ,
				 "Name of the output hit collection, default: TPCHitsNewTime"  ,
				 _outputColName ,
				 string("TPCHitsNewTime")) ;

    registerProcessorParameter("TimeCalculationMode",
			       "INFLEXION_POINT = 0 (default), FIT_PARABOLA = 1,  CENTRE_OF_GRAVITY = 2, FIRST_ABOVE_THRESHOLD = 3, AT_MAXIMUM = 4, CONSTANT_FRACTION=5",
			       _timeCalculationMode,
			       int(0));
   
    registerProcessorParameter("TimeCalculationParameter",
			       "fraction for constant fraction (default=0.9), no. of neighboring bins for center of gravity at maxbin, alpha/sigma for convolution",
			       _timeCalculationParameter,
			       double(0.9));

    registerProcessorParameter("ElectronicsParametersObjectName" ,
                               "Name of the conditions object in which the parameters of the read out electronics is stored",
                               _electrParaObjName,
                               string("GenericADCElectronicsParameters")
                              );

    registerProcessorParameter("ElectronicsParametersCollectionName" ,
                               "Name of the conditions collection in which the parameters of the read out electronics is stored",
                               _electrParaColName,
                               tpcconddata::GenericADCElectronicsParameters::getDefaultColName()
			       );
    
    registerProcessorParameter("PedestalCollectionName" ,
                               "Name of the conditions collection in which the pedestals are stored",
                               _pedestalColName,
                               tpcconddata::Pedestal::getDefaultColName()
			       );
    
    registerProcessorParameter("PulseStartPedestalWidthFactor",
                               string("A pulse starts if the ADC value is this factor times the ")+
			       string(" width of the pedestal (default: 4.0)"),
                               _pulseStartPedestalWidthFactor,
                               float(4.0));

    registerProcessorParameter("NoiseValuePedestalWidthFactor",
			       "The noise (cut) value is calculated as this factor times pedesta width (default: 3.0)",
			       _noiseValuePedestalWidthFactor,
                               float(3.0));

    registerOptionalParameter("IsPedestalSubtractedOverride",
                              "Optional:  Sets, if the data is already pedestal subtracted in the read-out electronics. This overrides any information provided by LCCD.",
                              _isPedestalSubtractedOverride,
                              true);

    registerOptionalParameter("PedestalValueOverride",
                              "Optional: The pedestal value to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
                              _pedestalValueOverride,
                              float(0.));
    
    registerOptionalParameter("PedestalWidthOverride",
                              "Optional: The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
                              _pedestalWidthOverride,
                              float(1.));
    
    registerOptionalParameter("ReadoutFrequencyOverride",
			      "Optional:  The readout frequency of every readout group (in MHz). This overrides any information provided by LCCD.",
			      _readoutFrequencyOverride,
			      float(20.));
    
    registerOptionalParameter("SignalPolarityOverride",
                              "The polarity of the signals (1 for positive, -1 for negative) produced by all readout groups. This overrides any information provided by LCCD.",
                              _polarityOverride,
                              1);
    
    registerOptionalParameter("InputTPCConditions",
			      "Optional: Name of the input collection containing the TPC conditions data",
			      _inputTPCConditionsCollectionName,
			      TPCConditions::getDefaultColName()) ;
    
    registerOptionalParameter("VDriftOverride" ,
			      "Optional: Set drift velocity in case there is no conditions data in mm/mu s"  ,
			      _driftVelocityOverride ,
			      float(0.)) ;
  }


  void HitTimeRecalculationProcessor::init()
  { 
    streamlog_out(DEBUG) << " init() called  " << endl;
    printParameters();
    
    // get drift velocity, either measured value from calibration
    // or nominal value given as process parameter
    if(parameterSet("VDriftOverride"))
      {
	streamlog_out(WARNING) << "the drift velocity is taken from the steering file. please change" << endl;
      }
    else // CAUTION! this is not only true for this value, but currently it's the only one in use
      {
	try
	  {
	    _tpcConditionsListener = new TPCConditionsListener(_inputTPCConditionsCollectionName);
	  }
	
	catch(lccd::LCCDException &)
	  {
	    streamlog_out(ERROR) << "Collection named \"" << _inputTPCConditionsCollectionName << "\" with TPCConditions is not available." << std::endl;
	    streamlog_out(ERROR) << "Use the ConditionsProcessor to provide it, or set the VDriftOverride parameter if this is not possible." << std::endl;
	    throw;
	  }
      }
    
    // The override parameter for the readout frequency has changed from Hz to MHz. Stop processing if an old (unreasonably high) frequency is detected.
    // 1e5 should be fine. A readout with more than 100 GHz can be excluded for the foreseeable future. A readout < 100 kHz (10 ��s samples) is not reasonable
    // for TPCs anyway, so we should be save to assume MHz for lower values.
    if (_readoutFrequencyOverride > 1e5) 
      {
	streamlog_out(ERROR) << "Unreasonably high readout frequency detected ("<< _readoutFrequencyOverride <<" MHz)."<<std::endl
			     << "You are probably using the old value in Hz. Plesase adapt this value to MHz." << std::endl;
	throw StopProcessingException(this);
      }
    
    PedestalListener * pedestalListener = NULL;
    
    // determine if the pedestal information is supplied by LCCD
    //If both the pedestal width and value are overridden no listener is necessary
    if ( !(parameterSet("PedestalWidthOverride")  && parameterSet("PedestalValueOverride")) )
      {
        pedestalListener = new PedestalListener(_pedestalColName);
      }
    
    ADCElectronicsListener * adcElectronicsListener = NULL;
    
    // determine if the polarity, maxADC value and readout frequency are supplied by LCCD
    // no listener needed if all three values overridden
    if (!(parameterSet("ReadoutFrequencyOverride") && 
	  parameterSet("SignalPolarityOverride") && 
	  parameterSet("IsPedestalSubtractedOverride")))
      {
	adcElectronicsListener = new ADCElectronicsListener(_electrParaObjName, _electrParaColName);
      }
    
    //create an instance of the PulseFinder. Some values are dummies as we need only a part of the functionality.
    //We set only the values relevant for the time estimator methods
    _pulseFinder = new PulseFinder(0,// _minimumPulseHeight,
				   0,//_minimumPulseLength,
				   _pulseStartPedestalWidthFactor,
				   0,//_pulseEndPedestalWidthFactor,
				   _noiseValuePedestalWidthFactor,
				   pedestalListener,
				   adcElectronicsListener,
				   0,//_saveNBinsBeforeStart,
				   0,//_saveNBinsAfterEnd,
				   _timeCalculationMode,
				   _timeCalculationParameter,
				   0);//_chargeCalculationMode);
	
    // now set the override parameters, if any
    if (parameterSet("PedestalValueOverride"))
      {
	_pulseFinder->setPedestalValueOverride(_pedestalValueOverride);
      }
    
    if (parameterSet("PedestalWidthOverride"))
      {
	_pulseFinder->setPedestalWidthOverride(_pedestalWidthOverride);
      }
    
    if (parameterSet("ReadoutFrequencyOverride"))
      {
	_pulseFinder->setReadoutFrequencyOverride(_readoutFrequencyOverride);
      }
    if (parameterSet("IsPedestalSubtractedOverride"))
      {
      _pulseFinder->set_isPedestalSubtractedOverride(_isPedestalSubtractedOverride);
      }

    if (parameterSet("SignalPolarityOverride"))
      {
	_pulseFinder->setPolarityOverride(_polarityOverride);
      }
    
    if (parameterSet("ReadoutFrequencyOverride"))
      {
	_pulseFinder->setReadoutFrequencyOverride(_readoutFrequencyOverride);
      }
    if (parameterSet("MaximumADCValueOverride"))
      {
	_pulseFinder->setMaxADCValueOverride(1024);
      }

    
  }
  
  
  void HitTimeRecalculationProcessor::processRunHeader(LCRunHeader*  /*run*/)
  {
    // do something for each run
  } 
  

  void HitTimeRecalculationProcessor::processEvent(LCEvent* evt)
  { 
    // use an LCCollectionVec instead of LCCollection so we can use iterators later
    LCCollectionVec* inputColVec = 0;
  
    // do a try / catch in case there is an empty event
    try
      {
	inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
      }
    catch (DataNotAvailableException &e)
      {
	// no input data, just return and process the next event
	return;
      }
   // create output hit collection and set the persistency flag
    LCCollectionVec* tpcHitCollection = new LCCollectionVec(LCIO::TRACKERHIT);
    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0);
    trkFlag.setBit(lcio::LCIO::RTHBIT_ID1);
    tpcHitCollection->setFlag(trkFlag.getFlag());
    
    // loop the data collection and do something
    for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
	 inputObjectIter < inputColVec->end() ;  inputObjectIter++ )
      {
	// dynamic cast the input object (LCObject) to what it really is e.g.
	lcio::TrackerHit * hit= dynamic_cast<lcio::TrackerHit *>( *inputObjectIter );
	TrackerPulse const * maxpulse=NULL;
	double maxcharge=0;
	//find pulse with maximum charge
	for ( LCObjectVec::const_iterator pulseIter = hit->getRawHits().begin() ;
	      pulseIter < hit->getRawHits().end() ; pulseIter++ )
	  {
	    TrackerPulse const * pulse = dynamic_cast<TrackerPulse const *>(*pulseIter);
	    if(pulse->getCharge()>maxcharge)
	      {
		maxcharge=pulse->getCharge();
		maxpulse=pulse;
	      }
	  }
	double newtime=_pulseFinder->calculatePulseTimeAndCharge(maxpulse->getTrackerData()).getTime();
	TrackerHitImpl* newhit = LCObjectCopier::copy(dynamic_cast<TrackerHit*>(*inputObjectIter)) ;
	newhit->setTime(newtime);
	double zPosition = Global::GEAR->getTPCParameters().getMaxDriftLength()
	  -  _getDriftVelocity() * newtime / 1000. ; // time is in [ns], drift velocity in [mm/mus]
	const double position[3] = { hit->getPosition()[0],hit->getPosition()[1],zPosition};
	newhit->setPosition(position);
	tpcHitCollection->addElement(newhit);
      }
      evt->addCollection(tpcHitCollection, _outputColName);
  }

  double HitTimeRecalculationProcessor::_getDriftVelocity()
  {
    if(parameterSet("VDriftOverride"))
      {
	return _driftVelocityOverride;
      }
    else
      {
	try
	  {
	    return _tpcConditionsListener->getDriftVelocity();
	  }
	catch(...)
	  {
	    streamlog_out(ERROR) << "[getDriftVelocity failed] The Listener for TPCConditions could not be used. Maybe the conditions object is not available?" << endl;
	    streamlog_out(ERROR) <<  "                          Possibly you need to use the override parameterin the steering file called (\"VDriftOverride\")." << endl;
	    abort();
	  }
      }
  }
  
  void HitTimeRecalculationProcessor::check(LCEvent*  /*evt*/)
  { 
    // check what is needed here
  }


  void HitTimeRecalculationProcessor::end()
  { 

  }
} // close namespace brackets
