#include "MMHitTimeCorrectionProcessor.h"
#include <iostream>
#include <cmath>
#include <vector>

#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <EVENT/TrackerPulse.h>
#include <EVENT/TrackerHit.h>

#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

#include <marlin/Exceptions.h>
#include <marlin/Global.h>

#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>

// ----- include for verbosity dependent logging ---------
#include "marlin/VerbosityLevels.h"

// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"

// ROOT
#include <TROOT.h>
#include <TF1.h>
#include <TGraph.h>
#include <MMPulseShapeFunction.h>
#include "TGraphErrors.h"
#include "TMath.h"
#include "TFitResultPtr.h"

// OTHER
#include "LCObjectCopier.h"

using namespace std;
using namespace lcio;
using namespace marlin;

namespace marlintpc
{

MMHitTimeCorrectionProcessor  aMMHitTimeCorrectionProcessor;

MMHitTimeCorrectionProcessor::MMHitTimeCorrectionProcessor() : Processor("MMHitTimeCorrectionProcessor")
{

    _description = "MMHitTimeCorrectionProcessor corrects the time estimate and z-coordinate for Micromegas hits.";

    registerInputCollection(LCIO::TRACKERHIT,
                            "InputTrackerHits",
                            "Name of the input hit collection.",
                            _inputHitCollectionName,
                            std::string("TPCHits"));

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OutputTrackerHits",
                             "Name of the output TrackerHits collection",
                             _outputHitCollectionName,
                             std::string("TPCHitsTimeCorrected"));

    registerProcessorParameter("WriteOutputToStorage",
                               "If true, the output hits collection is written to storage (default: true)",
                               _outputHitsPersistent,
                               true);

    registerProcessorParameter("ElectronicsParametersCollectionName",
                               "Name of the conditions collection in which the parameters of the read out electronics is stored",
                               _electrParaColName,
                               tpcconddata::GenericADCElectronicsParameters::getDefaultColName());

    registerProcessorParameter("ElectronicsParametersObjectName",
                               "Name of the conditions object in which the parameters of the read out electronics is stored",
                               _electrParaObjName,
                               string("GenericADCElectronicsParameters"));

    registerProcessorParameter("TPCConditionsCollectionName",
                               "Name of the conditions collection in which the TPC conditions are stored",
                               _tpcCondColName,
                               tpcconddata::TPCConditions::getDefaultColName());

    registerProcessorParameter("UseOverflowPulses",
                               "Set this parameter to true to also use overrange pulses for the hit time determination, default: false",
                               _useOverflowPulses,
                               false);

    registerProcessorParameter("DiscardHitsWithOverflow",
                               "Set this parameter to true to discard hits completely, if they have an overflow pulse (will be ignored, if UseOverflowPulses is also true), default: false",
                               _discardHitsWithOverflow,
                               false);

    registerProcessorParameter("TimeCalculationMethod",
                               "Method to calculate the time. Inflexion Point = 0, Max Bin = 1, Box Method = 2, Half-Max Bin = 3, Fixed-Value = 4, Weighted-Mean = 5, Gaussian Inflexion Point = 6, Gaussian Mean = 7, Gamma5 = 8. (DEFAULT: 0)",
                               _timeCalculationMethod,
                               int(0));

    registerOptionalParameter("VDriftOverride",
                              "This value overrides any drift velocity taken from conditions database if set",
                              _driftVelocityOverride,
                              float(75.6));

    registerOptionalParameter("ReadoutFrequencyOverride",
                              "This value overrides any readout frequency from conditions database if set; please give frequency in Hz",
                              _readoutFrequencyOverride,
                              float(2.5e9));

    registerOptionalParameter("TimeOffsetOverride",
                              "The t0 offset as determined from cosmics data (ns please!); overrides value taken from conditions database",
                              _timeOffsetOverride,
                              float(0.));

    registerProcessorParameter("PedestalCollectionName",
                               "Name of the conditions collection in which the pedestals are stored",
                               _pedestalColName,
                               tpcconddata::Pedestal::getDefaultColName()
                              );

    registerOptionalParameter("PedestalWidthOverride",
                              "Optional: The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
                              _pedestalWidthOverride,
                              float(1.));

    registerOptionalParameter("NNeighbours",
                              "Number of neighbour pads to be used in the time calculation on each side of the max (DEFAULT: 0)",
                              _nNeighbours,
                              int(0));

}



void MMHitTimeCorrectionProcessor::init()
{

    streamlog_out(DEBUG) << " init called  " << std::endl;

    // usually a good idea to
    printParameters();

    _nRun = 0;
    _nEvt = 0;

    // Check the validity of the Time Calculation input
    switch(_timeCalculationMethod)
    {
        case 0:
            break;
        case 1:
            break;
        case 2:
            break;
        case 3:
            break;
        case 4:
            break;
        case 5:
            break;
        case 6:
            break;
        case 7:
            break;
        case 8:
            break;
        case 9:
            break;
        default:
            streamlog_out(ERROR) << "Invalid choice of Time Calculation method. Must be one of:" << std::endl;
            streamlog_out(ERROR) << "0 -- Inflexion Point" << std::endl;
            streamlog_out(ERROR) << "1 -- Maximum Bin" << std::endl;
            streamlog_out(ERROR) << "2 -- Box Method" << std::endl;
            streamlog_out(ERROR) << "3 -- Half-Maximum Bin" << std::endl;
            streamlog_out(ERROR) << "4 -- Fixed Value" << std::endl;
            streamlog_out(ERROR) << "5 -- Weighted Mean" << std::endl;
            streamlog_out(ERROR) << "6 -- Gaussian Inflexion Point" << std::endl;
            streamlog_out(ERROR) << "7 -- Gaussian Mean" << std::endl;
            streamlog_out(ERROR) << "8 -- Gamma5 Fit" << std::endl;
            throw StopProcessingException(this);
    }

    //warn user if _useOverflowPulses==true and _discardHitsWithOverflow==true, since their functionalities exclude each other
    if(_useOverflowPulses && _discardHitsWithOverflow)
    {
        streamlog_out(WARNING) << "UseOverflowPulses and DiscardHitsWithOverflow both set true! Hits will not be discarded, when overlow pulses are used for time calculation!" << std::endl;
    }

    //determine if parameters need to be supplied by conditions database or if all values are overwritten
    if(!parameterSet("ReadoutFrequencyOverride"))
    {
        _adcElectronicsListener = new ADCElectronicsListener(_electrParaObjName, _electrParaColName);
    }
    else
    {
        streamlog_out(MESSAGE) << "ReadoutFrequencyOverride set!" << std::endl;
    }

    if(!(parameterSet("TimeOffsetOverride") && parameterSet("VDriftOverride")))
    {
        _tpcConditionsListener = new TPCConditionsListener(_tpcCondColName);
    }
    if(parameterSet("TimeOffsetOverride"))
    {
        streamlog_out(MESSAGE) << "TimeOffsetOverride set! Time offset set to: " << _timeOffsetOverride << std::endl;
        _timeOffset = _timeOffsetOverride;
    }
    if(parameterSet("VDriftOverride"))
    {
        streamlog_out(MESSAGE) << "DriftVelocityOverride set! Drift velocity set to: " << _driftVelocityOverride << std::endl;
        _driftVelocity = _driftVelocityOverride;
    }
    if(!(parameterSet("PedestalWidthOverride")))
    {
        _pedestalListener = new PedestalListener(_pedestalColName);
    }

}//end of init



void MMHitTimeCorrectionProcessor::processRunHeader(LCRunHeader*  /*run*/)
{

    ++_nRun;
}



void MMHitTimeCorrectionProcessor::processEvent(LCEvent* evt)
{

    //-- note: this will not be printed if compiled w/o MARLINDEBUG=1 !
    streamlog_out(DEBUG) << " processing event: " << evt->getEventNumber()
                         << " in run:  " << evt->getRunNumber() << std::endl;

    //Get the input collection, if it exists
    IMPL::LCCollectionVec* inputCol = NULL;
    try
    {
        inputCol = static_cast<IMPL::LCCollectionVec*>(evt->getCollection(_inputHitCollectionName));
    }
    catch(lcio::DataNotAvailableException e)
    {
        streamlog_out(WARNING) << _inputHitCollectionName << " collection not available" << std::endl;
        return;
    }

    //Verify there are hits in the collection
    if(inputCol->getNumberOfElements() == 0)
    {
        streamlog_out(MESSAGE) << "No hits in event " << evt->getEventNumber() << ". Nothing to do here." << std::endl;
        return;
    }

    //Here we verify that the user has properly stored the pulse spectrum.
    EVENT::TrackerHit* testHit = static_cast<EVENT::TrackerHit* >(inputCol->getElementAt(0));
    if(!testHit)
    {
        streamlog_out(ERROR) << "Error: The input collection does not contain TrackerHits (there was an error in the cast)." << std::endl;
        throw StopProcessingException(this);
    }

    if(testHit->getRawHits().size() == 0 || !testHit->getRawHits()[0])
    {
        streamlog_out(ERROR) << "Error: The input collection does not contain the TrackerPulses associated with the hits." << std::endl;
        throw StopProcessingException(this);
    }

    if(!static_cast<EVENT::TrackerPulse*>(testHit->getRawHits()[0])->getTrackerData())
    {
        streamlog_out(ERROR) << "Error: The input collection does not contain the TrackerData associated with the hits' pulses." << std::endl;
        throw StopProcessingException(this);
    }

    IMPL::LCCollectionVec* outputCol = new IMPL::LCCollectionVec(LCIO::TRACKERHIT);
    outputCol->setTransient(!_outputHitsPersistent);

    //get the parameters from the conditions database
    if(!parameterSet("ReadoutFrequencyOverride"))
    {
        _readoutFrequency = _adcElectronicsListener->getFrequency(0) * 1.E6;//Listener gives MHz -> convert to Hz since this is used in the rest of the processor
        streamlog_out(DEBUG) << "Readout frequency set to value from conditions database: " << _readoutFrequency << std::endl;
    }
    else
    {
        _readoutFrequency = _readoutFrequencyOverride;
        streamlog_out(DEBUG) << "Readout frequency set from override to: " << _readoutFrequencyOverride << std::endl;

    }

    if(!parameterSet("VDriftOverride"))
    {
        _driftVelocity = _tpcConditionsListener->getDriftVelocity();
        streamlog_out(DEBUG) << "Drift velocity set to value from conditions database: " << _driftVelocity << std::endl;
    }
    if(!parameterSet("TimeOffsetOverride"))
    {
        _timeOffset = _tpcConditionsListener->getTimeOffset() * 1.E3;//Listener returns microseconds -> convert to ns
        streamlog_out(DEBUG) << "Time offset set to value from conditions database: " << _timeOffset << std::endl;
    }

    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0);
    trkFlag.setBit(lcio::LCIO::RTHBIT_ID1);
    outputCol->setFlag(trkFlag.getFlag());

    //Calculate the correct times.
    _correctTimes(inputCol, outputCol);

    evt->addCollection(outputCol, _outputHitCollectionName);

    ++_nEvt;
}//end of processEvent()



void MMHitTimeCorrectionProcessor::check(LCEvent*  /*evt*/)
{
}



void MMHitTimeCorrectionProcessor::end()
{
}



void MMHitTimeCorrectionProcessor::_correctTimes(IMPL::LCCollectionVec* input, IMPL::LCCollectionVec* output)
{

    //Iterate through the hits in the collection
    for(IMPL::LCCollectionVec::iterator hitIt = input->begin(); hitIt != input->end(); ++hitIt)
    {
        const EVENT::LCObjectVec& pulses = (static_cast<IMPL::TrackerHitImpl*>(*hitIt))->getRawHits();

        const EVENT::TrackerPulse* maxPulse = NULL;

        float maxVal = -9E9;

        for(LCObjectVec::const_iterator pulseIt = pulses.begin(); pulseIt != pulses.end(); ++pulseIt)
        {

            const EVENT::TrackerPulse* tmp = static_cast<const EVENT::TrackerPulse*>(*pulseIt);

            //find the highest pulse, that is not overflow, unless overflow pulses should be used
            if((_useOverflowPulses || !pulseflag::isOverflow(tmp->getQuality())) && tmp->getCharge() > maxVal)
            {
                maxPulse = tmp;
                maxVal = tmp->getCharge();
            }
            //discard the complete hit, if a single pulse is overflow and this behaviour is wanted
            else if(!_useOverflowPulses && _discardHitsWithOverflow && pulseflag::isOverflow(tmp->getQuality()))
            {
                maxPulse = NULL;
                break;
            }

        }//End loop over pulses

        //If, somehow, all the pulses are overflow, then I don't know what to do...
        if(maxPulse == NULL) continue;

        std::pair<double, double> fixedTime;

        //Calculate the fixed time

        if(!parameterSet("NNeighbours") || _nNeighbours == 0)
        {
            fixedTime = _calculateCorrectedTime(maxPulse);
        }
        else
        {
            /* in case neighbour pulses should also be included,
             * loop over pulses again and calculate the times for
             * those which are close enough to the max pulse */

            // keep track of charge to calculate charge-weighted time
            double sumCharge = 0;
            double sumTimeCharge = 0;
            double sum2TimeErrCharge = 0; // FIXME!! - proper error with correlations?
            double tmpTime, tmpTimeErr, tmpCharge;
            for(LCObjectVec::const_iterator pulseIt = pulses.begin(); pulseIt != pulses.end(); ++pulseIt)
            {

                const EVENT::TrackerPulse* tmpPulse = static_cast<const EVENT::TrackerPulse*>(*pulseIt);
                int distanceToMax = abs(tmpPulse->getCellID0() - maxPulse->getCellID0());

                // use pulses in the desired pad range if they are not overflow, unless overflow pulses should be used
                if(distanceToMax <= _nNeighbours && (_useOverflowPulses || !pulseflag::isOverflow(tmpPulse->getQuality())))
                {
                    fixedTime = _calculateCorrectedTime(tmpPulse);
                    tmpTime = fixedTime.first;
                    tmpTimeErr = fixedTime.second;
                    tmpCharge = tmpPulse->getCharge();
                    sumCharge += tmpCharge;
                    sumTimeCharge += tmpCharge * tmpTime;
                    sum2TimeErrCharge += (tmpCharge * tmpTimeErr) * (tmpCharge * tmpTimeErr);
                    if(0 == distanceToMax)
                    {
                        streamlog_out(DEBUG4) << "max pulse time: " << tmpTime << " +- " << tmpTimeErr << std::endl;
                    }
                }
            }//End loop over pulses
            // calculate charge-weighted fixed time
            fixedTime.first = sumTimeCharge / sumCharge;
            fixedTime.second = sqrt(sum2TimeErrCharge) / sumCharge;

            streamlog_out(DEBUG4) << "charge-weighted time: " << fixedTime.first << " +- " << fixedTime.second << std::endl;
        }


        if(fixedTime.first == -9E9) continue;

        //Apply the time offset correction
        fixedTime.first -= _timeOffset;

        IMPL::TrackerHitImpl* curHit = static_cast<IMPL::TrackerHitImpl*>(*hitIt);
        IMPL::TrackerHitImpl* copyHit = LCObjectCopier::copy(curHit);

        double fixedZ = Global::GEAR->getTPCParameters().getMaxDriftLength() -
                        _driftVelocity * fixedTime.first / 1000.; // time is in [ns], drift velocity in [mm/mus]
        double fixedPos[3] = {curHit->getPosition()[0], curHit->getPosition()[1], fixedZ};
        EVENT::FloatVec fixedCov = EVENT::FloatVec(curHit->getCovMatrix());
        double fixedZErr = _driftVelocity * fixedTime.second / 1000.;
        fixedCov[5] = fixedZErr * fixedZErr;

        streamlog_out(DEBUG3) << " maxdriftlength : " << Global::GEAR->getTPCParameters().getMaxDriftLength()
                              << " , driftVelocity " << _driftVelocity << " , fixedTime.first " << fixedTime.first  << std::endl;
        streamlog_out(DEBUG3) << " fixedPosition : " << fixedPos[0]
                              << " , " << fixedPos[1] << " , " << fixedPos[2]  << std::endl;

        copyHit->setTime(fixedTime.first);
        copyHit->setPosition(fixedPos);
        copyHit->setCovMatrix(fixedCov);

        output->addElement(copyHit);

    }//End loop over hits

}//End _correctTimes()



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateCorrectedTime(const EVENT::TrackerPulse* thePulse)
{
    /*

      This is the general method to correct the time.
      It calls the chosen individual method to carry out the desired method of time calculation.
      It was implemented to avoid repeating the same long switch statement at two (or more?) points in the code.

    */

    std::pair<double, double> fixedTimeAndError;

    switch(_timeCalculationMethod)
    {
        case 0:
            fixedTimeAndError = _calculateInflexionPoint(thePulse);
            break;
        case 1:
            fixedTimeAndError = _calculateMaxPoint(thePulse);
            break;
        case 2:
            fixedTimeAndError = _calculateBoxMethod(thePulse);
            break;
        case 3:
            fixedTimeAndError = _calculateHalfMaxPoint(thePulse);
            break;
        case 4:
            fixedTimeAndError.first = 4000.;
            fixedTimeAndError.second = 0.;
            break;
        case 5:
            fixedTimeAndError = _calculateWeightedMean(thePulse);
            break;
        case 6:
            fixedTimeAndError = _calculateGaussianInflexion(thePulse);
            break;
        case 7:
            fixedTimeAndError = _calculateGaussianMean(thePulse);
            break;
        case 8:
            fixedTimeAndError = _calcPulseFitWithFunction(thePulse);
            break;
        case 9:
            fixedTimeAndError = _calculateWeightedMeanAtMax(thePulse);
            break;
    }

    return fixedTimeAndError;
}



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateInflexionPoint(const EVENT::TrackerPulse* thePulse)
{
    /*

      This is taken almost entirely from PulseFinderProcessor for GEMs.
      It uses a process of comparing derivatives to find the inflection point of the rising
      edge of the pulse. This, of course, assumes a nice, single pulse shape. Potentially still
      applicable later, but must be investigated further. It is used here since we are only
      concerned with approximate pulse times at the moment.

    */

    //Return val
    std::pair<double, double> timeAndError;

    std::vector<float> chargeVec(thePulse->getTrackerData()->getChargeValues());
    int readout_group = thePulse->getTrackerData()->getCellID1();
    int hardware_channel = thePulse->getTrackerData()->getCellID0();

    //RECENT//////////////
    float noise_value = 30.;
    if(parameterSet("PedestalWidthOverride"))
    {
        streamlog_out(DEBUG) << "Pedestal Value for channel: " << readout_group << ":" << hardware_channel
                             << " overridden to " << _pedestalWidthOverride << std::endl;
        noise_value = _pedestalWidthOverride;
    }
    else
    {
        if(_pedestalListener != NULL)
        {
            streamlog_out(DEBUG) << "Pedestal Width for channel: " << readout_group << ":" << hardware_channel
                                 << " = " << _pedestalListener->getPedestalWidth(hardware_channel, readout_group) << std::endl;
            noise_value = _pedestalListener->getPedestalWidth(hardware_channel, readout_group);
        }
        else
        {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinder]::getPedestalWidth-Exception: The Pedestal width is not defined or can not be found!"));
            noise_value = 30.;
        }
    }

    float time      = 0.0;
    float timeError = 0.0;
    float derivative     = 0.0;
    float norm           = 0.0;
    float sliceTime      = thePulse->getTrackerData()->getTime();

    // determine the mean value of the positive derivative == the best estimate for the pulse time
    if(chargeVec.size() > 1)
    {
        for(std::vector<float>::const_iterator timeBin = chargeVec.begin() + 1; timeBin < chargeVec.end(); ++timeBin)
        {
            derivative  = (*timeBin) - (*(timeBin - 1));

            sliceTime += 1.;

            // break out of calculation if decrease is larger than the noise
            if(derivative < (-noise_value))
                break;

            norm      += derivative;
            time += derivative * sliceTime;
        }
        if(norm > 0.)
            time /= norm;
        else // no iteration means a single slot that counts!
            time = sliceTime;
    }
    else
        time = sliceTime;

    // now determine the pulse time error -- which needs a second loop
    const float timeSum = (time - thePulse->getTrackerData()->getTime()) * norm;

    if(chargeVec.size() > 1)
    {
        for(std::vector<float>::const_iterator timeBin = chargeVec.begin() + 1; timeBin < chargeVec.end(); ++timeBin)
        {
            derivative  = (*timeBin) - (*(timeBin - 1));
            unsigned int index = timeBin - chargeVec.begin();

            // break out of calculation if decrease is larger than the noise
            if(derivative < (-noise_value))
                break;

            double derivativeVariance = 2 * noise_value * noise_value; //sigma q = noise value
            timeError += (index * norm - timeSum) * (index * norm - timeSum) * derivativeVariance;
        }
        timeError = sqrt(timeError) / (norm * norm);
    }
    else
        timeError = -1.;

    const float bin_width = 1. / _readoutFrequency * 1.E9;
    time = time * bin_width; //conversion in ns

    //Set the return values
    timeAndError.first = time; //Set the pulse time!
    timeAndError.second = timeError;

    return timeAndError;

}//End _calculateInflexionPoint()



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateMaxPoint(const EVENT::TrackerPulse* thePulse)
{
    /*

      Simply searches for the time value of the bin with the maximum charge.

    */

    //Return value
    std::pair<double, double> timeAndError;

    //Get the spectrum
    std::vector<float> chargeVec(thePulse->getTrackerData()->getChargeValues());

    int maxBin = 0;
    float maxVal = -9E9;

    //Find maximum amplitude bin
    for(unsigned int i = 0; i < chargeVec.size(); ++i)
    {
        if(chargeVec[i] > maxVal)
        {
            maxVal = chargeVec[i];
            maxBin = i;
        }
    }

    const float binWidth = 1. / _readoutFrequency * 1.E9;
    double pulseTime = 0.;
    float sliceTime = thePulse->getTrackerData()->getTime();
    //float sliceTime = 0.;
    pulseTime = (sliceTime + maxBin) * binWidth;

    timeAndError.first = pulseTime;
    timeAndError.second = binWidth / 2.;

    return timeAndError;

}//End _calculateMaxPoint();



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateHalfMaxPoint(const EVENT::TrackerPulse* thePulse)
{
    /*

      Simply searches for the time value of the bin with the maximum charge.

    */

    //Return value
    std::pair<double, double> timeAndError;

    //Get the spectrum
    std::vector<float> chargeVec(thePulse->getTrackerData()->getChargeValues());

    int maxBin = 0;
    float maxVal = -9E9;

    //Find maximum amplitude bin
    for(unsigned int i = 0; i < chargeVec.size(); ++i)
    {
        if(chargeVec[i] > maxVal)
        {
            maxVal = chargeVec[i];
            maxBin = i;
        }
    }

    float targetVal = maxVal / 2.;
    float diff = 9e9;
    int bestBin = 0;

    //Find half-maximum amplitude bin
    for(int i = 0; i < maxBin; ++i)
    {
        if(fabs(chargeVec[i] - targetVal) < diff)
        {
            diff = fabs(chargeVec[i] - targetVal);
            bestBin = i;
        }
    }

    const float binWidth = 1. / _readoutFrequency * 1.E9;
    double pulseTime = 0.;
    float sliceTime = thePulse->getTrackerData()->getTime();
    //float sliceTime = 0.;
    pulseTime = (sliceTime + bestBin) * binWidth;

    timeAndError.first = pulseTime;
    timeAndError.second = binWidth / 2.;

    return timeAndError;

}//End _calculateHalfMaxPoint();



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateWeightedMean(const EVENT::TrackerPulse* thePulse)
{
    /*

      Simply searches for the time value of the bin with the maximum charge.

    */

    //Return value
    std::pair<double, double> timeAndError;

    //Get the spectrum
    std::vector<float> chargeVec(thePulse->getTrackerData()->getChargeValues());
    double sliceTime = thePulse->getTrackerData()->getTime();  //Note: We're converting from int here

    double sumT = 0.;
    double sumT2 = 0.;
    double sumAmp = 0.;

    //Find maximum amplitude bin
    for(unsigned int i = 0; i < chargeVec.size(); ++i)
    {
        if(chargeVec[i] <= 0) continue;
        sumT += (sliceTime + (double)i) * chargeVec[i];
        sumT2 += (sliceTime + (double)i) * (sliceTime + (double)i) * chargeVec[i];
        sumAmp += chargeVec[i];
    }

    double meanT = sumT / sumAmp;
    double meanT2 = sumT2 / sumAmp;

    const float binWidth = 1. / _readoutFrequency * 1.E9;
    double pulseTime = meanT * binWidth;
    double pulseTimeErr2 = (meanT2 - meanT * meanT);

    timeAndError.first = pulseTime;
    timeAndError.second = sqrt(pulseTimeErr2 / sumAmp) * binWidth;

    return timeAndError;

}//End _calculateWeightedMean();



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateWeightedMeanAtMax(const EVENT::TrackerPulse* thePulse)
{
    const FloatVec charge_values = thePulse->getTrackerData()->getChargeValues();

    //find position of max or min, depending on the polarity in charge values

    vector<float>::const_iterator maxbin = std::max_element(charge_values.begin(), charge_values.end());

    int number = 3;
    TrackerDataImpl truncData;
    truncData.setTime(thePulse->getTrackerData()->getTime());
    truncData.setCellID0(thePulse->getTrackerData()->getCellID0());
    truncData.setCellID1(thePulse->getTrackerData()->getCellID1());
    FloatVec truncspec;
    float timeOffset = 0.;
    for(vector<float>::const_iterator itj = charge_values.begin(); itj < charge_values.end(); ++itj)
    {
        if(abs(distance(itj, maxbin)) <= number)
        {
            truncspec.push_back(*itj);
            if(truncspec.size() == 1)
            {
                timeOffset = (float) distance(charge_values.begin(), itj);
            }

        }
    }
    const FloatVec& setspec = truncspec;
    truncData.setChargeValues(setspec);
    //const TrackerDataImpl passpulse= truncPulse;
    TrackerPulseImpl truncPulse;
    truncPulse.setTime(thePulse->getTime());
    truncPulse.setCellID0(thePulse->getCellID0());
    truncPulse.setCellID1(thePulse->getCellID1());
    truncPulse.setTrackerData(&truncData);
    std::pair<double, double> timeAndError = _calculateWeightedMean(&truncPulse);

    float time = timeOffset;

    const float bin_width = 1. / _readoutFrequency * 1.E9;
    time = time * bin_width; //conversion in ns
    //          std::cout<<"1 "<<time<<std::endl;
    //          std::cout<<"2 "<<timeAndError.first<<std::endl;
    timeAndError.first += time;
    //std::cout<<"3 "<<timeAndError.first<<std::endl;
    return timeAndError;

}



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateBoxMethod(const EVENT::TrackerPulse* thePulse)
{
    /*

      Computes two areas, the bin that make the areas closest to equal is the time.
      areaDown is under the curve before the cut bin.
      areaUp is above the curve and below the max value after the cut bin.

    */

    //streamlog_out(MESSAGE) << "Using box method\n";
    //Return value
    std::pair<double, double> timeAndError;

    //Get the spectrum
    std::vector<float> chargeVec(thePulse->getTrackerData()->getChargeValues());

    int maxBin = 0;
    float maxVal = -9E9;

    //Find maximum amplitude bin
    for(unsigned int i = 0; i < chargeVec.size(); ++i)
    {
        if(chargeVec[i] > maxVal)
        {
            maxVal = chargeVec[i];
            maxBin = i;
        }
    }

    //Check if saturated, if so take center of saturation region as peak

    int numberOfPeakBins = 1; //If not saturated should stay at 1
    float saturationDerivative = 2.0; //If there are bins directly after the max bin with value within this region around the maximum then they are included
    if(pulseflag::isOverflow(thePulse->getQuality()))
    {
        for(unsigned int i = maxBin + 1; i < chargeVec.size(); ++i)
        {
            if(fabs(chargeVec[maxBin] - chargeVec[i]) < saturationDerivative)
            {
                ++numberOfPeakBins;
                maxVal += chargeVec[i];
            }
            else
                break;
        }
        maxVal /= numberOfPeakBins;
        maxBin += (numberOfPeakBins - 1) / 2; //i.e. if we find 1 extra peak bin then it will add 1/2 to the peak bin value
    }

    streamlog_out(DEBUG) << "Max Amplitude Bin: " << maxBin << "\n";

    int minBin = 0;

    while(minBin && (chargeVec[minBin - 1] > 30))
        --minBin;

    const float binWidth = 1. / _readoutFrequency * 1.E9;
    double pulseTime = 0.;
    float sliceTime = thePulse->getTrackerData()->getTime();

    double background = chargeVec[minBin ? minBin - 1 : 0]; //Assume for now

    double area = 0;
    double area2 = 0;
    double areaErrSum = 0;

    for(int i = 0; i <= maxBin; ++i)
    {
        area += (chargeVec[i] - background);
        area2 += (chargeVec[i] - background) * (chargeVec[i] - background);
        areaErrSum += chargeVec[i];
    }

    //float areaErr = sqrt(areaErrSum);

    double deltaT = area / (maxVal - background);

    //float deltaTErr = deltaT * sqrt(pow(areaErr/area, 2) + (maxVal + background)/pow(maxVal - background, 2));

    streamlog_out(DEBUG) << "Area " << area << " Max pulse Charge " << maxVal << " Delta Time " << deltaT << "\n";

    pulseTime = (sliceTime + maxBin - deltaT) * binWidth;

    //float pulseTimeErr = sqrt(maxBin + deltaT) * binWidth; //Assuming no significant error on binWidth or sliceTime
    double pulseTimeErr = sqrt(areaErrSum + area2 / maxVal) / maxVal * binWidth; //Assuming no significant error on binWidth or sliceTime

    streamlog_out(DEBUG) << "sliceTime " << sliceTime << " binWidth " << binWidth << " pulsetime " << pulseTime << "\n";

    timeAndError.first = pulseTime;
    timeAndError.second = pulseTimeErr;

    return timeAndError;

}// _calculateBoxMethod



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateGaussianInflexion(const EVENT::TrackerPulse* thePulse)
{

    //Return value
    std::pair<double, double> timeAndError;

    //Get the spectrum
    std::vector<float> chargeVec(thePulse->getTrackerData()->getChargeValues());

    int maxBin = 0;
    float maxVal = -9E9;

    //Find maximum amplitude bin
    for(unsigned int i = 0; i < chargeVec.size(); ++i)
    {
        if(chargeVec[i] > maxVal)
        {
            maxVal = chargeVec[i];
            maxBin = i;
        }

        if(0 < chargeVec[i])
        {
            streamlog_out(DEBUG2) << " found charge : " << chargeVec[i]
                                  << " in bin  " << i  << std::endl;
        }

    }

    //Define a range where it's mostly gaussian
    int lowerEdge = maxBin - 5;
    int upperEdge = maxBin + 5;

    // lowerEdge should not be before the start of the pulse
    if(0 > lowerEdge)
        lowerEdge = 0;

    //Use this later for a more refined approach
    //  int minBin = 0;
    //
    //  while(minBin && (chargeVec[minBin-1] > 30))
    //      --minBin;

    const float binWidth = 1. / _readoutFrequency * 1.E9;
    double sliceTime = thePulse->getTrackerData()->getTime();

    //Fit values
    vector<double> tVal;
    vector<double> ampVal;

    for(int i = lowerEdge; i <= upperEdge; ++i)
    {
        tVal.push_back((sliceTime + (double)i)*binWidth);
        ampVal.push_back((double)chargeVec[i]);

        streamlog_out(DEBUG2) << " sliceTime : " << sliceTime  << " readoutFrequency : " << _readoutFrequency
                              << " tval : " << tVal.back() << " , ampVal " << ampVal.back()  << std::endl;
    }

    gROOT->SetBatch(1);

    //TGraph to start, maybe a TGraphErrors later
    TGraph* gr = new TGraph(tVal.size(), &(tVal[0]), &(ampVal[0]));
    int r = gr->Fit("gaus", "Q");

    if(r != 0)
    {
        timeAndError.first = -9E9;
        timeAndError.second = -9E9;
        return timeAndError;
    }

    double mu = gr->GetFunction("gaus")->GetParameter(1);
    double sigma = gr->GetFunction("gaus")->GetParameter(2);

    double muE = gr->GetFunction("gaus")->GetParError(1);
    double sigmaE = gr->GetFunction("gaus")->GetParError(2);

    timeAndError.first = mu - sigma;
    timeAndError.second = sqrt(sigmaE * sigmaE + muE * muE);

    delete gr;
    return timeAndError;

}//End _calculateGaussianInflexion();



std::pair<double, double> MMHitTimeCorrectionProcessor::_calculateGaussianMean(const EVENT::TrackerPulse* thePulse)
{

    //Return value
    std::pair<double, double> timeAndError;

    //Get the spectrum
    std::vector<float> chargeVec(thePulse->getTrackerData()->getChargeValues());

    int maxBin = 0;
    float maxVal = -9E9;

    //Find maximum amplitude bin
    for(unsigned int i = 0; i < chargeVec.size(); ++i)
    {
        if(chargeVec[i] > maxVal)
        {
            maxVal = chargeVec[i];
            maxBin = i;
        }

        if(0 < chargeVec[i])
        {
            streamlog_out(DEBUG2) << " found charge : " << chargeVec[i]
                                  << " in bin  " << i  << std::endl;
        }

    }

    //Define a range where it's mostly gaussian
    int lowerEdge = maxBin - 5;
    int upperEdge = maxBin + 5;

    // lowerEdge should not be before the start of the pulse
    if(0 > lowerEdge)
        lowerEdge = 0;

    //Use this later for a more refined approach
    //  int minBin = 0;
    //
    //  while(minBin && (chargeVec[minBin-1] > 30))
    //      --minBin;

    const float binWidth = 1. / _readoutFrequency * 1.E9;
    double sliceTime = thePulse->getTrackerData()->getTime();

    //Fit values
    vector<double> tVal;
    vector<double> ampVal;

    for(int i = lowerEdge; i <= upperEdge; ++i)
    {
        tVal.push_back((sliceTime + (double)i)*binWidth);
        ampVal.push_back((double)chargeVec[i]);

        streamlog_out(DEBUG2) << " sliceTime : " << sliceTime  << " readoutFrequency : " << _readoutFrequency
                              << " tval : " << tVal.back() << " , ampVal " << ampVal.back()  << std::endl;
    }

    gROOT->SetBatch(1);

    //TGraph to start, maybe a TGraphErrors later
    TGraph* gr = new TGraph(tVal.size(), &(tVal[0]), &(ampVal[0]));
    int r = gr->Fit("gaus", "Q");

    if(r != 0)
    {
        timeAndError.first = -9E9;
        timeAndError.second = -9E9;
        return timeAndError;
    }

    double mu = gr->GetFunction("gaus")->GetParameter(1);
    double muE = gr->GetFunction("gaus")->GetParError(1);

    timeAndError.first = mu;
    timeAndError.second = muE;

    delete gr;
    return timeAndError;

}



std::pair<double, double> MMHitTimeCorrectionProcessor::_calcPulseFitWithFunction(const EVENT::TrackerPulse* thePulse)
{
    /*
      This routine is used to fit the signal line
      shape of MM TPC for the ILD detector
      In this function we calculate the amplitude maximum and the time of arrival
      of this maximum (done with function _calcPulseFitWithFunction )

    */

    //streamlog_out(MESSAGE) << "Using Pulse fit with function  method\n";

    //Return value
    std::pair<double, double> timeAndError;

    // Extract Pulse time samples
    const std::vector<float>& adc = thePulse->getTrackerData()->getChargeValues();
    float maxCharge = *(std::max_element(adc.begin(), adc.end()));

    MMPulseShapeFunction* pulse_shape_func = new MMPulseShapeFunction();
    TF1* gamma5 = new TF1("gamma5", pulse_shape_func, &MMPulseShapeFunction::pulse_shape, 2, 7, 4, "MMPulseShapeFunction", "pulse_shape");

    std::vector<float> binArray;
    std::vector<float> binErr;
    std::vector<float> adcErr;
    int   maxBin = 0;
    float maxVal = -9E9;

    for(unsigned int i = 0; i < adc.size(); ++i)
    {
        binArray.push_back(i);
        if(adc[i] > maxVal)
        {
            maxVal = adc[i];
            maxBin = i;
        }
    }

    // Set  Noise = 7 ADC channel
    for(unsigned int i = 0; i < adc.size(); ++i)
    {
        binErr.push_back(0.0);
        adcErr.push_back(7.0);
    }

    //  TGraphErrors* pulseGraph = new TGraphErrors(10, &(binArray[0]), &(adc[maxBin-5]), &binErr[0], &adcErr[maxBin-5]);
    TGraph* pulseGraph = new TGraph(10, &(binArray[0]), &(adc[maxBin - 5]));
    TFitResultPtr fitresult = 0;
    gamma5->SetParameters(maxCharge, 5.0, 5.0, 1.1);
    gamma5->FixParameter(1, 5.0);
    gamma5->SetParLimits(3, 0.5, 2.0);
    try
    {
        fitresult = pulseGraph->Fit("gamma5", "QNO", 0, 2, 7);
    }
    catch(std::exception& e)
    {
        fitresult = -1;
        streamlog_out(WARNING) << "gamma5 fit did not work" << std::endl;
    }
    double time, timeError;
    if(fitresult == -1)
    {
        time = -10000;
        timeError = 10000;
    }
    else
    {
        time = thePulse->getTrackerData()->getTime() + gamma5->GetParameter(2) + maxBin - 5;
        const float bin_width = 1. / _readoutFrequency * 1.E9;
        time = time * bin_width; //conversion in ns
        timeError = sqrt(pow(gamma5->GetParError(2) * bin_width, 2));

        /*
          std::cout << "Number of samples = "   << adc.size() << " maxCharge = "
          << maxCharge << " maxBin = " << maxBin << std::endl;
          std::cout << "thePulse->getTrackerData()->getTime() = " << thePulse->getTrackerData()->getTime()  << std::endl
          << " par[0] = " << gamma5->GetParameter(0) << std::endl
          << " par[1] = " << gamma5->GetParameter(1) << std::endl
          << " par[2] = " << gamma5->GetParameter(2) << "+/-" << gamma5->GetParError(2) << std::endl
          << " par[3] = " << gamma5->GetParameter(3) << std::endl
          << " time = "   << time << "+/-" << timeError << " ns"<< std::endl;
        */

    }

    timeAndError.first = time;
    timeAndError.second = timeError;

    delete gamma5;
    delete pulseGraph;
    delete pulse_shape_func;

    return timeAndError;

}// _calcPulseFitWithFunction



}//End namespace marlintpc
