#include "MMPadResponseFunction.h"

//C++
#include <vector>
#include <cmath>
#include <iostream>

namespace marlintpc
{

  //Set Data functions
  void MMPadResponseFunction::SetData(PointsByRow& aVec)
  {
    _dataByRow = aVec;
  }

  void MMPadResponseFunction::SetData(std::vector< std::pair<double, float> >& aVec)
  {
    PointsByRow().swap(_dataByRow); //Clear and reallocate any stored data
    _dataByRow.push_back(aVec);
  }

  //Normalize Data functions
  void MMPadResponseFunction::normalizeData(std::vector< std::pair<double, float> >& data)
  {
    double norm = calculateRowAmp(x0, _params, data);

    for(std::vector<std::pair<double, float> >::iterator pulseIt = data.begin(); 
        pulseIt != data.end(); ++pulseIt)
    {
      pulseIt->second /= norm;
    }
  }

  void MMPadResponseFunction::normalizeData(PointsByRow& data)
  {
    for(unsigned int i = 0; i < data.size(); i++)
    {
      normalizeData(data[i]);
    }
  }

  //Chi2 operator
  double MMPadResponseFunction::operator()(const double* x) const 
  {

    // Initialize the chi-square to 0.
    double chi2 = 0.;

    // For each row...
    // (Have to use a C-style for-loop here to keep track of the row number)
    for(unsigned int i = 0; i < _dataByRow.size(); i++)
    {

      double norm = 0.;
      if(_fitType == MMPadResponseFunction::FIT_MEAN)  //In this case, parameters are stored in the class
        norm = calculateRowAmp(i, x[0], _params);
      else if(_fitType == MMPadResponseFunction::FIT_PARAMS) //In this case, x0 is assumed to be 0 and parameters are provided
        norm = calculateRowAmp(i, 0., x);
      else
        norm = calculateRowAmp(i, x[0], &x[1]);  //In this case, x0 is the first value in x, and the rest are the parameters

      // Loop over all the pulses in the hit to calculate the chi-square 
      for(std::vector<std::pair<double, float> >::const_iterator pulseIt = _dataByRow[i].begin(); 
          pulseIt != _dataByRow[i].end(); ++pulseIt)
      {

        if(_restrict && fabs(pulseIt->first - x0) > 5.0 ) continue;  //Restricting the domain, if applicable

        double numer = 0.;
        if(_fitType == MMPadResponseFunction::FIT_MEAN)
          numer = pulseIt->second - norm*prf(pulseIt->first - x[0], _params);
        else if(_fitType == MMPadResponseFunction::FIT_PARAMS)
          numer = pulseIt->second - norm*prf(pulseIt->first, x);
        else
          numer = pulseIt->second - norm*prf(pulseIt->first - x[0], &x[1]);

        double denom = pulseIt->second; //Take sqrt(A) to be the error

        if(denom == 0)  denom = 30.;  //This is a basic attempt to deal with padded hits, where "pulses" of 0-amplitude 
        //have been added to increase the size of hits, as in the FTPC code

        if(pulseIt->second >= 3846) denom *= 9; //This is a basic attempt to handle overrange pulses. A more sophisticated technique should be implemented.

        chi2 += ( numer * numer ) / denom; 
      }
    }

    // Now return the chi-square value.
    return chi2;
  }    

  //This calculates the normalization factor, or Row Amplitude
  double MMPadResponseFunction::calculateRowAmp(double x0, const double * x, const std::vector< std::pair<double, float> >& data) const
  {
    double amp = 0.;

    //Sum of pulse amplitudes
    if(_fitType == MMPadResponseFunction::SUM_FIT)
    {
      double sumAmp = 0.;

      for(std::vector<std::pair<double, float> >::const_iterator pulseIt = data.begin();
          pulseIt != data.end(); ++pulseIt)
      {
        sumAmp += pulseIt->second;
      }
      amp = sumAmp;   
    }

    //Amplitude of gaussian fit. Note that this is approximate, and requires more pulses for best results!
    else if(_fitType == MMPadResponseFunction::GAUS_FIT)
    {
      //(Un)weighted least squares gaussian fit, with fixed mean.
      double sum_yx2 = 0.;
      double sum_yx4 = 0.;
      double sum_ylny = 0.;
      double sum_yx2lny = 0.;
      double sum_y = 0.;
      for(std::vector<std::pair<double, float> >::const_iterator pulseIt = data.begin();
          pulseIt != data.end(); ++pulseIt)
      {
        double x = pulseIt->first - x0;
        double y = pulseIt->second;
        double lny = log(y);
        //        y = 1.; //Unweighted if used, weighted if commented.
        sum_yx2 += x*x*y;
        sum_yx4 += x*x*x*x*y;
        sum_ylny += y*lny;
        sum_yx2lny += y*x*x*lny;
        sum_y += y;
      }

      double delta = sum_yx2*sum_yx2 - sum_yx4*sum_y;
      double beta = (sum_yx2*sum_yx2lny - sum_yx4*sum_ylny) / delta;

      amp = exp(beta);
    }

    //Classic Min-Chi2 approach. Found by minimizing the Chi2 with respect to the normalization constant.
    else if(_fitType == MMPadResponseFunction::CHI2_MIN)
    {
      double sumAmp = 0.;
      double sumPRF = 0.;

      for(std::vector<std::pair<double, float> >::const_iterator pulseIt = data.begin();
          pulseIt != data.end(); ++pulseIt)
      {
        double p = prf( pulseIt->first - x0, x);
        double a = pulseIt->second;
        sumAmp += p;
        sumPRF += p*p/a;
      }
      amp = sumAmp / sumPRF;    
    }

    //Default normalization: sum of pulse amplitudes / sum of PRF values at corresponding x-values.
    else
    {
      double sumAmp = 0.;
      double sumPRF = 0.;

      for(std::vector<std::pair<double, float> >::const_iterator pulseIt = data.begin();
          pulseIt != data.end(); ++pulseIt)
      {
        sumAmp += pulseIt->second;
        sumPRF += prf( pulseIt->first - x0, x);
      }
      amp = sumAmp / sumPRF;    
    }

    return amp;
  }

  //Various useful methods of calling the calculateRowAmp function.
  double MMPadResponseFunction::calculateRowAmp(int row, double x0, const double * x, const PointsByRow& data) const
  {

    return calculateRowAmp(x0, x, data[row]);

  }

  double MMPadResponseFunction::calculateRowAmp(int row, double x0, const double * x) const
  {

    return calculateRowAmp(row, x0, x, _dataByRow);

  }

  //Definition of the _validPRFs static variable.
  std::map<std::string, std::pair<int, int> > MMPadResponseFunction::_validPRFs = MMPadResponseFunction::generatePRFMap();

}//End namespace marlintpc
