/*
 * PedSubRawDataFilterProcessor.cc
 *
 *  Created on: Oct 11, 2011
 *      Author: caiazza
 */

#include "PedSubRawDataFilterProcessor.h"

#include "marlin/Exceptions.h"
#include <algorithm>

namespace marlintpc
{
PedSubRawDataFilterProcessor aPedSubRawDataFilterProcessor;

PedSubRawDataFilterProcessor::PedSubRawDataFilterProcessor() :marlin::Processor("PedSubRawDataFilterProcessor")
{
    /* the constructor call first the super constructor which the name
     * of the processor
     * this must by the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "PedSubRawDataFilterProcessor filters the event setting the return value if the conditions are satisfied ";

    /* register steering parameters: name, description, class-variable, default value
     * the type will by defined by the type of the default value
     * string, double, float and int are possible
     * please note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERRAWDATA,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TPCRawData"));

    registerOptionalParameter(  "MinimumNumberOfPulses",
                                "The minumum number of pulses in the event to be accepted",
                                _minNumPulses,
                                0);

    registerOptionalParameter(  "MinimumPeakAverage",
                                "The minimum average of the pulse amplitudes in the event for this to be accepted",
                                _minPeakAverage,
                                static_cast<float>(0));
}

void PedSubRawDataFilterProcessor::init()
{
    printParameters();

    _nRun = 0;
    _nEvt = 0;
    inputCol = 0;
}

void PedSubRawDataFilterProcessor::processRunHeader(lcio::LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2711 $");
    for (marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }

    _nRun++;
}

void PedSubRawDataFilterProcessor::processEvent(lcio::LCEvent* evt)
{
    if((_nEvt % 1000) == 0)
    {
        streamlog_out(DEBUG) << " Processing event " << _nEvt << std::endl;
    }
    // get the input collection from the event
    try
    {
        inputCol = evt->getCollection(_inputColName);
    }
    catch (lcio::DataNotAvailableException &l)
    {
        // if there is not input collection give a warning to the user
        // note: this is not a graceful way to handle this exception
        streamlog_out(WARNING) << "Warning of PedSubRawDataFilterProcessor::processEvent: input collection "
                << _inputColName << " cannot be found in event " << _nEvt << std::endl;
        throw marlin::SkipEventException( this);
    }

    try
    {
        int numPulses = inputCol->getNumberOfElements();
        if(parameterSet("MinimumNumberOfPulses"))
        {
            _numPulsesOK = numPulses >= _minNumPulses;
        }
        else
        {
            _numPulsesOK = true;
        }

        if(parameterSet("MinimumPeakAverage"))
        {
            float avPeak = 0;
            for(int idPulse = 0 ; idPulse < numPulses ; idPulse ++)
            {
                TrackerRawData* inputData = dynamic_cast<TrackerRawData*>(inputCol->getElementAt(idPulse));
                EVENT::ShortVec adcValues = inputData->getADCValues();
                EVENT::ShortVec::const_iterator maxPulse = std::max_element(adcValues.begin(), adcValues.end());
                avPeak += static_cast<float>(*maxPulse) / numPulses;
            }
            _peakAverageOK = avPeak >= _minPeakAverage;
        }
        else
        {
            _peakAverageOK = true;
        }
    }
    catch (std::exception &e)
    {
        streamlog_out(WARNING) << _processorName << "::processEvent: "<< evt->getEventNumber()
                <<"Unknown exception caught while computing the filter conditions: " << e.what() << std::endl;
        throw marlin::SkipEventException(this);
    }

    setReturnValue(_numPulsesOK && _peakAverageOK);

    //A little bit of cleanup
    //The pointed LCIO objects are taken care of by the LCIO framework
    inputCol=0;

    _nEvt++;
}

void PedSubRawDataFilterProcessor::check(lcio::LCEvent*  /*evt*/)
{
    // nothing to check here - could be used to fill checkplots in reconstruction processor
}



void PedSubRawDataFilterProcessor::end()
{
    //   nothing done here
}

}


