#include "PulseFilterProcessor.h"
#include <iostream>
#include <string>
#include <math.h>
#include <algorithm>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerData.h>
#include <EVENT/TrackerPulse.h>
#include <Exceptions.h>

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{

PulseFilterProcessor aPulseFilterProcessor;


PulseFilterProcessor::PulseFilterProcessor() : Processor("PulseFilterProcessor")
{
  // the processor description
  _description = "PulseFilterProcessor is intended to remove hits with certain properties ";
  
  /* register steering parameters: name, description, class-variable, default value

    for example:
  */

  registerInputCollection(    LCIO::TRACKERPULSE,
            "InputCollectionName" ,
            "Name of the input TrackerPulse collection, default: TPCPulses"  ,
            _inputColName ,
            string("TPCPulses")) ;

  
  registerOutputCollection(   LCIO::TRACKERPULSE,
			      "OutputCollectionName" ,
			      "Name of the output TrackerData collection, default: GoodTPCPulses",
			      _outputCollectionName ,
			      string("GoodTPCPulses")) ;

  registerProcessorParameter("ChargeVariation",
			     "Accept pulses with a charge variation larger than this value",
			     _chargeVariation,
			     float(0.0));

  registerProcessorParameter("SlopeVariation",
			     "Accept pulses with a slope change larger than this value",
			     _slopeVariation,
			     float(0.0));

}


void PulseFilterProcessor::init()
{ 
  streamlog_out(DEBUG) << " init() called  " << endl;
  
  // you might want to list the used parameters by invoking
  printParameters();

}


void PulseFilterProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/)
{
  // do something for each run
} 


void PulseFilterProcessor::processEvent(lcio::LCEvent* evt)
{ 
  // do something for each event, usually access the input collection:

  // use an LCCollectionVec instead of LCCollection so we can use iterators later
  LCCollectionVec* inputColVec = 0;

  // do a try / catch in case there is an empty event
  try
  {
    inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
  }
  catch (DataNotAvailableException &e)
  {
    // no input data, just return and process the next event
    return;
  }

  LCCollectionVec* outputColVec = new LCCollectionVec(LCIO::TRACKERPULSE);
  outputColVec->setSubset(true);

  // loop the data collection and do something 
  for( LCCollectionVec::iterator inputObjectIter = inputColVec->begin() ;
       inputObjectIter < inputColVec->end() ;
       inputObjectIter++ )
  {
    // dynamic cast the input object (LCObject) to what it really is e.g.
    TrackerPulse * thePulse = dynamic_cast<TrackerPulse *>( *inputObjectIter );
    TrackerData * theData = thePulse->getTrackerData();
    FloatVec theCharges = theData->getChargeValues();

    //Look for the difference in amplitude within the pulse, the slope of rise or fall
    float minCharge = 0.;
    float maxCharge = 0.;
    float chargeVar = 0.;
    float slopeVar = 0.;
    FloatVec::iterator chargeIter;
    vector<float> derivativeVec;

    for(chargeIter = theCharges.begin(); chargeIter < theCharges.end(); chargeIter++)
      {
	if(chargeIter == theCharges.begin())
	  {
	    minCharge = (*chargeIter);
	    maxCharge = (*chargeIter);
	  }
	else
	  {
            if((*chargeIter) < minCharge) minCharge = (*chargeIter);
            if((*chargeIter) > maxCharge) maxCharge = (*chargeIter);
	  }
	//calculate derivatives
	if(chargeIter!=theCharges.end())
	  derivativeVec.push_back(*(chargeIter+1)-*chargeIter);
      }

    chargeVar = fabs(maxCharge - minCharge);
    double maxPosDerivative=*(max_element(derivativeVec.begin(),derivativeVec.end()));
    double minPosDerivative=*(min_element(derivativeVec.begin(),derivativeVec.end()));
    slopeVar = maxPosDerivative-minPosDerivative;

    //flat pulses
    if(chargeVar > _chargeVariation)
      {
	//slow rising or falling
	if(slopeVar>_slopeVariation)
	  outputColVec->addElement(thePulse);
      }
    else
      {
	//This part was copied from the PulseFinderProcessor and changed slightly
	if (streamlog::out.write< streamlog::DEBUG >())
	  {
	    streamlog_out(DEBUG) << "Run/Event: "<<evt->getRunNumber()<<" "<<evt->getEventNumber()<<endl;
	    streamlog_out(DEBUG) << "[***] Removed a pulse found on [Pad/Module]: [" << theData->getCellID0()
			 << "/" << theData->getCellID1() << "], at the time " << thePulse->getTime() << " of length "
			 << (theData->getChargeValues()).size()  << " containing charge " << thePulse->getCharge()
			 << " with quality flag " << thePulse->getQuality()
			 << endl;

	    streamlog_out(DEBUG) << "[***] The pulse spectrum is : ";
	    for (vector<float>::const_iterator adcVal = theCharges.begin(); adcVal < theCharges.end(); ++adcVal)
	      {
		streamlog_out(DEBUG) << *adcVal << " ";
	      }
	    streamlog_out(DEBUG) << endl;
	  }
      }
  }

  evt->addCollection(outputColVec, _outputCollectionName);

}


void PulseFilterProcessor::check(lcio::LCEvent*  /*evt*/)
{ 
  // check what is needed here 
}


void PulseFilterProcessor::end()
{ 
  // do something at the end of the processing, e.g. cleaning up
}
} // close namespace brackets
