#include "PulseFinder.h"

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// gsl
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_multifit.h>

//LCIO
#include "EVENT/TrackerData.h"
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>

//Streamlog
#include <streamlog/streamlog.h>

//C++
#include <iostream>
#include <algorithm>
#include <iterator>
#include <cmath>

//ROOT
#include "TF1.h"
#include "TGraph.h"
#include "TMath.h"
#include "TFitResultPtr.h"
#include "TROOT.h"

using namespace std;
using namespace lcio;
using namespace streamlog;

namespace marlintpc
{

  PulseFinder::PulseFinder(float minimumPulseHeight,
			   int minimumPulseLength,
			   float pulseStartPedestalWidthFactor,
			   float pulseEndPedestalWidthFactor,
			   float noiseValuePedestalWidthFactor,
			   PedestalListener const * pedestalListener,
			   ADCElectronicsListener const * adcElectronicsListener,
			   int saveNBinsBeforeStart,
			   int saveNBinsAfterEnd,
			   int timeCalculationMode,
			   double timeCalculationParameter,
			   int chargeCalculationMode) :
    _pulseStartPedestalWidthFactor(pulseStartPedestalWidthFactor),
    _pulseEndPedestalWidthFactor(pulseEndPedestalWidthFactor), 
    _noiseValuePedestalWidthFactor(noiseValuePedestalWidthFactor), 
    _timeCalculationMode(timeCalculationMode),
    _COGAtMaxParameter(2),
    _ConstFracParameter(0.9),
    _chargeCalculationMode(chargeCalculationMode),
    _minimumPulseHeight(minimumPulseHeight),
    _minimumPulseLength(minimumPulseLength),
    _saveNBinsBeforeStart(saveNBinsBeforeStart),
    _saveNBinsAfterEnd(saveNBinsAfterEnd),
    _pedestalListener(pedestalListener),
    _adcElectronicsListener(adcElectronicsListener),
    _usePedestalValueOverride(false),
    _usePedestalWidthOverride(false),
    _use_isPedestalSubtractedOverride(false),
    _usePolarityOverride(false),
    _useReadoutFrequencyOverride(false),
    _useMaxADCValueOverride(false),
    // the override values don't matter here, they are not used because all the use-flags are false
    _maximumADCValueOverride(0.),
    _readoutFrequencyOverride(0.),
    _pedestalValueOverride(0.),
    _pedestalWidthOverride(0.),
    _isPedestalSubtractedOverride(true),
    _polarityOverride(1),
    _forceSpectrumSave(false),
    _switchOffSplitting(false),
    _maxPlateauLength(0)
  {
    // only allow known calculation modes
    if ( ( _timeCalculationMode !=  INFLEXION_POINT ) &&
	 ( _timeCalculationMode !=  FIT_PARABOLA ) &&
	 ( _timeCalculationMode !=  CENTRE_OF_GRAVITY ) &&
	 ( _timeCalculationMode !=  FIRST_ABOVE_THRESHOLD ) &&
	 ( _timeCalculationMode !=  AT_MAXIMUM ) &&
	 ( _timeCalculationMode !=  CONSTANT_FRACTION ) &&
	 ( _timeCalculationMode !=  GAMMA_FIT ) &&
	 ( _timeCalculationMode !=  COG_ATMAX ) &&
	 ( _timeCalculationMode !=  BOX ) &&
	 ( _timeCalculationMode !=  GAUSSIAN_INFLEXION )
    )
    {
      // extend this error message when introducing new modes
      streamlog_out(ERROR) << "Unknown time calculation mode " << _timeCalculationMode 
			   << ". Valid modes are: " << std::endl;
      streamlog_out(ERROR) << "INFLEXION_POINT = " << INFLEXION_POINT <<std::endl
			   << "FIT_PARABOLA = " << FIT_PARABOLA <<std::endl
			   << "CENTRE_OF_GRAVITY = " << CENTRE_OF_GRAVITY<<std::endl
			   << "FIRST_ABOVE_THRESHOLD = " << FIRST_ABOVE_THRESHOLD <<std::endl
			   << "CONSTANT_FRACTION = " << CONSTANT_FRACTION <<std::endl
      	  	   << "GAMMA_FIT = " << GAMMA_FIT <<std::endl
      	  	   << "COG_ATMAX = " << COG_ATMAX <<std::endl
      	  	   << "BOX = " << BOX <<std::endl
	  	       << "GAUSSIAN_INFLEXION = " << GAUSSIAN_INFLEXION <<std::endl;


      throw lcio::Exception("PulseFinder: Unknown timeCalculationMode");
    }
    
    if ( ( _chargeCalculationMode !=  SUMQ ) &&
	 ( _chargeCalculationMode !=  FIT_PARABOLA ) &&
	 ( _chargeCalculationMode !=  MAXQ ))
    {
      // extend this error message when introducing new modes
      streamlog_out(ERROR) << "Unknown charge calculation mode " << _chargeCalculationMode 
			   << ". Valid modes are: " << std::endl;
      streamlog_out(ERROR) << "SUMQ = " << SUMQ <<std::endl
			   << "FIT_PARABOLA = " << FIT_PARABOLA <<std::endl
			   << " MAXQ = " <<  MAXQ <<std::endl;

      throw lcio::Exception("PulseFinder: Unknown chargeCalculationMode");
    }
   
    switch(_timeCalculationMode)
    {
	case CONSTANT_FRACTION:
		_ConstFracParameter=timeCalculationParameter;
		break;
	case COG_ATMAX:
		_COGAtMaxParameter=(int)timeCalculationParameter;
		break;
    }
	 
  }

  void PulseFinder::setMaxADCValueOverride(float maxADCValue)
  {
    _maximumADCValueOverride = maxADCValue;
    _useMaxADCValueOverride =  true;
  }

  void PulseFinder::setPedestalValueOverride(float pedestalValue)
  {
    _pedestalValueOverride = pedestalValue;
    _usePedestalValueOverride =  true;
  }

  void PulseFinder::setPedestalWidthOverride(float pedestalWidth)
  {
    _pedestalWidthOverride = pedestalWidth;
    _usePedestalWidthOverride =  true;
  }

  void PulseFinder::setReadoutFrequencyOverride(float readoutFrequency)
  {
    _readoutFrequencyOverride = readoutFrequency;
    _useReadoutFrequencyOverride =  true;
  }

  void PulseFinder::set_isPedestalSubtractedOverride(bool isPedestalSubtracted)
  {
    _isPedestalSubtractedOverride = isPedestalSubtracted;
    _use_isPedestalSubtractedOverride =  true;
  }

  void PulseFinder::setPolarityOverride(int polarity)
  {
    _polarityOverride = polarity;
    _usePolarityOverride =  true;
  }

  void PulseFinder::setPlateauCutOff(int maxPlateauLength)
  {
    _maxPlateauLength=maxPlateauLength;
  }

  std::string PulseFinder::getRevisionNumber()
  {
    return "$Rev: 5101 $"; 
  }


  std::vector<TrackerPulse *> PulseFinder::findPulses(TrackerData* inputData)
  {
    std::vector<TrackerPulse *> pulseVector; // the return value

    // a vector with TrackerData and the quality flag for the pulses
    vector<pair<TrackerDataImpl*, int> > pulseCandidates;

    findPulseCandidates(inputData, &pulseCandidates);

    // fill the candiates into TrackerPulse objects
    // use the found pulses to calculate time and charge information and store the data
    for (vector<pair<TrackerDataImpl*, int> >::iterator pulseCandidateIter = pulseCandidates.begin();
	 pulseCandidateIter < pulseCandidates.end();
	 ++pulseCandidateIter)
    {
      TrackerDataImpl* trackerData = pulseCandidateIter->first;
      int quality                  = pulseCandidateIter->second;

      //ceate a TrackerPulse object
      TrackerPulseImpl* newPulse = new TrackerPulseImpl();
      newPulse->setCellID0(trackerData->getCellID0());
      newPulse->setCellID1(trackerData->getCellID1());

      Pulse myPulse = calculatePulseTimeAndCharge(trackerData);
      newPulse->setCharge(myPulse.getCharge());
      newPulse->setTime(myPulse.getTime());
      newPulse->setQuality(quality);
      vector<float> covarianceMatrix; // the "newly" added error information concerning the pulse
      covarianceMatrix.push_back(myPulse.getChargeError()*myPulse.getChargeError());
      covarianceMatrix.push_back(0.);
      covarianceMatrix.push_back(myPulse.getTimeError()*myPulse.getTimeError());
      newPulse->setCovMatrix(covarianceMatrix);

      if (streamlog::out.write< streamlog::DEBUG >())
      {
	streamlog_out(DEBUG) << "[***] Found a pulse on [Pad/Module]: [" << trackerData->getCellID0()
		     << "/" << trackerData->getCellID1()
		     << "], at the time " << newPulse->getTime() 
		     << " +- " << covarianceMatrix.at(2) << " of length "
		     << (trackerData->getChargeValues()).size() 
		     << " containing charge " << newPulse->getCharge()
		     << " +- " << covarianceMatrix.at(0)
		     << " with quality flag " << quality
		     << " (MultiplePulseCandidate:" << pulseflag::isMultiplePulseCandidate(quality) 
		     << ", "
		     << "AnomalousShape:" << pulseflag::isAnomalousShape(quality) << ", "
		     << "PlateauCutoff:" << pulseflag::isPlateauCutoff(quality) << ", "
		     << "Underflow:" << pulseflag::isUnderflow(quality) << ", "
		     << "Overflow:" << pulseflag::isOverflow(quality) << ")."
		     << "split:" << pulseflag::isSplit(quality) << ")."
		     << endl;

	streamlog_out(DEBUG) << "[***] The pulse spectrum is : ";

	vector<float> theADCSpectrum = trackerData->getChargeValues();
	for (vector<float>::const_iterator adcVal = theADCSpectrum.begin(); 
	     adcVal < theADCSpectrum.end(); ++adcVal)
	{
	  streamlog_out(DEBUG) << *adcVal << " ";
	}
	streamlog_out(DEBUG) << endl;
      }

      // store the adc spectrum connected to the pulse
      // this is implicitly done by incorporating /_forceSpectrumSave/ into /quality/
      if (quality != 0 || _forceSpectrumSave)
      {
	newPulse->setTrackerData(trackerData);
      }
      else // if not added to the pulse, delete the TrackerData object
      {
	delete trackerData;
      }

      pulseVector.push_back(newPulse);
    }// for pulseCandidateIter
    
      return pulseVector;
  }

  void PulseFinder::findPulseCandidates(TrackerData* thisData, vector<pair<TrackerDataImpl*, int> >* tempPulses)
  {
    /* the basic idea to find pulses is split into three steps:
       (1) simply look for candidate pulses: above values between start and end threshold
          due to the high frequency of this already check here for a minimum length
       (2) search the candidate pulses for possible multiple pulse and split, if necessary
       (3) finally check the pulse boundaries and assemble the pulse out of the bins before and after the pulse, if possible
    */

    // get the maximum ADC value
    const float maximumADCValue = getMaxADCValue(thisData->getCellID1());

    float startTime  = thisData->getTime();

    // To shorten the syntax we use a variable to store the spectrum.
    // It's important to use a reference and not make a copy, so the 
    // iterators can be used with it.
    vector<float> const & adcSpectrum = thisData->getChargeValues();

    /* this is a help construct for the data (including a typedef)
       simply store the start and end position of the pulse
    */
    typedef std::vector<float>::const_iterator timeIter;
    vector<pair<timeIter, timeIter> > theCandidatePulses;

    /* do the first task:
       simply look for the values above the start threshold and below end threshold
    */
    brutePulseSearch(thisData, &theCandidatePulses);

    /* do the check for multiple pulses,
       technically: divide the start- and end-iterator into a new pair
       TODO: in addition to pulse splitting adjust the charge at the boundaries
    */
    vector<InternalSplitPulse> theExtendedCandidatePulses =
      splitMultiplePulses(&theCandidatePulses,
			  _noiseValuePedestalWidthFactor*getPedestalWidth(thisData->getCellID0(),
									  thisData->getCellID1()) );

    /* as last part determine the possible boundaries of the pulse;
       - check wether the storing of the time bins before and after the end is possible
       - set the flags accordingly
    */

    // this is the supposed minimum time between two pulses
    int nominalSeparation =  _saveNBinsBeforeStart + _saveNBinsAfterEnd;

    /* now prepare the pulse candidates for storage:
     - assign quality flags
     - check boundaries at the beginning and the end
     - cut off if plateau is found, if this option is set
     - return the data construct for storage
    */
    
    for (vector<InternalSplitPulse>::const_iterator valueAddedPulseBoundary = theExtendedCandidatePulses.begin();
         valueAddedPulseBoundary < theExtendedCandidatePulses.end(); ++valueAddedPulseBoundary)
      {
	
	int quality = 0;

        // check if we're on the first or last pulse boundaries
        timeIter previous_pulse_end, next_pulse_start;

        // first special case: first pulse in question -> set previous end to the beginning of the ADCspectrum
        if (valueAddedPulseBoundary == theExtendedCandidatePulses.begin())
          previous_pulse_end = adcSpectrum.begin();
        else
          previous_pulse_end = (valueAddedPulseBoundary - 1)->endBin;
        // second special case: the end pulse -> set next start to the end of the adcSpectrum
        if (valueAddedPulseBoundary == (theExtendedCandidatePulses.end() - 1))
          next_pulse_start = adcSpectrum.end();
        else
          next_pulse_start = (valueAddedPulseBoundary + 1)->startBin;

        // check for a double-pulse candidate
        // this is a misnomer? this specifies if a pulse has a too close neighbour
	// Only check for nominal separation between pulses, not at begin and end of the sequence
	
	// distance to previous pulse
        if ( (valueAddedPulseBoundary != theExtendedCandidatePulses.begin()) &&
	     (distance(previous_pulse_end, valueAddedPulseBoundary->startBin) < nominalSeparation) )
          {
            pulseflag::setMultiplePulseCandidate(quality);
          }

	// distance to next pulse
        if ( (valueAddedPulseBoundary != (theExtendedCandidatePulses.end() - 1)) &&
	     (distance(valueAddedPulseBoundary->endBin, next_pulse_start) < nominalSeparation) )
	  {
            pulseflag::setMultiplePulseCandidate(quality);
          }

        // check for anomalous pulse shape -- TODO: how to determine this in general?
        if (false)
          {
            pulseflag::setAnomalousShape(quality);
          }

        // check for underflow
        //TODO: what is underflow? surely there will be some bin values of "zero" charge in simulations. perhaps this only makes sense when noise is included in the channel
        if (false)
          {
            pulseflag::setUnderflow(quality);
          }

        // check for overflow
	const timeIter max = max_element(valueAddedPulseBoundary->startBin, valueAddedPulseBoundary->endBin);
	float maxBinValue = maximumADCValue;
        // if data has already been pedestal subtracted, take this into account for overflow value maxBinValue ( = maximumADCValue)
        if (isPedestalSubtracted(thisData->getCellID1()))
          {
            maxBinValue = maximumADCValue - floor(getPedestalValue(thisData->getCellID0(), thisData->getCellID1()));
          }

        if (*max >= maxBinValue) // overflow in at least one time bin
          {
            pulseflag::setOverflow(quality);;
          }

        if (valueAddedPulseBoundary->isSplit)
          {
            pulseflag::setSplit(quality);
          }
	
        // now copy the data to a TrackerData object
        // determine the start and end positions for the copy:
        timeIter copy_start = valueAddedPulseBoundary->startBin;
        timeIter copy_end = valueAddedPulseBoundary->endBin;
	bool hasPlateau=false;
	float pedwidth=getPedestalWidth(thisData->getCellID0(),thisData->getCellID1());
	//Search if pulse has a plateau if requested and override copy_end position to cut off on plateau
	//TODO: check if something sits on plateau at later time
	if(_maxPlateauLength)
	  {
	    double noise=_noiseValuePedestalWidthFactor*pedwidth;
	    int plateau_length=0;
	    float plateau_charge=0;
	    timeIter plateau_start=valueAddedPulseBoundary->startBin;
	    for(timeIter bin=valueAddedPulseBoundary->startBin; bin<valueAddedPulseBoundary->endBin-1;bin++)
	      streamlog_out(DEBUG)<<*bin<<",";
	    streamlog_out(DEBUG)<<endl;
	   
	    for(timeIter bin=valueAddedPulseBoundary->startBin; bin<valueAddedPulseBoundary->endBin-1;bin++)
	      {
		double derivative=(*(bin+1)-*bin);
		if (fabs(derivative)<noise)
		  {
		    if (plateau_length==0)
		      plateau_start=bin;
		    plateau_length++;
		    plateau_charge+=*bin;
		  }
		else
		  {
		    plateau_length=0;
		    plateau_charge=0;
		    plateau_start=bin;
		  }
		if (plateau_length>_maxPlateauLength)
		  {
		    //only accept plateau if it is not in overflow
		    if(plateau_charge/plateau_length<maxBinValue-noise)
		      {
			copy_end = bin;
			hasPlateau=true;
			pulseflag::setPlateauCutoff(quality);
			streamlog_out(DEBUG)<<"Found plateau, starting at "<<*bin
					    <<" cutting after "<<distance(copy_start, copy_end)
					    <<" with charge "<< plateau_charge/plateau_length
					    <<", max= "<<maxBinValue<<endl;
			break;
		      }
		    else //keep looking
		      {
			streamlog_out(DEBUG)<<"Found plateau, in overrange"<<endl; 
			plateau_length=0;
			plateau_charge=0;
			plateau_start=bin;
		      } 
		  }
	      }
	  }
        

	//check if pulse criteria are met after the brute search and splitting.
	//if splitting occured on a plateau it can create pulses that are just a snippet of plateau and no real
	//pulse, or do not have the required maximum length or height or is flat
	const timeIter maxP = max_element(copy_start, copy_end);
	const timeIter minP = min_element(copy_start, copy_end);

	bool validpulse=true;
	streamlog_out(DEBUG)<<"Checking pulse: ";
	int polarity = getPolarity( thisData->getCellID1() );
	const bool negativePolarity = (polarity == -1);

	const bool meets_height_requirement = negativePolarity ? (*maxP <= _minimumPulseHeight) : (*maxP >= _minimumPulseHeight);
	const bool meets_length_requirement = (distance(copy_start,copy_end) >= _minimumPulseLength);
	const bool meets_chargeVariation = (negativePolarity ? ((*maxP-*minP)<=_minimumPulseHeight-pedwidth):((*maxP-*minP)>=_minimumPulseHeight-pedwidth));
	streamlog_out(DEBUG)<<"max="<<*maxP <<"("<<_minimumPulseHeight<<") , length="
			    <<distance(copy_start,copy_end)<<" ("<<_minimumPulseLength<<") , charge variation="
			    <<*maxP-*minP;
	if (!(meets_height_requirement && meets_length_requirement && meets_chargeVariation))
	  {
	    streamlog_out(DEBUG)<<"  ->Rejecting pulse"<<endl;
	    validpulse=false;
	  }
	else
	  streamlog_out(DEBUG)<<endl;
      	//if requirements are not met skip this pulse
	if(!validpulse)
	  continue;
	
	// check wether the start point can be moved in front to the nonimal position
        const int start_displacement = std::max(-_saveNBinsBeforeStart, static_cast<int>(distance(copy_start, previous_pulse_end)));
        advance(copy_start, start_displacement);
        // check wether the end point can be moved the nonimally requested position
	//if a plateau was found cut directly there without adding postsamples
	if (!hasPlateau)
	  advance(copy_end, min(_saveNBinsAfterEnd, static_cast<int>(distance(copy_end, next_pulse_start))));

 
	

       /* calculate the start time of the pulse to be stored:
        * this is calculated from the time of the pulse (see far above) and needs to be corrected for the modifications
        * within the pulse from its calculation
        */
        float currentTime = startTime + (valueAddedPulseBoundary->startBin - adcSpectrum.begin()) + start_displacement;

        TrackerDataImpl* toStoreTrackerData;
        toStoreTrackerData = new TrackerDataImpl();
        toStoreTrackerData->setCellID0(thisData->getCellID0());
        toStoreTrackerData->setCellID1(thisData->getCellID1());
        toStoreTrackerData->setTime(currentTime);
        toStoreTrackerData->chargeValues().resize(distance(copy_start, copy_end));

        copy(copy_start, copy_end, toStoreTrackerData->chargeValues().begin());

        pair<TrackerDataImpl*, int> tempPair(toStoreTrackerData, quality);
        tempPulses->push_back(tempPair);
      }

    return;
  }



  void PulseFinder::brutePulseSearch(TrackerData const * inputTrackerData,
				     vector<pair<vector<float>::const_iterator, 
						 vector<float>::const_iterator> >* theCandidatePulses)
  {
    typedef vector<float>::const_iterator timeIter;
    timeIter candStart, candEnd;

    // The actual thresholds where to start the pulse.
    // Calculated as a user defined multiple of the pedestalWidth.
    // Note: The pededestalWidth from LCCD can be overridden by the user.
    // FIXME: It should be thr = pedestal + factor*pedestalWidth,
    //        if the data is not pedestal subtracted.
    float pedestalWidth = getPedestalWidth(inputTrackerData->getCellID0(),
					   inputTrackerData->getCellID1());
    float startThreshold =  pedestalWidth * _pulseStartPedestalWidthFactor;
    float endThreshold   =  pedestalWidth * _pulseEndPedestalWidthFactor;

    int polarity = getPolarity( inputTrackerData->getCellID1() );

    vector<float> const & adcSpectrum = inputTrackerData->getChargeValues();

    // determine if the pulse is negative
    const bool negativePolarity = (polarity == -1);

    // also look if the pulse candidate meets the minimum requirements of length and height
    // loop through the pulse data looking for continuous areas of charge

    vector<float>::const_iterator current_value = adcSpectrum.begin();

    int candidate_length = 0;
    float candidate_max_height = 0.;
    bool inside_pulse = false;
    while (current_value < adcSpectrum.end())
      {
	if (!inside_pulse)
          {
            // we're not inside a pulse, look for the start
            const bool above_start_threshold = negativePolarity ? (*current_value <= startThreshold) : (*current_value >= startThreshold);
	   
            if (above_start_threshold)
              {
                inside_pulse = true;
                candidate_length = 1;
                candidate_max_height = *current_value;
		candStart = current_value;
              }
          }
        else
          {
            // we're inside a pulse, look for the end
            const bool below_end_threshold = negativePolarity ? (*current_value > endThreshold) : (*current_value < endThreshold);
            const bool end_of_data = (current_value + 1 == adcSpectrum.end());

            if (!below_end_threshold)
              {
                ++candidate_length;
                candidate_max_height = negativePolarity ? min(*current_value, candidate_max_height) : max(*current_value, candidate_max_height);
	
		// in case we are at the end of the ADC spectrum and still above the threshold 
		// we have to increase the iterator to correctly point to the positon after the last
		// sample in the pulse. It is safe to increase here because we are done with the
		// calculations.
		if (end_of_data)
		  {
		    ++current_value;
		  }
              }

            if (below_end_threshold || end_of_data)
              {
		streamlog_out(DEBUG1) <<std::endl;
	
		// determine if this pulse is a candidate
                const bool meets_height_requirement = negativePolarity ? (candidate_max_height <= _minimumPulseHeight) : (candidate_max_height >= _minimumPulseHeight);

                const bool meets_length_requirement = (candidate_length >= _minimumPulseLength);
	
                if (meets_height_requirement && meets_length_requirement)
                  {
                    streamlog_out(DEBUG1) << "Candidate accepted. Reason?: max height " << candidate_max_height << ", length " << candidate_length << endl;
                    theCandidatePulses->push_back(pair<timeIter, timeIter>(candStart, current_value));
                  }
                else
                  {
                    // this feels like fishing, the pulse wasn't big enough so toss it back
                    streamlog_out(DEBUG1) << "Candidate rejected. Reason?: below height " << !meets_height_requirement << ", below length " << !meets_length_requirement << endl;
                  }

                // reset the candidate
                inside_pulse = false;
	      }
          }

	// Do not increase the iterator if it already is end(). This would cause an invalid iterator.
	if (current_value != adcSpectrum.end())
	{
	  ++current_value;
	}
      }
  }



  vector<PulseFinder::InternalSplitPulse> PulseFinder::splitMultiplePulses(
	  vector<pair<vector<float>::const_iterator,
		      vector<float>::const_iterator> >* theCandidatePulses,
	  float noiseThreshold)
  {
    typedef vector<float>::const_iterator TimeIter;

    // create a temporary return object for the "new" separated pulses
    vector<InternalSplitPulse> tempPulseStorage;

    // loop all the incoming iterator pairs, i.e. pulse candidates
    for (vector<pair<TimeIter, TimeIter> >::iterator aPulse = theCandidatePulses->begin(); 
	 aPulse < theCandidatePulses->end(); ++aPulse)
      {
	//        vector<float> derivativeVec;
	//        vector<unsigned int> splitPositions;

	/* The algorithm implements a state machine which processes the incoming samples.
	 * There are four different states:
	 * - fallingOrPlateau : When a rising slope larger than the noise is 
	 *                      detected, the pulse is split.
	 * - risingEdge: rising slope larger than the noise threshold, no splitting.
	 * - slowRisingEdge: state was risingEdge but the slope now is smaler than the noise 
	 *                   treshold. If the next slope is 0 or negative the state is changed 
	 *                    to fallingOrPlateau.
	 * - dipInRisingEdge: state was risingEdge, but the slope now is 0 or negative, but still 
	 *                    larger than -noiseThreshold. If the next derivative is not rising 
	 *                    steeper than the noise treshold, the state is changed 
	 *                    to fallingOrPlateau.
	 */
	static int const fallingOrPlateau = 0;
	static int const risingEdge = 1;
	static int const slowRisingEdge = 2;
	static int const dipInRisingEdge = 3;

	// starting with risingEdge prevents from cutting low slopes at the beginning of the pulse
	int state = risingEdge;
	TimeIter beginIter = aPulse->first; // remember the start of the current split section
	//only try the splitting if it is not siwtched of
	if(_switchOffSplitting==false)
	  {
	    for (TimeIter binIter = aPulse->first; binIter < (aPulse->second-1); ++binIter)
	      {
		// calculate the derivative.
		float derivative = *(binIter + 1) - *binIter;
		
		switch (state)
		  {
		  case risingEdge:
		    if (derivative <= -noiseThreshold)
		      {
			// definitive a falling edge
			state = fallingOrPlateau;
		      }
		    else if (derivative <= 0)
		      {
			// slowly falling, might be noise if it's just one bin
			state = dipInRisingEdge;
		      }
		    else if (derivative <= noiseThreshold)
		      {
			// slowly rising (0 or negative are already handled)
			state = slowRisingEdge;
		      }
		    // nothing to do for (derivative > noiseThreshold), state already is risingEdge
		    break;
		    
		  case slowRisingEdge:
		    if (derivative <= 0)
		      {
			// as the slope was not steeply rising this 
			// is considered as plateau (three consecutive samples are the same within the noise)
			// if slowly falling, or falling if < - noiseThreshold
			state = fallingOrPlateau;
		      }
		    else if (derivative > noiseThreshold)
		      {
			// pulse is rising steeply again
			state = risingEdge;
		      }
		    // consecutive slowly rising is consideres as a rising trend, we keep the state
		    break;
		    
		  case dipInRisingEdge:
		    if (derivative > noiseThreshold)
		      {
			// pulse is rising steeply again
			state = risingEdge;
		      }
		    else
		      {
			// If the pulse is not rising steeply after the one slowly falling derivative,
			// the rising edge is ending here.
			state = fallingOrPlateau;
		      }
		    break;
		    
		  case fallingOrPlateau: 
		    if (derivative > noiseThreshold)
		      {
			// This is the split condition: steep rising slope after falling or plateau.
			// Perform the splitting and change the state to risingEdge.
			// Store the pulse up to the rising slope in the list of pulses
			// with the split flag turned on.
			tempPulseStorage.push_back(InternalSplitPulse(beginIter, binIter+1, true));
			beginIter = binIter+1; // remember where the next pulse starts
			state = risingEdge;
		      }
		    break;
		  default: ;
		    // should never happen, throw an exception
		  }
		
	      }// for binIter
	  }
	// Store the final (probaly only pulse), which ends at the end of the input sequence
	// If the beginIter is still  aPulse->first the pulse was not split (flag = false),
	// else the pulse was split -> flag is true
	tempPulseStorage.push_back(InternalSplitPulse(beginIter,  aPulse->second, 
						      (beginIter == aPulse->first ? false : true)));
      }// for aPulse

    return tempPulseStorage;
  }



  PulseFinder::Pulse PulseFinder::calculatePulseTimeAndCharge(const EVENT::TrackerData *pulse_data)
  {
    Pulse returnValue;

    // Check if the fit method is used.
    // Call the fit only once because fitting is expensive.
    if ( (_chargeCalculationMode == FIT_PARABOLA) || (_timeCalculationMode == FIT_PARABOLA) )
    {
      returnValue = (calculatePulseWithFit(pulse_data));
    }

    switch (_chargeCalculationMode)
    {
      case SUMQ:
	returnValue.set(calcPulseChargeSumQ(pulse_data));
	break;
      case FIT_PARABOLA:
	//nothing to do. returnValue already filled before
	break;
      case MAXQ:
	returnValue.set(calcPulseChargeMaxQ(pulse_data));
	break;
      default:
	// we already checked this in the constructor. Should not happen
	streamlog_out(ERROR) << "PulseFinder: Internal error: Unknown chargeCalculationMode "
			     << " should have been caught by the constructor!" << std::endl;
	throw lcio::Exception("PulseFinder: Internal error: Unknown chargeCalculationMode");
    }

    switch (_timeCalculationMode)
      {
      case INFLEXION_POINT:
	returnValue.set(calcPulseTimeInflexionPoint(pulse_data));
	break;
      case FIT_PARABOLA:
	//nothing to do. returnValue already filled before
	break;
      case CENTRE_OF_GRAVITY:
	returnValue.set(calcPulseTimeCentreOfGravity(pulse_data));
	break;
      case FIRST_ABOVE_THRESHOLD:
	returnValue.set(calcPulseTimeFirstBinAboveThreshold(pulse_data));
	break;
      case AT_MAXIMUM:
	returnValue.set(calcPulseTimeAtMaximum(pulse_data));
	break;
      case CONSTANT_FRACTION:
	returnValue.set(calcPulseTimeAtConstantFraction(pulse_data));
	break;
      case GAMMA_FIT:
      	returnValue.set(calcPulseTimeWithGammaFit(pulse_data));
	break;
      case COG_ATMAX:
      	returnValue.set(calcPulseTimeCOGAtMaximum(pulse_data));
    break;
      case BOX:
        returnValue.set(calcPulseTimeBoxMethod(pulse_data));
    break;
      case GAUSSIAN_INFLEXION:
        returnValue.set(calcuPulseTimeGaussianInflexion(pulse_data));
	break;
      default:
	// we already checked this in the constructor. Should not happen
	streamlog_out(ERROR) << "PulseFinder: Internal error: Unknown timeCalculationMode "
			     << " should have been caught by the constructor!" << std::endl;
	throw lcio::Exception("PulseFinder: Internal error: Unknown timeCalculationMode");
    }

    return returnValue;
  }


  PulseFinder::PulseTime PulseFinder::calcPulseTimeWithGammaFit(const lcio::TrackerData *thisPulse)
    {
    /*
     * Parameter 0: Pulse height at Maximum
     * Parameter 1: Order
     * Parameter 2: Start time
     * Parameter 3: Peak time
     * The maximum is at start time + peak time.
     */
	    const vector<float> &charge_values = thisPulse->getChargeValues();
		TF1* gamma4 = new TF1("gamma4","([0]*exp([1])*((x-[2])/[3])**[1]*exp(-[1]*((x-[2])/[3])))*(x>[2])",0,28);
	   // TF1 * gamma = new TF1("gamma","[3]*TMath::GammaDist(x,[0],[1],[2])");
	    float maxCharge= *(std::max_element(charge_values.begin(),charge_values.end()));
	   // gamma->SetParameters(3.731,1.14,1.003,4*maxCharge);
	   // gamma->FixParameter(0,3.731);
	   // gamma->FixParameter(2,1.003);
	    gamma4->SetParameters(maxCharge,4,0.5,3.5);
	    gamma4->FixParameter(1,4);
	    vector<float> binArray;
	    for(unsigned int i=0;i<charge_values.size();i++)
	    	binArray.push_back(i);
	    TGraph* pulseGraph = new TGraph(charge_values.size(),&(binArray[0]),&(charge_values[0]));
	    TFitResultPtr fitresult=0;
	    try
	    {
	    	fitresult=pulseGraph->Fit("gamma4","QNO");
	    }
	    catch(std::exception & e)
	    {
	    	fitresult=-1;
	    	streamlog_out(WARNING)<<"gamma4 fit did not work"<<endl;
	    }
	    double time,timeError;
	    PulseTime returnValue;
	    if (fitresult==-1)
	    {
	    	time =-10000;
	    	timeError=10000;
	    }
	    else
	    {
	    	time=thisPulse->getTime()+gamma4->GetParameter(2)+gamma4->GetParameter(3);
	    	const float bin_width = 1.E03 / getReadoutFrequency(thisPulse->getCellID1());
	    	time = time * bin_width; //conversion in ns
	    	timeError =sqrt(pow(gamma4->GetParError(2)*bin_width,2)+pow(gamma4->GetParError(3)*bin_width,2));

	    }
	    returnValue.setTime(time);
	    returnValue.setTimeError(timeError);
	    delete gamma4;
	    delete pulseGraph;

	    return returnValue;

    }

  PulseFinder::Pulse PulseFinder::calculatePulseWithFit(const EVENT::TrackerData *pulse_data)
  {
    // find the polarity of the pulse data
    const int polarity = getPolarity(pulse_data->getCellID1());

    const vector<float> &charge_values = pulse_data->getChargeValues();

    // fit a parabola to the log of the data around the maximum

    // find the maximum (or minimum)
    vector<float>::const_iterator extreme_point = (polarity == -1.) ? min_element(charge_values.begin(), charge_values.end()) : max_element(charge_values.begin(), charge_values.end());

    streamlog_out(DEBUG) << "Spectrum is ";

    // use 3 points on each side (at most)

    for (vector<float>::const_iterator i_bin = charge_values.begin();
	 i_bin != charge_values.end(); ++i_bin)
      {
        streamlog_out(DEBUG) << ", " << *i_bin;
      }

    streamlog_out(DEBUG) << endl;

    // check the begining pulses for non-negativity
    vector<float>::const_iterator fit_start = extreme_point;
    const int max_reverse = min(3, static_cast<int>(distance(charge_values.begin(), fit_start)));
    for (int i = 0; i < max_reverse; i++)
      {
        const float previous_value = *(fit_start - 1);
        if ((previous_value * polarity) > 0)
          {
            fit_start = fit_start - 1;
          }
        else
          {
            break;
          }
      }

    // and check the ending ones
    vector<float>::const_iterator fit_end = extreme_point;
    const int max_advance = min(3, static_cast<int>(distance(fit_end, charge_values.end())));
    for (int i = 0; i < max_advance; i++)
      {
        const float next_value = *(fit_end + 1);
        if ((next_value * polarity) > 0)
          {
            fit_end = fit_end + 1;
          }
        else
          {
            break;
          }
      }

    // initialize the fit
    const int point_count = distance(fit_start, fit_end) + 1;

    // make sure that the distance is greater than 3
    if (point_count == 1)
      {
        const int bin_time = static_cast<int>(pulse_data->getTime() + distance(charge_values.begin(), fit_start) + 0.5);
        const float time_bin_width = 1.E03 / getReadoutFrequency(pulse_data->getCellID1());
        const float fit_time = bin_time * time_bin_width;

        const float fit_amplitude = *fit_start;

        streamlog_out(DEBUG0) << "Time = 0.5 + " << distance(charge_values.begin(), fit_start) << " + " << pulse_data->getTime() << " = " << bin_time << "bins" << endl;
        streamlog_out(DEBUG0) << "Bin width = " << time_bin_width << endl;

        Pulse returnValue;
        returnValue.setCharge(fit_amplitude);
        returnValue.setTime(fit_time);

        return returnValue;
      }


    gsl_matrix *X = gsl_matrix_alloc(point_count, 3);
    gsl_vector *y = gsl_vector_alloc(point_count);
    gsl_vector *c = gsl_vector_alloc(3);
    gsl_matrix *covariance = gsl_matrix_alloc(3, 3);
    double minimum = 0.;

    streamlog_out(DEBUG0) << "Calculating pulse time using the points ";

    vector<float>::const_iterator current_charge = fit_start;

    for (int row = 0; current_charge <= fit_end; ++current_charge, ++row)
      {
        // prevent bad log values
        if ((*current_charge * polarity) > 0.001)
          {
            gsl_vector_set(y, row, std::log(*current_charge * polarity));
          }
        else
          {
            continue;
          }

        // may as well transform the fit in the t-direction
        gsl_matrix_set(X, row, 0, 1.);
        gsl_matrix_set(X, row, 1, row);
        gsl_matrix_set(X, row, 2, row*row);

        streamlog_out(DEBUG0) << '(' << row << ", " << *current_charge << ") ";
      }

    gsl_multifit_linear_workspace *workspace = gsl_multifit_linear_alloc(point_count, 3);

    // perform the fit
    gsl_multifit_linear(X, y, c, covariance, &minimum, workspace);

    streamlog_out(DEBUG0) << ". Got: " << gsl_vector_get(c, 2) << "x^2 + " << gsl_vector_get(c, 1) << "x + " << gsl_vector_get(c, 0) << endl;

    // calculate the time
    // convert the result to nanoseconds
    const double translated_bin = -0.5 * gsl_vector_get(c, 1) / gsl_vector_get(c, 2);
    const double bin_time = pulse_data->getTime() + translated_bin + distance(charge_values.begin(), fit_start) + 0.5;

    const double time_bin_width = 1.E03 / getReadoutFrequency(pulse_data->getCellID1());
    const double fit_time = bin_time * time_bin_width;

    streamlog_out(DEBUG0) << "Time = " << translated_bin << " + 0.5 + " << distance(charge_values.begin(), fit_start) << " + " << pulse_data->getTime() << " = " << bin_time << "bins" << endl;
    streamlog_out(DEBUG0) << "Bin width = " << time_bin_width << endl;

    // calculate the amplitude
    const float max_log = gsl_vector_get(c, 0) - (0.25 * gsl_vector_get(c, 1) * gsl_vector_get(c, 1) / gsl_vector_get(c, 2));
    const float fit_amplitude = exp(max_log);

    // clean up the fit
    gsl_multifit_linear_free(workspace);
    gsl_matrix_free(X);
    gsl_vector_free(y);
    gsl_vector_free(c);

    Pulse returnValue;
    returnValue.setCharge(fit_amplitude);
    returnValue.setTime(fit_time);

    return returnValue;
  }



  PulseFinder::PulseTime PulseFinder::calcPulseTimeInflexionPoint(const TrackerData* thisPulse)
  {
    vector<float> chargeVec = thisPulse->getChargeValues();

    // determine the polarity of the pulse from the electronics parameters
    const int polarity = getPolarity(thisPulse->getCellID1());

    // calculate the noise
    float noise_value = _noiseValuePedestalWidthFactor * getPedestalWidth(thisPulse->getCellID0(),
									  thisPulse->getCellID1());

    /*
     * If the pedestal width is not defined and
     */
    // if pedestal width is not defined its default value is -1
    // then set pedestal width equal to 1
    if (noise_value <= 0.)
      noise_value = _noiseValuePedestalWidthFactor;

    float time      = 0.0;
    float timeError = 0.0;
    float derivative     = 0.0;
    float norm           = 0.0;
    float sliceTime      = thisPulse->getTime();

    // determine the mean value of the positive derivative == the best estimate for the pulse time
    if (chargeVec.size() > 1)
      {
        for (vector<float>::const_iterator timeBin = chargeVec.begin() + 1; timeBin < chargeVec.end(); ++timeBin)
          {
            derivative  = (*timeBin) - (*(timeBin - 1));

            // make derivative sign insensitive
            derivative *= polarity;

            sliceTime += 1.;

            // break out of calculation if decrease is larger than the noise
            if (derivative < (-noise_value))
              break;

            norm      += derivative;
            time += derivative * sliceTime;
          }
        if (norm > 0.)
          time /= norm;
        else // no iteration means a single slot that counts!
          time = sliceTime;
      }
    else
      time = sliceTime;

    // now determine the pulse time error -- which needs a second loop
    const float timeSum = (time - thisPulse->getTime()) * norm;

    if (chargeVec.size() > 1)
      {
        for (vector<float>::const_iterator timeBin = chargeVec.begin() + 1; timeBin < chargeVec.end(); ++timeBin)
          {
            derivative  = (*timeBin) - (*(timeBin - 1));
            unsigned int index = timeBin - chargeVec.begin();

            // break out of calculation if decrease is larger than the noise
            if (derivative < (-noise_value))
              break;

            double derivativeVariance = 2*noise_value*noise_value; //sigma q = noise value
            timeError += (index * norm - timeSum) * (index * norm - timeSum) * derivativeVariance;
          }
        timeError = sqrt(timeError) / (norm * norm);
      }
    else
      timeError = -1.;

    const float bin_width = 1.E03 / getReadoutFrequency(thisPulse->getCellID1());
    time = time * bin_width; //conversion in ns

    PulseTime returnValue;
    returnValue.setTime(time);
    returnValue.setTimeError(timeError);

    return  returnValue;
  }



  PulseFinder::PulseCharge PulseFinder::calcPulseChargeSumQ(const TrackerData* thisPulse)
  {

    const int polarity = getPolarity(thisPulse->getCellID1());
    const FloatVec charge_values = thisPulse->getChargeValues();
    float pedestalWidth = getPedestalWidth(thisPulse->getCellID0(),
					   thisPulse->getCellID1());
    float         charge     = 0.0;
    // measurement error on the charge is the pedestalwidth! Statistical error on the charge according to poisson only makes sense 
    // if charge is given in units  of primary electrons. This conversion will happen on hit level.
    float         chargeErrorMeas  = 0.0;
 

    for (vector<float>::const_iterator adcVal = charge_values.begin(); adcVal < charge_values.end(); ++adcVal)
      {
        charge += (*adcVal);
        chargeErrorMeas +=pow(pedestalWidth,2);
      }

    // flip sign if polarity is negative
    charge *= polarity;
    PulseCharge returnValue;
    returnValue.setCharge(charge);
    returnValue.setChargeError(sqrt(chargeErrorMeas));
    return returnValue;
  }

  PulseFinder::PulseCharge PulseFinder::calcPulseChargeMaxQ(const TrackerData* thisPulse)
  {

    int readoutGroup = thisPulse->getCellID1();
    const FloatVec charge_values = thisPulse->getChargeValues();
    
    bool negativePolarity = ( getPolarity(readoutGroup) == -1 );

    // extremeQ an be max or min, depending on the polarity
    // Start with MaxADCValue for negative pulses, with 0.0 for positive pulses.
    float extremeQ = (negativePolarity ?  getMaxADCValue(readoutGroup) : 0.0) ; 

    for (vector<float>::const_iterator adcVal = charge_values.begin(); 
	 adcVal < charge_values.end(); ++adcVal)
    {
      extremeQ = (negativePolarity ? std::min(*adcVal,extremeQ) : std::max(*adcVal, extremeQ) );
    }

    //Perform pedestal correction if needed.
    //Especially for negative values the result would be the smallest measured ADC value otherwise
    if (!isPedestalSubtracted(readoutGroup))
    {
      extremeQ -= getPedestalValue(thisPulse->getCellID0(), readoutGroup);
    }

    PulseCharge returnValue;
    returnValue.setCharge( getPolarity(readoutGroup) *extremeQ );
    // The measurement error on a single ADC value obviously is the pedestal width.
    returnValue.setChargeError(getPedestalWidth(thisPulse->getCellID0(), readoutGroup ));

    return returnValue;
  }

  PulseFinder::PulseTime PulseFinder::calcPulseTimeAtMaximum(const TrackerData* thisPulse)
  {

    int readoutGroup = thisPulse->getCellID1();
    const FloatVec charge_values = thisPulse->getChargeValues();
    
    bool negativePolarity = ( getPolarity(readoutGroup) == -1 );

    //find position of max or min, depending on the polarity in charge values

    vector<float>::const_iterator maxbin=(negativePolarity ? std::min_element(charge_values.begin(),charge_values.end()) : std::max_element(charge_values.begin(),charge_values.end() ));
    float time=thisPulse->getTime()+distance(charge_values.begin(),maxbin);
    const float bin_width = 1.E03 / getReadoutFrequency(thisPulse->getCellID1());
    time = time * bin_width; //conversion in ns
    PulseTime returnValue;
    returnValue.setTime(time);
    //accuracy limited to time bin width
    returnValue.setTimeError(bin_width/sqrt(12.));
    return returnValue;
  }

 PulseFinder::PulseTime PulseFinder::calcPulseTimeCOGAtMaximum(const TrackerData* thisPulse)//, int number)
  {

    	int readoutGroup = thisPulse->getCellID1();
    	const FloatVec charge_values = thisPulse->getChargeValues();
    
    	bool negativePolarity = ( getPolarity(readoutGroup) == -1 );

    	//find position of max or min, depending on the polarity in charge values

    	vector<float>::const_iterator maxbin=(negativePolarity ? std::min_element(charge_values.begin(),charge_values.end()) : std::max_element(charge_values.begin(),charge_values.end() ));


	int number=_COGAtMaxParameter;
	TrackerDataImpl truncPulse;
	truncPulse.setTime(thisPulse->getTime());
	truncPulse.setCellID0(thisPulse->getCellID0());
	truncPulse.setCellID1(thisPulse->getCellID1());
	FloatVec truncspec;
	float timeOffset=0.;
	for (vector<float>::const_iterator itj=charge_values.begin();itj<charge_values.end();++itj)
	{
		if (abs(distance(itj,maxbin))<=number)
		{
			truncspec.push_back(*itj);
			if(truncspec.size()==1)
			{
				timeOffset=(float) distance(charge_values.begin(),itj);
			}
			
		}
	}
	const FloatVec& setspec = truncspec;
	truncPulse.setChargeValues(setspec);
	//const TrackerDataImpl passpulse= truncPulse;
	PulseTime res=calcPulseTimeCentreOfGravity(&truncPulse);
	
    	float time=timeOffset;
	
    	const float bin_width = 1.E03 / getReadoutFrequency(thisPulse->getCellID1());
    	time = time * bin_width; //conversion in ns
	time +=res.getTime();
    	PulseTime returnValue;
    	returnValue.setTime(time);
    	
    	returnValue.setTimeError(res.getTimeError());
    	return returnValue;
  }

  PulseFinder::PulseTime PulseFinder::calcPulseTimeAtConstantFraction(const TrackerData* thisPulse)
    {

	  int readoutGroup = thisPulse->getCellID1();
	  const FloatVec charge_values = thisPulse->getChargeValues();

	  //find position of max or min, depending on the polarity in charge values
	  bool negativePolarity = ( getPolarity(readoutGroup) == -1 );
	  double maxCharge=(negativePolarity ? *(std::min_element(charge_values.begin(),charge_values.end())) : *(std::max_element(charge_values.begin(),charge_values.end() )));
	  vector<float>::const_iterator current_value = charge_values.begin();
	  vector<float>::const_iterator timebin = charge_values.begin();
	  double fraction=_ConstFracParameter;
	  while (current_value < charge_values.end())
	  {
		  const bool above_start_threshold = negativePolarity ? (*current_value <= fraction*maxCharge) : (*current_value >= fraction*maxCharge);
		  if (above_start_threshold)
		  {
			  timebin=current_value;
			  break;
		  }
		  current_value++;
	  }
	  float time=thisPulse->getTime()+distance(charge_values.begin(),timebin)-( (*timebin-fraction*maxCharge) / (*timebin-(*(timebin-1))) );
	  const float bin_width = 1.E03 / getReadoutFrequency(thisPulse->getCellID1());
	  time = time * bin_width; //conversion in ns
	  PulseTime returnValue;
	  returnValue.setTime(time);
	  //accuracy limited to time bin width
	  returnValue.setTimeError(bin_width/sqrt(12.));
	  return returnValue;
    }

  PulseFinder::PulseTime PulseFinder::calcPulseTimeCentreOfGravity(const TrackerData* thisPulse)
  {
    const FloatVec charge_values = thisPulse->getChargeValues();
    float pedestalWidth = getPedestalWidth(thisPulse->getCellID0(),
					   thisPulse->getCellID1());
    vector<float>::const_iterator current_value = charge_values.begin();

    float sumQ=0;
    float sumtimeQ=0;
    int bin=0;
    float err_q=0,err_t=0;
    while (current_value < charge_values.end())
      {
	sumQ+=*current_value;
	sumtimeQ+=(*current_value)*bin;
	current_value++;
	bin++;
      }
    //second go for error calculation
    current_value = charge_values.begin();
    bin=0;//was missing, but should be here
    while (current_value < charge_values.end())
      {
	//error of charge is the pdestalwidth
	err_q+=pow((bin*sumQ-sumtimeQ)/pow(sumQ,2)*pedestalWidth,2);
	//error on time is timebinwidth=1 divided by sqrt(12)
	err_t+=pow((*current_value/sumQ*1./sqrt(12.)),2);
	current_value++;
	bin++;
      }
 

    float time=thisPulse->getTime()+sumtimeQ/sumQ;
    const float bin_width = 1.E03 / getReadoutFrequency(thisPulse->getCellID1());
    time = time * bin_width; //conversion in ns
    PulseTime returnValue;
    returnValue.setTime(time);
    float time_error=sqrt(err_q+err_t);

    returnValue.setTimeError(time_error*bin_width);
    return returnValue;
   }

  PulseFinder::PulseTime PulseFinder::calcPulseTimeFirstBinAboveThreshold(const TrackerData* thisPulse)
  {
    float pedestalWidth = getPedestalWidth(thisPulse->getCellID0(),
					   thisPulse->getCellID1());
    float startThreshold =  pedestalWidth * _pulseStartPedestalWidthFactor;
    
    int polarity = getPolarity( thisPulse->getCellID1() );

    const FloatVec charge_values = thisPulse->getChargeValues();

    // determine if the pulse is negative
    const bool negativePolarity = (polarity == -1);
    vector<float>::const_iterator current_value = charge_values.begin();
    vector<float>::const_iterator timebin = charge_values.begin();
    while (current_value < charge_values.end())
      {
	const bool above_start_threshold = negativePolarity ? (*current_value <= startThreshold) : (*current_value >= startThreshold);
	if (above_start_threshold)
	  {
	    timebin=current_value;
	    break;
	  }
	current_value++;
      }
    float time=thisPulse->getTime()+distance(charge_values.begin(),timebin);
    const float bin_width = 1.E03 / getReadoutFrequency(thisPulse->getCellID1());
    time = time * bin_width; //conversion in ns
    PulseTime returnValue;
    returnValue.setTime(time);
    //accuracy limited to time bin width
    returnValue.setTimeError(bin_width/sqrt(12.));
    return returnValue;
   }

  PulseFinder::PulseTime PulseFinder::calcPulseTimeBoxMethod(const EVENT::TrackerData * thisPulse)
    {
      /*

  		   Computes two areas, the bin that make the areas closest to equal is the time.
         areaDown is under the curve before the cut bin.
         areaUp is above the curve and below the max value after the cut bin.

       */

      //streamlog_out(MESSAGE) << "Using box method\n";

      //Get the spectrum
      std::vector<float> chargeVec(thisPulse->getChargeValues());

      int maxBin = 0;
      float maxVal = -9E9;

      //Find maximum amplitude bin
  		for(unsigned int i = 0; i < chargeVec.size(); ++i)
      {
  			if(chargeVec[i] > maxVal)
  			{
          maxVal = chargeVec[i];
          maxBin = i;
        }
      }

      //Check if saturated, if so take center of saturation region as peak

      int numberOfPeakBins = 1; //If not saturated should stay at 1
      float saturationDerivative = 2.0; //If there are bins directly after the max bin with value within this region around the maximum then they are included
      // if data has already been pedestal subtracted, take this into account for overflow value maxBinValue ( = maximumADCValue)
      double maxBinValue=getMaxADCValue(thisPulse->getCellID1());
      if (isPedestalSubtracted(thisPulse->getCellID1()))
      {
    	  maxBinValue -= floor(getPedestalValue(thisPulse->getCellID0(), thisPulse->getCellID1()));
      }

      if (maxVal >= maxBinValue) // overflow in at least one time bin
      {
  		for(unsigned int i = maxBin+1; i < chargeVec.size(); ++i)
        {
          if(fabs(chargeVec[maxBin] - chargeVec[i]) < saturationDerivative)
          {
  					++numberOfPeakBins;
            maxVal+= chargeVec[i];
          }
  				else
            break;
        }
        maxVal /= numberOfPeakBins;
        maxBin += (numberOfPeakBins - 1)/2; //i.e. if we find 1 extra peak bin then it will add 1/2 to the peak bin value
      }

  	  streamlog_out(DEBUG) << "Max Amplitude Bin: " << maxBin << "\n";

      int minBin = 0;

      while(minBin && (chargeVec[minBin-1] > 30))
        --minBin;

  		const float binWidth =1.E03 /getReadoutFrequency(thisPulse->getCellID1());
      double pulseTime = 0.;
      float sliceTime = thisPulse->getTime();

      double background = chargeVec[minBin ? minBin - 1 : 0]; //Assume for now

      double area = 0;
      double area2 = 0;
      double areaErrSum = 0;

  		for(int i = 0; i <= maxBin; ++i)
  		{
        area += (chargeVec[i] - background);
        area2 += (chargeVec[i] - background)*(chargeVec[i] - background);
        areaErrSum += chargeVec[i];
      }

      //float areaErr = sqrt(areaErrSum);

      double deltaT = area/(maxVal-background);

  		//float deltaTErr = deltaT * sqrt(pow(areaErr/area, 2) + (maxVal + background)/pow(maxVal - background, 2));

      streamlog_out(DEBUG) << "Area " << area << " Max pulse Charge " << maxVal << " Delta Time " << deltaT << "\n";

      pulseTime = (sliceTime + maxBin - deltaT)*binWidth;

      //float pulseTimeErr = sqrt(maxBin + deltaT) * binWidth; //Assuming no significant error on binWidth or sliceTime
      double pulseTimeErr = sqrt(areaErrSum + area2/maxVal)/maxVal * binWidth; //Assuming no significant error on binWidth or sliceTime

      streamlog_out(DEBUG) << "sliceTime " << sliceTime << " binWidth " << binWidth << " pulsetime " << pulseTime << "\n";

      PulseTime returnValue;
      returnValue.setTime(pulseTime);
      //accuracy limited to time bin width
      returnValue.setTimeError(pulseTimeErr);
      return returnValue;

    }// _calculateBoxMethod


  PulseFinder::PulseTime PulseFinder::calcuPulseTimeGaussianInflexion(const EVENT::TrackerData * thisPulse)
    {

      //Return value
      std::pair<double, double> timeAndError;

      //Get the spectrum
      std::vector<float> chargeVec(thisPulse->getChargeValues());

      int maxBin = 0;
      float maxVal = -9E9;

      //Find maximum amplitude bin
  		for(unsigned int i = 0; i < chargeVec.size(); ++i)
      {
  			if(chargeVec[i] > maxVal)
  			{
          maxVal = chargeVec[i];
          maxBin = i;
        }
      }

      //Define a range where it's mostly gaussian
      int lowerEdge = maxBin - 5;
      int upperEdge = maxBin + 5;

      //Use this later for a more refined approach
  		//	int minBin = 0;
      //
      //	while(minBin && (chargeVec[minBin-1] > 30))
      //	    --minBin;

		const float binWidth =1.E03 /getReadoutFrequency(thisPulse->getCellID1());
      double sliceTime = thisPulse->getTime();

      //Fit values
      double tVal[11];
  		double ampVal[11];

  		for(int i = lowerEdge; i <= upperEdge; ++i )
      {
        tVal[i-lowerEdge] = (sliceTime + (double)i)*binWidth;
        ampVal[i-lowerEdge] = (double)chargeVec[i];
      }

      gROOT->SetBatch(1);

      //TGraph to start, maybe a TGraphErrors later
      TGraph * gr = new TGraph(11, tVal, ampVal);
      int r = gr->Fit("gaus", "Q");

      if(r != 0)
      {
    	PulseTime returnValue;
    	     returnValue.setTime(-9E9);
    	      //accuracy limited to time bin width
    	      returnValue.setTimeError(-9E9);
        return returnValue;
      }

      double mu = gr->GetFunction("gaus")->GetParameter(1);
      double sigma = gr->GetFunction("gaus")->GetParameter(2);

      double muE = gr->GetFunction("gaus")->GetParError(1);
      double sigmaE = gr->GetFunction("gaus")->GetParError(2);


      PulseTime returnValue;
           returnValue.setTime(mu - sigma);
           //accuracy limited to time bin width
           returnValue.setTimeError(sqrt(sigmaE*sigmaE + muE*muE));
  		delete gr;
        return returnValue;

    }//End _calculateGaussianInflexion();

 

  /* Returns the pedestal value for a given TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
  */
  float PulseFinder::getPedestalValue(const int hardware_channel, const int readout_group)
  {
    if (_usePedestalValueOverride)
    {
        streamlog_out(DEBUG) << "Pedestal Value for channel: " << readout_group << ":" << hardware_channel
                << " overridden to " << _pedestalValueOverride << std::endl;
	return _pedestalValueOverride;
    }
    else
    {
      if (_pedestalListener != NULL)
      {
	streamlog_out(DEBUG) << "Pedestal Value for channel: " << readout_group 
			     << ":" << hardware_channel
			     << " = " 
			     << _pedestalListener->getPedestalValue(hardware_channel, readout_group)
			     << std::endl;
	return _pedestalListener->getPedestalValue(hardware_channel, readout_group);
      }
      else
      {
	throw tpcconddata::ConditionsObjectException(string("[PulseFinderProcessor]::getPedestalValue-Exception: The Pedestal value is not defined or can not be found!"));
	return 0.;
      }
    }
  }


  /* Returns the pedestal width for a given TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
   */
  float PulseFinder::getPedestalWidth(const int hardware_channel, const int readout_group)
  {
    if (_usePedestalWidthOverride)
      {
        streamlog_out(DEBUG) << "Pedestal Width for channel: " << readout_group << ":" << hardware_channel
                << " overridden to " << _pedestalWidthOverride << std::endl;
        return _pedestalWidthOverride;
      }
    else
      {
        if (_pedestalListener != NULL)
        {
            streamlog_out(DEBUG) << "Pedestal Width for channel: " << readout_group << ":" << hardware_channel
                    << " = " << _pedestalListener->getPedestalWidth(hardware_channel, readout_group) << std::endl;
            return _pedestalListener->getPedestalWidth(hardware_channel, readout_group);
        }
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinder]::getPedestalWidth-Exception: The Pedestal width is not defined or can not be found!"));
            return 0.;
          }
      }
  }



  /* Returns if the data has been pedestal subtracted in the read-out electronics or not
   */
  bool PulseFinder::isPedestalSubtracted(const int readout_group)
  {
    if (_use_isPedestalSubtractedOverride)
      {
        return _isPedestalSubtractedOverride;
      }
    else
      {
        if (_adcElectronicsListener != NULL)
          return _adcElectronicsListener->isPedestalSubtracted(readout_group);
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinder]::isPedestalSubtracted-Exception: If data is pedestal subtracted or not is either not defined or can not be found!"));
            return 0;
          }
      }
  }


  /* Returns the polarity of a TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
   */
  int PulseFinder::getPolarity(const int readout_group)
  {
    if (_usePolarityOverride)
      {
        return _polarityOverride;
      }
    else
      {
        if (_adcElectronicsListener != NULL)
          return _adcElectronicsListener->getPolarity(readout_group);
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinder]::getPolarity-Exception: The polarity is either not defined or can not be found!"));
            return 0;
          }
      }
  }



  /* Returns the readout frequency of a readout group (in Mhz) taking into consideration whether it should be overwritten by a processor parameter
   */
  float PulseFinder::getReadoutFrequency(const int readout_group)
  {
    if (_useReadoutFrequencyOverride)
      {

        return _readoutFrequencyOverride;
      }
    else
      {
        if (_adcElectronicsListener != NULL)
            //In the DB the frequency is supposed to be stired in MHz
          return _adcElectronicsListener->getFrequency(readout_group);
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinder]::getReadoutFrequency-Exception: The Readout Frequency is not defined or can not be found!"));
            return 0.;
          }
      }
  }



  /* Returns the maximum ADC value of a readout group taking into consideration whether it should be overwritten by a processor parameter
   */
  float PulseFinder::getMaxADCValue(const int readout_group)
  {
    if (_useMaxADCValueOverride)
      {
        return _maximumADCValueOverride;
      }
    else
      {
        if (_adcElectronicsListener != NULL)
          return (float)(_adcElectronicsListener->getMaxADCValue(readout_group));
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinder]::getMaxADCValue-Exception: The maximum ADC Value is not defined or can not be found!"));
            return 0.;
          }
      }
  }

  void PulseFinder::setForceSpectrumSave(bool doSave)
  {
    _forceSpectrumSave = doSave;
  }

  void PulseFinder::setSwitchOffSplitting(bool doNotSplit )
  {
    _switchOffSplitting = doNotSplit;
  }
}//namespace
