#include "PulseFinderProcessor.h"

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// Marlin
#include <marlin/ConditionsProcessor.h>
#include <marlin/Exceptions.h>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

//Streamlog
#include <streamlog/streamlog.h>

// MarlinTPC
#include "PedestalListener.h"
#include "ADCElectronicsListener.h"
#include "PulseFinder.h"

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace tpcconddata;

namespace marlintpc
{
  PulseFinderProcessor aPulseFinderProcessor;

  PulseFinderProcessor::PulseFinderProcessor() 
    : Processor("PulseFinderProcessor"), _pulseFinder(NULL)
  {

    _description = "PulseFinderProcessor: finds pulses out of the raw data and calculate the time and charge (in FADC) information";

    registerInputCollection(LCIO::TRACKERDATA,
                            "InputCollectionName",
                            "Name of the input TrackerData collection (default: TPCData)",
                            _inputColName,
                            string("TPCData"));

    registerOutputCollection(LCIO::TRACKERDATA,
                             "OutputTrackerDataCollectionName",
                             string("Name of the output TrackerData collection. ") +
                             string("Only the data related to a pulse will be stored")+
                             string("(default: TPCDataInPulses)"),
                             _outputDataColName,
                             string("TPCDataInPulses"));

    registerOutputCollection(LCIO::TRACKERPULSE,
                             "OutputTrackerPulseCollectionName",
                             "Name of the output TrackerPulse collection (default: TPCPulses)",
                             _outputPulseColName,
                             string("TPCPulses"));

    registerProcessorParameter("WriteOutputToStorage",
                               "if true the output collection is written (default: true)",
                               _outputIsPersistent,
                               true);

    registerProcessorParameter("ForceSpectrumSave",
                               "if true the ADC spectrum will be stored for all pulses (default true)",
                               _forceSpectrumSave,
                               true);

    registerProcessorParameter("SwitchOffSplitting",
                               "if true pulses will not be splitted if a double pulse is detected (default false)",
                               _switchOffSplitting,
                               false);

    registerProcessorParameter("PulseStartPedestalWidthFactor",
                               string("A pulse starts if the ADC value is this factor times the ")+
			       string(" width of the pedestal (default: 4.0)"),
                               _pulseStartPedestalWidthFactor,
                               float(4.0));
    
    registerProcessorParameter("PulseEndPedestalWidthFactor",
                               string("A pulse ends if the ADC value drops below this factor")+
			       string(" times the pedestal width (default: 3.0)"),
                               _pulseEndPedestalWidthFactor,
                               float(2.0));

    registerProcessorParameter("MinimumPulseHeight",
                               string("To accept a pulse its maximum must exceed this cut ")+
			       string("in adc counts (default: 12)"),
                               _minimumPulseHeight,
                               float(12.0));

    registerProcessorParameter("MinimumPulseLength",
                               string("Declares the minimal length of a pulse in time bins above the thresholds (default: 5)"),
                               _minimumPulseLength,
                               5);

    registerProcessorParameter("SaveNBinsBeforeStart",
                               "# bins saved before start threshold (except there is another pulse, default 2)",
                               _saveNBinsBeforeStart,
                               2);

    registerProcessorParameter("SaveNBinsAfterEnd",
                               "# bins saved after the end threshold (except there is another pulse, default 8)",
                               _saveNBinsAfterEnd,
                               8);

    registerProcessorParameter("NoiseValuePedestalWidthFactor",
                               "The noise (cut) value is calculated as this factor times pedesta width (default: 3.0)",
                               _noiseValuePedestalWidthFactor,
                               float(3.0));

    registerProcessorParameter("ElectronicsParametersObjectName" ,
                               "Name of the conditions object in which the parameters of the read out electronics is stored",
                               _electrParaObjName,
                               string("GenericADCElectronicsParameters")
                              );

    registerProcessorParameter("ElectronicsParametersCollectionName" ,
                               "Name of the conditions collection in which the parameters of the read out electronics is stored",
                               _electrParaColName,
                               tpcconddata::GenericADCElectronicsParameters::getDefaultColName()
                              );

    registerProcessorParameter("PedestalCollectionName" ,
                               "Name of the conditions collection in which the pedestals are stored",
                               _pedestalColName,
                               tpcconddata::Pedestal::getDefaultColName()
                              );

    registerProcessorParameter("ChargeCalculationMode" ,
			       "SUMQ = 0 (default), FIT_PARABOLA = 1, MAXQ = 2",
			       _chargeCalculationMode,
			       int(0));

    registerProcessorParameter("TimeCalculationMode",
			       "INFLEXION_POINT = 0 (default), FIT_PARABOLA = 1,  CENTRE_OF_GRAVITY = 2, FIRST_ABOVE_THRESHOLD = 3, AT_MAXIMUM = 4, CONSTANT_FRACTION=5, GAMMMA4_FIT=6, COG_ATMAX =7, BOX =8, GAUSSIAN_INFLEXION=9",
                              _timeCalculationMode,
			       int(0));
   
    registerProcessorParameter("TimeCalculationParameter",
			       "fraction for constant fraction (default=0.9), no. of neighboring bins for center of gravity at maxbin, alpha/sigma for convolution",
                              _timeCalculationParameter,
			       double(0.9));

    // now some parameters for overriding the default values obtained by listeners/conditions data
    registerOptionalParameter("MaximumADCValueOverride",
                              "Optional: Maximum number of the adc spectrum; to determine wether a pulse is maxed out. This single value overrides any value, from any readout group, which may have been taken from LCCD.",
                              _maximumADCValueOverride,
                              float(255.));

    registerOptionalParameter("PedestalValueOverride",
                              "Optional: The pedestal value to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
                              _pedestalValueOverride,
                              float(0.));


    registerOptionalParameter("PedestalWidthOverride",
                              "Optional: The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
                              _pedestalWidthOverride,
                              float(1.));

    registerOptionalParameter("IsPedestalSubtractedOverride",
                              "Optional:  Sets, if the data is already pedestal subtracted in the read-out electronics. This overrides any information provided by LCCD.",
                              _isPedestalSubtractedOverride,
                              true);


    registerOptionalParameter("SignalPolarityOverride",
                              "The polarity of the signals (1 for positive, -1 for negative) produced by all readout groups. This overrides any information provided by LCCD.",
                              _polarityOverride,
                              1);

    registerOptionalParameter("ReadoutFrequencyOverride",
                              "Optional:  The readout frequency of every readout group (in MHz). This overrides any information provided by LCCD.",
                              _readoutFrequencyOverride,
                              float(40.));
    // end of the override variables

    registerOptionalParameter("PlateauLengthCutOff",
                              "Optional:  The maximum length of a plateau within a pulse in time bins. This used to cut off strange pulses that go into a plateau.",
                              _maxPlateauLength,
                              int(5));
    // end of the override variables

  }



  void PulseFinderProcessor::init()
  {

    // usually a good idea to
    printParameters();
    
    // Some parameter names have changed. Check if the old ones are used and print an error message if this is the case.
    // We refuse to run because these setting would be ignored the silently the default values of the new parameters
    // would be used, which will lead to unwanted results without the user being aware of it.

    // The _parameters variable contains all the parameters from the steering file.
    if (_parameters->isParameterSet("IsPedestalSubstractedOverride") || // has typo "substracted" instead of "subtracted"
	_parameters->isParameterSet("PedestalWidthScale") ||
	_parameters->isParameterSet("PulseEndThreshold") ||
	_parameters->isParameterSet("PulseStartThreshold")||
	_parameters->isParameterSet("UseParabolicFitAmplitudeAsCharge?")||
	_parameters->isParameterSet("UseParabolicFitMaxAsTime?")||
	_parameters->isParameterSet("UseParabolicFitAmplitudeAsCharge")||
	_parameters->isParameterSet("UseParabolicFitMaxAsTime") )
    {
      streamlog_out(ERROR) << "You are using old parameter names!" << std::endl << std::endl
			   << "The folowing parameters have been renamed: " << std::endl
			   << "'IsPedestalSubstractedOverride' to 'IsPedestalSubtractedOverride' (subtracted without s after the b) " << std::endl
			   << "'PedestalWidthScale' to 'NoiseValuePedestalWidthFactor'"<< std::endl
			   << "'PulseEndThreshold' to 'PulseEndPedestalWidthFactor'"<< std::endl
			   << "'PulseStartThreshold' to 'PulseStartPedestalWidthFactor'"<< std::endl
			   << std::endl
			   << "The folowing parameters have been replaced: " << std::endl
			   << "'UseParabolicFitAmplitudeAsCharge?' (bool)  or "
			   << "'UseParabolicFitAmplitudeAsCharge' (bool) with "
			   << "'ChargeCalculationMode' (int)"<< std::endl
			   << "'UseParabolicFitMaxAsTime?' (bool)  or "
			   << "'UseParabolicFitMaxAsTime' (bool) with "
			   << "'TimeCalculationMode' (int)"<< std::endl
			   << "You can use tools/scripts/updatePulseFinderParameters.sh to update your steering files." << std::endl << std::endl;

      throw StopProcessingException(this);
    }
       
    // The override parameter for the readout frequency has changed from Hz to MHz. Stop processing if an old (unreasonably high) frequency is detected.
    // 1e5 should be fine. A readout with more than 100 GHz can be excluded for the foreseeable future. A readout < 100 kHz (10 ��s samples) is not reasonable
    // for TPCs anyway, so we should be save to assume MHz for lower values.
    if (_readoutFrequencyOverride > 1e5) 
    {
      streamlog_out(ERROR) << "Unreasonably high readout frequency detected ("<< _readoutFrequencyOverride <<" MHz)."<<std::endl
			   << "You are probably using the old value in Hz. Plesase adapt this value to MHz." << std::endl;
      throw StopProcessingException(this);
    }

    PedestalListener * pedestalListener = NULL;

    // determine if the pedestal information is supplied by LCCD
    //If both the pedestal width and value are overridden no listener is necessary
    if ( !(parameterSet("PedestalWidthOverride")  && parameterSet("PedestalValueOverride")) )
      {
        pedestalListener = new PedestalListener(_pedestalColName);
      }

    ADCElectronicsListener * adcElectronicsListener = NULL;

    // determine if the polarity, maxADC value and readout frequency are supplied by LCCD
    // no listener needed if all three values overridden
    if (!(parameterSet("ReadoutFrequencyOverride") && 
	  parameterSet("SignalPolarityOverride") && 
	  parameterSet("IsPedestalSubtractedOverride") && 
	  parameterSet("MaximumADCValueOverride")))
      {
        adcElectronicsListener = new ADCElectronicsListener(_electrParaObjName, _electrParaColName);
      }

    // we now have all the information to instantiate the PulseFinder
    _pulseFinder = new PulseFinder( _minimumPulseHeight,
				    _minimumPulseLength,
				    _pulseStartPedestalWidthFactor,
				    _pulseEndPedestalWidthFactor,
				    _noiseValuePedestalWidthFactor,
				    pedestalListener,
				    adcElectronicsListener,
				    _saveNBinsBeforeStart,
				    _saveNBinsAfterEnd,
				    _timeCalculationMode,
				    _timeCalculationParameter,
				    _chargeCalculationMode);
    
    // now set the override parameters, if any
    if (parameterSet("PedestalValueOverride"))
    {
      _pulseFinder->setPedestalValueOverride(_pedestalValueOverride);
    }

    if (parameterSet("PedestalWidthOverride"))
    {
      _pulseFinder->setPedestalWidthOverride(_pedestalWidthOverride);
    }

    if (parameterSet("MaximumADCValueOverride"))
    {
      _pulseFinder->setMaxADCValueOverride(_maximumADCValueOverride);
    }

    if (parameterSet("IsPedestalSubtractedOverride"))
    {
      _pulseFinder->set_isPedestalSubtractedOverride(_isPedestalSubtractedOverride);
    }

    if (parameterSet("SignalPolarityOverride"))
    {
      _pulseFinder->setPolarityOverride(_polarityOverride);
    }

    if (parameterSet("ReadoutFrequencyOverride"))
    {
      _pulseFinder->setReadoutFrequencyOverride(_readoutFrequencyOverride);
    }

    _pulseFinder->setForceSpectrumSave(_forceSpectrumSave);
    _pulseFinder->setSwitchOffSplitting(_switchOffSplitting);
    
    //switch on plateau cut off if requested
    if ( parameterSet("PlateauLengthCutOff") )
      {
	if(_maxPlateauLength<=0)
	  {
	    streamlog_out(ERROR) << "Unreasonably plateau length detected ("<< _maxPlateauLength<<std::endl
				 << "A length should larger than 0." << std::endl;
	    throw StopProcessingException(this);
	  }
      _pulseFinder->setPlateauCutOff(_maxPlateauLength);
      }   

  }



  void PulseFinderProcessor::processRunHeader(lcio::LCRunHeader* run)
  {
    run->parameters().setValue(_processorName + "_revision", "$Rev: 5472 $: PulseFinder "+ PulseFinder::getRevisionNumber());
    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
      {
        if (! i->second->isOptional() || i->second->valueSet())
          {
            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
          }
      }
  }



  void PulseFinderProcessor::processEvent(lcio::LCEvent* evt)
  {
    streamlog_out(DEBUG3) << "Finding the pulses in " << _inputColName << "." << endl;
    // use a CollectionVec so we can iterate
    LCCollectionVec * inputCollection;

    try
    {
      // this cast should be safe. The CollectionVec is the only implementation of
      // the abstract interface class.
      inputCollection = dynamic_cast<LCCollectionVec *>(evt->getCollection(_inputColName));
    }
    catch(DataNotAvailableException &e)
    {
        // if there is not input collection print a message and just continue. It's just an 
	// empty event, might happen...
        streamlog_out(MESSAGE) << "Input collection " << _inputColName 
	  		       << " cannot be found in event " << evt->getEventNumber() << endl;

	return; // just return, nothing special about an empty event.
    }

    // create the output collections for pulses and zero suppressed data (ADC spectra)
    IMPL::LCCollectionVec* outputColPulse = new IMPL::LCCollectionVec(LCIO::TRACKERPULSE);
    IMPL::LCCollectionVec* outputColData  = new IMPL::LCCollectionVec(LCIO::TRACKERDATA);

    // set the persistency flag as defined by the processor parameter
    outputColPulse->setTransient(!_outputIsPersistent);
    outputColData->setTransient(!_outputIsPersistent);

    // set the charge and time calculation mode as collection parameters for convenient access
    outputColPulse->parameters().setValue("ChargeCalculationMode",_chargeCalculationMode);
    outputColPulse->parameters().setValue("TimeCalculationMode",_timeCalculationMode);
    

    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0);
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1);
    trkFlag.setBit(lcio::LCIO::TRAWBIT_CM);
    outputColPulse->setFlag(trkFlag.getFlag());
    outputColData->setFlag(trkFlag.getFlag());

    // loop the input data 
    for (LCCollectionVec::const_iterator inputIter = inputCollection->begin();
	 inputIter != inputCollection->end(); ++inputIter)
    {
      TrackerData* inputTrackerData = dynamic_cast<TrackerData*>(*inputIter);

      //FIXME? If cast is not OK notify somehow or throw an exception
      if (inputTrackerData == NULL)    // cast failed. Seems to be something else in the collection
      {
	streamlog_out(ERROR) << "Collection " << _inputColName << " does not contain "
		     << "TrackerData objects" << std::endl;
	throw DataNotAvailableException("Wrong data type in the input collection");
      }

      streamlog_out(DEBUG2) << "Looking at Pad:" << inputTrackerData->getCellID0()
		   << " module:" << inputTrackerData->getCellID1() << endl;

      // ok, the PulseFinder does all the work for us
      std::vector<TrackerPulse *> pulses = _pulseFinder->findPulses(inputTrackerData);

      // all we have to do is store the pulses and corresponding tracker data into the collections
      for (vector<TrackerPulse *>::const_iterator pulseIter = pulses.begin();
	   pulseIter != pulses.end(); ++pulseIter)
      {
            outputColPulse->addElement(*pulseIter);
	    
	    if ((*pulseIter)->getTrackerData() != NULL)
	    {
	      outputColData->addElement((*pulseIter)->getTrackerData());
	    }
      }

    }// for inputIter

    // only store non-empty collections in the event
    if (!outputColPulse->empty())
    {
      evt->addCollection(outputColPulse, _outputPulseColName);
    }
    if (!outputColData->empty())
    {
      evt->addCollection(outputColData, _outputDataColName);
    }


      streamlog_out(DEBUG4) << "Found " << outputColPulse->getNumberOfElements() << " pulses." << endl;
  }



  void PulseFinderProcessor::check(lcio::LCEvent*  /*evt*/)
  {
  }



  void PulseFinderProcessor::end()
  {
    delete _pulseFinder;
  }



}// namespace marlintpc

