#ifndef PHOTOELECTRICXMLHELPER_H
#define PHOTOELECTRICXMLHELPER_H 1

// CLHEP
#include <CLHEP/Vector/TwoVector.h>

// GEAR
#include <gearxml/tinyxml.h>
#include <gear/TPCParameters.h>


// stl
#include <vector>
#include <map>
#include <set>

namespace marlintpc{
/**
 * This provides some tools on handling the xml file used to store information on the Photoelectric analysis.
 * \author K. Zenker (DESY)
 * \class PhotoelectricXMLFile
 * Up to now it allows to:\n
 * 1. Read photodot/line positions from a xml file and sort them by the distance to the center.\n
 * This is useful if clusters should be connected to the true photodot position, since in the
 * center of the magnet distortions are at minimum. Therefore one can start in the center.\n
 * 2. Add information about clusters connected to photodots\photoline.\n
 * 3. Add information about clusters not connected to photodots.\n
 * 4. Add start and end time to be used by processors to do a time cut on pulses.\n
  * \remark Keep in mind that most of the function return only pointers or iterators to internal
 * stored information. Thus you should not delete the PhotoelectricXMLFile object as long as you
 * want to use these information!
 */
struct Photodot {
  CLHEP::Hep2Vector _position;
  double _diameter;
  int _module;
  int _id;
  gear::TiXmlElement* _element;
  /**
   * Used to find a certain photodot by id.
   */
  bool operator== (Photodot a){return a._id == _id;}
};

struct Photoline {
  CLHEP::Hep2Vector _start;
  CLHEP::Hep2Vector _end;
  double _width;
  int _module;
  gear::TiXmlElement* _element;
  bool operator== (Photoline a){return a._module == _module;}
};

class PhotoelectricXMLFile {
private:
  gear::TiXmlDocument _photodotFile; ///< The xml file which is a copy of the xml file given in the constructor.
  /**
   * Main information about the photodots:
   * - TiXmlElement is the \c<photodot> element in the xml file
   * - Hep2Vector is the position of the photodot
   * - diameter
   * - id  The number in the xml file
   */
  std::vector<Photodot> _photodots;

  /**
   * Main information about the photolines:
   * - Hep2Vector start and end
   * - width of the photoline
   */
  std::vector<Photoline> _photolines;
  /**
   * This method is used to sort the photodots by the distance (first parameter).
   */
  static bool Comparison(std::pair<double, std::vector<Photodot>::const_iterator> i, std::pair<double, std::vector<Photodot>::const_iterator> j);
  /**
   * Add information about a cluster to an element.
   */
  void AddPhotodotProjection(gear::TiXmlElement* element, const CLHEP::Hep2Vector centroid, const std::set<gear::GlobalPadIndex> &pads);
  /**
   * Get photodot positions from the xml file and store them in the vector \c _photodots.
   * The index for each photodot is taken from the position in the xml file.
   */
  bool FillPhotodotInformation();

  /**
   * Get photoline start and end points from the xml file and store them in the vector \c _photolines.
   */
  bool FillPhotolineInformation();

public:
  /**
   * Constructor expecting a xml file with the true photodot positions stored as follows:
   * \code
   * <photoelectric_geometry>
   *   < photodot>
   *     <centroid>1256.77,-171.959</centroid>
   *   </photodot>
   * </photoelectric_geometry>
   * \endcode
   * \param xmlFile Photodot file name including the path
   */
  PhotoelectricXMLFile(std::string xmlFile);
  ~PhotoelectricXMLFile(){};
  /**
   * This methods sorts the photodots from the photodot file.
   * The distance to the given point (center_x,center_y) is calculated and than all dots
   * are sorted by this distance. Shortest distance is at the beginning. The corresponding
   * dot should be used to assign the clusters to the true position. This assumes that the
   * distortions of the magnetic field are minimal in the center of the magnet/the given point.
   * \param center_x x coordinate of the point with the most homogeneous magnetic field
   * \param center_y y coordinate of the point with the most homogeneous magnetic field
   * \return Returns sorted iterators to the photodot map. \n
   * The map key is the id of each photodot.
   * The pair consists of the xml element of this photodot and its position.
   */
  std::vector<std::pair<double, std::vector<Photodot>::const_iterator> > GetSortedPhotodots(double center_x, double center_y);

  /**
   * Photodots are sorted by the index found in the xml file.
   * \return A vector of iterators to the internal photodots.
   */
  std::vector<std::vector<Photodot>::const_iterator> GetPhotodots();

  /**
   * Photodots are sorted by the index found in the xml file.
   * \return A vector of iterators to the internal photolines.
   */
  std::vector<std::vector<Photoline>::const_iterator> GetPhotolines();

  /**
   * This adds the time information to the xml file. This can be later on be used by processors to apply a
   * time cut on pulses to reduce noise pulses. Will look like this:
   * \code
   * <photoelectric_geometry>
   *  <arrival_time start="1.400000000e+04" end="1.425000000e+04" />
   * </photoelectric_geometry>
   * \endcode
   * \param start The start time.
   * \param end The end time.
   */
  void SetTimeInfo(float start, float end);
  /**
   * This reads the time information from the xml file.
   * Expected xml format:
   * \code
   * <photoelectric_geometry>
   *  <arrival_time start="1.400000000e+04" end="1.425000000e+04" />
   * </photoelectric_geometry>
   * \endcode
   * \param start The start time.
   * \param end The end time.
   * \return true in case of success
   */
  bool GetTimeInfo(float &start, float &end);
  /**
   * Save the internal xml file. It consists of the information from the original
   * photodot file and possibly additional information (pulse time cut, cluster connected to each dot,...).
   * \param fileName The file name of the xml file including the path.
   */
  void SaveFile(std::string fileName){_photodotFile.SaveFile(fileName);};
  /**
   * This return the main element in the xml file.
   * It can be used to add additional information to the main (photoelectric_geometry) element.
   * \code
   * <photoelectric_geometry>
   *   <!-- add stuff here -->
   * </photoelectric_geometry>
   * \endcode
   */
  gear::TiXmlElement* GetPhotoElectricElement();
  /**
   * Add information about the associated pads to a photodot.
   * \code
   * <photoelectric_geometry>
   *   <photodot>
   *     <centroid x="192.1" y="12" />
   *     <projection>
   *      <pad pad_index="1769614" module_id="5" />
   *     </projection>
   *   </photodot>
   * </photoelectric_geometry>
   * \endcode
   */
  void AddPhotodotProjection(int photodot_id, const CLHEP::Hep2Vector centroid, const std::set<gear::GlobalPadIndex> &pads);
  /**
   * Add information about pads in a cluster not connected to a photodot.
   * \code
   * <photoelectric_geometry>
   *   <centroid x="192.1" y="12" />
   *   <projection>
   *    <pad pad_index="1769614" module_id="5" />
   *   </projection>
   * </photoelectric_geometry>
   * \endcode
   */
  void AddPhotodotProjection(const CLHEP::Hep2Vector centroid, const std::set<gear::GlobalPadIndex> &pads);
  /**
   * Get pads associated to each photodot.
   * If no pads are associated the set size is 0.
   * \return Map holding for each photodot a set of pads.
   */
  std::map<int, std::set<gear::GlobalPadIndex> > GetPhotodotPads();
  /**
   * Get pads associated to each photoline.
   * If no pads are associated the set size is 0.
   * \return Map holding for each photoline a set of pads.
   */
  std::map<int, std::set<gear::GlobalPadIndex> > GetPhotolinePads();
  /**
   * Add pads associated to each photoline.
   * \param module_id Give the module where the line belongs to.
   * \param pads Global pad indices of the associated pads.
   */
  void AddPhotolineProjection(int module_id, const std::set<gear::GlobalPadIndex> &pads);
};
}
#endif
