#ifdef USE_LCCD

// MarlinTPC
#include "AFTERRawDataConverterProcessor.h"

// Marlin
#include <marlin/Exceptions.h>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCFlagImpl.h>


using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{


AFTERRawDataConverterProcessor aAFTERRawDataConverterProcessor;


AFTERRawDataConverterProcessor::AFTERRawDataConverterProcessor() : Processor("AFTERRawDataConverterProcessor"),

        _input_tracker_raw_data_collection_name("ADC_CHAN_DATA"),
        _output_tracker_data_collection_name("TPCData"),
        _output_is_transient(true),
        _new_pulse_sentinal_bit(13),
        _bin_count_before_pulse(10),
        _bits_per_sample_override(12)
{

    /* description of the processor which will be displayed in the
    * steering file autometicly generated by
    * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
    */

    _description = "AFTERRawDataConverterProcessor splits the input raw data into smaller raw data chunks if it is zero-suppressed. This needs to happen because the (ALTRO?) Front End Mezzanine (FEM) card puts sentinal values into the ADC values which split the data. These need to be removed and the correct starting time for each collection of ADC values needs to be calculated.";

    /* register steering parameters: name, description, class-variable, default value
    * the type will by definded by the type of the default value
    * string, double, float and int are possible
    * plaese note to get a float you must cast in C++: (float) 1.3
    */
    registerInputCollection(LCIO::TRACKERRAWDATA ,
                            "InputTrackerRawData",
                            "Name of the input collection of TPC raw data from an AFTER FEM card as stored in the LCIO file.",
                            _input_tracker_raw_data_collection_name,
                            std::string("TPCSuppressedRawData"));

    registerOutputCollection(LCIO::TRACKERDATA ,
                             "OutputTrackerData",
                             "The name of the output collection of unsuppressed TPC tracker data.",
                             _output_tracker_data_collection_name,
                             std::string("TPCData"));

    registerOptionalParameter("SetOutputTransient",
                              "Controls whether the output collection should be saved in the LCIO file (false) or if it should be deleted at the end (true - default).",
                              _output_is_transient,
                              true);

    registerOptionalParameter("NewPulseSentinalBit" ,
                              "Specifies which bit in the raw ADC values designates when a new pulse is next in the data stream. The count starts from 1 ;in a 16 bit bin the bit will be in the range [1,16]. Defaults to bit 13.",
                              _new_pulse_sentinal_bit,
                              13);

    registerOptionalParameter("BinCountBeforePulse",
                              "Specifies the number of bins which are stored before the signal starts. The default value for the FEM card is 10 bins so it is the default in MarlinTPC as well.",
                              _bin_count_before_pulse,
                              10);

    /**registerOptionalParameter( "ElectronicsParametersCollectionName" ,
                "Name of the conditions collection in which the parameters of the read out electronics is stored",
                _electronic_parameters_collection_name,
                tpcconddata::GenericADCElectronicsParameters::getDefaultColName() );*/

    // now some parameters for overriding values which should be obtained from LCCD
    registerOptionalParameter("BitsPerSampleOverride",
                              "The number of bits per ADC sample (Default - 12). This overrides any information provided by LCCD. Currently this information can't be provided by LCCD so this parameter must be used.",
                              _bits_per_sample_override,
                              12);
    // end of the override variables
}


void AFTERRawDataConverterProcessor::init()
{

    // usually a good idea to
    printParameters();

    // do a sanity check...
    if (parameterSet("BitsPerSampleOverride"))
        {

            if (_bits_per_sample_override >= _new_pulse_sentinal_bit)
                {

                    streamlog_out(ERROR) << "The number of bits per data sample ( " << _bits_per_sample_override << ") must be less than the sentinal bit number (" << _new_pulse_sentinal_bit << "). There can't be a flag inside of the data!" << endl;
                }
        }
}


//FIXME: not used: void AFTERRawDataConverterProcessor::processRunHeader(lcio::LCRunHeader* run)
//{
//
//}

void AFTERRawDataConverterProcessor::processEvent(lcio::LCEvent * evt)
{


    LCCollection* input_suppressed_raw_data = evt->getCollection(_input_tracker_raw_data_collection_name);

    streamlog_out(MESSAGE) << "Processing " << input_suppressed_raw_data->getNumberOfElements() << " raw data pieces." << endl;

    // create the output collection with a string defining the TYPE (see lcio.h)
    LCCollectionVec* output_tracker_data = new LCCollectionVec(LCIO::TRACKERDATA);
    // set the transient flag as definded by the processor parameter
    output_tracker_data->setTransient((_output_is_transient != 0));
    output_tracker_data->setSubset(false);

    // set the flags that cellID1 should be stored
    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    output_tracker_data->setFlag(trkFlag.getFlag());

    // calculate the bit mask for the data
    // FIXME: include an option to get the data from LCCD and calculate the masks on a per-readout group basis
    const short data_mask = (1 << _bits_per_sample_override) - 1;

    const short new_pulse_flag = (1 << (_new_pulse_sentinal_bit - 1));

    int uncompressed_raw_data_count = 0;

    // loop through the data, unsuppressing it
    for (int i_raw_data = 0; i_raw_data < input_suppressed_raw_data->getNumberOfElements(); i_raw_data++)
        {

            TrackerRawData* raw_data = dynamic_cast<TrackerRawData*>(input_suppressed_raw_data->getElementAt(i_raw_data)) ;

            if (!raw_data)
                {

                    throw(Exception("Input collection " + _input_tracker_raw_data_collection_name + " does not contain TrackerRawData."));
                }

            // ShortVec is from LCIO
            ShortVec suppressed_adc_values = raw_data->getADCValues();

            ShortVec::const_iterator adc_value = suppressed_adc_values.begin();

            TrackerDataImpl *new_data = NULL;

            // start the first output TrackerRawData object
            if ((*adc_value & new_pulse_flag) == 0)
                {

                    // start a new non-suppressed signal
                    new_data = new TrackerDataImpl();
                    new_data->setCellID0(raw_data->getCellID0());
                    new_data->setCellID1(raw_data->getCellID1());
                    new_data->setTime(raw_data->getTime());

                    streamlog_out(DEBUG) << "Found unsuppressed TrackerRawData with CellIDs: {" << raw_data->getCellID0() << "," << raw_data->getCellID1() << "}" << endl;
                    uncompressed_raw_data_count += 1;
                }

            while (adc_value != suppressed_adc_values.end())
                {

                    if ((*adc_value & new_pulse_flag) != 0)
                        {

                            // for some reason there is a value of -999 at the end of the unsuppressed data...
                            if (*adc_value == -999)
                                {

                                    break;
                                }

                            // add the latest TrackerRawData to the collection
                            if (new_data != NULL)
                                {

                                    output_tracker_data->addElement(new_data);
                                }

                            streamlog_out(DEBUG) << "ADC value" << *adc_value << endl;

                            // start a new suppressed signal
                            new_data = new TrackerDataImpl();

                            new_data->setCellID0(raw_data->getCellID0());
                            new_data->setCellID1(raw_data->getCellID1());

                            // calculate the time
                            const int time_offset = (*adc_value & data_mask);
                            const int time = raw_data->getTime() + time_offset - _bits_per_sample_override;
                            new_data->setTime(time);

                            streamlog_out(DEBUG) << "Found suppressed TrackerRawData with CellIDs: {" << new_data->getCellID0() << "," << new_data->getCellID1() << "} starting at time " << new_data->getTime() << endl;

                            uncompressed_raw_data_count += 1;
                        }
                    else
                        {

                            // add the current adc value to the new signal
                            new_data->chargeValues().push_back(*adc_value);
                        }

                    adc_value += 1;
                }

            // add the final TrackerRawData to the collection
            output_tracker_data->addElement(new_data);
        }


    streamlog_out(MESSAGE) << "Created " << uncompressed_raw_data_count << " unsuppressed TrackerRawData objects." << endl;
    evt->addCollection(output_tracker_data, _output_tracker_data_collection_name);
}



// FIXME: not used: void AFTERRawDataConverterProcessor::check(lcio::LCEvent * evt)
//{
//    // nothing to check here - could be used to fill checkplots in reconstruction processor
//}


void AFTERRawDataConverterProcessor::end()
{

}

}

#endif // USE_LCCD
