#include "ElectronCloudFinderProcessor.h"

// C++
#include <cmath>
#include <stack>
#include <list>
#include <utility>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/LCFloatVec.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/LCFlagImpl.h>

// GEAR
/*#include <gearimpl/RectangularPadRowLayout.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>*/
#include <gear/TPCParameters.h>
/*#include <gearimpl/GlobalPadIndex.h>*/

#include <gear/PadRowLayout2D.h>
#include <gear/TPCModule.h>


// MARLIN
#include <marlin/Global.h>



using namespace marlin;
using namespace gear;
using namespace std;

namespace marlintpc
{

    ElectronCloudFinderProcessor aElectronCloudFinderProcessor;


    ElectronCloudFinderProcessor::ElectronCloudFinderProcessor() : Processor("ElectronCloudFinderProcessor")
    {


        /**
         * description of the processor which will be displayed in the
         * steering file autometicly generated by
         * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
         */
        _description = "ElectronCloudFinderProcessor calculates TrackerHits from TrackerPulses";


        // register steering parameters: name, description, class-variable, default value
        registerInputCollection(LCIO::TRACKERPULSE,
                                "InputTrackerPulses",
                                "Name of the input TrackerPulses collection",
                                _inputTrackerPulsesCollectionName,
                                std::string("TPCPulses"));

        registerOutputCollection(LCIO::TRACKERHIT,
                                 "OutputTrackerHits",
                                 "Name of the output TrackerHits collection",
                                 _outputTrackerHitsCollectionName,
                                 std::string("TPCHits"));

        registerProcessorParameter("SetOutputHitsTransient",
                                   "if true the output hits collection is set transient (default: false)",
                                   _outputHitsTransient,
                                   bool(false));
  
		registerProcessorParameter( "NegativeZDirection" ,
									"Set to true for the second half TPC, which look into the negative z direction (default: false)"  ,
									_negativeZDirection,
									bool(false));

        registerProcessorParameter("MaxTimeSpread",
                                   "Maximum time between pulses in a hit in ns (default: 200.)",
                                   _maxTimeSpread,
                                   float(200.));

        registerProcessorParameter("VDrift",
                                   "Optional: Set drift velocity in case there is no conditions data in mm/us",
                                   _vDrift,
                                   float(0.));

        registerOptionalParameter("MinHitSize",
                                  "Minimum size of hit / Minimum number of Pads (bigger than 0, default: 1)",
                                  _minHitSize,
                                  int(1));

        registerOptionalParameter("MaxHitSize",
                                  "Maximum size of hit / Maximum number of Pads (default: 1)",
                                  _maxHitSize,
                                  int(1));
    }


    void ElectronCloudFinderProcessor::init()
    {
        // Usually a good idea to
        printParameters();

	    //get max drift distance from gear
	    _maxDriftDistance = Global::GEAR->getTPCParameters().getMaxDriftLength ();
		/// Create an object which holds the information about the neighbourship of all pads.
		myNeighbouringPadHandler = new NeighbouringPadHandler();
    }

    void ElectronCloudFinderProcessor::processRunHeader(LCRunHeader* run)
    {
        run->parameters().setValue(_processorName + "_revision", std::string("$Rev: 1645 $"));

        for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
            {
                run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
            }
        }
    }



    void ElectronCloudFinderProcessor::processEvent(LCEvent * evt)
    {
        // some status output
        if (evt->getEventNumber() % 1000 == 0)
            streamlog_out(MESSAGE4) << "\r" << _processorName << " processing event " << evt->getEventNumber()
            << std::endl;

        // allocate the output collection (hits)
        LCCollectionVec* tpcHitsCollection = new LCCollectionVec(LCIO::TRACKERHIT);

        // set the transient flag as definded by the processor parameter
        tpcHitsCollection->setTransient(_outputHitsTransient);

        // get the pulses collections from the event
        LCCollectionVec* inputPulsesCollection = 0;
        try
        {
            inputPulsesCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection(_inputTrackerPulsesCollectionName));
        }
        catch (DataNotAvailableException &e)
        {
            // there are not TPC TrackerPuses in this event
            // just continue ...
            return;
        }

        // Prepare a list of pointers to the pulses with additional found flag.
        std::list< pair<TrackerPulse*, bool> > pulsesList;

        // A stack of iterators to the list elements above for the iterative version of a flood fill algorithm.
        std::stack< std::list< pair<TrackerPulse*, bool> >::iterator> pulsesStack;

        // fill list with pointers to pulses
        for (LCCollectionVec::iterator itPulse = inputPulsesCollection->begin();
                itPulse < inputPulsesCollection->end(); itPulse++)
        {
            TrackerPulse *thisPulse = dynamic_cast<TrackerPulse *>(*itPulse);

            pulsesList.push_back(make_pair(thisPulse, false));

        }
        pulsesList.sort(comparePulses);

        while (!pulsesList.empty())
        {
            // Retrieve the iterator to the first pulse+bool in the list and push it onto the stack.
            // We will use this as the first seed pulse later on
            std::list< pair<TrackerPulse*, bool> >::iterator itSearchPulse = pulsesList.begin();
            pulsesStack.push(itSearchPulse);

            // Mark the pulse as "found"
            (*itSearchPulse).second = true;

            TrackerHitImpl* hitCandidate = new TrackerHitImpl();

            while (!pulsesStack.empty())
            {
                std::list< pair<TrackerPulse*, bool> >::iterator itSeedPulse = pulsesStack.top();
                pulsesStack.pop();

                // Extract pulse object for convenience
                TrackerPulse* seedPulse = dynamic_cast<TrackerPulse*>((*itSeedPulse).first);

                const int seedPadIndex = seedPulse->getCellID0();
                const int seedModuleID = seedPulse->getCellID1();
                const float seedTime = seedPulse->getTime();
		streamlog_out(DEBUG4) << "SeedModule ID " << seedModuleID << " SeedPadIndex" << seedPadIndex << std::endl;		
                const int smallestNeighbour = myNeighbouringPadHandler->getSmallestNeighbouringPadIndex(seedModuleID, seedPadIndex);
                const int largestNeighbour = myNeighbouringPadHandler->getLargestNeighbouringPadIndex(seedModuleID, seedPadIndex);

                // Search preceding pulses in list for neighbourship
                for (std::list< pair<TrackerPulse*, bool> >::iterator itThisPulse = itSeedPulse;
                        itThisPulse != pulsesList.rend().base(); itThisPulse--)
                {
                    // Extract pulse object for convenience
                    TrackerPulse* thisPulse = dynamic_cast<TrackerPulse*>((*itThisPulse).first);

                    const int thisPadIndex = thisPulse->getCellID0();
                    const int thisModuleID = thisPulse->getCellID1();
                    const float thisTime = thisPulse->getTime();
                    const bool pulseNotFoundYet = ((*itThisPulse).second == false);

                    const bool outOfSearchRange = (thisModuleID != seedModuleID) || (thisPadIndex < smallestNeighbour);
                    if (outOfSearchRange)
                    {
                        break;
                    }

                    const bool pulseBelongsToHit = myNeighbouringPadHandler->areNeighbouringPads(seedModuleID, seedPadIndex, thisPadIndex) &&
                                                   pulseNotFoundYet &&
                                                   fabs(thisTime - seedTime) <= _maxTimeSpread;
                    if (pulseBelongsToHit)
                    {
                        pulsesStack.push(itThisPulse);

                        // Mark this pulse as found
                        (*itThisPulse).second = true;
                    }
                }

                // Search succeeding pulses in list for neighbourship
                for (std::list< pair<TrackerPulse*, bool> >::iterator itThisPulse = itSeedPulse;
                        itThisPulse != pulsesList.end(); itThisPulse++)
                {
                    // Extract pulse object for convenience
                    TrackerPulse* thisPulse = dynamic_cast<TrackerPulse*>((*itThisPulse).first);

                    const int thisPadIndex = thisPulse->getCellID0();
                    const int thisModuleID = thisPulse->getCellID1();
                    const float thisTime = thisPulse->getTime();
                    const bool pulseNotFoundYet = ((*itThisPulse).second == false);

                    const bool outOfSearchRange = (thisModuleID != seedModuleID) || (thisPadIndex > largestNeighbour);
                    if (outOfSearchRange)
                    {
                        break;
                    }

                    const bool pulseBelongsToHit = myNeighbouringPadHandler->areNeighbouringPads(seedModuleID, seedPadIndex, thisPadIndex) &&
                                                   pulseNotFoundYet &&
                                                   fabs(thisTime - seedTime) <= _maxTimeSpread;
                    if (pulseBelongsToHit)
                    {
                        pulsesStack.push(itThisPulse);

                        // Mark this pulse as found
                        (*itThisPulse).second = true;
                    }
                }

                // Add the seed pulse to the hit candidate's vector of pulses.
                // Besides, this seed pulse is not needed anymore and can be deleted from the search list.
                hitCandidate->rawHits().push_back(seedPulse);
                pulsesList.erase(itSeedPulse);
            } // Ends while. Continue with a new seed pulse which might have been pushed on the stack meanwhile,
            // otherwise the hit is complete and we can go on.

            // If the parameters were set, check wether hit fullfills the size criteria, it needs to be 1 pulse at least ...
            const unsigned int hitSize = hitCandidate->rawHits().size();
            bool hitIsRightSize = true;

			if (hitSize < 1)
			{
				hitIsRightSize = false;
			}
			
			if (parameterSet("MinHitSize"))
            {
                if (hitSize < static_cast<unsigned int>(_minHitSize))
				{
                    hitIsRightSize = false;
                }
            }

            if (parameterSet("MaxHitSize"))
            {
                if (hitSize > static_cast<unsigned int>(_maxHitSize))
                {
                    hitIsRightSize = false;
                }
            } 

			// All checks done.
            if (hitIsRightSize)
            {

                // hitCandidate is valid

                // calculate the hit from the cluster
                // the sums needed for centre of gravity:
                double sumXPhiQ = 0; // sum(x*q) or sum(phi*q)
                double sumYRQ = 0; // sum(y*q) or sum(r*q)
                double sumZQ = 0;    // sum(z*q)
                double sumQ = 0;  // sum(q)
                double xPhi = 0;   // coordinate along pad row
                double yR = 0;   // coordinate orthogonal to pad row

				// From now on work only on the module which contains the hit,
				// for this extract the first pulse from the hit to get the gear settings ...
				EVENT::LCObjectVec::iterator pulseIter = hitCandidate->rawHits().begin();
				const TrackerPulse* firstPulse = dynamic_cast<TrackerPulse*>(*pulseIter);
				
				const gear::TPCModule &module =
					marlin::Global::GEAR->getTPCParameters().getModule(firstPulse->getCellID1());
		
				const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();
				

                // loop over all pulses in cluster and sum up
                for (pulseIter = hitCandidate->rawHits().begin();
                        pulseIter < hitCandidate->rawHits().end(); pulseIter++)
                {
                    // from now on work only on the module which contains the hit

                    // Raw hits is a vector of LCObjects.
                    // We have to cast the iterator to TrackerPulse to access the member functions.
                    TrackerPulse* pulse = dynamic_cast<TrackerPulse*>(*pulseIter);

                    // get pad coordinates of current pulse from GEAR layout
                    Vector2D padCenterGlobal = module.getPadCenter(pulse->getCellID0());
                    Vector2D padCenterLocal = module.globalToLocal(padCenterGlobal[0],padCenterGlobal[1]);

                    // get x, y / r, phi coordinate from first pad
                    switch (localPadPlane.getCoordinateType())
                    {
                        case PadRowLayout2D::CARTESIAN:
                            xPhi = padCenterLocal[0]; // x
							yR = padCenterLocal[1]; // y
                            break;
                        case PadRowLayout2D::POLAR:
                            xPhi = padCenterLocal[1]; // phi
							yR = padCenterLocal[0]; // r
                            break;
                        default:
                            throw gear::Exception("ElectronCloudFinder: unkown coordinate type");
                    };

                    // summ correctly if handling pads from the first and thr fourth quadrant
                    // in polar coordinates (wrap around from 2 pi to 0)
                    if ((pulseIter != hitCandidate->rawHits().begin())  // for the first pad there is nothing to compare with
                            && (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR))  // problem only occurs in polar coordinates
                    {
                        // there can be thwo cases
                        //1.
                        if ((sumXPhiQ / sumQ < M_PI_2) // pulse up to now is in first quadrant (or in fourth, already mapped to <0)
                                && (xPhi > 1.5*M_PI))  // current pulse in fourth quadrant
                            xPhi -= 2 * M_PI; // map to range [-pi:pi] so averaging works

                        // 2.
                        // no else needed; statements are mutually excluse because due to first condition
                        // even if xPhi was changed in the first statement
                        if ((sumXPhiQ / sumQ > 1.5* M_PI) // pulse up to now is in fourth quadrant (or in first, already mapped to >2*pi)
                                && (xPhi < M_PI_2))  // current pulse in first quadrant
                            xPhi += 2 * M_PI; // map to range [pi:3*pi] so averaging works
                    }

                    // add to the sums
                    sumXPhiQ += xPhi * pulse->getCharge();  // x * q or phi * q
                    sumYRQ += yR * pulse->getCharge();
                    sumZQ += pulse->getTime() * _vDrift / 1000 // t * vDrift = z // 1000 ns/us
                             * pulse->getCharge() ;   // * q
                    sumQ += pulse->getCharge() ;    // q

                }// for (pulse)

                // global 2D position, cartesian or polar
                gear::Vector2D positionGlobal2D;

                switch (localPadPlane.getCoordinateType())
                {
                    case PadRowLayout2D::CARTESIAN:
                        // in cartesian coordinates xPhi is x, yR is y
                        positionGlobal2D = module.localToGlobal(sumXPhiQ / sumQ , sumYRQ / sumQ);
                        break;
                    case PadRowLayout2D::POLAR:
                        positionGlobal2D = module.localToGlobal(sumYRQ / sumQ , sumXPhiQ / sumQ);
                        break;
                };

                // the positon in global cartesian for lcio
                double positionGlobal[3];
                switch (module.getTPCCoordinateType())
                {
                    case PadRowLayout2D::CARTESIAN:
                        // in cartesian coordinates xPhi is x, yR is y
                        positionGlobal[0] = positionGlobal2D[0];
                        positionGlobal[1] = positionGlobal2D[1];
                        break;
                    case PadRowLayout2D::POLAR:
                        // convert polar to cartesian
                        positionGlobal[0] = positionGlobal2D[0] * cos(positionGlobal2D[1]);
                        positionGlobal[1] = positionGlobal2D[0] * sin(positionGlobal2D[1]);
                        break;
                };

                // sumZQ contains the weighted drift distance
                // calculate z coordinate wrt. origin, which is at the cathode
                positionGlobal[2] = _maxDriftDistance - sumZQ / sumQ;

                // the second half TPC look in the negative z direction
                // (rotated by 180 degress around y axis wrt. first endplate)
                // multiply x and z with -1 in this case
                if (_negativeZDirection)
                {
                    positionGlobal[0] *= -1.;
                    positionGlobal[2] *= -1.;
                }

                // set position and dEdx of the hit
                hitCandidate->setPosition(positionGlobal);
//                hitCandidate->setdEdx(sumQ);
                hitCandidate->setEDep(sumQ);

                // add the valid hit to the hit collection
                tpcHitsCollection->addElement(hitCandidate);
            } // if (hitIsRightSize)
            else
            {
                streamlog_out(MESSAGE2) << "Charge cloud rejected, does not fulfil size criteria (" << hitSize << ")." << endl;
                delete hitCandidate;
            }
        }






        // All loops are finished, add the collections to the event
        if (tpcHitsCollection->getNumberOfElements() > 0)
        {
            evt->addCollection(tpcHitsCollection, _outputTrackerHitsCollectionName);

            streamlog_out(MESSAGE4) << "Found " << tpcHitsCollection->getNumberOfElements()
            << " charge clouds in event." << std::endl;

            // dump the found hits in MESSAGE1 level
            if (streamlog::out.write< streamlog::MESSAGE1 > ())
            {
                for (IMPL::LCCollectionVec::iterator hitIter = tpcHitsCollection->begin();
                        hitIter != tpcHitsCollection->end(); hitIter++)
                {
                    TrackerHit *hit = 0;
                    try
                    {
                        hit = dynamic_cast<EVENT::TrackerHit*>(*hitIter);
                    }
                    catch (std::bad_cast &)
                    {
                        streamlog::out() << "ERROR: bad cast when dumping hit coordinates."
                        << std::endl;
                        break;
                    }

                    const int numberOfPulsesInHit = hit->getRawHits().size();

                    const double * hitPos = hit->getPosition();
                    streamlog_out(MESSAGE1) << "Found charge cloud at " << hitPos[0] << "\t"
                    << hitPos[1] << "\t" << hitPos[2] << " with " << numberOfPulsesInHit << " pulses." << std::endl;

                }

            }// if MESSAGE1
        } // tpcHitsCollection not empty
        else
        {
            delete tpcHitsCollection;
        }
    }



    void ElectronCloudFinderProcessor::check(LCEvent *  /*evt*/)
    {
        // nothing to check here - could be used to fill checkplots in reconstruction processor
    }



    void ElectronCloudFinderProcessor::end()
    {

    }



    bool ElectronCloudFinderProcessor::comparePulses(
        pair<TrackerPulse*, bool> firstSortingPair,
        pair<TrackerPulse*, bool> secondSortingPair)
    {
        // For comparing the two pulses we need CellIDs 1 and 0 and time of each signal:
        int firstModuleID = (firstSortingPair).first->getCellID1();
        int secondModuleID = (secondSortingPair).first->getCellID1();

        int firstPadIndex = (firstSortingPair).first->getCellID0();
        int secondPadIndex = (secondSortingPair).first->getCellID0();

        float timeOfFirstSignal = (firstSortingPair).first->getTime();
        float timeOfSecondSignal = (secondSortingPair).first->getTime();

        // and then we just compare ...
        if ((firstModuleID <= secondModuleID) &&
                (firstPadIndex <= secondPadIndex) &&
                (timeOfFirstSignal < timeOfSecondSignal))
        {
            return true;
        }
        else
        {
            return false;
        }
    } // End compareMeasurements(...).

}// namespace marlintpc
