#include <NeighbouringPadHandler.h>

// C++
#include <algorithm>
#include <vector>
#include <map>

// GEAR
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/PadRowLayout2D.h"
#include "gearimpl/Vectors.h"
#include "gear/TPCModule.h"

// Marlin
#include "marlin/Global.h"

using namespace marlin;
namespace marlintpc
{
// ******************************************************
// * Implementation of the NeighbouringPadsObject class *
// ******************************************************

  NeighbouringPadsObject::NeighbouringPadsObject()
  {
  }

  NeighbouringPadsObject::NeighbouringPadsObject(
      const NeighbouringPadsObject& originalObject)
  {
    _thisPadsPadIndex = originalObject.getPadIndex();

    _leftRowCoordinate = originalObject.getLeftSide(); ///< x or phi-coordinate of the left corner of the pad
    _rightRowCoordinate = originalObject.getRightSide(); ///< x or phi-coordinate of the right corner of the pad

    _precedingPads = originalObject.getPrecedingPads();
    _succeedingPads = originalObject.getSucceedingPads();
  }

  NeighbouringPadsObject::NeighbouringPadsObject(const int padIndex)
  {
    _thisPadsPadIndex = padIndex;
  }

  std::vector<int>
  NeighbouringPadsObject::getPrecedingPads() const
  {
    return _precedingPads;
  }

  std::vector<int>
  NeighbouringPadsObject::getSucceedingPads() const
  {
    return _succeedingPads;
  }

  void
  NeighbouringPadsObject::addNeighbouringPad(const int padIndex)
  {
    if (padIndex < _thisPadsPadIndex)
      {
        _precedingPads.push_back(padIndex);
      }

    if (padIndex > _thisPadsPadIndex)
      {
        _succeedingPads.push_back(padIndex);
      }
    // otherwise, e.g. if pad indices are equal, do nothing

    //DEBUG: std::cout << "Pad: " << _thisPadsPadIndex << " Neighbour:" << padIndex << std::endl;

  }

  void
  NeighbouringPadsObject::setPadSideRowCoordinates(double cLeft, double cRight)
  {
    _leftRowCoordinate = cLeft;
    _rightRowCoordinate = cRight;
  }

  double
  NeighbouringPadsObject::getLeftSide() const
  {
    return _leftRowCoordinate;
  }

  double
  NeighbouringPadsObject::getRightSide() const
  {
    return _rightRowCoordinate;
  }

  int
  NeighbouringPadsObject::getPadIndex() const
  {
    return _thisPadsPadIndex;
  }

// ******************************************************
// * Implementation of the NeighbouringPadHandler class *
// ******************************************************

  NeighbouringPadHandler::NeighbouringPadHandler()
  {
    // In the constructor the map with information about neighbourship of pads shall filled automatically from the GEAR geometry.
    //
    // Each pad can have neighbours in the preceding and succeeding row. However it's not neccessary to iterate both rows, but only
    // the preceding one and then inform both pad objects that they are neighbours. That's because for the preceding neighbour of
    // the current pad, the current pad ist a succeeding neighbour and the preceding pad object already exists due to the algorithm.
    //
    // After that the immediate left and right neighbour of the pad is added and the pad object is stored in a map for the module's
    // pad plane. The pad objects held by this map are consecutively modiefied by the algorithm, until all pads have been iterated
    // over and thus all neighbourship information is complete. The map is then made the inner part of the more complex container
    // structure used to access the neighbourship information by moduleIndex and padIndex.
    // This procedure is repeated for all modules.

    // For convenience shorten the definition of the module vector
    typedef const std::vector<gear::TPCModule*> GearModuleVector;

    // Some preparation ...
    const gear::TPCParameters &tpc_parameters =
        Global::GEAR->getTPCParameters();

    GearModuleVector &modules = tpc_parameters.getModules();

    // Iterate through all modules of the GEAR file ...
    for (GearModuleVector::const_iterator itModule = modules.begin();
        itModule < modules.end(); itModule++)
          {
      // Extract module pointer from iterator, for clearer writing
      gear::TPCModule* module = *itModule;

      // Get some information about the module ...
      int moduleID = module->getModuleID();
      const int numberOfRowsInModule = module->getNRows();
      const gear::PadRowLayout2D &localPadPlane = module->getLocalPadLayout();
      unsigned short IndexXPhi = 0; //Index for the coordinate along a row, it is 0 for x (cartesian layout) and 1 for phi (polar layout)

      switch (localPadPlane.getCoordinateType())
        {
      case gear::PadRowLayout2D::CARTESIAN:
        IndexXPhi = 0; // x
        break;
      case gear::PadRowLayout2D::POLAR:
        IndexXPhi = 1; // phi
        break;
      default:
        throw gear::Exception("TopoFinder: unkown coordinate type");
        };

      // Create a map for the pad plane of this module.
      std::map<int, NeighbouringPadsObject> padPlaneMap;

      // Iterate through the pad plane by row and pad ...
      for (int rowNumber = 0; rowNumber < numberOfRowsInModule; rowNumber++)
        {
        // For convenience shorten the definition of the pad vector
        typedef const std::vector<int> GearPadVector;

        GearPadVector &padsInRow = module->getPadsInRow(rowNumber);

        for (GearPadVector::const_iterator itPad = padsInRow.begin();
            itPad < padsInRow.end(); itPad++)
              {
          // Get and calculate information about the pad
          const int padNumber = module->getPadNumber(*itPad);
          const int padIndex = module->getPadIndex(rowNumber, padNumber);
 //         std::cout << "PadIndex: " <<  padIndex << " Module: " << moduleID << " Row: " << rowNumber << " Pad: " << padNumber << std::endl;
          const double halfPadWidth = module->getPadWidth(padIndex) / 2.0;
          const gear::Vector2D padCenterGlobal = module->getPadCenter(padIndex);
          const gear::Vector2D padCenter = module->globalToLocal(padCenterGlobal[0],padCenterGlobal[1]);
          const double leftSide = padCenter[IndexXPhi] - halfPadWidth;
          const double rightSide = padCenter[IndexXPhi] + halfPadWidth;

          // Create a NeighbouringPadsObject for the current pad.
          NeighbouringPadsObject currentPadObject(padIndex);

          // Store left and right coordinate of the pad for finding neighbours in later iterations.
          currentPadObject.setPadSideRowCoordinates(leftSide, rightSide);

          // Search in the preceding row for neighbouring pads ...
          if (rowNumber > 0)
            {
              GearPadVector &padsInRowSearchVector = module->getPadsInRow(
                  rowNumber - 1);

              for (GearPadVector::const_iterator itSearchPad =
                  padsInRowSearchVector.begin();
                  itSearchPad < padsInRowSearchVector.end(); itSearchPad++)
                    {
                const int searchPadNumber = module->getPadNumber(*itSearchPad);
                const int searchPadIndex = module->getPadIndex(rowNumber - 1,
                    searchPadNumber);

                // We can use the already stored NeighbouringPadsObject of the last row:
                NeighbouringPadsObject* searchPadObject = &(padPlaneMap.find(
                    searchPadIndex)->second); //getNeighbouringPadsObject(moduleID, searchPadIndex);

                //TODO: Neighbour condition for circular layouts is a little more complicated:
                const bool neighbourFound = (searchPadObject->getLeftSide()
                    <= rightSide)
                    && (searchPadObject->getRightSide() >= leftSide);

                if (neighbourFound)
                  {
/*                    std::cout << std::endl;
                    std::cout << "Neighbour found!" << std::endl;
                    std::cout << "pRow: " << rowNumber << " sRow: "
                        << rowNumber - 1 << " pPad: " << padNumber << " sPad: "
                        << searchPadNumber << std::endl;
                    std::cout << "pPadIndex: " << padIndex << " sPadIndex: "
                        << searchPadIndex << std::endl;
                    std::cout << "halfPadWidth: " << halfPadWidth << " X0: "
                        << padCenter[0] << " X1: " << padCenter[1] << std::endl;
                    std::cout << "sleftSide: " << searchPadObject->getLeftSide()
                        << " <= prightSide: " << rightSide << std::endl;
                    std::cout << "srightSide: "
                        << searchPadObject->getRightSide() << " >= pleftSide: "
                        << leftSide << std::endl;*/
                    // Add the found neighbour to the list of the current pad.
                    currentPadObject.addNeighbouringPad(searchPadIndex);
                    // Add the current pad to the list of the found neighbouring pad respectively,
                    // thus we save a forward search.
                    searchPadObject->addNeighbouringPad(padIndex);
                  }
              }
          }

        // Add the immediate predecessor of the pad in the same row. First pad in row will cause an exception in some layouts.
        try
          {
            currentPadObject.addNeighbouringPad(
                module->getLeftNeighbour(padIndex));
          }
        catch (...)
          {
            // If so, just continue.
          }

        // Add the immediate successor of the pad in the same row. Last pad in row will cause an exception in some layouts.
        try
          {
            currentPadObject.addNeighbouringPad(
                module->getRightNeighbour(padIndex));
          }
        catch (...)
          {
            // If so, just continue.
          }

        // Store the pad information inside the inner map.
        padPlaneMap[padIndex] = currentPadObject;

      } // ... ends pad iteration.

  } // ... ends row iteration.

// Store the inner map of the module's pad plane inside the global map of the entire end plate.
_endPlateMap[moduleID] = padPlaneMap;
} // ... ends module iteration.
std::cout << "Preceding 65536: " << getNeighbouringPadsObject(3,65536)->getPrecedingPads().size() << std::endl;
std::cout << "Succeding 65536: " << getNeighbouringPadsObject(3,65536)->getSucceedingPads().size() << std::endl;
}

bool
NeighbouringPadHandler::areNeighbouringPads(int moduleIndex, int padIndex1,
int padIndex2) const
{
bool pad2IsNeighbourOfPad1 = false;

 // Before starting any searches check, wether padIndex2 is within the upper and
 // lower limits of the neighbouring pad indices of padIndex1
const bool padIndexIsWithinLimits = (padIndex2
>= getSmallestNeighbouringPadIndex(moduleIndex, padIndex1))
&& (padIndex2 <= getLargestNeighbouringPadIndex(moduleIndex, padIndex1));

if (padIndexIsWithinLimits)
{
if (padIndex2 < padIndex1)
  {
    std::vector<int> predecessors = getPrecedingNeighbouringPads(moduleIndex,
        padIndex1);

    for (std::vector<int>::const_iterator itPadIndex = predecessors.begin();
        itPadIndex < predecessors.end(); itPadIndex++)
          {
      if (*itPadIndex == padIndex2)
        {
          pad2IsNeighbourOfPad1 = true;
          break;
        }
    }
}
else
{
  std::vector<int> successors = getSucceedingNeighbouringPads(moduleIndex,
      padIndex1);

  for (std::vector<int>::const_iterator itPadIndex = successors.begin();
      itPadIndex < successors.end(); itPadIndex++)
        {
    if (*itPadIndex == padIndex2)
      {
        pad2IsNeighbourOfPad1 = true;
        break;
      }
  }
}
}

return pad2IsNeighbourOfPad1;
}

int
NeighbouringPadHandler::getSmallestNeighbouringPadIndex(int moduleIndex,
int padIndex) const
{
  int SNPIndex = 0;
  
  if(_endPlateMap.find(moduleIndex)->second.find(padIndex)->second.getPrecedingPads().size()>0)
  {
    SNPIndex = *(_endPlateMap.find(moduleIndex)->second.find(padIndex)->second.getPrecedingPads().begin());
  }
  else if(padIndex == 0) 
  {
      SNPIndex = 0;
  }
  
  return(SNPIndex);
}

int
NeighbouringPadHandler::getLargestNeighbouringPadIndex(int moduleIndex,
int padIndex) const
{
return _endPlateMap.find(moduleIndex)->second.find(padIndex)->second.getSucceedingPads().back();
}

std::vector<int>
NeighbouringPadHandler::getPrecedingNeighbouringPads(int moduleIndex,
int padIndex) const
{
return _endPlateMap.find(moduleIndex)->second.find(padIndex)->second.getPrecedingPads();
}

std::vector<int>
NeighbouringPadHandler::getSucceedingNeighbouringPads(int moduleIndex,
int padIndex) const
{
return _endPlateMap.find(moduleIndex)->second.find(padIndex)->second.getSucceedingPads();
}

NeighbouringPadsObject*
NeighbouringPadHandler::getNeighbouringPadsObject(int moduleIndex, int padIndex)
{
return &_endPlateMap.find(moduleIndex)->second.find(padIndex)->second;
}
} // ... ends namespace marlintpc
