//LCIO
#include <lcio.h>
#include <EVENT/LCIO.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackerHitImpl.h>
//MarlinTPC
#include "PhotodotReconstructionProcessor.h"
#include "ElectronCloud.h"


//Marlin
#include <marlin/Global.h>

//C++
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <cmath>
#include <iomanip>
#include <set>
#include <cassert>

//GEAR
#include <gearimpl/RectangularPadRowLayout.h>
#include <gearimpl/FixedDiskLayoutBase.h>
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gearxml/tinyxml.h>

//Root
#include <TMath.h>

using namespace marlin;
using namespace gear;
using namespace lcio;

namespace marlintpc {

// Create an instance of the processoer to register it with Marlin
PhotodotReconstructionProcessor aPhotodotReconstructionProcessor;

/* Call the constructor to initialize the private member values
 */
PhotodotReconstructionProcessor::PhotodotReconstructionProcessor() :
    Processor("PhotodotReconstructionProcessor"),
    /* Variables to be initialized:
     */
    _geometry_filename("photo_geometry.txt"), _input_pulses_collection_name(
        "TPCPulses"), _output_reconstuctedPhototdots_collection_name("ReconstructedPhotodots"),
        _output_photodot_collection_name("Photodots"), _pulses(0),
    /* TPC geometry parameters:
     */
    _diffusion(1.), _omega_tau(10.), _drift_velocity(40.), _tpc_outer_radius(
        1500.), _tpc_inner_radius(305.), _outer_z(100.), _inner_z(10.), _sigma(
        0.034692), _useGaus(false)

{

  /* Description of the processor which will be displayed in the steering
   * file automatically generated by Marlin.
   */
  _description =
      "PhotodotReconstructionProcessor will compare the location of the centre of a deposited electron cloud with the position of the photodot that generated it. Photodot positions are adjusted for the presence of a known magnetic field.";

  registerInputCollection(LCIO::TRACKERPULSE, "InputTrackerPulses",
      "Input collection carrying a deposited photelectric pulse",
      _input_pulses_collection_name, std::string("TPCPulses"));

  registerInputCollection(LCIO::LCGENERICOBJECT, "InputTPCElectronClouds",
      "Input collection carrying the ElectronClouds drifted through the TPC",
      _input_hitPosition_collection_name, std::string("TPCElectronClouds"));

  registerOutputCollection(LCIO::TRACKERHIT, "OutputReconstructedPhotodots",
      "Output collection of the reconstructed photodot positions.",
      _output_reconstuctedPhototdots_collection_name, std::string("TPCReconstructedPhotodots"));

  registerOutputCollection(LCIO::TRACKERHIT, "OutputPhotodots",
      "Output collection of the photodot positions.",
      _output_photodot_collection_name, std::string("TPCPhotodots"));

  registerOptionalParameter("GeometryFilename",
      "The name of the file containing the positions of photodots",
      _geometry_filename, std::string("photo_geometry.txt"));

  registerProcessorParameter("TransverseDiffusion",
      "Transverse diffusion of photodots along the drift distance.", _diffusion,
      50.);

  registerProcessorParameter("OmegaTau",
      "The gas parameterization for this section of the TPC", _omega_tau, 10.);

  registerProcessorParameter("DriftVelocity",
      "The parameterized drift velocity in this section of the TPC (mm/us)",
      _drift_velocity, 40.);

  registerProcessorParameter("TPCOuterRadius",
      "The outer radius of the TPC (mm)", _tpc_outer_radius, 1500.);

  registerProcessorParameter("TPCInnerRadius",
      "The inner radius of the TPC (mm)", _tpc_inner_radius, 300.);

  registerProcessorParameter("OuterZ",
      "The absolute value of the z coordinate of the outermost position of the TPC (mm)",
      _outer_z, 100.);

  registerProcessorParameter("InnerZ",
      "The absolute value of the z coordinate of the innermost position of the TPC (mm)",
      _inner_z, 10.);

  registerProcessorParameter("useGaus",
      "If true a gausian distribution on each row is assumed when searching for the hit centre. Thus the diffusion must be known. If false the charge weighted mean is claculated.",
      _useGaus, false);
  //Pad layout information is initialized to disallowed values to cover errors
//  _padLayoutType = -1;
//  _padLayout = NULL;

} // /Constructor

/* Destructor
 */
PhotodotReconstructionProcessor::~PhotodotReconstructionProcessor() {
}

void PhotodotReconstructionProcessor::init() {
  streamlog_out(DEBUG)<< "PhotodotReconstructionProcessor::init\n";

  // Usually a good idea to
  printParameters();

  _diffusion *= _sigma;

} // /init

void PhotodotReconstructionProcessor::processRunHeader(
    lcio::LCRunHeader *  /*run*/) {
  streamlog_out(DEBUG)<< "PhotodotReconstructionProcessor::processRunHeader()\n";

  /* Get the photodot geometry
   */
  TiXmlDocument document ( _geometry_filename.c_str() );

  if( !document.LoadFile() ) {

    streamlog_out( ERROR ) << "Couldn't load projected geometry file " << _geometry_filename << "." << std::endl;
    return;
  } else {
    streamlog_out( MESSAGE ) << "Loaded projected geometry file " << _geometry_filename << "." << std::endl;
  }
  // get a list of the input photoelectric dots
  TiXmlElement *photodot_element = document.RootElement()->FirstChildElement("photodot");
  if(photodot_element == 0){
    streamlog_out( WARNING ) << "Could not find photodot information in the given photodot file." << std::endl;
  }
  double diameter,x,y;
  for ( ; photodot_element; photodot_element = photodot_element->NextSiblingElement( "photodot" ) ) {
    photodot_element->QueryDoubleAttribute("diameter",&diameter);
    photodot_element->QueryDoubleAttribute("x",&x);
    photodot_element->QueryDoubleAttribute("y",&y);
    _photodots.push_back(Photodot(x, y, diameter));
    streamlog_out(DEBUG) << "Added a photodot at " << x << ", " << y << std::endl;
  }
} // /processRunHeader

void PhotodotReconstructionProcessor::processEvent(lcio::LCEvent * evt) {
  streamlog_out(DEBUG)<< "IdealPhotodotComparisonProcessor::processEvent\n";

  try
  {

    // Get the input collection
    LCCollection * input_pulses = evt->getCollection(_input_pulses_collection_name);

    // Initialize _pulsesByIndex
    _pulses = new std::multimap<gear::GlobalPadIndex, TrackerPulse * >;

    /* Copy each input pulse into the map
     */
    for (int iRawData = 0;
        iRawData < input_pulses->getNumberOfElements();
        iRawData++)
    {

      // Retrieve a pulse
      TrackerPulse * input_pulse = dynamic_cast< TrackerPulse * >
      (input_pulses->getElementAt(iRawData));

      if (input_pulse)// only proceed if the pulse is charged
      {

        // Insert the pulse into the map
        streamlog_out(DEBUG) << "Added pulse: " << input_pulse->getCellID0() << " on module " << input_pulse->getCellID1() << " with charge " << input_pulse->getCharge() << std::endl;
        gear::GlobalPadIndex pair(input_pulse->getCellID0(),input_pulse->getCellID1());
        _pulses->insert(std::make_pair(pair,input_pulse));

      } // / if valid input pulse

      else

      throw(EVENT::Exception("Input collection "
              + _input_pulses_collection_name
              + "does not contain TrackerPulses."));

    } // /for each input pulse

  } // /try getting input pulses

  catch (DataNotAvailableException &e)
  {

    streamlog_out(ERROR) << e.what() << std::endl;
    throw e;

  } // /catch getting input pulses

/* FIXME: to-do comment in this block below produces unused-variable-warning.
 * Please check and either delete or use the code!
 *
  // get the collection of expected hits
  LCCollection* electron_cloud_collection;
  try{
    electron_cloud_collection = evt->getCollection(_input_hitPosition_collection_name);
  } catch (EventException &exception) {
    streamlog_out(DEBUG4) << "Exception: " << exception.what() << std::endl;
    throw(EVENT::Exception("Input collection "
                  + _input_hitPosition_collection_name
                  + "was not found."));
  }

  // Determine which pulses are in a valid time range.
  //ToDo Check if this is needed and were to get the mean time from
  //findPhotodotPulses(electron_cloud_collection);
*/

  streamlog_out(MESSAGE) << "Processing " << _photodots.size()
  << " pulses." << std::endl;

  streamlog_out(MESSAGE) << "Processing photodots" << std::endl;
  LCCollectionVec *Photodots_collection = new LCCollectionVec( LCIO::TRACKERHIT );
  LCCollectionVec *recoPhotodots_collection = new LCCollectionVec( LCIO::TRACKERHIT );
  for (unsigned int i = 0; i < _photodots.size(); i++)
//  for (unsigned int i = 130; i < 131; i++)
  {
    streamlog_out(MESSAGE) << "Processing dot " << i << std::endl;
    // add expected photodot positions
    TrackerHitImpl *Photodot = new TrackerHitImpl();
    double real_pos[3] = {_photodots[i]._x,_photodots[i]._y,0};
    Photodot->setPosition(real_pos);
    Photodots_collection->addElement(Photodot);
    /* Find the nearest hit to the photodot.
     */
    int Module, closestPad;
    bool isOnModule = false;
    if(Global::GEAR->getTPCParameters().getCoordinateType() == PadRowLayout2D::CARTESIAN)
      isOnModule = Global::GEAR->getTPCParameters().isInsideModule(_photodots[i]._x, _photodots[i]._y);
    else if (Global::GEAR->getTPCParameters().getCoordinateType() == PadRowLayout2D::POLAR)
      isOnModule = Global::GEAR->getTPCParameters().isInsideModule(_photodots[i]._r, _photodots[i]._phi);
    else
      throw(gear::Exception("Unknown pad layout."));
    if(!isOnModule)
      continue;
    findNearestHit(_photodots[i], Module, closestPad);

    /* Find the central column of the nearest hit.
     */
    int hitCentreColumn = findCentreColumn(Module, closestPad);

    /* Find the geometric centre of the nearest hit
     * Its output is (x,y) or (r,phi) as appropriate
     */
    TrackerHitImpl *recoPhotodot = new TrackerHitImpl();
    double pos[3];
    if(_useGaus){
      Vector2D hitCentre = findGausHitCentre(Module, hitCentreColumn);
      pos[0] = hitCentre[0];
      pos[1] = hitCentre[1];
      pos[2] = 0;
    } else {
      float covMatrix[5];
      Vector3D hitCentre = findMeanHitCentre(Module, hitCentreColumn, covMatrix);
      pos[0] = hitCentre.x();
      pos[1] = hitCentre.y();
      pos[2] = hitCentre.z();
      recoPhotodot->setCovMatrix(covMatrix);
    }
    recoPhotodot->setPosition(pos);
    // time is taken from the pulse on the centre pad with the heighest charge
    std::multimap<gear::GlobalPadIndex , EVENT::TrackerPulse * >::iterator start, end , highestCharge;
    start = _pulses->equal_range(gear::GlobalPadIndex(hitCentreColumn,Module)).first;
    end = _pulses->equal_range(gear::GlobalPadIndex(hitCentreColumn,Module)).second;
    highestCharge = start;
    if(std::distance(start,end) > 1){
      streamlog_out(DEBUG4) << "Found more than one pulse on the centre pad. (" << std::distance(start,end) << ")." << std::endl;
      while(start != end){
        if(start->second->getCharge() > highestCharge->second->getCharge())
          highestCharge = start;
        start++;
      }
    }
    recoPhotodot->setTime(highestCharge->second->getTime());
    recoPhotodot->setType(i);
    recoPhotodots_collection->addElement(recoPhotodot);
  } // /for each photodot
  evt->addCollection(recoPhotodots_collection, _output_reconstuctedPhototdots_collection_name);
  evt->addCollection(Photodots_collection, _output_photodot_collection_name);
} // /processEvent()

void PhotodotReconstructionProcessor::check(lcio::LCEvent *  /*evt*/) {
}

void PhotodotReconstructionProcessor::end() {
}

void PhotodotReconstructionProcessor::findNearestHit(Photodot thisPhotodot, int &module, int &pad) {

  double distanceToPad = 999999999.9;


  /* Analyse each pulse for the closest one
   */

  for (std::multimap<gear::GlobalPadIndex, TrackerPulse *>::iterator iPulse =
      _pulses->begin(); iPulse != _pulses->end(); iPulse++) {
    //get module
    const gear::TPCModule & tpcModule = Global::GEAR->getTPCParameters().getModule(iPulse->second->getCellID1());
    // Get the centre of the pad being tested:
    Vector2D currentPad = tpcModule.getPadCenter(iPulse->second->getCellID0());
    // Distance to that pad
    double padDistance = 0.;

    /* Calculate the distance to that pad in the global coordinate system since dots are given in global coordinates
     */
    switch (tpcModule.getPadLayoutType()) {

    case PadRowLayout2D::POLAR:
      padDistance = thisPhotodot._r * thisPhotodot._r
          + currentPad[0] * currentPad[0]
          - 2 * thisPhotodot._r * currentPad[0]
              * cos(fabs(thisPhotodot._phi - currentPad[1]));
      break;

    case PadRowLayout2D::CARTESIAN:
      padDistance = (thisPhotodot._x - currentPad[0])
          * (thisPhotodot._x - currentPad[0])
          + (thisPhotodot._y - currentPad[1])
              * (thisPhotodot._y - currentPad[1]);
      break;
    } // /switch calculate pad distance

    if (padDistance < distanceToPad) {
      module = iPulse->second->getCellID1();
      pad = iPulse->second->getCellID0();
      distanceToPad = padDistance;
    } // /if change closest pulse

  } // /for each pulse

  return ;

} // /findNearestHit()

int PhotodotReconstructionProcessor::findCentreColumn(int module, int pad) {

  /* This loop looks at the existence of the adjacent pad (assigning the pad
   * to nextPulse if it exists), then checks if the charge on that pad exceeds
   * the charge on the current pad. If it does, the loop repeats and reassigns
   * nextPad to pad.
   * Only the given module will be used.
   */
  bool newPad = true;
  // get the module where the current pulse is on
  const gear::TPCModule & tpcModule = Global::GEAR->getTPCParameters().getModule(module);
  int nextPad = pad;
  // Use an iterator to limit the number of map searches
  std::multimap<gear::GlobalPadIndex, EVENT::TrackerPulse *>::iterator thisPulse;
  while (newPad) {

    // Update the IDs of the pulse being tested
    pad = nextPad;
    newPad = false;

    // Check the left pad
    try{
      nextPad = tpcModule.getLeftNeighbour(pad);
      thisPulse = _pulses->find(gear::GlobalPadIndex(nextPad, module));
      if( thisPulse != _pulses->end() &&
          (*thisPulse).second->getCharge() >
          _pulses->find(gear::GlobalPadIndex(pad, module))->second->getCharge())
        newPad = true;
    } catch(...) {
//      streamlog_out(DEBUG4) << e.what() << std::endl;
      streamlog_out(MESSAGE) << "No left pad found" << std::endl;
    }
    // Check the right pad
    if ( !newPad ){
      try{
        nextPad = tpcModule.getRightNeighbour(pad);
        thisPulse = _pulses->find(gear::GlobalPadIndex(nextPad, module));
        if( thisPulse != _pulses->end() &&
            (*thisPulse).second->getCharge() >
            _pulses->find(gear::GlobalPadIndex(pad, module))->second->getCharge())
          newPad = true;
//      } catch(gear::Exception &e) {
      } catch(...) {
//        streamlog_out(DEBUG4) << e.what() << std::endl;
        streamlog_out(MESSAGE) << "No right pad found" << std::endl;
      }
    }
    if( !newPad ) {
      Vector2D pos;
      switch (tpcModule.getLocalPadLayout().getPadLayoutType())
      {

      case PadRowLayout2D::POLAR:
        pos = tpcModule.getPadCenter(pad);
        pos = tpcModule.globalToLocal(pos[0],pos[1]);
        pos[0] -= tpcModule.getPadHeight(pad);
        pos = tpcModule.localToGlobal(pos[0],pos[1]);
        nextPad = tpcModule.getNearestPad(pos[0],pos[1]);
        if (_pulses->find(gear::GlobalPadIndex(nextPad = tpcModule.getNearestPad(pos[0], pos[1]), module)) != _pulses->end() && tpcModule.getRowNumber(nextPad) != tpcModule.getRowNumber(pad))
          newPad = true;


        break;

      case PadRowLayout2D::CARTESIAN:
        pos = tpcModule.getPadCenter(pad);
        pos = tpcModule.globalToLocal(pos[0],pos[1]);
        pos[1] -= tpcModule.getPadHeight(pad);
        pos = tpcModule.localToGlobal(pos[0],pos[1]);
        if (_pulses->find(gear::GlobalPadIndex(nextPad = tpcModule.getNearestPad(pos[0], pos[1]), module)) != _pulses->end())
            newPad = true;
        break;

      }// /switch pad type
    }
  } // /while searching for new pads

  // The current pad is now the bottom of the hit's centre column
  return pad;

} // /findCentreColumn()

Vector2D PhotodotReconstructionProcessor::findGausHitCentre(int module, int centreColumn) {
  const gear::TPCModule & tpcModule = Global::GEAR->getTPCParameters().getModule(module);
  // Centre of the row in the each case (to be calculated repeatedly)
  double phiHitCentre = 0;
  double xHitCentre = 0;

  // Total charge
  double totalCharge = 0;

  // Charge on the lower part of the hit
  double lowerCharge = 0;

  // Vector holding the charge of each row
  std::vector<double> chargeByRow;

  // Distance to the lower edge of the upper part of the hit
  double upperRowBase = 0;

  // cell ID of the next row to be tested
  int thisRowCentre = centreColumn;

  // temporary position used for switching between local and global coodinate system
  Vector2D pos;
  /* Find the sum of the charges in each row
   */
  bool foundNext = true;
  while (foundNext){
    if(_pulses->find(gear::GlobalPadIndex(thisRowCentre, module)) == _pulses->end())
      break;
    // Sum of the charges on each side
    double leftChargeSum, rightChargeSum;
    leftChargeSum = rightChargeSum = 0;
    streamlog_out(DEBUG) << "Pad " << thisRowCentre << "\t " << tpcModule.getPadCenter(thisRowCentre)[0] << ", " << tpcModule.getPadCenter(thisRowCentre)[1]
                         << "\t R-Phi: " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisRowCentre)[0],tpcModule.getPadCenter(thisRowCentre)[1])[0]
                         << ", " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisRowCentre)[0],tpcModule.getPadCenter(thisRowCentre)[1])[1]
                         << "\tCharge: " << _pulses->find(gear::GlobalPadIndex(thisRowCentre, module))->second->getCharge() << std::endl;
    /* sum the charges to the right of the centre column
     */
    int thisPad = 0;
    try{
      thisPad = tpcModule.getRightNeighbour(thisRowCentre);
    } catch(gear::Exception &e) {
      streamlog_out(DEBUG4) << e.what() << std::endl;
    }
    while (_pulses->find(gear::GlobalPadIndex(thisPad, module)) != _pulses->end()){
      streamlog_out(DEBUG) << "Right: Pad " << thisPad << "\t " << tpcModule.getPadCenter(thisPad)[0] << ", " << tpcModule.getPadCenter(thisPad)[1]
                           << "\t R-Phi: " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisPad)[0],tpcModule.getPadCenter(thisPad)[1])[0]
                           << ", " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisPad)[0],tpcModule.getPadCenter(thisPad)[1])[1]
                           << "\tCharge: " << _pulses->find(gear::GlobalPadIndex(thisPad, module))->second->getCharge() << std::endl;
      rightChargeSum += getPadCharge(module,thisPad);
      try{
        thisPad = tpcModule.getRightNeighbour(thisPad);
      } catch(gear::Exception &e) {
        // stop if no right pad is found any more
        streamlog_out(DEBUG4) << e.what() << std::endl;
        break;
      }

    }

    /* sum the charges to the left of the centre column
     */
    thisPad = thisRowCentre;
    while (_pulses->find(gear::GlobalPadIndex(thisPad, module)) != _pulses->end()){
      streamlog_out(DEBUG) << "Left: Pad " << thisPad << "\t " << tpcModule.getPadCenter(thisPad)[0] << ", " << tpcModule.getPadCenter(thisPad)[1]
                           << "\t R-Phi: " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisPad)[0],tpcModule.getPadCenter(thisPad)[1])[0]
                           << ", " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisPad)[0],tpcModule.getPadCenter(thisPad)[1])[1]
                           << "\t Charge: " << _pulses->find(gear::GlobalPadIndex(thisPad, module))->second->getCharge() << std::endl;
      leftChargeSum += getPadCharge(module,thisPad);
      try{
        thisPad = tpcModule.getLeftNeighbour(thisPad);
      } catch(gear::Exception &e) {
        // stop if no right pad is found any more
        streamlog_out(DEBUG4) << e.what() << std::endl;
        break;
      }

    }

    chargeByRow.push_back(leftChargeSum + rightChargeSum);
    totalCharge += leftChargeSum + rightChargeSum;
    bool failedMoveUp = false;
    int tmpPad;
    /* Find the next row, and other operations as appropriate
     */
    switch (tpcModule.getLocalPadLayout().getPadLayoutType()) {

    case PadRowLayout2D::POLAR:
      // get global position
      pos = tpcModule.getPadCenter(thisRowCentre);
      // switch to local position
      pos = tpcModule.globalToLocal(pos[0],pos[1]);
      /* First, find the row's hit centre in a polar layout
       */
      streamlog_out(DEBUG) << "Pad width is " << tpcModule.getPadWidth(thisRowCentre) << " r is " << pos[0] << " and phi is " << pos[1] << std::endl;
      streamlog_out(DEBUG) << "Charge on the central pad is: " << _pulses->find(gear::GlobalPadIndex(thisRowCentre, module))->second->getCharge()
                           << "\nLeft charge is: " << leftChargeSum
                           << "\nRight charge is: " << rightChargeSum
                           << "\nDiffusion is set to: " << _diffusion << std::endl;
      streamlog_out(DEBUG) << "phiCentre is: " << (pos[1] + 0.5 * tpcModule.getPadWidth(thisRowCentre) - std::sqrt(2.) * tpcModule.getPadWidth(thisRowCentre)* _diffusion
          * TMath::ErfInverse(2 * leftChargeSum / (leftChargeSum + rightChargeSum) - 1.0)) << std::endl;
      phiHitCentre +=
      // Get the upper side of this pad (boundary between left and right)
          (pos[1]
              + 0.5 * tpcModule.getPadWidth(thisRowCentre) -
          // Get the charge centre for this row
              std::sqrt(2.) * _diffusion
              // translate diffusion in mm into diffusion in rad
              * tpcModule.getPadWidth(thisRowCentre) / (2 * pos[0] * TMath::Sin(tpcModule.getPadWidth(thisRowCentre)/2.))
                  * TMath::ErfInverse(
                      2 * leftChargeSum / (leftChargeSum + rightChargeSum)
                          - 1.0))
          // Prepare the weighted mean
              * (leftChargeSum + rightChargeSum);

      /* Now find the next row
       */
      pos[0] += tpcModule.getPadHeight(thisRowCentre);
      pos = tpcModule.localToGlobal(pos[0],pos[1]);
      streamlog_out(DEBUG) << "Before moving the row pad index is: " << thisRowCentre << "\t in row " << tpcModule.getRowNumber(thisRowCentre) << std::endl;
      tmpPad = tpcModule.getNearestPad(pos[0],pos[1]);
      if(tpcModule.getRowNumber(tmpPad) == tpcModule.getRowNumber(thisRowCentre)){
        streamlog_out(WARNING) << "Moving failed for pad " << tmpPad << std::endl;
        failedMoveUp = true;
      }

      else{
        thisRowCentre = tmpPad;
        streamlog_out(DEBUG) << "After moving the row pad index is: " << thisRowCentre << "\t in row " << tpcModule.getRowNumber(thisRowCentre) <<std::endl;
      }
      break;

    case PadRowLayout2D::CARTESIAN:
      //TODO: Go to the local coordinate system and check in which direction rows are defined
      /* Find the row's hit centre
       */
      xHitCentre +=
      // Get the upper edge of this pad)
          (tpcModule.getPadCenter(thisRowCentre)[0]
              + 0.5 * tpcModule.getPadWidth(thisRowCentre) -
          // Get the charge centre for this row
              std::sqrt(2.) * _diffusion
                  * TMath::ErfInverse(
                      2 * leftChargeSum / (leftChargeSum + rightChargeSum)
                          - 1.0))
          // Prepare the weighted mean
              * (leftChargeSum + rightChargeSum);

      /* Now find the next row
       */
      thisRowCentre = tpcModule.getNearestPad(
          (tpcModule.getPadCenter(thisRowCentre))[0],
          (tpcModule.getPadCenter(thisRowCentre))[1]
              + tpcModule.getPadHeight(thisRowCentre));
      break;

    } // /switch

    // don't repeat if already on the topmost row
    if (failedMoveUp){
      foundNext = false;
    // check if there is a pulse in thisRowCentre (which is now on the upper row)
    // if not check if there is a pulse next or next to next to the pad in the upper row
    // this is not possible if the row index is 0 because than there was no upper row...
    } else if (_pulses->find(gear::GlobalPadIndex(thisRowCentre, module)) == _pulses->end()){
      // get the pads in the next row
      foundNext = false;
      try{
        if(_pulses->find(gear::GlobalPadIndex(tpcModule.getLeftNeighbour(thisRowCentre), module)) != _pulses->end()){
            thisRowCentre = tpcModule.getLeftNeighbour(thisRowCentre);
            foundNext = true;
        }
      } catch (gear::Exception &e){
        streamlog_out(MESSAGE) << "No left pad found in charge collection. (" << e.what() << ")." << std::endl;
      }
      try{
        if(!foundNext && _pulses->find(gear::GlobalPadIndex(tpcModule.getRightNeighbour(thisRowCentre), module)) != _pulses->end()){
            thisRowCentre = tpcModule.getLeftNeighbour(thisRowCentre);
            foundNext = true;
        }
      } catch (gear::Exception &e){
        streamlog_out(MESSAGE) << "No right pad found in charge collection. (" << e.what() << ")." << std::endl;
      }
      try{
        if(!foundNext && _pulses->find(gear::GlobalPadIndex(tpcModule.getLeftNeighbour(tpcModule.getLeftNeighbour(thisRowCentre)), module)) != _pulses->end()){
            thisRowCentre = tpcModule.getLeftNeighbour(thisRowCentre);
            foundNext = true;
        }
      } catch (gear::Exception &e){
        streamlog_out(MESSAGE) << "No next to left pad found in charge collection. (" << e.what() << ")." << std::endl;
      }
      try{
        if(!foundNext && _pulses->find(gear::GlobalPadIndex(tpcModule.getRightNeighbour(tpcModule.getRightNeighbour(thisRowCentre)), module)) != _pulses->end()){
            thisRowCentre = tpcModule.getLeftNeighbour(thisRowCentre);
            foundNext = true;
        }
      } catch (gear::Exception &e){
        streamlog_out(MESSAGE) << "No next to right pad found in charge collection. (" << e.what() << ")." << std::endl;
      }
    }

  } // do sum the charges on a line

  std::vector<double>::iterator iCharge;

  /*
   * Calculate the centre of the hit
   */
  Vector2D hitCentre;

  switch (tpcModule.getLocalPadLayout().getPadLayoutType()) {

  case PadRowLayout2D::POLAR:

    /* Calculate the weighted mean of the phi-positions found for the rows
     */
    phiHitCentre /= totalCharge;

    /* Columns:
     */
    iCharge = chargeByRow.begin();

    double rHitCentre;
    // get global position
    pos = tpcModule.getPadCenter(centreColumn);
    // switch to local position
    pos = tpcModule.globalToLocal(pos[0],pos[1]);

    /*
     * If there is only one row put the hit in the centre.
     * Since lowerCharge in this case is 0 the position set here (upperRowBase)
     * is the final position.
     * Else put the hit in the middle and than do charge weighting.
     */
    if (chargeByRow.size() > 2){
      lowerCharge = *(iCharge+1);
      upperRowBase = pos[0] + 1.5 * tpcModule.getPadHeight(centreColumn);
    } else if (chargeByRow.size() > 1){
      lowerCharge = *iCharge;
      upperRowBase = pos[0] + 0.5 * tpcModule.getPadHeight(centreColumn);
    }else {
      upperRowBase = pos[0];
    }

    rHitCentre =
    // Get the lower edge of the upper half of the hit
        (upperRowBase -
        // Get the radial charge centre (radial diffusion must be found)
            std::sqrt(2.) * _diffusion
                * TMath::ErfInverse(2 * lowerCharge / totalCharge - 1.0));
    // switch back to global position
    pos = tpcModule.localToGlobal(rHitCentre,phiHitCentre);
    hitCentre = Vector2D(pos[0], pos[1]);

    break;

    /* Cartesian case
     */
  case PadRowLayout2D::CARTESIAN:
    //TODO: Go to the local coordinate system and check in which direction rows are defined
    /* Calculate the weighted mean of the x-positions for the rows
     */
    xHitCentre /= totalCharge;

    /* Columns:
     */
    iCharge = chargeByRow.begin();

    double yHitCentre;

    /* Make sure there is sufficient charge on the bottom row for an accurate
     * result.
     */
    if ((chargeByRow.size() > 2)
        && ((lowerCharge += *iCharge) / totalCharge < 0.01)) {
      // central row
      ++iCharge;
      lowerCharge += *iCharge;
      upperRowBase = tpcModule.getPadCenter(centreColumn)[1]
          + 1.5 * tpcModule.getPadHeight(centreColumn);

    } // /if the charge on the bottom row is too small
    else

      upperRowBase = tpcModule.getPadCenter(centreColumn)[1]
          + 0.5 * tpcModule.getPadHeight(centreColumn);

    yHitCentre =
    // Get the lower edge of the upper half of the hit
        (upperRowBase -
        // Get the y charge centre
            std::sqrt(2.) * _diffusion
                * TMath::ErfInverse(2 * lowerCharge / totalCharge - 1.0));

    hitCentre = Vector2D(xHitCentre, yHitCentre);

    break;

  } // /switch pad layout type

  return hitCentre;

} // /findGausHitCentre()

Vector3D PhotodotReconstructionProcessor::findMeanHitCentre(int module, int centreColumn,   float* covMatrix) {
  const gear::TPCModule & tpcModule = Global::GEAR->getTPCParameters().getModule(module);

  // Vector holding the pad positions and the corresponding charge
  HitCandidate thisHit(Global::GEAR->getTPCParameters(), _drift_velocity);

  // cell ID of the next row to be tested
  int thisRowCentre = centreColumn;

  // temporary position used for switching between local and global coodinate system
  Vector2D pos;
  /* Find the sum of the charges in each row
   */
  bool foundNext = true;
  while (foundNext){
    if(_pulses->find(gear::GlobalPadIndex(thisRowCentre, module)) == _pulses->end())
      break;


    streamlog_out(DEBUG) << "Pad " << thisRowCentre << "\t " << tpcModule.getPadCenter(thisRowCentre)[0] << ", " << tpcModule.getPadCenter(thisRowCentre)[1]
                         << "\t R-Phi: " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisRowCentre)[0],tpcModule.getPadCenter(thisRowCentre)[1])[0]
                         << ", " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisRowCentre)[0],tpcModule.getPadCenter(thisRowCentre)[1])[1]
                         << "\tCharge: " << _pulses->find(gear::GlobalPadIndex(thisRowCentre, module))->second->getCharge() << std::endl;
    /* sum the charges to the left of the centre column
     */
    int thisPad = thisRowCentre;
    while (_pulses->find(gear::GlobalPadIndex(thisPad, module)) != _pulses->end()){
      streamlog_out(DEBUG) << "Left: Pad " << thisPad << "\t " << tpcModule.getPadCenter(thisPad)[0] << ", " << tpcModule.getPadCenter(thisPad)[1]
                           << "\t R-Phi: " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisPad)[0],tpcModule.getPadCenter(thisPad)[1])[0]
                           << ", " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisPad)[0],tpcModule.getPadCenter(thisPad)[1])[1]
                           << "\t Charge: " << _pulses->find(gear::GlobalPadIndex(thisPad, module))->second->getCharge() << std::endl;
      // add all pulses on the centre pad
      for (std::multimap< gear::GlobalPadIndex , EVENT::TrackerPulse * >::iterator iPulse =
              _pulses->lower_bound(gear::GlobalPadIndex(thisPad, module));
              iPulse != _pulses->upper_bound(gear::GlobalPadIndex(thisPad, module));
              iPulse++) {
        thisHit.addPulse(iPulse->second);
      }
      try{
        thisPad = tpcModule.getLeftNeighbour(thisPad);
      } catch(gear::Exception &e) {
        // stop if no right pad is found any more
        streamlog_out(DEBUG4) << e.what() << std::endl;
        break;
      }

    }
    /* sum the charges to the right of the centre column
     */
    try{
      thisPad = tpcModule.getRightNeighbour(thisRowCentre);
    } catch(gear::Exception &e) {
      streamlog_out(DEBUG4) << e.what() << std::endl;
    }
    while (_pulses->find(gear::GlobalPadIndex(thisPad, module)) != _pulses->end()){
      streamlog_out(DEBUG) << "Right: Pad " << thisPad << "\t " << tpcModule.getPadCenter(thisPad)[0] << ", " << tpcModule.getPadCenter(thisPad)[1]
                           << "\t R-Phi: " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisPad)[0],tpcModule.getPadCenter(thisPad)[1])[0]
                           << ", " << tpcModule.globalToLocal(tpcModule.getPadCenter(thisPad)[0],tpcModule.getPadCenter(thisPad)[1])[1]
                           << "\tCharge: " << _pulses->find(gear::GlobalPadIndex(thisPad, module))->second->getCharge() << std::endl;
      for (std::multimap< gear::GlobalPadIndex , EVENT::TrackerPulse * >::iterator iPulse =
              _pulses->lower_bound(gear::GlobalPadIndex(thisPad, module));
              iPulse != _pulses->upper_bound(gear::GlobalPadIndex(thisPad, module));
              iPulse++) {
        thisHit.addPulse(iPulse->second);
      }
      try{
        thisPad = tpcModule.getRightNeighbour(thisPad);
      } catch(gear::Exception &e) {
        // stop if no right pad is found any more
        streamlog_out(DEBUG4) << e.what() << std::endl;
        break;
      }

    }
    /* Find the next row, and other operations as appropriate
     */
    bool failedMoveUp = false;
    int tmpPad;
    switch (tpcModule.getLocalPadLayout().getPadLayoutType()) {

    case PadRowLayout2D::POLAR:
      // get global position
      pos = tpcModule.getPadCenter(thisRowCentre);
      // switch to local position
      pos = tpcModule.globalToLocal(pos[0],pos[1]);
      /* First, find the row's hit centre in a polar layout
       */
      streamlog_out(DEBUG) << "Pad width is " << tpcModule.getPadWidth(thisRowCentre) << " r is " << pos[0] << " and phi is " << pos[1] << std::endl;
      streamlog_out(DEBUG) << "Charge on the central pad is: " << _pulses->find(gear::GlobalPadIndex(thisRowCentre, module))->second->getCharge()
                           << "\nSo far the cluster consists of: " << thisHit.getWidth() << " pulses (not necessary the number of pads!)." << std::endl;
      /* Now find the next row
       */
      pos[0] += tpcModule.getPadHeight(thisRowCentre);
      pos = tpcModule.localToGlobal(pos[0],pos[1]);
      streamlog_out(DEBUG) << "Before moving the row pad index is: " << thisRowCentre << "\t in row " << tpcModule.getRowNumber(thisRowCentre) << std::endl;
      tmpPad = tpcModule.getNearestPad(pos[0],pos[1]);
      if(tpcModule.getRowNumber(tmpPad) == tpcModule.getRowNumber(thisRowCentre)){
        streamlog_out(WARNING) << "Moving failed for pad " << tmpPad << std::endl;
        failedMoveUp = true;
      } else {
        thisRowCentre = tmpPad;
        streamlog_out(DEBUG) << "After moving the row pad index is: " << thisRowCentre << "\t in row " << tpcModule.getRowNumber(thisRowCentre) <<std::endl;
      }
      break;

    case PadRowLayout2D::CARTESIAN:
      //TODO: Check in which direction rows are defined
      /* Now find the next row
       */
      // get global position
      pos = tpcModule.getPadCenter(thisRowCentre);
      // switch to local position
      pos = tpcModule.globalToLocal(pos[0],pos[1]);
      pos[1] += tpcModule.getPadHeight(thisRowCentre);
      tmpPad = tpcModule.getNearestPad(pos[0],pos[1]);
      if(tpcModule.getRowNumber(tmpPad) == tpcModule.getRowNumber(thisRowCentre)){
        streamlog_out(WARNING) << "Moving failed for pad " << tmpPad << std::endl;
        failedMoveUp = true;
      } else {
        thisRowCentre = tmpPad;
        streamlog_out(DEBUG) << "After moving the row pad index is: " << thisRowCentre << "\t in row " << tpcModule.getRowNumber(thisRowCentre) <<std::endl;
      }

      break;

    } // /switch

    // don't repeat if already on the topmost row
    if (failedMoveUp){
      foundNext = false;
    // check if there is a pulse in thisRowCentre (which is now on the upper row)
    // if not check if there is a pulse next or next to next to the pad in the upper row
    // this is not possible if the row index is 0 because than there was no upper row...
    } else if (_pulses->find(gear::GlobalPadIndex(thisRowCentre, module)) == _pulses->end()){
      // get the pads in the next row
      foundNext = false;
      try{
        if(_pulses->find(gear::GlobalPadIndex(tpcModule.getLeftNeighbour(thisRowCentre), module)) != _pulses->end()){
            thisRowCentre = tpcModule.getLeftNeighbour(thisRowCentre);
            foundNext = true;
        }
      } catch (gear::Exception &e){
        streamlog_out(MESSAGE) << "No left pad found in charge collection. (" << e.what() << ")." << std::endl;
      }
      try{
        if(!foundNext && _pulses->find(gear::GlobalPadIndex(tpcModule.getRightNeighbour(thisRowCentre), module)) != _pulses->end()){
            thisRowCentre = tpcModule.getLeftNeighbour(thisRowCentre);
            foundNext = true;
        }
      } catch (gear::Exception &e){
        streamlog_out(MESSAGE) << "No right pad found in charge collection. (" << e.what() << ")." << std::endl;
      }
      try{
        if(!foundNext && _pulses->find(gear::GlobalPadIndex(tpcModule.getLeftNeighbour(tpcModule.getLeftNeighbour(thisRowCentre)), module)) != _pulses->end()){
            thisRowCentre = tpcModule.getLeftNeighbour(thisRowCentre);
            foundNext = true;
        }
      } catch (gear::Exception &e){
        streamlog_out(MESSAGE) << "No next to left pad found in charge collection. (" << e.what() << ")." << std::endl;
      }
      try{
        if(!foundNext && _pulses->find(gear::GlobalPadIndex(tpcModule.getRightNeighbour(tpcModule.getRightNeighbour(thisRowCentre)), module)) != _pulses->end()){
            thisRowCentre = tpcModule.getLeftNeighbour(thisRowCentre);
            foundNext = true;
        }
      } catch (gear::Exception &e){
        streamlog_out(MESSAGE) << "No next to right pad found in charge collection. (" << e.what() << ")." << std::endl;
      }
    }

  } // do sum the charges on a line

  /*
   * Calculate the centre of the hit
   * \remark This can also be done by the HitCandidate, but in this no uncertainty
   * is calculated and the hit is forced on the centre of a row!
   */
  Vector3D hitCentre = calculatePositionsAndErrors(thisHit, covMatrix);

  return hitCentre;

} // /findMeanHitCentre()

Vector3D PhotodotReconstructionProcessor::calculatePositionsAndErrors(const HitCandidate& cand, float* covMatrix){
  // calculate the center of gravity for the measurement coordinate
  const TrackerPulse* maxPulse              = cand.getMaximumPulse();
  const gear::TPCModule& module             = Global::GEAR->getTPCParameters().getModule(maxPulse->getCellID1());
  const gear::PadRowLayout2D& localPadPlane = module.getLocalPadLayout();

  // check whether the module geometry is described in cartesian coordinates
  // if not, then throw an exception; this isn't in the manual -- or: the convention is to use cartesian coordinates in LCIO and the global geometry
  if(module.getCoordinateType() != PadRowLayout2D::CARTESIAN)
    throw EVENT::Exception("The module does not return cartesian coordinates; this is against the general convention. Better fix it and catch the exception.");

  // now check the coordinate system type of the pad plane; this defines the natural system for all calculations
  // define two bools for easier reading
  bool cartesianPadPlane = (localPadPlane.getCoordinateType()  == PadRowLayout2D::CARTESIAN);
  bool polarPadPlane = !cartesianPadPlane;

  /*** NOTE: if the pad plane is polar, then the globalToLocal transformation also changes the coordinate type in GEAR !!! ***/

  std::vector<TrackerPulse*> thePulses = cand.getPulses();

  double chargeInPulse  = 0.;
  double chargeErrorInPulse  = 0.;
  double totalCharge    = 0.;

  double localPosition[2]  = { 0., 0.};
  double localPositionError[2]  = { 0., 0.};

  Vector2D localPadCenter, globalPadCenter;

  for(std::vector<TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse)
  {
    globalPadCenter = module.getPadCenter((*aPulse)->getCellID0());
    localPadCenter  = module.globalToLocal(globalPadCenter[0], globalPadCenter[1]);
    chargeInPulse = (*aPulse)->getCharge();
    totalCharge += chargeInPulse;
    for(unsigned int i = 0; i < 2; ++i)
    {
      localPosition[i] += chargeInPulse * localPadCenter[i];
    }
  }

  Vector2D globalPosition2D =  module.localToGlobal(localPosition[0] / totalCharge, localPosition[1] / totalCharge);
  double xPosition = globalPosition2D[0];
  double yPosition = globalPosition2D[1];

  //  coordinate system: z=0 at cathode, for large prototype the anode is in negative z-direction
  //  z-position is therefore: zAnode (max drift length, positive) - driftTime * driftVelocity
  double zPosition = Global::GEAR->getTPCParameters().getMaxDriftLength()
                                       -  _drift_velocity * (maxPulse->getTime()) / 1000. ; // time is in [ns], drift velocity in [mm/mus]

  // error calculation :: distinguish two cases!
  // as the variance is the square of the standard deviation, one needs to be careful with the squaring
  covMatrix[5]  =  pow(_drift_velocity,2) * (maxPulse->getCovMatrix())[2] / 1000000.; // error propagation from maximum pulse
  double tempDebug1 =0.0;
  double tempDebug2 =0.0;
  if(polarPadPlane)
  {
    for(std::vector<TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse)
    {
      globalPadCenter = module.getPadCenter((*aPulse)->getCellID0());
      localPadCenter  = module.globalToLocal(globalPadCenter[0], globalPadCenter[1]);
      chargeInPulse = (*aPulse)->getCharge();
      chargeErrorInPulse = ((*aPulse)->getCovMatrix())[0];
      streamlog_out(DEBUG4) << "chargeErrorInPulse =" << chargeErrorInPulse << std::endl;
      localPositionError[1] += pow(chargeInPulse*module.getPadWidth((*aPulse)->getCellID0())/totalCharge,2)/12
          +pow((localPadCenter[1]*totalCharge - localPosition[1]) / (totalCharge*totalCharge),2)*chargeErrorInPulse;
      tempDebug1+=pow(chargeInPulse*module.getPadWidth((*aPulse)->getCellID0())/totalCharge,2) / 12;
      tempDebug2+=pow((localPadCenter[1]*totalCharge - localPosition[1]) / (totalCharge*totalCharge),2)*chargeErrorInPulse;

    }
    //  localPositionError[1]=tempDebug2;
    streamlog_out(DEBUG4) << "totalCharge =" << totalCharge << std::endl;
    streamlog_out(DEBUG4) << "localPosition[0]/totalCharge =" << localPosition[0]/totalCharge << std::endl;
    streamlog_out(DEBUG4) << "localPosition[1]/totalCharge =" << localPosition[1]/totalCharge << std::endl;
    streamlog_out(DEBUG4) << "error from pad width =" << sqrt(tempDebug1) << std::endl;
    streamlog_out(DEBUG4) << "error from charge=" << sqrt(tempDebug2) << std::endl;
    double meanPhi = localPosition[1] / totalCharge;
    streamlog_out(DEBUG4) << "meanPhi="<< meanPhi<< std::endl;
    double meanRadius = localPosition[0] / totalCharge;
    double radiusErrorSquare = pow(module.getRowHeight(module.getRowNumber(maxPulse->getCellID0())), 2) / 12.;
    covMatrix[0] = pow(cos(meanPhi), 2) * radiusErrorSquare + pow(meanRadius, 2) * pow(sin(meanPhi), 2) * localPositionError[1];
    covMatrix[1] = sin(meanPhi) * cos(meanPhi) * (radiusErrorSquare - pow(meanRadius, 2) * localPositionError[1]);
    covMatrix[2] = pow(sin(meanPhi), 2) * radiusErrorSquare + pow(meanRadius, 2) * pow(cos(meanPhi), 2) * localPositionError[1];

  }
  else if(cartesianPadPlane)
  {
    for(std::vector<TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse)
    {
      globalPadCenter = module.getPadCenter((*aPulse)->getCellID0());
      localPadCenter  = module.globalToLocal(globalPadCenter[0], globalPadCenter[1]);
      chargeInPulse = (*aPulse)->getCharge();
      chargeErrorInPulse = ((*aPulse)->getCovMatrix())[0];
      localPositionError[0] += pow(chargeInPulse*module.getPadWidth((*aPulse)->getCellID0())/totalCharge,2)/12
          +pow((localPadCenter[0]*totalCharge-localPosition[0])/(totalCharge*totalCharge)*chargeErrorInPulse,2);

    }
    covMatrix[0] = localPositionError[0];
    covMatrix[2] = pow(module.getRowHeight(module.getRowNumber(maxPulse->getCellID0())),2) / 12.;
  }

  return Vector3D(xPosition, yPosition, zPosition);
}

void PhotodotReconstructionProcessor::findPhotodotPulses(LCCollection* electron_cloud_collection) {

  /* The upper and lower limits of the times to be considered to be valid
   * pulses.
   * These values are set in case the more complex method defined below
   * doesn't work -- this will happen if the photopulses were not created at
   * time 0.
   */
  double arrivalTime_low = 64100.;
  double arrivalTime_high = 64400.;

  /* Valid arrival times are calculated from the expected drift times for
   * each photodot.
   */
  double currentLowTime = 999999999.9;
  double currentHighTime = 0.;
  double meanTime = 0.;
  int nClouds = electron_cloud_collection->getNumberOfElements();
  for( int cloud_number = 0; cloud_number < nClouds; cloud_number++ ) {
    ElectronCloud *current_cloud = new ElectronCloud(dynamic_cast<LCGenericObject*>( electron_cloud_collection->getElementAt( cloud_number ) ));
    assert(current_cloud);
    currentLowTime = (current_cloud->get_time()*1000)< currentLowTime ?
        current_cloud->get_time() : currentLowTime;
    currentHighTime = (current_cloud->get_time()*1000) > currentHighTime ?
        current_cloud->get_time() : currentHighTime;
    meanTime += current_cloud->get_time() / nClouds;
    // diameter is set to zero and not needed in the end
    _expected_photodots.push_back(
            Photodot(current_cloud->get_x_position(), current_cloud->get_y_position(),0));
    delete current_cloud;
  }

  /* The mean arrival time for expected photodots is for some reason 70ns
   * over the arrival time for images.
   */
//  meanTime -= 0.070;
  /* 50ns added to give a wider range to search over, to ensure all pulses
   * are covered.
   */
  double timeRange = currentHighTime - currentLowTime + 0.05;

  /* It *appears* that the most prevalent pulse time in the event occurs at
   * 2/5 of the distribution.
   */
  arrivalTime_low = meanTime - 0.4 * timeRange;
  arrivalTime_high = meanTime + 0.6 * timeRange;
  /* Check that there are no pulses outside the desired range and that no
   * pad gets multiple pulses
   */
  for (std::multimap<gear::GlobalPadIndex , EVENT::TrackerPulse *>::iterator iPulse =
      _pulses->begin(); iPulse != _pulses->end(); iPulse++) {

    if (iPulse->second->getTime() < arrivalTime_low * 1.E3
        || iPulse->second->getTime() > arrivalTime_high * 1.E3) {

      streamlog_out(WARNING)<< "Pulse at module " << iPulse->first.getModuleID()
      << " and pad " << iPulse->first.getPadIndex()
      << " deleted due to invalid time ("
      << iPulse->second->getTime() << ")"
      << std::endl;

      _pulses->erase(iPulse);

    } // /if the pulse's time is invalid
  } // /for each pulse

} // /indPhotodotPulses()

double PhotodotReconstructionProcessor::getPadCharge(int module, int padIndex) {

  double totalCharge = 0;

  /* The summing only needs to occur if there exists more than one pad
   * with a certain index
   */
  if (_pulses->count(gear::GlobalPadIndex(padIndex, module)) > 1) {

    streamlog_out(WARNING)<< "Pulse on pad #" << padIndex
    << " is a duplicate." << std::endl;

    for (std::multimap< gear::GlobalPadIndex , EVENT::TrackerPulse * >::iterator iPulse =
        _pulses->lower_bound(gear::GlobalPadIndex(padIndex, module));
        iPulse != _pulses->upper_bound(gear::GlobalPadIndex(padIndex, module));
        iPulse++) {
      totalCharge += iPulse->second->getCharge();
    }

  } // if multiple pulses exist at one index
  else {
    std::multimap<gear::GlobalPadIndex, EVENT::TrackerPulse * >::iterator it =  _pulses->find(gear::GlobalPadIndex(padIndex, module));
    if(it != _pulses->end()){
      streamlog_out(DEBUG) << "Found pad " << padIndex << " on module " << module << std::endl;
      totalCharge = _pulses->find(gear::GlobalPadIndex(padIndex, module))->second->getCharge();
    } else
      streamlog_out(WARNING) << "Could not find pad " << padIndex << " on module " << module << std::endl;
  }
  return totalCharge;

} // findPhotodotPulses()

} // /namespace marlintpc
