#ifdef USE_LCCD

// MarlinTPC
#include "PhotoelectricLikelihoodFittingProcessor.h"
#include "PhotoelectricClusterTools.h"

// Marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>

// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

using namespace CLHEP;
using namespace EVENT;
using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace std;

namespace marlintpc
{


PhotoelectricLikelihoodFittingProcessor aPhotoelectricLikelihoodFittingProcessor;


PhotoelectricLikelihoodFittingProcessor::PhotoelectricLikelihoodFittingProcessor() : Processor("PhotoelectricLikelihoodFittingProcessor"),

        _input_seed_photodot_collection_name("TPCSeedPhotodots"),
        _input_seed_photoline_collection_name("TPCSeedPhotolines"),
        _output_photodot_collection_name("TPCPhotodots"),
        _output_photoline_collection_name("TPCPhotolines"),
        _output_is_transient(true),
        _sigma(2.5)
{

    /* description of the processor which will be displayed in the
    * steering file autometicly generated by
    * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
    */

    _description = "PhotoelectricLikelihoodFittingProcessor performs a likelihood fit on photoelectric data.";

    /* register steering parameters: name, description, class-variable, default value
    * the type will by definded by the type of the default value
    * string, double, float and int are possible
    * plaese note to get a float you must cast in C++: (float) 1.3
    */

    registerInputCollection(LCIO::TRACKERHIT,
                            "InputSeedPhotodots",
                            "The name of the input collection of TrackerHits which represent seed photodots.",
                            _input_seed_photodot_collection_name,
                            _input_seed_photodot_collection_name);

    registerInputCollection(LCIO::TRACK ,
                            "InputSeedPhotolines",
                            "The name of the input collection of Tracks which represent seed photolines.",
                            _input_seed_photoline_collection_name,
                            _input_seed_photoline_collection_name);

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OutputPhotodots",
                             "The name of the output collection of TPCHits which represent fitted photodots.",
                             _output_photodot_collection_name,
                             _output_photodot_collection_name);

    registerOutputCollection(LCIO::TRACK,
                             "OutputPhotolines",
                             "The name of the output collection of TPCHits which represent fitted photolines.",
                             _output_photoline_collection_name,
                             _output_photoline_collection_name);

    registerOptionalParameter("SetOutputTransient",
                              "Controls whether the output collection should be saved in the LCIO file (false) or if it should be deleted at the end (true - default).",
                              _output_is_transient,
                              true);

    registerOptionalParameter("Sigma",
                              "The expected diffusion of photodot charge due to the drift distance and initial defocussing.",
                              _sigma,
                              _sigma);

}


void PhotoelectricLikelihoodFittingProcessor::init()
{

    // usually a good idea to
    printParameters();

    // load AIDA (if it's available)
    try
        {

            _photodot_fit_results_tuple = AIDAProcessor::tupleFactory(this)->create("PhotodotFitResults", "Photodot Fit Results", "int i_dot, float x, float y, int n_pads, bool fit_status", "");

        }
    catch (lcio::Exception &e)
        {

            streamlog_out(DEBUG) << e.what() << endl;
            return;
        }

}


void PhotoelectricLikelihoodFittingProcessor::processRunHeader(lcio::LCRunHeader*  /*run*/)
{

}


void PhotoelectricLikelihoodFittingProcessor::processEvent(lcio::LCEvent * evt)
{

    // the input seed photodots
    LCCollection *seed_photodot_collection = evt->getCollection(_input_seed_photodot_collection_name);
    const int seed_photodot_count = seed_photodot_collection->getNumberOfElements();

    // the output fit photodots
    LCCollectionVec* output_fit_photodots = new LCCollectionVec(LCIO::TRACKERHIT);
    output_fit_photodots->setTransient(_output_is_transient);
    output_fit_photodots->setSubset(false);

    // loop through the seed photodots, fitting them
    for (int i_dot = 0; i_dot < seed_photodot_count; i_dot++)
        {

            TrackerHit *seed_photodot = dynamic_cast< TrackerHit* >(seed_photodot_collection->getElementAt(i_dot));

            assert(seed_photodot);

            // check to see if any fitting needs to occur
            if ( seed_photodot->getRawHits().size() == 0 )
                {
                    streamlog_out(DEBUG) << "Skipping photodot " << i_dot << ". It didn't receive any data this event." << endl;
                    _photodot_fit_results_tuple->fill(0, i_dot);
                    _photodot_fit_results_tuple->fill(3, 0);
                    _photodot_fit_results_tuple->fill(4, false);
                    _photodot_fit_results_tuple->addRow();
                    continue;
                }

            // create a TrackerHit to store the results
            TrackerHitImpl *fit_photodot = new TrackerHitImpl;

            output_fit_photodots->addElement(fit_photodot);

            bool isValid = doLikelihoodFit(*seed_photodot, fit_photodot, _sigma);

            if (seed_photodot->getRawHits().size() == 1)
                {

                    streamlog_out(DEBUG) << "Skipping photodot " << i_dot << ". It only received data from one pad this event." << endl;
                }

            if (_photodot_fit_results_tuple)
                {

                    _photodot_fit_results_tuple->fill(0, i_dot);
                    _photodot_fit_results_tuple->fill(1, fit_photodot->getPosition()[0]);
                    _photodot_fit_results_tuple->fill(2, fit_photodot->getPosition()[1]);
                    _photodot_fit_results_tuple->fill(3, (int) seed_photodot->getRawHits().size());
                    _photodot_fit_results_tuple->fill(4, isValid);
                    _photodot_fit_results_tuple->addRow();
                }
        }

    evt->addCollection(output_fit_photodots, _output_photodot_collection_name);
}



void PhotoelectricLikelihoodFittingProcessor::check(lcio::LCEvent *  /*evt*/)
{

    // nothing to check here - could be used to fill checkplots in reconstruction processor
}


void PhotoelectricLikelihoodFittingProcessor::end()
{

}

}

#endif // USE_LCCD
