// Marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>


// GEAR
#include <gearxml/tinyxml.h>

//stl
#include <algorithm>

#include "PhotoelectricXMLHelper.h"

using namespace std;
using namespace CLHEP;
using namespace marlintpc;

PhotoelectricXMLFile::PhotoelectricXMLFile(string xmlFile){
  if (!_photodotFile.LoadFile(xmlFile.c_str())) {

    // create the base document

    // create the XML declaration node
    gear::TiXmlDeclaration *declaration = new gear::TiXmlDeclaration("1.0", "", "");

    // create a node to hold the photogeometry element
    gear::TiXmlElement *photogeometry_element = new gear::TiXmlElement(
        "photoelectric_geometry");

    _photodotFile.LinkEndChild(declaration);
    _photodotFile.LinkEndChild(photogeometry_element);

  }
  FillPhotodotInformation();
  FillPhotolineInformation();
}

bool PhotoelectricXMLFile::FillPhotodotInformation(){
  streamlog_out(DEBUG) << "Start filling vector with photodot information... " << endl;
  gear::TiXmlElement *photogeometry_element = _photodotFile.FirstChildElement(
            "photoelectric_geometry");
  if(photogeometry_element == 0){
    streamlog_out( WARNING ) << "Could not find photodot information in the given photodot file." << endl;
    return false;
  }
  gear::TiXmlElement *photodot_element = photogeometry_element->FirstChildElement(
        "photodot");
  if(photodot_element == 0){
    streamlog_out( WARNING ) << "Could not find photodot information in the given photodot file." << endl;
    return false;
  }
  int photodot = 0;
  for (; photodot_element;
    photodot_element = photodot_element->NextSiblingElement("photodot"),photodot++) {
    // get the centroid
    Photodot dot;
    photodot_element->QueryDoubleAttribute("diameter",&dot._diameter);
    photodot_element->QueryIntAttribute("number",&dot._id);
    photodot_element->QueryIntAttribute("module",&dot._module);
    double x,y;
    photodot_element->QueryDoubleAttribute("x",&x);
    photodot_element->QueryDoubleAttribute("y",&y);
    dot._position = Hep2Vector(x,y);
    dot._element = photodot_element;
    _photodots.push_back(dot);
    streamlog_out(DEBUG) << "Add photodot: " << " id=" << dot._id << " module=" << dot._module
        << " at position: " << dot._position << endl;
  }
  streamlog_out(MESSAGE) << "Found " << _photodots.size() << " photodots in the given xml file." << endl;
  return true;
}

bool PhotoelectricXMLFile::FillPhotolineInformation(){
  streamlog_out(DEBUG) << "Start filling vector with photoline information... " << endl;
  gear::TiXmlElement *photogeometry_element = _photodotFile.FirstChildElement(
            "photoelectric_geometry");
  if(photogeometry_element == 0){
    streamlog_out( WARNING ) << "Could not find photodot information in the given photodot file." << endl;
    return false;
  }
  gear::TiXmlElement *photodot_element = photogeometry_element->FirstChildElement(
        "photoline");
  if(photodot_element == 0){
    streamlog_out( WARNING ) << "Could not find photoline information in the given photodot file." << endl;
    return false;
  }
  int photodot = 0;
  for (; photodot_element;
    photodot_element = photodot_element->NextSiblingElement("photoline"),photodot++) {
    // get the centroid
    Photoline line;
    photodot_element->QueryDoubleAttribute("width",&line._width);
    double x,y;
    photodot_element->QueryDoubleAttribute("start_point_x",&x);
    photodot_element->QueryDoubleAttribute("start_point_y",&y);
    line._start = Hep2Vector(x,y);
    photodot_element->QueryDoubleAttribute("end_point_x",&x);
    photodot_element->QueryDoubleAttribute("end_point_y",&y);
    line._end = Hep2Vector(x,y);
    photodot_element->QueryIntAttribute("module",&line._module);
    line._element = photodot_element;
    _photolines.push_back(line);
    streamlog_out(DEBUG) << "Add photoline: " << " module=" << line._module
            << " from " << line._start << " to "<< line._end << endl;
  }
  streamlog_out(MESSAGE) << "Found " << _photolines.size() << " photolines in the given xml file." << endl;
  return true;
}

vector<pair<double, vector<Photodot>::const_iterator> > PhotoelectricXMLFile::GetSortedPhotodots(double center_x, double center_y){
  vector<pair<double, vector<Photodot>::const_iterator> > elements;
  streamlog_out(MESSAGE) << "Considering the following position for sorting photodots: " << center_x << "," << center_y << endl;
  Hep2Vector center(center_x, center_y);
  for(vector<Photodot>::const_iterator it = _photodots.begin(); it != _photodots.end(); it++){
    elements.push_back(make_pair(((*it)._position - center).mag2(),it));
  }
  sort(elements.begin(),elements.end(),PhotoelectricXMLFile::Comparison);
  return elements;
}

vector<vector<Photodot>::const_iterator> PhotoelectricXMLFile::GetPhotodots(){
  vector<vector<Photodot>::const_iterator> dots;
  for(vector<Photodot>::const_iterator it = _photodots.begin(); it != _photodots.end(); it++){
    dots.push_back(it);
  }
  return dots;
}

vector<vector<Photoline>::const_iterator> PhotoelectricXMLFile::GetPhotolines(){
  vector<vector<Photoline>::const_iterator> lines;
  for(vector<Photoline>::const_iterator it = _photolines.begin(); it != _photolines.end(); it++){
    lines.push_back(it);
  }
  return lines;
}

bool PhotoelectricXMLFile::Comparison(pair<double, vector<Photodot>::const_iterator> i, pair<double, vector<Photodot>::const_iterator> j){
  return (i.first < j.first);
}

void PhotoelectricXMLFile::SetTimeInfo(float start, float end){
  gear::TiXmlElement *photogeometry_element = _photodotFile.FirstChildElement(
        "photoelectric_geometry");
  gear::TiXmlElement *arrival_time_element = new gear::TiXmlElement("arrival_time");
  arrival_time_element->SetDoubleAttribute("start", start);
  arrival_time_element->SetDoubleAttribute("end", end);
  photogeometry_element->LinkEndChild(arrival_time_element);
}

bool PhotoelectricXMLFile::GetTimeInfo(float &start, float &end){
  gear::TiXmlElement *arrival_element = _photodotFile.RootElement()->FirstChildElement("arrival_time");
  if (! arrival_element){
    return false;
  }

  arrival_element->QueryDoubleAttribute("start", &start);
  arrival_element->QueryDoubleAttribute("end", &end);
  return true;
}

gear::TiXmlElement* PhotoelectricXMLFile::GetPhotoElectricElement() {
  gear::TiXmlElement *photogeometry_element = _photodotFile.FirstChildElement(
          "photoelectric_geometry");
  return photogeometry_element;
}

void PhotoelectricXMLFile::AddPhotodotProjection(gear::TiXmlElement *element, const CLHEP::Hep2Vector centroid, const set<gear::GlobalPadIndex> &pads){
  gear::TiXmlElement *projection_element = new gear::TiXmlElement("projection");
  gear::TiXmlElement *centroid_element = new gear::TiXmlElement("centroid");
  centroid_element->SetDoubleAttribute("x", centroid.x());
  centroid_element->SetDoubleAttribute("y", centroid.y());
  projection_element->LinkEndChild(centroid_element);

  for (set<gear::GlobalPadIndex>::const_iterator global_index = pads.begin();
      global_index != pads.end(); global_index++) {

    gear::TiXmlElement *pad_element = new gear::TiXmlElement("pad");
    pad_element->SetAttribute("pad_index", global_index->getPadIndex());
    pad_element->SetAttribute("module_id", global_index->getModuleID());

    projection_element->LinkEndChild(pad_element);
  }

  element->LinkEndChild(projection_element);
}

void PhotoelectricXMLFile::AddPhotodotProjection(int photodot_id, const CLHEP::Hep2Vector centroid, const set<gear::GlobalPadIndex> &pads){
  Photodot tmp;
  tmp._id = photodot_id;
  gear::TiXmlElement *element = (*find(_photodots.begin(), _photodots.end(), tmp))._element;
  if(element == 0){
    streamlog_out( WARNING ) << "Could not find photodot element for photodot " << photodot_id << endl;
    return;
  }
  AddPhotodotProjection(element, centroid, pads);
}

void PhotoelectricXMLFile::AddPhotodotProjection(const CLHEP::Hep2Vector centroid, const set<gear::GlobalPadIndex> &pads){
  gear::TiXmlElement *element = GetPhotoElectricElement();
  AddPhotodotProjection(element, centroid, pads);
}


void PhotoelectricXMLFile::AddPhotolineProjection(int module_id, const set<gear::GlobalPadIndex> &pads){
  Photoline tmp;
  tmp._module = module_id;
  vector<Photoline>::iterator line = find(_photolines.begin(), _photolines.end(),tmp);
  gear::TiXmlElement* projection_element;
  if(line != _photolines.end()){
    projection_element = (*line)._element;
  } else {
    streamlog_out( WARNING ) << "Could not find photoline element for module " << module_id << endl;
    return;
  }
  for (set<gear::GlobalPadIndex>::const_iterator global_index = pads.begin();
      global_index != pads.end(); global_index++) {

    gear::TiXmlElement *pad_element = new gear::TiXmlElement("pad");
    pad_element->SetAttribute("pad_index", global_index->getPadIndex());
    pad_element->SetAttribute("module_id", global_index->getModuleID());

    projection_element->LinkEndChild(pad_element);
  }
}

std::map<int, std::set<gear::GlobalPadIndex> > PhotoelectricXMLFile::GetPhotodotPads(){
  std::map<int, std::set<gear::GlobalPadIndex> > map;
  if(_photodots.size() == 0)
    return map;
  for(vector<Photodot>::const_iterator it = _photodots.begin();
      it != _photodots.end(); it++){
    std::set<gear::GlobalPadIndex> pads;
    gear::TiXmlElement *element = (*it)._element;
    gear::TiXmlNode *projection_node = element->FirstChild("projection");

    if (!projection_node){
      streamlog_out(DEBUG) << "photodot " << (*it)._id << " doesn't have a projection." << endl;
      map[(*it)._id] = pads;
      continue;
    }

    gear::TiXmlElement *pad_element = projection_node->FirstChildElement("pad");
    // add the pads
    if (!pad_element){
      streamlog_out(DEBUG) << "photodot " << (*it)._id << " doesn't have a pads in the projection." << endl;
      map[(*it)._id] = pads;
      continue;
    }
    for (; pad_element; pad_element = pad_element->NextSiblingElement("pad")){

      int pad_index;
      int module_id;

      pad_element->QueryIntAttribute("pad_index", &pad_index);
      pad_element->QueryIntAttribute("module_id", &module_id);

      gear::GlobalPadIndex new_index(pad_index, module_id);
      pads.insert(new_index);
    }
    map[(*it)._id] = pads;
  }
  return map;
}

std::map<int, std::set<gear::GlobalPadIndex> > PhotoelectricXMLFile::GetPhotolinePads(){
  std::map<int, std::set<gear::GlobalPadIndex> > map;
  if(_photolines.size() == 0){
    streamlog_out(DEBUG) << "No photoline found in the photodot file" << endl;
    return map;
  }
  for (vector<Photoline>::const_iterator line = _photolines.begin(); line != _photolines.end(); line++){
    gear::TiXmlElement *element = (*line)._element;
    gear::TiXmlElement *pad_element = element->FirstChildElement("pad");
    std::set<gear::GlobalPadIndex> pads;
    if(!pad_element){
      streamlog_out(DEBUG) << "No pads for photoline " << (*line)._module << "found." << endl;
      map[(*line)._module] = pads;
    }
    for (; pad_element; pad_element = pad_element->NextSiblingElement("pad")){
      int pad_index;
      int module_id;
      pad_element->QueryIntAttribute("pad_index", &pad_index);
      pad_element->QueryIntAttribute("module_id", &module_id);
      gear::GlobalPadIndex new_index(pad_index, module_id);
      pads.insert(new_index);
    }
    map[(*line)._module] = pads;
  }
  return map;
}

