#include "GridPixPixelClusterProcessor.h"

namespace marlintpc{

GridPixPixelClusterProcessor aGridPixPixelClusterProcessor;

GridPixPixelClusterProcessor::GridPixPixelClusterProcessor() : Processor("GridPixPixelClusterProcessor"){

	registerInputCollection(LCIO::TRACKERHIT,"InputCollection","InputCollection containing TrackerHits",_inputCol,std::string("TimePixHitCandidates"));

	registerOutputCollection(LCIO::TRACKERHIT,"OutputHitsCollection","OutputCollection containing TrackerHits",_outputHits,std::string("TimePixHitsClustered"));

	registerOutputCollection(LCIO::TRACKERDATA,"OutputDataCollection","OutputCollection containing TrackerData",_outputData,std::string("TimePixDataClustered"));

	registerProcessorParameter("SetOutputTransient","if not 0 the output collection is set transient",_outputIsTransient,0);

	registerProcessorParameter("MaxSharedFraction","Maximum charge fraction to be treated as shared charge",_maxSharedFraction,double(.2));

	registerProcessorParameter("MinSharingCharge","Minimum charge for which neighbors are accepted as shared charge",_minSharingCharge,int(0));

	registerOptionalParameter("MinChargeOnPixel","If set pixels with less charge will be killed after clustering",_minChargeOnPixel,int(2500));

	registerProcessorParameter("MinNoPixels","Minimum number of pixels in a hit",_minNoPixel,int(10));

	_description = "";

}

GridPixPixelClusterProcessor::~GridPixPixelClusterProcessor(){

}

void GridPixPixelClusterProcessor::init(){

	printParameters();

	//create histograms
	_histoVec.push_back(new TH1D("pixels","Number of pixels",250,0,500));
	_histoVec.push_back(new TH1D("chargeperpixel","Charge per pixel",500,0,200000));
	_histoVec.push_back(new TH1D("chargeperpixel_out","Charge per thrown out pixel",20,0,_minChargeOnPixel));

	_scatterVec.push_back(new TH2D("occupancy_out","Positions of thrown out pixels",256,-.5,255.5,256,-.5,255.5));

}

void GridPixPixelClusterProcessor::processRunHeader( lcio::LCRunHeader* run ){

	run->parameters().setValue( _processorName + "_revision", "$Rev: 007 $");

	for ( marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
		if ( ! i->second->isOptional() || i->second->valueSet() ) {
			run->parameters().setValue( _processorName + "_" + i->second->name(),
					i->second->value() );
		}
	}

}

void GridPixPixelClusterProcessor::processEvent( lcio::LCEvent* evt ){

	lcio::LCCollectionVec* inputCol(NULL);

	try{

		inputCol = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCol));

	}
	catch(DataNotAvailableException &e){

		streamlog_out(MESSAGE3) << "InputCollection not available" << std::endl;
		return;

	}

	if(inputCol == NULL){

		streamlog_out(ERROR2) << "InputCollection vector is 0" << std::endl;
		return;

	}

	lcio::LCCollectionVec* outputHits = new lcio::LCCollectionVec(LCIO::TRACKERHIT);
	lcio::LCCollectionVec* outputData = new lcio::LCCollectionVec(LCIO::TRACKERDATA);

	outputHits->setTransient( (_outputIsTransient != 0) );
	outputData->setTransient( (_outputIsTransient != 0) );

	//put all pixels of one hit into map (key = cellid0)
	for(lcio::LCCollectionVec::iterator iterCol = inputCol->begin(); iterCol != inputCol->end(); iterCol++)
	{

		lcio::TrackerHit* hit = dynamic_cast<lcio::TrackerHit*>(*iterCol);

		if(hit == NULL)
		{
			streamlog_out(ERROR2) << "Dynamic cast failed!" << std::endl;
			return;
		}

		const lcio::LCObjectVec* rawData = & hit->getRawHits();

		int cellID1 = (dynamic_cast<lcio::TrackerData*>((*rawData)[0]))->getCellID1();

		gear::TPCModule const &module = marlin::Global::GEAR->getTPCParameters().getModule( cellID1 );

		// get some geometry information from gear
		// This processor implicitly assumes a rectangular matrix of pads/pixels
		_nRow = module.getNRows();
		_nColumn = module.getNPads() / _nRow;

		std::list<int> listAllPixels;
		std::map<int,int> mapAllPixels;
		std::map<int,std::list<int>*> mapAllNeighbors;

		for(lcio::LCObjectVec::const_iterator iterData = rawData->begin(); iterData != rawData->end(); iterData++)
		{

			lcio::TrackerData* data = dynamic_cast<lcio::TrackerData*>(*iterData);

			int cellID0 = data->getCellID0();
			lcio::FloatVec adcValues = data->getChargeValues();

			for(int iADC = 0; iADC != static_cast<int>(adcValues.size()); iADC++)
			{

				listAllPixels.push_back(cellID0 + iADC);
				mapAllPixels[cellID0 + iADC] = adcValues[iADC];
				mapAllNeighbors[cellID0 + iADC] = new std::list<int>;

			}

		}

		listAllPixels.sort(CompareByCharge(&mapAllPixels));
		listAllPixels.reverse();

		//put cellid0 of neighbors into vector for every pixel of hit
		for(std::map<int,int>::iterator iterPixel = mapAllPixels.begin(); iterPixel != mapAllPixels.end(); iterPixel++)
		{

			std::map<int,int>::iterator iterWhile = iterPixel;
			iterWhile++;

			bool endOfRow = ( ((iterPixel->first) % _nColumn) == (_nColumn - 1) );
			bool beginOfRow = ( ((iterPixel->first) % _nColumn) == 0 );
			bool endOfColumn = ( ((iterPixel->first) / _nColumn ) == (_nRow - 1) );

			bool goOn(true);

			while( goOn ){

				//break if last pixel is reached!
				if( iterWhile == mapAllPixels.end() ){ break; }

				int distance = iterWhile->first - iterPixel->first;

				if( ( (distance == 1) && (!endOfRow) ) || ( (distance == _nColumn) && (!endOfColumn) ) ){

					mapAllNeighbors[iterPixel->first]->push_back(iterWhile->first);
					mapAllNeighbors[iterWhile->first]->push_back(iterPixel->first);

				}

				if( ( (distance == (_nColumn + 1)) && (!endOfRow) && (!endOfColumn) ) || ( (distance == (_nColumn - 1)) && (!beginOfRow) && (!endOfColumn) ) ){

					mapAllNeighbors[iterPixel->first]->push_back(iterWhile->first);
					mapAllNeighbors[iterWhile->first]->push_back(iterPixel->first);

				}

				//stop while-loop when next pixels cannot be neighbors
				if( distance > (_nColumn +1) ){ goOn = false;}

				iterWhile++;

			}//while( goOn )

		}//for(std::map<int,int>::iterator iterPixel = mapAllPixels.begin(); iterPixel != mapAllPixels.end(); iterPixel++)

		streamlog_out(DEBUG4) << "run" << evt->getRunNumber() << "evt" << evt->getEventNumber() << ": before " << mapAllPixels.size() << std::endl;

		for(std::list<int>::iterator iterList = listAllPixels.begin(); iterList != listAllPixels.end(); iterList++)
		{

			if( (mapAllPixels[*iterList] >= _minSharingCharge) && (mapAllNeighbors[*iterList]->size() > 0) )
			{

				int tranferCharge(0);

				for(std::list<int>::iterator iterNeighborList = mapAllNeighbors[*iterList]->begin(); iterNeighborList != mapAllNeighbors[*iterList]->end(); iterNeighborList++)
				{

					if(mapAllPixels[*iterNeighborList] < _maxSharedFraction*mapAllPixels[*iterList])
					{

						tranferCharge += mapAllPixels[*iterNeighborList];
						mapAllPixels.erase(*iterNeighborList);

						for(std::list<int>::iterator iterNeighbor = mapAllNeighbors[*iterNeighborList]->begin(); iterNeighbor != mapAllNeighbors[*iterNeighborList]->end(); iterNeighbor++)
						{

							mapAllNeighbors[*iterNeighbor]->remove(*iterNeighborList);

						}

						delete mapAllNeighbors[*iterNeighborList];
						mapAllNeighbors.erase(*iterNeighborList);

						listAllPixels.remove(*iterNeighborList);

						iterNeighborList--;

					}

				}

				mapAllPixels[*iterList] += tranferCharge;

			}

		}//for(std::list<int>::iterator iterList = listAllPixels.end(); iterList != listAllPixels.begin(); iterList--)

		streamlog_out(DEBUG4) << "run" << evt->getRunNumber() << "evt" << evt->getEventNumber() << ": after " << mapAllPixels.size() << std::endl;

		lcio::TrackerHitImpl* newHit = new TrackerHitImpl;

		for(std::map<int,int>::iterator iterPixels = mapAllPixels.begin(); iterPixels != mapAllPixels.end(); iterPixels++)
		{

			if( !parameterSet("MinChargeOnPixel") || iterPixels->second > _minChargeOnPixel)
			{

				lcio::TrackerDataImpl* newData = new lcio::TrackerDataImpl;

				_histoVec[1]->Fill(iterPixels->second);


				newData->setCellID1(cellID1);
				newData->setCellID0(iterPixels->first);
				newData->chargeValues().push_back(iterPixels->second);

				outputData->addElement(newData);

				newHit->rawHits().push_back(newData);

			}
			else
			{

				_histoVec[2]->Fill(iterPixels->second);

				int x = iterPixels->first % _nColumn;
				int y = (iterPixels->first - x) / _nColumn;

				_scatterVec[0]->Fill(x,y);

			}

		}

		if((int)newHit->rawHits().size() >= _minNoPixel)
		{
			_histoVec[0]->Fill(newHit->rawHits().size());
			outputHits->addElement(newHit);
		}
		else
		{
			delete newHit;
		}

		//clean up memory
		for(std::map<int,std::list<int>*>::iterator iterMap = mapAllNeighbors.begin(); iterMap != mapAllNeighbors.end(); iterMap++)
		{

			delete iterMap->second;

		}

	}

	if(outputData->size() > 0)
	{
		evt->addCollection(outputData,_outputData);
	}
	else
	{
		delete outputData;
	}

	if(outputHits->size() > 0)
	{
		evt->addCollection(outputHits,_outputHits);
	}
	else
	{
		delete outputHits;
	}

}

void GridPixPixelClusterProcessor::check( lcio::LCEvent* evt ){

	lcio::LCCollectionVec* hitCol(NULL);
	lcio::LCCollectionVec* hitColIn(NULL);

	try{

		hitCol = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_outputHits));
		hitColIn = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCol));

	}
	catch(DataNotAvailableException &e){

		return;

	}

	std::stringstream histoTitleBefore;
	std::stringstream histoTitleAfter;

	histoTitleBefore << "run" << evt->getRunNumber() << "evt" << evt->getEventNumber() << "before";
	histoTitleAfter << "run" << evt->getRunNumber() << "evt" << evt->getEventNumber() << "after";

	TH2D* eventBefore = new TH2D(histoTitleBefore.str().c_str(),histoTitleBefore.str().c_str(),256,-0.5,255.5,256,-0.5,255.5);
	TH2D* eventAfter = new TH2D(histoTitleAfter.str().c_str(),histoTitleAfter.str().c_str(),256,-0.5,255.5,256,-0.5,255.5);

	//create eventBefore
	for(lcio::LCCollectionVec::iterator iterHit = hitColIn->begin(); iterHit != hitColIn->end(); iterHit++)
	{

		lcio::TrackerHit* hit = dynamic_cast<lcio::TrackerHit*>(*iterHit);

		const lcio::LCObjectVec* rawData = & hit->getRawHits();

		int cellID1 = (dynamic_cast<lcio::TrackerData*>((*rawData)[0]))->getCellID1();

		gear::TPCModule const &module = marlin::Global::GEAR->getTPCParameters().getModule( cellID1 );

		// get some geometry information from gear
		// This processor implicitly assumes a rectangular matrix of pads/pixels
		_nRow = module.getNRows();
		_nColumn = module.getNPads() / _nRow;

		for(lcio::LCObjectVec::const_iterator iterData = rawData->begin(); iterData != rawData->end(); iterData++)
		{

			lcio::TrackerData* data = dynamic_cast<lcio::TrackerData*>(*iterData);

			int cellID0 = data->getCellID0();
			lcio::FloatVec adcValues = data->getChargeValues();

			for(int iADC = 0; iADC != static_cast<int>(adcValues.size()); iADC++)
			{

				int x = (cellID0+iADC) % _nColumn;
				int y = (cellID0+iADC-x) / _nColumn;

				eventBefore->Fill(x,y,adcValues[iADC]);

			}

		}

	}

	//create eventAfter
	for(lcio::LCCollectionVec::iterator iterHit = hitCol->begin(); iterHit != hitCol->end(); iterHit++)
	{

		lcio::TrackerHit* hit = dynamic_cast<lcio::TrackerHit*>(*iterHit);

		const lcio::LCObjectVec* rawData = & hit->getRawHits();

		int cellID1 = (dynamic_cast<lcio::TrackerData*>((*rawData)[0]))->getCellID1();

		gear::TPCModule const &module = marlin::Global::GEAR->getTPCParameters().getModule( cellID1 );

		// get some geometry information from gear
		// This processor implicitly assumes a rectangular matrix of pads/pixels
		_nRow = module.getNRows();
		_nColumn = module.getNPads() / _nRow;

		for(lcio::LCObjectVec::const_iterator iterData = rawData->begin(); iterData != rawData->end(); iterData++)
		{

			lcio::TrackerData* data = dynamic_cast<lcio::TrackerData*>(*iterData);

			int cellID0 = data->getCellID0();
			lcio::FloatVec adcValues = data->getChargeValues();

			for(int iADC = 0; iADC != static_cast<int>(adcValues.size()); iADC++)
			{

				int x = (cellID0+iADC) % _nColumn;
				int y = (cellID0+iADC-x) / _nColumn;

				eventAfter->Fill(x,y,adcValues[iADC]);

			}

		}

	}

	//write eventdisplays and clean up memory
	marlin::AIDAProcessor::tree(this);

	eventBefore->Write();
	eventAfter->Write();

	delete eventBefore;
	delete eventAfter;

}

void GridPixPixelClusterProcessor::end(){

	marlin::AIDAProcessor::tree(this);

	//write histograms to file and clean up memory
	for(std::vector<TH1D*>::iterator iterHisto = _histoVec.begin(); iterHisto != _histoVec.end(); iterHisto++)
	{

		(*iterHisto)->Write();
		delete *iterHisto;

	}

	for(std::vector<TH2D*>::iterator iterHisto = _scatterVec.begin(); iterHisto != _scatterVec.end(); iterHisto++)
	{

		(*iterHisto)->Write();
		delete *iterHisto;

	}

}

}//namespace marlintpc
