#include "GridPixXrayObjectCalculatorProcessor.h"

namespace marlintpc{

class excentricity : public ROOT::Minuit2::FCNBase {

public:

	excentricity(const std::vector<double>&  /*xInput*/, const std::vector<double>&  /*yInput*/, const double& posX, const double& posY, const double& pitchXInput, const double& pitchYInput) : X(posX),Y(posY),x(xInput),y(yInput),pitchX(pitchXInput),pitchY(pitchYInput){

	}

	double operator() (const std::vector<double>&  /*p*/) const {

		double sumX(0.);
		double sumY(0.);
		double sumXSqr(0.);
		double sumYSqr(0.);

		for(int i = 0; i < static_cast<int>(x.size()); i++){

			double newX = std::cos(p[0])*(x[i]-X)*pitchX - std::sin(p[0])*(y[i]-Y)*pitchY;
			double newY = std::sin(p[0])*(x[i]-X)*pitchX + std::cos(p[0])*(y[i]-Y)*pitchY;

			sumX += newX;
			sumY += newY;
			sumXSqr += newX*newX;
			sumYSqr += newY*newY;

		}

		double rmsX = std::sqrt((sumXSqr/static_cast<double>(x.size())) - (sumX*sumX/static_cast<double>(x.size())/static_cast<double>(x.size())));
		double rmsY = std::sqrt((sumYSqr/static_cast<double>(y.size())) - (sumY*sumY/static_cast<double>(y.size())/static_cast<double>(y.size())));

		double exc = rmsX / rmsY;

		return -exc;

	}

	//not needed as error is irrelevant
	double Up() const {return 0.5;}

private:

	double X;
	double Y;
	std::vector<double> x;
	std::vector<double> y;
	double pitchX;
	double pitchY;

};//class excentricity

GridPixXrayObjectCalculatorProcessor aGridPixXrayObjectCalculatorProcessor;

GridPixXrayObjectCalculatorProcessor::GridPixXrayObjectCalculatorProcessor() : Processor("GridPixXrayObjectCalculatorProcessor"), _minuit(NULL), _excentricity(NULL){

	_description = "Creates GridPixXrayObjects from TrackerHits and calculates their rotAngle";

	registerInputCollection(LCIO::TRACKERHIT,"InputCollection","InputCollection containing TrackerHits",_inputCol,std::string("TimePixHitCandidates"));

	registerOutputCollection(LCIO::TRACKERHIT,"OutputHitCollection","OutputCollection containing TrackerHits",_outputHits,std::string("TimePixHits"));

	registerOutputCollection(LCIO::LCGENERICOBJECT,"OutputXrayCollection","OutputCollection containing GridPixXrayObjects",_outputXrays,std::string("GridPixXrayObjects"));

	registerOutputCollection(LCIO::LCRELATION,"OutputCollection","OutputCollection containing LCRelations with GridPixXrayObjects and TrackerHits",_outputCol,std::string("TimepixXrayObjects"));

	registerInputCollection(LCIO::LCGENERICOBJECT,"ModeMapCollectionName","Name of the mode map collection",_modeMapColName,std::string("TimePixModeMap"));

	registerInputCollection(LCIO::LCGENERICOBJECT,"ModeMapCorrCollectionName","Name of the collection containing the correction to the mode map",_modeMapCorrName,std::string("TimePixModeMapCorr"));

	registerProcessorParameter("Shutter","length of shutter signal in us",_shutter,(double) 108.);

	registerProcessorParameter("MultipleShutter","Allow multiple shutters during one acquisition",_multiShutter,false);

	registerProcessorParameter("ReadoutFrequency","readout frequency of TimePix chip in MHz",_readoutFrequency,(double) 100.);

	registerProcessorParameter("DriftVelocity","drift velocity of gas in mm/us",_driftVelocity,(double) 26.7);

	registerProcessorParameter("SetOutputTransient","if not 0 the output collection is set transient",_outputIsTransient,0);

	registerProcessorParameter("minNoPixels","minimum number of pixels in one cluster",_minNoPixels,(int) 10);

}

GridPixXrayObjectCalculatorProcessor::~GridPixXrayObjectCalculatorProcessor(){}

void GridPixXrayObjectCalculatorProcessor::init(){

	printParameters();

	_minuit = new TFitterMinuit;

	_timePixMap = TimePixMapHandlerProcessor::getTimePixMap();

	_numberXrays = 0;

	_histoRotAngle = new TH1D("rotangle","rotAngle",100,0.,4*std::atan(1.));

	_histoExcentricity = new TH1D("excentricity","Excentricity",200,0.,5.);

	_scatterExcentricity = new TH2D("excentrictybvsa","Excentricity before vs excentricity after",200,0.,5.,200,0.,5.);

}

void GridPixXrayObjectCalculatorProcessor::processRunHeader( lcio::LCRunHeader* run ){

	run->parameters().setValue( _processorName + "_revision", "$Rev: 2000 $");

	for ( marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
		if ( ! i->second->isOptional() || i->second->valueSet() ) {
			run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
		}
	}

}

void GridPixXrayObjectCalculatorProcessor::processEvent( lcio::LCEvent* evt ){

	lcio::LCCollectionVec* inputCol(NULL);

	try{
		inputCol = dynamic_cast<lcio::LCCollectionVec*>(evt->getCollection(_inputCol));
	}
	catch(DataNotAvailableException &e){
		streamlog_out(MESSAGE2) << "processEvent: input collection " <<_inputCol<<" cannot be found in event "<<evt->getEventNumber() <<std::endl;
		return;
	}

	if(inputCol == NULL){
		streamlog_out(ERROR2) << "InputCollection vector is NULL" << std::endl;
		return;
	}

	lcio::LCCollectionVec* outputHits = new lcio::LCCollectionVec(LCIO::TRACKERHIT);
	lcio::LCCollectionVec* outputXrays = new lcio::LCCollectionVec(LCIO::LCGENERICOBJECT);

	lcio::LCCollectionVec* outputCol = new lcio::LCCollectionVec(LCIO::LCRELATION);

	outputHits->setTransient( (_outputIsTransient != 0) );
	outputXrays->setTransient( (_outputIsTransient != 0) );

	outputCol->setTransient( (_outputIsTransient != 0) );

	for(lcio::LCCollectionVec::iterator iterHit = inputCol->begin(); iterHit != inputCol->end(); iterHit++)
	{

		lcio::TrackerHit* rawHit = dynamic_cast<lcio::TrackerHit*>(*iterHit);

		const lcio::LCObjectVec* rawDataVec = & rawHit->getRawHits();

		int cellID1 = dynamic_cast<lcio::TrackerData*>((*rawDataVec)[0])->getCellID1();

		gear::TPCModule const &module = marlin::Global::GEAR->getTPCParameters().getModule( cellID1 );

		int nRows = module.getNRows();
		int nColumns = module.getNPads() / nRows;
		double pitchX = module.getPadPitch(0);
		double pitchY = module.getRowHeight(0);

		lcio::TrackerHitImpl* hit = new TrackerHitImpl;
		GridPixXrayObject* xray = new GridPixXrayObject;

		int clustersize(0);

		int sumChargeInCluster(0);
		int sumTimeInCluster(0);

		int noTOTPixel(0);
		int noTIMEPixel(0);

		double sumX(0.);
		double sumY(0.);
		double sumX2(0.);
		double sumY2(0.);
		double sumXY(0.);

		std::vector<double> xVec;
		std::vector<double> yVec;

		for(lcio::LCObjectVec::const_iterator iterData = rawDataVec->begin(); iterData != rawDataVec->end(); iterData++)
		{

			lcio::TrackerData* raw = dynamic_cast<lcio::TrackerData*>(*iterData);

			int cellID0 = raw->getCellID0();

			double x = cellID0 % nColumns;
			double y = static_cast<int>((cellID0 - x) / nColumns);

			lcio::FloatVec chargeValues = raw->getChargeValues();

			for(lcio::FloatVec::iterator iterCharge = chargeValues.begin(); iterCharge != chargeValues.end(); iterCharge++)
			{

				int mode = ((_timePixMap->find( tpcconddata::TimePixPixelMode::PixelKey( cellID1, cellID0 ) ))->second)->getPixelMode();
				int status = ((_timePixMap->find( tpcconddata::TimePixPixelMode::PixelKey( cellID1, cellID0 ) ))->second)->getPixelStatus();

				if( status == tpcconddata::TimePixPixelMode::INTACT ){

					if( mode == tpcconddata::TimePixPixelMode::TOT ){

						sumChargeInCluster += *iterCharge;

						noTOTPixel++;

					}

					if( mode == tpcconddata::TimePixPixelMode::TIMEPIX ){

						sumTimeInCluster += *iterCharge;

						noTIMEPixel++;

					}

					sumX += x;
					sumX2 += x*x;
					xVec.push_back(x);

					sumY += y;
					sumY2 += y*y;
					yVec.push_back(y);

					sumXY += x*y;

					clustersize++;

				}

				x++;

				if( x == nColumns ){

					x = 0.;
					y++;

				}

			}//for(lcio::FloatVec::iterator iterCharge = chargeValues->begin(); iterCharge != chargeValues->end(); iterCharge++)

				hit->rawHits().push_back( raw );

		}//for(lcio::LCObjectVec::const_iterator iterData = rawDataVec->begin(); iterData != rawDataVec->end(); iterData++)

			if( clustersize < _minNoPixels ){

				delete hit;
				delete xray;

				continue;

			}

			if( sumChargeInCluster < 0 )
			{
				streamlog_out(WARNING2) << "Total charge of Xray in event " << evt->getEventNumber() << " is off integer limits (causes sign flip due to overflow), setting total charge to " << std::numeric_limits<int>::max() << " std::numeric_limits<int>::max()" << std::endl;

				sumChargeInCluster = std::numeric_limits<int>::max();
			}

			double posX = sumX / clustersize;
			double posY = sumY / clustersize;

			double rmsX = std::sqrt(sumX2/clustersize-posX*posX);
			double rmsY = std::sqrt(sumY2/clustersize-posY*posY);

			double rotAngleEstimate = std::atan((sumXY/clustersize) - posX*posY / (rmsX * rmsX));

			if(rotAngleEstimate < 0.){rotAngleEstimate += 8*std::atan(1.);}

			if(rotAngleEstimate > 4*std::atan(1.)){rotAngleEstimate -= 4*std::atan(1.);}

			rmsX = rmsX*pitchX;
			rmsY = rmsY*pitchY;

			_minuit->Clear(); //deletes the actual function on the heap

			_excentricity = new excentricity(xVec,yVec,posX,posY,pitchX,pitchY);

			_minuit->SetMinuitFCN(_excentricity);
			_minuit->SetParameter(0,"rotAngle",rotAngleEstimate,1.,-4*std::atan(1.),4*std::atan(1.));
			_minuit->SetPrintLevel(-1);
			_minuit->CreateMinimizer();

			int minimize = _minuit->Minimize();

			double rotAngle(0.);

			if(minimize != 0){

				streamlog_out(MESSAGE4) << "Event: " << evt->getEventNumber() << " Minuit2 failed -> setting rotAngle to -1" << std::endl;
				rotAngle = -1.;

			}
			else{

				rotAngle = _minuit->GetParameter(0);

				if(rotAngle < 0){ rotAngle += 8*std::atan(1.); }
				if(rotAngle > 4*std::atan(1.)){ rotAngle -= 4*std::atan(1.); }

			}

			double meanDriftTimeCluster(0.);

			if( noTIMEPixel > 0 ){

				meanDriftTimeCluster = _shutter - (sumTimeInCluster/noTIMEPixel)/_readoutFrequency;

			}

			double chargeCluster = sumChargeInCluster*static_cast<double>(clustersize) / noTOTPixel;

			double pos[3] = { 0, 0, 0 };

			pos[0] = (nColumns-(posX+0.5))*pitchX;
			pos[1] = (posY+0.5)*pitchY;
			pos[2] = meanDriftTimeCluster*_driftVelocity;

			gear::Vector2D globalPos = module.localToGlobal( pos[0] , pos[1] );
			pos[0] = globalPos[0];
			pos[1] = globalPos[1];

			hit->setPosition( pos );
			hit->setEDep( sumChargeInCluster );
			hit->setTime( meanDriftTimeCluster*1000 );

			xray->setCenterX(pos[0]);
			xray->setCenterY(pos[1]);

			xray->setPixels(clustersize);
			xray->setCharge(chargeCluster);
			xray->setRotAngle(rotAngle);

			std::vector<double> par;
			par.push_back(rotAngle);

			_histoRotAngle->Fill(rotAngle);
			_histoExcentricity->Fill(-(*_excentricity)(par));

			_scatterExcentricity->Fill(rmsX/rmsY,-(*_excentricity)(par));

			_numberXrays++;

			outputHits->addElement( hit );
			outputXrays->addElement( xray );

			lcio::LCRelationImpl* xrayObject = new LCRelationImpl(xray,hit);

			outputCol->addElement(xrayObject);

	}//for(lcio::LCCollectionVec::iterator iterHit = inputCol->begin(); iterHit != inputCol->end(); iterHit++

	if( !outputCol->empty() ){

		evt->addCollection(outputHits,_outputHits);
		evt->addCollection(outputXrays,_outputXrays);

		evt->addCollection(outputCol,_outputCol);

	}
	else{

		delete outputCol;

		delete outputHits;
		delete outputXrays;

	}

}

void GridPixXrayObjectCalculatorProcessor::end(){

	streamlog_out(MESSAGE4) << "Found " << _numberXrays << " XrayObjects in " << _inputCol << std::endl;

	delete _minuit;

	marlin::AIDAProcessor::tree(this);

	_histoRotAngle->Write();
	delete _histoRotAngle;

	_histoExcentricity->Write();
	delete _histoExcentricity;

	_scatterExcentricity->Write();
	delete _scatterExcentricity;

}

}//namespace marlintpc
