#include "InGridSimpleClusterFinder.h"
#include "TimePixMapHandlerProcessor.h"
#include "TimePixPixelMode.h"

// C++
#include <iostream>
#include <string>
#include <sstream>
#include <cmath>

// include what you need from lcio
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCRelationImpl.h>
#include "UTIL/LCTOOLS.h"
#include "UTIL/LCRelationNavigator.h"
#include <EVENT/LCEvent.h>
#include <IMPL/LCFlagImpl.h>

#include <TMath.h>

// Aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterFactory.h>
#include <RAIDA/RAIDAUtil.h>
#include <RAIDA/IBrushStyleROOT.h>
#include <RAIDA/IPlotterFactoryROOT.h>
#include <AIDA/IMarkerStyle.h>

//marlin
#include <marlin/ConditionsProcessor.h>
#include <marlin/Global.h>

// GEAR
#include <gear/GEAR.h>
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

using namespace lcio;
using namespace marlin;
using namespace tpcconddata;



namespace marlintpc
{
  
  InGridSimpleClusterFinder aInGridSimpleClusterFinder;
  
  
  InGridSimpleClusterFinder::InGridSimpleClusterFinder() : Processor("InGridSimpleClusterFinder"), _timePixMap(0)
  {
    
    _description = "InGridSimpleClusterFinder makes a TimePixHit from every activated Pixel, which is not NOISY or DEAD. ";
    
    registerInputCollection(LCIO::TRACKERDATA,"InputCollectionName","Name of the input TrackerData collection, default: TPCData",_inputColName,std::string("TimePixZeroSuppressedData"));
    
    registerOutputCollection(LCIO::TRACKERHIT,"OutputCollectionName","Name of the output  collection",_outputColName,std::string("TimePixHits"));
    
    registerOutputCollection(LCIO::TRACKERDATA,"OutputTrackerDataCollectionName","Name of the output collection with sepatated tracker data objects (default TimePixSepData)",_outputTrackerDataColName,std::string("TimePixSepData"));
    
    registerProcessorParameter("ReadoutFrequency","readout frequency of TimePix chip in MHz",_readoutFrequency,(double) 55.);
    
    registerProcessorParameter("Shutter","length of shutter signal in us",_shutter,(double) 50.);    
    
    registerProcessorParameter("DriftVelocity","drift velocity of gas in mm/us",_driftVelocity,(double) 8.7);
    
    registerProcessorParameter("Median Filter","use median filter to suppress noisy pixels in time mode? 1:true (default), 0:false",_medianfilter,(int) 1);	   										  
    
    registerProcessorParameter("SetOutputTransient","if not 0 the output collection is set transient",_outputIsTransient,0);   
    
  }
  
  
  void InGridSimpleClusterFinder::init()
  { 
    streamlog_out(DEBUG) << " init() called  " << std::endl;
    
    // you might want to list the used parameters by invoking
    printParameters();
    
    _timePixMap = TimePixMapHandlerProcessor::getTimePixMap();
    
    
  }
  
  
  void InGridSimpleClusterFinder::processRunHeader(LCRunHeader* run)
  {
    run->parameters().setValue( _processorName + "_revision", "$Rev: 1785 $");
    
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
      if ( ! i->second->isOptional() || i->second->valueSet() ) {
	run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
      }
    }
  } 
  
  
  void InGridSimpleClusterFinder::processEvent(LCEvent* evt)
  { 
    if ( evt->getEventNumber() %100==0 )
    {
      streamlog_out(MESSAGE4) << "Event:  " << evt->getEventNumber() << std::endl;
    }
    // do something for each event, usually access the input collection:
    
    // use an LCCollectionVec instead of LCCollection so we can use iterators later
    LCCollectionVec* inputCol = 0;
    
    // do a try / catch in case there is an empty event
    try{
      inputCol = dynamic_cast<LCCollectionVec *>( evt->getCollection( _inputColName ) );
    }
    catch (DataNotAvailableException &e){
      
      // no input data, just return and process the next event
      return;
    }
    streamlog_out(DEBUG0)<<"trycatch"<<std::endl;
    
    // create output Collections for TrackerHits and corrected TrackerData
    LCCollectionVec* outputCol = new LCCollectionVec( LCIO::TRACKERHIT );
    LCCollectionVec* dataCol = new LCCollectionVec( LCIO::TRACKERDATA );
    
    //makes the outputs transient in dependence of the steeringfile. Standard is not transient.
    outputCol->setTransient( _outputIsTransient );
    dataCol->setTransient( _outputIsTransient );
    
    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
    outputCol->setFlag( trkFlag.getFlag() );
    dataCol->setFlag( trkFlag.getFlag() );
    
    // loop the data collection and do something 
    for( LCCollectionVec::iterator inputObjectIter = inputCol->begin() ;
    inputObjectIter < inputCol->end() ;
    inputObjectIter++ )
    {
      // dynamic cast the input object (LCObject) to what it really is e.g.
      TrackerDataImpl * myTrackerData = dynamic_cast<TrackerDataImpl *>( *inputObjectIter );
      
      int chipID = myTrackerData->getCellID1();
      
      // get the corresponding module from gear
      gear::TPCModule const &module = Global::GEAR->getTPCParameters().getModule( chipID );
      
      // get some geometry information from gear
      // This processor implicitly assumes a rectangular matrix of pads/pixels
      int nRows = module.getNRows();
      int nColumns = module.getNPads() / nRows;
      double pitchX = module.getPadPitch(0); // there always should be a pad 0
      double pitchY = module.getRowHeight(0); // there always should be a row 0
      
      //get chargeVec from the TrackerData
      FloatVec chargeVec = myTrackerData->getChargeValues();
      FloatVec temp;
      
      //loop over the activated Pixels which are next to each other
      for( unsigned int iCharge = 0; iCharge < chargeVec.size(); ++iCharge ){
	
	//get PixelID and Pixelmode and Pixelstatus
	int pixelID = myTrackerData->getCellID0() + iCharge;
	int mode = ((_timePixMap->find( TimePixPixelMode::PixelKey( chipID, pixelID ) ))
	->second)->getPixelMode();
	int status = (( _timePixMap->find( TimePixPixelMode::PixelKey( myTrackerData->getCellID1(), pixelID ) ) )->second )->getPixelStatus();
	
	//only Intact pixels are counted. Dead or Noisy pixels are not further interesting
	if(  status == TimePixPixelMode::INTACT ){
	  
	  //create new LCIOObjects
	  TrackerHitImpl* hit = new TrackerHitImpl;
	  TrackerDataImpl* newData = new TrackerDataImpl;
	  
	  //changing the old TrackerData to TrackerHits and the corresponding TrackerData (only one entry in the chargeVec)
	  double pos[3] = {0,0,0};
	  pos[0]=( nColumns-1-pixelID%nColumns )*pitchX;
	  pos[1]=static_cast<int>( ( pixelID-( pixelID%nColumns ) )/nColumns )*pitchY;
	  
	  //In dependence of the Mode of the Chip the counted clockcircles are set to EDep or Time of the TrackerData 
	  if (mode == TimePixPixelMode::TIMEPIX)
	  {
	    int pixelTime=chargeVec[ iCharge ];
	    pos[2]=(_shutter-pixelTime/_readoutFrequency)*_driftVelocity;
	    hit->setTime( ( _shutter-pixelTime/_readoutFrequency)*1000  );
	    hit->setEDep( 0 );
	  }
	  else if ( mode == TimePixPixelMode::TOT )
	  {
	    hit->setEDep( chargeVec[iCharge] );
	    hit->setTime( 0);
	    pos[2]=0;
	  }
	  
	  //position correction from gear
	  gear::Vector2D globalPos = module.localToGlobal( pos[0] , pos[1] );
	  pos[0] = globalPos[0];
	  pos[1] = globalPos[1];
	  
	  //temp Floatvector for storung the chargeVec in the new Dtaa the correct way
	  temp.push_back(chargeVec[iCharge]);
	  





	  //setting the values of the output data and adding the objects to the output collection
	  newData->setChargeValues( temp );
	  newData->setCellID0( pixelID );
	  newData->setCellID1( chipID );
	  hit->setPosition( pos );
	  hit->rawHits().push_back( newData );
	  hit->setCellID0( pixelID );
	  hit->setCellID1( chipID );
	  temp.clear();

















	  dataCol->addElement(newData);
	  outputCol->addElement( hit );
	  
	}//if(  status == TimePixPixelMode::INTACT )
      }//for( unsigned int iCharge = 0; iCharge < chargeVec.size(); ++iCharge )
      
    }//for( LCCollectionVec::iterator inputObjectIter = inputCol->begin() ;inputObjectIter < inputCol->end() ; inputObjectIter++ )
    
    if (_medianfilter==1)
    {
      double * times = new double[dataCol->size()];
      double * weight = new double[dataCol->size()];
      for (unsigned int iter=0;iter<dataCol->size();iter++)
      {
	times[iter]=0;
	weight[iter]=0;
      }
      
      for (unsigned int iterator=0; iterator < dataCol->size() ; iterator++ )
      {
	
	TrackerDataImpl * myTrackerData = dynamic_cast<TrackerDataImpl *>( dataCol->getElementAt(iterator) );
	if(((_timePixMap->find( TimePixPixelMode::PixelKey( myTrackerData->getCellID1(), myTrackerData->getCellID0()) ))
	  ->second)->getPixelMode()==TimePixPixelMode::TIMEPIX)
	{
	  times[iterator]=(myTrackerData->getChargeValues())[0];
	  weight[iterator]=1;
	}
      }
      
      double median=TMath::Median(dataCol->size(),times,weight);
      delete [] times;
      delete [] weight;
      
      for(unsigned int iterator=0; iterator < dataCol->size() ; iterator++)
      {
	
	TrackerDataImpl * myTrackerData = dynamic_cast<TrackerDataImpl *>( dataCol->getElementAt(iterator) );
	if(((_timePixMap->find( TimePixPixelMode::PixelKey( myTrackerData->getCellID1(), myTrackerData->getCellID0()) ))
	  ->second)->getPixelMode()==TimePixPixelMode::TIMEPIX)
	{
	  if (fabs((myTrackerData->getChargeValues())[0]-median)>120)
	  {
	    dataCol->removeElementAt(iterator);
	    outputCol->removeElementAt(iterator);
	    iterator--;
	  }
	}
      }
    }
    
    evt->addCollection( dataCol, _outputTrackerDataColName);
    evt->addCollection( outputCol, _outputColName );
    
  }
  
  
  void InGridSimpleClusterFinder::check(LCEvent*  /*evt*/)
  {
    // check what is needed here 
  }
  
  void InGridSimpleClusterFinder::end()
  { 
    streamlog_out(DEBUG0)<<"SimpleIngridClusterFinder end"<<std::endl;
  }
  
} // close namespace brackets
