#include "TimePixClusterFinderProcessor.h"

// C++
#include <iostream>
#include <queue>
#include <string>
#include <cmath>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/LCFlagImpl.h>
#include "UTIL/LCTOOLS.h"

// Aida
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud1D.h>
#include <AIDA/IHistogram1D.h>

//Root
#include "TROOT.h"
#include "TFile.h"
#include "TH2F.h"
#include "THStack.h"

#endif

using std::string;
using std::queue;
using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{


TimePixClusterFinderProcessor aTimePixClusterFinderProcessor ;


TimePixClusterFinderProcessor::TimePixClusterFinderProcessor() :
        Processor("TimePixClusterFinderProcessor")
{
    /* the constructor call first the super constructor which the name
     * of the processor
     * this must by the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "TimePixClusterFinderProcessor converts Trackerdata of TimePix into TrackerHits by arranging adjacent pixels to clusters" ;


    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERDATA,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TimePixZeroSuppressedData")) ;

    registerOutputCollection(LCIO::TRACKERDATA,
                             "OutputCollectionNameSplitData" ,
                             "Name of the split  data collection"  ,
                             _outputColNameData ,
                             std::string("TimePixSplitData")) ;

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OutputCollectionName" ,
                             "Name of the output collection"  ,
                             _outputColName ,
                             std::string("TimePixHitCandidates")) ;

    registerProcessorParameter("nColumn" ,
                               "number of columns of the chip"  ,
                               _nColumn ,
                               (int) 256) ;

    registerProcessorParameter("nRow" ,
                               "number of rows of the chip"  ,
                               _nRow ,
                               (int) 256) ;

    registerProcessorParameter("minNoPixel" ,
                               "minimum number of pixels in one cluster"  ,
                               _minNoPixel ,
                               (int) 10) ;

    registerProcessorParameter("SetOutputTransient" ,
                               "if not 0 the output collection is set transient"  ,
                               _outputIsTransient,
                               0);
}


void TimePixClusterFinderProcessor::init()
{

    // usually a good idea to
    printParameters() ;

    _nRun = 0 ;

//  hs = new THStack*[10000];

    histoncl =   AIDAProcessor::histogramFactory(this)->
                 createHistogram1D("histoncl",
                                   "010707_001.siy.300.310 number of clusters per event before separtion",
                                   100, 0, 100);
}

void TimePixClusterFinderProcessor::processRunHeader(lcio::LCRunHeader* run)
{


    run->parameters().setValue(_processorName + "_revision", "$Rev: 2626 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }

    _nRun++ ;

}

void TimePixClusterFinderProcessor::processEvent(lcio::LCEvent * evt)
{


    streamlog_out(MESSAGE1) << "Event " << evt->getEventNumber() << std::endl;

    streamlog_out(DEBUG2) << "processEvent( " << evt->getEventNumber() << " )" << std::endl;

    // get the input collection from the event
    LCCollection* inputCol = 0;
    try
        {
            inputCol = evt->getCollection(_inputColName) ;
        } //end of try {

    catch (DataNotAvailableException)
        {
            // if there is not input collection give a warning to the user
            streamlog_out(WARNING2) << " processEvent:"
            << " input collection " << _inputColName
            << " cannot be found in event " << evt->getEventNumber() << std::endl;
            return;
        }

    // The key of this map is the chip ID. For each chip there is a vector which contains the
    // corresponding pulses
    std::map<int, std::vector<TrackerData*> > dataVecMap;

    // copy the pointers from the input collection to the map
    for (int iInput = 0; iInput < inputCol->getNumberOfElements(); iInput++)
        {
            TrackerData* data = dynamic_cast<TrackerData*>(inputCol->getElementAt(iInput)) ;
            if (data != 0)
                {
                    // put the data entry into the vector of the corresponding chip (cellID1)
                    dataVecMap[data->getCellID1()].push_back(data);
                }
        }

//      streamlog_out(DEBUG0)<<" DataVec.size() = "<<DataVec.size()<<std::endl;


    // create the output collection with a string defining the TYPE (see lcio.h)
    LCCollectionVec* outputCol = new LCCollectionVec(LCIO::TRACKERHIT);


    // set the transient flag as definded by the processor parameter
    outputCol->setTransient((_outputIsTransient != 0));

    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    outputCol->setFlag(trkFlag.getFlag());

    // loop the chips
    for (std::map<int,  std::vector<TrackerData*> >::iterator dataVecMapIter = dataVecMap.begin();
            dataVecMapIter != dataVecMap.end(); dataVecMapIter++)
        {

            //Now search for adjacent areas of Pixels
            std::queue< TrackerData*> dataQueue;

            // the value of the map entry (iterator->second) is the data vector
            std::vector<TrackerData*> & dataVec = dataVecMapIter->second;

            for (std::vector<TrackerData*>::iterator dataIter = dataVec.begin();
                    dataIter < dataVec.end(); dataIter++)
                {

                    TrackerHitImpl* hit = new TrackerHitImpl();

                    //push first entry in queue
                    dataQueue.push(*dataIter);
                    hit->rawHits().push_back(*dataIter);

                    int nrPixels = (*dataIter)->getChargeValues().size();;

                    //erase from vector
                    dataVec.erase(dataIter);
                    dataIter--;

                    while (!dataQueue.empty())
                        {

                            int begin1 = (dataQueue.front()->getCellID0()) % _nColumn;
                            //if ( begin1 < 0 ) begin1+=1;

                            int end1 = (dataQueue.front()->getCellID0()) % _nColumn + dataQueue.front()->getChargeValues().size() - 1;

                            int row1 = static_cast<int>((dataQueue.front()->getCellID0()) / _nRow);

                            //look for adjacent data to this range
                            for (std::vector<TrackerData*>::iterator j = dataVec.begin();
                                    j < dataVec.end(); j++)
                                {
                                    int begin2 = ((*j)->getCellID0()) % _nColumn;
                                    //if ( begin2 < 0 ) begin2+=1;

                                    int end2 = ((*j)->getCellID0()) % _nColumn + (*j)->getChargeValues().size() - 1;

                                    int row2 = static_cast<int>(((*j)->getCellID0()) / _nRow);

                                    if (row2 > row1 + 1)
                                        break;

                                    //adjacent data below
                                    bool low1 =
                                        (begin2  <  begin1 - 1) &&
                                        (end2    >  end1   + 1) &&
                                        (row1 == row2 + 1);

                                    bool low2 =
                                        (begin2  >= begin1 - 1) &&
                                        (begin2  <= end1   + 1) &&
                                        (row1 == row2 + 1);

                                    bool low3 =
                                        (end2  >= begin1 - 1) &&
                                        (end2  <= end1   + 1) &&
                                        (row1 == row2 + 1);


                                    //adjacent data above
                                    bool up1 =
                                        (begin2  <  begin1 - 1) &&
                                        (end2    >  end1   + 1) &&
                                        (row1 == row2 - 1);


                                    bool up2 =
                                        (begin2  >= begin1 - 1) &&
                                        (begin2  <= end1   + 1) &&
                                        (row1 == row2 - 1);


                                    bool up3 =
                                        (end2  >= begin1 - 1) &&
                                        (end2  <= end1   + 1) &&
                                        (row1 == row2 - 1);


                                    streamlog_out(DEBUG0) << "begin1 : " << begin1 << "  end1 : " << end1 << std::endl;
                                    streamlog_out(DEBUG0) << "begin2 : " << begin2 << "  end2 : " << end2 << std::endl;
                                    streamlog_out(DEBUG0) << "row1 : " << row1 << "  row2 : " << row2 << std::endl;

                                    if (low1 || low2 || low3 || up1 || up2 || up3)
                                        {
                                            streamlog_out(DEBUG0) << "fabs(row1-row2)  = " << fabs(row1 - row2) << std::endl;

                                            dataQueue.push(*j);
                                            hit->rawHits().push_back(*j);
                                            nrPixels += (*j)->getChargeValues().size();

                                            dataVec.erase(j);
                                            if (j < dataVec.end()) j--;

                                        }// if( low1 || low2 || low3 || up1 || up2 || up3 )


                                } //end of for ( std::vector<TrackerData*>::iterator j = dataVec.begin();

                            dataQueue.pop();

                        }//end of while( !dataQueue.empty() ) {

                    if (nrPixels > _minNoPixel - 1)
                        outputCol->addElement(hit);
                    else
                        delete hit;

                }// dataIter

        } // for dataVecMapIter

    evt->addCollection(outputCol, _outputColName);

    streamlog_out(DEBUG0) << "end processEvent(" << evt->getEventNumber() << ")" << std::endl;

}


void TimePixClusterFinderProcessor::check(lcio::LCEvent * evt)
{

    streamlog_out(DEBUG2) << "check(" << evt->getEventNumber() << ")" << std::endl;

    //LCTOOLS::dumpEvent( evt );


    LCCollection* HitCol;

    try
        {
            HitCol = evt->getCollection(_outputColName);
        }
    catch (DataNotAvailableException)
        {
            return;
        }


    std::stringstream sEvt;
    sEvt << evt->getEventNumber();
    std::string stackname = "clustersevt" + sEvt.str();


    int n = HitCol->getNumberOfElements();

    //changes into directory of this processor in aida-output-file
    AIDAProcessor::tree(this);

    TH2F** histoCluster = new TH2F*[n];
    THStack *hs = new THStack(stackname.c_str() , stackname.c_str());

    for (int i = 0; i < HitCol->getNumberOfElements(); i++)
        {

            TrackerHit* Hit = dynamic_cast<TrackerHit*>(HitCol->getElementAt(i)) ;

            std::stringstream si;
            si << i;
            std::string histoname = "histo" + si.str();

            histoCluster[i] = new TH2F(histoname.c_str() , histoname.c_str() , 256, 0, 256, 256, 0, 256);
            histoCluster[i]->SetLineColor(i + 3);

            if (Hit != 0)
                {

                    std::vector<LCObject*> dataVec = Hit->getRawHits();
                    std::vector<LCObject*>::iterator j;

                    for (j = dataVec.begin(); j != dataVec.end(); j++)
                        {

                            TrackerData* Data = dynamic_cast<TrackerData*>(*j) ;

                            if (Data != 0)
                                {

                                    int cellid0 = Data->getCellID0();
                                    int x = cellid0 % _nColumn;
                                    int y = static_cast<int>((cellid0 - x) / _nColumn) + 1;

                                    FloatVec ChargeVec  = Data->getChargeValues();
                                    FloatVec::iterator sv;

                                    for (sv = ChargeVec.begin(); sv != ChargeVec.end(); sv++)
                                        {

                                            histoCluster[i]->Fill(x, y);
                                            x++;
                                            if (x == _nColumn)
                                                {
                                                    x = 0;
                                                    y++;
                                                }
                                        }

                                    ChargeVec.clear();
                                }
                        }
                }

            hs->Add(histoCluster[i]);

        }

    hs->Write();
    delete hs;

    for (int i = 0; i < HitCol->getNumberOfElements(); i++)
        {
            histoCluster[i]->SetDirectory(0);
            delete histoCluster[i];
        }

    delete[] histoCluster;

    streamlog_out(DEBUG2) << "end check(" << evt->getEventNumber() << ")" << std::endl;
}


void TimePixClusterFinderProcessor::end()
{

    /*
    streamlog_out(MESSAGE4) << "TimePixClusterFinderProcessor::end()  " << name()
      << " processed " << evt->getEventNumber() << " events in " << _nRun << " runs "
      << std::endl ;
    */
}

}


