#include "TimePixMultiChipBoardDividerProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerRawDataImpl.h>
#include "UTIL/LCTOOLS.h"
#include <IMPL/LCFlagImpl.h>

//Marlin
#include <marlin/Exceptions.h>

//AIDA
#include <AIDA/IAnalysisFactory.h>
#include <AIDA/ITreeFactory.h>
#include <AIDA/ITree.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud1D.h>
#include <AIDA/IHistogram2D.h>


using namespace lcio ;
using namespace marlin ;
namespace marlintpc
{


TimePixMultiChipBoardDividerProcessor aTimePixMultiChipBoardDividerProcessor ;


TimePixMultiChipBoardDividerProcessor::TimePixMultiChipBoardDividerProcessor() :
        Processor("TimePixMultiChipBoardDividerProcessor")
{
    /* the constructor call first the super constructor which the name
     * of the processor
     * this must by the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "TimePixMultiChipBoardDividerProcessor converts MultiTrackerRawData of multi_TimePix_Chips into TrackerRawData of each TimePix" ;


    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERRAWDATA,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TimePixRawData")) ;

    registerOutputCollection(LCIO::TRACKERRAWDATA,
                             "OutputCollectionName" ,
                             "Name of the output collection"  ,
                             _outputColName ,
                             std::string("TimePixMultiBoardDividerRawData")) ;

    registerProcessorParameter("PixelsPerChipX" ,
                               "Number of columns of the chip "  ,
                               _pixelsPerChipX ,
                               (int) 256) ;

    registerProcessorParameter("PixelsPerChipY" ,
                               "Number of rows of the chip "  ,
                               _pixelsPerChipY ,
                               (int) 256) ;

    registerProcessorParameter("ChipsPerBoardX" ,
                               "Number of chips in a row "  ,
                               _chipsPerBoardX ,
                               (int) 2) ;

    registerProcessorParameter("ChipsPerBoardY" ,
                               "Number of chops in a column "  ,
                               _chipsPerBoardY ,
                               (int) 2) ;

    registerProcessorParameter("SetOutputTransient" ,
                               "if not 0 the output collection is set transient"  ,
                               _outputIsTransient,
                               0);


}


void TimePixMultiChipBoardDividerProcessor::init()
{

    // usually a good idea to
    printParameters() ;

}

void TimePixMultiChipBoardDividerProcessor::processRunHeader(lcio::LCRunHeader* run)
{


    run->parameters().setValue(_processorName + "_revision", "$Rev:  $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(),
                                               i->second->value());
                }
        }
}

void TimePixMultiChipBoardDividerProcessor::processEvent(lcio::LCEvent * evt)
{

    streamlog_out(MESSAGE1) << "Event " << evt->getEventNumber() << std::endl;


    // get the input collection from the event
    LCCollection* inputCol = 0;

    try
        {

            inputCol = evt->getCollection(_inputColName) ;

        } //end of try {

    catch (DataNotAvailableException)
        {
            // if there is not input collection give a warning to the user
            streamlog_out(MESSAGE3) << "Warning of TimePixMultiChipBoardDividerProcessor::processEvent:"
            << " input collection "
            << _inputColName << " cannot be found in event "
            << evt->getEventNumber() << std::endl;
            return;
        }


    // create the output collection with a string defining the TYPE (see lcio.h)
    LCCollectionVec* outputCol = new LCCollectionVec(LCIO::TRACKERRAWDATA);

    // set the transient flag as definded by the processor parameter
    outputCol->setTransient((_outputIsTransient != 0));

    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    outputCol->setFlag(trkFlag.getFlag());

    // get the number of elements in the input collection,
    int nRawData = inputCol->getNumberOfElements();


    //and loop over it
    for (int iRawData = 0; iRawData < nRawData ; iRawData++)
        {

            /* getElement returns only a LCObject
             * you have to cast it to the get the right type
             * please use dynamic_cast which will return 0 if the object cannot be casted
             */
            TrackerRawData* boardRawData =
                dynamic_cast<TrackerRawData*>(inputCol->getElementAt(iRawData)) ;

            if (boardRawData != 0)
                {
                    //only proceed if the cast was successful
                    ShortVec rawDataADCVal = boardRawData->getADCValues();

                    for (int column = 0; column < _chipsPerBoardX; column++)
                        {
                            for (int row = 0; row < _chipsPerBoardY; row++)
                                {

                                    ShortVec chipData(_pixelsPerChipY*_pixelsPerChipX);
                                    int startpkt = column * _pixelsPerChipX + (_chipsPerBoardX * _pixelsPerChipX * _pixelsPerChipY * row);

                                    //get data of chip on position row and column
                                    for (int rowInChip = 0; rowInChip < _pixelsPerChipY ; rowInChip++)
                                        {
                                            for (int columnInChip = 0 ; columnInChip < _pixelsPerChipX ; columnInChip++)
                                                {
                                                    // i is the index in the rawData vector of the multi chip board
                                                    int i = startpkt + rowInChip * _chipsPerBoardX * _pixelsPerChipX + columnInChip;

                                                    // n is the index in the rawData vector of the chip
                                                    int n = rowInChip * _pixelsPerChipX + columnInChip;

                                                    chipData[n] = rawDataADCVal[i];
                                                } //for columInChip
                                        } // for rowInChip

                                    // now the data for the chip is copied
                                    // create the raw data object
                                    TrackerRawDataImpl *chipRawData = new TrackerRawDataImpl;

                                    // the cellID0 is 0 because it is non zero suppressed data
                                    chipRawData->setCellID0(0);

                                    // the cellID1 is the chip ID
                                    // it is computed as boardID*nChipsPerBoard + chipNumberOnThisBoard
                                    chipRawData->setCellID1(boardRawData->getCellID1()*_chipsPerBoardX * _chipsPerBoardY
                                                            + row * _chipsPerBoardX + column);

                                    chipRawData->setTime(boardRawData->getTime());
                                    chipRawData->setADCValues(chipData);

                                    outputCol->addElement(chipRawData);
                                } // for row
                        } // for column

                }// if ( boardRawData != 0 )

        } // iRawData

    /* after converting all raw data objects
     * add the collection to the event and asign a name to it
     * please note again: DO NOT DELETE the collection this will be done by LCIO
     */
    if (outputCol->size() > 0)
        evt->addCollection(outputCol, _outputColName);
    else
        {   // have to delete the collction if not added to the event
            // the outputCol deletes all the LCObjects it contains
            delete outputCol;
        }

}// processEvent()


void TimePixMultiChipBoardDividerProcessor::check(lcio::LCEvent *  /*evt*/)
{
//
//  LCCollection* TrackerRawDataCol;
//     try{
//    TrackerRawDataCol = evt->getCollection( _outputColName );
//      }
//      catch(lcio::DataNotAvailableException)
//      {
//    streamlog_out(WARNING0) << "Empty event: no zero suppredded TimePix data  in event "
//                <<evt->getEventNumber()<<std::endl;
//    return;
//      }
//
//      //LCTOOLS::dumpEvent( evt );
//
//      std::stringstream sEvt;
//      sEvt << evt->getEventNumber();
//
//      std::string histoname = "clustersevt" + sEvt.str();
//
//
//      AIDAProcessor::histogramFactory(this)->destroy(histoCluster);
//
//      histoCluster = AIDAProcessor::histogramFactory(this)->
//  createHistogram2D( histoname.c_str(), histoname.c_str(),
//             256, 0, 256, 256, 0, 256);
//
//      for( int i = 0; i < TrackerRawDataCol->getNumberOfElements(); i++ ) {
//
//  TrackerRawData* RawData =
//    dynamic_cast<TrackerRawData*>( TrackerRawDataCol->getElementAt( i ) ) ;
//
//  if ( RawData != 0 ) {
//
//    int cellid0 = RawData->getCellID0();
//    int x = cellid0%_nColumn;
//    int y = static_cast<int>((cellid0-x)/_nColumn)+1;
//
//    ShortVec ADCVec  = RawData->getADCValues();
//
//    for ( ShortVec::iterator sv = ADCVec.begin(); sv != ADCVec.end(); sv++ ) {
//
//        histoCluster->fill( x, y, *sv );
//      x++;
//      if ( x == _nColumn ) {
//        x = 0;
//        y++;
//      }
//    }
//
//
//  }
//
//
//      }
//
//
//      std::string histonameraw = "clustersevtraw" + sEvt.str();
//
//      LCCollection* TrackerRawDataColRaw = evt->getCollection( _inputColName );
//
//      AIDAProcessor::histogramFactory(this)->destroy(histoClusterRaw);
//
//      histoClusterRaw = AIDAProcessor::histogramFactory(this)->
//  createHistogram2D( histonameraw.c_str(), histonameraw.c_str(),
//             256, 0, 256, 256, 0, 256);
//
//      for( int i = 0; i < TrackerRawDataColRaw->getNumberOfElements(); i++ ) {
//
//  TrackerRawData* RawData =
//    dynamic_cast<TrackerRawData*>( TrackerRawDataColRaw->getElementAt( i ) ) ;
//
//  if ( RawData != 0 ) {
//
//    int cellid0 = RawData->getCellID0();
//    int x = cellid0%_nColumn;
//    int y = static_cast<int>((cellid0-x)/_nColumn)+1;
//
//    ShortVec ADCVec  = RawData->getADCValues();
//
//    for ( ShortVec::iterator sv = ADCVec.begin(); sv != ADCVec.end(); sv++ ) {
//
//      histoClusterRaw->fill( x, y, *sv );
//      x++;
//      if ( x == _nColumn ) {
//        x = 0;
//        y++;
//      }
//    }
//
//
//  }
//
//
//      }
//
//      // AIDAProcessor::tree(this)->commit();
//
}//check()



void TimePixMultiChipBoardDividerProcessor::end()
{
    /*
    streamlog_out(MESSAGE4) << "TimePixMultiChipBoardDividerProcessor::end()  " << name()
      << " processed " << evt->getEventNumber()
      << " events in " << _nRun << " runs "
      << std::endl ;
    */

}

}
