#include "TimePixPixelInterpolationProcessor.h"
#include "TimePixPixelMode.h"
#include "TimePixMapHandlerProcessor.h"

// C++
#include <iostream>
#include <cmath>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <EVENT/TrackerHit.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include "UTIL/LCTOOLS.h"
#include <IMPL/LCFlagImpl.h>

// Aida
#ifdef MARLIN_USE_AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/IHistogram1D.h>
#include <AIDA/IHistogram2D.h>
#endif

//marlin
#include <marlin/ConditionsProcessor.h>

using std::string;
using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{


TimePixPixelInterpolationProcessor aTimePixPixelInterpolationProcessor ;


TimePixPixelInterpolationProcessor::TimePixPixelInterpolationProcessor() :
        Processor("TimePixPixelInterpolationProcessor")
{
    /* the constructor call first the super constructor which the name
     * of the processor
     * this must by the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
    _description = "TimePixPixelInterpolationProcessor replaces the adc values of dead or noisy pixels  "
                   "with the mean value of the surrounding pixels. " ;


    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERDATA,
                            "InputCollectionName" ,
                            "Name of the input collection"  ,
                            _inputColName ,
                            std::string("TimePixZeroSuppressedData")) ;

    registerOutputCollection(LCIO::TRACKERDATA,
                             "OutputCollectionName" ,
                             "Name of the output collection"  ,
                             _outputColName,
                             std::string("TimePixInterpolData")) ;

    registerProcessorParameter("NColumn" ,
                               "Number of columns of the chip "  ,
                               _nColumn ,
                               (int) 256) ;

    registerProcessorParameter("NRow" ,
                               "Number of rows of the chip "  ,
                               _nRow ,
                               (int) 256) ;

    registerProcessorParameter("MinNoPixelToInterpolate" ,
                               "Minimum number of surrounding pixels needed to interpolate "  ,
                               _minNoPixelToInterpolate,
                               (int) 2) ;

    registerProcessorParameter("OnlyRegardNonZeroPixels" ,
                               "For interpolation only Pixels with ADC-Values larger than 0 are taken into account "  ,
                               _onlyRegardNonZeroPixels,
                               (int) 1) ;

    registerProcessorParameter("SetOutputTransient" ,
                               "if not 0 the output collection is set transient"  ,
                               _outputIsTransient,
                               0);


    _timePixMap = 0;


}


void TimePixPixelInterpolationProcessor::init()
{

    streamlog_out(DEBUG2) << "TimePixPixelInterpolationProcessor::init()" << std::endl;

    // usually a good idea to
    printParameters() ;

    _chipID = 0;
}


void TimePixPixelInterpolationProcessor::processRunHeader(lcio::LCRunHeader* run)
{

    streamlog_out(DEBUG2) << "TimePixPixelInterpolationProcessor::processRunHeader("
    << run->getRunNumber() << ")" << std::endl;

    run->parameters().setValue(_processorName + "_revision", "$Rev: 944 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }


}

void TimePixPixelInterpolationProcessor::processEvent(lcio::LCEvent * evt)
{

    streamlog_out(MESSAGE1) << "Event " << evt->getEventNumber() << std::endl;

    streamlog_out(DEBUG2) << "TimePixPixelInterpolationProcessor::processEvent( "
    << evt->getEventNumber() << " ) " << std::endl;

    //get the current TimePixMap from the MapHandlerProcessor
    _timePixMap = TimePixMapHandlerProcessor::getTimePixMap();

    // get the input collection from the event
    LCCollection* inputCol = 0;
    try
        {
            inputCol = evt->getCollection(_inputColName) ;
        }

    catch (DataNotAvailableException)
        {
            // if there is no input collection give a warning to the user
            streamlog_out(WARNING2)
            << "Warning of TimePixPixelInterpolationProcessor::processEvent: input collection "
            << _inputColName << " cannot be found in event " << evt->getEventNumber() << std::endl;

            return;
        }




    //create map of inputCol with PixelKey as key
    _inputMap.clear();

    for (int j = 0; j < inputCol->getNumberOfElements(); j++)
        {
            TrackerData* Data = dynamic_cast<TrackerData*>(inputCol->getElementAt(j)) ;
            if (Data != 0)    //Data == 0 means the cast failed
                {

                    int cellid0 = Data->getCellID0();
                    FloatVec ChargeVec = Data->getChargeValues();

                    for (FloatVec::iterator f = ChargeVec.begin(); f != ChargeVec.end(); f++)
                        {

                            _inputMap[ TimePixPixelMode::PixelKey(Data->getCellID1(), cellid0)] = *f;
                            cellid0++;

                        }
                }
        }

    //create map of outputCol with PixelKey as key
    std::map< TimePixPixelMode::PixelKey, float> outputMap;

    //loop over the input collection to fill the output collection

    for (int j = 0; j < inputCol->getNumberOfElements(); j++)
        {

            TrackerData* Data = dynamic_cast<TrackerData*>(inputCol->getElementAt(j)) ;
            if (Data != 0)    //Data == 0 means the cast failed
                {

                    FloatVec GetChargeVec = Data->getChargeValues() ;
                    int pixelID = Data->getCellID0();

                    for (FloatVec::iterator c = GetChargeVec.begin();
                            c != GetChargeVec.end(); c++)
                        {


                            int status =
                                ((_timePixMap->find(TimePixPixelMode::PixelKey(Data->getCellID1(),
                                                                               pixelID)))
                                 ->second)->getPixelStatus();


                            if (status == TimePixPixelMode::INTACT)     //pixel is ok, add to ChargeVec
                                {
                                    streamlog_out(DEBUG4) << "intact :" << Data->getCellID1() << " " << pixelID << " " << *c << std::endl;
                                    outputMap[ TimePixPixelMode::PixelKey(Data->getCellID1(), pixelID)] = *c;
                                }
                            else   //pixel is not ok, needs to be interpolated
                                {

                                    if (status == TimePixPixelMode::NOISY)
                                        {

                                            streamlog_out(DEBUG4) << "noisy :" << std::endl;
                                            float newadc = interpolatePixel(Data->getCellID1(), pixelID);
                                            if (newadc > 0)
                                                outputMap[ TimePixPixelMode::PixelKey(Data->getCellID1(), pixelID)] = newadc;
                                        }
                                } //else

                            pixelID++;


                        } // for ( FloatVec::iterator c = ( Data->getChargeValues() ).begin();

                } //if Data!=0


        } //for for ( int j = 0; j < inputCol->getNumberOfElements();

    streamlog_out(DEBUG4) << "_inputMap.size() " << _inputMap.size() << std::endl;
    streamlog_out(DEBUG4) << "outputMap.size() " << outputMap.size() << std::endl;

    //dead pixels must be handled separately since they are not in the zerosuppressed data!
    //loop over the _timePixMap and find all dead pixels:
    for (TimePixMap::iterator t = _timePixMap->begin(); t != _timePixMap->end(); t++)
        {

            if (((*t).second)->getPixelStatus() == TimePixPixelMode::DEAD)
                streamlog_out(DEBUG4) << "dead: " << std::endl;
            float newadc = 0;
            if (((*t).second)->getPixelStatus() == TimePixPixelMode::DEAD)
                newadc = interpolatePixel(((*t).first).ChipID, ((*t).first).PixelID);

            if (newadc > 0)
                outputMap[ TimePixPixelMode::PixelKey(((*t).first).ChipID, ((*t).first).PixelID)] = newadc;
        }


    //now loop over outputMap and create outpuCollection from it
    LCCollectionVec* outputCol = new LCCollectionVec(LCIO::TRACKERDATA);

    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    outputCol->setFlag(trkFlag.getFlag());

    unsigned int safetycounter = 0;
    for (std::map< TimePixPixelMode::PixelKey, float>::iterator i = outputMap.begin();
            i != outputMap.end(); i++)
        {

            TrackerDataImpl* DataImpl = new TrackerDataImpl();
            DataImpl->setCellID1(((*i).first).ChipID);
            DataImpl->setCellID0(((*i).first).PixelID);

            FloatVec ChargeVec;
            ChargeVec.push_back((*i).second);

            int pixelID = ((*i).first).PixelID;
            int oldPixelID = pixelID;

            while (safetycounter < outputMap.size())
                {

                    pixelID = ((*i).first).PixelID;
                    if (pixelID == oldPixelID + 1 &&
                            pixelID % _nColumn != 0)      //Pixel belongs to the same TrackerData
                        ChargeVec.push_back((*i).second);
                    else break; //TrackerData is complete leave while
                    safetycounter++;
                    i++;

                }

            safetycounter++;

            DataImpl->setChargeValues(ChargeVec);

            outputCol->addElement(DataImpl);

        }

    evt->addCollection(outputCol, _outputColName);

    _inputMap.clear();
    outputMap.clear();

}


void TimePixPixelInterpolationProcessor::check(lcio::LCEvent * evt)
{

    streamlog_out(DEBUG2) << "end of TimePixPixelInterpolationProcessor::checkEvent( "
    << evt->getEventNumber() << " )" << std::endl;

}



void TimePixPixelInterpolationProcessor::end()
{
    /*
    streamlog_out(MESSAGE4) << "TimePixPixelInterpolationProcessor::end()  " << name()
      << " processed " << evt->getEventNumber() << " events in " << evt->getRunNumber() << " runs "
      << std::endl ;
    */
}

float TimePixPixelInterpolationProcessor::interpolatePixel(int chipID, int pixelID)
{

    int mode =
        ((_timePixMap->find(TimePixPixelMode::PixelKey(chipID, pixelID)))->second)->getPixelMode();

    std::vector<int> NeighborVec;

    //fill with the PixelIDs of the surrounding pixel:
    NeighborVec.push_back(pixelID - _nColumn - 1);    // |-----------|
    NeighborVec.push_back(pixelID - _nColumn);        // | 0 | 1 | 2 |
    NeighborVec.push_back(pixelID - _nColumn + 1);    // |-----------|
    NeighborVec.push_back(pixelID - 1);               // | 3 | x | 4 |
    NeighborVec.push_back(pixelID + 1);               // |-----------|
    NeighborVec.push_back(pixelID + _nColumn - 1);    // | 5 | 6 | 7 |
    NeighborVec.push_back(pixelID + _nColumn);        // |-----------|
    NeighborVec.push_back(pixelID + _nColumn + 1);    //


    //set all pixelIDs to -1 that can not be used to interpolate
    for (unsigned int n = 0; n < NeighborVec.size(); n++)
        {

            if (n == 0 || n == 3 || n == 5)
                {

                    //the pixel lies on the left edge of the chip
                    //the left neighbors therefore are on the other side of the chip
                    //and should not be used
                    if (NeighborVec[n] % _nColumn == _nColumn - 1)
                        NeighborVec[n] = -1;

                }

            if (n == 2 || n == 4 || n == 7)
                {

                    //the pixel lies on the right edge of the chip
                    //the right neighbors therefore are on the other side of the chip
                    //and should not be used
                    if (NeighborVec[n] % _nColumn == 0)
                        NeighborVec[n] = -1;

                }

            //these pixels are not on the chip anymore
            if (NeighborVec[n] > (_nColumn*_nRow - 1) || NeighborVec[n] < 0)
                NeighborVec[n] = -1;

        } //for ( int n = 0; n < NeighborVec.size(); n++ ){

    //calculate mean of the counting surrounding pixels

    int noPixel = 0;
    int noNonZeroPixel = 0;
    float sumADC = 0;
    streamlog_out(DEBUG4) << "pixelID = " << pixelID << "\n............." << std::endl;

    for (std::vector<int>::iterator iN = NeighborVec.begin();
            iN < NeighborVec.end(); iN++)
        {

            streamlog_out(DEBUG4) << "pixelID neighbor " << *iN << std::endl;
            if (*iN > 0)
                {

                    int mode2 =
                        ((_timePixMap->find(TimePixPixelMode::PixelKey(chipID, *iN)))
                         ->second)->getPixelMode();
                    int status2 =
                        ((_timePixMap->find(TimePixPixelMode::PixelKey(chipID, *iN)))
                         ->second)->getPixelStatus();

                    streamlog_out(DEBUG4) << "mode = " << mode << "  mode2 = " << mode2 << std::endl;
                    streamlog_out(DEBUG4) << "status2 = " << status2 << "(INTACT : "
                    << TimePixPixelMode::INTACT << ") " << std::endl;
                    if (mode == mode2 && status2 == TimePixPixelMode::INTACT)
                        {
                            //pixel can be used to interpolate

                            streamlog_out(DEBUG4) << "sumADC = " << sumADC << " + "
                            << _inputMap[ TimePixPixelMode::PixelKey(chipID, *iN)];

                            if (_inputMap[ TimePixPixelMode::PixelKey(chipID, *iN)] > 0  || _onlyRegardNonZeroPixels == 0)
                                {
                                    noPixel++;
                                    noNonZeroPixel++;
                                }
                            else if (mode2 == TimePixPixelMode::TOT)
                                {
                                    noPixel++;
                                };
                            // THIS IF STATEMENT WAS ADDED BY C:BREZINA:
                            // ONLY PIXEL WITH ADC VALUE != 0 ARE TAKEN INTO ACCOUNT FOR INTERPOLATION
                            sumADC += _inputMap[ TimePixPixelMode::PixelKey(chipID, *iN)];
                            streamlog_out(DEBUG4) << " = " << sumADC << " with " << noPixel << " pixels" << std::endl;

                        }

                } // if ( *iN > 0 )

        } //end for ( std::vector<int>::iterator iN = NeighborVec.begin();

    streamlog_out(DEBUG4) << " nopixel: " << noPixel << std::endl;
    float returnvalue;

    //set mean of the counting surrounding pixels as new charge value
    if (((noPixel >= _minNoPixelToInterpolate) && (mode != TimePixPixelMode::TOT)) || (mode == TimePixPixelMode::TOT && noNonZeroPixel >= 2))
        {
            streamlog_out(DEBUG4) << "interpolated from " << noPixel << " pixels" << std::endl;
            returnvalue = sumADC / static_cast<float>(noPixel) ;
        }

    //if we cannot calculate interpolated charge value we must keep the old one!
    else
        {
            streamlog_out(DEBUG4) << "pixel could not be interpolated only"
            << noPixel << "available for interpolation" << std::endl;
            // CHANGED BY C.BREZINA
            // PIXEL IS DEAD OR NOISY THERE FOR ADC VALUE IS INTERPOLATED OR 0!!!
            returnvalue = 0.0; //_inputMap[ TimePixPixelMode::PixelKey( chipID, pixelID ) ] ;
        }


    streamlog_out(DEBUG4) << "interpolated value : " << returnvalue << std::endl;

    return returnvalue;

}


}



