#ifndef KALTRACKINGTEMPLATE_H
#define KALTRACKINGTEMPLATE_H

#include <map>
#include <list>
#include <EVENT/TrackerHit.h>
#include <IMPL/TrackImpl.h>
#include "TPCHitMap.h"
#include "TPCHitDomain.h"
#include <gear/TPCParameters.h>
#include <kaldet/GearTPCKalDetector.h>
#include <kaldet/GearTPCHit.h>
#include <gear/GearMgr.h>
#include <TMatrixD.h>
#include "TVector3.h"

#include "KalTrackingBase.h"

namespace marlintpc
{

/** This is a base KalDet tracking class. It provides common methods for tracking.
 *
 *  It is templated for a container which contain the track state. For compatilibity with 
 *  with TKalTrack it has to imolement SetOwner(), Clear() and Last(). Last may just return 
 *  a null pointer.
 *
 *  @author Bo Li, Tsinghua University
 *  @author Keisuke Fujii, KEK
 */
  template <class ContainerT> class KalTrackingTemplate : public KalTrackingBase
{
    public:
        KalTrackingTemplate(const gear::GearMgr &gearMgr, 
			unsigned int minTrackHits, 
			unsigned int maxSkipRows, 
			double maxDeltaChi2,
			int hitDistanceFactor,
			int startingHitMode,
			double vDrift,
			std::string modulesNotUsed);

	virtual ~KalTrackingTemplate();

	/**  Find all possible tracks in the given hit collection
	 */
	virtual std::list<EVENT::Track*>  findTracks(EVENT::LCCollection const * hitCollection);


        /** Set maximum delta chi2
         */
        void setMaxDeltaChi2 (double max){_maxDeltaChi2 = max;}

        /** Set minimum number of hits on track
         */
        void setMinTrackHits(unsigned int min)
        {
            _minTrackHits = min;
        }

        /** Set maximum track skipping row
         */
        void setMaxSkipRows(unsigned int max)
        {
            _maxSkipRows = max;
        }

    protected:
	/** Convenience function for printing. Implemented for helix and straight tracks.
	 */
        virtual void printStateVector() = 0;

        virtual IMPL::TrackImpl* getTrack(ModuleRowHit firstHit);

        /** Search a proper hit to track
         */
        virtual void searchTrackHit();

        /** Search track starting from hit
         */
        virtual bool searchTrack(ModuleRowHit hit) = 0;

        /** Add a new site to track and update track parameters.
	 *  The track is given as an argument so it can also be used for track candidates and
	 *  not only for the class-wide variable _kalTrack.
	 *
	 *  Only to be called after the track has been initalised with a first track state.
         */
        virtual bool doAddAndFilter(ContainerT * track, ModuleRowHit hit) = 0;

		virtual double getTrackTangentAndLayerAngle(ContainerT * trackCandidate) = 0;


	/** Create a track candidate by making a KalHit from the ModuleRowHit and
	 *  creating a TrackSite from it. Afterwards it copies the _kalTrack of this class
	 *  and adds and filters with the new site.
	 *  It also hands over the ownership of the new track, so don't forget to delete it.
	 *  Attention: The new track is not the owner of the sites, as they belong to the
	 *  original _kalTrack, except for the newly deleted site. Before deleting the 
	 *  candidate do not forget to delete the last added site on the track!
	 *
	 *  If the add and filter fails you get a 0-pointer;
	 */
	//	virtual ContainerT * createTrackCandidate( ModuleRowHit hit ) = 0;

	/** Get the deltaChi2 from the track candidate
	 */
	virtual double calculateDeltaChi2(ContainerT * trackCandidate) = 0;

	/** Reset parameters and data of track
         */
	virtual void reset();

        /** Hit list of the track
         */
	//std::list<ModuleRowHit> _hitsOnTrack;
	std::vector<ModuleRowHit> _hitsOnTrack;
    
        /**
         */
        gear::GearMgr const & _gearMgr;

        /** An area on the TPC readout plane, now it is (module, row)
         */
        TPCHitDomain* _domain;


        /** TPC geometry
         */
        const gear::TPCParameters&    _tpcParameters;

        /** The last hit that is added to track
         */
	//        ModuleRowHit _lastBestHit;

        /** Minimum number of hits on track
         */
        unsigned int _minTrackHits;

	//FIXME:use enum
	/** track direction
	 */
	int _direction;//+1: index increaing; -1: index decreasing
	static const int INWARDS=-1;
	static const int OUTWARDS=1;

        /** Maximum track skipping row
         */
        unsigned int _maxSkipRows;

        /** Maximum delta chi2
         */
        double _maxDeltaChi2;

		int _hitDistanceFactor;

		int _startingHitMode;
  
	    double _vDrift; //< The drift velocity
	
        kaldet::GearTPCKalDetector _kalDetector;
    
        TKalDetCradle _detectorCradle;

        /** The container of all arranged hits
         */
        TPCHitMap _tpcHitMap;

        virtual void setLCIOTrackParameters(IMPL::TrackImpl* trk) = 0;

	ContainerT * _containerTrack;

	/// Create a kalman hit from module, row and the lcio hit (ModuleRowHit object)
	virtual kaldet::GearTPCHit * createGearHit( ModuleRowHit const & mhr ) const;

	/// A struct as return value for the extrapolation into the next layer
	struct Intersection
	{
	  /// The coordinates of the intersection
	  TVector3 coordinates;
	  /// The status whether the intersection calculation succeeded
	  enum Status{IntersectionOK, OutOfDetector, NoIntersection, 
		      NoIntersectionInActiveRegion} status;

	  /// Convenience constructor
	  Intersection(TVector3 _coordinates_, Status _status_) 
	  :  coordinates(_coordinates_), status(_status_)
	  {}
	};

	/// Use the current track state of the containerTrack to extrapolate into the next layer.
	/// The distance can be positive or negative, depending if we move 
	/// inwards or outwards, or 0 to get another intersection on the same layer.
	virtual Intersection extrapolateIntoNextLayer(int targetDistance,
						      bool useFarHit) = 0;

	virtual const kaldet::GearTPCMeasLayer* extrapolateIntoNextLayer(int targetDistance) = 0;

    private:
        KalTrackingTemplate(const KalTrackingTemplate& orig);
      
        KalTrackingTemplate& operator = (const KalTrackingTemplate& orig);
};

}
#endif  //KALTRACKINGTEMPLATE_H
