#ifndef ROWBASEDFASTHOUGHTRANSFORMATIONPROCESSOR_H
#define ROWBASEDFASTHOUGHTRANSFORMATIONPROCESSOR_H

#include "RowTripletBasedTrackFinderProcessor.h"

#include <marlin/Processor.h>
#include <marlin/Global.h>
#include <lcio.h>
#include <string>

#include <gear/TPCModule.h>

// LCIO:
#include "EVENT/TrackerHit.h"

#include "TFile.h"
#include "TTree.h"
#include "TMath.h"
#include "TVectorD.h"
#include "TMatrixD.h"
#include "TMatrixDSym.h"

namespace marlintpc {
/// Track finder based on Fast Hough Transformation.

/**
 * Fast Hough transformation to construct tracks from space points.
 *
 * The average track direction is in (local) X direction. 
 * The measured coordinates have to be shifted and rescaled to: X->u in [-1, +1],
 * Y,Z -> v,w in [-0.5, 0.5]. The resolution in v and w should be comparable.
 * The track projections v(u), w(u) 
 * are parametrized by series of Legendre polynomials L_i, in the (XY) bending plane up to 
 * order 2 and in the (XZ) plane with order 1.:
 *   - v(u) = a_0 * L_0(u) + a_1 * L_1(u) (+ a_2 *L_2(u))
 *   - w(u) = b_0 * L_0(u) + b_1 * L_1(u)
 *   - with L_0(u) = 1., L_1(u) = u, L_2(u) = 1.5*u*u - 0.5
 *
 * The track parameters a_i, b_i are in [-0.5, +0.5] (the initial hypercube).
 * Each space point defines 2 hyperplanes (v(u)-v_meas=0, w(u)-w_meas=0) in the 5 (or 4)
 * dimensional parameter space (a_i, b_i) with normal vectors (L_i(u), 0.) and (0., L_i(u)).
 * The unit normal vector is used to calculate the distance 
 * of a hyperplane to the center of a hypercube. They intersect if the
 * absolute value of the distance times the largest component of the unit normal vector is
 * smaller than 0.5. The root cube is recursivly split for each dimension in two child cubes.
 * Only childs containing a minimum number of rows are considered further until the hit
 * density in a cube is compatible with a single track. The search stops with the first found 
 * track candidate.
 * 
 * Extended for timepix data (allow multiple hits per row).
 *
 * \param  "InputHits":string           The name of the input collection of TPC hits (default: "TPCHits")
 * \param  "OutputTracks":string        The name of the output collection with the found tracks (default: "FHTTracks")
 * \param  "CenterXYMeasurement":double Center of XY (anode plane) measurement (default 60.)
 * \param  "RangeXYMeasurement":double  Range of XY (anode plane) measurement (default 200.)
 * \param  "CenterXZMeasurement":double Center of XY (drift time) measurement (default 300.)
 * \param  "RangeXZMeasurement":double  Range of XY (drift time) measurement (default 600.)
 * \param  "BFieldScaleFactor":double   \c Optional parameter, scales magnetic field (map), use 1.0 (default) for field ON or 0.0 for field OFF
 * \param  "UseXYMeasurement":bool      \c Optional parameter, flag for using XY (anode plane) measurement (default true)
 * \param  "UseXZMeasurement":bool      \c Optional parameter, flag for using XZ (drift time) measurement (default true)
 * \param  "MinimumRows":int            \c Optional parameter, minimal number of (correlated) rows (default 8)
 * \param  "MaximumRowDifference":int   \c Optional parameter, row correation parameter (max. distance of correlated rows) (default 1)
 * \param  "FractionOfRows":float       \c Optional parameter, relative minimum cube content (rows >= fraction * (correlated) rows) (default 0.8)
 * \param  "MaximumCubes":int           \c Optional parameter, maximum number of (hyper) cubes to check (default 250)
 * \param  "MinimumLevel":int           \c Optional parameter, minimum number of (hyper) cubes splittings (default 5)
 * \param  "MaximumLevel":int           \c Optional parameter, maximum number of (hyper) cubes splittings (default 8)
 * \param  "EfficienyCut":float         \c Optional parameter, minimum hit density (hits/tracklength) (default 0.8)
 * \param  "PurityCut":float            \c Optional parameter, maximum hit density (hits/tracklength) (default 1.1)
 * \param  "MaxHitsPerRow":int          \c Optional parameter, mximum number of hits per row (default 1)
 * \param  "UnusedHitMatchingChi2Cut":double  \c Optional parameter, Chi2 cut for matching unused hits in XY and Z (default 20.)
 * \param  "UnusedHitMaxGapCut:int  \c Optional parameter, maximal (row) gap for matching unused hits in XY and Z (default 4)
 * \param  "EncodedModuleID":bool  \c Optional parameter, flag for encoding of module ID in CellID0 (default true)
 * \param  "ReferencePointAtPca":bool \c Optional parameter,  use PCA as reference point, else 1. hit (default false)
 * \param  "TimePixFlag":bool \c Optional parameter, is timepix data(default false)
 *
 * @author  C. Kleinwort (140205)
 * 
 * \par Credits:
 * The processor skeleton was generated by the script \c createProcessor.py
 *
 * \todo
 *   - (average) track direction different from X direction ?
 *   - hit cleaning for segment creation ?
 *   - replace hard coded radial row number for timepix data ?
 *
 * A detailed description is available as <a href="http://www-flc.desy.de/lcnotes/">lcnote</a> LC-TOOL-2014-006.
 *
 */
class RowBasedFastHoughTransformationProcessor: public marlin::Processor {
	public:

		virtual Processor* newProcessor() {
			return new RowBasedFastHoughTransformationProcessor;
		}

		RowBasedFastHoughTransformationProcessor();

		virtual void init();

		virtual void processRunHeader(EVENT::LCRunHeader* run);

		virtual void processEvent(EVENT::LCEvent* evt);

		virtual void check(EVENT::LCEvent* evt);

		virtual void end();

	protected:
		/* the place for protected and private member data and functions */
		std::string _inputColName; ///< Name of the input collection
		std::string _outputColName;  ///< Name of the output collection
		double _bfieldScaleFactor; ///< scale factor for magnetic field (default: 1.0)
		bool _useXY; ///< Use XY (anode plane) measurement
		bool _useXZ; ///< Use XZ (drift time) measurement
		double _centerXY; ///< Center of XY (anode plane) measurement
		double _rangeXY; ///< Range of XY (anode plane) measurement
		double _centerXZ; ///< Center of XZ (drift time) measurement
		double _rangeXZ; ///< Range of XZ (drift time) measurement
		int _minRow; ///< Minimal number of (correlated) rows
		float _fracRow; ///< Relative minimum cube content
		int _maxRowDiff; ///< Row correation parameter
		int _maxCube; ///< Maximum number of (hyper) cubes to check
		int _minLevel; ///< Minimum number of (hyper) cubes splittings
		int _maxLevel; ///< Maximum number of (hyper) cubes splittings
		float _effCut; ///< Minimum hit density (hits/tracklength)
		float _purCut; ///< Maximum hit density (hits/tracklength)
		int _maxHitsPerRow; ///< Maximum number hits per row
		double _unusedCut; ///< Chi2 cut for matching unused hits in XY and Z
		int _maxGap; ///< Cut for (row) distance to segment for matching unused hits in XY and Z
		bool _encodedModuleID; ///< Module ID is encoded in CellID0
		bool _refAtPCA; ///< Use Pca as reference point (else 1. hit)
		bool _timePixFlag; ///< Is timepix data

	private:
		int _scaleBits; ///< number of scale bits (2^(_scaleBits) = 1.0)
		double _Bzc; ///< magnetic field strength (Bz*c)
		double _Xcenter; ///< TPC center, X coordinate
		double _Ycenter; ///< TPC center, Y coordinate
};

typedef std::vector<int> intListType;
typedef std::vector<double> doubleListType;
typedef std::vector<doubleListType> directionsType;
/// Row based hyperplane.
/**
 * Has hyperplane directions and hit (pointer).
 * Used to check intersection with hypercubes.
 * The distances to the childs (proected on the normal) and
 * the corresponding cuts are stored as (scaled) integers.
 */
class rb_HyperPlane {
	public:
		rb_HyperPlane(int, rb_Hit*, directionsType&, double = 0.75);
		void print() const;
		int getScaleBits() const;
		int getRow() const;
		rb_Hit* getHit() const;
		intListType getCut() const;
		int getCut(int) const;
		std::vector<intListType> getSteps() const;
		int getStep(int, int) const;

	private:
		int _scaleBits; ///< number of scale bits (2^(_scaleBits) = 1.0)
		rb_Hit* _hit; ///< pointers to hit
		intListType _cut; ///< distance cut for maximal component
		std::vector<intListType> _steps; ///< steps (distance to childs projected on normal)
};

typedef std::vector<rb_HyperPlane*> hyperPlaneListType;

int nCube; ///< cube counter

/// Row based hypercubes.
/**
 * Has entries with distances and hyperplanes.
 * Subdivided recursively until first track candidate found or
 * maximal (splitting) level reached.
 */
class rb_HyperCube {
	public:
		rb_HyperCube(unsigned int*, hyperPlaneListType&, intListType&, unsigned int = 0);
		void print() const;
		hitListType divide(int, int, int, int, float, float);

	private:
		unsigned int _setup[2];  ///< setup (number of parameters for pad plane (3), drift time measurement (2))
		std::vector<rb_HyperPlane*> _planes; ///< list of hyperplanes
		intListType _distances; ///< list distances
		int _level; ///< (splitting) level
		int _numChild; ///< number of child cubes
		unsigned int _dimension; ///< (measurement) dimension (1 or 2)
};

} // namespace marlintpc
#endif // ROWBASEDFASTHOUGHTRANSFORMATIONPROCESSOR_H
