#ifndef TPCHITMAP_H
#define TPCHITMAP_H

#include <map>
#include <set>
#include <string>
#include <limits>
#include <sstream>

#include <gear/TPCParameters.h>
#include <EVENT/LCEvent.h>
#include <IMPL/TrackerHitImpl.h>
#include <TVector3.h>
#include <kaldet/GearTPCKalDetector.h>
#include <kaldet/GearTPCCylinderMeasLayer.h>



namespace marlintpc
{

class TPCHitDomain;
 
/** This is hits management class for TPC.
 *
 *  @author Bo Li, Tsinghua University
 *  @author Keisuke Fujii, KEK
 */

struct MarlinTPCHit
{
	EVENT::TrackerHit* hit;
	bool isUsed;	

	MarlinTPCHit(EVENT::TrackerHit* _hit_ = 0, bool _isUsed_ = false)
		        : hit(_hit_), isUsed(_isUsed_)
	{}

	//We just compare the z coordinate and hit pointer, ignore isUsed.
	//Then, if a MarlinTPCHit type hit, which has the same z and pointer with
	//a MarlinTPCHit in a container(say, a set), can not be added.
	bool operator<(const MarlinTPCHit &h) const
	{
		//Mainly sort the hits according to the z coordinate
		if( sqrt(hit->getPosition()[2]) == sqrt(h.hit->getPosition()[2]) )
		{
			return hit < h.hit;
		}
		else
		{
			return sqrt(hit->getPosition()[2]) < sqrt(h.hit->getPosition()[2]);
		}
	}

	bool operator == (const MarlinTPCHit &h) const
	{
		return hit == h.hit;
	}

	MarlinTPCHit& operator = (const MarlinTPCHit &h) 
	{
		hit = h.hit;
		isUsed = h.isUsed;

		return *this;
	}
};

typedef std::set< MarlinTPCHit >::iterator HitSetIter;
typedef std::pair<int, int> ModuleRow;
typedef std::map< ModuleRow, std::set< MarlinTPCHit > > HitContainer;

/**
 * Struct to hold module and row and a hit in this row.
 * It can be invalid if the hit pointer is 0. In this case module and row
 * have no meaning.
 */

struct ModuleRowHit
{
    int module;
    int row;
	MarlinTPCHit tpcHit;

    ModuleRowHit( int _module_ = std::numeric_limits<int>::max(),
		          int _row_    = std::numeric_limits<int>::max(),
		          MarlinTPCHit _tpcHit_= MarlinTPCHit() )
                  : module(_module_), row(_row_), tpcHit(_tpcHit_)
    {}

	bool operator ==(const ModuleRowHit &h) const
	{
		return module == h.module && row == h.row && tpcHit == h.tpcHit;
	}

    double const * getHitPosition() const { return tpcHit.hit->getPosition(); }
	EVENT::TrackerHit* getTrackerHit() const {return tpcHit.hit;}

	bool isUsed() const {return tpcHit.isUsed;}
};

/// A helper function to allow sorting of a std::list of ModuleRowHits
/// according to their distance to a reference position.
class ModuleRowHitListSorter
{
    public:
        ModuleRowHitListSorter(TVector3 refPos): _referencePosition(refPos){}
   
	    // the comparison operator for sort
        bool operator()(ModuleRowHit const & mrh1, ModuleRowHit const & mrh2) const;

    private:
        TVector3 _referencePosition;
};

class TPCHitMap
{
    public:
        TPCHitMap(gear::TPCParameters        const * tpcParameters, 
				  kaldet::GearTPCKalDetector const * det,
				  std::string                modulesNotUsed = std::string("-1")); 

        virtual ~TPCHitMap();

        /** Fill hits to this data container and arrange them by module and row.
	     *  In case a hit is not inside a module or on a pad it is ignored, but a
		 *  warning is printed. For reconstructed data this should never happen,
		 *  but when directly simulating 3D hit they might be in an insensitive area.
		 *  We just filter them here. If we would throw and excetion (the better way for
		 *  reconstructed data, because this should not happen) we would force simulations
		 *  to filter the hits in insensitive areas before filling the hit map, which
		 *  would break current code.
         */   
		void fillEventHits(EVENT::LCCollection const * inputHits);

        /** Remove hits from the container and empties the input hit list as the 
		 *  ModuleRowHits in it are invalidated.
         */
        void removeHits(std::vector<ModuleRowHit>& hitlist);

        /** Remove a hit from the container
         */
        void removeHit(ModuleRowHit hit);

        /** Clear all hits
         */
        void clear();

        /** Get hit size of container
         */
        size_t size() const ;

        /** Get the first hit from the container
         */
        ModuleRowHit getFirstHit(int mode = 0);

	    /** Check if the container is empty.
	     */
	    bool isEmpty() const;
 
        /** Get Hits from a TPC domain
         */
	    std::list<ModuleRowHit> getHitsInDomain(std::set< ModuleRow > const & moduleRowsInDomain);

		ModuleRowHit getHitInDomain(std::set< ModuleRow > const & moduleRowsInDomain, 
										ModuleRowHit hit);

        static double getHitsDistance(ModuleRowHit& hit1, ModuleRowHit& hit2);

	    /** Get a list of all the module/rows pairs which have at least one hit
	     */
	    std::list< ModuleRow > getActiveModuleRows();

	    /** Get all hits in one moduleRow. Throws an exception if there are no hits on 
	     *  the modleRow.
	     */
	    std::set< MarlinTPCHit > const & getHitsInModuleRow(int module, int row);

		/** Set the used flag in the hits container
		 */
		void setHitUsed(const ModuleRowHit& hit, bool used = true);

		int getLayerIndex(int module, int row);

    protected:
        /** The hits container
         */
        HitContainer _hitContainer;

	    std::map< int, std::set<ModuleRow> > _indexModuleRow;

	    gear::TPCParameters const *_tpcParameters;

	    size_t _nHits; ///< The total number of hits currently stored in the container

		kaldet::GearTPCKalDetector const *	_kalDetector;

		std::set<int> _modulesNotUsedSet;

};

}
#endif //TPCHITMAP_H
