#ifndef TIMEPIXHOUGHTRANSFORMIMPROCESSOR_H
#define TIMEPIXHOUGHTRANSFORMIMPROCESSOR_H

#include <marlin/Processor.h>
#include <lcio.h>

// CPP
#include <string>
#include <cmath>
#include <iostream>
#include <vector>
#include <algorithm>

// ROOT
#include "TH2D.h"
#include "TFile.h"

// LCIO
#include <gear/PadRowLayout2D.h>
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <EVENT/TrackerHit.h>
#include <Exceptions.h>

#include "TrackFitterFactory.h"
#include "TrackFitterBase.h"

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{
/** Take an inputCollection of TrackerHits,
 *  treat those TrackerHits in groups (based on their CellID1(cihpID))
 *  and reconstruct Tracks for each group separately.
 *  Which CellID1 is used can be specified setting "ProcessedChipID", default is -1 meaning all CellID1s will be used.\n
 * \n
 * \n
 * Straight line in xy-plane is parameterized as:\n
 * y = x * tan(\phi) + D0/cos(\phi) = x * \frac{sin(\phi)}{cos(\phi)} + \frac{D0}{cos(\phi)}\n
 * or in Hough space:\n
 * D0 = y * cos(\phi) - x * sin(\phi)\n
 * \n
 * @author Martin Rogowski, University of Bonn
 *
 */

class TimePixHoughTransformNormalProcessor : public marlin::Processor
{
 public:
  
  virtual Processor*  newProcessor() { return new TimePixHoughTransformNormalProcessor; }
  
  TimePixHoughTransformNormalProcessor();
  
  virtual void init();
  
  virtual void processRunHeader(lcio::LCRunHeader* run);
  
  virtual void processEvent(lcio::LCEvent* evt);
  
  virtual void check(lcio::LCEvent* evt);
  
  virtual void end();
  
  /* typedefs */
  typedef std::vector< TrackImpl * > TrackImplVec;

  /* my functions */
  /**
   * Fills the Hough Space in XY
   */
  virtual void FillHistogram();
  /**
   * For each entry greater than _nHitsOnTrack; Store the corresponding phi, D0 and the hits on this Track.
   */
  virtual void GetTrackProperties();
  /**
   * For each track property, create a track.
   */
  virtual void CreateTracks();
  /**
   * For each track add the corresponding hits.
   * Starting from the Track with the greatest number of hits on Track.
   */
  virtual void AddHits();
  /**
   * remove tracks with less hits than _nHitsOnTrack
   */
  virtual void RemoveSparseTracks( TrackImplVec & ImplVec, TrackImplVec::iterator itFirst, TrackImplVec::iterator itLast );
  /**
   * add every element in range to output
   */
  virtual void OutputAddElement( LCCollectionVec *& outputCol, TrackImplVec::iterator itFirst, TrackImplVec::iterator itLast );

  /** nested helper class that stores the properties of a straight track */
  struct TrackProperties
    {
        double _Phi; ///< angle between the track and the x-axis in the x-y-projection
        double _D0; ///< impact parameter of the track in the x-y-plane (defines the Point of closest approach, PCA)
        int _hitsontrack; ///< number of hits on the track

        TrackProperties( double phivalue = 0, double d0value = 0
                        , int hitsontrackvalue = 0)
        {
            _Phi = phivalue;
            _D0 = d0value;
            _hitsontrack = hitsontrackvalue;
        };

        bool operator <( TrackProperties const trackproperties ) const
        {
            return (_hitsontrack < trackproperties._hitsontrack);
        }

        bool operator >( TrackProperties const trackproperties ) const
        {
            return (_hitsontrack > trackproperties._hitsontrack);
        }

        bool operator == ( TrackProperties const trackproperties ) const
        {
            return(     _D0 == trackproperties._D0
                    &&   _Phi == trackproperties._Phi
                    &&   _hitsontrack == trackproperties._hitsontrack
                    );
        }
    };

  static bool sortPredicate(const TrackProperties& lhs, const TrackProperties& rhs)
  {
      // locate largest as first element
      return lhs._hitsontrack > rhs._hitsontrack;
  }

  static bool sortTrackImpl( const TrackImpl * lhs, const TrackImpl * rhs )
  {
      // locate largest as first element
      return lhs->getTrackerHits().size() > rhs->getTrackerHits().size();
  }

 protected:
  TH1D * _hMaxConvoluted;

  /* the place for protected and private member data and functions */
  std::string _inputColName;  ///< Name of the input collection
  std::string _outputColName; ///< Name of the output collection

  LCCollectionVec * _inputColVec;

  /* parameters: Track */
  int _nHitsOnTrack;            ///< minimum number if hits on track
  int _ChipID;                  ///< only the chip with _ChipID (CellID1) will be processed; if set to -1 all chipIDs will be processed
  double _maxDistanceToTrackY;   ///< Maximum distance of a hit to the track in mm in the xy plane
  TrackerHitVec _hitVec;         ///< store TRackerHits from _inputColVec

  /* containers */
  std::vector< TrackProperties > _trackPropertiesY; ///< storing track properties for xy plane
  TrackImplVec _trackImplsY; ///< storing in xy plane found tracks

  TH2D * _histoHoughSpace; ///< one Hough space histogram
  std::vector< TH2D * > _histoHoughSpaceVec; ///< vector of Hough space histograms

  /* parameters: Hough space XY*/
  int   _houghSpaceNBinsPhi; ///< The number of bins in the hough space histogram in 'phi' direction (default 100)
  float _houghSpaceMaxPhi;   ///< The half range (HoughSpaceMinPhi .. HoughSpaceMaxPhi) of the Hough space histogram in 'phi' direction (default +0.5 pi)
  float _houghSpaceMinPhi;   ///< The half range (HoughSpaceMinPhi .. HoughSpaceMaxPhi) of the Hough space histogram in 'phi' direction (default -0.5 pi)
  int   _houghSpaceNBinsD0;  ///< The number of bins in the hough space histogram in 'D0' direction (default 50)
  float _houghSpaceMaxD0;   ///< The half range (-max..max) of the Hough space histogram in 'D0' direction (default 65)

  /*  */
  std::string _fileName; ///< Histograms will be written to this file
  int _MaxHisto; ///< Maximum number of saved Histograms
  int _CurrentHisto; /// < Number of the current Histogram

  /** give the status which is set to the transient flag of the output collection
   * it is an int instead of a boolean, because the processor parameter cannot be a boolean
   */
  int _outputIsTransient;

};
} // namespace marlintpc
#endif // HOUGHTRAFOIMPROCESSOR_H
