#ifndef TIMEPIXLOCALROADSEARCHPROCESSOR_H
#define TIMEPIXLOCALROADSEARCHPROCESSOR_H

#include "RowTripletBasedTrackFinderProcessor.h"

#include <marlin/Processor.h>
#include <marlin/Global.h>
#include <lcio.h>
#include <string>

#include <gear/TPCModule.h>

// LCIO:
#include "EVENT/TrackerHit.h"

#include "TFile.h"
#include "TTree.h"
#include "TMath.h"
#include "TVectorD.h"
#include "TMatrixD.h"
#include "TMatrixDSym.h"

namespace marlintpc {
/// Track finder for TimePix data based on local road search.

/**
 * Local road search with 'macro' pixels to construct tracks from space points.
 *
 * For each timepix chip the pixel space is reorganized into larger bins (e.g. of size 32*32 pixels) (to reduce the combinatorics). 
 * Each bin can contain several 'macro' pixels seperated in drift time (Z). The macro pixel position (and shape) is definded 
 * by the center of gravity (and variance) of the positions of the contributing pixels.
 * Qualified pairs of macro pixels (cut on contents and distance) are used to define (linear) search roads. Road consistent
 * with a track segment (in longitudinal and transverse direction) define macro pixel segments. With each found segment the
 * road search is stopped, the corresponding macro pixels removed (marked as used) and the road search is restarted with
 * the remaining macro pixels.
 * In a hierarchical way (to reduce combinatorics and extrapolation) the macro pixel segment are combined into larger
 * segments: chips to octoboards, octoboards to (mechnical) modules and finally all modules.
 * For this equivalence classes of matching segments are used. The matching is defined by the compatibility of the track
 * parameters of two segments at their (relative) midpoint.
 * The (local) coordinates may be shifted by an (XY) offset into a new system. Track directions are defined relative to the origin
 * of that system. 
 *
 * \param  "InputHits":string           The name of the input collection of TPC hits (default: "TPCHits")
 * \param  "OutputTracks":string        The name of the output collection with the found tracks (default: "FHTTracks")
 * \param  "MacroPixelSize":int         \c Optional parameter, macro pixel (bin) size (default 32)
 * \param  "MinTrackLevel":int          \c Optional parameter, min. track level for output (default 1) (0: single chip segment finder, 1,2,3: chip,octoboard,module combiner)
 * \param  "MinZmeas":double            \c Optional parameter, min. measured Z value (default 0.0)
 * \param  "MaxZmeas":double            \c Optional parameter, max. measured Z value (default 600.)
 * \param  "SigmaZ":double              \c Optional parameter, resolution in Z (for clustering) (default 3.0)
 * \param  "Xoffset":double             \c Optional parameter, offset in x of (local) coordinate system (default 0.)
 * \param  "Yoffset":double             \c Optional parameter, offset in y of (local) coordinate system (default 1500.)
 * \param  "BFieldScaleFactor":double   \c Optional parameter, scales magnetic field (map), use 1.0 (default) for field ON or 0.0 for field OFF
 * \param  "MaxMult":int                \c Optional parameter, max. multiplicity for road seeding bins (default 3)
 * \param  "MinPixels":int              \c Optional parameter, min. number of pixels for road seeding maxro pixel (default 3)
 * \param  "MinDist":int                \c Optional parameter, min. distance for road seeding bins (default 2)
 * \param  "MaxPull2XY":double          \c Optional parameter, max. pull squared to road in XY (default 15.)
 * \param  "MaxPull2Z":double           \c Optional parameter, max. pull squared to road in Z (default 15.)
 * \param  "MaxChi2":double             \c Optional parameter, max. Chi2/Ndf for road (default 5.0)
 * \param  "MaxSvar":double             \c Optional parameter, max. relative arc-length variance for road (default 0.15)
 * \param  "MaxGap":double              \c Optional parameter, max. (arc-length) gap for road (default 4.0)
 * \param  "Chi2CutChip":double         \c Optional parameter, max. Chi2/Ndf for chip segment matching (default 30.)
 * \param  "Chi2CutOcto":double         \c Optional parameter, max. Chi2/Ndf for octoboard segment matching (default 30.)
 * \param  "Chi2CutMod":double          \c Optional parameter, max. Chi2/Ndf for module segment matching (default 50.)
 * \param  "DistCut":double             \c Optional parameter, max. relative (to track length sum) distance (of centers) for segment matching (default 2.0)
 * \param  "ReferencePointAtPca":bool   \c Optional parameter, use PCA as reference point, else 1. hit (default false)
 *
 * @author  C. Kleinwort (160604)
 * 
 * \par Credits:
 * The processor skeleton was generated by the script \c createProcessor.py
 *
 * \todo
 *   - variable octoboard to (mechanical) module mapping (from GEAR file?)
 *
 * A detailed description is available as <a href="http://www-flc.desy.de/lcnotes/">lcnote</a> LC-TOOL-2016-xxx.
 *
 */
class TimePixLocalRoadSearchProcessor: public marlin::Processor {
	public:

		virtual Processor* newProcessor() {
			return new TimePixLocalRoadSearchProcessor;
		}

		TimePixLocalRoadSearchProcessor();

		virtual void init();

		virtual void processRunHeader(EVENT::LCRunHeader* run);

		virtual void processEvent(EVENT::LCEvent* evt);

		virtual void check(EVENT::LCEvent* evt);

		virtual void end();

	protected:
		/* the place for protected and private member data and functions */
		std::string _inputColName; ///< Name of the input collection
		std::string _outputColName;  ///< Name of the output collection
		double _bfieldScaleFactor; ///< scale factor for magnetic field (default: 1.0)
		int _minTrackLevel; ///< min. track level for output (default 1)
		int _mpSize; ///< macro pixel (bin) size (default: 32)
		double _minZmeas; ///< min. measured Z value (default 0.)
		double _maxZmeas; ///< max. measured Z value (default 600.)
		double _sigmaZ; ///< resolution in Z (default: 3.0)
		double _xOffset; ///< offset in x of (local) coordinate system (default 0.)
		double _yOffset; ///< offset in y of (local) coordinate system (default 1500.)
		int _maxMult; ///< max. multiplicity for road seeding bins (default: 3)
		int _minPixels; ///< min. number of pixels for road seeding maxro pixel (default: 3)
		int _minDist; ///< min. distance for road seeding bins (default: 2)
		double _maxPull2XY; ///< max. pull squared to road in XY (default: 15.)
		double _maxPull2Z; ///< max. pull squared to road in Z (default: 15.)
		double _maxChi2; ///< max. Chi2/Ndf for road (default 5.)
		double _maxSvar; ///< max. relative arc-length variance for road (default: 0.15)
		double _maxGap; ///< max. (arc-length) gap for road (default: 4.0)
		int _octoOffset; ///< octoboard number offset (in module=(octo+offset)/scale) (default 8)
		int _octoScale; ///< octoboard number scale (in module=(octo+offset)/scale) (default 12)
		double _chi2CutChip; ///< max. Chi2/Ndf for chip segment matching (default: 30.)
		double _chi2CutOcto; ///< max. Chi2/Ndf for octoboard segment matching (default: 30.)
		double _chi2CutMod; ///< max. Chi2/Ndf for module segment matching (default: 50.)
		double _distCut; ///< max. relative (to track length sum) distance (of centers) for segment matching (default: 2.0)
		bool _refAtPCA; ///< Use Pca as reference point (else 1. hit)

	private:
		double _Bzc; ///< magnetic field strength (Bz*c)
		double _Xcenter; ///< TPC center, X coordinate
		double _Ycenter; ///< TPC center, Y coordinate
};

/// TimePix pixel.
class tp_Pixel {
	public:
		tp_Pixel(const int, const EVENT::TrackerHit&, const double, const double);
		void print() const;
		int getHitNum() const;
		int getChip() const;
		int getRow() const;
		int getCol() const;
		double getX() const;
		double getY() const;
		double getZ() const;
		void getPos(double*) const;
		double getSigma2XY() const;
		double getSigma2Z() const;
		double getArcLength() const;
		bool getUsed() const;
		void setUsed();
		void setArcLength(const double);

	private:
		const int _hitNum; ///< input hit collection index
		const int _chipNum; ///< chip number
		const int _rowNum; ///< row number
		const int _colNum; ///< column number
		const double _posX; ///< X position
		const double _posY; ///< Y position
		const double _posZ; ///< Z position
		const double _sigma2XY; ///< resolution (squared) for XY measurement
		const double _sigma2Z; ///< resolution (squared) for Z measurement
		bool _used; ///< used flag
		double _sarc; ///< arc length (from segment fitting)

};

typedef std::vector<tp_Pixel*> pixelListType;
typedef std::map<int, pixelListType> chipPixelMapType;

/// TimePix macro pixel.
class tp_MacroPixel {
	public:
		tp_MacroPixel(const int, pixelListType&, const int);
		~tp_MacroPixel();
		void print() const;
		int getId() const;
		int getSize() const;
		int getEntries() const;
		int getRow() const;
		int getCol() const;
		double getX() const;
		double getY() const;
		double getZ() const;
		void getPos(double*) const;
		double getVarXY(const double, const double) const;
		double getVarZ() const;
		double getSigma2XY() const;
		double getSigma2Z() const;
		bool getUsed() const;
		void setUsed();
		pixelListType getPixels() const;

	private:
		const int _mpId; ///< ID (row/col bin: z bin)
		const int _size; ///< size
		const int _entries; ///< entries
		pixelListType _pixels; ///< pixels
		bool _used; ///< used flag
		double _posX; ///< X position
		double _posY; ///< Y position
		double _posZ; ///< Z position
		double _varXX; ///< variance XX
		double _varXY; ///< variance XY
		double _varYY; ///< variance YY
		double _varZZ; ///< variance ZZ
		double _sigma2XY; ///< resolution (squared) for XY measurement
		double _sigma2Z; ///< resolution (squared) for Z measurement
		int _locRow; ///< local row
		int _locCol; ///< local col

};

typedef std::vector<tp_MacroPixel*> macroPixelListType;
typedef std::pair<tp_MacroPixel*, tp_MacroPixel*> macroPixelPairType;
typedef std::vector<macroPixelPairType> macroPixelPairListType;
typedef std::map<int, macroPixelPairListType> macroPixelPairListMapType;

/// TimePix (macro) pixel bin.
class tp_PixelBin {
	public:
		tp_PixelBin(const int, const int, pixelListType&, const int, const double, const double);
		~tp_PixelBin();
		void print() const;
		int getId() const;
		int getNumPix() const;
		int getNumMPix() const;
		macroPixelListType getMacroPixels() const;

	private:
		const int _binId; ///< bin Id (row, col)
		const int _size; ///< size
		int _numPix; ///< number of pixels in macro pixels
		macroPixelListType _macroPixels; ///< macro pixel
};

typedef std::vector<tp_PixelBin*> pixelBinListType;

/// TimePix (macro) pixel road.
class tp_PixelRoad {
	public:
		tp_PixelRoad(macroPixelPairType, pixelBinListType&, const double*);
		~tp_PixelRoad();
		void print() const;
		bool isValid() const;
		int getNumPix() const;
		int getNumMPix() const;
		double getChi2() const;
		macroPixelListType getMacroPixels() const;

	private:
		const int _mpId1; ///< ID of macro pixel 1
		const int _mpId2; ///< ID of macro pixel 2
		bool _valid; ///< flag for valid road
		int _numPix; ///< number of pixels in macro pixels
		double _chi2ByNdf; ///< sum of chi2 / Ndf
		std::vector<double> _sList; ///< arc length list
		macroPixelListType _macroPixels; ///< macro pixel
};

/// TimePix pixel segment.
class tp_PixelSegment {
	public:
		tp_PixelSegment(const int, const int, const double, macroPixelListType&);
		~tp_PixelSegment();
		void print() const;
		int getNumPar() const;
		int getLevel() const;
		double getChi2() const;
		int getNdf() const;
		double getLength() const;
		void getRefPoint(double*) const;
		void getPar(double *) const;
		const TVectorD& getPar() const;
		const TMatrixDSym& getCov() const;
		bool match(tp_PixelSegment*, const double, const double) const;
		macroPixelListType getMacroPixels() const;
		void getLCIOStateAtRefPoint(const double*, TVectorD&, TMatrixDSym&) const;

	private:
		const int _segId; ///< segment ID
		const int _level; ///< segment level
		const double _bzc; ///< magnetic field strength (Bz*c)
		const int _npar; ///< number of parameters (5: helix, 4: line)
		macroPixelListType _macroPixels; ///< macro pixels
		double _refX; ///< X of reference point
		double _refY; ///< Y of reference point
		double _refZ; ///< Z of reference point
		double _length; ///< segment (arc) length
		int _ndf; ///< number of degrees of freedom of segment fit
		double _chi2; ///< chi2 from segment fit
		TVectorD _parameters; ///< parameter vector
		TMatrixDSym _covariance; ///< covariance matrix
};

typedef std::vector<tp_PixelSegment*> pixelSegmentListType;
typedef std::map<int, pixelSegmentListType> pixelSegmentListMapType;

/// TimePix pixel chip.
class tp_PixelChip {
	public:
		tp_PixelChip(const int, pixelListType&, const int, const double);
		~tp_PixelChip();
		void print() const;
		int getId() const;
		int getNumPix() const;
		int getNumMPix() const;
		int getNumBins() const;
		int findSegments(const double, const int*, const double*);
		pixelSegmentListType getPixelSegments() const;

	private:
		const int _chipId; ///< chip Id
		const int _size; ///< size
		int _numPix; ///< number of pixels in macro pixels
		int _numMPix; ///< number of macro pixels
		pixelBinListType _pixelBins; ///< (macro) pixel bins
		pixelSegmentListType _pixelSegments; ///< pixel segments
};

typedef std::vector<tp_PixelChip*> pixelChipListType;

/// TimePix pixel segment combiner.
class tp_PixelSegmentCombiner {
	public:
		tp_PixelSegmentCombiner(const double);
		pixelSegmentListType run(const int, const int, pixelSegmentListType&, const double, const double);

	private:
		double _bzc; ///< magnetic field strength (Bz*c)
};

} // namespace marlintpc
#endif // TIMEPIXLOCALROADSEARCHPROCESSOR_H
