#ifndef TRACK_FITTER_SIMPLE_CHI_SQUARE_PADS_H
#define TRACK_FITTER_SIMPLE_CHI_SQUARE_PADS_H

#include <string>
#include <marlin/StringParameters.h>
#include <EVENT/LCRunHeader.h>
#include <EVENT/Track.h>
#include <IMPL/TrackImpl.h>
#include <EVENT/TrackerHit.h>
#include "TrackFitterSimpleChiSquare.h"

namespace gear
{
class TPCParameters;
class PadRowLayout2D;
}//namespace gear

namespace marlintpc
{
/** \f$\chi^2\f$ fitter which takes into account the pad geometry.
 *  It reimplements the \c getResiduals functions to calculate the residuals
 *  along the pad rows and not perpendicular to the track.
 *
 *  Currently the actual fitter is not implemented yet.
 *  But one can instanciate the fitter to use the
 *  \c getResiduals functions, which is reimplemented to caculate the residuals
 *  along the pad rows.
 *
 *  \todo: Implement the fitter function.
 */

class TrackFitterSimpleChiSquarePads : public TrackFitterSimpleChiSquare
{
    public:

        /** The constructor needs the pad geometry, which is taken grom GEAR.
         *  In addition it requires the defocussing and diffusion parameters
         *  for the chi square function.
         */
        TrackFitterSimpleChiSquarePads(gear::TPCParameters const *gearTPCParameters,
                                       double transDefocussing, double longDefocussing,
                                       double transDiffusion, double longDiffusion);

        virtual ~TrackFitterSimpleChiSquarePads() {}

        /// Well defined interface to create a new instance of the track fitter
        static TrackFitterBase * getInstance(LCParameters const * parameters);

        virtual std::string getRevision() const;

        /// Information what fitter implementation this is.
        virtual unsigned char getFitterType() const;


        /** Return the distance of the hit \c testHitNumber of the testTrack
         *  to the point of closest approach on the referenceTrack wrt.\ the
         *  hit.
         *  If no referenceTrack is given the
         *  distance to the testTrack is returned.
         *
         *  Throws EVENT::DataNotAvailableException in case there is no hit number
         *  "testHitNumber" on testHit.
         *
         *  Calls  calculateResidual(EVENT::TrackerHit *testHit, EVENT::Track *trackWithoutTestHit)
         *  in this implementation.
         */
        virtual DoubleVec calculateResiduals(EVENT::Track const * testTrack,
                                             unsigned int testHitNumber,
                                             EVENT::Track const * referenceTrack = NULL) const;

        /** Return the distance of the given hit to the test track
         *  where this hit has been removed.
         *  The residual is calculated as the distance
         *  along the pad row.
         *
         */
        virtual DoubleVec calculateResiduals(EVENT::TrackerHit const * testHit,
                                             EVENT::Track const * trackWithoutTestHit) const;

        /** Perform the actual fitting. Uses TrackFitterSimpleChiSquare::fitTrack, using the residuals
         *  from this class, and then setting the correct fitter type.
         */
        virtual IMPL::TrackImpl *fitTrack(EVENT::Track const * seedTrack) const;


    protected:

        /// A pointer to the GEAR TPC parameters
        gear::TPCParameters const *_TPCParameters;

        /// A pointer to the GEAR pad layout
        gear::PadRowLayout2D const * _padLayout;

        /// The type of pad layout (PadRowLayout2D::CARTESIAN or PadRowLayout2D::POLAR)
        int _padLayoutType;

        /// Called from both constructors to set the GEAR information
        /// (originally there were two, right now this is obsolete)
        void  initialiseGEARGeometry(gear::TPCParameters const *gearTPCParameters);

        /// Calculate the angular distance on a circle.
        /// The maximum can be \f$\pi\f$.
        //  double distanceOnCircle(double phi1, double phi2) const;

    private:
        /// The private variable pointintg to the instance you get using getInstance function.
        /// This class is not explicitly a singleton, but if you use getInstance it behaves like one
        /// to avoid overhead by allocating and deallocating memory.
        static TrackFitterSimpleChiSquarePads * _theSimpleChiSquarePadsFitter;
};

}// namespace marlintpc

#endif // TRACK_FITTER_BASE_H
