#ifdef USE_TRACK_FITTER_KALMAN
#include "KalTrackingHelix.h"
#include "TPCHitMap.h"

//GEAR
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#include <gear/BField.h>
#include <gear/PadRowLayout2D.h>
#include <gearimpl/TPCModuleImpl.h>
#include <gearxml/GearXML.h>

#include <marlin/Global.h>
#include <streamlog/streamlog.h>

//LCIO
#include <UTIL/Operators.h>
#include <lcio.h>

//KalTest
#include <kaltest/TKalTrackSite.h>
#include <kaltest/TKalTrackState.h>
#include <kaltest/TVKalState.h>
#include <kaltest/TKalDetCradle.h>
#include <kaltest/THelicalTrack.h>
#include <kaltest/TVTrack.h>

//KalDetector, MeasLayer and KalHit
#include <kaldet/GearTPCKalDetector.h>
#include <kaldet/GearTPCCylinderMeasLayer.h>
#include <kaldet/GearTPCHit.h>

#include <TVirtualPad.h>
#include <TPolyMarker3D.h>
#include <TMath.h>

#include <fstream>
#include <memory> 

//FIXME: not used: static const Bool_t gkDir = kIterForward;

using namespace std;
using namespace gear;
using namespace EVENT;
using namespace kaldet;

namespace marlintpc
{

/** helper function to restrict the range of the azimuthal angle to ]-pi,pi]*/
double KalTrackingHelix::toBaseRange(double phi)
{
    while (phi <= -M_PI)
        {
            phi += 2. * M_PI ;
        }
    while (phi >   M_PI)
        {
            phi -= 2. * M_PI ;
        }
    return phi ;
}


KalTrackingHelix::KalTrackingHelix(const gear::GearMgr &gearMgr, 
				   int minTrackHits, 
				   int maxSkipRows, 
				   double maxDeltaChi2,
				   int hitDistanceFactor,
				   int hitStartingMode,
				   double vDrift,
				   std::string modulesNotUsed)
  : KalTrackingTemplate<TKalTrack>(gearMgr, minTrackHits, maxSkipRows, maxDeltaChi2, 
								   hitDistanceFactor, hitStartingMode, vDrift, modulesNotUsed)
{
}

gear::Vector2D KalTrackingHelix::getNearestPointInNextLayer(ModuleRowHit firstHit)
{
    // Extrapolate the hit into the next pad row (not measurement layer, this might be the
    // neighbouring cylinder segment at the same radius). 
	//
	// Firstly, convert global coordinate to local coordinate.
    const double *globalFirstPos = firstHit.getHitPosition();
    TPCModule const &module = _tpcParameters.getModule(firstHit.module);
    gear::Vector2D localCoordinates 
		           = module.globalToLocal( globalFirstPos[0], globalFirstPos[1] );

	// for making seed, it is OK if we just increase the row index
    int newRow = firstHit.row + 1;

	//FIXME
	if(newRow>=module.getNRows())
	{
		throw gear::Exception("index out of Cradle");
	}

    // FIXME: pad 0 in row might not exist
    int firstPadInRow = module.getPadIndex( newRow, 0 );
	
    gear::Vector2D localPadCentre = module.getLocalPadLayout().getPadCenter(firstPadInRow);
    gear::Vector2D extrapolatedLocalCoordinates;

    switch (module.getLocalPadLayout().getCoordinateType())
    {
      case PadRowLayout2D::CARTESIAN:
   	// x stays the same
	extrapolatedLocalCoordinates[0]=localCoordinates[0];
	// y is from the new row
	extrapolatedLocalCoordinates[1]=localPadCentre[1];
	break;

      case PadRowLayout2D::POLAR:
	// r is from the new row
	extrapolatedLocalCoordinates[0]=localPadCentre[0];
	// phi stays the same
	extrapolatedLocalCoordinates[1]=localCoordinates[1];
	break;

      default:
	streamlog_out(ERROR) << "KalTrackingStraight::searchTrack"
			     << "Unknown coordinate type of the module"<< endl;
	throw gear::UnknownParameterException("Unknown coordinate type");
    }

    gear::Vector2D extrapolatedGlobalGearCoordinates = 
      module.localToGlobal(extrapolatedLocalCoordinates[0],
			   extrapolatedLocalCoordinates[1]);

	return extrapolatedGlobalGearCoordinates;
}

bool KalTrackingHelix::searchTrack(ModuleRowHit firstHit)
{
    streamlog_out(DEBUG3) << "------------------------------------" << endl
    << "In KalTrackingHelix::SearchTrack()"       << endl
    << "------------------------------------" << endl;

    bool foundTrack = false;

	const double *firstPos = firstHit.getHitPosition();

    streamlog_out(DEBUG3) << "hit1_x=" << firstPos[0] 
			  << " hit1_y=" << firstPos[1] << " hit1_z=" <<  firstPos[2] << endl;

	//We reqiure the track hits are in the same module.
	//There are two searching directions for making the initial track
	for(int seedDir = 1; seedDir >= -1; seedDir-=2)
	{
        streamlog_out(DEBUG3) << "seed direction: " << seedDir << endl;
		if(firstHit.row+seedDir<=0)continue;
        if(foundTrack)break;

	    set< pair<int, int> > moduleRowsInDomain1 
			                  = _domain->getModuleRows(firstHit.module, firstHit.row+seedDir);

	    ModuleRowHit defautHit;	
	    ModuleRowHit secondHit = _tpcHitMap.getHitInDomain( moduleRowsInDomain1, firstHit);
		if(secondHit==defautHit)continue;

	
		const double *secondPos = secondHit.getHitPosition();
	
		streamlog_out(DEBUG3) << "hit2_x=" << secondPos[0] << " hit2_y=" << secondPos[1] 
							      << " hit2_z=" << secondPos[2] << endl;

		if(secondHit.row+seedDir<0)continue;
			
		//Require the module is the same for track seed
		if(secondHit.module!=firstHit.module)continue;
	      
	    set< pair<int, int> > moduleRowsInDomain2 = _domain->getModuleRows(
					           secondHit.module, secondHit.row+seedDir); 
	
	    ModuleRowHit thirdHit = _tpcHitMap.getHitInDomain( moduleRowsInDomain2, secondHit );    
		if(thirdHit==defautHit)continue;
			
	    const double *thirdPos = thirdHit.getHitPosition();
		if(thirdHit.module!=firstHit.module)continue;

		//Since the three starting hits are in nearby rows, z diffrece of hit 1 and 2 should 
		//be almostly the same with that of hit 2 and 3.
		double delta12 = TPCHitMap::getHitsDistance(firstHit, secondHit);
		double delta23 = TPCHitMap::getHitsDistance(secondHit, thirdHit);
			   
		streamlog_out(DEBUG3) << "hit3_x=" << thirdPos[0] << " hit3_y=" << thirdPos[1] 
				              << " hit3_z" << thirdPos[2] << endl;

		streamlog_out(DEBUG3) << "delta12: " << delta12 << ", delta23: " << delta23 << endl;	   

		if(delta12>_hitDistanceFactor*delta23 || delta23>_hitDistanceFactor*delta12)
		{
			streamlog_out(DEBUG3) << "Abandon this try because of distances are not consistent" << endl;
			continue;
		}
	
		//reset track
		reset();
	
		//make the initial site
		TKalTrackSite* initialSite = createInitialSite(firstHit, secondHit, thirdHit);
	
		// now the track is initialized and we can start filtering
		_containerTrack->SetOwner(true);
		_containerTrack->Add(initialSite);
	 
		streamlog_out(DEBUG3) << "add the 1st hit" << endl;
	    
		_hitsOnTrack.push_back(firstHit);
		if (!doAddAndFilter(_containerTrack, firstHit))break;  //add the 1st hit
	
		streamlog_out(DEBUG3) << "add the 2nd hit" << endl;
	    _hitsOnTrack.push_back(secondHit);
		if (!doAddAndFilter(_containerTrack, secondHit))break;//add the 2nd hit
	
		streamlog_out(DEBUG3) << "add the 3rd hit" << endl;
		
		_hitsOnTrack.push_back(thirdHit);
		if (!doAddAndFilter(_containerTrack, thirdHit))break;//add the 3rd hit

		_direction = seedDir;
		
		searchTrackHit();//it's recursive
		
	 	if (_hitsOnTrack.size() >= _minTrackHits)
		{
			foundTrack = true;
	    }
		
	    if(foundTrack)break;
	}

    return foundTrack;
}



TKalTrackSite* KalTrackingHelix::createInitialSite(ModuleRowHit& mrh1,
											   ModuleRowHit& mrh2,	
											   ModuleRowHit& mrh3)
{ 
	TVTrackHit * initialKalmanHit = createGearHit(mrh1); 
	TKalTrackSite * initialSite = new TKalTrackSite(*initialKalmanHit);

	TKalMatrix initialStateVector = calculateStartHelix(mrh1, mrh2, mrh3);

    static TKalMatrix initialCovarianceMatrix(kSdim, kSdim);

    for (Int_t i = 0; i < kSdim; i++)
    {
 	   initialCovarianceMatrix(i, i) = 1.e6;
    }
 					   
    // Overwrite the errors with huge errors so the initial hit does not affect the filtering.
    // The real first hit will be added later.

    (*initialKalmanHit)(0, 1) = 1.e6;
    (*initialKalmanHit)(1, 1) = 1.e6;

    initialSite->SetHitOwner(true); // site owns hit
    initialSite->SetOwner(true);    // site owns states

    initialSite->Add( new TKalTrackState(initialStateVector,
 			        initialCovarianceMatrix,
 					*initialSite, 
 					TVKalSite::kPredicted));

    initialSite->Add( new TKalTrackState(initialStateVector,
     				initialCovarianceMatrix,
					*initialSite, 
					TVKalSite::kFiltered));

	return initialSite;
}

double KalTrackingHelix::getTrackTangentAndLayerAngle(TKalTrack* track)
{
    streamlog_out(DEBUG3) << "In KalTrackingHelix::getTrackTangentAndLayerAngle()" << endl;
	
    TKalTrackSite& from = dynamic_cast<TKalTrackSite&>(track->GetCurSite());
    const TVTrackHit& hit = from.GetHit(); 

    const TVMeasLayer& ml_to = hit.GetMeasLayer();

    unique_ptr<TVTrack> help(&static_cast<TKalTrackState &>
                               (from.GetCurState()).CreateTrack()); // tmp track
    TVTrack &hel = *help;

    const TVSurface *sfp = dynamic_cast<const TVSurface *>(&ml_to);


    TVector3 xto;  //crossing point
    double   fito; //angle relative to pivot

    sfp->CalcXingPointWith(hel, xto, fito, 0);
    TMatrixD dxdphi = hel.CalcDxDphi(fito); 

    streamlog_out(DEBUG3) << "Vector of tangent: x=" << dxdphi(0,0) << " y=" << dxdphi(1,0) << endl;
    streamlog_out(DEBUG3) << "Vector of pos: x=" << xto(0) << " y=" << xto(1) << endl;
    streamlog_out(DEBUG3) << "Vector of hit: x=" << from.GetPivot().X() << " y=" << from.GetPivot().Y() << endl;
    streamlog_out(DEBUG3) << "sv: " << from.GetCurState()[2][0] << endl;

    TVector3 v1(dxdphi(0,0), dxdphi(1,0), 0.);
	v1 = v1.Unit();

	const TPCModule& tpcModule = _tpcParameters.getNearestModule(xto(0), xto(1));
	const Vector2D& moduleOffset = tpcModule.getOffset();

	xto(0) += moduleOffset[0];
	xto(1) += moduleOffset[1];
	TVector3 v2 = xto.Unit();

	//See the definitons of positive helix and negtive helix in KalTest. 
	//Actually they just have a pi difference.
	double scalarProd = v1 * v2 * ( from.GetCurState()[2][0]/hel.GetPtoR() > 0. ? -1 : 1 );

    return scalarProd;
}

bool KalTrackingHelix::doAddAndFilter(TKalTrack * track, ModuleRowHit hit) 
{
	//FIXME:we don't use the track in the function parameters list
    streamlog_out(DEBUG3) << "In KalTrackingHelix::doAddAndFilter()" << endl;
    GearTPCHit *gearHit = createGearHit(hit);

    TKalTrackSite * site = new TKalTrackSite(*gearHit);
    site->SetHitOwner();

	//copy the track just for debug
	//TKalTrack  aTrack(*track);
	//aTrack.SetOwner(false);

    if (track->AddAndFilter(*site))
    {
      if (track->GetEntries() >= 5) // only test the chi2 for more than 3 hits
      {
	     double deltachi2 = track->GetCurSite().GetDeltaChi2();
		 
		 if (deltachi2 < _maxDeltaChi2)
	     {
			 streamlog_out(DEBUG3) << "good: deltaChi2=" << deltachi2 << endl;
			 streamlog_out(DEBUG3) << "now the hit number in TKalTrack is: " << track->GetEntries() << endl; 

			 return true;
		 }
		 else
		 {
			 streamlog_out(DEBUG3) << "fail: deltaChi2=" << deltachi2 << endl; 
			 delete site;
			 return false;
		 }
      }
      else // less than 3 hits, always return true.
      { 
		  streamlog_out(DEBUG3) << "less than or equals 3 hits. deltaChi2=" 
			                    << track->GetCurSite().GetDeltaChi2() << endl;
		  _tpcHitMap.setHitUsed(hit);
		  return true;
      }
    }
    else // addAndFilter failed
    {
      streamlog_out(DEBUG3) << "Hit is not added into TKalTrack" << endl;
      delete site;
      return false;
    }
}


TMatrixD KalTrackingHelix::calculateStartHelix(ModuleRowHit const & hit1, 
											   ModuleRowHit const & hit2,
											   ModuleRowHit const & hit3)
{
   streamlog_out(DEBUG3) << "In KalTrackingHelix::calcStartHelix()" << endl;

   TVector3 x1( hit1.getHitPosition() );
   TVector3 x2( hit2.getHitPosition() );
   TVector3 x3( hit3.getHitPosition() );

   streamlog_out(DEBUG3) << "x1: x=" << x1.X() << " x1: y=" << x1.Y() << " x1: z=" << x1.Z() << endl;
   streamlog_out(DEBUG3) << "x2: x=" << x2.X() << " x2: y=" << x2.Y() << " x2: z=" << x2.Z() << endl;
   streamlog_out(DEBUG3) << "x3: x=" << x3.X() << " x3: y=" << x3.Y() << " x3: z=" << x3.Z() << endl;

   static const TVector3 ez(0., 0., 1.);
   TVector3 x12 = x2 - x1;
   TVector3 x13 = x3 - x1;
   TVector3 x23 = x3 - x2;

   x12.SetZ(0.);
   x13.SetZ(0.);
   x23.SetZ(0.);

   Double_t x12mag = x12.Mag();
   Double_t x13mag = x13.Mag();
   Double_t x23mag = x23.Mag();
   x12 = x12.Unit();
   x13 = x13.Unit();
   x23 = x23.Unit();

   //Double_t sinHalfPhi23 = x13.Cross(x23).Z();
   Double_t sinHalfPhi23 = x12.Cross(x13).Z();
   Double_t cosHalfPhi23 = 0.5 * (x13mag/x12mag + (1. - x23mag/x12mag)*(x12mag + x23mag)/x13mag);
   Double_t halfPhi23 = TMath::ATan2(sinHalfPhi23, cosHalfPhi23);

   Double_t r  = -0.5 * x23mag / sinHalfPhi23;//mm2m

   TVector3 xc = 0.5 * (x2 + x3) + r * cosHalfPhi23 * x23.Cross(ez);

   //if (dir == kIterBackward) r = -r;

   //the z component of the b field at xc
   double b = _gearMgr.getBField().at(gear::Vector3D( xc[0], xc[1], xc[2]) )[2];
   const double alpha = 1e9/TMath::C()/b;

   TMatrixD sv(kSdim,1);
   sv(0,0) = 0.;
   sv(1,0) = TMath::ATan2(r * (xc.Y() - x1.Y()), r * (xc.X() - x1.X()));
   sv(2,0) = alpha / r *1000;
   sv(3,0) = 0.;
   sv(4,0) = (x2.Z() - x3.Z()) / (r * 2 * halfPhi23);

   if(b==0.) sv(2,0) = 0;

   const double PI2 = 2*TMath::Pi();

   while(sv(1,0)<0  )sv(1,0)+=PI2;
   while(sv(1,0)>PI2)sv(1,0)-=PI2;

   streamlog_out(DEBUG3) << "the calculated initial track parameters are:" << endl
       << "drho(mm): " << sv(0,0) << endl
       << "phi0: " << sv(1,0) << endl
       << "kappa: " << sv(2,0) << endl
       << "dz(mm): " << sv(3,0) << endl
       << "tanlambda: " << sv(4,0) << endl
       << "radius(m): "   << r << endl;

   if (kSdim == 6) sv(5, 0) = 0.;   // t0

   return sv;
}

//distance can be positive or negtive, and 0.
KalTrackingTemplate<TKalTrack>::Intersection KalTrackingHelix::
                               extrapolateIntoNextLayer(int targetDistance, bool useFarHit)
{
    streamlog_out(DEBUG3) << "In extrapolateIntoNextLayer" << endl;

   	// the last hit on the track
    ModuleRowHit mrh = _hitsOnTrack.back();
    GearTPCMeasLayer const * thisMeasLayer = _kalDetector.GetMeasLayer(mrh.module, mrh.row);
    int thisMeasLayerIndex = thisMeasLayer->GetIndex(); // the index in the cradle
    
    // Determine the layer index which is distance away, whithout counting inactive layers.
    // Do we have to count the index and the distance separately.
    int nextIndex = thisMeasLayerIndex;
    int currentDistance = 0;

    // incease/decrease the distance until we reach the target value
    TVMeasLayer const * nextMeasLayer=0;
    if (targetDistance==0)
    { 
      // we are already there and know that the layer is active
      nextMeasLayer=thisMeasLayer;
    }
    else
    {
      do
      //while (currentDistance != targetDistance)
      {  
        if(targetDistance > 0)
	{
	  nextIndex++;
	  if ( nextIndex >= _detectorCradle.GetEntries() )
	  {   
	    return Intersection( TVector3(), Intersection::OutOfDetector);
	  }   
	  // currentDistance++;
	}
	else // this is safe, we know targetDistance is not 0 here
	{
	  nextIndex--;
	  if(nextIndex<0)
	  {
	    return Intersection( TVector3(), Intersection::OutOfDetector);
	  }
	  // currentDistance--;
	}

	nextMeasLayer = dynamic_cast<TVMeasLayer const *> (_detectorCradle.At(nextIndex));
   
	if (nextMeasLayer==0)
	{
	  throw bad_cast();
	}
       
	// only count active layers for the distance
	if (!nextMeasLayer->IsActive())
	{
	  // An inactive layer. Just ignore it and continue with the next layer.
	  // Do not increase/decrease the current distance.
	  continue;
	}

	if(targetDistance >0)
	  ++currentDistance;
	else
	  --currentDistance;

	streamlog_out(DEBUG3) << "targetDistance: " << targetDistance << ", currentDistance: "<< currentDistance << endl;
      }while (currentDistance != targetDistance);
    }//else targetDistance==0

    // try to cast to a gearTPCClinderMeasLayer, 
    GearTPCCylinderMeasLayer const * cylinderMeasLayer
      = dynamic_cast<GearTPCCylinderMeasLayer const *>(nextMeasLayer);

    if ( cylinderMeasLayer==0 )
    {
      //FIXME: Implementation for planar layer missing
      throw gear::NotImplementedException("Only cylindrical layers are implemented");
    }

    streamlog_out(DEBUG3) << "the next layer index is: " << nextIndex << endl;
    double angle = 0.;

    TKalTrackState& currentState = dynamic_cast<TKalTrackState& >( _containerTrack->GetCurSite().GetCurState() );
    TVTrack& track = currentState.CreateTrack(); 

    //calculate the intersection with the layer
    TVector3 xingPoint;  
    int calcXingRetVal = cylinderMeasLayer->CalcXingPointWith(track,
							      xingPoint, 
							      angle, 
							      (useFarHit? 2 : 0) );
    // use mode 2 to get the far hit, or mode 0 to get the near hit

    // first delete the track before throwing an exception to avoid memory leaks. Use autopointers if this should change in future!
    delete &track;

    streamlog_out(DEBUG3) << "end of extrapolateIntoNextLayer()" << endl;
    if (calcXingRetVal!=0)
    {
      // no intersection could be calculated
      return Intersection( xingPoint, Intersection::NoIntersection);
    }
    
    return  Intersection( xingPoint, Intersection::IntersectionOK);
}


//distance can be positive or negtive, and 0.
const GearTPCMeasLayer* KalTrackingHelix::extrapolateIntoNextLayer(int targetDistance)
{
    streamlog_out(DEBUG3) << "In extrapolateIntoNextLayer" << endl;

   	// the last hit on the track
    ModuleRowHit mrh = _hitsOnTrack.back();

    GearTPCMeasLayer const* thisMeasLayer = _kalDetector.GetMeasLayer(mrh.module, mrh.row);
    GearTPCMeasLayer const* nextMeasLayer = 0;
    
	int thisMeasLayerIndex = thisMeasLayer->GetIndex(); // the index in the cradle
    
    // Determine the layer index which is distance away, whithout counting inactive layers.
    // Do we have to count the index and the distance separately.
    int nextIndex = thisMeasLayerIndex;
    int currentDistance = 0;

    // incease/decrease the distance until we reach the target value

    if (targetDistance==0)
    { 
		// we are already there and know that the layer is active
		nextMeasLayer = thisMeasLayer;
    }
    else
    { 
		do //while (currentDistance != targetDistance)
		{  
			if(targetDistance > 0) 
			{
				nextIndex++;
				
				if ( nextIndex >= _detectorCradle.GetEntries() )
				{   
					//if out of detector return null pointer.
					return 0;
				}   
			}
			else // this is safe, we know targetDistance is not 0 here
			{
			    nextIndex--;

				if(nextIndex==-1)
				{ 
					if(_direction==-1)
					{ 
						nextIndex = 0; 
						_direction = 1; //change the searching direction
					}
					else
					{
					    throw gear::UnknownParameterException("Unknown layer index");
					}
				}
			}
			
			nextMeasLayer = dynamic_cast<GearTPCMeasLayer const*> (_detectorCradle.At(nextIndex));
   
			if (nextMeasLayer==0)
			{
				throw bad_cast();
			}
       
			// only count active layers for the distance
	
			if (!nextMeasLayer->IsActive())
			{
				// An inactive layer. Just ignore it and continue with the next layer.
				// Do not increase/decrease the current distance.
				continue;
			}

			if(targetDistance >0)
				++currentDistance;
			else
				--currentDistance;

			streamlog_out(DEBUG3) << "targetDistance: " << targetDistance 
								  << ", currentDistance: "<< currentDistance << endl;
		}while (currentDistance != targetDistance);
    }//else targetDistance==0

    streamlog_out(DEBUG3) << "end of extrapolateIntoNextLayer()" << endl;

	return nextMeasLayer;
}


void KalTrackingHelix::printStateVector()
{
    streamlog_out(DEBUG4) << "dr=" << _containerTrack->GetCurSite().GetCurState()[0][0]
    << " phi0=" << _containerTrack->GetCurSite().GetCurState()[1][0]
    << " kappa=" << _containerTrack->GetCurSite().GetCurState()[2][0]
    << " dz=" << _containerTrack->GetCurSite().GetCurState()[3][0]
    << " tanl=" << _containerTrack->GetCurSite().GetCurState()[4][0]
    << endl
    << " : (ndf,chi2)=(" << _containerTrack->GetNDF()
    << " ," << _containerTrack->GetChi2() << ")"
    << endl;
}

void KalTrackingHelix::setLCIOTrackParameters(IMPL::TrackImpl* trk)
{
    streamlog_out(DEBUG3) << "------------------------------------" << endl
						  << "In KalTrackingHelix::setLCIOTrackParameters()" << endl
						  << "------------------------------------" << endl;

    //fg: smooth back to the beginning - choose site 4, as the the first 3
    //sites are probably not very reliable

    //_containerTrack->SmoothBackTo(4);

    TKalTrackState& trkState = dynamic_cast<TKalTrackState&>( _containerTrack->GetCurSite().GetCurState() );

    TVKalSite &cursite = _containerTrack->GetCurSite();

    Int_t    ndf  = _containerTrack->GetNDF();
    Double_t chi2 = _containerTrack->GetChi2();
    Double_t confidenceLevel   = TMath::Prob(chi2, ndf);


    //============== convert parameters to LCIO convention ====

    Double_t kappa  = trkState(2, 0);

	if(kappa!=0.) {
    THelicalTrack helix = trkState.GetHelix() ;


    // get the 5x5 sub matrix of the covariance matrix
    const TMatrixD& covK =  trkState.GetCovMat() ;

    //  --- uncomment to get parameters with reference point at origin - typically `not for tbeam...
    // double dPhi ;
    // helix.MoveTo(  TVector3( 0., 0., 0. ) , dPhi , 0 , &covK ) ;

    //-------------------------------------


    double phi       =    toBaseRange(helix.GetPhi0() + M_PI / 2.) ;
    double omega     =    1. / helix.GetRho()  ;
    double d0        =  - helix.GetDrho() ;
    double z0        =    helix.GetDz()   ;
    double tanLambda =    helix.GetTanLambda()  ;

    //covariance matrix in LCIO - stored as lower triangle matrix where the order of parameters is:
    // d0, phi, omega, z0, tan(lambda). So we have cov(d0,d0), cov( phi, d0 ), cov( phi, phi), ...


    // get alpha vor conversion of covariance parameters
    double alpha = kappa / omega;

    EVENT::FloatVec cov(15)  ;
    cov[ 0] =   covK(0 , 0)   ;   //   d0,   d0

    cov[ 1] = - covK(1 , 0)   ;   //   phi0, d0
    cov[ 2] =   covK(1 , 1)   ;   //   phi0, phi

    cov[ 3] = - covK(2 , 0) * alpha   ;   //   omega, d0
    cov[ 4] =   covK(2 , 1) * alpha   ;   //   omega, phi
    cov[ 5] =   covK(2 , 2) * alpha * alpha  ;   //   omega, omega

    cov[ 6] = - covK(3 , 0)   ;   //   z0  , d0
    cov[ 7] =   covK(3 , 1)   ;   //   z0  , phi
    cov[ 8] =   covK(3 , 2) * alpha   ;   //   z0  , omega
    cov[ 9] =   covK(3 , 3)   ;   //   z0  , z0

    cov[10] = - covK(4 , 0)   ;   //   tanl, d0
    cov[11] =   covK(4 , 1)   ;   //   tanl, phi
    cov[12] =   covK(4 , 2) * alpha  ;   //   tanl, omega
    cov[13] =   covK(4 , 3)   ;   //   tanl, z0
    cov[14] =   covK(4 , 4)   ;   //   tanl, tanl

    trk->setD0(d0) ;
    trk->setPhi(phi) ;    // fi0  - M_PI/2.  ) ;
    trk->setOmega(omega) ;
    trk->setZ0(z0) ;
    trk->setTanLambda(tanLambda) ;
    trk->setCovMatrix(cov) ;

    float pivot[3] ;
    pivot[0] = ((TKalTrackSite&) cursite).GetPivot()(0) ;
    pivot[1] = ((TKalTrackSite&) cursite).GetPivot()(1) ;
    pivot[2] = ((TKalTrackSite&) cursite).GetPivot()(2) ;

	 //((TKalTrackSite&) cursite).GetMeasNoiseMat().Print();

    trk->setReferencePoint(pivot) ;

    streamlog_out(DEBUG3) << " helical track parameters: "
    << " chi2/ndf: " << chi2 / ndf
    << " chi2: " <<  chi2
    << " conv. level: " << confidenceLevel << endl
	<< " kappa: " << kappa

    << "\t D0 "          <<  d0         <<  "[+/-" << sqrt(cov[0]) << "] "
    << "\t Phi :"        <<  phi        <<  "[+/-" << sqrt(cov[2]) << "] "
    << "\t Omega "       <<  omega      <<  "[+/-" << sqrt(cov[5]) << "] "
    << "\t Z0 "          <<  z0         <<  "[+/-" << sqrt(cov[9]) << "] "
    << "\t tan(Lambda) " <<  tanLambda  <<  "[+/-" << sqrt(cov[14]) << "] "

    << "\t pivot : [" << pivot[0] << ", " << pivot[1] << ", "  << pivot[2]
    << " - r: " << sqrt(pivot[0]*pivot[0] + pivot[1]*pivot[1]) << "]"
    << endl ;
	}
	else {

	TVector3 orig(0., 0., 0.);
	double t = 0.;
	std::unique_ptr<TVTrack> help(&trkState.CreateTrack()); // tmp track
	TVTrack &hel = *help;
	hel.MoveTo(orig, t);

    double phi       =    toBaseRange(hel.GetPhi0() + M_PI / 2.)  ;
    double omega     =    0.;
    double d0        =    -hel.GetDrho();//trkState(0, 0);
    double z0        =    hel.GetDz();//trkState(3, 0);
    double tanLambda =    hel.GetTanLambda();//trkState(4, 0);

    trk->setD0(d0) ;
    trk->setPhi(phi) ; 
    trk->setOmega(omega) ;
    trk->setZ0(z0) ;
    trk->setTanLambda(tanLambda) ;

	float pivot[3] = {0., 0., 0.};
    trk->setReferencePoint(pivot) ;

    streamlog_out(DEBUG3) << " straight track parameters: "
    << " chi2/ndf: " << chi2 / ndf
    << " chi2: " <<  chi2
    << " conv. level: " << confidenceLevel << endl
	<< " kappa: " << kappa

    << "\t D0 "          <<  d0       
    << "\t Phi :"        <<  phi     
    << "\t Omega "       <<  omega  
    << "\t Z0 "          <<  z0       
    << "\t tan(Lambda) " <<  tanLambda 

    << "\t pivot : [" << pivot[0] << ", " << pivot[1] << ", "  << pivot[2]
    << " - r: " << sqrt(pivot[0]*pivot[0] + pivot[1]*pivot[1]) << "]"
    << endl ;
	}

    trk->subdetectorHitNumbers().push_back(_hitsOnTrack.size()) ;
    trk->subdetectorHitNumbers().push_back(_hitsOnTrack.size()) ;
    trk->setChi2(chi2) ;
    trk->setNdf(ndf) ;


#if LCIO_VERSION_GE( 1 , 60 )
    streamlog_out(DEBUG4) << lcio::header(trk) << endl ;
#else
    streamlog_out(DEBUG4) << lcio::header(*trk) << endl ;
#endif

    streamlog_out(DEBUG4) << lcio::lcshort((lcio::Track*)trk) << endl ;
}

double KalTrackingHelix::calculateDeltaChi2(TKalTrack * trackCandidate)
{
  return trackCandidate->GetCurSite().GetDeltaChi2();
}

bool ModuleRowHitListSorter::operator()(ModuleRowHit const & mrh1,
							  ModuleRowHit const & mrh2) const
{
  //double const * pos1 = mrh1.hit->getPosition();
  double const * pos1 = mrh1.getHitPosition();
  double distanceSquare1 = (_referencePosition[0]-pos1[0]) * (_referencePosition[0]-pos1[0]) 
                          +(_referencePosition[1]-pos1[1]) * (_referencePosition[1]-pos1[1])
                          +(_referencePosition[2]-pos1[2]) * (_referencePosition[2]-pos1[2]);

  double const * pos2 = mrh2.getHitPosition();
  double distanceSquare2 = (_referencePosition[0]-pos2[0]) * (_referencePosition[0]-pos2[0]) 
                          +(_referencePosition[1]-pos2[1]) * (_referencePosition[1]-pos2[1])
                          +(_referencePosition[2]-pos2[2]) * (_referencePosition[2]-pos2[2]);

  return (distanceSquare1 < distanceSquare2);
}
  
}
#endif
