#ifdef USE_TRACK_FITTER_KALMAN
#include "KalTrackingTemplate.h"
 
#include <IMPL/TrackImpl.h>
#include <EVENT/LCCollection.h>
#include <streamlog/streamlog.h>

#include <kaldet/GearTPCKalDetector.h>
#include <kaldet/GearTPCCylinderMeasLayer.h>
#include <limits>

#include "TPCHitMap.h"

#include <StraightTrackParameters.h>

#include <gear/BField.h>

using namespace IMPL;
using namespace gear;
using namespace std;
using namespace EVENT;
using namespace kaldet;

namespace marlintpc
{

template <class ContainerT> KalTrackingTemplate<ContainerT>::KalTrackingTemplate(
				 const GearMgr &gearMgr, unsigned int minTrackHits,
                 unsigned int maxSkipRows, double maxDeltaChi2, 
				 int hitDistanceFactor, int startingHitMode, double vDrift, 
				 std::string modulesNotUsed)
                : _gearMgr(gearMgr), 
                  _tpcParameters(gearMgr.getTPCParameters()),
                  _minTrackHits(minTrackHits), 
				  _direction(OUTWARDS), 
				  _maxSkipRows(maxSkipRows),
                  _maxDeltaChi2(maxDeltaChi2),
				  _hitDistanceFactor(hitDistanceFactor),
				  _startingHitMode(startingHitMode),
		          _vDrift(vDrift),
                  _kalDetector(gearMgr),
				  _tpcHitMap(&gearMgr.getTPCParameters(), &_kalDetector, modulesNotUsed)
{   
    // We have to register the kalDetector with the cradle.
    // Afterwards we close the cradle (no more detectors can be added, we just have this one)
    // and sort it so all the Kalman layers are in the right order according to their sorting 
    // policy in the cradle.
    // After this the order in the cradle and the kalDetector are not necessarily the same any more.
    _detectorCradle.Install(_kalDetector);
    _detectorCradle.Close();
    _detectorCradle.Sort();

    //the hit domain can only be initialized after the cradle is intalled and sorted.
    _domain = new TPCHitDomain(&_detectorCradle, &_kalDetector, &gearMgr.getTPCParameters());

    _containerTrack = new ContainerT;
}

template <class ContainerT> KalTrackingTemplate<ContainerT>::~KalTrackingTemplate()
{
  streamlog_out(DEBUG3) << "In KalTrackingTemplate destructor" << endl;
  delete _domain;
  delete _containerTrack;
  streamlog_out(DEBUG3) << "end of KalTrackingTemplate destructor" << endl;
}

// Do not implement them. They cannot / will not be used because they are private
//KalTrackingTemplate::KalTrackingTemplate(const KalTrackingTemplate&)
//{
//}
//
//KalKalTrackingTemplate& KalTrackingTemplate::operator =(const KalTrackingTemplate&)
//{
//}

template <class ContainerT> 
list<EVENT::Track*> KalTrackingTemplate<ContainerT>::findTracks(
                                                      EVENT::LCCollection const * hitCollection)
{
	streamlog_out(DEBUG3)<< "======================================" << endl
						  << "In findTrack()"                         << endl
					      << "======================================" << endl;

    _tpcHitMap.fillEventHits(hitCollection);

    list<EVENT::Track*> trackList;

    while (!_tpcHitMap.isEmpty())
    {
       TrackImpl* aTrack = getTrack(_tpcHitMap.getFirstHit(_startingHitMode));

	   
       if (aTrack == 0)
	   {   
		   streamlog_out(DEBUG3) << "reset the track" << endl;
		   reset();
		   streamlog_out(DEBUG3) << "no problem with reset" << endl;
		   continue;
	   }

	   streamlog_out(DEBUG4) << "Found a track" << endl;
       printStateVector();

       trackList.push_back(aTrack);
    }

    return trackList;
}

template <class ContainerT> 
TrackImpl* KalTrackingTemplate<ContainerT>::getTrack(ModuleRowHit firstHit)
{
    streamlog_out(DEBUG3) << "======================================" << endl
                          << "In getTrack()"                         << endl
                          << "======================================" << endl;
	reset();

    if (searchTrack(firstHit))
    {
        streamlog_out(DEBUG3) << endl << "A track is found, containing "
			  << _hitsOnTrack.size()  << " hits" << endl;

        //save track
        TrackImpl* aTrack = new TrackImpl();
        // FIXME: set the track parameters

        // fill the hits into the track
        for (vector<ModuleRowHit>::iterator mrhIter = _hitsOnTrack.begin(); 
  		     mrhIter != _hitsOnTrack.end(); ++mrhIter)
        {
          aTrack->addHit( mrhIter->getTrackerHit() );
        }

        setLCIOTrackParameters(aTrack);

        // after deleting the hits the hitLists is void,
		// FIXME::put it to setLCIOTrackParameters
		//
        //remove hits belonging to the track from map
        _tpcHitMap.removeHits(_hitsOnTrack);
        _hitsOnTrack.clear();

        return aTrack;
    }
    else
    {
		//set the MarlinTPCHit in TPCHitMap not used
        for (vector<ModuleRowHit>::iterator mrhIter = _hitsOnTrack.begin(); 
  		     mrhIter != _hitsOnTrack.end(); ++mrhIter)
        {
          _tpcHitMap.setHitUsed( *mrhIter, false );
        }
		_hitsOnTrack.clear();

        //if not find a track, remove this hit
        _tpcHitMap.removeHit(firstHit);

		streamlog_out(DEBUG3) << "Found no track in this try, return null track" << endl;

        return 0;
    }
}

template <class ContainerT>
void KalTrackingTemplate<ContainerT>::searchTrackHit()//recursive
{
    streamlog_out(DEBUG3) << "---------------------------" << endl
                          << "In SearchTrackHit()"         << endl
                          << "---------------------------" << endl;

    // search for the next domain, max allow maxSkippedRows
    
    // the hit we are going to add.
    // Currently invalid because the pointer to the TrackerHit is 0.
    ModuleRowHit hit;

    //the initial value for seaeching a hit is according to the current direction
    int distance; // start with +- 1

    // If we just turned aroud we already checked the near hit, so we want the
    // far one this time.

	// Now we just compare every hit in the hit domain, and the used hits will never
	// come into this domain. So we don't need this flag any more.
    //bool useFarHit = false;

    // do not incearse the skip in the loop header.
    // In case we do not have an intersection and want to turn around we want 
    // to rerun with the same layer without increasing the skip
    unsigned int  skippedRows      = 1; //skipped rows in one direction
    unsigned int  totalSkippedRows = 1; //total skipped rows 

	bool findHit     = false;

	ModuleRowHit bestHit;

	while(1)
    {
		distance = _direction*skippedRows;
		streamlog_out(DEBUG3) << "SkippedRows: " << skippedRows << ", distance: " << distance << endl;

		const GearTPCMeasLayer* nextLayer = extrapolateIntoNextLayer(distance);

		//check the next measLayer
		if(!nextLayer)
		{
			streamlog_out(DEBUG3) << "End of tracking, since no proper measlayer any more" << endl;
			return;
		}

		set< pair<int, int> > moduleRowsInDomain = nextLayer->GetModuleRows();

        if ( moduleRowsInDomain.empty() )
        {
			streamlog_out(DEBUG3) << "No proper modulerows contains the extrapolated hit" << endl;
        }

        list<ModuleRowHit> hitList = _tpcHitMap.getHitsInDomain(moduleRowsInDomain);

		streamlog_out(DEBUG3) << "modulerow size is " 
						      << moduleRowsInDomain.size()
							  << ", and found " << hitList.size() 
							  << " hits in this domain." << endl;

        if ( hitList.empty() )
        {
			streamlog_out(DEBUG3) << "No hit in this domain" << endl;
        }

	    //choose the best hit which has minimum delta chi2
		double minChi2 = _maxDeltaChi2;
		/** TODO:(1) minHitsDistance is set further below, but not used.
		 *  the assignment below may be unnecessary, please check!
		 *  Commented, due to "set but not used" warning
		 */
		/// double minHitsDistance = numeric_limits<double>::max();
		list<ModuleRowHit> goodHitsList;

        for(list<ModuleRowHit>::iterator iter = hitList.begin(); iter != hitList.end(); iter++)	
		{ 
			ContainerT trackCandidate(*_containerTrack);
			trackCandidate.SetOwner(false); // a track candidate does not own the sites!
	  
		    if(doAddAndFilter(&trackCandidate, *iter))
			{
                /** TODO:(3) commented, as aHit does not seem to be
                 * used either:
                 */
				///ModuleRowHit aHit = *(_hitsOnTrack.rbegin());

				double candidateDChi2 = calculateDeltaChi2(&trackCandidate);
				/** TODO:(2) commented, as hitsDistance does not seem to be
				 * used either:
				 */
				///double hitsDistance =  TPCHitMap::getHitsDistance( *iter, aHit );

				if(candidateDChi2 > _maxDeltaChi2)
				{
					streamlog_out(DEBUG3) << "Filtered hit has a too large deltaCh2: " 
										  << candidateDChi2 << " maxDeltaChi2 is "
										  << _maxDeltaChi2 << endl;
				}
				else
				{
					//compare delta chi2
					if(candidateDChi2 < minChi2)
					{
						double angle = getTrackTangentAndLayerAngle(&trackCandidate);

						streamlog_out(DEBUG3) << "scalar product: " << angle
											  << ", direction: " << _direction << endl;

						//Obviously, the sign of track angle with layer should be consistent with 
						//that of tracking direction
						if(angle * (double)_direction > 0)
						{
							findHit = true;
							bestHit = *iter; 
							minChi2 = candidateDChi2;
							///TODO: this assignment may be unnecessary
							/// (see decl. above)
							/// minHitsDistance = hitsDistance;
							goodHitsList.push_back(*iter);
						}
						else
						{
							streamlog_out(DEBUG3) << "But the direction is wrong, hit is not added." << endl;
						}
					}
				}
		
				// The track candidate has done it's job. We now get rid of the dynamically
				// allocated last site because the track candidate is not the owner and
				// does not delete it when it goes out of scope.
				delete trackCandidate.Last();
			}

		}//for

		if(findHit)
		{
            //if(goodHitsList.size()>=2)streamlog_out(DEBUG3) << "Proper hits >=1 !!!" << endl;

			break;
		}
		else
		{
			streamlog_out(DEBUG3) << "Find no proper hit in the layer, skip row" << endl;

			//print the expected hit position
			

		    int lastLayerIndex= _tpcHitMap.getLayerIndex( (*_hitsOnTrack.rbegin()).module,
											 (*_hitsOnTrack.rbegin()).row );

			//
			if((int)skippedRows == lastLayerIndex-1 && _direction == -1)
			{
				//At layer 0, but not find hit yet.
				//Change to another direction, and set skippedRows to 0, 
				//the totalSkipped to maxSkipRows.
				_direction *= -1;
				skippedRows = 0;

				if(totalSkippedRows>_maxSkipRows)
				{
			        streamlog_out(DEBUG3) << "break here" << endl;
					break;
				}
				else
				{
					totalSkippedRows = _maxSkipRows;
					continue;
				}
			}

		    skippedRows++;
			totalSkippedRows++;

			//TODO:: what if index greater than module index of detector
			//if()
			//{
			//}
			
			if(skippedRows > _maxSkipRows) 
			{
				_direction *= -1; //change searching direction
				skippedRows = 0;
			}

			unsigned int maxSkip = 2*_maxSkipRows+1;
			if(totalSkippedRows > maxSkip)
			{
				streamlog_out(DEBUG3) << "total skipped row: "  << totalSkippedRows 
									  << ", maxSkip row: " << maxSkip 
					                  << ", stop skipping here" << endl;
				break;
			}
		}
    }//while skippedRows

	if(findHit)
	{
		// now add and filter the best hit with the real track.  
		doAddAndFilter(_containerTrack, bestHit); 
		_hitsOnTrack.push_back(bestHit);
		_tpcHitMap.setHitUsed(bestHit);

		//debug:
		//printStateVector();

		streamlog_out(DEBUG3) << "The hit number in the track now is " << _hitsOnTrack.size() << endl;
		
		streamlog_out(DEBUG3) << "Recurse to the next layer" << endl;
		searchTrackHit();
	}
	else
	{
		//skip all possible rows and still find no hit, then stop searching
		streamlog_out(DEBUG3) << "return no hit" << endl;
		return;
	}
}

/*
// for a generic container we just delete it and create a new one
template <class ContainerT>
void KalTrackingTemplate<ContainerT>::reset()
{
	_containerTrack->SetOwner(true);
    delete _containerTrack;

    _containerTrack = new ContainerT;

    _hitsOnTrack.clear();
}
*/

// special implementation for a KalTrack: We have to make it own the track states and clear it
// so the track states are also deleted.
template <>
void KalTrackingTemplate<TKalTrack>::reset()
{
    //Make the kalTrack owner of all the hits on it and all clear so they are all deleted.
    //Afterwards delete the track and create a fresh one (to reset all the internal valuables properly)
    _containerTrack->SetOwner(true);
    _containerTrack->Clear();
	_direction = OUTWARDS;
    delete _containerTrack;

    _containerTrack = new TKalTrack;

    for (vector<ModuleRowHit>::iterator mrhIter = _hitsOnTrack.begin(); 
  		     mrhIter != _hitsOnTrack.end(); ++mrhIter)
    {
          _tpcHitMap.setHitUsed( *mrhIter, false );
    }

    _hitsOnTrack.clear();
}

template <>
void KalTrackingTemplate<StraightTrackParameters>::reset()
{
	cout << "StraightTrackParameters reset" << endl;
	_direction = OUTWARDS;

    for (vector<ModuleRowHit>::iterator mrhIter = _hitsOnTrack.begin(); 
  		     mrhIter != _hitsOnTrack.end(); ++mrhIter)
    {
          _tpcHitMap.setHitUsed( *mrhIter, false );
    }

    _hitsOnTrack.clear();

    const double P1_init[4] = {1.0e10, 0, 0, 1.0e10};
	const double sv1_init[2] = {0, 0}; 
	const double PHI_init[4] = {1, 0, 0, 1}; 

	_containerTrack->_P1(0, 0) = P1_init[0];
	_containerTrack->_P1(0, 1) = P1_init[1];
	_containerTrack->_P1(1, 0) = P1_init[2];
	_containerTrack->_P1(1, 1) = P1_init[3];

	_containerTrack->_sv1(0, 0) = sv1_init[0];
	_containerTrack->_sv1(1, 0) = sv1_init[1];

	_containerTrack->_M_unit(0, 0) = PHI_init[0];
	_containerTrack->_M_unit(0, 1) = PHI_init[1];
	_containerTrack->_M_unit(1, 0) = PHI_init[2];
	_containerTrack->_M_unit(1, 1) = PHI_init[3];

	_containerTrack->_H[0][1] = 1;
	_containerTrack->_H_t[1][0] = 1;

	_containerTrack->_R(0, 0) = 0.4;
}

template <class ContainerT>
GearTPCHit * KalTrackingTemplate<ContainerT>::createGearHit( ModuleRowHit const & mrh ) const
{
  const double* hitPos = mrh.getHitPosition();
  TVector3 xv(hitPos[0], hitPos[1], hitPos[2]);

  TrackerHit* tHit = mrh.getTrackerHit();
  int module = mrh.module;
  int row   = mrh.row;

  GearTPCMeasLayer const * measlayerPtr = _kalDetector.GetMeasLayer(module, row);

  // FIXME: this assumes that the measurement space is two-dimensional
  double   meas [2];
  double   dmeas[2];

  TKalMatrix mv   = measlayerPtr->XvToMv(xv);
  double   z    = mv(1, 0);

  meas [0] = mv(0, 0);
  meas [1] = z;

  //FIXME: currently we take the measurement error from the layer, so all hits get the same
  // (z-depenent) error. This should be taken from the reconstructed hit, once it is set there correctly.
  dmeas[0] = measlayerPtr->GetSigmaX(z);
  dmeas[1] = measlayerPtr->GetSigmaZ(z);

  ///FIXME: This only uses the z-component
  ///FIXME: Use GlobalFieldProcessor instead?
  double b =  _gearMgr.getBField().at( Vector3D(hitPos[0], hitPos[1], hitPos[2]) )[2];
  
  streamlog_out(DEBUG3) << " (module,row)=(" << module << "," << row << ")" << endl
			<< "xv=" << xv.X() << " " << xv.Y() << " " << xv.Z() << endl
			<< "meas[0]=" << meas[0] << " meas[1]=" << meas[1] << endl;
  
  // the measlayerPtr 
  return measlayerPtr->createHit(meas, dmeas, static_cast<void *>(tHit), b, _vDrift);
}

template class KalTrackingTemplate<TKalTrack>;
template class KalTrackingTemplate<StraightTrackParameters>;

}// namespace
#endif
