#ifdef USE_TRACK_FITTER_KALMAN
#include "TPCHitDomain.h"
#include <streamlog/streamlog.h>

#include <kaldet/GearTPCCylinderMeasLayer.h>
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

using namespace kaldet;
using namespace std;

namespace marlintpc
{
  TPCHitDomain::TPCHitDomain(TKalDetCradle const * kalDetCradle, 
			     kaldet::GearTPCKalDetector const * kalDetector,
			     gear::TPCParameters const * tpcParameters)
    :_kalDetCradle(kalDetCradle), _kalDetector(kalDetector), _tpcParameters(tpcParameters)
	{}

	set< pair<int, int> > TPCHitDomain::getModuleRows ( TVector3 const & extrapolatedHitPosition) const
	{
		// In principle the question which modules and rows are "near" the extrapolated hit position
		// can be complicated. In the current implementation we use the fact that the KalDetector
		// has perfect cylinders and each layer known which module/rows belong to it. The hits
		// can only be on those module/rows.

		// Determine the kalman layer which belongs to the hit position.
		// First we need to know which module and row the hit position is on. We have to convert
		// the incoming hit position, which is cartesian, to the TPC coordinate system.

		vector<double> tpcCoordinates(3);
		
		switch (_tpcParameters->getCoordinateType())
		{
			case gear::PadRowLayout2D::CARTESIAN:
			//streamlog_out(DEBUG3) << "CARTESIAN" << endl;
			
			tpcCoordinates[0]=extrapolatedHitPosition[0];
			tpcCoordinates[1]=extrapolatedHitPosition[1];
			break;
      
			case gear::PadRowLayout2D::POLAR:
			tpcCoordinates[0]=sqrt( extrapolatedHitPosition[0]*extrapolatedHitPosition[0] + 
				extrapolatedHitPosition[1]*extrapolatedHitPosition[1] );
	
			tpcCoordinates[1]=atan2(extrapolatedHitPosition[1],extrapolatedHitPosition[0]);
			break;
      
			default:
			throw gear::UnknownParameterException("Unknown coordinate type");
		}
    
		tpcCoordinates[2]=extrapolatedHitPosition[2];

		gear::GlobalPadIndex globalPadIndex = _tpcParameters->getNearestPad(tpcCoordinates[0],
									tpcCoordinates[1],
									tpcCoordinates[2]);
		int moduleID = globalPadIndex.getModuleID();

		//
		//It seems that there is a problem when we get nearest pad near the boundary.
		//So if the extrapolated hit is not inside the nearest module or pad, return no modulerows.
		//Then the will skip it. 
		//
    
		if(!_tpcParameters->isInsidePad(tpcCoordinates[0], tpcCoordinates[1])) 
		{
			cout << "ext hit is not inside pad" << endl;

			set< pair<int, int> > modulerows;
			return modulerows;
		}

	    int row = _tpcParameters->getModule( moduleID ).getRowNumber( globalPadIndex.getPadIndex() );
		streamlog_out(DEBUG3) << "module: " << moduleID << ", row: " << row << endl;    

		return getModuleRows(moduleID, row);
	}

    set< pair<int, int> > TPCHitDomain::getModuleRows ( int moduleID, int row) const
	{
		const gear::TPCModule& module = _tpcParameters->getModule(moduleID);

		if(row >= module.getNRows())
		{
			set< pair<int, int> > modulerows;
		    return modulerows;
		}
   
		GearTPCMeasLayer const * domainMeasLayer = _kalDetector->GetMeasLayer(moduleID, row);

		if ( !domainMeasLayer->IsPerfect() )
		{
			streamlog_out(ERROR) << "Layer is not a perfect measurement layer.\n"
							     << "TPCHitDomain currently only works with perfect layers, sorry."
								 << endl;

			throw gear::NotImplementedException("TPCHitDomain only works with perfect layers");
		}

		// copy the set of module rows.
		return domainMeasLayer->GetModuleRows();
	}


}// namespace marlintpc


#endif //USE_TRACK_FITTER_KALMAN 
