#include "TPCHit_FixCellIDs_forTracking.h"
#include <iostream>
#include <cmath>

#include <EVENT/LCCollection.h>
#include <IMPL/TrackerHitImpl.h>
#include "IMPL/LCCollectionVec.h"
#include <UTIL/BitField64.h>
#include <UTIL/ILDConf.h>
#include <UTIL/LCTOOLS.h>
#include "LCObjectCopier.h"

//---- GEAR ----
#include "marlin/Global.h"
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/TPCModule.h"
#include "gear/BField.h"
#include "gearimpl/Vector3D.h"

using namespace lcio ;
using namespace marlin ;


namespace marlintpc
{
    TPCHit_FixCellIDs_forTracking aTPCHit_FixCellIDs_forTracking ;


    TPCHit_FixCellIDs_forTracking::TPCHit_FixCellIDs_forTracking() : Processor("TPCHit_FixCellIDs_forTracking") ,
        _nRun(0), _nEvt(0)
    {

        // modify processor description
        _description = "fix CellID0 for old TrackerHits ..." ;


        // register steering parameters: name, description, class-variable, default value
        registerInputCollection(LCIO::TRACKERHIT,
                                "TPCCollectionName" ,
                                "Name of the TPC TrackerHit collection"  ,
                                _tpcInputColName ,
                                std::string("AllTPCTrackerHits")
                               );

        registerOutputCollection(LCIO::TRACKERHIT,
                                 "OutputTrackerHits" ,
                                 "Name of the output TrackerHits collection"  ,
                                 _tpcOutputColName ,
                                 std::string("FixedCellIDTPCHits")) ;

        registerProcessorParameter("WriteOutputToStorage" ,
                                   "if true, the output hits collection is written to storage (default: true)"  ,
                                   _outputHitsPersistent,
                                   true);
                                   
        registerOptionalParameter("EncodeModuleID",
                                  "Encode module ID in CellID0",
                                  _encodeModuleID,
                                  bool(false));
    }


    void TPCHit_FixCellIDs_forTracking::init()
    {

        streamlog_out(DEBUG) << "   init called  " << std::endl ;

        // usually a good idea to
        printParameters() ;

        _nRun = 0 ;
        _nEvt = 0 ;

    }


    void TPCHit_FixCellIDs_forTracking::processRunHeader(lcio::LCRunHeader*  /*run*/)
    {

        _nRun++ ;
    }



    void TPCHit_FixCellIDs_forTracking::processEvent(lcio::LCEvent * evt)
    {


        UTIL::BitField64 encoder(ILDCellID0::encoder_string) ;


        //=====================================================================================================
        //      TPC
        //=====================================================================================================


        LCCollection* tpcCol = 0 ;

        try
            {
                tpcCol =  dynamic_cast<LCCollection*>(evt->getCollection(_tpcInputColName));

            }
        catch(lcio::DataNotAvailableException& e)
            {

                streamlog_out(DEBUG5) <<  " input collection not in event : " << _tpcInputColName << "   - nothing to do for TPC hits  !!! " << std::endl ;
            }

        if(tpcCol)
            {

                const gear::TPCParameters& gearTPC = Global::GEAR->getTPCParameters() ;

                UTIL::BitField64 encoder(lcio::ILDCellID0::encoder_string) ;

                int nHit = tpcCol->getNumberOfElements()  ;

                // create output hit collection and set the persistency flag
                LCCollectionVec* tpcHitCollection = new LCCollectionVec(LCIO::TRACKERHIT);
                tpcHitCollection->setTransient(!_outputHitsPersistent);

            	// set the flags that cellID1 should be stored
            	lcio::LCFlagImpl trkFlag(0);
            	trkFlag.setBit(lcio::LCIO::RTHBIT_ID1);
            	tpcHitCollection->setFlag(trkFlag.getFlag());

                for(int i = 0; i < nHit ; i++)
                    {

                        TrackerHitImpl* hit = LCObjectCopier::copy(dynamic_cast<TrackerHit*>(tpcCol->getElementAt(i))) ;

                        const double* pos = hit->getPosition();
                        // Note: here we take the Cartesian coordinates as specified in the gear file!
                        const gear::TPCModule& tpcmodule = gearTPC.getNearestModule(pos[0], pos[1]);


                        int mod = tpcmodule.getModuleID();
                        int pad = tpcmodule.getNearestPad(pos[0], pos[1]);
                        int row = tpcmodule.getRowNumber(pad);

			//specific implementation for the 7 modules of the LP1
			//we need to find a module for the lower, middle and upper row
			//to create global pad row numbers
			int npads_lowerRow=0; //module 0,1
			int npads_middleRow=0; //module 2,3,4
			try			  
			  {
			    npads_lowerRow=gearTPC.getModule(0).getNRows();
			  }
			catch(gear::Exception& e)
			  {
			    try
			      {
				npads_lowerRow=gearTPC.getModule(1).getNRows();
			      }
			    catch(gear::Exception& e){}
			  }
			try{
			  npads_middleRow=gearTPC.getModule(2).getNRows();
			}
			catch(gear::Exception& e)
			  {
			    try{
			      npads_middleRow=gearTPC.getModule(3).getNRows();
			    }
			    catch(gear::Exception& e)
			      {
				try{
				  npads_middleRow=gearTPC.getModule(4).getNRows();
				}
				catch(gear::Exception& e){}
			      } 
			  }
			
			int row_global =
                            // Modules 0 and 1 get the same row index
                            (mod == 0 || mod == 1) ? row :
                            /* For modules 2, 3 and 4 the row get increased
                             * by the number of rows in the modules 0/1
                             */
                            ((mod == 2 || mod == 3 || mod == 4) ? npads_lowerRow + row :
                             /* For modules 5 and 6 the row get increased
                              * by the number of rows in the modules 0/1 and 2/3/4
                              */
                              npads_lowerRow + npads_middleRow + row);


                        encoder.reset() ;  // reset to 0
                        encoder[ILDCellID0::subdet] = ILDDetID::TPC ;
                        encoder[ILDCellID0::layer] = row_global ;
                        if (_encodeModuleID) encoder[ILDCellID0::module] = mod ;
                        
                        int CellID = encoder.lowWord() ;
                        hit->setCellID0(CellID);

                        tpcHitCollection->addElement(hit);

                    }
                tpcHitCollection->parameters().setValue("CellIDEncoding" , ILDCellID0::encoder_string) ;
                evt->addCollection(tpcHitCollection, _tpcOutputColName);

            }

        streamlog_out(DEBUG3) << "   processing event: " << evt->getEventNumber()
                              << "   in run:  " << evt->getRunNumber() << std::endl ;
        _nEvt ++ ;
    }



    void TPCHit_FixCellIDs_forTracking::check(lcio::LCEvent *  /*evt*/)
    {
        // nothing to check here - could be used to fill checkplots in reconstruction processor
    }


    void TPCHit_FixCellIDs_forTracking::end()
    {

        streamlog_out(MESSAGE) << "TPCHit_FixCellIDs_forTracking::end()  " << name()
                               << " processed " << _nEvt << " events in " << _nRun << " runs "
                               << std::endl ;

    }

}
