#include "TPCHit_FixErrors_forTracking.h"
#include <iostream>
#include <cmath>

#include <EVENT/LCCollection.h>
#include <IMPL/TrackerHitImpl.h>
#include "IMPL/LCCollectionVec.h"
#include <UTIL/BitField64.h>
#include <UTIL/ILDConf.h>
#include <UTIL/LCTOOLS.h>
#include "LCObjectCopier.h"

//---- GEAR ----
#include "marlin/Global.h"
#include "gear/GEAR.h"
#include "gear/TPCParameters.h"
#include "gear/TPCModule.h"
#include "gear/BField.h"
#include "gearimpl/Vector3D.h"

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
#include "TPCConditions.h"


using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

    TPCHit_FixErrors_forTracking aTPCHit_FixErrors_forTracking ;


    TPCHit_FixErrors_forTracking::TPCHit_FixErrors_forTracking() : Processor("TPCHit_FixErrors_forTracking") ,
        _nRun(0), _nEvt(0), _tpcConditionsListener(0)
    {

        // modify processor description
        _description = "fix errors  for  TrackerHits from testbeam ..." ;


        // register steering parameters: name, description, class-variable, default value
        registerInputCollection(LCIO::TRACKERHIT,
                                "TPCCollectionName" ,
                                "Name of the TPC TrackerHit collection"  ,
                                _tpcInputColName ,
                                std::string("TPCHits")
                               );



        registerOutputCollection(LCIO::TRACKERHIT,
                                 "OutputTrackerHits" ,
                                 "Name of the output TrackerHits collection"  ,
                                 _tpcOutputColName ,
                                 std::string("FixedErrorTPCHits")) ;

        registerProcessorParameter("WriteOutputToStorage" ,
                                   "if true, the output hits collection is written to storage (default: true)"  ,
                                   _outputHitsPersistent,
                                   true);

    	registerOptionalParameter("InputTPCConditions",
    			"Optional: Name of the input collection containing the TPC conditions data",
    			_inputTPCConditionsCollectionName,
    			tpcconddata::TPCConditions::getDefaultColName()) ;

        registerProcessorParameter("offsetx", "sigma0 from fit in xy", _offsetx, double(0.02));
        registerProcessorParameter("slopex", "diffusion term from fit in xy", _slopex, double(0.0004));
        registerProcessorParameter("offsetz", "sigma0 from fit in z", _offsetz, double(0.4));
        registerProcessorParameter("slopez", "diffusion term from fit in z", _slopez, double(0.));
	registerProcessorParameter("offsetr", "sigma0 of radius", _offsetr, double(0.));
	registerProcessorParameter("sloper", "diffusion term for radius", _sloper, double(0.));

    }


    void TPCHit_FixErrors_forTracking::init()
    {

    	streamlog_out(DEBUG) << "   init called  " << std::endl ;
    	if(parameterSet("InputTPCConditions"))
    	{
    		try
    		{
    			_tpcConditionsListener = new TPCConditionsListener(_inputTPCConditionsCollectionName);
    			_offsetx=_tpcConditionsListener->getTransDefocussing();
    			_slopex=_tpcConditionsListener->getTransDiffusionCoef();
    			_offsetz=_tpcConditionsListener->getLongDefocussing();
    			_slopez=_tpcConditionsListener->getLongDiffusionCoef();
    		}

    		catch(lccd::LCCDException &)
    		{
    			streamlog_out(ERROR) << "Collection named \"" << _inputTPCConditionsCollectionName << "\" with TPCConditions is not available." << std::endl;
    			streamlog_out(ERROR) << "Use the ConditionsProcessor to provide it, or set the resolution parameter (and comment the InputTPCConditions parameter) if this is not possible." << std::endl;
    			throw;
    		}
    	}
    	else
    	{
    		streamlog_out(WARNING) << "The resolution parameters are taken from the XML input" << std::endl;
    	}
        // usually a good idea to
        printParameters() ;

        _nRun = 0 ;
        _nEvt = 0 ;

    }


    void TPCHit_FixErrors_forTracking::processRunHeader(lcio::LCRunHeader*  /*run*/)
    {

        _nRun++ ;
    }



    void TPCHit_FixErrors_forTracking::processEvent(lcio::LCEvent * evt)
    {


        //=====================================================================================================
        //      TPC
        //=====================================================================================================


        LCCollection* tpcCol = 0 ;

        try
            {
                tpcCol =  dynamic_cast<LCCollection*>(evt->getCollection(_tpcInputColName));

            }
        catch(lcio::DataNotAvailableException& e)
            {

                streamlog_out(DEBUG5) <<  " input collection not in event : " << _tpcInputColName << "   - nothing to do for TPC hits  !!! " << std::endl ;
            }

        if(tpcCol)
            {

                const gear::TPCParameters& gearTPC = Global::GEAR->getTPCParameters() ;

                int nHit = tpcCol->getNumberOfElements()  ;

                // create output hit collection and set the persistency flag
                LCCollectionVec* tpcHitCollection = new LCCollectionVec(LCIO::TRACKERHIT);
                tpcHitCollection->setTransient(!_outputHitsPersistent);
                
                // encoder for ModuleID
                UTIL::BitField64 encoder(lcio::ILDCellID0::encoder_string) ;
                
                // set the flags that cellID1 should be stored
                lcio::LCFlagImpl trkFlag(0);
                trkFlag.setBit(lcio::LCIO::RTHBIT_ID1);
                tpcHitCollection->setFlag(trkFlag.getFlag());

                for(int i = 0; i < nHit ; i++)
                    {

                        TrackerHitImpl* h = LCObjectCopier::copy(dynamic_cast<TrackerHitImpl*>(tpcCol->getElementAt(i))) ;
                        const double* pos = h->getPosition();
                        // Determine the covariance matrix according to the given resolution
                        double zdrift = gearTPC.getMaxDriftLength() - (std::abs(pos[2]));

                        // Sigma squared along Z
                        double zSigmaSquare = _offsetz  +  _slopez * zdrift;
			// Sigma squared in RPhi plane (azimuthal direction)
			double phiSigmaSquare = _offsetx + _slopex * zdrift;
			// Sigma squared in RPhi plane (radial direction)
			double rSigmaSquare = _offsetr + _sloper * zdrift;

                        // The phi angle of the hit. 
			const gear::TPCModule &module = Global::GEAR->getTPCParameters().getNearestModule(h->getPosition()[0],h->getPosition()[1]);
                        /* better: use ModuleID from encoder(CellID0)
                        encoder.setValue(h->getCellID0());
                        const gear::TPCModule &module = Global::GEAR->getTPCParameters().getModule(encoder[ILDCellID0::module]) ; // use encoder
                        */

                        double phi; // measurement direction
                        if (module.getLocalPadLayout().getCoordinateType() == gear::PadRowLayout2D::POLAR) {
  			   gear::Vector2D correction = module.getOffset();
                           //CHK (azimuthal) measurement direction = phi(hit) + pi/2 !
                           phi =  atan2(pos[0] - correction[0], correction[1] - pos[1]);
                           //std::cout << " Polar " << phi << std::endl;
                        } else {
                           //CHK (azimuthal) measurement direction = phi(module) + pi/2 !
                           phi = module.getAngle(); // + 0.5*M_PI; ??
                           //std::cout << " Cartesian " << phi << std::endl;
                        }

                        float covMatrix[6] = {0., 0., 0., 0., 0., 0.};

			covMatrix[0] = pow(cos(phi), 2) * phiSigmaSquare + pow(sin(phi), 2) * rSigmaSquare;
			covMatrix[1] = sin(phi) * cos(phi) * (phiSigmaSquare - rSigmaSquare);
			covMatrix[2] = pow(sin(phi), 2) * phiSigmaSquare + pow(cos(phi), 2) * rSigmaSquare;
			covMatrix[5] = float(zSigmaSquare);

                        h->setCovMatrix(covMatrix);
                        tpcHitCollection->addElement(h);
                    }
                evt->addCollection(tpcHitCollection, _tpcOutputColName);

            }

        streamlog_out(DEBUG3) << "   processing event: " << evt->getEventNumber()
                              << "   in run:  " << evt->getRunNumber() << std::endl ;

        _nEvt ++ ;
    }



    void TPCHit_FixErrors_forTracking::check(lcio::LCEvent *  /*evt*/)
    {
        // nothing to check here - could be used to fill checkplots in reconstruction processor
    }


    void TPCHit_FixErrors_forTracking::end()
    {

        streamlog_out(MESSAGE) << "TPCHit_FixErrors_forTracking::end()  " << name()
                               << " processed " << _nEvt << " events in " << _nRun << " runs "
                               << std::endl ;

    }


}
