#include "TimePixHoughTransformNormalProcessor.h"

namespace marlintpc
{

TimePixHoughTransformNormalProcessor aTimePixHoughTransformNormalProcessor;


TimePixHoughTransformNormalProcessor::TimePixHoughTransformNormalProcessor() : Processor("TimePixHoughTransformNormalProcessor")
{
  // the processor description
    _description =  "TimePixHoughTransformProcessor is intended to take an inputCollection of TrackerHits, "
                    "treat those TrackerHits in groups (based on their CellID1), "
                    "and reconstruct Tracks for each group separately. Which CellID1 is used can be specified setting \"ProcessedChipID\", "
                    "default is -1 meaning all CellID1s will be used.\n"
                    "Straight line is parameterized as:\n"
                    "\ty = x * tan(phi) + \frac{D0}{cos(phi)} = x * \frac{sin(phi)}{cos(phi)} + \frac{D0}{cos(phi)}\n"
                    "or in Hough space:\n"
                    "\tb = y * cos(phi) - x * sin(phi).";

        registerInputCollection(    LCIO::TRACKERHIT,
            "InputCollectionName" ,
            "Name of the input TrackerData collection, default: TimePixHits"  ,
            _inputColName ,
            string("TimePixHits")) ;


        registerOutputCollection(   LCIO::TRACK,
            "OutputCollectionName" ,
            std::string("Name of the output Track collection( default: TimePixTrackCandidates)"),
            _outputColName ,
            std::string("TimePixTrackCandidates")) ;

        registerProcessorParameter(
                "SetOutputTransient" ,
                "if not 0 the output collection is set transient (default 0)"  ,
                _outputIsTransient,
                0);

        registerProcessorParameter(
                "NHitsOnTrack" ,
                "minimum number of hits on one track"  ,
                _nHitsOnTrack,
                50);

        registerProcessorParameter(
                "MaximumDistanceToTrack" ,
                "Maximum distance of a hit to the track in mm in the xy plane"  ,
                _maxDistanceToTrackY,
                3.);

        registerProcessorParameter(
                "HoughSpaceNBinsPhi" ,
                "The number of bins in the Hough space histogram in 'phi' direction (default 200)"  ,
                _houghSpaceNBinsPhi,
                200);//100

        registerProcessorParameter(
                "HoughSpaceMaxPhi" ,
                "The half range (HoughSpaceMinPhi .. HoughSpaceMaxPhi) of the Hough space histogram in 'phi' direction (default +1.0 pi)"  ,
                _houghSpaceMaxPhi,
                float(+0.6 * 4.*atan(1.)));//+0.6 * pi

        registerProcessorParameter(
                "HoughSpaceMinPhi" ,
                "The half range (HoughSpaceMinPhi .. HoughSpaceMaxPhi) of the Hough space histogram in 'phi' direction (default -0.5 pi)"  ,
                _houghSpaceMinPhi,
                float(-0.5 * 4.*atan(1.)));//-0.5 * pi

        registerProcessorParameter(
                "HoughSpaceNBinsD0" ,
                "The number of bins in the Hough space histogram in 'D0' direction (default 200)"  ,
                _houghSpaceNBinsD0,
                200);//100

        registerProcessorParameter(
                "HoughSpaceMaxD0" ,
                "The half range (-max..max) of the Hough space histogram in 'D0' direction (default 65)"  ,
                _houghSpaceMaxD0,
                float(65.));

        registerProcessorParameter(
                "ProcessedChipID" ,
                "only the chip with _ChipID (CellID0) will be processed; if set to -1 all chipIDs will be processed (default: -1)"  ,
                _ChipID,
                (int) -1);

        registerProcessorParameter(
                "FileName" ,
                "Histograms will be written to this file (default: \"HoughSpaceXZ.root\")"  ,
                _fileName,
                string("HoughSpace.root"));

        registerProcessorParameter(
                "MaxHisto" ,
                "maximum number of events for which histograms will be saved (default: 30)"  ,
                _MaxHisto,
                (int) 100);

}


void TimePixHoughTransformNormalProcessor::init()
{ 
  streamlog_out(DEBUG) << " init() called  " << endl;

  /* initialize some variables */
  _CurrentHisto = 0;

  // you might want to list the used parameters by invoking
  printParameters();

  TFile file( _fileName.c_str(), "recreate");
  file.Close();

  std::ostringstream os;
  os << "MaxComvoluted_" << _ChipID;

  _hMaxConvoluted = new TH1D( os.str().c_str(),"maximal value bin in convoluted space;value of max bin;occurences",100,1,1000);

}


void TimePixHoughTransformNormalProcessor::processRunHeader(LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 4245 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
    {
        if (! i->second->isOptional() || i->second->valueSet())
        {
            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
} 


void TimePixHoughTransformNormalProcessor::processEvent(LCEvent* evt)
{
     // use an  LCCollection so we can use the index later
     // inputCol declared a member variable (so it can be used without passing to other functions)
     _inputColVec = 0;

     if( evt->getEventNumber() % 100 == 0 )
     {
         streamlog_out(MESSAGE9) << "ChipID: " << _ChipID << " Event: " << evt->getEventNumber() << std::endl;
     }
     else
     {
         streamlog_out(DEBUG) << "ChipID: " << _ChipID << " Event: " << evt->getEventNumber() << std::endl;
     }

     //
     // do a try / catch in case there is an empty event
     try
     {
         _inputColVec = dynamic_cast< LCCollectionVec * >(evt->getCollection( _inputColName ));
     }
     catch (DataNotAvailableException &e)
     {
         streamlog_out(WARNING0) << "EvtNo. " << evt->getEventNumber() << ", e.what(): " << e.what() << std::endl;
         return;
     }
     //
     // no in collection
     if( _inputColVec->getNumberOfElements() == 0 )
     {
         streamlog_out(WARNING0) << "InputCollection is empty" << std::endl;
         return;
     }

     // create the  output collections with a string defining the TYPE (see lcio.h)
     LCCollectionVec* outputCol = new LCCollectionVec(LCIO::TRACK);

     // add hits to track collection
     LCFlagImpl trkFlag(0) ;
     trkFlag.setBit(LCIO::TRBIT_HITS) ;
     outputCol->setFlag(trkFlag.getFlag());


     // set the transient flag as defined by the processor parameter
     outputCol->setTransient( (_outputIsTransient != 0) );

     /* get a new instance of the histogram */
     std::stringstream sName;
     sName << "HoughSpaceXY_Evt_#" << evt->getEventNumber() << "_ChipID_#" << _ChipID;
     _histoHoughSpace = new TH2D( sName.str().c_str()
                                , "d0 = cos(#Phi)*y - sin(#Phi)*x;#Phi [rad];d0 [mm]"
                                , _houghSpaceNBinsPhi, _houghSpaceMinPhi , _houghSpaceMaxPhi
                                , _houghSpaceNBinsD0, -_houghSpaceMaxD0, _houghSpaceMaxD0);
     _histoHoughSpace->SetStats(0);



     /* save hits to seperate vector */
     // get hits in this event
     for( LCCollectionVec::iterator inputObjectIter = _inputColVec->begin()
             ; inputObjectIter < _inputColVec->end()
             ; inputObjectIter++ )
     {
         // dynamic cast the input object (LCObject) to what it really is e.g.
         TrackerHit * myTrackerHit = dynamic_cast< TrackerHit * >( *inputObjectIter );

         // only hits with matching CellID0(chipID), unless specified differently
         if( (_ChipID != -1) && (_ChipID != (*myTrackerHit).getCellID1()) ) continue;

         _hitVec.push_back( myTrackerHit );
     }

     /* fill histogram */
     FillHistogram();

     /* get tracks and hits from histogram */
     GetTrackProperties();

     /* create tracks from properties */
     CreateTracks();

     /* add hits to tracks */
     AddHits();

     /* erase tracks with too less hits */
     RemoveSparseTracks( _trackImplsY, _trackImplsY.begin(), _trackImplsY.end() );

     /* store found tracks in outputCol */
     streamlog_out(MESSAGE4) << _trackImplsY.size() << " tracks found" << std::endl;
     OutputAddElement( outputCol, _trackImplsY.begin(), _trackImplsY.end() );



     /* add output collection */
     if( !outputCol->empty() )
     {
         evt->addCollection(outputCol, _outputColName);
     }
     else
     {
         delete outputCol;
     }

     /* clear vectors */
     _trackPropertiesY.clear();
     _trackImplsY.clear();


     return;
}



void TimePixHoughTransformNormalProcessor::RemoveSparseTracks( TrackImplVec & ImplVec, TrackImplVec::iterator  /*itFirst*/, TrackImplVec::iterator  /*itLast*/ )
{

    for( TrackImplVec::iterator itTrack = ImplVec.begin()
            ; itTrack != ImplVec.end()
            ; itTrack++ )
    {
        if( (int)(*itTrack)->getTrackerHits().size() < _nHitsOnTrack )
        {
            streamlog_out(DEBUG0) << "removing track with " << (int)(*itTrack)->getTrackerHits().size() << " < " << _nHitsOnTrack << std::endl;
            ImplVec.erase( itTrack );
            itTrack--;
        }
    }

    streamlog_out(MESSAGE1) << "found " << ImplVec.size() << " tracks with more than " << _nHitsOnTrack << " hits" << std::endl;

    return;
}



void TimePixHoughTransformNormalProcessor::OutputAddElement( LCCollectionVec *& outputCol, TrackImplVec::iterator itFirst, TrackImplVec::iterator itLast )
{


    if( itFirst != itLast )
    {
        streamlog_out(MESSAGE4) << "found track with " << (*itFirst)->getTrackerHits().size() << " hits"
                        << "\nphi " << (*itFirst)->getPhi()
                        << "\nd0  " << (*itFirst)->getD0()
                        << std::endl;

        outputCol->addElement( *itFirst );
        OutputAddElement( outputCol, ++itFirst, itLast );
    }
    else
    {
        streamlog_out(MESSAGE4) << outputCol->size() << " with more than " << _nHitsOnTrack << " hits" << std::endl;
    }

    return;
}



void TimePixHoughTransformNormalProcessor::FillHistogram()
{
    // loop hits
    for( TrackerHitVec::iterator itHit = _hitVec.begin()
            ; itHit < _hitVec.end()
            ; itHit++ )
    {
        const double * pos = (*itHit)->getPosition();

        // loop phi in b = cos(phi)*y - sin(phi)*x
        for( int nPhi = 0; nPhi < _houghSpaceNBinsPhi-1; nPhi++ )
        {
            double fPhi = _houghSpaceMinPhi + nPhi * ( _houghSpaceMaxPhi - _houghSpaceMinPhi ) / (double)( _houghSpaceNBinsPhi - 1 );
            double fInt = cos(fPhi) * pos[1] - sin(fPhi) * pos[0];

            _histoHoughSpace->Fill( fPhi, fInt );
        }
    }

    return;
}




void TimePixHoughTransformNormalProcessor::GetTrackProperties()
{
    //ConvoluteHistogram( _histoHoughSpace );

    /* loop entries of histogram, phi and d0; since histogram is choosen with a too large range, we don't need the edges[2,GetNBins()-1] */
    for( int nPhi = 3; nPhi < -1 + _histoHoughSpace->GetXaxis()->GetNbins(); nPhi++ )
    for( int nD0  = 3; nD0  < -1 + _histoHoughSpace->GetYaxis()->GetNbins(); nD0++  )
    {
        // save a trackProperty, if sufficient hits on track
        int nHitsOnTrack = _histoHoughSpace->GetBinContent( nPhi, nD0 );
        if( _nHitsOnTrack < nHitsOnTrack )
        {
            double fPhi = _histoHoughSpace->GetXaxis()->GetBinCenter( nPhi );
            double fD0  = _histoHoughSpace->GetYaxis()->GetBinCenter( nD0 );

            streamlog_out(DEBUG2) << "------------------------------------------"
                          << "\nSlope: " << tan(fPhi) << "\nIntersection: " << fD0/cos(fPhi)
                          << "\nPhi: " << fPhi << "\nD0: " << fD0
                          << "\nHitsOnTrack: " << nHitsOnTrack << std::endl;
            _trackPropertiesY.push_back( TrackProperties( fPhi, fD0, nHitsOnTrack ) );
        }
    }

    /* save histogram */
    if( _CurrentHisto < _MaxHisto )
    {
        _CurrentHisto++;
        _histoHoughSpaceVec.push_back( _histoHoughSpace );
    }
    else
    {
        //_histoHoughSpace->Delete();// Clear();
        delete _histoHoughSpace;
    }

    // sort TrackProperties
    streamlog_out(DEBUG2) << "number of tracks: " << _trackPropertiesY.size() << std::endl;
    std::sort( _trackPropertiesY.begin(), _trackPropertiesY.end(), sortPredicate );

    return;
}


void TimePixHoughTransformNormalProcessor::CreateTracks()
{
    if( _trackPropertiesY.empty() ) return;

    for( unsigned int nTrk = 0; nTrk < _trackPropertiesY.size(); nTrk++ )
    {
        if( nTrk < 5 )
            streamlog_out(DEBUG5) << "number of hits " << _trackPropertiesY[ nTrk ]._hitsontrack << std::endl;
        TrackProperties trkProp = _trackPropertiesY[ nTrk ];

        double fPhi = trkProp._Phi;
        double fD0  = trkProp._D0;

        // define Track in the 2-Dimensional readout plane
        TrackImpl * trkImpl = new TrackImpl();
        trkImpl->setPhi( fPhi );
        trkImpl->setD0( fD0 );

        // add track
        _trackImplsY.push_back( trkImpl );
    }
    _trackImplsY.resize( _trackPropertiesY.size() );

    return;
}



void TimePixHoughTransformNormalProcessor::AddHits()
{

    // first try to get a fitter, according to the fitter type defined in the track
    unsigned char fitterType = TrackFitterBase::FitterTypes::SIMPLEMINIMIZER;
    LCParameters const & trackCollectionParameters = _inputColVec->getParameters();
    TrackFitterBase *trackFitter = TrackFitterFactory::getTrackFitter( fitterType , &trackCollectionParameters);

    streamlog_out(DEBUG8) << "starting with " << _hitVec.size() << " hits" << std::endl;

    // loop hits
    for( TrackerHitVec::iterator itHit = _hitVec.begin()
            ; itHit < _hitVec.end()
            ; itHit++ )
    {
        //
        bool bAddedHit = false;

        // loop tracks
        for( TrackImplVec::iterator itTrk = _trackImplsY.begin()
                ; itTrk < _trackImplsY.end() && !bAddedHit
                ; itTrk++ )
        {

            // calculate residuals
            EVENT::DoubleVec globalResiduals = trackFitter->calculateResiduals( *itHit, *itTrk );
            // store hit to track if residuals are less than _maxDistanceToTrack
            if( pow(globalResiduals[0], 2) < pow(_maxDistanceToTrackY, 2) )
            {
                (*itTrk)->addHit( *itHit );
                bAddedHit = true;
            }
        }// for( unsigned int nTrk = 0; nTrk < _trackImpls.size(); ntrk++ )

        if( bAddedHit )
        {
            _hitVec.erase( itHit );
            itHit--;
        }
    }// for( TrackerHitVec::iterator itHit = _hitVec.begin()

    // these hits will not be used anymore
    streamlog_out(DEBUG8) << "left with " << _hitVec.size() << " hits" << std::endl;
    _hitVec.clear();


    return;
}

void TimePixHoughTransformNormalProcessor::check(LCEvent*  /*evt*/)
{ 

}


void TimePixHoughTransformNormalProcessor::end()
{ 
    //open file to save histograms
     TFile file( _fileName.c_str(), "update");
     if( !file.IsOpen() )
     {
         streamlog_out(WARNING0) << "file could not be opened" << std::endl;
     }
     else
     {
         streamlog_out(MESSAGE4) << "file opened" << std::endl;

         _hMaxConvoluted->Write();
         delete _hMaxConvoluted;

         for( std::vector< TH2D * >::iterator it = _histoHoughSpaceVec.begin()
                 ; it < _histoHoughSpaceVec.end()
                 ; it++)
         {
             if( (*it) == NULL )
                 continue;
             (*it)->GetXaxis()->SetLabelSize(0.04);
             (*it)->GetYaxis()->SetLabelSize(0.04);
             (*it)->GetXaxis()->SetTitleOffset( 0.7 );
             (*it)->GetYaxis()->SetTitleOffset( 0.7 );
             (*it)->GetXaxis()->SetTitleSize( 0.05 );
             (*it)->GetYaxis()->SetTitleSize( 0.05 );
             (*it)->Write();
             (*it)->Delete();
             streamlog_out(DEBUG5) << "histogram written" << std::endl;
         }

         // close file after histograms have been written
         streamlog_out(MESSAGE4) << "file closing " << _fileName << std::endl;
         file.Close();
         streamlog_out(MESSAGE4) << "file closed" << std::endl;
     }
     _histoHoughSpaceVec.clear();

     file.Close();
}


} // close namespace brackets
