#include "TrackFitterSimpleChiSquare.h"
#include "TrackFitterFactory.h"
#include <marlin/StringParameters.h>
#include <EVENT/LCRunHeader.h>
#include <gear/TPCParameters.h>

#include <streamlog/streamlog.h>

#include "LCObjectCopier.h"

#include <Minuit2/MnUserParameters.h>
#include <Minuit2/MnMigrad.h>
#include <Minuit2/FunctionMinimum.h>

#include "SimpleChiSquareResidualsFunction.h"

//#define CHI_SQUARE_TYPE_STRING "SimpleChiSquare"

namespace marlintpc
{
TrackFitterSimpleChiSquare * TrackFitterSimpleChiSquare::_theSimpleChiSquareFitter = 0;

TrackFitterSimpleChiSquare::TrackFitterSimpleChiSquare(double transDefocussing, double longDefocussing,
                                                       double transDiffusion, double longDiffusion,
                                                       double zAnode)
        : _transDefocussing(transDefocussing), _longDefocussing(longDefocussing),
        _transDiffusion(transDiffusion), _longDiffusion(longDiffusion), _zAnode(zAnode)
{
}

TrackFitterBase * TrackFitterSimpleChiSquare::getInstance(LCParameters const * parameters)
{
    // try to read the fitter parameters from the LCParameters
    EVENT::FloatVec transDefocusingVals;
    parameters->getFloatVals("SimpleChiSquare_TransDefocussing", transDefocusingVals);

    float transDefocussing;
    if (!transDefocusingVals.empty())
        {
            transDefocussing = transDefocusingVals[0];
        }
    else
        transDefocussing = 1.;

    EVENT::FloatVec longDefocusingVals;
    parameters->getFloatVals("SimpleChiSquare_LongDefocussing", longDefocusingVals);

    double longDefocussing;
    if (!longDefocusingVals.empty())
        {
            longDefocussing = longDefocusingVals[0];
        }
    else
        longDefocussing = 1.;

    // for the diffusion we can accept the default of 0.0 if parameter not found
    double transDiffusion = parameters->getFloatVal("SimpleChiSquare_TransDiffusionCoef");
    double longDiffusion  = parameters->getFloatVal("SimpleChiSquare_LongDiffusionCoef");

    // if the parameter is not found it is assumed to be 0
    double zAnode = parameters->getFloatVal("SimpleChiSquare_ZAnode");

    if (_theSimpleChiSquareFitter)
        {
            // reinitialise the fitter
            new(_theSimpleChiSquareFitter) TrackFitterSimpleChiSquare(transDefocussing, longDefocussing,
                                                                      transDiffusion, longDiffusion,
                                                                      zAnode);
        }
    else
        {
            _theSimpleChiSquareFitter = new TrackFitterSimpleChiSquare(transDefocussing, longDefocussing,
                                                                       transDiffusion, longDiffusion,
                                                                       zAnode);
        }

    return _theSimpleChiSquareFitter;
}

std::string TrackFitterSimpleChiSquare::getRevision() const
{
    return std::string("$Rev: 2253 $ ; TrackFitterBase ") + TrackFitterBase::getRevision();
}

unsigned char TrackFitterSimpleChiSquare::getFitterType() const
{
    return TrackFitterBase::FitterTypes::SIMPLECHISQUARE;
}


double TrackFitterSimpleChiSquare::getTransDefocussing() const
{
    return _transDefocussing;
}

double TrackFitterSimpleChiSquare::getLongDefocussing() const
{
    return _longDefocussing;
}

double TrackFitterSimpleChiSquare::getTransDiffusionCoef() const
{
    return _transDiffusion;
}

double TrackFitterSimpleChiSquare::getLongDiffusionCoef() const
{
    return _longDiffusion;
}

double TrackFitterSimpleChiSquare::getZAnode() const
{
    return _zAnode;
}

IMPL::TrackImpl * TrackFitterSimpleChiSquare::fitTrack(EVENT::Track const * seedTrack) const
{
    streamlog_out(DEBUG2) << "TrackFitterSimpleChiSquare: copying track" << std::endl;
    // create a copy of the track which is going to be fitted
    IMPL::TrackImpl * fittedTrack = LCObjectCopier::copy(seedTrack);

    streamlog_out(DEBUG2) << "TrackFitterSimpleChiSquare: setting minuit parameters" << std::endl;

    // prepare the starting parameters of the fit
    ROOT::Minuit2::MnUserParameters minuitParameters;
    if (_useD0Start)
        {
            minuitParameters.Add("D0", _d0Start,           0.1);
        }
    else
        {
            minuitParameters.Add("D0", seedTrack->getD0(), 0.1);
        }

    if (_usePhiStart)
        {
            minuitParameters.Add("Phi", _phiStart,           0.1);
        }
    else
        {
            minuitParameters.Add("Phi", seedTrack->getPhi(), 0.1);
        }

    if (_useOmegaStart)
        {
            minuitParameters.Add("Omega", _omegaStart ,          0.01);
        }
    else
        {
            minuitParameters.Add("Omega", seedTrack->getOmega(), 0.01);
        }

    if (!_noZFit)
        {

            if (_useTanLambdaStart)
                {
                    minuitParameters.Add("TanLambda", _tanLambdaStart,         0.1);
                }
            else
                {
                    minuitParameters.Add("TanLambda", seedTrack->getTanLambda(), 0.1);
                }

            if (_useZ0Start)
                {
                    minuitParameters.Add("Z0", _z0Start,           0.1);
                }
            else
                {
                    minuitParameters.Add("Z0", seedTrack->getZ0(), 0.1);
                }
        }

    //one full circle should be enough, but depending how the track parameters are set
    //there might be negative starting values...
    minuitParameters.SetLimits("Phi", -2*M_PI, 2*M_PI);

    /// !!!FIXME!!! don't hardcode these limits
    minuitParameters.SetLimits("D0", -100, 100);
    minuitParameters.SetLimits("Omega", -1., 1.);

    if (!_noZFit)
        {
            minuitParameters.SetLimits("TanLambda", -100, 100);
            minuitParameters.SetLimits("Z0", -1000, 1000);
        }

    streamlog_out(DEBUG2) << "TrackFitterSimpleChiSquare: create theFCN" << std::endl;

    SimpleChiSquareResidualsFunction theFCN(seedTrack, this, _zAnode);

    streamlog_out(DEBUG2) << "TrackFitterSimpleChiSquare: creating migrad" << std::endl;
    // create the minuit instance
    ROOT::Minuit2::MnMigrad migrad(theFCN,
                                   minuitParameters);

    if (_fixD0)
        {
            migrad.Fix("D0");
        }
    if (_fixPhi)
        {
            migrad.Fix("Phi");
        }
    if (_fixOmega)
        {
            migrad.Fix("Omega");
        }

    if (!_noZFit)
        {
            if (_fixTanLambda)
                {
                    migrad.Fix("TanLambda");
                }
            if (_fixZ0)
                {
                    migrad.Fix("Z0");
                }
        }

//    migrad.Fix(2);
//    migrad.Fix(3);

    streamlog_out(DEBUG2) << "TrackFitterSimpleChiSquare: performing fit" << std::endl;

    // perform the minimisation
    ROOT::Minuit2::FunctionMinimum fitResult = migrad();

    if (fitResult.IsValid())
        {
            streamlog_out(DEBUG2) << "TrackFitterSimpleChiSquare: setting track parameters" << std::endl;
            streamlog_out(MESSAGE0) << "TrackFitterSimpleChiSquare: chi square is " << fitResult.Fval()
            << " , NDoF is " << fittedTrack->getTrackerHits().size()
            << std::endl;

            // although in the documentation the << operator is not implemented.
            // streamlog_out(DEBUG4) << fitResult << std::endl;

            fittedTrack->setD0(fitResult.UserParameters().Value("D0"));
            fittedTrack->setPhi(fitResult.UserParameters().Value("Phi"));
            fittedTrack->setOmega(fitResult.UserParameters().Value("Omega"));

            if (!_noZFit)   // take Z0 and tanLambda from fit
                {
                    fittedTrack->setTanLambda(fitResult.UserParameters().Value("TanLambda"));
                    fittedTrack->setZ0(fitResult.UserParameters().Value("Z0"));
                }
            else // tanLambda and Z0 have not been fitted. Take them from the seed track or the given, fixed value
                {
                    if (_useTanLambdaStart)
                        {
                            fittedTrack->setTanLambda(_tanLambdaStart);
                        }
                    else
                        {
                            fittedTrack->setTanLambda(seedTrack->getTanLambda());
                        }

                    if (_useZ0Start)
                        {
                            fittedTrack->setZ0(_z0Start);
                        }
                    else
                        {
                            fittedTrack->setZ0(seedTrack->getZ0());
                        }
                }

            fittedTrack->setChi2(fitResult.Fval());

        }
    else
        {
            streamlog_out(WARNING1) << "TrackFitterSimpleChiSquare: Fit failed, chi square is "
            << fitResult.Fval() << std::endl;

            fittedTrack->setTypeBit(FITFAILEDBIT);
        }      // don't change the seed parameters. The ChiSquare is zero, so these tracks can be
    // identified.
    float refPoint[3]={0,0,0};
    fittedTrack->setReferencePoint(refPoint);


    // finally set the fitter type in the trackType word
    setTrackFitterType(this->getFitterType(), fittedTrack);

    return fittedTrack;

}


}//namespace marlintpc
