#include "TrackFitterSimpleHelixProcessor.h"
#include <iostream>
#include <string>
#include <cmath>
#include <algorithm>

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include "EVENT/LCIO.h"
#include "EVENT/Track.h"
#include "EVENT/LCCollection.h"
#include <IMPL/LCCollectionVec.h>
#include "IMPL/TrackImpl.h"
#include <IMPL/LCFlagImpl.h>
#include <Exceptions.h>

// Marlin
#include "marlin/Global.h"
#include <gear/TPCParameters.h>
#include "TrackFitterBase.h"

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{

TrackFitterSimpleHelixProcessor aTrackFitterSimpleHelixProcessor;


TrackFitterSimpleHelixProcessor::TrackFitterSimpleHelixProcessor() : Processor("TrackFitterSimpleHelixProcessor"),_trackCollectionParameters(0)
{

	_description = "Fits a simple Helix fit to the TrackerHits in the track candidates collection!";


	registerInputCollection(    LCIO::TRACK,
			"InputCollectionName" ,
			"Name of the input TrackerData collection, default: TPCSeedTracks"  ,
			_inputColName ,
			string("TPCSeedTracks")) ;


	registerOutputCollection(   LCIO::TRACK,
			"OutputCollectionName" ,
			"Name of the output TrackerData collection, default: TPCHelixTracks",
			_outputCollectionName ,
			string("TPCHelixTracks")) ;

	registerProcessorParameter("SetOutputTransient" ,
			"if not 0 the output hits collection is set transient (default: 0)",
			_outputIsTransient,
			int(0));

	registerOptionalParameter("D0Start",
			"Use this start value for D0 instead of the value from the seed track",
			_d0Start,
			float(0.));

	registerOptionalParameter("PhiStart",
			"Use this start value for phi instead of the value from the seed track",
			_phiStart,
			float(0.));

	registerOptionalParameter("OmegaStart",
			"Use this start value for omega instead of the value from the seed track",
			_omegaStart,
			float(0.));

	registerOptionalParameter("TanLambdaStart",
			"Use this start value for tanLambda instead of the value from the seed track",
			_tanLambdaStart,
			float(0.));

	registerOptionalParameter("Z0Start",
			"Use this start value for z0 instead of the value from the seed track",
			_z0Start,
			float(0.));

}


void TrackFitterSimpleHelixProcessor::init()
{ 
	streamlog_out(DEBUG) << " init() called  " << endl;

	printParameters();

	// Initialize the collection parameters
	_trackCollectionParameters = new IMPL::LCParametersImpl();

	// set the parameters for this fitter
	_trackCollectionParameters->setValue("TrackFitterType",
			"SimpleHelixFit");

}


void TrackFitterSimpleHelixProcessor::processRunHeader(LCRunHeader* run)
{
	run->parameters().setValue(_processorName + "_revision", "$Rev:$");

	for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
	{
		if (! i->second->isOptional() || i->second->valueSet())
		{
			run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
		}
	}


} 


void TrackFitterSimpleHelixProcessor::processEvent(LCEvent* evt)
{ 

	streamlog_out(DEBUG4) << "processEvent" << std::endl;

	IMPL::LCCollectionVec* seedTracks = 0;
	try
	{
		// the tracker hit collection
		seedTracks =
				dynamic_cast<IMPL::LCCollectionVec*>
		(evt->getCollection(_inputColName));
	}
	catch (DataNotAvailableException &)
	{
		streamlog_out(MESSAGE1) << "No seed tracks collection \"" << _inputColName
				<< "\" in event " << evt->getEventNumber() << std::endl;
		return;
	}

	// a collection to hold the new track
	IMPL::LCCollectionVec* outputTrackCollection = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);

	// set flag that the hits are stored in with the track
	// make sure the hit collection is not transient
	LCFlagImpl trkFlag(0) ;
	trkFlag.setBit(LCIO::TRBIT_HITS) ;
	outputTrackCollection->setFlag(trkFlag.getFlag());

	// set parameters for the output collection
	// (rather complicated because the assignment operator cannot be use (it's lcio :-( )

	StringVec integerKeys;
	_trackCollectionParameters->getIntKeys(integerKeys);
	for (StringVec::iterator it = integerKeys.begin();
			it < integerKeys.end(); it ++)
	{
		outputTrackCollection->parameters().setValue(*it,
				_trackCollectionParameters->getIntVal(*it));
	}

	StringVec floatKeys;
	_trackCollectionParameters->getFloatKeys(floatKeys);
	for (StringVec::iterator it = floatKeys.begin();
			it < floatKeys.end(); it ++)
	{
		outputTrackCollection->parameters().setValue(*it,
				_trackCollectionParameters->getFloatVal(*it));
	}

	StringVec stringKeys;
	_trackCollectionParameters->getStringKeys(stringKeys);
	for (StringVec::iterator it = stringKeys.begin();
			it < stringKeys.end(); it ++)
	{
		outputTrackCollection->parameters().setValue(*it,
				_trackCollectionParameters->getStringVal(*it));
	}

	streamlog_out(DEBUG3) << "requesting fitter" << std::endl;

	// request a fitter from the factory
	// a pointer to this event is passed so conditions data can be taken into
	// account
	TrackFitterBase *fitter = TrackFitterFactory::getTrackFitter(TrackFitterBase::FitterTypes::SIMPLEHELIXFIT,
			_trackCollectionParameters);

    if (parameterSet("D0Start"))
        {
            fitter->setD0Start(_d0Start);
        }
    if (parameterSet("PhiStart"))
        {
            fitter->setPhiStart(_phiStart);
        }
    if (parameterSet("OmegaStart"))
        {
            fitter->setOmegaStart(_omegaStart);
        }
    if (parameterSet("TanLambdaStart"))
        {
            fitter->setTanLambdaStart(_tanLambdaStart);
        }
    if (parameterSet("Z0Start"))
        {
            fitter->setZ0Start(_z0Start);
        }

	streamlog_out(DEBUG3) << "fitter is at " << fitter << " fitter type "
			<< static_cast<unsigned short>(fitter->getFitterType()) << std::endl;

	//loop all tracks
	for (IMPL::LCCollectionVec::iterator trackIter = seedTracks->begin();
			trackIter < seedTracks->end() ; trackIter++)
	{
		EVENT::Track *inputTrack = dynamic_cast<EVENT::Track *>(*trackIter);

		if (inputTrack == 0)
		{
			streamlog_out(WARNING2) << "bad_cast: Track candidates collection"
					<< _inputColName
					<< " does not contain EVENT::Tracks!" << std::endl;
			throw DataNotAvailableException("No track candidates found.");
		}


		streamlog_out(DEBUG3) << "fitting track" << std::endl;

		IMPL::TrackImpl* fittedTrack = fitter->fitTrack(inputTrack);

		streamlog_out(DEBUG3) << "fitting done" << std::endl;

		//determine dEdx:
		//===============
		//FIXME: This currently calculates the mean charge per hit,
		// not the dE/dx because it is difficult to calculate the actual
		// track length in a general formula

		TrackerHitVec const & hitVec = fittedTrack->getTrackerHits();

		//      double phi =  outputTrack->getPhi();
		//      double d0 =  outputTrack->getD0();
		double Q_sum = 0;

		for (TrackerHitVec::const_iterator hit = hitVec.begin();
				hit < hitVec.end(); hit++)
		{
			Q_sum += (*hit)->getEDep();
		}

		fittedTrack->setdEdx(Q_sum / hitVec.size());

		outputTrackCollection->push_back(fittedTrack);

	} // loop tracks
	evt->addCollection(outputTrackCollection, _outputCollectionName);

}


void TrackFitterSimpleHelixProcessor::check(LCEvent*  /*evt*/)
{ 
	// check what is needed here
}


void TrackFitterSimpleHelixProcessor::end()
{ 
	// do something at the end of the processing, e.g. cleaning up
}
} // close namespace brackets
