#ifdef USE_TRACK_FITTER_KALMAN
#include "TrackMakingKalmanFilterProcessor.h"

#include <cmath>
#include <map>
#include <list>

#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <marlin/Global.h>
#include <marlin/AIDAProcessor.h>
#include <marlin/Exceptions.h>

#include <gearimpl/RectangularPadRowLayout.h>
#include <gear/PadRowLayout2D.h>
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gearimpl/TPCParametersImpl.h>
#include <gear/BField.h>

#include "KalTrackingBase.h"
#include "KalTrackingHelix.h"
#include "KalTrackingStraight.h"


using namespace marlin ;
using namespace std;

namespace marlintpc
{

TrackMakingKalmanFilterProcessor aTrackMakingKalmanFilterProcessor;

TrackMakingKalmanFilterProcessor::TrackMakingKalmanFilterProcessor() : Processor("TrackMakingKalmanFilterProcessor")
{
    _description = "For simple track finding and fitting by Kalman filter!";

    registerInputCollection(LCIO::TRACKERHIT,
                            "InputTrackerHits" ,
                            "Name of the Input TrackerHits collection"  ,
                            _InputTrackerHitsCollectionName ,
                            std::string("TPCHits")) ;

    registerOutputCollection(LCIO::TRACK,
                             "OutputTracks" ,
                             "Name of the output Tracks collection"  ,
                             _outputTracksCollectionName ,
                             std::string("TPCTracks")) ;

    registerProcessorParameter("MinTrackHits" ,
                               "Minimum number of hits on track (default: 15)",
                               _minTrackHits ,
                               int(15)) ;

    registerProcessorParameter("MaxSkipRows" ,
                               "Maximum number of subsequnetly missing hits (default: 3, maximum: 5)",
                               _maxSkipRows ,
                               int(3)) ;

    registerProcessorParameter( "MaxDeltaChi2" ,
                               "Maximum number of chi2 increment (default: 30.)",
                               _maxDeltaChi2 ,
                               double(30.) ) ;

    registerProcessorParameter( "ModulesNotUsed" ,
                                "The ID of modules which are not used in track reconstruction",
                                _modulesNotUsed ,
                                std::string("-1") ) ;

     registerProcessorParameter("DriftVelocityOverride" ,
                               "Use this drift velocity instead of the one from conditions data [mm/us]",
                               _vDrift ,
                               double(79.) ) ;

   registerProcessorParameter("TrackType" ,
                               "Track type (default: 1, 0=straight line 1=helix)",
                               _trackType ,
                               int(1)) ;

   registerProcessorParameter("HitDistanceFactor" ,
                               "Factor for comparing the hit distances when building track seed(default: 2)",
                               _hitDistanceFactor ,
                               int(2)) ;

   registerProcessorParameter("StartingHitMode" ,
                               "Mode of getting the first hit(defaul: 1, by layer index)" ,
                               _startingHitMode ,
                               int(1)) ;
}


TrackMakingKalmanFilterProcessor::~TrackMakingKalmanFilterProcessor()
{
}


void TrackMakingKalmanFilterProcessor::processRunHeader(LCRunHeader *run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2545 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }
}


void TrackMakingKalmanFilterProcessor::init()
{
    printParameters();

    _tpcParameters = & Global::GEAR->getTPCParameters();

    // _hitMap = new TPCHitMap(_tpcParameters, _InputTrackerHitsCollectionName);
    if (!parameterSet("DriftVelocityOverride"))
    {
      streamlog_out(WARNING) << "Parameter DriftVelocityOverride not set."
			     << " This is currently the only way to get it in (conditions data"
			     << " not implemented yet). I will use " << _vDrift 
			     << " mm/us." << std::endl;
    }

    switch (_trackType)
    {
      case 0: // straight line
	_trackTuple = AIDAProcessor::tupleFactory(this)->create("tracktuple",
								"tracktuple",
								"double slope, double intercept");
	_kalTracking = new KalTrackingStraight(*Global::GEAR, 
					       _minTrackHits, 
					       _maxSkipRows, 
					       _maxDeltaChi2,
						   _hitDistanceFactor, 
						   _startingHitMode,
					       _vDrift,
						   _modulesNotUsed);
	break;

      case 1: // helix
	_trackTuple = AIDAProcessor::tupleFactory(this)->create("tracktuple",
								"tracktuple",
					"double kappa, int ndf, double chi2, double cl");
	_kalTracking = new KalTrackingHelix(*Global::GEAR, 
					     _minTrackHits, 
					     _maxSkipRows, 
					     _maxDeltaChi2,
						 _hitDistanceFactor,
						 _startingHitMode,
					     _vDrift,
						 _modulesNotUsed);
	break;

      default:
	streamlog_out(ERROR) << "Unknown track type " << _trackType << "!"
			     << " Only 0 (straight line) and 1 (helix) are allowed." 
			     << std::endl;
	throw StopProcessingException(this);
    }

    if(_maxSkipRows > 5)
    {
      streamlog_out(WARNING) << "Attention: MaxSkipRows > 5 might lead to bad tracking results"
			     << std::endl;
    }

}


void TrackMakingKalmanFilterProcessor::processEvent(LCEvent *evt)
{
    LCCollection * inputCollection;
    try
    {
      inputCollection = evt->getCollection(_InputTrackerHitsCollectionName);
    }
    catch (DataNotAvailableException &e)
    {
      // no input data, just return and process the next event
      return;
    }

    LCCollectionVec *_tpcTracksCollection = new LCCollectionVec(LCIO::TRACK);//deletion is left to the event

    //fg: this ensures that the hits will be written to the LCIO file
    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(LCIO::TRBIT_HITS) ;
    _tpcTracksCollection->setFlag(trkFlag.getFlag()) ;

    //FIXME
    //fill map with hits of this event
    std::list<EVENT::Track*> tracksList = _kalTracking->findTracks(inputCollection);

    for (std::list<EVENT::Track*>::iterator trackIter = tracksList.begin();
	 trackIter != tracksList.end(); ++trackIter)
    {
	   	// we are lazy and don't use a switch here.
        // In init we made sure that it can only be either of the two. 
        // If you introcude more track types please change to swith instead of
        // nested if/then/else.
		Track *track = *trackIter;
	   	if (_trackType == 0)
		{
			double slope = tan(track->getPhi());
			double intercept =  track->getD0() / sin(track->getPhi());
	        
			_trackTuple -> fill(0, slope);
			_trackTuple -> fill(1, intercept);
			_trackTuple -> addRow();
	    }
        else
        {
		  	double omega = track->getOmega();
			double b = Global::GEAR->getBField().at( gear::Vector3D( 0., 0., 0. ))[2];
			double kappa;
			
			if(b==0.) {
				if(omega==0.) {
					kappa = 0.;
				}
				else {
					streamlog_out(ERROR3) << "B=0, but kappa is not 0!" << endl;
					continue;
				}
			} 
			else { 
				kappa = omega * 1000 // in meter
				         *1.e9/(b * TMath::C());
			}

			double chi2 = track->getChi2();
			int    ndf  = track->getNdf();
			double confidenceLevel = TMath::Prob(chi2, ndf);

	        _trackTuple -> fill(0, kappa);
		    _trackTuple -> fill(1, ndf);
	        _trackTuple -> fill(2, chi2);
	        _trackTuple -> fill(3, confidenceLevel);
	        _trackTuple -> addRow();
        }

        _tpcTracksCollection->addElement(track);
    }

    //save track collection
    int trackNum = _tpcTracksCollection->getNumberOfElements();
    if (trackNum != 0)
    {
      evt->addCollection(_tpcTracksCollection, _outputTracksCollectionName);
      streamlog_out(DEBUG3) << "Event " << evt->getEventNumber()
			    << "  has " << trackNum << " tracks"
			    << endl;
    }
    else
    {
      // we have to delete the output collection if we do not add it to the event
      delete _tpcTracksCollection;
    }
}


Processor* TrackMakingKalmanFilterProcessor::newProcessor()
{
    return new TrackMakingKalmanFilterProcessor();
}


void TrackMakingKalmanFilterProcessor::check(LCEvent * /*evt*/)
{
}


void TrackMakingKalmanFilterProcessor::end()
{
    delete _kalTracking;

    streamlog_out(DEBUG4) << "End of TrackMakingKalmanFilterProcessor" << endl;
}

} // namespace marlintpc
#endif
