#ifndef CLOUDTOSIMTRACKERHITPROCESSOR_H
#define CLOUDTOSIMTRACKERHITPROCESSOR_H

#include <marlin/Processor.h>
#include <lcio.h>
#include <string>

//GSL
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

namespace marlintpc
{
/**  CloudToSimTrackerHitProcessor:
 * This processor takes as input a collection of TPCElectronClouds and returns a collection
 * of SimTrackerHits. This collection can be used e.g. by the digitisation/GEMProcessor.
 * In order to convert the TPCCloud into SimTrackerHits the longitudinal and transverse diffusion
 * of the cloud is used to distibute the electrons of the cloud randomly (gaus). Thus a cloud with n
 * electrons results on n SimTrackerHits with EDep of the ionisation energy of the drift gas.
 * The default corresponds to the ionization energy of Ar.
 * It is also possible to create only one SimTrackerHit for each cloud with EDep=n*2.6e-08.
 * The drift velocity is needed to calculate the time step after the z step corresponding to
 * the longitudinal diffusion was set.
 * //TODO: change this to be used with other gases
 * 
 * @author Klaus Zenker, DESY
 *
 */

class CloudToSimTrackerHitProcessor : public marlin::Processor
{
 public:
  
  virtual Processor*  newProcessor() { return new CloudToSimTrackerHitProcessor; }
  
  CloudToSimTrackerHitProcessor();
  
  virtual void init();
  
  virtual void processRunHeader(lcio::LCRunHeader* run);
  
  virtual void processEvent(lcio::LCEvent* evt);
  
  virtual void check(lcio::LCEvent* evt);
  
  virtual void end();
  
  
 protected:
  /* the place for protected and private member data and functions */
  std::string _inputColName; ///< Name of the input collection

  std::string _outputColName; ///< Name of the output collection

  bool _singleElectrons; ///< If true for n electrons in the cloud n SimTrackerHits with EDep=2.6e-08 are produced.

  double _ionisationEnergy; ///< The ionisation energy of the drift gas

  gsl_rng *_r;///<Random generator for landau and gaussian

  double _vdrift;///<Drift velocity used to calculate the time step in DistibuteElectrons

  void DistributeElectrons(double* pos, double &time, double difftrans, double difflong); //< Smear a point according to the given diffusion
};
} // namespace marlintpc
#endif // CLOUDTOSIMTRACKERHITPROCESSOR_H
