#include "CloudToSimTrackerHitProcessor.h"
#include <iostream>
#include <string>

// LCIO
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimTrackerHitImpl.h>
#include <EVENT/LCRunHeader.h>

// Marlin
#include <marlin/Exceptions.h>

// MarlinTPC
#include "ElectronCloud.h"

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace EVENT;
using namespace IMPL;


namespace marlintpc
{

CloudToSimTrackerHitProcessor aCloudToSimTrackerHitProcessor;


CloudToSimTrackerHitProcessor::CloudToSimTrackerHitProcessor() : Processor("CloudToSimTrackerHitProcessor"),
    _singleElectrons(true)
{
  // the processor description
  _description = string("CloudToSimTrackerHitProcessor is intended to create from a given collection)")
               + string( "of TPCElectronClouds SimTrackerHits.")
               + string("Here either one SimTrackerHit per cloud is produced or n SimTrackerHit for n electrons in the cloud.");
  
  registerInputCollection(    LCIO::LCGENERICOBJECT,
      "InputCollectionName" ,
      "Name of the input TrackerData collection, default: TPCElectronClouds"  ,
      _inputColName ,
      string("TPCElectronClouds")) ;

  registerOutputCollection(   LCIO::SIMTRACKERHIT,
      "OutputCollectionName" ,
      "Name of the output SimTrackerHit collection, default: TPCDataInPulses",
      _outputColName ,
      string("TPCDriftedElectrons"));

  registerOptionalParameter( "UseSingleElectrons?" ,
      "If true for each electron in the cloud on SimTrackerHit is produced and the position distributed randomly according to transverse and longitudinal diffusion.",
      _singleElectrons,
     bool(true));
  registerOptionalParameter( "Ionisation energy",
      "The energy stored for each SimTrackerHit corresponds to the number of electron associated to the hit and the ionisation energy of the drift gas.",
      _ionisationEnergy,
      double(2.6e-08));
  registerOptionalParameter( "Drift velocity",
      "Set the drift velocity (mm/us) of the considered drift gas.",
      _vdrift,
      double(40));
}


void CloudToSimTrackerHitProcessor::init()
{ 
  streamlog_out(DEBUG) << " Start converting TPCElectronClouds to SimTrackerHits...  " << endl;
  _r = gsl_rng_alloc(gsl_rng_mt19937);
  printParameters();

}


void CloudToSimTrackerHitProcessor::processRunHeader(LCRunHeader* run)
{
  //Revision logging
  run->parameters().setValue(_processorName + "_revision", "$Rev$");

  //Parameter logging
  for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++) {

    if (!i->second->isOptional() || i->second->valueSet())
      run->parameters().setValue(_processorName + "_" + i->second->name(),
          i->second->value());
  }
} 


void CloudToSimTrackerHitProcessor::processEvent(LCEvent* evt)
{ 
  // use an LCCollectionVec instead of LCCollection so we can use iterators later
  LCCollectionVec* electron_cloud_collection = 0;

  // do a try / catch in case there is an empty event
  try
  {
    electron_cloud_collection = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
  }
  catch (DataNotAvailableException &e)
  {
    // no input data, just return and process the next event
    return;
  }
  LCCollectionVec *simtrackerhit_collection= new LCCollectionVec( LCIO::SIMTRACKERHIT );
  simtrackerhit_collection->setTransient(true);
  for( int cloud_number = 0; cloud_number < electron_cloud_collection->getNumberOfElements(); cloud_number++ ) {
    ElectronCloud *current_cloud = dynamic_cast<ElectronCloud*>( electron_cloud_collection->getElementAt( cloud_number ) );
    streamlog_out(DEBUG) << " Working on cloud with ID: " << current_cloud->get_cloud_id() << endl;
    SimTrackerHitImpl *simtrackerhit;
    double position[3] = {current_cloud->get_x_position(), current_cloud->get_y_position(), current_cloud->get_z_position()};
    // time in the cloud is given in us, but it is expected to be in ns later on...
    double time = current_cloud->get_time()*1000;
    if(_singleElectrons){
      streamlog_out(DEBUG1) << " ...I will produce " << current_cloud->get_electron_count() << " SimTrackerHits." << endl;
      streamlog_out(DEBUG4) << "Mean position is "
                    << position[0] << "mm, "
                    << position[1] << "mm, "
                    << position[2] << "mm at the time " << time << "ns" << endl;
      streamlog_out(DEBUG4) << "transverse_stdev:  " << current_cloud->get_transverse_stdev() << " mm" << endl;
      streamlog_out(DEBUG4) << "longitudinal_stdev:  " << current_cloud->get_longitudinal_stdev() << " us" << endl;
      for(int electron = 0; electron < current_cloud->get_electron_count(); electron++){
        simtrackerhit = new SimTrackerHitImpl();
        double tmp_position[3] = {position[0], position[1], position[2]};
        double tmp_time = time;
        DistributeElectrons(tmp_position, tmp_time, current_cloud->get_transverse_stdev(), current_cloud->get_longitudinal_stdev());
        streamlog_out(DEBUG4) << "New position is "
                      << tmp_position[0] << "mm, "
                      << tmp_position[1] << "mm, "
                      << tmp_position[2] << "mm at the time " << tmp_time << "ns" << endl;

        simtrackerhit->setPosition(tmp_position);
        simtrackerhit->setTime(tmp_time);
        simtrackerhit->setEDep(_ionisationEnergy);
        simtrackerhit_collection->addElement(simtrackerhit);
     }
    } else {
      streamlog_out(DEBUG) << " ...I will produce only one SimTrackerHits." << endl;
      simtrackerhit = new SimTrackerHitImpl();
      simtrackerhit->setPosition(position);
      // time is up to here given in us, but it is expected to ns later on...
      simtrackerhit->setTime(time);
      // set the energy of the hit to the energy one would need to produce this amount of electrons in a certain gas in order to use it with digitisation/GEMProcessor
      simtrackerhit->setEDep(current_cloud->get_electron_count() * _ionisationEnergy);
      simtrackerhit_collection->addElement(simtrackerhit);
    }
  }
  try {
    evt->addCollection( simtrackerhit_collection, _outputColName );
  }
  catch( ... ) {

    throw SkipEventException( this );
  }
}


void CloudToSimTrackerHitProcessor::check(LCEvent*  /*evt*/)
{ 

}


void CloudToSimTrackerHitProcessor::end()
{ 

}

void CloudToSimTrackerHitProcessor::DistributeElectrons(double* pos, double &time, double transverse_stdev, double longitudinal_stdev){
  pos[0] += gsl_ran_gaussian(_r, transverse_stdev); //mm
  pos[1] += gsl_ran_gaussian(_r, transverse_stdev); //mm
  double t_step = gsl_ran_gaussian(_r, longitudinal_stdev); //us
  pos[2] += t_step * _vdrift;
  time += t_step * 1000; // ns
}
} // close namespace brackets
