/*
 * ConvertAltroraw2lcio.cpp
 *
 *  Created on: Jul 6, 2012
 *      Author: caiazza
 */
#include <iostream>
#include <sstream>
#include <fstream>
#include <string>
#include <list>
#include <vector>

//LCIO
#include "lcio.h"
#include "UTIL/LCSplitWriter.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"

#include "TAltroConverter.hh"
#include "AltroConverterExceptions.hh"

void PrintUsage();

std::vector<std::ifstream*> CreateInputStreams(std::string& inputfile);

std::string GenerateOutFileName(std::string& inputfile);

/**
 * Program to convert the Altro Raw data to LCIO
 *
 * @param argc Number of command line arguments
 * @param argv List of the arguments
 * @return
 */
int
main(int argc, char* argv[])
{
    /*
     * Basic program flow:
     * Parse the command line for options
     * Creates the list of filenames
     * Open the files and put them in a vector of istream pointers
     * Opens the LCWriter
     * Creates the converter object
     * Sets all the necessary parameters
     * Run the decode function
     * Prints eventual messages
     * Finish
     */

    std::cout << "Starting the conversion from ALTRO Raw Data to LCIO stream" << std::endl;

    //Listing the options the program needs to be defined
    std::list<std::string> inputfilenames;
    std::list<std::string> outputfilenames;
    bool useSplitWriter = true;
    bool writeToDB = false;
    std::string DBInit;
    std::string GeneralElectronicInfoFolder = "/LP1/T24_1/DESY/Grid_GEM_Module_V.2/ADCElectronicsParameters";
    std::string ExtendedElectronicInfoFolder = "/LP1/T24_1/DESY/Grid_GEM_Module_V.2/AltroExtElectronicsParameters";
    bool OverrideZSThreshold = false;
    int ZSThresholdOverride = -1;

    // Invalid usage of the program with no arguments.
    if(argc == 1)
    {
        PrintUsage();
        return 0;
    }

    //Parsing the command line
    int argindex = 1;
    while(argindex<argc)
    {
        std::string option = argv[argindex];

        if(option.at(0) == '-')
        {
            //Detected option marker, switch between the different options.
            if(option == "-I")
            {
                while((++argindex < argc) && (argv[argindex][0] != '-'))
                {
                    inputfilenames.push_back(std::string(argv[argindex]));
                }

            }

            if(option == "-O")
            {
                while((++argindex < argc) && (argv[argindex][0] != '-'))
                {
                    outputfilenames.push_back(std::string(argv[argindex]));
                }
            }

            if(option.find('S') != std::string::npos)
            {
                std::cout << "Enabling the LCIO Split Writer. WARNING The Split writer only "
                        << "allows for a single data run to be processed in the program" << std::endl;
                useSplitWriter = true;
                ++argindex;
            }else if(option.find('s') != std::string::npos)
            {
                std::cout << "Enabling the standard LCWriter." << std::endl;
                useSplitWriter = false;
                ++argindex;
            }

            //Checking for the option to enable the DB writing
            if(option.find('D') != std::string::npos)
            {
                if((++argindex <argc) && (argv[argindex][0] != '-'))
                {
                    DBInit = argv[argindex];
                    writeToDB = true;
                }
                else
                {
                    std::cout << "Mandatory DB initialization missing after -D option \n\n Usage:" << std::endl;
                    PrintUsage();
                    return 0;
                }
                ++argindex;
            }

            //Checking the eventual custom name for the General Info Folder
            if(option.find('G') != std::string::npos)
            {
                if((++argindex <argc) && (argv[argindex][0] != '-'))
                {
                    GeneralElectronicInfoFolder = argv[argindex];
                }
                else
                {
                    std::cout << "-G option specified without the folder name following \n\n Usage:" << std::endl;
                    PrintUsage();
                    return 0;
                }
                ++argindex;
            }

            //Checking the eventual custom name for the General Info Folder
            if(option.find('E') != std::string::npos)
            {
                if((++argindex <argc) && (argv[argindex][0] != '-'))
                {
                    ExtendedElectronicInfoFolder = argv[argindex];
                }
                else
                {
                    std::cout << "-E option specified without the folder name following \n\n Usage:" << std::endl;
                    PrintUsage();
                    return 0;
                }
                ++argindex;
            }

            //Checking if the zero suppression threshold was overridden
            if(option.find('Z') != std::string::npos)
            {
                if((++argindex <argc) && (argv[argindex][0] != '-'))
                {
                    OverrideZSThreshold = true;
                    std::istringstream ZSString(argv[argindex]);
                    ZSString >> ZSThresholdOverride;
                }
                else
                {
                    std::cout << "-E option specified without the folder name following \n\n Usage:" << std::endl;
                    PrintUsage();
                    return 0;
                }
                ++argindex;
            }
        }
        else
        {
            // The only possibility of an argument not in an option group are the filenames
            while(option.at(0) != '-')
            {
                /*
                 * Special case representing command option 1
                 */
                if(argindex == 2 && argc == 3)
                {
                    outputfilenames.push_back(option);
                }
                else
                {
                    inputfilenames.push_back(option);
                }

                if(++argindex == argc)
                {
                    break;
                }
                option = argv[argindex];
            }
        }
    }

    if(useSplitWriter && (inputfilenames.size() > 1))
    {
        std::cout << "ERROR: The Split Writer only supports single run data conversion. Abort"
                << std::endl;
        return 0;
    }

    //Opening the files
    std::list<std::string>::iterator itnames = inputfilenames.begin();

    std::vector<std::pair<std::vector<std::ifstream*>, std::string > > ConversionList;
    for(; itnames != inputfilenames.end(); ++itnames)
    {
        std::vector<std::ifstream*> InputStreams = CreateInputStreams(*itnames);
        if(outputfilenames.empty())
        {
            ConversionList.push_back(std::make_pair(InputStreams, GenerateOutFileName(*itnames)));
        }
        else
        {
            ConversionList.push_back(std::make_pair(InputStreams, outputfilenames.front()));
            outputfilenames.pop_front();
        }
    }

    //Local pointers to be used in the try clause and be deleted if an exception is caught in the catch
    lcio::LCWriter* lciowriter = 0;
    lccd::DBInterface* DBGenericInfo = 0;
    lccd::DBInterface* DBExtendedInfo = 0;
    try
    {
        if(useSplitWriter)
        {
            lciowriter = new lcio::LCSplitWriter(lcio::LCFactory::getInstance()->createLCWriter(),
                    2040109465);
        }
        else
        {
            lciowriter = lcio::LCFactory::getInstance()->createLCWriter();
        }

        if(writeToDB)
        {
            DBGenericInfo = new lccd::DBInterface(DBInit, GeneralElectronicInfoFolder, true);
            DBExtendedInfo = new lccd::DBInterface(DBInit, ExtendedElectronicInfoFolder, true);
        }


        std::vector<std::pair<std::vector<std::ifstream*>, std::string > >::iterator itConv = ConversionList.begin();
        for(; itConv!= ConversionList.end(); ++itConv)
        {
            std::cout << "Starting conversion job " << (itConv - ConversionList.begin()) << std::endl;
            std::cout << itConv->second << std::endl;
            lciowriter->open(itConv->second);
            marlintpc::TAltroConverter Converter(itConv->first, lciowriter);
            if(OverrideZSThreshold)
            {
                Converter.SetZSThresholdOverride(ZSThresholdOverride);
            }
            Converter.Decode();
            if(writeToDB)
            {
                Converter.ToDB(DBGenericInfo, DBExtendedInfo);
            }
            Converter.Print(std::cout);
            lciowriter->close();

            std::vector<std::ifstream*>::iterator itStreams = itConv->first.begin();
            for(; itStreams!= itConv->first.end(); ++itStreams)
            {
                (*itStreams)->close();
                delete (*itStreams);
                (*itStreams) = 0;
            }

        }

        delete lciowriter;
        lciowriter = 0;
        delete DBGenericInfo;
        DBGenericInfo = 0;
        delete DBExtendedInfo;
        DBExtendedInfo = 0;
    }
    catch(lcio::Exception & e)
    {
        delete lciowriter;
        lciowriter = 0;
        delete DBGenericInfo;
        DBGenericInfo = 0;
        delete DBExtendedInfo;
        DBExtendedInfo = 0;
        std::cout << e.what();
    }

}

void PrintUsage()
{
    std::cout << "DESCRIPTION" << std::endl;
    std::cout << "\t\t This program converts altro raw data to LCIO data in the slcio format."<< std::endl
            << "\t\t The program requires at least one input file name. If the -I option is not " <<
            "set the program assumes the last argument is the input file name" << std::endl;
    std::cout << "\t\t Each of the input files must be the first of the data files produced by "
            << "the DAQ, which contains the start of run data block." << std::endl;
    std::cout << "\t\t If any of the input file names has the format <prefix>_0.dat, the program "
            << "will attempt to load all the other files with the same prefix and increasing counter "
            << "located in the same folder."
            << "Files that are not named sequentially as described above are considered as single "
            << "data files of uncorrelated runs and will be converted separately." <<std::endl;
    std::cout << "\t\t If the option -O is not specified or if the number of output file names is less than "
            << "the number of input file names each input files that doesn't have a matching output file name "
            << "will be written to an slcio file whose name is obtained substituting the input file extension "
            << "or adding the .slcio extension in case of no extension in the input file" << std::endl;
    std::cout << "\nSYNOPSIS" << std::endl;
    std::cout << "\t\t ConvertAltroraw2lcio INPUTFILE OUTPUTFILE" << std::endl;
    std::cout << "\t\t ConvertAltroraw2lcio [options] INPUTFILE[INPUTFILES ...]" << std::endl;
    std::cout << "\t\t ConvertAltroraw2lcio [options]" << std::endl;
    std::cout << "\nOPTIONS" <<std::endl;
    std::cout << "\t\t -I FILES"
            <<"\n\t\t\t List of input file names. See the description for more information" <<std::endl;
    std::cout << "\t\t -O OUTPUTFILES"
            << "\n\t\t\t List of the output file name to create. Each output file will be matched with "
            << "an input file." <<std::endl;
    std::cout << "\t\t -S \t Uses the LCIO Split Writer. The LCIO Split Writer only support one "
            << "run per process. Multiple file input is disabled." << std::endl;
    std::cout << "\t\t -s \t Uses the standard LCWriter. Conversion of multiple run are allowed but no "
            << "check on the output file size is performed." << std::endl;
    std::cout << "\t\t -D DBInit"
            << "\n\t\t\t Enables the program to upload the electronics information to the database."
            << "It REQUIRES the CondDB initialization string as argument in the form "
            << "host:databasename:login:password[:port]." <<std::endl;
    std::cout << "\t\t -G GenericElectronicInfoFolder"
            << "\n\t\t\t Defines the path to the folder where to store the GenericElectronic Condition Data."
            << "If this option is not defined and the database is to be used, a default name is set."
            <<" Path format: /database/folder"<< std::endl;
    std::cout << "\t\t -E ExtendedElectronicInfoFolder"
            << "\n\t\t\t Defines the path to the folder where to store the ExtendedElectronic Condition Data."
            << "If this option is not defined and the database is to be used, a default name is set"
            << " Path format: /database/folder" << std::endl;


}

std::vector<std::ifstream*> CreateInputStreams(std::string& inputfile)
{
    std::vector<std::ifstream*> result;
    if((inputfile.length() > 6) && (inputfile.substr(inputfile.length() -6) == "_0.dat"))
    {
        std::string Basename = inputfile.substr(0, inputfile.length() -6);
        int counter = 0;
        std::ifstream* file = new std::ifstream(inputfile.c_str(), std::ifstream::binary);
        while(file->good())
        {
            result.push_back(file);
            std::ostringstream NextFilename;
            NextFilename << Basename << "_" << ++counter << ".dat";
            //std::cout << NextFilename.str() << std::endl;
            file = new std::ifstream(NextFilename.str().c_str(), std::ifstream::binary);
        }
        delete file;

        if(result.empty())
        {
            throw marlintpc::TAltroConverterCLException
            (std::string("Invalid input file detected: ") + inputfile);
        }
    }
    else
    {
        std::ifstream* file = new std::ifstream(inputfile.c_str(), std::ifstream::binary);
        if(!file->good())
        {
            delete file;
            throw marlintpc::TAltroConverterCLException
            (std::string("Invalid input file detected: ") + inputfile);
        }
        result.push_back(file);
    }
    return result;
}

std::string GenerateOutFileName(std::string& inputfile)
{
    std::stringstream outfilename;

    if((inputfile.length() > 6) && (inputfile.substr(inputfile.length() -6) == "_0.dat"))
    {
        outfilename << inputfile.substr(0, inputfile.length() -6) << ".slcio";
    }
    else
    {
        size_t lastdotpos = inputfile.find_last_of(".");
        outfilename << inputfile.substr(0, lastdotpos) << ".slcio";
    }
    return outfilename.str();
}
