/*
 * ConvertMMraw2lcio.cpp
 *
 *  Created on: Mar 3, 2014
 *      Author: zenker
 *
 *  This converter is based on a version used for the test beam in Feb. 2014.
 *  It converts *.acq files for a certain run into slcio files.
 *  It is also possible to put the data of one run stored in different
 *  raw data files into one slcio file by using the merge option.
 */

//
//
// NOTE:
//  - Compress mode can be obtained by parsing contents of the CFG file,
//    which is stored in LCIO Run Header.
//    (eg) <compress>false</compress>           <--- compress = 0
//
// standard libraries
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <dirent.h>
#include <errno.h>
#include <arpa/inet.h>
#include <limits>
#include <getopt.h>

//MarlinTPC
#include "ADCChannelMapping.h"

// GEAR
#include <gearxml/tinyxml.h>
#include <gear/GEAR.h>

//ILCUTIL
#include "streamlog/streamlog.h"

// LCIO
#include <lcio.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCFlagImpl.h>
#include <UTIL/LCTOOLS.h>
#include <UTIL/LCTime.h>

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"

using namespace std;
using namespace lcio;

//----------------------------------------------------------
// declare constants
//----------------------------------------------------------

const int kNumTimeBins = 512;
const int kChanPerChip = 72;
const int kNRows = 24;
const int kNCols = 72;
const int kNMods = 7;
const int kChannelMax = kChanPerChip * kNRows;


// Macros to interpret DataPacket header
#define GET_LOS_FLAG(word)   (((word) & 0x0400)>>10)
#define GET_SYNCH_FAIL(word) (((word) & 0x0800)>>11)

// Macros to interpret DataPacket read back arguments
#define GET_RB_COMPRESS(word) (((word) & 0x2000)>>13)
#define GET_RB_ARG2(word)     (((word) & 0x1E00)>>9)
#define GET_RB_ARG1(word)     (((word) & 0x01FF))

#define GET_RT(word)           (((word) & 0x0C00)>>10)
#define GET_BIN_INDEX(word)    (((word) & 0x01FF))
#define GET_FLAG(word)         (((word) & 0x3000)>>12)
#define GET_DCC_INDEX(word) (((word) & 0x03F0)>>4)
#define GET_FEM_INDEX(word) (((word) & 0x000F))
#define GET_VERSION(word)      (((word) & 0xC000)>>14)

struct EventHeader {
  int eventSize;
  int eventNumb;
};

struct DataPacketHeader {
  unsigned short int size;
  unsigned short int dcc_hdr;    // To change if the data after May 2010
  unsigned short int hdr;
  unsigned short int args;
  unsigned short int ts_h;
  unsigned short int ts_l;
  unsigned short int ecnt;
  unsigned short int scnt;
};

//For convenience
struct NewDataHeader {
  unsigned short int args;
  unsigned short int wordcount;
};

struct DataPacketEnd {
  unsigned short int crc1;
  unsigned short int crc2;
};

struct Pulse {

  std::vector<unsigned short int> adc;
  unsigned short int counts;
  unsigned short int time;

};

/**
 * Structure used to construct a time stamp for each event.
 * The current event time is stored in \c end.
 * Thus at the end of the run \c end corresponds to the last time of the run.
 */
struct TimeStamp {
  string RunUID;   // Run UID in the beginning of the ACQ file
  unsigned int timeBin; // 32bit timeStamp
  long nHits; // counts how often 32bit register for the time stamp was filled
  unsigned int offset; // this is the timeStamp of the first event
  const unsigned short clockCycle; //ns
  int runNumber;
  int subRunNumber;
  LCTime start;
  LCTime end;
  TimeStamp():clockCycle(10){
    nHits = 0;
    offset = 0;
    timeBin = 0;
    runNumber = 0;
    subRunNumber = 0;
    start = LCTime(0);
    end = LCTime(0);
  }
  bool Update(unsigned int tmpStamp){
    if (tmpStamp < timeBin)
      nHits++;
    timeBin = tmpStamp;
    if(start.unixTime() == 0){
      offset = tmpStamp;
      char* pEnd;
      long int date[6];
      date[0] = strtol(RunUID.c_str() + 1, &pEnd, 10); // year
      date[1] = strtol(pEnd + 1, &pEnd, 10); // month
      date[2] = strtol(pEnd + 1, &pEnd, 10); // day
      date[3] = strtol(pEnd + 1, &pEnd, 10); // hour
      date[4] = strtol(pEnd + 1, &pEnd, 10); // minute
      date[5] = strtol(pEnd + 1, &pEnd, 10); // second
      start = LCTime(date[0], date[1], date[2], date[3], date[4], date[5]);
    }
    long64 timeSinceStart = timeBin;
    timeSinceStart += nHits * numeric_limits<unsigned int>::max() - offset;
    timeSinceStart *= clockCycle;
//    long64 timeSinceStart =  (timeBin - offset + nHits * numeric_limits<unsigned int>::max()) * clockCycle;
    LCTime tmp(start.timeStamp());
    tmp += timeSinceStart;
    // check if after updating the time stamp is increased
    if(end.timeStamp() > tmp.timeStamp()){
      streamlog_out(ERROR) << "Something went wrong in the calculation of the event time!" << endl;
      streamlog_out(ERROR) << "Time last event: " << end.getDateString() << " \t next event: " << tmp.getDateString() << endl;
      return false;
    }
    end = tmp;
    return true;
  }
};

struct CFGParameter {
  int readMode;
  int compress;   // 0 = non comp,  1 = compressed
  int numTimeBins;
  float samplingPeriod;   // ns
  int shaping;
  CFGParameter(){
    readMode = 0;
    compress = 0;
    numTimeBins = 0;
    samplingPeriod = 0;
    shaping = 0;
  }
};

/**
 * Read all files in the given directory which include the run string and end with ".acq".
 * \param dir The directory with the raw data files.
 * \param run The run number used to find the corresponding raw data files.
 */
vector<string> GetFilesInDir(const string dir, const string run) {
  vector<string> files;
  DIR *dp;
  struct dirent *dirp;
  if ((dp = opendir(dir.c_str())) == NULL) {
    stringstream message;
    message << "Error(" << errno << ") opening " << dir.c_str();
    throw Exception(message.str());
  }
  string file_name;
  while ((dirp = readdir(dp)) != NULL) {
    file_name = dirp->d_name;
    std::size_t found = file_name.find(run);
    if (found != string::npos) {
      // get length length
      int len = strlen(file_name.c_str());
      if (len >= 4 && strcmp(".acq", &(file_name[len - 4])) == 0) {
        file_name.insert(0, dir);
        files.push_back(file_name);
      }
    }
  }
  closedir(dp);
  streamlog_out(MESSAGE) << "Opened the following Files in path " << dir.c_str()
      << " (identifier: " << run.c_str() << "):" << endl;
  vector<string>::iterator it = files.begin();
  while (it != files.end()) {
    streamlog_out(MESSAGE) << "-->" << (*it).c_str() << endl;
    it++;
  }
  return files;
}

/***
 * Open a raw data file and extract time stamp information.
 * In addition the CFG file name is constructed. It is assumed that it is called:
 * CFG_run.xml
 *
 * \remark The Time stamp UID, runNumber, and subRunNumber are read for every run file which is in principle not necessary.
 *         Only the subRunNumber needs to be updated and it is not used at the moment...
 * \param file Raw data file name.
 * \param dir Directory of the raw data files (needed to find the CFG file).
 * \param cfgFileName Here the name of the config file is stored.
 * \param fileData This is the opened raw data file.
 * \param timeStamp Here the time stamp information is added.
 */
bool OpenAcqFile(const string file, const string dir, string &cfgFileName, FILE* &fileData, TimeStamp &timeStamp) {
  // Get run number from the filename
  streamlog_out(MESSAGE) << "Open file " << file.c_str() << endl;
  sscanf((file.substr(file.find_last_of("/")+1,file.size())).c_str(), "RUN_%5d.%5d", &timeStamp.runNumber, &timeStamp.subRunNumber);

  // CFG filename
  char tmp[160];
  sprintf(tmp, "%sCFG_%05d.xml", dir.c_str(), timeStamp.runNumber);
  cfgFileName = string(tmp);
  streamlog_out(MESSAGE) << "CFG filename = " << cfgFileName.c_str() << endl;

  // Open the file
  fileData = fopen(file.c_str(), "rb");
  if (!fileData) {
    streamlog_out(ERROR) << " Can't open file: " << file.c_str() << endl;
    return false;
  }

  //
  // Read Run UID characters
  //
  char tmp_runID[21];
  fread(tmp_runID, 1, 20, fileData);
  timeStamp.RunUID = string(tmp_runID);
  timeStamp.RunUID.erase(timeStamp.RunUID.size()-2,timeStamp.RunUID.size());
  streamlog_out(MESSAGE) << "RunUID = " << timeStamp.RunUID.c_str() << endl;
  return true;
}

template<size_t NMods, size_t ChannelMax>
bool DecodeNextEvent(int &eventNumber, FILE* fileData, TimeStamp &timeStamp, vector<Pulse> (&Data)[NMods][ChannelMax]) {
  EventHeader * evtHead = new EventHeader;
  DataPacketHeader * packetHeader = new DataPacketHeader;
  //----------------------------------------------
  // Read event header or quit at the end of file
  //----------------------------------------------
  if (fread(evtHead, sizeof(EventHeader), 1, fileData) != 1) {
    fclose(fileData);
    return false; // End of program
  }

  //Convert to from network byte things
  evtHead->eventSize = ntohl(evtHead->eventSize); // size of current event
  eventNumber = evtHead->eventNumb = ntohl(evtHead->eventNumb); // current event number
  streamlog_out(DEBUG) << "Start decoding event " << eventNumber << endl;
  int eventSize = evtHead->eventSize;
  int rawDataSize = eventSize - sizeof(EventHeader) - sizeof(DataPacketEnd);
  int nByteTot = rawDataSize;
  int nByteRead = 0;
  int nPacket = 0;
  bool timeIsSet = false;
  while (nByteRead < nByteTot) // nByteTot represents the size of the full event, not only the current packet
  {
    int nlinesread = 0;
    nPacket++;
    // Read Packet Header of the current packet
    fread(packetHeader, sizeof(DataPacketHeader), 1, fileData);
    nByteRead += sizeof(DataPacketHeader);

    // # of data words in the packet except header and trailer
    int kTrailerSize = sizeof(DataPacketEnd);
    int kPktHdrSize = sizeof(DataPacketHeader) / sizeof(short); // 8
    int nTotPktWords = ntohs(packetHeader->size) / sizeof(unsigned short int);

    int nlinespacket = nTotPktWords - kPktHdrSize - kTrailerSize;

    // Check sync fail
    if (GET_LOS_FLAG((ntohs(packetHeader->hdr))))
      streamlog_out(WARNING) << "LOS FLAG SET" << endl;
    if (GET_SYNCH_FAIL((ntohs(packetHeader->hdr))))
      streamlog_out(WARNING) << "SYNC FAIL FLAG SET" << endl;

    // Check compress mode
    int compressed = GET_RB_COMPRESS(ntohs(packetHeader->args));

    int version = GET_VERSION(ntohs(packetHeader->dcc_hdr)); // old is 0, new is 1
    int dccFlag = GET_FLAG(ntohs(packetHeader->dcc_hdr));
    int replyType = GET_RT(ntohs(packetHeader->dcc_hdr));
    int femID = GET_FEM_INDEX(ntohs(packetHeader->dcc_hdr));

    // Check dcc flags
    if (version != 1) {
      streamlog_out(WARNING) << "This is NOT a version 2 format data" << endl;
    }

    // If EoE or Non-FEM data, read out all remaining packet data and
    // exit from the loop since there is nothing to be done.
    //Figure This one out later
    if (dccFlag == 2 || replyType == 1) {
      int nWordsToRead = nTotPktWords - kPktHdrSize;    // - dTrailerSize;
      unsigned short int * tData = new unsigned short int[nWordsToRead];
      fread(tData, sizeof(unsigned short int), nWordsToRead, fileData);
      nByteRead += sizeof(unsigned short int) * nWordsToRead;
      delete[] tData;
      break;
    }

    int physChannel = -10;
    int bin_index;
    bool firstChan = true;
    unsigned short int * pData = 0;

    while (nlinesread < nlinespacket) {
      // Get args and sample count
      unsigned short int args, sampleCount;
      if (firstChan) {
        streamlog_out(DEBUG) << "--- First Chan Data Block ----" << endl;
        args = ntohs(packetHeader->args);
        sampleCount = ntohs(packetHeader->scnt);
        if(!timeIsSet){
          // shift ts_h by 16 bits and add ts_l to get the 32 bit time stamp
          //  unsigned int tmpStamp = ((packetHeader->ts_h) << 16) | packetHeader->ts_l;
          unsigned int tmpStamp = ntohs(packetHeader->ts_h) << 16;
          tmpStamp += ntohs(packetHeader->ts_l);
          if(!timeStamp.Update(tmpStamp)){
            streamlog_out(MESSAGE) << "Working on event " << eventNumber << endl;
            throw Exception("Failed to update the time stamp!");
          }
          timeIsSet = true;
        }
        firstChan = false;
      } else {
        streamlog_out(DEBUG) << "--- New Chan Data Block ----" << endl;
        NewDataHeader newData;
        try {
          fread(&newData, sizeof(NewDataHeader), 1, fileData);
        } catch (std::bad_alloc& ba) {
          std::cerr << "Error. Bad Alloc: " << ba.what() << std::endl;
        }
        nlinesread += sizeof(NewDataHeader) / sizeof(unsigned short int);
        nByteRead += sizeof(NewDataHeader);

        if (((ntohs(newData.args) & 0xE000) != 0x2000)
            || ((ntohs(newData.wordcount) & 0xF000) != 0x4000)) {
          streamlog_out(ERROR) << "something wrong happened... exiting" << endl;
          return false;
        }

        args = ntohs(newData.args) & 0x1FFF; // No need but...
        sampleCount = ntohs(newData.wordcount) & 0x0FFF;
      }

      int arg1 = GET_RB_ARG1(args);
      int arg2 = GET_RB_ARG2(args);

      // compute asic#
      streamlog_out(DEBUG) << "Packet " << nPacket << endl;
      streamlog_out(DEBUG) << "arg1 = " << arg1 << ", arg2 = " << arg2 << endl;
      int channelN = arg1 / 6;
      int asicN = 10 * (arg1 % 6) / 2 + arg2;

      // From now on only data words until next channel data.
      // Thus read the data words.
      int sampleCountRead = sampleCount;
      if (sampleCount % 2 == 1)
        sampleCountRead++;
      streamlog_out(DEBUG) << "SampleCountsRead = " << sampleCountRead << endl;

      if (sampleCount == 0)
        continue;

      try {
        pData = new unsigned short int[sampleCountRead];
      } catch (std::bad_alloc& ba) {
        streamlog_out(ERROR) << "Error2. Bad Alloc: " << ba.what() << std::endl;
      }
      fread(pData, sizeof(unsigned short int), sampleCountRead, fileData);

      nlinesread += sampleCountRead;

      streamlog_out(DEBUG) << "nlinesread= " << nlinesread << endl;

      nByteRead += sampleCountRead * sizeof(unsigned short int);

      unsigned short int binIndex = ntohs(pData[0]);

      streamlog_out(DEBUG) << "Getting bin index" << endl;
      bin_index = GET_BIN_INDEX(binIndex);
      streamlog_out(DEBUG) << "Bin index: " << bin_index << endl;
      streamlog_out(DEBUG) << "new word count: " << sampleCount << endl;

      int t0 = 0;  // will be updated for compressed data.
      //
      // Handle data
      //
      bool chstatus = false;

      // compute physical channel (0-71)
      if (channelN > 2 && channelN < 15) {
        physChannel = channelN - 3;
        chstatus = true;
      } else if (channelN > 15 && channelN < 28) {
        physChannel = channelN - 4;
        chstatus = true;
      } else if (channelN > 28 && channelN < 53) {
        physChannel = channelN - 5;
        chstatus = true;
      } else if (channelN > 53 && channelN < 66) {
        physChannel = channelN - 6;
        chstatus = true;
      } else if (channelN > 66) {
        physChannel = channelN - 7;
        chstatus = true;
      }      //else
             //  physChannel  = asicN*100 + channelN;        // cell id for non-phys data
             //If Something goes wrong, look here first!!

      int thisPhysChannel = asicN * kNCols + physChannel;
      if ((thisPhysChannel < 0) || (thisPhysChannel >= 1728)) {
        delete[] pData;
        continue;
      }

      if (physChannel >= 0 && physChannel < 72 && chstatus) {
        if (compressed)      ////compressed data
        {
          // Calculate time bin of the first sample of the first pulse
          const int kPreThr = 10; // # of sample before threshold
          int nskip;          // in case of scnt < 10,
          if (bin_index < kPreThr) {
            t0 = 0;            // time bin id of the first valid sample = 0
            nskip = kPreThr - t0; // first few samples are just paddings
          } else {
            t0 = bin_index - kPreThr;
            nskip = 0;           // all samples are valid --> no need to skip
          }

          if (sampleCount <= kNumTimeBins) {

            Data[femID][thisPhysChannel].push_back(Pulse());

            Data[femID][thisPhysChannel].back().counts = sampleCount - nskip
                - 2;
            Data[femID][thisPhysChannel].back().time = t0;
            Data[femID][thisPhysChannel].back().adc = std::vector<
                unsigned short int>();

            for (int ismpl = nskip + 1; ismpl < sampleCount - 1; ismpl++) {
              //unsigned short int int datum = ntohs(pData.data[ismpl]);
              unsigned short int datum = ntohs(pData[ismpl]);

              if ((datum & 0x1000) == 0) {
                //Leave this stuff to later
                /*                if ((datum & 0x1000) == 0)   // Check Bin Index Word mask
                 {
                 // Just in case
                 if ((datum & 0xF000) != 0) {
                 cout << "*** ERROR: Wrong sample word: 0x" << hex
                 << datum << dec << ": nskip=" << nskip << endl;
                 }
                 }
                 */
                //    const int kOffset = 250;
                //                int datumC = datum - kOffset;
                Data[femID][thisPhysChannel].back().adc.push_back(datum);
              } else {
                bin_index = GET_BIN_INDEX(datum);
                if (bin_index < kPreThr) {
                  t0 = 0;           // time bin id of the first valid sample = 0
                  nskip = kPreThr - t0; // first few samples are just paddings
                } else {
                  t0 = bin_index - kPreThr;
                  nskip = 0;        // all samples are valid --> no need to skip
                }
                Data[femID][thisPhysChannel].push_back(Pulse());

                Data[femID][thisPhysChannel].back().counts = sampleCount - nskip
                    - 2;
                Data[femID][thisPhysChannel].back().time = t0;
                Data[femID][thisPhysChannel].back().adc = std::vector<
                    unsigned short int>();
              }
            }
          }
        } else //Non zero-suppressed
        {
          Data[femID][thisPhysChannel].push_back(Pulse());

          Data[femID][thisPhysChannel].back().counts = sampleCount;
          Data[femID][thisPhysChannel].back().time = 0;
          Data[femID][thisPhysChannel].back().adc = std::vector<
              unsigned short int>();

          for (int ismpl = 0; ismpl < sampleCount; ismpl++)
            Data[femID][thisPhysChannel].back().adc.push_back(pData[ismpl]);
        }
      }
      delete[] pData;
    }

    DataPacketEnd * packetEnd = new DataPacketEnd;
    // Read the trailer words
    fread(packetEnd, sizeof(DataPacketEnd), 1, fileData);
    nByteRead += sizeof(DataPacketEnd);

    delete packetEnd;

  }  //end main channel
  if (evtHead)
    delete evtHead;
  if (packetHeader)
    delete packetHeader;

  streamlog_out(DEBUG) << "End of decode..." << endl;
  return true;
}

/**
 * Write information of an event from the raw data file to an lcio file.
 * The collection to be written is called ADC_CHAN_DATA.
 * In the original Converter there was also the possibility to write two more collections:
 * - RESET_CHAN_DATA Data of channels 0,1,2
 * - FPN_CHAN_DATA Data of noise channels, which were hard coded.
 * \param runNumber Run number of the event.
 * \param eventNumber Event number.
 * \param lcWrt The writer with the open lcio file.
 * \param timeStamp The time stamp information of the event.
 * \param Data The data to write.
 */
template<size_t NMods, size_t ChannelMax>
bool EncodeThisEvent(int eventNumber, LCWriter* lcWrt, const TimeStamp timeStamp, vector<Pulse> (&Data)[NMods][ChannelMax]) {
  /***
   * NOISE and RESET collections are disabled, but could be added by uncommenting...
   */
  streamlog_out(DEBUG) << "Start Encoding Event " << eventNumber << endl;

  //
  // Create new lcio event
  //
  LCEventImpl* evt = new LCEventImpl();

  evt->setRunNumber(timeStamp.runNumber);
  evt->setEventNumber(eventNumber);
  evt->setDetectorName("LP-TPC");
  streamlog_out(DEBUG) << "Time stamp of the event: " << timeStamp.end.getDateString().c_str() << endl;
  evt->setTimeStamp(timeStamp.end.timeStamp());

  // Create a collection of ADC samples to put into the event
  IMPL::LCCollectionVec* rawADCCollection;
  rawADCCollection = new IMPL::LCCollectionVec(LCIO::TRACKERRAWDATA);
  rawADCCollection->setTransient(false);  // we want to store it to a file
  /*
  // Second Collection for RESET channels.
  IMPL::LCCollectionVec* resetChanCollection;
  resetChanCollection = new IMPL::LCCollectionVec(LCIO::TRACKERRAWDATA);
  resetChanCollection->setTransient(false);  // we want to store it to a file

  // Third Collection for NOISE channels
  IMPL::LCCollectionVec* noiseChanCollection;
  noiseChanCollection = new IMPL::LCCollectionVec(LCIO::TRACKERRAWDATA);
  noiseChanCollection->setTransient(false);  // we want to store it to a file
  */
  //NEED TO SET A FLAG TO USE MODULE IDs!!
  LCFlagImpl flag;
  flag.setBit(LCIO::TRAWBIT_ID1);
  rawADCCollection->setFlag(flag.getFlag());

  TrackerRawDataImpl* rawData; // = new TrackerRawDataImpl;

  for (int mod = 0; mod < kNMods; mod++) {
    // For each channel (more channels than pads),
    // create a data vector, and re-store data
    for (int chn = 0; chn < kChannelMax; chn++) {
      //
      // Skip for channels without hit
      if (!Data[mod][chn].size())
        continue;
      EVENT::ShortVec *ADCValues;

      for (vector<Pulse>::iterator it = Data[mod][chn].begin();
          it != Data[mod][chn].end(); it++) {
        rawData = new TrackerRawDataImpl;

        ADCValues = new EVENT::ShortVec();
        for (vector<unsigned short int>::iterator it_adc = (*it).adc.begin(); it_adc != (*it).adc.end(); it_adc++){
          ADCValues->push_back(*it_adc);
        }
        rawData->setADCValues(*ADCValues);
        rawData->setTime((*it).time);
        rawData->setCellID1(mod);    // module number
        rawData->setCellID0(chn);
        /*
         int asicN = chn%kNCols;
         int channelN = chn - asicN;

         int thisPhysChannel = asicN*kNCols + physChannel;

         if ( channelN <= 2 ) {
         //              rawData->setCellID0( chn );
         resetChanCollection->push_back(rawData);
         }
         else if ( channelN==15 || channelN==28 || channelN==53 || channelN==66 ) {
         noiseChanCollection->push_back(rawData);
         }
         else {
         */
        rawADCCollection->push_back(rawData);
        delete ADCValues;
//        }
      }
    }
  }

  // add the collection of raw data's to the event
  ostringstream name;
  name << "ADC_CHAN_DATA";  // to make this appear before other coll's in dump
  evt->addCollection(rawADCCollection, name.str());
  /*
  ostringstream name2;
  name2 << "RESET_CHAN_DATA";
  evt->addCollection(resetChanCollection, name2.str());

  ostringstream name3;
  name3 << "FPN_CHAN_DATA";
  evt->addCollection(noiseChanCollection, name3.str());
  */
  static int nCount = 0;
  if (nCount % 50 == 0)
    LCTOOLS::dumpEvent(evt);
  nCount++;

  try {
    lcWrt->writeEvent(evt);
  } catch (IOException& e) {
    streamlog_out(ERROR) << e.what() << endl;
  }

  delete evt;
  return true;
}

/**
 * Here some inforamtion from the CFG file are retrieved. See CFGParameter to see which parameters are retrieved.
 * \param filename The file name including path of the config file.
 * \param cfgParameter Here the retrieved parameters are stored.
 *
 */
void ParseConfigFile(const string filename, CFGParameter &cfgParameter) {
  const int kBaseFrequency = 100;  // MHz
  try {
    gear::TiXmlDocument document;
    document.LoadFile(filename);
    gear::TiXmlElement *daq = document.FirstChildElement("Daq");
    if (!daq)
      throw gear::Exception("Could not find element <Daq>.");
    gear::TiXmlElement *dcc = daq->FirstChildElement("DCC");
    if (!dcc)
      throw gear::Exception("Could not find element <DCC>.");
    gear::TiXmlElement *fem = dcc->FirstChildElement("FEM");
    if (!fem)
      throw gear::Exception("Could not find element <FEM>.");
    string id = fem->Attribute("id");
    // find FEM with id="*"
    while (fem && id.compare("*")) {
      fem = dcc->NextSiblingElement("FEM");
      id = fem->Attribute("id");
    }
    if (!fem || id.compare("*"))
      throw gear::Exception("Could not find element <FEM id=\"*\">");
    gear::TiXmlElement *module = fem->FirstChildElement("Module");
    if (!module)
      throw gear::Exception("Could not find element <Module>");
    gear::TiXmlElement *tmpElement = module->FirstChildElement("readMode");
    if (tmpElement
        && !string(tmpElement->FirstChild()->Value()).compare("time"))
      cfgParameter.readMode = 0;
    else if (tmpElement
        && !string(tmpElement->FirstChild()->Value()).compare("space"))
      cfgParameter.readMode = 1;
    else
      throw gear::Exception(
          "Could not find element <readMode> with value of either space or time.");
    tmpElement = module->FirstChildElement("compress");
    if (tmpElement
        && !string(tmpElement->FirstChild()->Value()).compare("true"))
      cfgParameter.compress = 1;
    else if (tmpElement
        && !string(tmpElement->FirstChild()->Value()).compare("false"))
      cfgParameter.compress = 0;
    else
      throw gear::Exception(
          "Could not find element <compress> with value of either true or false.");
    tmpElement = module->FirstChildElement("nTimeBuckets");
    if (!tmpElement)
      throw gear::Exception("Could not find element <nTimeBuckets>");
    cfgParameter.numTimeBins = strtol(tmpElement->FirstChild()->Value(), 0, 16);
    tmpElement = module->FirstChildElement("divideScaClockBy");
    if (!tmpElement)
      throw gear::Exception("Could not find element <divideScaClockBy>");
    long int sampling = strtol(tmpElement->FirstChild()->Value(), 0, 16);
    cfgParameter.samplingPeriod = 1000. * sampling / kBaseFrequency; // ns : 1/f = 1/(B/d)

    gear::TiXmlElement *fec = fem->FirstChildElement("FEC");
    if (!fec)
      throw gear::Exception("Could not find element <FEC>");
    gear::TiXmlElement *asic = fec->FirstChildElement("ASIC");
    if (!asic)
      throw gear::Exception("Could not find element <ASIC>");
    id = asic->Attribute("id");
    while (asic && id.compare("*")) {
      asic = fec->NextSiblingElement("FEM");
      id = asic->Attribute("id");
    }
    if (!asic || id.compare("*") != 0)
      throw gear::Exception("Could not find element <ASIC id=\"*\">");
    tmpElement = asic->FirstChildElement("enableShaping");
    if (tmpElement
        && !string(tmpElement->FirstChild()->Value()).compare("true"))
      cfgParameter.shaping = 1;
    else if (tmpElement
        && !string(tmpElement->FirstChild()->Value()).compare("false"))
      cfgParameter.shaping = 0;
    else
      throw gear::Exception(
          "Could not find element <enableShaping> with value of either true or false.");
    streamlog_out(MESSAGE) << "readMode      = " << cfgParameter.readMode << endl;
    streamlog_out(MESSAGE) << "compress      = " << cfgParameter.compress << endl;
    streamlog_out(MESSAGE) << "sampling (ns) = " << cfgParameter.samplingPeriod << endl;
    streamlog_out(MESSAGE) << "nTimeBuckets  = " << cfgParameter.numTimeBins << endl;
    streamlog_out(MESSAGE) << "shaping       = " << cfgParameter.shaping << endl;
  } catch (gear::Exception& e) {
    streamlog_out(ERROR) << "Error in parsing file " << filename << endl;
    streamlog_out(ERROR) << e.what() << endl;
  } catch (...) {
    streamlog_out(ERROR) << "Unknown error in parsing file " << filename << endl;
  }

}

/**
 * Write the header of the lcio file. It includes also writing the hole config file to the header.
 * cfgParameters are retrieved from the config file during this step.
 * \remark The config file should not be written to the header ?!
 * \param lcWrt The writer with the lcio file to write the header to.
 * \param cfgFilename The name including path of the config file in order to copy it to the header and to extract the config parameters.
 * \param timeStamp The time stamp used to add information to the header.
 * \param cfgParameter Here the extracted parameter of the config files are stored.
 */
bool EncodeHeader(LCWriter* lcWrt, string cfgFilename, TimeStamp &timeStamp, CFGParameter &cfgParameter) {
  //Create a RUN Header for the output
  LCRunHeaderImpl* runHdr = new LCRunHeaderImpl;

  // Run Number
  runHdr->setRunNumber(timeStamp.runNumber);

  // Detector Name
  string detName("LP-TPC");
  runHdr->setDetectorName(detName);

  // Description
  stringstream description;
  runHdr->setDescription(timeStamp.RunUID);

  // Active Sub Detector
  string tpcName("DriftChamberLP");
  runHdr->addActiveSubdetector(tpcName);

  //
  // Optional Parameters
  //
  // Add RunUID string to keep time stamp : (eg) R2008.10.03-12:38:31
  runHdr->parameters().setValue("RunUID", timeStamp.RunUID);
  runHdr->parameters().setValue("RUN Start Date", timeStamp.start.getDateString());
  runHdr->parameters().setValue("RUN Start Timestamp", timeStamp.start.unixTime());
  runHdr->parameters().setValue("RUN End Date", timeStamp.end.getDateString());
  runHdr->parameters().setValue("RUN End Timestamp", timeStamp.end.unixTime());

  streamlog_out(MESSAGE) << "Run start: " << timeStamp.start.getDateString() << endl;
  streamlog_out(MESSAGE) << "Run end: " << timeStamp.end.getDateString() << endl;

  // Sub Run Number
  runHdr->parameters().setValue("SubRunNumber", timeStamp.subRunNumber);

  // Parse Config File
  //------------------
  ParseConfigFile(cfgFilename, cfgParameter);

  // Read Mode : <-- Use frontendChannel
  runHdr->parameters().setValue("ReadMode", cfgParameter.readMode);

  // Compress : <-- Use frontendChannel
  runHdr->parameters().setValue("Compress", cfgParameter.compress);

  // Number of Time bins
  runHdr->parameters().setValue("NumTimeBins", cfgParameter.numTimeBins);

  // Sampling Period
  runHdr->parameters().setValue("SamplingPeriod", cfgParameter.samplingPeriod);

  // Shaping
  runHdr->parameters().setValue("Shaping", cfgParameter.shaping);

  //
  // Copy contents of CFG file as a string vector
  //
  // NOTE: We may be able to parse read mode and compress here, but actually
  //       they can be set individually for each channel.
  //       The info is encoded in packet header and saved into channelID1
  //       so that it is better use them later during analysis.
  //
  ifstream cfgFile(cfgFilename.c_str());

  if (cfgFile) {
    string line;
    StringVec lines;

    // Build string vector
    while (getline(cfgFile, line)) {
      lines.push_back(line);
    }
    runHdr->parameters().setValues("CFG.xml", lines);
  } else {
    streamlog_out(ERROR) << "Can't find CFG file " << cfgFilename.c_str() << endl;
    streamlog_out(ERROR) << "  Skip copying CFG file" << endl;
  }

  // Write header to the file.
  lcWrt->writeRunHeader(runHdr);
  delete runHdr;

  return true;
}

void PrintHelp(){
  printf("Required options:\n");
  printf("\t-d --inputDir<dir> Directory with raw data (.acq extension).\n");
  printf("\t-r --run <run_number> with or without leading zeros.\n");
  printf("Optional options:\n");
  printf("\t-v --verbose Print debug output.\n");
  printf("\t-o --outputDir <dir> Output directory of the converted files [default: raw data directory].\n");
  printf("\t-m --mergeFiles Merge all events of one run into a single slcio file.\n");
  printf("\t-c --channelMapping <channelMappingFile> Give the channel mapping file and the DBInit to write information to the data base.\n");
  printf("\t-D --dbinit <dbinit> Give the channel mapping file and the DBInit to write information to the data base.\n"
         "\t                     It REQUIRES the CondDB initialization string as argument in the form host:databasename:login:password[:port]\n");
  printf("\t-C --channelMappingDir <dbdir> Give the name of the directory in the data base where to write the channel mapping information [default: /LP1/T24_1/MicroMegas/channelmapping].\n");
  printf("\t-n --channelMappingCol <collection> Give the name channel mapping collection in the channel mapping lcio file [default: TPCChannelMapping].\n");
  printf("\t-E --extentRunTime <time[s]> Add a certain time before and after the run to be sure there is an entry in the run data base. Default: 1s.\n");
}

int main(int argc, char **argv) {
  if(argc == 1)
  {
      PrintHelp();
      return 0;
  }

  streamlog::out.init( std::cout , "MMraw2lcioConverter" ) ;
  streamlog::out.addLevelName<streamlog::MESSAGE>() ;
  streamlog::out.addLevelName<streamlog::DEBUG>() ;
  streamlog::logscope scope( streamlog::out ) ;
  scope.setLevel("MESSAGE");

  string dir = "";
  string run = "";
  string* outputDir = 0;
  string cfgFileName = "";
  string channelMapLCIOFile = "";
  string channelMapColName = "";
  string DBInit = "";
  string channelMapDir = "";
  bool mergeRunFiles = false;
  int maxevents = -1;
  // by default add 1 second before and after the run in run data base when the channel mapping is written
  int extentRunTime = 1;
  // Get options
  // Interpreting the command line
  static struct option long_options[] =
      {
          { "events", required_argument, 0, 'e' },
          { "inputDir", required_argument, 0, 'd' },
          { "outputDir", required_argument, 0, 'o' },
          { "mergeFiles", no_argument, 0, 'm' },
          { "run", required_argument, 0, 'r' },
          { "verbose", no_argument, 0, 'v' },
          { "dbinit", required_argument, 0, 'D' },
          { "channelMapping", required_argument, 0, 'c' },
          { "channelMappingDir", required_argument, 0, 'C'},
          { "channelMappingCol", required_argument, 0, 'n'},
          { "extentRunTime", required_argument, 0 ,'E'},
      };

  int OptionId;
  int OptionIndex;

  while(( OptionId = getopt_long(argc, argv, "vme:d:o:r:D:c:E:", long_options, &OptionIndex) ) != -1)
  {
      switch(OptionId)
      {
      case 'e':
      {
        maxevents = atoi(optarg);
        streamlog_out(MESSAGE) << "Maximum number of events to convert is set to: " << maxevents << "." << endl;
        break;
      }
      case 'd':
      {
        dir = string(optarg);
        streamlog_out(MESSAGE) << "Raw data directory is set to: " << dir << "." << endl;
        break;
      }
      case 'o':
      {
        outputDir = new string(optarg);
        streamlog_out(MESSAGE) << "Output directory is set to: " << dir << "." << endl;
        break;
      }
      case 'm':
      {
        mergeRunFiles = true;
        streamlog_out(MESSAGE) << "Will merge events of all *.acq files into one sclio file." << endl;
        break;
      }
      case 'r':
      {
        run = string(optarg);
        streamlog_out(MESSAGE) << "The following run will be converted: " << run << "." << endl;
        break;
      }
      case 'v':
      {
        scope.setLevel("DEBUG");
        streamlog_out(MESSAGE) << "Tunrned on the verbose output." << endl;
        break;
      }
      case 'D':
      {
        DBInit = std::string(optarg);
        streamlog_out(MESSAGE) << "DBinit is set to: " << DBInit << "." << endl;
        break;
      }
      case 'c':
      {
        channelMapLCIOFile = string(optarg);
        streamlog_out(MESSAGE) << "Channel mapping file is set to be: " << channelMapLCIOFile << "." << endl;
        break;
      }
      case 'n':
      {
        channelMapColName = string(optarg);
        streamlog_out(MESSAGE) << "Channel mapping collection is set to be: " << channelMapColName << "." << endl;
        break;
      }
      case 'C':
      {
        channelMapDir = string(optarg);
        streamlog_out(MESSAGE) << "Channel mapping directory in the data base is set to: " << channelMapDir << "." << endl;
        break;
      }
      case 'E':
        extentRunTime = atoi(optarg);
        break;
      case '?':
      {
        PrintHelp();
        return 0;
      }
      }
  }
  if( run == "" || dir == ""){
    streamlog_out(ERROR) << "Sorry I need more information." << endl;
    PrintHelp();
    return 1;
  }
  if( (DBInit != "" && channelMapLCIOFile == "") || (DBInit == "" && channelMapLCIOFile != "")){
    streamlog_out(ERROR) << "If you want to write ChannelMapping information to the data base DBinit and the channel mapping file are needed!" << endl;
    return 1;
  }
  if(dir[strlen(dir.c_str())-1] != '/')
    dir.append("/");
  if(outputDir && (*outputDir)[strlen(outputDir->c_str())-1] != '/')
    outputDir->append("/");
//  RunNumber = atoi(run.c_str());
  vector<string> files;
  try {
    files = GetFilesInDir(dir, run);
  } catch (Exception &e){
    streamlog_out(ERROR) << e.what() << endl;
  }
  if (files.size() == 0) {
    streamlog_out(ERROR) << "Failed to construct file names..." << endl;
    return 1;
  }
  LCWriter* lcWrt = 0;
  int eventloop = 0;
  FILE* fileData = 0;
  TimeStamp timeStamp;
  CFGParameter cfgParameter;
  vector<Pulse> Data[kNMods][kChannelMax];
  for (vector<string>::iterator it = files.begin(); it != files.end(); it++) {
    // Open ACQ file
    if (!OpenAcqFile(*it,dir,cfgFileName,fileData,timeStamp)) {
      streamlog_out(ERROR) << "Failed to open raw data file: " << *it << endl;
      return 1;
    }
    if (it == files.begin() || !mergeRunFiles) {
      // Create an LCWriter
      lcWrt = LCFactory::getInstance()->createLCWriter();
      lcWrt->setCompressionLevel(0);

      // Make LCIO filename
      string lcioFilename;
      if(outputDir){
        streamlog_out(MESSAGE) << "Output directory is set to: " << outputDir->c_str() << endl;
        lcioFilename = *outputDir;
        lcioFilename.append(*it,(*it).find_last_of("/")+1,(*it).size()-(*it).find_last_of("/")+1);
      } else {
        streamlog_out(MESSAGE) << "No output directory given. Will use input directory." << endl;
        lcioFilename.append(*it);
      }
      lcioFilename.erase(lcioFilename.end()-3,lcioFilename.end());
      lcioFilename.append("slcio");

      // Opem LCIO file
      lcWrt->open(lcioFilename, LCIO::WRITE_NEW);

      for (int iMod = 0; iMod < kNMods; iMod++) {
        for (int ich = 0; ich < kChannelMax; ich++) {
          for (int t = 0; t < kNumTimeBins; t++) {
            Data[iMod][ich] = std::vector<Pulse>();
          }
        }
      }
    }
    while (true) {
      // Reset event
      for (int iMod = 0; iMod < kNMods; iMod++) {
        for (int ich = 0; ich < kChannelMax; ich++) {
          for (int t = 0; t < kNumTimeBins; t++) {
            Data[iMod][ich].clear();
          }
        }
      }
      // Get event
      int eventNumber;
      if (DecodeNextEvent(eventNumber, fileData, timeStamp, Data) != true) {
        streamlog_out(MESSAGE) << "End of file reached " << endl;
        break;
      } else {
        eventloop++;
        // encode and resave
        if(!EncodeThisEvent(eventNumber, lcWrt, timeStamp, Data)){
          streamlog_out(ERROR) << "Failed to write event information." << endl;
          return 1;
        }
      }
      if (maxevents > 0 && eventloop > (maxevents-1)){
        streamlog_out(MESSAGE) << "Reached maximum number of events, which was set to: " << maxevents << endl;
        it = files.end()-1;
        break;
      }

    }  // End of main event loop !!!
  } // End loop
  // Create and write header at the end to add the start and end time of the run
  EncodeHeader(lcWrt, cfgFileName, timeStamp, cfgParameter);
  streamlog_out(MESSAGE) << "Transfered " << eventloop << " events." << endl;
  lcWrt->close();

  // Write channelmapping to data base
  if(channelMapLCIOFile != "")
  {
    LCReader* lcreader = 0;
    LCEvent* evt = 0;
    try{
      lcreader = lcio::LCFactory::getInstance()->createLCReader();
      lcreader->open(channelMapLCIOFile);
      evt = lcreader->readNextEvent();
    } catch (Exception &e){
      streamlog_out(ERROR) << "Failed to open the channel mapping file: " << channelMapLCIOFile << endl;
      streamlog_out(ERROR) << e.what() << endl;
      delete lcreader;
      lcreader = 0;
    }
    if(lcreader) {
      if(channelMapColName == "")
        channelMapColName = tpcconddata::ADCChannelMapping::getDefaultColName();;
      if(channelMapDir == "")
        channelMapDir = "/LP1/T24_1/MicroMegas/channelmapping";
      try{
        lcio::LCCollection* ChanMapCol = evt->getCollection(channelMapColName);
        streamlog_out(MESSAGE) << "\n\n Found channel mapping collection: " << channelMapColName << ".\n" << std::endl;
        lccd::DBInterface* DBChanMap = new lccd::DBInterface(DBInit, channelMapDir, true);
        streamlog_out(MESSAGE) << "Will use " << timeStamp.RunUID << endl;
        streamlog_out(MESSAGE) << "Will use " << timeStamp.start.getDateString() << "\t-\t" << timeStamp.end.getDateString() << endl;
        streamlog_out(MESSAGE) << "I will add " << extentRunTime << "s before and after the run before filling the data base." << endl;
        timeStamp.start += -1 * extentRunTime * 1000000000;
        timeStamp.end += extentRunTime * 1000000000;
        DBChanMap->storeCollection(timeStamp.start.timeStamp(), timeStamp.end.timeStamp(), ChanMapCol, "Channel Mapping information");
        delete DBChanMap;
        DBChanMap = 0;
        streamlog_out(MESSAGE) << "\n\n Storage of the channel mapping COMPLETE.\n" << std::endl;
      } catch (Exception &e) {
        streamlog_out(ERROR) << "Failed to load channel mapping collection " << channelMapColName << std::endl;
        streamlog_out(ERROR) << e.what() << std::endl;
      }
    }
  }
  delete lcWrt;
  return 0;
}

