/*
 * TIRCUBlock.hh
 *
 *  Created on: Jul 11, 2012
 *      Author: caiazza
 */

#ifndef TIRCUBLOCK_HH_
#define TIRCUBLOCK_HH_


#include "TIAltroBlock.hh"

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"

namespace marlintpc
{
/**
 * Interface for the class decoding the ALTRO RCU Block
 */
class TIRCUBlock : public TIAltroBlock
{
public:
    /**
     * Virtual destructor
     */
    virtual ~TIRCUBlock(){}

    /**
     * Add Data to an LCCollection
     * @param Collection
     */
    virtual void AddData(lcio::LCCollection* Collection) const = 0;

    /**
     * Adds one or more additional element to a collection containing the General Electronic info.
     * These element should be LCGeneric Object of type GenericADCElectronicsParameters
     * @param Collection where to store the element
     */
    virtual void AddGeneralElectronicInfo(lcio::LCCollection* Collection) const = 0;

    /**
     * Adds one or more additional element to a collection containing the Extended Electronic info.
     * @param Collection where to store the element
     */
    virtual void AddExtendedElectronicsInfo(lcio::LCCollection* Collection) const = 0;
};

/**
 * Implementation for the class decoding the ALTRO RCU Block for version 4.2+
 */
class TRCUBlock_V42: public TIRCUBlock
{
public:
    TRCUBlock_V42(std::vector<uint32_t>& Content, const std::vector<uint32_t>::iterator beginRCU );

    virtual ~TRCUBlock_V42();

    /**
     * Get the block lenght in number of 32 bit words
     * @return number of 32 bit words in the block
     */
    virtual uint32_t GetLength() const;

    /**
     * Verifies that the content of the block matches the expected content of the type of block we build.
     * @return True if the content of the block matches the expectation, False otherwise
     */
    virtual bool CheckIntegrity() const;

    /**
     * Returns a constant reference to the content of the block
     * @return vector of 32 bit data words
     */
    virtual const std::vector<uint32_t>& Content() const;

    /**
     * Returns a data word that identifies the block type
     * @return 32 bit data word identifying the block
     */
    virtual uint32_t GetIdentifier() const;

    /**
     * Allows to set additional parameters to the RunHeader.
     * @param RunHeader
     */
    virtual void SetAdditionalParameters(lcio::LCRunHeader* RunHeader);

    /**
     * Add Data to an LCCollection
     * @param Collection
     */
    virtual void AddData(lcio::LCCollection* Collection) const;

    /**
     * Adds one or more additional element to a collection containing the General Electronic info.
     * These element should be LCGeneric Object of type GenericADCElectronicsParameters
     * @param Collection where to store the element
     */
    virtual void AddGeneralElectronicInfo(lcio::LCCollection* Collection) const;

    /**
     * Adds one or more additional element to a collection containing the Extended Electronic info.
     * @param Collection where to store the element
     */
    virtual void AddExtendedElectronicsInfo(lcio::LCCollection* Collection) const;

    /**
     * Returns an iterator to the first word in the content vector after the end of this RCU block.
     * @return Iterator to past the end element of the content vector
     */
    virtual std::vector<uint32_t>::iterator GetBlockEnd() const;

private:
    std::vector<uint32_t>* _content;

    std::vector<uint32_t>::iterator _BlockStart;

    static const uint32_t WORDCODEMASK = 0xfc000000;
    static const uint32_t WORDVALUEMASK = ~WORDCODEMASK;
    static const uint32_t TRAILERLENGTHMASK = 0x0000007f;
    static const uint32_t POLARITYMASK = 0x00000010;
    static const uint32_t FREQUENCYCODEMASK = 0x000000f0;
    static const uint32_t ZEROSUPPRESSIONMASK = 0x00080000;
    static const uint32_t GLITCHFILTERMASK = 0x00003000;
    static const uint32_t PRESAMPLESMASK = 0x00060000;
    static const uint32_t POSTSAMPLESMASK = 0x0001c000;
    static const uint32_t PRETRIGGERSAMPLESMASK = 0x00f00000;
    static const uint32_t RCUADDRESSMASK = 0x0000ff80;
    static const uint32_t RCUFWVERMASK = 0x03ff0000;
    static const uint32_t RCUERRORSTATUSMASK = 0x0ffff000;


};//end of implementation for version 4.2

}


#endif /* TIRCUBLOCK_HH_ */
