/*
 * TEventBlock_V42.cpp
 *
 *  Created on: Jul 10, 2012
 *      Author: caiazza
 */

#include "TIEventBlock.hh"
#include "TIRCUBlock.hh"
#include <map>
#include <string>
#include <sstream>

//LCIO
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCFlagImpl.h"

namespace marlintpc
{

TEventBlock_V42::TEventBlock_V42(std::vector<uint32_t>& Content):
        _content(&Content),
        _RCUDataStart(7)
{
    if(!CheckIntegrity())
    {
        std::stringstream message;
        message << "Failed integrity check while analyzing the Event Block" << std::endl;
        message << "Dumping the block Content" << std::hex << std::endl;
        std::vector<uint32_t>::iterator itContent = _content->begin();
        for(; itContent != _content->end(); ++itContent)
        {
            message << *itContent << std::endl;
        }
        message << std::dec;
        throw TAltroConverterDecodeException("TEventBlock_V42", message.str());
    }
}

TEventBlock_V42::~TEventBlock_V42()
{
    delete _content;
}

uint32_t
TEventBlock_V42::GetLength() const
{
    return _content->size();
}

bool
TEventBlock_V42::CheckIntegrity() const
{
    bool result = true;
    result = result && (_content != 0)
            && (GetLength() == _content->at(0) + 1)
            && (_content->at(2) == 0x22222222);

    return result;
}

const std::vector<uint32_t>&
TEventBlock_V42::Content() const
{
    return (*_content);
}

uint32_t
TEventBlock_V42::GetIdentifier() const
{
    return (*_content)[2];
}

lcio::LCEvent*
TEventBlock_V42::CreateLCEvent(std::string& DataCollectionName)
{
    lcio::LCEventImpl* Event = new lcio::LCEventImpl();
    Event->setEventNumber(static_cast<int>(GetNEvent()));
    Event->setTimeStamp(GetTimestamp().timeStamp());
    Event->setWeight(1.0);
    Event->parameters().setValue("DBOX Trigger Number", static_cast<int>((*_content)[4]));
    Event->parameters().setValue("TLU Event Number", static_cast<int>((*_content)[5]));

    lcio::LCCollectionVec* RawData = new lcio::LCCollectionVec(lcio::LCIO::TRACKERRAWDATA);
    lcio::LCFlagImpl trkFlag(0);
    trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 );
    RawData->setFlag( trkFlag.getFlag() );

    try
    {
        std::vector<uint32_t>::iterator RCUStart = _content->begin() + _RCUDataStart;
        while(RCUStart != _content->end())
        {
            TRCUBlock_V42 RCUBlock((*_content), RCUStart);
            RCUBlock.AddData(RawData);
            RCUStart = RCUBlock.GetBlockEnd();
        }
    }
    catch(TAltroConverterException &e)
    {
        std::stringstream message;
        message << "failed decoding the event block" << std::endl;
        message << "Dumping the block Content" << std::hex << std::endl;
        std::vector<uint32_t>::iterator itContent = _content->begin();
        for(; itContent != _content->end(); ++itContent)
        {
            message << *itContent << std::endl;
        }
        message << std::dec;
        std::cout << message.str() << std::endl;
        delete RawData;
        throw TAltroConverterDecodeException("TEventBlock_V42", message.str());
    }

    Event->addCollection(RawData, DataCollectionName);
    return Event;
}

lcio::LCTime
TEventBlock_V42::GetTimestamp() const
{
    return lcio::LCTime(static_cast<EVENT::long64>((*_content)[5]));
}

uint32_t
TEventBlock_V42::GetNEvent() const
{
    return (*_content)[3];
}

lcio::LCCollection*
TEventBlock_V42::GetGeneralElectronicPar() const
{
    lcio::LCCollectionVec* GenericParCollection = new lcio::LCCollectionVec( lcio::LCIO::LCGENERICOBJECT );
    std::vector<uint32_t>::iterator RCUStart = _content->begin() + _RCUDataStart;
    while(RCUStart != _content->end())
    {
        TRCUBlock_V42 RCUBlock((*_content), RCUStart);
        RCUBlock.AddGeneralElectronicInfo(GenericParCollection);
        RCUStart = RCUBlock.GetBlockEnd();
    }
    return GenericParCollection;
}

lcio::LCCollection*
TEventBlock_V42::GetExtendedElectronicPar() const
{
    lcio::LCCollectionVec* ExtendedParCollection = new lcio::LCCollectionVec( lcio::LCIO::LCGENERICOBJECT );
    std::vector<uint32_t>::iterator RCUStart = _content->begin() + _RCUDataStart;
    while(RCUStart != _content->end())
    {
        TRCUBlock_V42 RCUBlock((*_content), RCUStart);
        RCUBlock.AddExtendedElectronicsInfo(ExtendedParCollection);
        RCUStart = RCUBlock.GetBlockEnd();
    }
    return ExtendedParCollection;
}

void
TEventBlock_V42::SetAdditionalParameters(lcio::LCRunHeader*  /*RunHeader*/)
{

}

void
TEventBlock_V42::SetRCUDataStart(uint32_t Word)
{
    _RCUDataStart = Word;
}

} // end namespace

