/*
 * Altro2LcioConverter.cpp
 *
 *  Created on: Jul 14, 2011
 *      Author: caiazza
 */

//Collecting all common includes and declarations
#include "Altro2LcioConverter.hh"
#include <stdlib.h>

using namespace A2Lconverter;

///Some utility function

void
A2Lconverter::checkCommandLine(int argc, char* argv[], ProgramOptions* theOptions)
        throw (std::runtime_error)
{
    //Command line checks
    //Check if there are too many or too few arguments in the command line
    if (argc < 2 || argc > 9)
    {
        std::string error("ERROR: Wrong number of arguments while calling this routine. \n");
        throw BadCommandLineException(error);
    }

    //Check for the different possible options that can be set from the command line
    for (int nArg = 1; nArg < (argc - 1); nArg++)
    {
        std::string argVal = argv[nArg];
        if (argVal == "-r")
        {
            theOptions->printRunHeader = true;
        }
        else if (argVal == "-x")
        {
            theOptions->printeXtensiveInfo = true;
        }
        else if (argVal == "-t")
        {
            theOptions->timestampmode = atoi(argv[nArg + 1]);
            if (!(theOptions->timestampmode == 1 || theOptions->timestampmode == 2
                    || theOptions->timestampmode == 3))
            {
                std::string error("ERROR: Time Stamp mode not recognized.\n");
                throw BadCommandLineException(error);
            }
            nArg++;
        }
        else if (argVal == "-e")
        {
            theOptions->whatEventPrint = atoi(argv[nArg + 1]);
            if (theOptions->whatEventPrint <= 0)
            {
                std::string error("ERROR: The event number must be a positive integer.\n");
                throw BadCommandLineException(error);
            }
            nArg++;
        }
        else if (argVal == "-c")
        {
            if (nArg + 1 >= argc - 1)
            {
                std::string error("ERROR: Missing output file name or input file name.\n");
                throw BadCommandLineException(error);
            }
            theOptions->outputFileName = argv[nArg + 1];
            theOptions->doConvert = true;
            nArg++;
        }
        else
        {
            std::string error("ERROR: Unexpected parameter detected in the routine call.\n ");
            throw BadCommandLineException(error);
        }
    }

    std::string tmpFileName = argv[argc - 1];
    size_t lastdotpos = tmpFileName.find_last_of('.');
    tmpFileName.erase(lastdotpos);
    size_t lastuspos = tmpFileName.find_last_of('_');
    theOptions->lastFileNumber = atoi(tmpFileName.substr(lastuspos + 1).c_str());
    tmpFileName.erase(lastuspos);
    theOptions->baseFileName = tmpFileName;
}

void
A2Lconverter::printusage(char* exeName)
{
    std::cerr
            << std::endl
            << "Usage: \n"
            << exeName
            << " [-r] [-x] [-t <1|2|3 (default)>] [-e <event number>] [-c <outfile.slcio>] infile.dat"
            << std::endl << std::endl;
    std::cerr << "-r: Visualize the Run Header data" << std::endl;
    std::cerr
            << "-x: Visualize extended run data extracting information from other block than the run header alone"
            << std::endl;
    std::cerr << "-t: Defines the mode to time stamp the events" << std::endl;
    std::cerr << "\t1: Time stamp from the event block" << std::endl;
    std::cerr << "\t2: One time stamp for all the event derived from the run header time stamp"
            << std::endl;
    std::cerr
            << "\t3: Different time stamp for each event calculated from the duration of the run. This is the default mode."
            << std::endl;
    std::cerr << "-e: Visualize the data relative to the event specified" << std::endl;
    std::cerr << "-c: Perform the data conversion. outfile.slcio is the name of the output file."
            << " The extension of the file will be automatically be set to slcio" << std::endl;
    std::cerr
            << "infile.dat: is the first of the raw data file of the run to convert. If the run is logged in more than"
            << "one data file the other data file must be located in the same folder of the first and will be loaded automatically"
            << "if the filename has a format: <base name>_<sequential file number>.dat"
            << std::endl;
}

void
A2Lconverter::GenerateOutFileName(ProgramOptions & theOptions)
{
    size_t lastdotpos = theOptions.outputFileName.find_last_of('.');
    if (lastdotpos == std::string::npos)
        theOptions.outputFileName += ".slcio";
    else if (!(theOptions.outputFileName.substr(lastdotpos) == ".slcio"))    //The extension is not slcio
    {
        theOptions.outputFileName += ".slcio";
    }
}

void
A2Lconverter::vectorDelete(std::vector<std::ifstream*> theVector)
{
    for (std::vector<std::ifstream*>::iterator it = theVector.begin(); it != theVector.end(); it++)
    {
        if ((*it)->is_open())
            (*it)->close();
        delete (*it);
    }
}

/*
 * END of the utility functions
 * Start of the main routine of the program
 */

// TODO adapt routine for only two command line options
// TODO include command line options to add at convert time a custom detector name and description
// TODO include a way to add additional parameters to the run header via a txt(ini file) or xml file
int
main(int argc, char* argv[])
{

    //Initialize the Program control variables
    ProgramOptions theOptions;
    theOptions.printRunHeader = false;
    theOptions.printeXtensiveInfo = false;
    theOptions.timestampmode = 3;
    theOptions.whatEventPrint = -1;
    theOptions.doConvert = false;
    theOptions.baseFileName = "";
    theOptions.outputFileName = "";

    //Input files
    std::vector<std::ifstream*> inputFileVec;    //Contains a vector of the input file
    std::ifstream* inputFile;    //Points to the stream I'm currently working on

    //Data buffer and control variables
    AltroBlock* bufferBlock = 0;
    AltroBlock* decodingBlock = 0;
    EVENT::long64 timestampSkip = 0;

    //LCIO objects
    lcio::LCWriter* lciowriter = 0;

    //Storage objects for particular blocks
    SORBlock* runHeader = 0;
    EORBlock* runTrailer = 0;

    //BEGIN ALGORYTHM//
    // Control the correctness of the command line arguments and fill the internal struct with the command options
    try
    {
        checkCommandLine(argc, argv, &theOptions);
    }
    catch (std::runtime_error &e)
    {
        std::cerr << e.what() << std::endl;
        printusage(argv[0]);
        return 2;
    }

    //Tries to open all the raw files and put the stream pointers into the input file vector
    //Once the vector is filled it will set the working file to the first file and check that it still open
    try
    {
        do
        {
            inputFile = new std::ifstream;                  //creating the new stream
            inputFile->clear();                             //Clears the error flags in the input file
            inputFile->exceptions(std::ifstream::badbit);   //Open the file throwing exceptions in case of badbit only

            // Reconstruct the input file name of the next file to open assuming that all the files are in the same folder
            // and that they are named as such: basename_filenumber.dat
            //This is the standard naming of the Altro file as they are saved from the DAQ
            std::ostringstream inputFileName;
            inputFileName << theOptions.baseFileName << "_" << theOptions.lastFileNumber << ".dat";
            theOptions.lastFileNumber++;
            inputFile->open(inputFileName.str().c_str(), std::ios_base::binary);    //Opens the file as a binary file
            //Check that the file is correctly open and stores the handle to the file in input file vector
            if (inputFile->is_open())
            {
                inputFileVec.push_back(inputFile);    //Copying the stream pointer in the vector
                std::cout << "Added file " << inputFileName.str() << " to the input file list."
                        << std::endl;
            }
        } while (inputFile->is_open());
        delete inputFile;       //Delete the last input file as it was not possible to open it

        // Sets the working input file to the first file of the list
        inputFile = inputFileVec[0];
        std::cout << std::endl;
    }
    catch (std::ifstream::failure &f)
    {
        std::cerr << "ERROR: Problem accessing one of the input file to convert" << std::endl;
        std::cerr << f.what() << std::endl;
        vectorDelete(inputFileVec);
        return 2;
    }
    catch (std::exception &e)
    {
        std::cerr << "ERROR: Problem when creating the input file list." << std::endl;
        std::cerr << e.what() << std::endl;
        vectorDelete(inputFileVec);
        return 2;
    }assert(inputFile->is_open());
    assert(inputFile == inputFileVec[0]);
    assert(inputFile->good());

    //Generate the output file name and create the lcwriter object only if I need to convert
    if (theOptions.doConvert)
    {
        try
        {
            GenerateOutFileName(theOptions);
            lciowriter = new UTIL::LCSplitWriter(lcio::LCFactory::getInstance()->createLCWriter(),
                    2040109465);
            lciowriter->open(theOptions.outputFileName);
        }
        catch (std::exception& e)
        {
            std::cerr << "ERROR: Exception caught when trying to create the output LCIO file"
                    << std::endl;
            std::cerr << e.what() << std::endl;
            if (lciowriter != 0)
            {
                lciowriter->close();
                delete lciowriter;
                lciowriter = 0;
            }
            vectorDelete(inputFileVec);
            return 2;
        }assert(inputFile->is_open());
        assert(lciowriter !=0);
    }

    //It process the run header and trailer
    try
    {
        //It was set up to 0 in the initialization. There is no reason why it should be not 0.
        assert(runHeader == 0);

        runHeader = new SORBlock(inputFile);
        if (!(runHeader->CheckIntegrity()))
        {
            std::string err =
                    "ERROR: Unable to verify the integrity of the run header. The file might be corrupted or wrong.\n";
            throw UnhandledBlockException(err);
        }
        runHeader->Decode();
        EventBlock::baseTStamp = runHeader->GetTimestampns();
        if (theOptions.printRunHeader || theOptions.printeXtensiveInfo)
            std::cout << runHeader->PrintSummary() << std::endl;
        if (theOptions.doConvert)
            runHeader->Write2LCIO(lciowriter);
    }
    catch (AltroBlockException &e)
    {
        std::cerr << "ERROR: Exception caught while elaborating the run header." << std::endl;
        std::cerr << e.what() << std::endl;
        if (runHeader != 0)
        {
            delete runHeader;
            runHeader = 0;
        }
        if (lciowriter != 0)
        {
            lciowriter->close();
            delete lciowriter;
            lciowriter = 0;
        }
        vectorDelete(inputFileVec);
        return 2;
    }assert(runHeader != 0);
    assert(inputFile->is_open());

    //Now I look for the EOR block to acquire the extended info necessary to calculate the timestamps
    //TODO Print the extensive summary if required
    try
    {
        inputFile = inputFileVec.back();    //change the input file to the last file
        inputFile->clear();
        if (!(inputFile->is_open()))
        {
            throw AltroBlockException(
                    "ERROR: The last input file was not open when trying to access the EOR block");
        }
        inputFile->exceptions(std::ifstream::badbit);    //not calling exception if it's end of file
        inputFile->seekg(0, std::ios::end);
        // the EOR record should start 6 word before the end in all data formats
        int startEORoffset = static_cast<int>(inputFile->tellg()) - 6 * AltroBlock::WORD32;

        //No reason why the run trailer should be not 0
        assert(runTrailer==0);
        runTrailer = new EORBlock(inputFile, startEORoffset);
        if (!(runTrailer->CheckIntegrity()))    //check that the file is accessible and the block id is correct
        {
            std::string err =
                    "ERROR: Unable to verify the integrity of the run trailer. The file might be corrupted or wrong.\n";
            throw UnhandledBlockException(err);
        }
        runTrailer->Decode();    // read the data from the file

        switch (theOptions.timestampmode)
        {
            case 1:
                //in this case I flag this value with -1 to use the timestamp from the event block
                timestampSkip = -1;
                break;

            case 2:
                //in this mode I'm using the same time stamp for all events coming from the run Header
                timestampSkip = 0;
                break;

            case 3:
                //in this mode I will add a fixed number of ns to each event
                timestampSkip = static_cast<EVENT::long64>((runTrailer->GetTimestampns()
                        - runHeader->GetTimestampns()) / runTrailer->GetTotEvents());break;

                default:
                //This should not be possible
                        throw std::runtime_error("Unrecognized time stamp mode.");
                    }

        EventBlock::tstampSkip = timestampSkip;
    }
    catch (AltroBlockException &a)
    {
        std::cerr << "ERROR: Exception caught while elaborating the End of run block " << std::endl;
        std::cerr << a.what() << std::endl;
        if (runTrailer != 0)
        {
            delete runTrailer;
            runTrailer = 0;
        }
        if (runHeader != 0)
        {
            delete runHeader;
            runHeader = 0;
        }
        if (lciowriter != 0)
        {
            lciowriter->close();
            delete lciowriter;
            lciowriter = 0;
        }
        vectorDelete(inputFileVec);
        return 2;
    }
    catch (std::exception &e)
    {
        std::cerr << "ERROR: Exception caught while elaborating the End of run block " << std::endl;
        std::cerr << e.what() << std::endl;
        if (runHeader != 0)
        {
            delete runHeader;
            runHeader = 0;
        }
        if (lciowriter != 0)
        {
            lciowriter->close();
            delete lciowriter;
            lciowriter = 0;
        }
        vectorDelete(inputFileVec);
        return 2;
    }assert(runHeader != 0);
    assert(runTrailer !=0);
    assert(inputFile->is_open());

    //Now the main decoding. The outer for cycle runs over the input files
    //TODO Implement the summary of the event data
    try
    {
        for (std::vector<std::ifstream*>::iterator itFile = inputFileVec.begin();
                itFile != inputFileVec.end(); itFile++)
        {
            inputFile = *itFile;
            inputFile->clear();
            inputFile->exceptions(std::ifstream::badbit);
            inputFile->seekg(0, std::ios::beg);
            if (itFile == inputFileVec.begin())
            {
                inputFile->seekg(runHeader->GetBlockEnd(), std::ios::beg);
            }

            while (!inputFile->eof())
            {
                if (bufferBlock != 0)
                    delete bufferBlock;
                if (decodingBlock != 0)
                    delete decodingBlock;

                bufferBlock = new AltroBlock(inputFile);
                if (!(bufferBlock->CheckIntegrity()))
                {
                    throw AltroBlockException(
                            "Block Integrity check failed. Cannot access the raw data");
                }
                decodingBlock = bufferBlock->Identify();
                decodingBlock->Decode();
                if (theOptions.doConvert)
                {
                    decodingBlock->Write2LCIO(lciowriter);
                }
                inputFile->seekg(decodingBlock->GetBlockEnd(), std::ios::beg);
                inputFile->peek();

                if (bufferBlock != 0)
                {
                    if (bufferBlock == decodingBlock)
                        decodingBlock = 0;
                    delete bufferBlock;
                    bufferBlock = 0;
                }
                if (decodingBlock != 0)
                {
                    delete decodingBlock;
                    decodingBlock = 0;
                }
            } //End of the block cycle
        } //End of the input file cycle
    }
    catch (AltroBlockException &a)
    {
        std::cerr << "ERROR: Exception caught while performing the decoding of the file."
                << std::endl;
        std::cerr << a.what() << std::endl;
        if (bufferBlock != 0)
        {
            if (bufferBlock == decodingBlock)
                decodingBlock = 0;
            delete bufferBlock;
            bufferBlock = 0;
        }
        if (decodingBlock != 0)
        {
            delete decodingBlock;
            decodingBlock = 0;
        }
        if (runTrailer != 0)
        {
            delete runTrailer;
            runTrailer = 0;
        }
        if (runHeader != 0)
        {
            delete runHeader;
            runHeader = 0;
        }
        if (lciowriter != 0)
        {
            lciowriter->close();
            delete lciowriter;
            lciowriter = 0;
        }
        vectorDelete(inputFileVec);
        return 2;
    }
    catch (std::exception &e)
    {
        std::cerr << "ERROR: Exception caught while elaborating the End of run block " << std::endl;
        std::cerr << e.what() << std::endl;
        if (bufferBlock != 0)
        {
            if (bufferBlock == decodingBlock)
                decodingBlock = 0;
            delete bufferBlock;
            bufferBlock = 0;
        }
        if (decodingBlock != 0)
        {
            delete decodingBlock;
            decodingBlock = 0;
        }
        if (runTrailer != 0)
        {
            delete runTrailer;
            runTrailer = 0;
        }
        if (runHeader != 0)
        {
            delete runHeader;
            runHeader = 0;
        }
        if (lciowriter != 0)
        {
            lciowriter->close();
            delete lciowriter;
            lciowriter = 0;
        }
        vectorDelete(inputFileVec);
        return 2;
    }assert(bufferBlock == 0);
    assert(decodingBlock == 0);

    if (runTrailer != 0)
    {
        delete runTrailer;
        runTrailer = 0;
    }
    if (runHeader != 0)
    {
        delete runHeader;
        runHeader = 0;
    }
    if (lciowriter != 0)
    {
        lciowriter->close();
        delete lciowriter;
        lciowriter = 0;
    }
    vectorDelete(inputFileVec);
    return 0;
}

/**
 * End of MAIN
 */

