/*
 * AltroBlock.hh
 *
 *  Created on: Jul 20, 2011
 *      Author: caiazza
 */

#ifndef ALTROBLOCK_HH_
#define ALTROBLOCK_HH_

#include "AltroBlocksLib.h"
#include <stdint.h>

/**
 * Base class for the Altro decoder
 */
namespace A2Lconverter
{

class AltroBlock
{
public:

    /**
     * This enum list all the possible blocks managed by the class to identify the type of the class itself
     */
    enum BlockIds
    {
        UNKNOWN = 0x00000000,
        BLOCK_SOR = 0x11111111,
        BLOCK_EOR = 0x33333333,
        BLOCK_POR = 0x11112222,
        BLOCK_COR = 0x11113333,
        BLOCK_EVENT = 0x22222222,
        BLOCK_BOF = 0x44444444,
        BLOCK_EOF = 0x55555555,
        BLOCK_RCU = 0xFFFFFFFF,
        BLOCK_DATA = 0x1111FFFF
    };

    static const uint32_t WORD32 = 4;

    /**
     * Class constructors
     * The first constructor it's the one that it should normally employed to use this class.
     * It gets a copy of the starting address of the block from which to start reading the file.
     * It is to be used when you want to feed the block the stream from which it begins.
     * The routine will automatically initialize the necessary variables.
     * It can be used as a default constructor in which case the stream pointer will be set to 0
     * to be reassigned later when initializing the object
     * but it can be useful if one wants to include this class in an array of blocks.
     * This parameter is implemented as a const pointer because you're not supposed to change the stream
     * you're working on during the object construction
     * As an option the constructor can also get a streampos which is interpreted as an offset from the
     * current read point on the stream from which to start to read.
     * The system will usually read an amount of data based on the correct identification of the block
     * but in case you want to explicitly limit the amount of data to read you can also set the third parameter
     * to a nonzero value. If the third parameter is 0 as in the default the system will read the amount of data
     * determined by the block identity.
     */

    AltroBlock(std::ifstream* const inputFile = 0);

    AltroBlock(std::ifstream* const inputFile, int start_address,
            uint32_t length = 0);

    /**
     * This constructor is implemented to allow the user to directly connect a vector containing the block data to
     * to the block class. This version of the constructor use the input reference to fill the inner vector before
     * performing any decoding
     */
    AltroBlock(const std::vector<uint32_t> & refContent);

    /**
     * Copy constructor
     */
    AltroBlock(const AltroBlock& anotherBlock);

    /**
     * Destructor
     */
    virtual
    ~AltroBlock();

    /**
     * Block Identifier
     * This function identify which kind of data are stored in the block and return a pointer of a new object
     * of the corrected derived type to the caller
     */
    AltroBlock*
    Identify();

    /**
     * Decode command
     * Instruct the class to read the data. You can optionally pass one integer parameter that will be interpreted
     *  differently by the different subclasses
     */
    virtual void
    Decode();

    /**
     * LCIOWriter: This is a virtual routine that writes the block content in the appropriate section of the LCIO file
     * using the LCIO writer object obtained in input.
     * This routine must be redefined in each subclass
     */

    virtual void
    Write2LCIO(IO::LCWriter* lcwriter) const
    {
    }

    /**
     * This function writes down a summary of the content of the block and returns it as a string stream to the caller
     */
    virtual std::string
    PrintSummary() const
    {
        return "";
    }

    /**
     * This function verifies that the block is correctly formed.
     * In the case of the base class it just verify that the stream is open or that the content vector is not 0
     */
    virtual bool
    CheckIntegrity() const;

    /**
     * Get functions for different block members
     */
    BlockIds
    GetId() const;
    uint16_t
    GetDataformat() const;
    uint16_t
    GetDAQVersion() const;
    UTIL::LCTime
    GetTimestamp() const;
    EVENT::long64
    GetTimestampns() const;
    uint32_t
    GetBlockLength() const;
    uint32_t
    GetBlockEnd() const;
    uint32_t
    GetBlockStart() const;

protected:
    std::ifstream* inputFile;
    int start;    ///The position in the stream from where the block starts
    uint32_t length;
    int end;     ///The position in the stream where the block ends
    BlockIds id;
    mutable UTIL::LCTime timestamp;
    std::vector<uint32_t> content;

    static uint16_t dataformat;
    static uint16_t daqversion;

    /**
     * This function reads the next four bytes in the block stream and formats them in a int value
     * inverting the byte ordering from big to little endian
     */
    uint32_t
    ReadWord32() const;

    /**
     * This function perform the byte ordering inversion from big to little endian starting from an uint32_t input
     */
    uint32_t
    InvertWord32(uint32_t &value) const;

    /**
     * This function fills the content vector with the data coming from the input stream up to a maximum of "length" 32 bit words
     * starting from the current position of the get pointer.
     * If eof is encountered before reading the maximum number of word it will return an ios::failure exception
     * By default the function cleans the content vector before starting. As an option it's possible to append the new data to the vector
     */
    void
    FillContent(uint32_t length, bool append = false);
};

/**
 * Exception class for the ALTRO block library
 */
class AltroBlockException : public std::runtime_error
{
public:
    AltroBlockException()
            : std::runtime_error("Error in the handling of the Altro Block")
    {

    }

    AltroBlockException(const std::string &err)
            : std::runtime_error(err)
    {

    }

    AltroBlockException(const char* err)
            : std::runtime_error(err)
    {

    }
};

class UnhandledBlockException : public AltroBlockException
{
public:
    UnhandledBlockException()
            : AltroBlockException("Unknown type of block encountered")
    {

    }

    UnhandledBlockException(const std::string &err)
            : AltroBlockException(err)
    {

    }
};

}

#endif /* ALTROBLOCK_HH_ */
