/*
 * AltroBlock.cpp
 *
 *  Created on: Jul 20, 2011
 *      Author: caiazza
 */

#include "AltroBlocksLib.h"

namespace A2Lconverter
{

uint16_t AltroBlock::daqversion = 0;
uint16_t AltroBlock::dataformat = 1;

AltroBlock::AltroBlock(std::ifstream * const inputFile)
        : inputFile(inputFile),
                start((inputFile == 0) ? 0 : static_cast<int>(inputFile->tellg())),length(0),
                end(start + length * WORD32), id(AltroBlock::UNKNOWN),
                timestamp(0), content(0)
                {

                }

AltroBlock::AltroBlock(std::ifstream* const inputFile, int start_address, uint32_t length)
        : inputFile(inputFile), start(start_address), length(length), end(start + length * WORD32),
                id(AltroBlock::UNKNOWN), timestamp(0), content(0)
{

}

AltroBlock::AltroBlock(const std::vector<uint32_t> & refContent)
        : inputFile(0), start(0), length(0), end(0), id(AltroBlock::UNKNOWN), timestamp(0),
                content(refContent)
{

}

AltroBlock::AltroBlock(const AltroBlock& anotherBlock)
        : inputFile(0), start(0), length(0), end(0), id(anotherBlock.id), timestamp(0), content(0)
{
    if (anotherBlock.content.size() == 0)
    {
        inputFile = anotherBlock.inputFile;
        start = anotherBlock.start;
    }
    else
    {
        content = anotherBlock.content;
    }
}

AltroBlock::~AltroBlock()
{

}

AltroBlock*
AltroBlock::Identify()
{
    AltroBlock* theNewBlock;
    inputFile->seekg(start + (2 * WORD32), std::ios::beg);
    switch (ReadWord32())
    {
        case (AltroBlock::BLOCK_SOR):
            id = AltroBlock::BLOCK_SOR;
            theNewBlock = new SORBlock(inputFile, start);
            break;
        case (AltroBlock::BLOCK_EOR):
            id = AltroBlock::BLOCK_EOR;
            theNewBlock = new EORBlock(inputFile, start);
            break;
        case (AltroBlock::BLOCK_BOF):
            id = AltroBlock::BLOCK_BOF;
            theNewBlock = new BOFBlock(inputFile, start);
            break;
        case (AltroBlock::BLOCK_EOF):
            id = AltroBlock::BLOCK_EOF;
            theNewBlock = new EOFBlock(inputFile, start);
            break;
        case (AltroBlock::BLOCK_POR):
            id = AltroBlock::BLOCK_POR;
            theNewBlock = new PORBlock(inputFile, start);
            break;
        case (AltroBlock::BLOCK_COR):
            id = AltroBlock::BLOCK_COR;
            theNewBlock = new CORBlock(inputFile, start);
            break;
        case (AltroBlock::BLOCK_EVENT):
            id = AltroBlock::BLOCK_EVENT;
            theNewBlock = new EventBlock(inputFile, start);
            break;
        default:
            id = AltroBlock::UNKNOWN;
            inputFile->seekg(start + (2 * WORD32), std::ios::beg);
            std::ostringstream err;
            err
                    << "Unidentifiable block located when trying to identify the generic ALTRO block. Identifier read: "
                    << std::hex << ReadWord32() << "\n";
            std::string tmp = err.str();
            inputFile->seekg(start, std::ios::beg);
            throw UnhandledBlockException(tmp);
    }
    inputFile->seekg(start, std::ios::beg);    //Go back to the block start
    return theNewBlock;
}

void
AltroBlock::Decode()
{
    if (content.size() == 0)
    {
        if (inputFile != 0)
        {
            inputFile->clear();
            inputFile->exceptions(std::ifstream::badbit | std::ifstream::eofbit);
            inputFile->seekg(start, std::ios::beg);
            length = ReadWord32() + 1;
            end = start + length * WORD32;
            inputFile->seekg(start, std::ios::beg);
            FillContent(length);
        }
        else
            throw AltroBlockException(
                    "ERROR: Unable to decode a block not linked to any content or any input file");
    }
    else
    {
        length = content.size();
        end = start + length * WORD32;
    }
}

//If the content vector is not empty and if the input file is open and readable the Block passes the test
bool
AltroBlock::CheckIntegrity() const
{
    if (content.size() != 0)
    {
        return true;
    }
    else if (inputFile != 0)
    {
        if (inputFile->is_open())
        {
            inputFile->seekg(start, std::ios::beg);
            inputFile->peek();
            return inputFile->good();
        }
    }
    return false;
}

AltroBlock::BlockIds
AltroBlock::GetId() const
{
    return id;
}

uint16_t
AltroBlock::GetDAQVersion() const
{
    return daqversion;
}

uint16_t
AltroBlock::GetDataformat() const
{
    return dataformat;
}

UTIL::LCTime
AltroBlock::GetTimestamp() const
{
    return timestamp;
}

EVENT::long64
AltroBlock::GetTimestampns() const
{
    return timestamp.timeStamp();
}

uint32_t
AltroBlock::GetBlockLength() const
{
    return length;
}

uint32_t
AltroBlock::GetBlockEnd() const
{
    return end;
}

uint32_t
AltroBlock::GetBlockStart() const
{
    return start;
}

uint32_t
AltroBlock::ReadWord32() const
{
    char tmp[WORD32];
    inputFile->read(tmp, WORD32);
    return static_cast<uint32_t>(static_cast<unsigned char>(tmp[0]))
            | static_cast<uint32_t>(static_cast<unsigned char>(tmp[1])) << 8
            | static_cast<uint32_t>(static_cast<unsigned char>(tmp[2])) << 16
            | static_cast<uint32_t>(static_cast<unsigned char>(tmp[3])) << 24;
}

uint32_t
AltroBlock::InvertWord32(uint32_t &value) const
{
    return (value & 0xff000000) | (value & 0x00ff0000) << 8 | (value & 0x0000ff00) << 16
            | (value & 0x000000ff) << 24;
}

void
AltroBlock::FillContent(uint32_t length, bool append)
{
    if (!append && content.size() != 0)
    {
        content.clear();
    }
    content.reserve(length);
    for (uint32_t i = 0; i < length; i++)
    {
        content.push_back(ReadWord32());
    }
}

}

