/*
 * RCUBlock.cpp
 *
 *  Created on: Jul 27, 2011
 *      Author: caiazza
 */

#include "AltroBlocksLib.h"

namespace A2Lconverter
{

const RCUBlock::TrailerMasks RCUBlock::V410MASKS = { 0xfc000000, 0x00000000, 0x04000000, 0x08000000,
        0x0c000000, 0x14000000, 0x18000000, 0x1c000000, 0x20000000, 0x24000000 };

const RCUBlock::TrailerMasks RCUBlock::V420MASKS = { 0xfc000000, 0x80000000, 0x84000000, 0x88000000,
        0x8c000000, 0x90000000, 0x94000000, 0x98000000, 0x9c000000, 0xe0000000 };

RCUBlock::TrailerMasks const * RCUBlock::Masks = 0;

const std::string RCUBlock::RCUCOLLECTIONNAME  = "RCU_Registers";
const std::string RCUBlock::DATACOLLECTIONNAME = "AltroRawData";

RCUBlock::RCUBlock(std::ifstream * const inputFile)
        : AltroBlock(inputFile)
{
    if (AltroBlock::daqversion == 0)
    {
        throw AltroBlockException("ERROR: RCUBlock initialized with no DAQ Version defined.");
    }

    if (Masks == 0)
    {
        if (AltroBlock::daqversion < 420)
        {
            Masks = &V410MASKS;
        }
        else
        {
            Masks = &V420MASKS;
        }
    }
}
RCUBlock::RCUBlock(std::ifstream* const inputFile, int start_address, uint32_t length)
        : AltroBlock(inputFile, start_address, length)
{
    if (AltroBlock::daqversion == 0)
    {
        throw AltroBlockException("ERROR: RCUBlock initialized with no DAQ Version defined.");
    }

    if (Masks == 0)
    {
        if (AltroBlock::daqversion < 420)
        {
            Masks = &V410MASKS;
        }
        else
        {
            Masks = &V420MASKS;
        }
    }
}

RCUBlock::RCUBlock(const AltroBlock & theMotherBlock)
        : AltroBlock(theMotherBlock)
{

    if (this->CheckIntegrity())
    {
        id = AltroBlock::BLOCK_RCU;
    }
    else
    {
        throw AltroBlockException(
                "The Altro Block copied in the RCU block doesn't appear to be a valid RCU block");
    }

    if (AltroBlock::daqversion == 0)
    {
        throw AltroBlockException("ERROR: RCUBlock initialized with no DAQ Version defined.");
    }

    if (Masks == 0)
    {
        if (AltroBlock::daqversion < 420)
        {
            Masks = &V410MASKS;
        }
        else
        {
            Masks = &V420MASKS;
        }
    }
}

RCUBlock::RCUBlock(const std::vector<uint32_t> & refContent)
        : AltroBlock(refContent)
{
    if (AltroBlock::daqversion == 0)
    {
        throw AltroBlockException("ERROR: RCUBlock initialized with no DAQ Version defined.");
    }

    if (Masks == 0)
    {
        if (AltroBlock::daqversion < 420)
        {
            Masks = &V410MASKS;
        }
        else
        {
            Masks = &V420MASKS;
        }
    }
}

RCUBlock::RCUBlock(const std::vector<uint32_t>::const_iterator refStart)
        : AltroBlock()
{
    if (AltroBlock::daqversion == 0)
    {
        throw AltroBlockException("ERROR: RCUBlock initialized with no DAQ Version defined.");
    }
    length = *refStart + 1;
    content.reserve(length);

    for (std::vector<uint32_t>::const_iterator it = refStart; it < refStart + length; it++)
    {
        content.push_back(*it);
    }
    if (Masks == 0)
    {
        if (AltroBlock::daqversion < 420)
        {
            Masks = &V410MASKS;
        }
        else
        {
            Masks = &V420MASKS;
        }
    }
}

void
RCUBlock::Decode()
{
    try
    {
        AltroBlock::Decode();
        StatusErrors = ((content[6] & 0x0ffff000) >> 12 );

        //Get the trailer length
        if ((*(content.end() - 1) & Masks->WORDCODEMASK) == Masks->RCUIDMASK)
        {
            trailer.TrailerLength = *(content.end() - 1) & 0x0000007f;
            trailer.RCUAddress = (*(content.end() - 1) & 0x0000ff80) >> 7;
            trailer.RCUFirmware = (*(content.end() - 1) & 0x03ff0000) >> 16;
        }
        else
        {
            std::ostringstream oss;
            oss << "ERROR: The RCU trailer's last word id code "
                    << (*(content.end() - 1) & (Masks->WORDCODEMASK))
                    << "doesn't match any of the supported code ids for the last trailer word.";
            std::string tmp = oss.str();
            throw AltroBlockException(tmp);
        }

        //Fill the trailer objects
        std::vector<uint32_t>::const_iterator it = (content.end() - trailer.TrailerLength);
        if ((*it & (Masks->WORDCODEMASK)) == Masks->PAYLOADMASK)
        {
            trailer.PayloadLength = *it & (~(Masks->WORDCODEMASK));
            it++;
        }
        else
            throw AltroBlockException("ERROR: unrecognized RCU Trailer Word ID");

        if ((*it & (Masks->WORDCODEMASK)) == Masks->ERRORREG1MASK)
        {
            trailer.ErrorReg1 = *it & (~(Masks->WORDCODEMASK));
            it++;
        }
        else
            throw AltroBlockException("ERROR: unrecognized RCU Trailer Word ID");

        if ((*it & (Masks->WORDCODEMASK)) == Masks->ERRORREG2MASK)
        {
            trailer.ErrorReg2 = *it & (~(Masks->WORDCODEMASK));
            it++;
        }
        else
            throw AltroBlockException("ERROR: unrecognized RCU Trailer Word ID");

        if ((*it & (Masks->WORDCODEMASK)) == Masks->ERRORREG3MASK)
        {
            trailer.ErrorReg3 = *it & (~(Masks->WORDCODEMASK));
            it++;
        }
        else
            throw AltroBlockException("ERROR: unrecognized RCU Trailer Word ID");

        if (AltroBlock::daqversion < 420)
        {
            it++;
        }

        if ((*it & (Masks->WORDCODEMASK)) == Masks->ACTFECAMASK)
        {
            trailer.ActFECA = *it & (~(Masks->WORDCODEMASK));
            it++;
        }
        else
            throw AltroBlockException("ERROR: unrecognized RCU Trailer Word ID");

        if ((*it & (Masks->WORDCODEMASK)) == Masks->ACTFECBMASK)
        {
            trailer.ActFECB = *it & (~(Masks->WORDCODEMASK));
            it++;
        }
        else
            throw AltroBlockException("ERROR: unrecognized RCU Trailer Word ID");

        if ((*it & (Masks->WORDCODEMASK)) == Masks->RDOCONFIG1MASK)
        {
            trailer.RDOConfig1 = *it & (~(Masks->WORDCODEMASK));
            it++;
        }
        else
            throw AltroBlockException("ERROR: unrecognized RCU Trailer Word ID");

        if ((*it & (Masks->WORDCODEMASK)) == Masks->RDOCONFIG2MASK)
        {
            trailer.RDOConfig2 = *it & (~(Masks->WORDCODEMASK));
            it++;
        }
        else
            throw AltroBlockException("ERROR: unrecognized RCU Trailer Word ID");

    }
    catch (std::ifstream::failure &f)
    {
        std::ostringstream err;
        err
                << "ERROR: Exception caught while decoding the RCU block. The file ended before it was supposed to"
                << std::endl;
        err << f.what() << std::endl;
        std::string tmp = err.str();
        throw AltroBlockException(tmp);
    }
    catch (std::exception &e)
    {
        std::ostringstream err;
        err << "ERROR: Exception caught while decoding the RCU block. " << std::endl;
        err << e.what() << std::endl;
        std::string tmp = err.str();
        throw AltroBlockException(tmp);
    }
}

std::string
RCUBlock::PrintSummary() const
{
    std::ostringstream oss;
    oss << std::dec << "RCU Block summary" << std::endl << std::endl;
    return oss.str();
}

bool
RCUBlock::CheckIntegrity() const
{
    if (AltroBlock::CheckIntegrity())
    {
        if (content.size() == 0)
        {
            inputFile->seekg(start + 2 * WORD32, std::ios::beg);
            return (ReadWord32() == AltroBlock::BLOCK_RCU);
        }
        else
        {
            return (content[2] == AltroBlock::BLOCK_RCU);
        }
    }
    return false;
}

void
RCUBlock::AddCollections(EVENT::LCEvent* lcevent) const
{
    //Create and then fill the collection containing the RCU registers
    lcio::LCCollectionVec* RCUParameterColl = 0;
    try
    {
        RCUParameterColl = dynamic_cast<lcio::LCCollectionVec*>(lcevent->getCollection(RCUCOLLECTIONNAME));
    }
    catch (EVENT::DataNotAvailableException &e)
    {
        RCUParameterColl = new lcio::LCCollectionVec(lcio::LCIO::LCINTVEC);
        lcevent->addCollection(RCUParameterColl, RCUCOLLECTIONNAME);
        RCUParameterColl->parameters().setValue("Parameter 1: ", "RCUID");
        RCUParameterColl->parameters().setValue("Parameter 2: ", "RCU Address");
        RCUParameterColl->parameters().setValue("Parameter 3: ", "RCU Firmware Version");
        RCUParameterColl->parameters().setValue("Parameter 4: ", "RCU Status and Error Register");
        RCUParameterColl->parameters().setValue("Parameter 5: ", "RCU Error Register 1");
        RCUParameterColl->parameters().setValue("Parameter 6: ", "RCU Error Register 2");
        RCUParameterColl->parameters().setValue("Parameter 7: ", "RCU Error Register 3");
        RCUParameterColl->parameters().setValue("Parameter 8: ", "RCU Active FEC List A");
        RCUParameterColl->parameters().setValue("Parameter 9: ", "RCU Active FEC List B");
        RCUParameterColl->parameters().setValue("Parameter 10: ", "RCU Readout Configuration 1");
        RCUParameterColl->parameters().setValue("Parameter 11: ", "RCU Readout Configuration 2");
    }
    AddRCUCollection(RCUParameterColl);
    int RCUNum = RCUParameterColl->getNumberOfElements() - 1;

    lcio::LCCollectionVec* altrocollection = 0;
    try
    {
        altrocollection = dynamic_cast<lcio::LCCollectionVec*>(lcevent->getCollection(DATACOLLECTIONNAME));
    }
    catch (EVENT::DataNotAvailableException &e)
    {
        altrocollection = new lcio::LCCollectionVec(lcio::LCIO::TRACKERRAWDATA);
        lcio::LCFlagImpl trkFlag(0);
        trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 );
        altrocollection->setFlag( trkFlag.getFlag() );
        lcevent->addCollection(altrocollection, DATACOLLECTIONNAME);
    }
    AddDataCollection(altrocollection, RCUNum);

}


//TODO Put the RCU data in the parameters of the Data collection together with the description using a pair in the struct
void
RCUBlock::AddRCUCollection(EVENT::LCCollection* RCUParameterColl) const
{
    EVENT::LCIntVec* RCUParameters = new EVENT::LCIntVec();
    RCUParameters->reserve(11);
    RCUParameters->push_back(RCUID);
    RCUParameters->push_back(trailer.RCUAddress);
    RCUParameters->push_back(trailer.RCUFirmware);
    RCUParameters->push_back(StatusErrors);
    RCUParameters->push_back(trailer.ErrorReg1);
    RCUParameters->push_back(trailer.ErrorReg2);
    RCUParameters->push_back(trailer.ErrorReg3);
    RCUParameters->push_back(trailer.ActFECA);
    RCUParameters->push_back(trailer.ActFECB);
    RCUParameters->push_back(trailer.RDOConfig1);
    RCUParameters->push_back(trailer.RDOConfig2);
    RCUParameterColl->addElement(RCUParameters);
}

void
RCUBlock::AddDataCollection(EVENT::LCCollection* altrocollection, int RCUNum) const
{
    if(trailer.PayloadLength !=0)
    {
        std::vector<uint32_t>::const_iterator dataStart = content.begin()+10;
        DataBlock ADCData = DataBlock(dataStart, trailer.PayloadLength);
        ADCData.AddDataCollection(altrocollection, RCUNum);
    }
}

}

