#ifndef HEP_REP_OUTPUT_PROCESSOR_H
#define HEP_REP_OUTPUT_PROCESSOR_H 1

#include <marlin/Processor.h>
#include <lcio.h>
#include <string>

/** HepRepOutputProcessor: This processor generates a HepRep Output file for a HepRep based event display program.
	* The processor extracts geometrical information of the detector in the init() function, so it can be used for geometry checks as well.
	* By means of parameters the user can decide which objects from the LCIO file shall be drawn.
	*
	*  \todo Color or pattern coding of the quality flags should be implemented for the pads.
	*
	*  <h4>Input - Prerequisites</h4>
	*  A valid GEAR file.
	*
	*  <h4>Output</h4>
	*  A HepRep xml file.
	*
	*  @param SimTrackerHitCollections list of the input SimTrackerHit collections.
	*  @param TrackerDataCollections list of the input SimTrackerHit collections.
	*  @param TrackerHitCollections list of the input SimTrackerHit collections.
	*  @param TrackerPulseCollections list of the input SimTrackerHit collections.
	*  @param TrackCollections list of the input SimTrackerHit collections.
	*  @param EventNumbersToDraw A space separated list of event numbers to be drawn.
	*  @param OutputFileName Name of the HepRep output file.
	*  @param PulseColorScheme Determines how the pulses are color coded:
            \li RAINBOW = 0, does rainbow colors between minimum and maximum charge.
			\li REDBLUE = 1, original color coding by J.Abernathy, gives a reddish value for charges > 0 and a blueish one for charges < 0
			\li TOPOGRAPHIC = 2, color coding as known from maps - blue for charges < 0 and green>yellow>orange>red for charges >0
			\li GREYSCALE = 3, grey between minimum (dark grey) and maximum charge (light grey)
			\li REDBLUESCALE = 4, \todo not yet implemented shall become a nicer implemetation of a red blue scale.
			\li COLORBRIGHTNESS = 5, \todo not yet implemented, shall vary between a given color and "black"
			\li COLORWHITE = 6, \todo not yet implemented, shall bleach a given color to "white"
	*  @param Color - \todo not yet implemented, will be needed for some future color schemes.

	*  @author	\li J. Abernathy (UVic)
				\li Oliver Sch&auml;fer, University of Rostock
	*  @version Revision $Rev$
 */
class HepRepOutputProcessor : public marlin::Processor
{

    public:

		/// The constructor. Parameters will be registered here.
        HepRepOutputProcessor();


        /// The Destructor.
        ~HepRepOutputProcessor();


        /// Return a new instance of this processor
        Processor * newProcessor();


		/// Processor initialisation. GEAR Geometry is drawn here.
        void init();


		/// RunHeader processing. Revision and parameter logging.
        void processRunHeader (lcio::LCRunHeader *run);


        /** Called for every event - the working horse.
        */
        virtual void processEvent (lcio::LCEvent *evt);


        /** Called for every event - right after processEvent() has been called for all processors.
        */
        virtual void check (lcio::LCEvent *evt);


        /** Called after data processing for clean up in the inverse order of the init() method so that resources allocated in the first processor also will be available for all following processors.
        */
        virtual void end();

    private:

        enum ColorScheme
        {
            RAINBOW = 0,
			REDBLUE = 1,
			TOPOGRAPHIC = 2,
			GREYSCALE = 3,
			REDBLUESCALE = 4,
			COLORBRIGHTNESS = 5,
			COLORWHITE = 6
        };

		int _pulseColorScheme;

		std::vector<std::string> _namesOfSimTrackerHitCollections;
		std::vector<std::string> _namesOfTrackerDataCollections;
		std::vector<std::string> _namesOfTrackerHitCollections;
		std::vector<std::string> _namesOfTrackerPulseCollections;
		std::vector<std::string> _namesOfTrackCollections;
		std::string _nameOfOutputFile;

		virtual void drawHepRepPad (int moduleID, int padIndex); ///<
		virtual void setHepRepPadColor (double padCharge, float minCharge, float maxCharge, int colorScheme, int pulseQuality); ///<

		// the HepRepXML writer
		HepRepXMLWriter _hepRepWriter;

		/** The event numbers to display **/
		std::vector<int> _eventsToDisplay;

		bool _drawPadPlane;
};

#endif // HEP_REP_OUTPUT_PROCESSOR_H



