#ifndef INTERSECTION_CALCULATOR_H
#define INTERSECTION_CALCULATOR_H

#include <exception>

namespace marlintpc
{

/** A helper class to calculate the intersections of
 *  \li two circles
 *  \li a circle with a straigt line
 *  \li two straight lines
 *  
 *  The idea is to calculate the intersection of a pad row with a track (straight or helix).
 *  The first part of the function name is the type of pad row (circle or horizontal line),
 *  the second part the type of track (circle or straight line). The pad plane is descibed by one local coordinate
 *  (radius \c R for circle and \c Y for rectangular pad rows, which is a horizontal line). The pad rows are 
 *  concentric around the origin of the local pad coordinate system, so no further parameter is needed. 
 *  The track circle is described by centre (x,y) and a radius. 
 *
 *  \attention The straight line is described by the inclination with respect to the y axis:<br>
 *  \f$ x = y\cdot inclination + xAxisIntersect\f$<br>
 *  This is done so tracks exactly perpendicular to the pad row can be described. Also for small
 *  inclinations with respect to the y axsis this is numerically more stable than \f$ y=mx+b\f$.
 *  For the large prototype the tracks are usually along the x axis so the parametrisation is done the other way.
 *  The decision which parametrisation is taken is done in circle_straightLine(double R, double x0, double y0, double trackPhi)
 *  based on the track angle.
 *
 *  In case there is no intersection, an IntersectionCalculator::NoIntersectionException is thrown.
 */

class IntersectionCalculator
{
 public:
  /** Calculate the intersection of a circular pad row with a track circle.
   *\param R The radius of the pad row.
   *\param centreX The x coordinate of the tracks helix centre.
   *\param centreY The y coordinate of the tracks helix centre.
   *\param r The radius of the track helix.
   *
   * Return value:
   * \li The two intersections, descibed by the two phi values on the pad row.
   * As the local coordinates are polar, (R, phi1) and (R, phi2) are the intersection points in 
     in local pad coordinates. The first argument of the pair is alwas the numerically smaller value.
   */
    static std::pair<double,double> circle_circle(double R, double centreX, double centreY, double r);

  /** Calculate the intersection of a circular pad row with a straight track parameterised in the 
      direction parallel to the y axis (x = inclination*y + xAxisIntersect).
   * \param R The radius of the pad row.
   * \param inclination Inclination of the track with respect to the y-axis
   * \param xAxisIntersect Intersection point of the track with the x-axis
   *
   * Return value:
   * \li The two intersections, each as an (x,y) pair (=local coorinates, converted to cartesian,
   * (r*cos(phi), r*sin(phi)) )
   * The first value has the smaller y coordinate.
   */
    static std::pair<std::pair<double,double>,std::pair<double,double> >  circle_straightLine_xy(double R, double inclination, double xAxisIntersect);

    
  /** Calculate the intersection of a circular pad row with a straight track.
      Depending on the angle of the track either circle_straightLine_inx_phi or
      circle_straightLine_iny_phi is called.

   * \param R The radius of the pad row.
   * \param x0,y0 A point on the track in local padplane coordinates
   * \param phi The angle of the track in local padplane coordinates
   *
   * Return value:
   * \li The two intersections, descibed by the two phi values on the pad row.
   * The first argument of the pair is alwas the numerically smaller value.
   */
    static std::pair<double,double>  circle_straightLine(double R, double x0, double y0, double trackPhi);

 
  /** Only kept for backwards compatibility. Use circle_straightLine instead!
      This function is now called circle_straightLine_inx_phi as a private member function
      accessable through circle_straightLine.
  */
    static std::pair<double,double>  circle_straightLine_phi(double R, double inclination, double xAxisIntersect);

    /** Calculate the intersection of a straight line (pad row) with a circle.
     *\param Y The y coordinate the pad row.
     *\param centreX The x coordinate of the tracks helix centre.
     *\param centreY The y coordinate of the tracks helix centre.
     *\param r The radius of the track helix.
     * 
     * Return value:
     * \li The two x coordinates where the circle intersects with the horizontal line
     */
    static std::pair<double,double>  horizonalLine_circle(double  Y,  double centreX, double centreY, double r);

    /** Calculate the intersection of a straight line (pad row) with a straight track parameterised
     * in the direction perpendicular to the pad rows (x = inclination*y + xAxisIntersect).
     *\param Y The y coordinate the pad row.
     * \param inclination Inclination of the track with respect to the y-axis
     * \param xAxisIntersect Intersection point of the track with the x-axis
     *
     * Return value:
     * \li The x coordinate of the intersection.
     */    
    static double horizontalLine_straightLine(double  Y,  double inclination, double xAxisIntersect);

    /// A nested exception which is thrown in case there is no intersection
    class NoIntersectionException : public std::exception 
    {
	protected:
	    std::string message ;
	    
	    NoIntersectionException(){  /*no_op*/ ; } 
	    
	public: 
	    virtual ~NoIntersectionException() throw() { /*no_op*/; } 
	    
	    NoIntersectionException( const std::string& text ){
		message = "NoIntersectionExceptionException: " + text ;
	    }
 
	    virtual const char* what() const  throw() { return  message.c_str() ; } 
    };

 private:
    /// Private constructor: No need to generate an instance, all member functions are static
    IntersectionCalculator(){}

    /** Calculate the intersection of a circular pad row with a straight track parameterised in the 
      direction parallel to the y axis (x = inclination*y + xAxisIntersect).
   * \param R The radius of the pad row.
   * \param inclination Inclination of the track with respect to the y-axis
   * \param xAxisIntersect Intersection point of the track with the x-axis
   *
   * Return value:
   * \li The two intersections, descibed by the two phi values on the pad row.
   * The first argument of the pair is alwas the numerically smaller value.
   */
    static std::pair<double,double>  circle_straightLine_inx_phi(double R, double inclination, double xAxisIntersect);

  /** Calculate the intersection of a circular pad row with a straight track parameterised in the 
      direction parallel to the x axis (y = inclination*x + yAxisIntersect).
   * \param R The radius of the pad row.
   * \param inclination Inclination of the track with respect to the y-axis
   * \param xAxisIntersect Intersection point of the track with the x-axis
   *
   * Return value:
   * \li The two intersections, descibed by the two phi values on the pad row.
   * The first argument of the pair is alwas the numerically smaller value.
   */
    static std::pair<double,double>  circle_straightLine_iny_phi(double R, double inclination, double yAxisIntersect);

};

} //namespace marlintpc
#endif //INTERSECTION_CALCULATOR_H
