#include <iostream>
#include <stdlib.h>     /* atoi */

#include <TCanvas.h>
#include <TROOT.h>
#include <TApplication.h>
#include <TAxis.h>
#include <TGraph.h>
#include <TFile.h>

#include "MediumMagboltz.hh"
#include "SolidBox.hh"
#include "GeometrySimple.hh"
#include "ComponentConstant.hh"
#include "Sensor.hh"
#include "TrackHeed.hh"
#include "Plotting.hh"

using namespace Garfield;

int main(int argc, char * argv[]) {

  // switch between different gas mixtures:
  // 0=P5, 1=TDR, 2=T2K, ...
  int gasChoice;
  if (2==argc)
    gasChoice=atoi(argv[1]);
  else
    gasChoice=0;

  // Make a medium
  MediumMagboltz* gas = new MediumMagboltz();

  std::string gasName, gasComponents; 

  // add your gas mixture here
  switch(gasChoice){
    case 0: {  gasName="P5"; 
      gasComponents="(Ar:CH4 95:5)"; 
      gas->SetComposition("ar", 95., "methane",5.); 
      break; }
    case 1: {  gasName="TDR"; 
        gasComponents="(Ar:CH4:CO2 95:3:2)"; 
        gas->SetComposition("ar", 95., "methane", 3., "co2",2.);  
        break; }
    case 2: {  gasName="T2K"; 
      gasComponents="(Ar:CF4:iCH10 95:3:2)"; 
      gas->SetComposition("ar", 95., "tetrafluoromethane", 3., "isobutane",2.);   // T2K
      break; }
    case 3: {  gasName="testTF"; 
      gasComponents="(Ar:CF4 95:5)"; 
      gas->SetComposition("ar", 95., "tetrafluoromethane", 5.);
      break; }
    case 4: {  gasName="testIso"; 
      gasComponents="(Ar:iCH10 95:5)"; 
      gas->SetComposition("ar", 95., "isobutane", 5.);
      break; }
    default: {
      std::cout << "No gas chosen. Using plain argon." << std:: endl;
      gasName="Ar"; 
      gasComponents="(Ar 100)"; 
      gas->SetComposition("ar", 100.); 
    }
  }

  gas->SetTemperature(293.15);
  gas->SetPressure(760.);


  // Detector geometry
  // Gap [cm]
  const double width = 0.585;
  SolidBox* box = new SolidBox(width / 2., 0., 0., width / 2.,10., 10.);
  GeometrySimple* geo = new GeometrySimple();
  geo->AddSolid(box, gas);

  // Make a component
  ComponentConstant* comp = new ComponentConstant();
  comp->SetGeometry(geo);
  comp->SetElectricField(240., 0., 0.);
  //   comp->SetMagneticField(1., 0., 0.);

  // Make a sensor
  Sensor* sensor = new Sensor();
  sensor->AddComponent(comp);

  // Initial track position and direction 
  double x0 = 0., y0 = 0., z0 = 0., t0 = 0.;
  double dx0 = 1., dy0 = 0., dz0 = 0.; 

  // Track class
  TrackHeed* track = new TrackHeed();
  track->SetSensor(sensor);
  track->SetParticle("electron");
  // track->DisableDeltaElectronTransport();

 


  // loop get cluster density for different momenta 
  const double electronMass = 0.510998910e6;
  double momentum;
  double m_over_p;

  // trying to find a loop iteration that uses nicely spaced values in log-x plots
  unsigned int ipoint=0, nbase=4, nminor=100;

  TGraph* grClusterDensity = new TGraph(nbase*nminor + 1);
  grClusterDensity->SetNameTitle("grClusterDensity", 
				 Form("%s %s;momentum/mass;number of cluster per cm", gasName.c_str(), gasComponents.c_str()));
  grClusterDensity->SetMarkerStyle(22);

  for (unsigned int ibase=0; ibase<nbase; ibase++){
    for (unsigned int iminor=0; iminor<nminor; iminor++){
      m_over_p=pow(10,ibase+(iminor/(1.0*nminor)));
      momentum=electronMass*m_over_p;
      track->SetMomentum(momentum);
      track->NewTrack(x0, y0, z0, t0, dx0, dy0, dz0);
      if (0==iminor){ 
	std::cout << "for momentum/mass=" << m_over_p 
		  << ": cluster density " << track->GetClusterDensity() << " cm-1" << std::endl;
      }
      grClusterDensity->SetPoint(ipoint, m_over_p, track->GetClusterDensity());
      ipoint++;
    }
  }
  m_over_p=pow(10,nbase); // and one more, to end at a nice x-value
  momentum=electronMass*m_over_p;
  track->SetMomentum(momentum);
  track->NewTrack(x0, y0, z0, t0, dx0, dy0, dz0);
  std::cout << "for momentum/mass=" << m_over_p 
	    << ": cluster density " << track->GetClusterDensity() << " cm-1" << std::endl;
  grClusterDensity->SetPoint(ipoint, m_over_p, track->GetClusterDensity());


  // draw and save
  grClusterDensity->SaveAs(Form("ClusterDensity240VcmDelta_%s.root", gasName.c_str()));

  TCanvas *cCD= new TCanvas("cCD", Form("cluster density %s", gasName.c_str()), 900, 600);
  cCD->SetLogx();
  grClusterDensity->Draw("AP");
  grClusterDensity->GetYaxis()->SetRangeUser(0, grClusterDensity->GetYaxis()->GetXmax());
  cCD->SaveAs(Form("ClusterDensity240VcmDelta_%s.png", gasName.c_str()));

  return 0;
}

