/*
 * TPCFieldSettings2LCCD.cc
 *
 *  Created on: Oct 29, 2012
 *      Author: amunnich
 */

/**
 * This tool is to be used to insert TPC FieldSettings data in the LCCD database
 * or in an slcio file for offline use.
 * It will take its input from a text file listing the validity time and the values of
 * the condition data to insert.
 * This tool is specific for the TPCFieldSettings object as defined in revision 2814
 *
 * USAGE:
 * To write data to a lcio simple file:
 * TPCFieldSettings2LCCD.exe -S outfile conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]
 *
 * To write data to a database:
 * TPCFieldSettings2LCCD.exe -D connectionString CondDBfolder conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]
 *
 * FORMATS:
 * Text file:
 * validSinceDate validSinceTime validTillDate validTillTime NGEMS gemVoltages transferFields inductionField anodeVoltage 
 * cathodeVoltage guardRingVoltage Bfield gemStatus DriftFieldStatus MagneticFieldStatus
 * Date format: YYYY-MM-DD
 * Time Format: HH:MM:SS where the number of seconds can be a decimal value
 * Connection String: hostname:database:login:password[:port]
 * ConDBfolder: /database/folder
 */

/*
 * TODO Add -h option to print the usage information
 */

//standard inclusions
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <stdexcept>
#include <cstdlib>
#define NDEBUG
#include "assert.h"
#include <getopt.h>

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"

//ILCUTIL
#include "streamlog/streamlog.h"

//LCIO
#include "lcio.h"
#include "Exceptions.h"
#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"
#include "IMPL/LCTOOLS.h"

//MarlinTPC
#include "FieldSetting.h"

/*
 * Structure encapsulating the TPCFieldSetting object and the LCTime object to separate the phase of
 * file reading and parsing from the writing phase
 */

struct TPCFieldrecord
{
    tpcconddata::FieldSetting* theConditions;
    lcio::LCTime validSince;
    lcio::LCTime validTill;
};

/*
 * Prints the usage informations
 */
void printUsage()
{
    std::cout<<"Usage:"<<std::endl;
    std::cout<<"To write data to a lcio simple file:"<<std::endl;
    std::cout<<"  TPCFieldSetting2LCCD.exe -S outfile \\"<<std::endl;
    std::cout<<"  conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]"<<std::endl;
    std::cout<<"To write data to a database:"<<std::endl;
    std::cout<<"  TPCFieldSetting2LCCD.exe -D connectionString folder \\"<<std::endl;
    std::cout<<"  conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]"<<std::endl;
    std::cout<<"Formats:"<<std::endl;
    std::cout<<" Text file: validSinceDate validSinceTime validTillDate validTillTime NGEMS gemVoltages transferFields inductionField anodeVoltage cathodeVoltage guardRindVoltage Bfield gemStatus DriftFieldStatus MagneticFieldStatus" << std::endl;
    std::cout<<" Date format: YYYY-MM-DD" << std::endl;
    std::cout<<" Time Format: HH:MM:SS where the number of seconds can be a decimal value" << std::endl;
    std::cout<<" Connection String: hostname:database:login:password[:port]"<<std::endl;
    std::cout<<" folder: /database/folder"<<std::endl;
    std::cout << "If you want to add a certain time before and after the run in the data base "
          << "give the option to be sure that the condition will be found use -e time[s]." << std::endl;
}

//Function that converts the date and time strings from the input file in an LCTime object from which to obtain the timestamp
lcio::LCTime convert2LCTime(std::string Date, std::string Time)
{
    //Variable delcaration
    std::vector<std::string> tokens;
    int Year;
    int Month;
    int Day;
    int Hour;
    int Minute;
    int Second;
    EVENT::long64 NanoSecond;
    lcio::LCTime timestamp;

    //TODO More meaningful, if any format checking for date and time
    //Date format: YYYY-MM-DD
    if(Date.size() != 10)
    {
        std::string error = std::string("The date string is not correctly formatted");
        std::invalid_argument e(error);
        throw  e;
    }
    std::stringstream datestream(Date);
    std::string token;
    while(std::getline(datestream, token, '-'))
    {
        tokens.push_back(token);
    }

    //TODO More meaningful, if any format checking for date and time
    //Time Format: HH:MM:SS[.ddd]
    if(Time.size() < 8)
    {
        std::string error = std::string("The time string is not correctly formatted");
        std::invalid_argument e(error);
        throw  e;
    }

    std::stringstream timestream(Time);
    while(std::getline(timestream, token, ':'))
    {
        tokens.push_back(token);
    }

    //Create a new stream using the strings that I got from the parsing then using the stream I/O
    //mechanism to feed the data in the variables using the internal operator(<<) for teh conversion
    std::stringstream convStream;
    double tempSec;
    convStream << tokens[0] << " " << tokens[1] << " " << tokens[2] << " " << tokens[3] << " " << tokens[4] << " " << tokens[5];
    convStream >> Year >> Month >> Day >> Hour >> Minute >> tempSec;
    //Retrieve the integral and the decimal part of the variable indicating the seconds
    Second = static_cast<int>(tempSec);
    NanoSecond = static_cast<EVENT::long64>((tempSec - static_cast<EVENT::long64>(tempSec)) * 1E9);

    timestamp=lcio::LCTime(Year,Month,Day,Hour,Minute,Second);
    timestamp+= NanoSecond;
    return timestamp;
}

//Main program
int main(int argc, char **argv)
{
  streamlog::out.init(std::cout, "TPCConditions2LCCD");
  streamlog::out.addLevelName<streamlog::MESSAGE>();
  streamlog::out.addLevelName<streamlog::DEBUG>();
  streamlog::logscope scope(streamlog::out);
  scope.setLevel("MESSAGE");
  //Program variable declaration
  //Program steering options
  bool writeSimpleFile=false;
  bool write2DB=false;
  std::string dbInitString;
  std::string dbFolder;
  std::string outFileName;
  std::vector<std::string> inFileNames;
  int extentRun = -1; // add this number of seconds before and after the run

  // Other variables
  lccd::DBInterface* dbConnection = 0;
  lcio::LCCollectionVec* conditionsColl = 0;
  const std::string COLLNAME = std::string("FieldSetting");
  lcio::LCWriter* lcWrt = 0;
  lcio::LCEventImpl*  evt = 0;
  std::vector<TPCFieldrecord> recordVector; //Vector containing the elements to fill in the db or file

  // Get options
  // Interpreting the command line
  static struct option long_options[] =
  {
      { "outputFile", required_argument, 0, 'S' },
      { "extendRun", required_argument, 0, 'e' },
      { "dbInit", required_argument, 0, 'D' },
      { "dbFolder", required_argument, 0, 'F'},
  };

  int OptionId;
  int OptionIndex;
  while ((OptionId = getopt_long(argc, argv, "S:e:D:F:", long_options,
      &OptionIndex)) != -1) {
    switch (OptionId) {
    case 'S': {
      writeSimpleFile=true;
      outFileName = std::string(optarg);
      streamlog_out(MESSAGE) << "Output file name is set to: " << outFileName
          << "." << std::endl;
      try
      {
          streamlog_out(MESSAGE) << "Creating the event writer to write the file" << std::endl;
          lcWrt = lcio::LCFactory::getInstance()->createLCWriter() ;    //Create LCWriter
          lcWrt->open( outFileName, lcio::LCIO::WRITE_NEW ) ;
          evt = new lcio::LCEventImpl() ;
      }
      catch(lcio::Exception &e)
      {
        streamlog_out(ERROR) << "TPCConditions2LCCD: ERROR while creating the LCWriter and the event objects" <<std::endl;
        streamlog_out(ERROR) << e.what() << std::endl;
        if(lcWrt != 0)
        {
            lcWrt->close();
            delete lcWrt;
            lcWrt = 0;
        }
        if(evt != 0)
        {
            delete evt;
            evt = 0;
        }
        return 1;
      }assert(lcWrt != 0);
      break;
    }
    case 'e': {
      extentRun = atoi(optarg);
      streamlog_out(MESSAGE) << "Will add " << extentRun
          << " seconds before and after the run." << std::endl;
      break;
    }
    case 'D': {
      write2DB=true;
      dbInitString = std::string(optarg);
      streamlog_out(MESSAGE) << "Got the following dbinit string: "
          << dbInitString << "." << std::endl;
      break;
    }
    case 'F': {
      dbFolder = std::string(optarg);
      streamlog_out(MESSAGE) << "Got the following db Folder: "
          << dbFolder << "." << std::endl;
      break;
    }
    case '?': {
      printUsage();
      return 0;
    }
    }
  }
  // read the remaining arguments, which should be the conditions files
  for (int i = optind; i < argc; i++) {
    streamlog_out(MESSAGE) << "Adding conditions file: " << argv[i] << std::endl;
    inFileNames.push_back(argv[i]);
  }
  if(inFileNames.size() == 0){
    printUsage();
    if(lcWrt != 0)
    {
        lcWrt->close();
        delete lcWrt;
        lcWrt = 0;
    }
    if(evt != 0)
    {
        delete evt;
        evt = 0;
    }
    return(1);
  }
  if(write2DB){
    try
    {
        dbConnection = new lccd::DBInterface(dbInitString, dbFolder, write2DB);
    }
    catch(std::exception &e)
    {
      streamlog_out(ERROR) << "ERROR while creating the DB connection" << std::endl;
      streamlog_out(ERROR) << e.what() << std::endl;
      if(dbConnection != 0)
      {
          delete dbConnection;
          dbConnection = 0;
      }
      return 1;
    } assert(dbConnection != 0);

  }

  //loop over the input files
  //TODO Add a try...catch clause around the file reading and parsing procedure
  for (std::vector<std::string>::iterator inFile = inFileNames.begin(); inFile != inFileNames.end(); ++inFile)
  {
      //Open the input file and connects it to a read buffer
      std::filebuf infile;
      infile.open(inFile->c_str(),std::ios::in);
      std::istream inputstream(&infile);
      streamlog_out(MESSAGE) << " Reading file #" << (inFile - inFileNames.begin() + 1) <<" : "<< *inFile << std::endl;
      int linenumber = 0; //<- Spy variables for debugging info

      //Read all the lines of the file until the end
      while(!inputstream.eof())
      {
          //Take a line from the stream and feeds it to a string
          std::string line;
          line.clear();
          std::getline(inputstream, line);

          // check if the next line is a comment or if it's an empty line
          // If the line is valid associate it to a stream so we can use the >> operator
          // to do the string conversions into the variables
          if (*line.begin() != '#' && line.size() != 0)
          {
              //Local variables creation
              std::stringstream linestream(line);
              tpcconddata::FieldSetting* theConditions = 0;
              lcio::LCTime validSince;
              lcio::LCTime validTill;
              std::string validSinceDate;
              std::string validSinceTime;
              std::string validTillDate;
              std::string validTillTime;
  int NGEMs=0;
  std::vector<float> GemVoltages;
  GemVoltages.clear();
  std::vector<float> TransferFields;
  TransferFields.clear();
  float InductionField=0;
  float AnodeVoltage=0;
  float CathodeVoltage=0;
  float GuardRingVoltage=0;
  float BField=0;
  int GEMStatus=0;
  int DriftFieldStatus=0;
  int MagneticFieldStatus=0;

 

              linestream >> validSinceDate >> validSinceTime >> validTillDate >> validTillTime
       >> NGEMs ;
  for(int g=0;g<NGEMs;g++)
    {
      float temp=0;
      linestream >> temp;
      GemVoltages.push_back(temp);
    }
  for(int t=0;t<NGEMs-1;t++)
    {
      float temp=0;
      linestream >> temp;
      TransferFields.push_back(temp);
    }

  linestream >> InductionField >> AnodeVoltage >> CathodeVoltage >> GuardRingVoltage
       >> BField  >> GEMStatus >> DriftFieldStatus >> MagneticFieldStatus;

              // test if everything went fine
              if (linestream.fail())
              {
                  streamlog_out(MESSAGE) << "TPCFieldSetting2LCCD: ERROR reading input file at line " << linenumber << " : " <<line << std::endl;

                  if(dbConnection != 0)
                  {
                      delete dbConnection;
                      dbConnection = 0;
                  }
                  if(lcWrt != 0)
                  {
                      lcWrt->close();
                      delete lcWrt;
                      lcWrt = 0;
                  }
                  if(evt != 0)
                  {
                      delete evt;
                      evt = 0;
                  }
                  return 1;
              }

              //Create TPC Condition Object
              try
              {
                  //Create the condition object
                  theConditions = new tpcconddata::FieldSetting(NGEMs);
      theConditions->setGEMStatus( GEMStatus );
      theConditions->setDriftFieldStatus( DriftFieldStatus );
      theConditions->setMagneticFieldStatus( MagneticFieldStatus );
      for(int g=0;g<NGEMs;g++)
        theConditions->setGEMVoltage( g, GemVoltages[g] );
      for(int t=0;t<NGEMs-1;t++)
        theConditions->setTransferField( t, TransferFields[t] );
      theConditions->setInductionField( InductionField );
      theConditions->setCathodeVoltage( CathodeVoltage );
      theConditions->setAnodeVoltage( AnodeVoltage );
      theConditions->setGuardRingVoltage( GuardRingVoltage );
      theConditions->setMagneticField( BField );

                  validSince = convert2LCTime(validSinceDate, validSinceTime);
                  validTill = convert2LCTime(validTillDate, validTillTime);
                  streamlog_out(MESSAGE) << "Valid From: " << validSince.timeStamp() << " to " << validTill.timeStamp() << std::endl;
                  TPCFieldrecord thisRecord = {theConditions, validSince, validTill};
                  recordVector.push_back(thisRecord);
              }
              catch(std::exception &e)
              {
                  streamlog_out(ERROR) << "TPCConditions2LCCD: ERROR while creating the LCIO objects" <<std::endl;
                  streamlog_out(ERROR) << e.what() << std::endl;
                  if(dbConnection != 0)
                  {
                      delete dbConnection;
                      dbConnection = 0;
                  }
                  if(lcWrt != 0)
                  {
                      lcWrt->close();
                      delete lcWrt;
                      lcWrt = 0;
                  }
                  if(evt != 0)
                  {
                      delete evt;
                      evt = 0;
                  }
                  if(theConditions != 0)
                  {
                      delete theConditions;
                      theConditions = 0;
                  }
                  if(recordVector.size() != 0)//Cleanup the vector
                  {
                      for(std::vector<TPCFieldrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
                      {
                          delete it->theConditions;
                      }
                  }
                  return 1;
              }assert(theConditions != 0); assert(recordVector.size() != 0);
          }
          else
          {
              streamlog_out(DEBUG) << "Skipping comment or empty line " << linenumber << " : "  << line <<  std::endl;
          }
          linenumber++;
      }
      //Close the input file
      infile.close();
  }

  //If write to file I create a single collection containing multiple TPC Condition objects
  if(writeSimpleFile)
  try
  {
      //Create collection
      conditionsColl = new lcio::LCCollectionVec(lcio::LCIO::LCGENERICOBJECT);
      //Fill the collection
      for(std::vector<TPCFieldrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
      {
          conditionsColl->addElement(it->theConditions);
          it->theConditions = 0;
      }
      //Add the collection to the event
      evt->addCollection(conditionsColl,COLLNAME);
      //Dump the event to screen
      lcio::LCTOOLS::dumpEvent( evt ) ;
      // write event to file
      lcWrt->writeEvent( evt ) ;
  }
  catch(lcio::Exception & e)
  {
      streamlog_out(ERROR) << "TPCConditions2LCCD: ERROR while creating the LCIO collection" <<std::endl;
      streamlog_out(ERROR) << e.what() << std::endl;
      if(lcWrt != 0)
      {
          delete lcWrt;
          lcWrt = 0;
      }
      if(evt != 0)
      {
          delete evt;
          evt = 0;
      }
      if(recordVector.size() != 0)//Cleanup the vector
      {
          for(std::vector<TPCFieldrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
          {
              if(it->theConditions != 0) delete it->theConditions;
          }
      }
      return 1;
  }
  else if (write2DB)
  {
      try
      {
          //Create one collection containing one element for each vector element and store it between the given timestamps
          for(std::vector<TPCFieldrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
          {
              //TODO Add an output for each line written to the DB
              conditionsColl = new lcio::LCCollectionVec(lcio::LCIO::LCGENERICOBJECT);
              conditionsColl->addElement(it->theConditions);
              if(extentRun != -1){
                it->validSince += -1 * extentRun * 1000000000;
                it->validTill += extentRun * 1000000000;
              }
              dbConnection->storeCollection(it->validSince.timeStamp(), it->validTill.timeStamp(), conditionsColl, COLLNAME);
          }
      }
      catch(std::exception &e)
      {
          streamlog_out(ERROR) << "TPCConditions2LCCD: ERROR while storing the condition object in the DB" <<std::endl;
          streamlog_out(ERROR) << e.what() << std::endl;
          if(dbConnection != 0)
          {
              delete dbConnection;
              dbConnection = 0;
          }
          if(recordVector.size() != 0)//Cleanup the vector
          {
              for(std::vector<TPCFieldrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
              {
                  if(it->theConditions != 0) delete it->theConditions;
              }
          }
          if(conditionsColl != 0)
          {
              delete conditionsColl;
              conditionsColl = 0;
          }
          return 1;
      }
  }

  //Close DB and clean up
  if(dbConnection != 0)
  {
      delete dbConnection;
      dbConnection = 0;
  }
  if(lcWrt != 0)
  {
      lcWrt->close();
  }
  if(evt != 0)
  {
      delete evt;
  }

  return 0;
}//End main


