#ifndef TimePixSimpleAlternativeReaderProcessor_h
#define TimePixSimpleAlternativeReaderProcessor_h 1

#include "marlin/DataSourceProcessor.h"
#include <fstream>

namespace marlin{

/** Reads TimePix data stored in format x y value
 *  Writes data to TrackerRawDataCollection named TimePixRawData
 *  Does NOT make use of the zero suppressed input files, so output data is NOT zero suppressed!
 *  Currently only works for one chip
 *
 *
 *  <h4>Input - Prerequisites</h4>
 *  TimePix data files (format: x y value).
 *
 *  <h4>Output</h4>
 *  \li A runheader
 *  \li Events with TimePixRawData (lcio::TrackerRawData)
 *  \li First event contains TimePixModeMap collection
 *
 * @param FileListName   Text file with list of input TimePix files
 * \attention{TimePix file names must be absolute pathes!}
 * @param OutputCollectionName Name of the output collection, defaut \c TimePixRawData
 * @param DetectorName Detector name for run header, default "TPC"
 * @param Description  Data descripton for run header, default "no description given"
 * @param SubDetectorName Sub-detector name for run header, default "TimePix"
 * @param ChipID Interer ID number of the chip (lcio::TrackerRawData::CellID1),
 *   should be 0 as it's only one chip in the current version
 * @param NRow Number of rows on the chip, default 256
 * @param NColumn Number of columns on the chip, default 256
 * @param MaxRecordNumber Maximum number of events 100000
 * @param SetOutputTransient If not 0 the output collection is set transient
 * @param RunNumber Run number for run header, default 0
 *
 * @author Simone Zimmermann (University of Bonn), Christoph Krieger (University of Bonn)
 */

class TimePixSimpleAlternativeReaderProcessor : public DataSourceProcessor {

public:

	TimePixSimpleAlternativeReaderProcessor() ;

	virtual TimePixSimpleAlternativeReaderProcessor*  newProcessor() ;


	/** Creates events with TrackerRawData collections from the TimePix raw data input files and
	 *  calls all active processors' processEvent() and processRunHeader Method.
	 *
	 */
	virtual void readDataSource( int numEvents ) ;


	virtual void init() ;
	virtual void end() ;


protected:

	std::string _outputColName;

	bool _isFirstEvent;
	std::string _listName ;
	std::ifstream _infile;
	int _chipID;
	int _runNr;
	int _evtNr;
	int _nRow;
	int _nColumn;
	int _maxRecordN;
	std::string _detName;
	std::string _dataDescription;
	std::string _subDetName;

	/** give the status which is set to the transient flag of the output collection
	 * it is an int instead of a boolean, because the processor parameter cannot be a boolean
	 */
	int _outputIsTransient;

};


} // end namespace marlin 

#endif
