#include "TPCHitToTrackerDataConverterProcessor.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <EVENT/TPCHit.h>
#include <IMPL/TPCHitImpl.h>

// Aida
#ifdef MARLIN_USE_AIDA
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogramFactory.h>
#include <AIDA/ICloud1D.h>
#endif

using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{

    TPCHitToTrackerDataConverterProcessor aTPCHitToTrackerDataConverterProcessor ;

    TPCHitToTrackerDataConverterProcessor::TPCHitToTrackerDataConverterProcessor() : Processor("TPCHitToTrackerDataConverterProcessor") {
        /* the constructor call first the super constructor which the name
         * of the processor
         * this must by the same as the class name.
         */

        /* description of the processor which will be displayed in the
         * steering file autometicly generated by
         * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
         */
        _description = "TPCHitToTrackerDataConverterProcessor converts data in the (deprecated) TPCHit format to TrackerData. By setting the parameter IgnoreFirstNBins to a non-zero integer value N, the first N bins will be disregarded." ;

        /* register steering parameters: name, description, class-variable, default value
         * the type will by definded by the type of the default value
         * string, double, float and int are possible
         * plaese note to get a float you must cast in C++: (float) 1.3
         */
        registerProcessorParameter( "InputCollectionName" ,
            "Name of the input collection"  ,
            _inputColName ,
            std::string("TPCFADC") ) ;

        registerProcessorParameter( "OutputCollectionName" ,
            "Name of the output collection"  ,
            _outputColName ,
            std::string("TPCConvertedRawData") ) ;

        registerProcessorParameter( "SetOutputTransient" ,
            "if not 0 the output collection is set transient"  ,
            _outputIsTransient,
            1);

        registerProcessorParameter( "IgnoreFirstNBins",
            "Number of bins at the beginning of the pulse that should not be stored",
            _ignoreFirstNBins,
            0);

    }

    void TPCHitToTrackerDataConverterProcessor::init() {
        // usually a good idea to
        printParameters() ;
        _nRun = 0 ;
        _nEvt = 0 ;
    }

    void TPCHitToTrackerDataConverterProcessor::processRunHeader(lcio::LCRunHeader* run) {
        run->parameters().setValue( _processorName + "_revision", "$Rev: 5472 $" );
        for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
            if ( ! i->second->isOptional() || i->second->valueSet() ) {
                run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
            }
        }
        _nRun++ ;
    }

    void TPCHitToTrackerDataConverterProcessor::processEvent(lcio::LCEvent* evt ) {
        // this gets called for every event
        _nEvt++;
        if (_nEvt%100 == 0) {
            std::cout<<_nEvt<<"\r";
            std::cout<<std::flush;
        }

        //read input data
        LCCollectionVec* InputCollection = dynamic_cast<LCCollectionVec*> (evt->getCollection( _inputColName ) ) ;
        LCCollectionVec* OutputCollection =  new LCCollectionVec(LCIO::TRACKERDATA);

        if( InputCollection != 0 ) {
            // input collection will not be saved in outputfile
            InputCollection->setTransient( true );
            int nPulse = InputCollection->getNumberOfElements()  ;
            int nSlice = 0;
            for(int iPulse=0; iPulse< nPulse ; iPulse++) {
                TPCHit* oldPulse = dynamic_cast<TPCHit*>( InputCollection->getElementAt( iPulse ) ) ;
                if (oldPulse) {
                    TrackerDataImpl* newPulse = new TrackerDataImpl();
                    nSlice   = oldPulse->getNRawDataWords();
                    newPulse->setTime( oldPulse->getRawDataWord(0) );
                    newPulse->setCellID0( oldPulse->getCellID() );
                    newPulse->setCellID1( 0 );
                    //newPulse->adcValues().resize( nSlice-1 );
                    for (int iSlice=_ignoreFirstNBins; iSlice< nSlice; iSlice++) {
                        // start with _ignoreFirstNBins, to get rid of unwanted bins at the beginning of the pulse
                        newPulse->chargeValues().push_back(oldPulse->getRawDataWord( iSlice ));
                    }
                    OutputCollection->addElement( newPulse );
                } 
            }  
            evt->addCollection( OutputCollection, _outputColName );
        } //end if( InputCollection != 0 )
    }

    void TPCHitToTrackerDataConverterProcessor::check(lcio::LCEvent *  /*evt*/ ) {
        // nothing to check here - could be used to fill checkplots in reconstruction processor
    }

    void TPCHitToTrackerDataConverterProcessor::end() {
        //   std::cout << "TPCHitRawDataToDataConverter::end()  " << name()
        //         << " processed " << _nEvt << " events in " << _nRun << " runs "
        //         << std::endl ;
    }

}
