#include "TestFieldSettingProcessor.h"
#include <iostream>
#include <string>
#include <math.h>
// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerData.h>

//marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>
//#include <Exceptions.h>

//lccd
#include <lccd_exceptions.h>

//gear
#include <gear/TPCParameters.h>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>

// AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/ITupleFactory.h>
#include <AIDA/ITuple.h>

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace gear;

namespace marlintpc {

TestFieldSettingProcessor aTestFieldSettingProcessor;

TestFieldSettingProcessor::TestFieldSettingProcessor() :
    Processor("TestFieldSettingProcessor"), _fieldSettingListener(NULL) {
  // the processor description
  _description =
      "TestFieldSettingProcessor is intended to test whether the data base has the field settings stored for a certain time."
          "These data can be stored in an root file (if an AIDA processor is active) or they are just printed. If TPCPulses are given also drift velocity is calculated."
          "Therefore it might be useful to specify a certain time window.";

  /* register steering parameters: name, description, class-variable, default value
   */

  registerInputCollection(LCIO::TRACKERRAWDATA, "InputCollectionRawData",
      "Name of the input collection", _inputColNameRawData,
      std::string("AltroRawData"));

  registerInputCollection(LCIO::TRACKERPULSE, "InputCollectionPulseData",
      "Name of the input collection. If this collection is available also the drift velocity is calculated. Therefore also set maxDriftLength",
      _inputColNamePules, std::string("TPCPulses"));

  registerOptionalParameter("InputTPCConditions",
      "Optional: Name of the input collection containing the TPC conditions data",
      _inputFieldSettingCollectionName,
      tpcconddata::FieldSetting::getDefaultColName());
  registerOptionalParameter("maxDriftLength",
      "Optional: The expected drift length for the signal in the lcio file (needed to calculate the drift velocity and only necessary if TPCPulses are available [cm].",
      _maxDriftLength, (double) 56.76);
  registerOptionalParameter("PulseTimeLowerLimit",
      "Optional: The lower time limit for TPCPulses to be considered for the calculation of the drift time [ns].",
      _startTime, (double) 0);
  registerOptionalParameter("PulseTimeUpperLimit",
      "Optional: The upper time limit for TPCPulses to be considered for the calculation of the drift time [ns].",
      _endTime, (double) 50000);
  registerProcessorParameter("TPCLengthOverride",
      "Give the length of the maximum drift distance in the TPC to calculate the drift field from the cathode/anode voltage [cm]. Otherwise it is taken from the gearfile: <maxDriftLength value= .../>.",
      _driftLength, (double) 56.76);
  registerProcessorParameter("PulseChargeCut",
      "Optional: Set a charge cut for the TPCPulses to be considered for the calculation of the drift time.",
      _pulseCharge, (double) -1);
}

void TestFieldSettingProcessor::init() {
  streamlog_out(DEBUG)<< " init() called  " << endl;

  // you might want to list the used parameters by invoking
  printParameters();
  try
  {
    _fieldSettingListener = new FieldSettingListener(_inputFieldSettingCollectionName);
  }

  catch(lccd::LCCDException &)
  {
    streamlog_out(ERROR) << "Collection named \"" << _inputFieldSettingCollectionName << "\" with FieldSetting is not available." << std::endl;
    throw StopProcessingException(this);;
  }

  _fieldData = 0;
  _driftVelocity = 0;
  _currentSettings = 0;
  _tmpPulseTime = 0;
  try
  {
    _fieldData = AIDAProcessor::tupleFactory(this)->create("FieldData", "Field data retrieved from the data base",
        "int run, int event,"
        "int driftFieldStatus, int magneticFieldStatus, int  nGEMs,"
        "float gem_voltage1, float gem_voltage2, float gem_voltage3,"
        "float transferField1, float transferField2, float inductionField, float driftField,"
        "float anodeVoltage, float cathodeVoltage, float guardRingVoltage,"
        "float magneticField", "");
  }
  catch(lcio::Exception &e)
  {
    streamlog_out(MESSAGE) << "No AIDA processor available. The field data won't be stored in a root file." << std::endl;
  }
  try
  {
    _driftVelocity = AIDAProcessor::tupleFactory(this)->create("driftVelocity", "Calculated drift velocity per condition.",
        "double driftVelocity, double driftVelocityRMS, double driftField", "");
    AIDAProcessor::tree(this)->mkdir("pulseData");
  }
  catch(lcio::Exception &e)
  {
    streamlog_out(MESSAGE) << "No AIDA processor available. The drift velocity data won't be stored in a root file." << std::endl;
  }
  if(!parameterSet("TPCLengthOverride")){
    const TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
    _driftLength = tpc_parameters.getMaxDriftLength () / 10.;
    streamlog_out(MESSAGE) << "Drift length is taken from the gear file and set to: " << _driftLength << std::endl;
  }
}

void TestFieldSettingProcessor::processRunHeader(LCRunHeader*  /*run*/) {

}

void TestFieldSettingProcessor::processEvent(LCEvent* evt) {
  LCCollection* input_pulses = 0;
  try {
    input_pulses = evt->getCollection(_inputColNamePules);
  } catch (lcio::DataNotAvailableException &e) {
    streamlog_out(DEBUG)<< "No Pulse data available." << std::endl;
  }
  try{
    // fill data if conditions changed and create histogram if pulses are available
    if (_currentSettings != _fieldSettingListener->getFieldSetting()) {
      // if histogram exist calculate drift velocity and fill it into the tuple
      if (_tmpPulseTime != 0 && _driftVelocity != 0) {
        double v = _driftLength * 1000. / _tmpPulseTime->mean();
        double dv = _driftLength * _tmpPulseTime->rms() * 1000. / pow(_tmpPulseTime->mean(),2);
        _driftVelocity->fill(0, v); // cm/mus
        _driftVelocity->fill(1, dv); // cm/mus
        _driftVelocity->fill(2, _lastDriftField);
        _driftVelocity->addRow();
      }
      // only create histogram if pulses are available
      if (input_pulses != 0 && _driftVelocity != 0) {
        std::stringstream sstm;
        sstm << "pulseData/PulseTimeHistogram_" << evt->getRunNumber() << "_" << evt->getEventNumber();
        try {
          _tmpPulseTime =
              AIDAProcessor::histogramFactory(this)->createHistogram1D(
                  sstm.str().c_str(), "Pulse time", 1000, _startTime, _endTime);
        } catch (lcio::Exception &e) {
          streamlog_out(DEBUG)<< "No AIDA processor available. The field data won't be stored in a root file." << std::endl;
        }
      }
      _currentSettings = _fieldSettingListener->getFieldSetting();
      lcio::LCTime time(evt->getTimeStamp());
      streamlog_out(MESSAGE)<< "Run: " << evt->getRunNumber() << " \t Event: " << evt->getEventNumber() <<
          "\t Start: " << time.getDateString().c_str() << std::endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getAnodeVoltage=" << _fieldSettingListener->getAnodeVoltage() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getCathodeVoltage=" << _fieldSettingListener->getCathodeVoltage() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getDriftFieldStatus=" << _fieldSettingListener->getDriftFieldStatus() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getFieldSetting=" << _fieldSettingListener->getFieldSetting() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getGEMVoltage(0)=" << _fieldSettingListener->getGEMVoltage(0) <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getGEMVoltage(1)=" << _fieldSettingListener->getGEMVoltage(1) <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getGEMVoltage(2=" << _fieldSettingListener->getGEMVoltage(2) <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getGuardRingVoltage()=" << _fieldSettingListener->getGuardRingVoltage() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getInductionField()=" << _fieldSettingListener->getInductionField() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getMagneticField()=" << _fieldSettingListener->getMagneticField() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getMagneticFieldStatus()=" << _fieldSettingListener->getMagneticFieldStatus() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getNGEMs()=" << _fieldSettingListener->getNGEMs() <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getTransferField(1)=" << _fieldSettingListener->getTransferField(1) <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getTransferField(0)=" << _fieldSettingListener->getTransferField(0) <<endl;
      streamlog_out(MESSAGE)<< "_fieldSettingListener->getDriftField(" << _driftLength << ")=" << _fieldSettingListener->getDriftField(_driftLength) <<endl;
      _lastDriftField = _fieldSettingListener->getDriftField(_driftLength);
      if (_fieldData != 0) {
        _fieldData->fill(0, evt->getRunNumber());
        _fieldData->fill(1, evt->getEventNumber());
        _fieldData->fill(2, _fieldSettingListener->getDriftFieldStatus());
        _fieldData->fill(3, _fieldSettingListener->getMagneticFieldStatus());
        _fieldData->fill(4, _fieldSettingListener->getNGEMs());
        _fieldData->fill(5, _fieldSettingListener->getGEMVoltage(0));
        _fieldData->fill(6, _fieldSettingListener->getGEMVoltage(1));
        _fieldData->fill(7, _fieldSettingListener->getGEMVoltage(2));
        _fieldData->fill(8, _fieldSettingListener->getTransferField(0));
        _fieldData->fill(9, _fieldSettingListener->getTransferField(1));
        _fieldData->fill(10, _fieldSettingListener->getInductionField());
        _fieldData->fill(11, _fieldSettingListener->getDriftField(_driftLength));
        _fieldData->fill(12, _fieldSettingListener->getAnodeVoltage());
        _fieldData->fill(13, _fieldSettingListener->getCathodeVoltage());
        _fieldData->fill(14, _fieldSettingListener->getGuardRingVoltage());
        _fieldData->fill(15, _fieldSettingListener->getMagneticField());
        _fieldData->addRow();
      }
    }
  } catch (lccd::LCCDException &e){
    streamlog_out(WARNING) << "Failed to load conditions data in run " << evt->getRunNumber() <<  " event " << evt->getEventNumber() << endl;
    streamlog_out(WARNING) << e.what() << endl;
    return;
  } catch (...){
    streamlog_out(ERROR) << "There was an unknown exception in run " << evt->getRunNumber() <<  " event " << evt->getEventNumber() << endl;
    return;
  }

  if (input_pulses != 0 && _tmpPulseTime != 0) {
    for (int i_tracker_pulse = 0;
        i_tracker_pulse < input_pulses->getNumberOfElements();
        i_tracker_pulse++) {
      TrackerPulse* input_pulse =
          dynamic_cast<TrackerPulse*>(input_pulses->getElementAt(
              i_tracker_pulse));
      if (!input_pulse) {
        throw(lcio::Exception(
            "Input collection " + _inputColNamePules
                + " does not contain TrackerPulses."));
      }
      if(_pulseCharge < 0 || input_pulse->getCharge() > _pulseCharge)
        _tmpPulseTime->fill(input_pulse->getTime());
    }
  }
}

void TestFieldSettingProcessor::check(LCEvent*  /*evt*/) {

}

void TestFieldSettingProcessor::end() {
  if (_tmpPulseTime != 0 && _driftVelocity != 0) {
    /***
     * All pulses are filled in _tmpPulseTime - the time cut is used to set the range
     * of the histogram and the mean is only calculated in the range of the histogram
     * not considering over- and underflow entries.
     ***/
    double v = _driftLength * 1000. / _tmpPulseTime->mean();
    double dv = _driftLength * _tmpPulseTime->rms() * 1000. / pow(_tmpPulseTime->mean(),2);
    _driftVelocity->fill(0, v); // cm/mus
    _driftVelocity->fill(1, dv); // cm/mus
    _driftVelocity->fill(2, _lastDriftField);
    _driftVelocity->addRow();
  }
}
} // close namespace brackets
