#include "TestTPCDistortionsMapProcessor.h"
#include "TPCDistortionMeasurement.h"
#include "MarlinTPCExceptions.hh"

#include <iostream>
#include <string>
#include <vector>

// GEAR
#include <gear/TPCParameters.h>

// lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <EVENT/TrackerData.h>
#include <Exceptions.h>

//marlin
#include <marlin/Exceptions.h>
#include <marlin/Global.h>

// root
#include "TCanvas.h"
#include "TArrow.h"
#include "TSystem.h"
#include "TVector2.h"
#include "TApplication.h"
#include "TMath.h"
#include "TROOT.h"
#include "TH2F.h"

using namespace lcio;
using namespace marlin;
using namespace std;

namespace marlintpc
{

TestTPCDistortionsMapProcessor aTestTPCDistortionsMapProcessor;


TestTPCDistortionsMapProcessor::TestTPCDistortionsMapProcessor() : Processor("TestTPCDistortionsMapProcessor"),
    _draw(true), _scale(1.), _cartesian(false), _showModule(3), _moduleInterpolation(false)
{
  // the processor description
  _description = "TestTPCDistortionsMapProcessor is intended to check if TPCDistortionMeasurement data can be retrieved from the data base. If yes a simple plot of the distortions is produced.";
  
  registerOptionalParameter("InputTPCConditions",
        "Optional: Name of the input collection containing the TPC conditions data",
        _inputDistortionsCollectionName,
        tpcconddata::TPCDistortionMeasurement::getDefaultColName());
  registerOptionalParameter("ScaleArrows",
      "Optional: Scale the distortion arrows by a certain factor.",
      _scale,
      (double) 1.);
  registerOptionalParameter("UseCartesianCoordinates",
      "Optional: Show the arrows in global x-y coordinate system (default: r-phi)",
      _cartesian,
      (bool) false);
  registerOptionalParameter("ShowModule",
      "Optional: A close up of a certain module is shown. Per default it is the central module (3). Set another here if you want.",
      _showModule,
      (int) 3);
  registerOptionalParameter("DoModuleInterpolation",
      "Optional: If false only dots from one module are used."
      "Thus at the module boundary no shift can be calculated, since for doing this dots from neighboring modules are needed."
      "If true the interpolation is done, but wrong results might appear at module boundaries in case of local field distortions there!",
      _moduleInterpolation,
      (bool) false);

}


void TestTPCDistortionsMapProcessor::init()
{ 
  streamlog_out(DEBUG) << " init() called  " << endl;
  
  // you might want to list the used parameters by invoking
  printParameters();

  try
  {
    _distortionsListener = new TPCDistortionsMapListener(_inputDistortionsCollectionName);
  }  catch(lccd::LCCDException &)
  {
    streamlog_out(ERROR) << "Collection named \"" << _inputDistortionsCollectionName << "\" with Distortions is not available." << endl;
    throw StopProcessingException(this);
  }
  return;
}


void TestTPCDistortionsMapProcessor::processRunHeader(LCRunHeader*  /*run*/)
{
  // do something for each run
}


void TestTPCDistortionsMapProcessor::processEvent(LCEvent*  /*evt*/)
{
  if (!_draw) {
    return;
  }

  vector<tpcconddata::TPCDistortionMeasurement*> distortionsData;
  try {
    distortionsData = _distortionsListener->getTPCDistortionsMap();
  } catch (lcio::DataNotAvailableException &e) {
    streamlog_out(DEBUG)<< "No distortions data available." << endl;
  }
  // need to create a ROOT application to draw a canvas, it does not need to be run
  TApplication* app = new TApplication("app", new int(0), new char* );
  app->SetReturnFromRun( kFALSE );
  if (gROOT->IsBatch()) {
    gROOT->SetBatch( kFALSE );
  }
  TCanvas* c1 = new TCanvas( "Field distortions", "Field distortions", 1400, 700 );
  TCanvas* c2 = new TCanvas( "Field distortions central module", "Field distortions on the central module in local coordinates", 1400, 700 );

  // Steering of the arrows
  TArrow* arr = new TArrow();
  arr->SetArrowSize(0.02);
  arr->SetAngle(10);
  arr->SetOption(">");
  arr->SetLineWidth(2);
  arr->SetLineColor(kRed);

  // get the global plane extend
  const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
  vector<double> extend = tpc_parameters.getPlaneExtent();
  TVector2 min;
  TVector2 max;
  if(tpc_parameters.getCoordinateType() == gear::PadRowLayout2D::CARTESIAN){
    min.Set(extend.at(0), extend.at(2));
    max.Set(extend.at(1), extend.at(3));
  } else if(tpc_parameters.getCoordinateType() == gear::PadRowLayout2D::POLAR){
    min.SetMagPhi(extend.at(0),extend.at(2));
    max.SetMagPhi(extend.at(1), extend.at(3));
  }
  // create a histogram with the global plain extend to put the arrows in later on and to have axes
  TH2F* h_c0, *h_c1;
  if(_cartesian){
    h_c0 = new TH2F("hist_global_c0","Distortions map",100,min.Y(),max.Y(),100,min.X(), max.X());
    h_c1 = new TH2F("hist_global_c1","Distortions map",100,min.Y(),max.Y(),100,min.X(), max.X());
    streamlog_out(DEBUG) << "Set histogram boundaries to: x -> " << min.Y() << " - " << max.Y()
        << "\t y -> " << min.X() << " - " << max.X() << endl;
    h_c0->SetTitle("Distortions in c0 direction in the global coordinate system;y [mm];x [mm];distortion along c_{0}");
    h_c1->SetTitle("Distortions in c1 direction in the global coordinate system;y [mm];x [mm];distortion along c_{1}");
  } else {
    h_c0 = new TH2F("hist_global_c0","Distortions map",100,TVector2::Phi_mpi_pi(min.Phi())*TMath::RadToDeg(),TVector2::Phi_mpi_pi(max.Phi())*TMath::RadToDeg(),100,min.Mod(),max.Mod());
    h_c1 = new TH2F("hist_global_c1","Distortions map",100,TVector2::Phi_mpi_pi(min.Phi())*TMath::RadToDeg(),TVector2::Phi_mpi_pi(max.Phi())*TMath::RadToDeg(),100,min.Mod(),max.Mod());
    streamlog_out(DEBUG) << "Set histogram boundaries to: x -> " << TVector2::Phi_mpi_pi(min.Phi())*TMath::RadToDeg() << " - " << TVector2::Phi_mpi_pi(max.Phi())*TMath::RadToDeg()
            << "\t y -> " << min.Mod() << " - " << max.Mod() << endl;
    h_c0->SetTitle("Distortions in c_{0} direction in the global coordinate system;#phi [deg];r [mm];distortion along c_{0}");
    h_c1->SetTitle("Distortions in c_{1} direction in the global coordinate system;#phi [deg];r [mm];distortion along c_{1}");
  }
  vector<double> shift_values[2];
  for(int bin_x = 1; bin_x <= h_c0->GetNbinsX(); bin_x++){
    for(int bin_y = 1; bin_y <= h_c0->GetNbinsY(); bin_y++){
      try{
        gear::Vector2D tmp_shift = _distortionsListener->getEstimatedShift2D(CLHEP::Hep3Vector(h_c0->GetYaxis()->GetBinCenter(bin_y),h_c0->GetXaxis()->GetBinCenter(bin_x),0), _moduleInterpolation);
        h_c0->SetBinContent(bin_x, bin_y, tmp_shift[0]);
        h_c1->SetBinContent(bin_x, bin_y, tmp_shift[1]);
        shift_values[0].push_back(tmp_shift[0]);
        shift_values[1].push_back(tmp_shift[1]);
      } catch (InvalidDataException &e){
        h_c0->SetBinContent(bin_x, bin_y,-10);
        h_c1->SetBinContent(bin_x, bin_y,-10);
      }
    }
  }
  for(int i = 0; i < 2; i++){
    sort(shift_values[i].begin(),shift_values[i].end());
    shift_values[i].insert(shift_values[i].begin(),(*shift_values[i].begin())-TMath::Abs((*shift_values[i].begin())*0.1));
    shift_values[i].push_back((*(shift_values[i].end()-1))+TMath::Abs((*(shift_values[i].end()-1))*0.1));
    streamlog_out(MESSAGE) << "Found local distortion c_" << i << " ranging from " << (*shift_values[i].begin()) << " - " << *(shift_values[i].end()-1) << endl;
    streamlog_out(MESSAGE) << "Set range for local distortion c_" << i << " to " << (*shift_values[i].begin()) << " - " << *(shift_values[i].end()-1) << endl;
  }

  c1->Divide(2,1);
  c1->cd(1);
  h_c0->Draw("colz");
  h_c0->GetZaxis()->SetRangeUser((*shift_values[0].begin()),*(shift_values[0].end()-1));
  c1->cd(2);
  h_c1->Draw("colz");
  h_c1->GetZaxis()->SetRangeUser((*shift_values[1].begin()),*(shift_values[1].end()-1));

  // create histogram with local coordinates of the central module
  const gear::TPCModule& local_module = tpc_parameters.getModule(_showModule);
  vector<double> v_local_module_extend = local_module.getLocalModuleExtent(); // x_min, x_max, y_min, y_max
  gear::Vector2D local_module_extend[2] = {gear::Vector2D(v_local_module_extend.at(0),v_local_module_extend.at(2)), gear::Vector2D(v_local_module_extend.at(1),v_local_module_extend.at(3))};
  TH2F* h_local_c0 = new TH2F("hist_local_c0","Distortions map",100,local_module_extend[0][1],local_module_extend[1][1],100,local_module_extend[0][0], local_module_extend[1][0]);
  TH2F* h_local_c1 = new TH2F("hist_local_c1","Distortions map",100,local_module_extend[0][1],local_module_extend[1][1],100,local_module_extend[0][0], local_module_extend[1][0]);
  h_local_c0->SetTitle("Distortions in c_{0} direction for the central module in local coordinates;c_{1};c_{0};distortion along c_{0}");
  h_local_c1->SetTitle("Distortions in c_{1} direction for the central module in local coordinates;c_{1};c_{0};distortion along c_{1}");
  for(int bin_x = 1; bin_x <= h_local_c0->GetNbinsX(); bin_x++){
    for(int bin_y = 1; bin_y <= h_local_c0->GetNbinsY(); bin_y++){
      try{
        gear::Vector2D global_coordinates = local_module.localToGlobal(h_local_c0->GetYaxis()->GetBinCenter(bin_y),h_local_c0->GetXaxis()->GetBinCenter(bin_x));
        streamlog_out(DEBUG) << "bins: " << bin_x << ":" << bin_y << endl;
        gear::Vector2D tmp_shift = _distortionsListener->getEstimatedShift2D(CLHEP::Hep3Vector(global_coordinates[0],global_coordinates[1],0));
        h_local_c0->SetBinContent(bin_x, bin_y, tmp_shift[0]);
        h_local_c1->SetBinContent(bin_x, bin_y, tmp_shift[1]);
      } catch (InvalidDataException &e){
        h_local_c0->SetBinContent(bin_x, bin_y,-10);
        h_local_c1->SetBinContent(bin_x, bin_y,-10);
      }
    }
  }
  c2->Divide(2,1);
  c2->cd(1);
  h_local_c0->Draw("colz");
  h_local_c0->GetZaxis()->SetRangeUser((*shift_values[0].begin()),*(shift_values[0].end()-1));
  c2->cd(2);
  h_local_c1->Draw("colz");
  h_local_c1->GetZaxis()->SetRangeUser((*shift_values[1].begin()),*(shift_values[1].end()-1));

  streamlog_out(MESSAGE) << "Found information for " << distortionsData.size() << " photodots." << endl;
  for(vector<tpcconddata::TPCDistortionMeasurement*>::iterator it = distortionsData.begin();
        it != distortionsData.end(); it++){
    streamlog_out(DEBUG) << "Found true position: " << (*it)->getDoubleVal(0) << "," << (*it)->getDoubleVal(1) << endl;
    streamlog_out(DEBUG) << "Found reconstructed position: " << (*it)->getDoubleVal(3) << "," << (*it)->getDoubleVal(4) << endl;
    TVector2 photodot((*it)->getDoubleVal(0),(*it)->getDoubleVal(1));
    TVector2 data((*it)->getDoubleVal(3),(*it)->getDoubleVal(4));
    TVector2 delta(data-photodot);
    gear::Vector2D local[2];
    const gear::TPCModule& module = tpc_parameters.getNearestModule((*it)->getDoubleVal(3),(*it)->getDoubleVal(4));
    local[0] = module.globalToLocal((*it)->getDoubleVal(0),(*it)->getDoubleVal(1));
    local[1] = module.globalToLocal((*it)->getDoubleVal(3),(*it)->getDoubleVal(4));
    if(module.getLocalPadLayout().getCoordinateType() == gear::PadRowLayout2D::POLAR && (local[0][1] > TMath::Pi()/2.)){
      local[0][1] -= TMath::Pi()*2;
    }
    TVector2 photodot_local(local[0][1],local[0][0]);
    TVector2 data_local(local[1][1],local[1][0]);
    TVector2 delta_local(data_local-photodot_local);
    for( int pad = 1; pad < 3; pad++){
      c1->cd(pad);
      if(_cartesian){
        arr->DrawArrow(photodot.Y(),photodot.X(),(photodot+_scale*delta).Y(),(photodot+_scale*delta).X());
      } else {
        arr->DrawArrow(TVector2::Phi_mpi_pi(photodot.Phi())*TMath::RadToDeg(),photodot.Mod(),TVector2::Phi_mpi_pi((photodot+_scale*delta).Phi())*TMath::RadToDeg(),(photodot+_scale*delta).Mod());
      }
      if(module.getModuleID() == _showModule){
        c2->cd(pad);
        arr->DrawArrow(photodot_local.X(),photodot_local.Y(),(photodot_local+_scale*delta_local).X(),(photodot_local+_scale*delta_local).Y());
      }
    }
  }


  // need some wait loop to keep canvas open
  streamlog_out(MESSAGE) << "Press return to continue ..." << endl;
  int flag = fcntl(1, F_GETFL, 0);
  int key = 0;
  while(true) {
    gSystem->ProcessEvents();
    fcntl(1, F_SETFL, flag | O_NONBLOCK);
    key = getchar();
      if((key == '\n') || (key == '\r'))
        break;
      usleep(1000);
  }
  fcntl(1, F_SETFL, flag);

  // draw only once
  _draw = false;

  /* could also use this:
   *   throw marlin::StopProcessingException(this);
   * but better let the user set the total number of events to be processed since maybe he wants other processors to go on...
   */
}


void TestTPCDistortionsMapProcessor::check(LCEvent*  /*evt*/)
{ 
  // check what is needed here 
}


void TestTPCDistortionsMapProcessor::end()
{ 
  // do something at the end of the processing, e.g. cleaning up
}
} // close namespace brackets
