#include "TimePixXYReaderProcessor.h"
#include "TimePixPixelMode.h"

// Marlin
#include "marlin/ProcessorMgr.h"

// LCIO
#include <IO/LCWriter.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCFlagImpl.h>
#include <UTIL/LCTOOLS.h>


namespace marlin{



TimePixXYReaderProcessor aTimePixXYReaderProcessor ;


TimePixXYReaderProcessor::TimePixXYReaderProcessor() : DataSourceProcessor("TimePixXYReaderProcessor"),
		_chipIDs(1,0),
		_evtNr(0),
		_activatedPixels(0),
		_nPixels(256),
		_nBoardsPerFec(4),
		_nChipsPerBoard(8),
		_isFirstEvent(true){


	_description = std::string("Reads TimePix raw data files as input and creates LCIO events with TrackerRawData collections.")
	+ std::string(" Make sure to not specify any LCIOInputFiles in the steering in order to read TimePix raw data files.") ;


	registerProcessorParameter( "FileListName" ,
			"Input file with list of TimePix files (default: TimePixData.txt)"  ,
			_listName ,
			std::string("TimePixData.txt") ) ;

	registerOutputCollection( LCIO::TRACKERRAWDATA,
			"OutputCollectionName" ,
			"Name of the output collection (default: TimePixRawData)"  ,
			_outputColName,
			std::string("TimePixRawData") ) ;

	registerProcessorParameter( "NChipRows" ,
			"Number of chip rows (default: 1) "  ,
			_nChipRow,
			int(1) ) ;

	registerProcessorParameter( "NChipColumns" ,
			"Number of chip columns (default: 1)"  ,
			_nChipColumn,
			int(1) );

	registerProcessorParameter( "SetOutputTransient" ,
			"If true the output collection is set transient and not added to the file output (default: false)"  ,
			_outputIsTransient,
			bool(false) );

	registerProcessorParameter( "MinPixels",
			"Minimum number of activated pixels in one event. Events with less pixels are skipped (default: 10)",
			_minPixels,
			int(10) );

	registerProcessorParameter( "ReadTimeStamps",
			"If true timestamps are extracted from Timepix frames filenames instead from file itself and stored in the corresponding events (default: false)",
			_enableTimeStamps,
			bool(false) );

	registerProcessorParameter( "LengthOfTimestamp",
			"Number of digits of timestamp (default: 9)",
			_lengthOfTimestamp,
			int(9) );

	registerProcessorParameter( "CharsAfterTimestamp",
			"Number of characters following timestamp in Timepix frame file name (default: 4)",
			_charsAfterTimestamp,
			int(4) );

	registerProcessorParameter("NoChipHeaders",
			"Set to true if frames have no header lines separating data of different chips (default: false)",
			_noChipHeaders,
			bool(false));

	registerProcessorParameter("MaxPixelsPerChip",
			"Maximum number of pixels read out per chip, depends on readout system (default: 4096)",
			_maxPixelsPerChip,
			int(4096));

	registerOptionalParameter( "RunNumber",
			"Set RunNumber (default: 0)",
			_runNr,
			int(0) );

	registerOptionalParameter( "Description" ,
			"Description of data (default: No description given)"  ,
			_dataDescription ,
			std::string( "No description given" ) ) ;

	registerOptionalParameter( "DetectorName" ,
			"Name of the detector (default: TPC)"  ,
			_detName ,
			std::string("TPC") );

	registerOptionalParameter( "SubDetectorName" ,
			"Name of the subdetector (default: TimePix)"  ,
			_subDetName ,
			std::string("TimePix") );

	registerOptionalParameter( "ChipIDs" ,
			"ID number(s) of the chip(s) (default: 0) "  ,
			_chipIDs,
			_chipIDs ) ;

} // end of C'tor



TimePixXYReaderProcessor*  TimePixXYReaderProcessor::newProcessor() {

	return new TimePixXYReaderProcessor ;
}



void TimePixXYReaderProcessor::init() {


	// Print parameters to screen
	printParameters() ;


	// Try to read in the file containing the data file list.
	try{

		_inFile.open( _listName.c_str() );

	}catch (IOException& e ){

		streamlog_out(ERROR) << e.what() <<std::endl;
	}


	if( !_inFile.good() ){

		_inFile.close();
		streamlog_out(ERROR) << "!" << _listName << ".good()" << std::endl;
	}

	if( !parameterSet( "ChipIDs" ) ){

		const unsigned nChips = _nChipColumn * _nChipRow;
		_chipIDs.resize( nChips );

		for( unsigned iChip = 0; iChip < nChips; ++iChip ){

			_chipIDs[iChip] = iChip;
		}

	} // end of if( !parameterSet( "ChipIDs" ) )


} // end of void TimePixXYReaderProcessor::init()





void TimePixXYReaderProcessor::readDataSource( int /*numEvents*/ ) {


	while(_inFile.good() && !_inFile.eof() )
	{

		char dataFile[7200];
		_inFile >> dataFile;

		const bool lineIsEmpty = (dataFile[0] == '\0');
		if(lineIsEmpty)
		{
			continue; //skip empty lines
		}

		// Create new event
		LCEventImpl*  evt = new LCEventImpl() ;
		evt->setRunNumber(  _runNr   );
		evt->setEventNumber( _evtNr );
		evt->setDetectorName( _detName );

		//extract timestamp from filename if option is set and from file properties itself if not
		long int timestamp;

		if(_enableTimeStamps)
		{
			std::string filename(dataFile);
			filename.erase(filename.length()-_charsAfterTimestamp,_charsAfterTimestamp);
			filename.erase(0,filename.length()-_lengthOfTimestamp);

			timestamp = std::strtol(filename.c_str(),NULL,10);
		}
		else
		{
			struct stat st;
			int error = stat(dataFile,&st);

			if(error == 0){timestamp = st.st_mtime;}
			else{timestamp = 0;}
		}

		evt->setTimeStamp((long64)timestamp);

		streamlog_out(DEBUG4) << "Timestamp of event " << _evtNr << ": " << evt->getTimeStamp() << std::endl;

		// Create collection from data file
		IMPL::LCCollectionVec* outCol = this->readXYFile( dataFile );

		if( outCol == NULL || (_activatedPixels < static_cast<unsigned>(_minPixels)) )
		{

			delete outCol;
			delete evt;
			continue;
		}

		evt->addCollection( outCol , _outputColName ) ;

		if( _isFirstEvent )
		{

			LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ;

			runHdr->setRunNumber( _runNr );
			runHdr->setDetectorName( _detName );
			runHdr->setDescription( _dataDescription );
			runHdr->addActiveSubdetector( _subDetName );

			// Log revision
			runHdr->parameters().setValue( _processorName + "_revision", "$Rev: 5472 $" );

			// Log parameters
			for( ProcParamMap::iterator iPara = _map.begin(); iPara != _map.end(); ++iPara )
			{

				if( ! iPara->second->isOptional() || iPara->second->valueSet() )
				{

					runHdr->parameters().setValue( _processorName + "_" + iPara->second->name(), iPara->second->value() );
				}
			}

			ProcessorMgr::instance()->processRunHeader( runHdr ) ;
			delete runHdr;

			_isFirstEvent = false;

		} // end of if( _isFirstEvent )


		if( streamlog::out.write< streamlog::DEBUG2 >() )
		{
			LCTOOLS::dumpEvent( evt ) ;
		}

		ProcessorMgr::instance()->processEvent(evt) ;

		delete evt ;

		++_evtNr;

	} // end of while(true)

} // end of readDataSource( int numEvents )





IMPL::LCCollectionVec* TimePixXYReaderProcessor::readXYFile( const std::string& fileName ){

	_activatedPixels = 0;

	std::ifstream dataFileStream;
	dataFileStream.open( fileName.c_str() );

	if( !dataFileStream.good() ){

		streamlog_out(ERROR) << "!" << fileName << ".good()" << std::endl;
		dataFileStream.close();

		return NULL;
	}

	IMPL::LCCollectionVec* trackerRawDataCol = new IMPL::LCCollectionVec( LCIO::TRACKERRAWDATA );

	// Set the flags to store cellID1
	lcio::LCFlagImpl trkFlag(0) ;
	trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
	trackerRawDataCol->setFlag( trkFlag.getFlag() );

	// Set the transient flag as defined by the processor parameter
	trackerRawDataCol->setTransient( _outputIsTransient );

	int fecID(-1);
	int boardID(-1);
	int chipID(-1);
	int globalChipID(-1);

	//at start pixelsRead should be bigger than pixelsOnChip so that the header is read if available
	int pixelsOnChip(-1);
	int pixelsRead(0);

	//special treatment for the first pixel read in as there is no previous one available obviously
	bool isFirstPixel(true);

	TrackerRawDataImpl* rawData(NULL);
	EVENT::ShortVec* adcValues(NULL);

	int prevX(-1);
	int prevY(-1);
	int prevGlobalChipID(-1);

	while(dataFileStream.good() && !dataFileStream.eof())
	{
		//read header if all hits of previous chip have been read
		if( (pixelsRead >= pixelsOnChip) && !_noChipHeaders )
		{
			//peek first character of line to see what part of header it is
			char peek = dataFileStream.peek();

			switch(peek)
			{
			case 'F': //header line containing fecID looks like FEC <fecID> \n
				dataFileStream.ignore(256,' ');
				dataFileStream >> fecID;
				streamlog_out(DEBUG0) << "FEC: " << fecID << std::endl;
				dataFileStream.ignore(256,'\n');
				break;
			case 'B': //header line containing boardID looks like Board <boardID> \n
				dataFileStream.ignore(256,' ');
				dataFileStream >> boardID;
				streamlog_out(DEBUG0) << "Board: " << boardID << std::endl;
				dataFileStream.ignore(256,'\n');
				break;
			case 'C': //header line containing chipID looks like Chip <chipID>, Hits: <pixelsOnChip> FPGA <fpgaFlag> \n
				dataFileStream.ignore(256,' ');
				dataFileStream >> chipID;
				chipID--; //chipID in TOS starts from 1 while in MarlinTPC it starts from 0
				streamlog_out(DEBUG0) << "Chip: " << chipID << std::endl;
				dataFileStream.ignore(256,':');
				dataFileStream >> pixelsOnChip;
				streamlog_out(DEBUG0) << "Hits: " << pixelsOnChip << std::endl;
				if(pixelsOnChip > _maxPixelsPerChip){pixelsOnChip = _maxPixelsPerChip;} //only _maxPixelsPerChip at maximum are written to file
				dataFileStream.ignore(256,'\n'); //fpgaFlag is not read in at the moment as there is now convenient way to store it in the data collection
				pixelsRead = 0; //reset number of pixels read
				break;
			default: //takes care of old header line version of format <chipID> <pixelsOnChip>\n
				fecID = 0; //for old headers there is always only one fec
				boardID = 0; //and one board
				dataFileStream >> chipID;
				chipID--; //chipID in TOS starts from 1 while in MarlinTPC it starts from 0
				streamlog_out(DEBUG0) << "Chip: " << chipID << std::endl;
				dataFileStream >> pixelsOnChip;
				streamlog_out(DEBUG0) << "Hits: " << pixelsOnChip << std::endl;
				if(pixelsOnChip > _maxPixelsPerChip){pixelsOnChip = _maxPixelsPerChip;} //only _maxPixelsPerChip at maximum are written to file
				dataFileStream.ignore(256,'\n');
				pixelsRead = 0; //reset number of pixels read
				break;
			}

			//build global chipID using fecID and boardID
			globalChipID = (fecID * _nBoardsPerFec + boardID) * _nChipsPerBoard + chipID;

		}
		else
		{
			int tempX(-1), tempY(-1), tempV(-1);

			//read in one pixel per line
			dataFileStream >> tempX;
			dataFileStream >> tempY;
			dataFileStream >> tempV;
			dataFileStream.ignore(256,'\n');

			int cellID0 = _nPixels * (tempY % _nPixels) + (tempX % _nPixels);

			if(tempX < 0 || tempY < 0 || tempV < 0){continue;}

			if(_noChipHeaders)
			{
				globalChipID = getChipID(tempX,tempY);
			}

			if(isFirstPixel)
			{
				//create new trackerrawdata object
				rawData = new TrackerRawDataImpl;
				rawData->setCellID0(cellID0);
				rawData->setCellID1(globalChipID);

				adcValues = new EVENT::ShortVec;

				isFirstPixel = false;

				streamlog_out(DEBUG0) << "Creating new rawdata object with cellID0 " << cellID0 << " and cellID1 " << globalChipID << std::endl;
			}
			else if( ( (prevGlobalChipID != globalChipID) || (prevY != tempY) || (prevX != tempX-1) ) && !isFirstPixel ) //if cannot add pixel to trackerrawdata object anymore, close it and create a new one
			{
				//set adc values of trackerrawdata object by copying values
				rawData->setADCValues( *adcValues );
				//add trackerrawdata abject to collection, collection takes ownership
				trackerRawDataCol->push_back( rawData );

				//clear up memory before creating a new shortvec
				delete adcValues;

				streamlog_out(DEBUG0) << "Adding rawdata object to collection" << std::endl;

				//create new trackerrawdata object
				rawData = new TrackerRawDataImpl;
				rawData->setCellID0(cellID0);
				rawData->setCellID1(globalChipID);

				adcValues = new EVENT::ShortVec;

				streamlog_out(DEBUG0) << "Creating new rawdata object with cellID0 " << cellID0 << " and cellID1 " << globalChipID << std::endl;
			}

			adcValues->push_back(tempV);

			streamlog_out(DEBUG0) << "Adding pixel with cellID0 " << cellID0 << " and cellID1 " << globalChipID << " to rawdata object" << std::endl;

			prevX = tempX;
			prevY = tempY;
			prevGlobalChipID = globalChipID;

			pixelsRead++;
			_activatedPixels++;
		}
	}

	//set adc values of trackerrawdata object by copying values
	rawData->setADCValues( *adcValues );
	//add trackerrawdata abject to collection, collection takes ownership
	trackerRawDataCol->push_back( rawData );

	//clear up memory
	delete adcValues;

	streamlog_out(DEBUG0) << "Adding rawdata object to collection" << std::endl;

	if( trackerRawDataCol->empty() ){

		delete trackerRawDataCol;
		trackerRawDataCol = NULL;
	}

	return trackerRawDataCol;

} // end of IMPL::LCCollectionVec* TimePixXYReaderProcessor::readXYFile( const std::string& fileName )


int TimePixXYReaderProcessor::getChipID( const int& col, const int& row ) const {

	const int chipInRow = row / _nPixels;
	const int chipInCol = col / _nPixels;

	const int entry = ( chipInRow * _nChipColumn ) + chipInCol ;

	return _chipIDs[entry];

} // end of int TimePixXYReaderProcessor::getChipID( const int& col, const int& row ) const





} // end of namespace marlin
