#ifndef ChannelCorrectionNew_h
#define ChannelCorrectionNew_h 1

//C++
#include <iostream>
#include <string>
#include <utility>
#include <bitset>

//LCIO
#include "lcio.h"
#include "UTIL/LCFixedObject.h"

#define ChannelCorrectionNewNINTVals 5       // pad id, module id and quality bit
#define ChannelCorrectionNewNFLOATVals 3     // channel gain calibration value, channel time offset and gain calibration offset
#define ChannelCorrectionNewNDOUBLEVals 0

namespace tpcconddata
{

    class ChannelCorrectionNew;

    typedef ChannelCorrectionNew LCChannelCorrectionNew;

    /** Class that combines information on ChannelCorrectionNews for storage.
     *  The pad ID and the readout module ID (Gear) as well as the channel ID and readout unit (hardware)
     *  are stored. In addition channel correction and status values are stored into a single object of type
     *  LCFixedObject.
     *  This object can be used to be stored in a LCCollectionVec for conditions data.
     *  Values stored besides the idetifiers are gain calibration factor (default: 1.0),
     *  time offset (default: 0.0), and a quality bit (as integer) with (from "left to right")
     *  the first bit to mark dead channels, the second for noisy channels and two additional ones for free
     *  use.
     *
     * @author R. Diener (DESY)
     */

    class ChannelCorrectionNew : public UTIL::LCFixedObject<ChannelCorrectionNewNINTVals,ChannelCorrectionNewNFLOATVals,ChannelCorrectionNewNDOUBLEVals>
    {

        public:

            /** Constructor which sets all values of the object at once.
             *  If padID/moduleID or channelID/readouGroup are unknown, set them to -1. In this case,
             *  an tpcconddata::ConditionsObjectException if tried to access these values.
             */
            ChannelCorrectionNew(int padID, int moduleID, int channelID, int readoutGroup, int quality, float calibrationValue, float timeOffset, float calibrationOffset = 0);

            /** Constructor wich only sets pad identifier by padID/moduleID and channelID/readouGroup.
             *  If padID/moduleID or channelID/readouGroup are unknown, set them to -1. In this case,
             *  an tpcconddata::ConditionsObjectException if tried to access these values.
             *  Automatically set are: Quality=0, calibration=1.0, time offset=0.0.
             */
            ChannelCorrectionNew(int padID, int moduleID, int channelID, int readoutGroup);

            /** 'Copy constructor' needed to interpret LCCollection read from file/database.
             */
            ChannelCorrectionNew(EVENT::LCObject* obj):UTIL::LCFixedObject<ChannelCorrectionNewNINTVals,
                                                               ChannelCorrectionNewNFLOATVals,
                                                               ChannelCorrectionNewNDOUBLEVals>(obj) { }

            /** Important for memory handling */
            virtual ~ChannelCorrectionNew();


            // the class interface:
            /** Returns a pair of integers of which the first element ist the pad ID and
             *  the second element the module ID as given in Gear. If the number "-1" is
             *  stored for one these (placeholder for unknown), a tpcconddata::ConditionsObjectException is thrown
             */
            std::pair<int,int> getGearID() const;
            /** Returns a pair of integers of which the first element ist the channel ID and
             *  the second element the readout group as given from the readout electronics. If the number "-1" is
             *  stored for one these (placeholder for unknown), a tpcconddata::ConditionsObjectException is thrown
             */
            std::pair<int,int> getHardwareID() const;

            /** Get gain calibration factor, should be given in the way, that if you divide the recorded charge by this number, the correct charge will result */
            float getCalibrationFactor() const;
            /** time offset should be given in the way, that if you subtract this value from the recorded pulse time the correct one will result */
            float getTimeOffset() const;
            /** gain calibration offset should be given in the way, that if you subtract this value from the recorded charge the correct one will result */
            float getCalibrationOffset() const;
            int   getQuality() const;
            bool  isDead();
            bool  isNoisy();
            /** One  of two not predefined quality bits */
            bool  isAddonBitASet();
            /** One  of two not predefined quality bits */
            bool  isAddonBitBSet();

            /** Get gain calibration factor, should be given in the way, that if you divide the recorded charge by this number, the correct charge will result */
            void setCalibrationFactor(float calibrationValue);
            /** time offset should be given in the way, that if you subtract this value from the recorded pulse time the correct one will result */
            void setTimeOffset(float timeOffset);
            /** gain calibration offset should be given in the way, that if you subtract this value from the recorded charge the correct one will result */
            void setCalibrationOffset(float calibrationOffset);
            void setQuality(int quality) ;
            void setIsDead(bool toggle) ;
            void setIsNoisy(bool toggle) ;
            /** Sets the first of two not predefined quality bits */
            void setAddonBitA(bool toggle) ;
            /** Sets the second of two not predefined quality bits */
            void setAddonBitB(bool toggle) ;

            static std::string getRevision();
            static std::string getDefaultColName();

            void print(  std::ostream& os = std::cout ) const;

            // -------- need to implement abstract methods from LCGenericObject
            const std::string getTypeName() const
            {
                return std::string("ChannelCorrectionNew");
            }
            const std::string getDataDescription() const
            {
                return std::string("i:padID,moduleID,channelID,readoutGroup,quality,f:calibrationValue,timeOffset,calibrationOffset");
            }

        protected:

            void setGearID(int padID, int moduleID);
            void setHardwareID(int channelID, int readoutGroup);
            std::bitset<4> _qualityBit;

    };// end of class

    std::ostream &operator<<(std::ostream &os, const ChannelCorrectionNew &p);


} //end namespace tpcconddata
#endif
