/*
 * TPCDistortionMeasurement.h
 *
 *  Created on: Apr 2, 2014
 *      Author: zenker
 */
#ifndef TPCDistortionMeasurement_h
#define TPCDistortionMeasurement_h 1

//C++
#include "iostream"
#include "string"

// CLHEP
#include <CLHEP/Vector/TwoVector.h>

//LCIO
#include "lcio.h"
#include "EVENT/LCGenericObject.h"
#include "IMPL/LCGenericObjectImpl.h"

namespace tpcconddata {

  class TPCDistortionMeasurement;

  /** class to store the distortions measured with the photoelectric system using a laser.
   *  The class is also able to handle the module ID. It is fully backward compatible
   *  to the version without module ID, although the implementation has changed from
   *  LCFixedObject to LCGenerigObject to provide the required flexibility.
   *
   *  Like the LCFixedObject this thing contains an LCGenericObject, but is not derived from it.
   *
   *  @author K. Zenker (DESY)
   */
  class TPCDistortionMeasurement: public EVENT::LCGenericObject {

  public:

    /** The constructor with all the member functions given.
     *  Use this to create on object on the heap.
     *  @param truePosition           True position of the photodot
     *  @param reconstructedPosition  The reconstructed position of the photodot
     *  @param dot_id                 The id of the photodot
     */
    TPCDistortionMeasurement(CLHEP::Hep3Vector* truePosition, CLHEP::Hep3Vector* reconstructedPosition, int dot_id);
    /** The constructor to be used with LCObjects from a collection.
     */
    TPCDistortionMeasurement(EVENT::LCObject* obj);

    /** The desctructor.
     */
    virtual ~TPCDistortionMeasurement();
    CLHEP::Hep3Vector getTruePosition() const;
    CLHEP::Hep3Vector getReconstructedPosition() const;

    int getDotID() const;

    static std::string getRevision();
    static std::string getDefaultColName();

    IMPL::LCGenericObjectImpl* obj() const {
      return _myObj;
    }
    virtual int id() const {
      return _myObj->id();
    }

    void print(std::ostream& os = std::cout) const;

    // -------- need to implement abstract methods from LCGenericObject
    virtual const std::string getTypeName() const {
      return std::string("TPCDistortionMeasurement");
    }

    virtual const std::string getDataDescription() const {
      return std::string("d:trueX,trueY,trueZ,recoX,recoY,recoZ");
    }

    // the other functions inherited from LCGenericObject
    // just pass them to the implementation
    virtual int getNInt() const {
      return _myObj->getNInt();
    }
    virtual int getNFloat() const {
      return _myObj->getNFloat();
    }
    virtual int getNDouble() const {
      return _myObj->getNDouble();
    }

    virtual int getIntVal(int i) const {
      return _myObj->getIntVal(i);
    }
    virtual float getFloatVal(int i) const {
      return _myObj->getFloatVal(i);
    }
    virtual double getDoubleVal(int i) const {
      return _myObj->getDoubleVal(i);
    }

    virtual bool isFixedSize() const {
      return _myObj->isFixedSize();
    }


  protected:
    bool _createObj;
    IMPL::LCGenericObjectImpl* _myObj;

  };

std::ostream &operator<<(std::ostream &os, const TPCDistortionMeasurement &acm);
}
#endif /* TPCDistortionMeasurement_H_ */
