/* ChannelCorrectionNew.cc
 * $Id$
 */
#include <bitset>
#include <sstream>
#include <iostream>
#include <string.h>

#include "ChannelCorrectionNew.h"
#include "TPCCondDataExeptions.h"

namespace tpcconddata{

    std::string ChannelCorrectionNew::getRevision()
    {
        return std::string("$Rev: 5156 $");
    }


    std::string ChannelCorrectionNew::getDefaultColName()
    {
        return std::string("TPCChannelCorrectionNew");
    }


    ChannelCorrectionNew::ChannelCorrectionNew(int padID, int moduleID, int channelID, int readoutGroup, int quality, float calibrationValue, float timeOffset, float calibrationOffset)
    {
        setGearID(padID, moduleID);
        setHardwareID(channelID, readoutGroup);
        setQuality(quality);
        setCalibrationFactor(calibrationValue);
        setTimeOffset(timeOffset);
        setCalibrationOffset(calibrationOffset);
    }


    ChannelCorrectionNew::ChannelCorrectionNew(int padID, int moduleID, int channelID, int readoutGroup)
    {
        setGearID(padID, moduleID);
        setHardwareID(channelID, readoutGroup);
        setQuality(0);
        setCalibrationFactor(1.0);
        setTimeOffset(0.0);
        setCalibrationOffset(0.0);
    }


    ChannelCorrectionNew::~ChannelCorrectionNew() { /* no op*/  }


    std::pair<int,int> ChannelCorrectionNew::getGearID() const
    {
        if ( getIntVal(0)!=-1 && getIntVal(1)!=-1 )
        {
            return std::make_pair(getIntVal(0), getIntVal(1));
        }
        else
        {
            std::stringstream message;
            message << "TPCCondData::ChannelCorrectionNew::getGearID(): padID and/or moduleID not set (out of range!)";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
    }



    std::pair<int,int> ChannelCorrectionNew::getHardwareID() const
    {
        if ( getIntVal(2)!=-1 && getIntVal(3)!=-1 )
        {
            return std::make_pair(getIntVal(2), getIntVal(3));
        }
        else
        {
            std::stringstream message;
            message << "TPCCondData::ChannelCorrectionNew::getHardwareID(): channelID and/or readoutGroup not set (out of range!)";
            throw tpcconddata::CondDataNotAvailableException(message.str());
        }
    }



    int ChannelCorrectionNew::getQuality() const
    {
        return getIntVal(4);

    }



    bool ChannelCorrectionNew::isDead()
    {
        return (getIntVal(4)>>0) & 1;
    }



    bool ChannelCorrectionNew::isNoisy()
    {
        return (getIntVal(4)>>1) & 1;
    }



    bool ChannelCorrectionNew::isAddonBitASet()
    {
        return (getIntVal(4)>>2) & 1;
    }



    bool ChannelCorrectionNew::isAddonBitBSet()
    {
        return (getIntVal(4)>>3) & 1;
    }



    float ChannelCorrectionNew::getCalibrationFactor() const
    {
        return getFloatVal( 0 );
    }


    float ChannelCorrectionNew::getTimeOffset() const
    {
        return getFloatVal( 1 );
    }


    float ChannelCorrectionNew::getCalibrationOffset() const
    {
        return getFloatVal( 2 );
    }


    void ChannelCorrectionNew::setCalibrationFactor(float calibrationValue)
    {
        _obj->setFloatVal( 0, calibrationValue );
    }


    void ChannelCorrectionNew::setTimeOffset(float timeOffset)
    {
        _obj->setFloatVal( 1, timeOffset );
    }


    void ChannelCorrectionNew::setCalibrationOffset(float calibrationOffset)
    {
        _obj->setFloatVal(2, calibrationOffset);
    }


    void ChannelCorrectionNew::setQuality(int quality)
    {
        if ( quality > 15 || quality < 0)
        {
            std::stringstream message;
            message << "TPCCondData::ChannelCorrectionNew::setQuality(int): Integer to set quality bit must be between 0 and 15, but is: "<<quality<<" (out of range!)";
            throw tpcconddata::ConditionsObjectException(message.str());
        }
        _qualityBit=std::bitset<4>(quality);
        _obj->setIntVal( 4, quality );
    }



    void ChannelCorrectionNew::setIsDead(bool toggle)
    {

        _qualityBit.set(0, toggle);
        _obj->setIntVal( 4, static_cast<int>(_qualityBit.to_ulong()) );
    }



    void ChannelCorrectionNew::setIsNoisy(bool toggle)
    {

        _qualityBit.set(1, toggle);
        _obj->setIntVal( 4, static_cast<int>(_qualityBit.to_ulong()) );
    }



    void ChannelCorrectionNew::setAddonBitA(bool toggle)
    {

        _qualityBit.set(2, toggle);
        _obj->setIntVal( 4, static_cast<int>(_qualityBit.to_ulong()) );
    }



    void ChannelCorrectionNew::setAddonBitB(bool toggle)
    {

        _qualityBit.set(3, toggle);
        _obj->setIntVal( 4, static_cast<int>(_qualityBit.to_ulong()) );
    }



    void ChannelCorrectionNew::print(  std::ostream& os ) const
    {
        std::streamsize PrecisionBefore = os.precision();
        std::streamsize WidthBefore = os.width();
        os.precision(3);  os.setf(std::ios_base::fixed);
        os << "ChannelCorrectionNew: pad:" << getGearID().first << "/module:" << getGearID().second <<
                "//channel:" << getHardwareID().first << "/readoutGroup:" << getHardwareID().second <<
                " quality:" << getQuality()<< std::fixed;
        os << " calibrationValue: " << getCalibrationFactor() << " timeOffset: " << getTimeOffset() << " calibrationOffset: " << getCalibrationOffset() << std::endl;
        os.unsetf( std::ios_base::fixed );
        os.precision( PrecisionBefore );
        os.width( WidthBefore );
    }



    void ChannelCorrectionNew::setGearID(int padID, int moduleID)
    {
        _obj->setIntVal( 0, padID );
        _obj->setIntVal( 1, moduleID );
    }



    void ChannelCorrectionNew::setHardwareID(int channelID, int readoutGroup)
    {
        _obj->setIntVal( 2, channelID );
        _obj->setIntVal( 3, readoutGroup );
    }


    std::ostream &operator<<(std::ostream &os, const ChannelCorrectionNew &p) {
	p.print(os);
	return os;
    }

}
