#include "lcio.h"

#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCRunHeaderImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "IMPL/MCParticleImpl.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cstdio>
#include <typeinfo>
#include <cmath>
#include <cstdlib>

#include <CLHEP/Random/RandGauss.h>

using namespace std ;
using namespace lcio ;

void printUsage()
{
  cout << "generateElectronsForLP1 energy phi zpos nEvents [outfile.slcio]" <<endl<<endl;
  cout <<"* Generates electrons coming from the 1200,50,zpos with constant z position." << endl;
  cout <<"* Use 0..2pi to define a fixed phi angle." << endl;
    cout <<"* Outfile name is optional, " << " default is MCElectrons.slcio"<< endl;
    cout << "* Energy is in GeV" << endl;
}

int main(int argc, char **argv)
{

  if (argc!=5 && argc!=6 )
  {
    printUsage();
    return(1);
  }

  
   // get outfile name
  string outFileName;
  if (argc==6)
    outFileName=argv[5];
  else
    outFileName="LP1Electrons.slcio";

  double energy = std::atof( argv[1] );
  double phi_in =  std::atof( argv[2] );
  double zpos =  std::atof( argv[3] );
  double lambda =  0;
  unsigned int nEvents = std::atoi( argv[4] );

  static const double m = 0.0005; // the electron mass in GeV

  if (energy < m)
  {
    cout << "Error, energy cannot be smaller than mass of the electron ( "
	 << m << " GeV)!" << endl;
    return -1;
  }

  // open outfile			     
  try{
    // create sio writer
    LCWriter* lcWrt = LCFactory::getInstance()->createLCWriter() ;
    
    try{  lcWrt->open( outFileName, LCIO::WRITE_NEW ) ;
    }
    catch( IOException& e ){    
      cout << e.what() << endl ;
      return 0 ;
    }
      
      // create runheader and write it
      // the information should come from a file
      // or be given interactively in a future version
      LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
      int runNr=0;
      runHdr->setRunNumber( runNr) ;
      runHdr->setDetectorName ("LPTPC");      
      lcWrt->writeRunHeader( runHdr ) ;
      
      // loop the events, generate 100,000 events
      for (unsigned int eventNr=0; eventNr < nEvents; eventNr++)
      {
	// preparation: calculate the momentum
	  double p[3];
	  double phi=phi_in*M_PI/180;

	  p[0]=sqrt( energy*energy - m*m ) * cos(phi) *cos(lambda);
	  p[1]=sqrt( energy*energy - m*m ) * sin(phi) *cos(lambda);
	  p[2]=sqrt( energy*energy - m*m ) * sin(lambda);


       // now start the real work...
	// create new lcio event
	LCEventImpl*  evt = new LCEventImpl() ;
	
	evt->setRunNumber(  runNr   ) ;
	evt->setEventNumber( eventNr ) ;
      
	// we want to store a collection of trackerRawData
	LCCollectionVec* mcElectronsVec = new LCCollectionVec( LCIO::MCPARTICLE )  ;
	
	MCParticleImpl *electron =  new MCParticleImpl;
	electron->setPDG(11); // it's a electron
	electron->setCharge(-1); // electrons have chage -1
	electron->setMass( m ); // the electron mass
	electron->setGeneratorStatus(1); // a stable particle in the detector
	electron->setMomentum( p ); // the the corresponding momentum 
	//beam spread of 1mm in all directions except z to avoid binning effects in simulation
	double randx=CLHEP::RandGauss::shoot( 0,1 );
	double randy=CLHEP::RandGauss::shoot( 0,1 );
	double randz=CLHEP::RandGauss::shoot( 0,5 );
	double vtx[3]={1200+randx,50+randy,zpos+randz};
	electron->setVertex(vtx);
	// add the electron
	mcElectronsVec->addElement( electron );

	// add the rawData collection to the event
	evt->addCollection( mcElectronsVec , "MCElectrons" ) ;

	// write the event to the file
	lcWrt->writeEvent( evt ) ;
	
	// dump the event to the screen 
	LCTOOLS::dumpEvent( evt ) ;

	// ------------ IMPORTANT ------------- !
	// we created the event so we need to delete it ...
	delete evt ;
	// -------------------------------------

      } // for eventNr

      //      delete runHdr ;
      
      lcWrt->close() ;
      
  } catch( std::exception& ex){
    
    cout << " an excpetion occured: " << endl ;
    cout << "  " << ex.what() << endl ;
    return 1 ;
  }
     
return 0;

}
