#include <iostream>
#include <cmath>
#include <complex>
#include "helix.h"
#include "TMath.h"
#include "TRandom.h"

using namespace std;

Circle::Circle(double x0, double y0, double r)
{
   _c[0] = x0;
   _c[1] = y0;

   _r    = r;
}

Circle::Circle()
{
   setCenter(0., 0.);
   setR(1.);
}

bool Circle::calcXPWithCircle(Circle& c, vector<Vector2D>& xp)
{
  Vector2D cc = c.getCenter();
  double   r  = c.getR();

  double distance = sqrt( (_c[0]-cc[0])*(_c[0]-cc[0]) + (_c[1]-cc[1])*(_c[1]-cc[1]) );

  cout << "circle1: " << "x=" << _c[0] << " y=" << _c[1] << " R=" << _r << endl
       << "circle2: " << "x=" << cc[0] << " y=" << cc[1] << " R=" <<  r << endl
       << "distance=" << distance << endl;

  if(distance < (_r+r) && distance > fabs(_r-r)){
    double A1 = -2*_c[0];
    double B1 = -2*_c[1];
    double C1 = _c[0]*_c[0] + _c[1]*_c[1] - _r*_r;

    double A2 = -2*cc[0];
    double B2 = -2*cc[1];
    double C2 = cc[0]*cc[0] + cc[1]*cc[1] - r*r;

    Vector2D xp1, xp2;

    if(B1 != B2){
       double k1 = -(A1-A2)/(B1-B2);
       double k2 = -(C1-C2)/(B1-B2);

       double a1  = 1+k1*k1;
       double a2  = A1+B1*k1+2*k1*k2;
       double a3  = k2*k2+B1*k2+C1;

       double ds  = sqrt(a2*a2-4*a1*a3);

       xp1[0] = (-a2 + ds)/(2.*a1);
       xp1[1] = k1*xp1[0] + k2;

       xp2[0] = (-a2 - ds)/(2.*a1);
       xp2[1] = k1*xp2[0] + k2;
    }
    else{
      xp1[0] = xp2[0] = -(C1-C2)/(A1-A2);
      xp1[1] = sqrt(_r*_r - (xp1[0] - _c[0])*(xp1[0] - _c[0]));
      xp2[1] = -xp1[1];
    }


    xp.push_back(xp1);
    xp.push_back(xp2);

    cout << "two solutions:" << endl;
    cout << "xp1:(x,y)=" << xp1[0] << ", " << xp1[1] << endl;
    cout << "xp2:(x,y)=" << xp2[0] << ", " << xp2[1] << endl;

    return true;
  }
  else
  {
     cout << "no solution!!!" << endl;
     return false;
  }
}

const double MAXZ = 2200.;
const double MINZ = 0.;

Helix::Helix(double xc, double yc, double r,
             TVector3 X0,
             double lambda, double sign)
            :Circle(xc, yc, r), 
             _lambda(lambda), 
             _sign(sign), 
             _X0(X0), 
             _MAXZ(MAXZ), 
             _MINZ(MINZ)
{
}

Helix::Helix()
      :Circle(0, 0, 1),
       _lambda(0.),
       _sign(-1),
       _X0(TVector3(1., 0., 0)),
       _MAXZ(MAXZ),
       _MINZ(MINZ)
{
}

bool Helix::calcXPWithLayer(Circle& c, vector<TVector3>& xp)
{
   vector<Vector2D> xpp;

   if(calcXPWithCircle(c, xpp))
   {
     double phi0 = atan2(_X0.Y(), _X0.X());
     double z0 = _X0.Z();
     double tanl = tan(_lambda);

     Vector2D cc = getCenter();


     for(int n = 0; n < 2; n++){
       double xx = xpp.at(n)[0];
       double yy = xpp.at(n)[1];

       double theta  = atan2(yy, xx);

       // resolution
       theta += gRandom->Gaus(0., 0.1) / c.getR();//rphi resolution: 0.1 mm


       xx = c.getR() * cos(theta);
       yy = c.getR() * sin(theta);

       double phi = atan2(yy - cc[1], xx - cc[0]);
       TVector3 is;

       double z;
       
       if(phi > phi0){
          z = z0 + (phi-phi0)*getR() * tanl + gRandom->Gaus(0., 0.2);
          if(z < _MAXZ){
            is.SetXYZ(xx, yy, z);
            xp.push_back(is);
          }
       }

       if(tanl == 0.) continue;

       int k = 1;

       while(1){
          z = z0 + ( 2*k*TMath::Pi() + (phi-phi0) )*getR() * tanl + gRandom->Gaus(0., 0.1);
          if(z > _MAXZ)break;

          cout << "z = " << z << endl;
          is.SetXYZ(xx, yy, z);
          xp.push_back(is);  

          k++;
       }

     }
   }

   if(xp.size()>0)
   {
     for(unsigned int n = 0; n < xp.size(); n++)
     {
       cout << "xxp.X = " << xp.at(n).X() << endl;
     }
     return true;
   }
   else 
     return false;
}
