#include "lcio.h"

#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h"
#include "IMPL/LCRunHeaderImpl.h"
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "IMPL/TrackerRawDataImpl.h"
#include "IMPL/TrackerDataImpl.h"
#include "UTIL/LCTime.h"
#include "IMPL/LCFlagImpl.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cstdio>
#include <typeinfo>
#include <cmath>
#include <cstdlib>
#include <climits>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

//using namespace std ;
//using namespace lcio ;

void printUsage()
{
    std::cout << "generateRawData [outfile.slcio]" <<std::endl;
    std::cout <<" outfile name is optional, "
        << "default is TPCPedestalRawData.slcio"<< std::endl;
}

unsigned time_seed()
{
    // make a seed for the rand48 random varibales by taking the hash of the system time
    time_t now = time ( 0 );
    unsigned char *p = (unsigned char *)&now;
    unsigned seed = 0;
    size_t i;
    
    for ( i = 0; i < sizeof now; i++ )
        seed = seed * ( UCHAR_MAX + 2U ) + p[i];
    return seed;
}

double rand48_gaussian (const double sigma)
{
    // gives back gaussian distributed random variables
    double x, y, r2;
    do
    {
        /* choose x,y in uniform square (-1,-1) to (+1,+1) */
        x = -1 + 2 * drand48();
        y = -1 + 2 * drand48();
        /* see if it is in the unit circle */
        r2 = x * x + y * y;
    }
    while (r2 > 1.0 || r2 == 0);
    /* Box-Muller transform */
    return sigma * y * sqrt (-2.0 * log (r2) / r2);
}

int main(int argc, char **argv)
{
    // initialize the random seed
    srand ( time_seed() );
    // initialize and seed the gsl random generator
    gsl_rng * gslR = gsl_rng_alloc (gsl_rng_taus2);
    gsl_rng_set (gslR, time_seed());


    if (argc!=1 && argc!=2 )
    {
        printUsage();
        return(1);
    }

    // get outfile name
    std::string outFileName;
    if (argc==2)
        outFileName=argv[1];
    else
        outFileName="TPCPedestalRawData.slcio";

    //int day=1;
    int days=0;
    int hour=0;
    int minute=0;
    int second=0;    
    
    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0);
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1);

        
    // open outfile
    try
    {
        // create sio writer
        lcio::LCWriter* lcWrt = lcio::LCFactory::getInstance()->createLCWriter() ;

        try
        {
            lcWrt->open( outFileName, lcio::LCIO::WRITE_NEW ) ;
        }
        catch( lcio::IOException& e )
        {
            std::cout << e.what() << std::endl ;
            return 0 ;
        }

        // create runheader and write it
        // the information should come from a file
        // or be given interactively in a future version
        lcio::LCRunHeaderImpl* runHdr = new lcio::LCRunHeaderImpl ;
        int runNr=0;
        runHdr->setRunNumber( runNr ) ;

        //      string detName("demo")  ;
        std::string detName("SimulationTPC");
        runHdr->setDetectorName( detName ) ;

        std::stringstream description ;
        description << "simulation tpc rawData";
        runHdr->setDescription( description.str()  ) ;

        std::string tpcName("simulationTpc") ;
        runHdr->addActiveSubdetector( tpcName ) ;

        lcWrt->writeRunHeader( runHdr ) ;

        // loop the events, generate 1000 events
        int maxEvents=1000;
        
        for (int eventNr=0; eventNr<maxEvents; eventNr++)
        {
            // create new lcio event
            lcio::LCEventImpl*  evt = new lcio::LCEventImpl() ;

            evt->setRunNumber(  runNr   ) ;
            evt->setEventNumber( eventNr ) ;
            evt->setDetectorName( detName ) ;
            //create a time from the event number so that every n seconds there will be one event
            int timeEventNr=eventNr*10;
            hour=(int)(timeEventNr/3600);
            days=0;
            while (hour > 23) {
                hour-=24;
                days+=1;
                timeEventNr-=86400;
            }
            minute=(int)((timeEventNr-(hour*3600))/60);
            second=(int)(timeEventNr-(hour*3600)-(minute*60));
            if ( eventNr%100 == 0 )
            {
                std::cout<<"Time of event ";
                if ( eventNr < 100 ) std::cout<<"  ";
                std::cout<<eventNr<<": 2001, 1, "<<1+days<<", "<<hour<<", "<<minute<<", "<<second<<std::endl;
            }
            lcio::LCTime evtTime (2001, 1, 1+days, hour, minute, second);
            evt->setTimeStamp(evtTime.timeStamp());

            // we want to store a collection of trackerRawData
            lcio::LCCollectionVec* trackerDataVec = new lcio::LCCollectionVec( lcio::LCIO::TRACKERRAWDATA )  ;
            trackerDataVec->setFlag(trkFlag.getFlag());

            int integer=0;

            
            //loop over 2 modules
            for (int modulenum = 0; modulenum < 2 ; modulenum++ ){
                //place one hit on each pad; loops over rows and pads in row    
                for(int padnum = 0; padnum<10 ; padnum++)
                {
                    lcio::TrackerRawDataImpl* Data = new lcio::TrackerRawDataImpl;
                    Data->setCellID0(padnum);
                    Data->setCellID1(modulenum);
                    integer++;
    
                    EVENT::ShortVec ADCValues;
                    
                    double sliceValue=0.0;
                    //filling the vector with ADCvalues with pedestal offset and gaussian distributed width
                    double sliceWidth=((double)(modulenum+1));
                    
                    for(int i=0; i<512; i++)
                    {
                        sliceValue=(float)modulenum*3.0+5.0+rand48_gaussian(sliceWidth);
                        if (sliceValue>255.0) sliceValue=255.0;
                        if (sliceValue<0.0) sliceValue=0.0;
                        ADCValues.push_back(static_cast<short>(sliceValue+0.50));
                    }

                    // vary the pulse height with a scale parameter which is distributed randomly
                    // according to a Landau distribution truncated below 0 and above 4
                    // the average scale parameter "heightModifier" should be very close to 1
                    double heightModifier = gsl_ran_landau(gslR)/5.10+0.50;
                    while ( !(heightModifier<=4.0 && heightModifier >=0.0)) {
                        heightModifier=gsl_ran_landau(gslR)/5.10+0.50;
                    }
                    // find a random start slice between 10 and 460                    
                    int startSlice=10+(int)((drand48()*450+0.5));
                    
                    // fill a pulse derived from an average pulse shape in the ADC vector
                    // this is then scaled and added to the noise already in the slices
                    ADCValues[startSlice]  =static_cast<short>((1.5*heightModifier)+ADCValues[startSlice]);
                    ADCValues[startSlice+1]=static_cast<short>((5.5*heightModifier)+ADCValues[startSlice+1]);
                    ADCValues[startSlice+2]=static_cast<short>((10.5*heightModifier)+ADCValues[startSlice+2]);
                    ADCValues[startSlice+3]=static_cast<short>((15.5*heightModifier)+ADCValues[startSlice+3]);
                    ADCValues[startSlice+4]=static_cast<short>((30.5*heightModifier)+ADCValues[startSlice+4]);
                    ADCValues[startSlice+5]=static_cast<short>((50.5*heightModifier)+ADCValues[startSlice+5]);
                    ADCValues[startSlice+6]=static_cast<short>((60.5*heightModifier)+ADCValues[startSlice+6]);
                    ADCValues[startSlice+7]=static_cast<short>((50.5*heightModifier)+ADCValues[startSlice+7]);
                    ADCValues[startSlice+8]=static_cast<short>((30.5*heightModifier)+ADCValues[startSlice+8]);
                    ADCValues[startSlice+9]=static_cast<short>((15.5*heightModifier)+ADCValues[startSlice+9]);
                    ADCValues[startSlice+10]=static_cast<short>((5.5*heightModifier)+ADCValues[startSlice+10]);
                    ADCValues[startSlice+11]=static_cast<short>((1.5*heightModifier)+ADCValues[startSlice+11]);
                    
                    // check for overflow values and cut them
                    for (int iSlice=startSlice ; iSlice<=startSlice+11 ; iSlice++ ){
                        if (ADCValues[iSlice] > 255 ) ADCValues[iSlice] = 255;
                        if (ADCValues[iSlice] < 0 ) ADCValues[iSlice] = 0;
                    }
    
                    Data->setADCValues(ADCValues);
                    Data->setTime(0);
    
                    trackerDataVec->push_back(Data);
                } //        for(int j = 0; j<512 ; j++)
            } //  (int modulenum = 0; modulenum <3 ; modulenum++)

            // add the rawData collection to the event
            evt->addCollection( trackerDataVec , "TPCRawData" ) ;

            // write the event to the file
            lcWrt->writeEvent( evt ) ;

            // dump the event to the screen
            // lcio::LCTOOLS::dumpEvent( evt ) ;

            // ------------ IMPORTANT ------------- !
            // we created the event so we need to delete it ...
            delete evt ;
            // -------------------------------------

        }                        // for eventNr

        delete runHdr ;

        lcWrt->close() ;
        lcio::LCTime evtTimeStart (2010, 1, 1, 0, 0, 0);
        lcio::LCTime evtTimeStop (2010, 1, 1+days, hour, minute, second);
        
        std::cout<<"first event time: "<<evtTimeStart.timeStamp()<<std::endl;
        std::cout<<" last event time: "<<evtTimeStop.timeStamp()<<std::endl;

    } // try
    catch( std::exception& ex)
    {
        std::cout << " an exception occured: " << std::endl ;
        std::cout << "  " << ex.what() << std::endl ;
        return 1 ;
    }
    return 0;
}
