
#include "KalTrackingHelixTester.h"

#include <gearimpl/TPCParametersImpl.h>
#include <gearimpl/ConstantBField.h>
#include <gearimpl/FixedPadAngleDiskLayout.h>
#include <gearimpl/TPCModuleImpl.h>
#include <gearimpl/GearMgrImpl.h>
#include <lcio.h>

#include <IMPL/TrackerHitImpl.h>

#include <kaldet/GearTPCMeasLayer.h>

#include "KalTrackingHelix.h"

#include <iostream>
#include <cmath>
#include <list>
#include <memory>

using namespace gear;
using namespace lcio;
using namespace kaldet;

namespace marlintpc
{

KalTrackingHelixTester::KalTrackingHelixTester()
{
  // a tpc with 600 mm drift length and cartesian global coodinates
  TPCParametersImpl * tpcParametersImpl = new TPCParametersImpl(600, PadRowLayout2D::CARTESIAN);
  PadRowLayout2D *localPadLayoutImpl = new FixedPadAngleDiskLayout(100 /*rMin*/, 
								   1000 /*rMax*/, 
								   9 /*nRow*/,
								   0. /*phiMin*/,
								   2*M_PI /*phiMax*/, 
								   100 /*nPadsInRow*/);
  TPCModuleImpl * moduleImpl = new TPCModuleImpl(17 /*moduleID, just something not 0*/,
					 localPadLayoutImpl, 
					 tpcParametersImpl->getCoordinateType(),
					 40 /*MHz readout frequency*/);
  moduleImpl->setOffset(135, 0); // some arbitrary offset in x, must be 0 in y in current KalDet 
  tpcParametersImpl->addModule( moduleImpl );

  // We need a pointer to TPCParametersImpl to be able to add the module. 
  // As we don't want all the functionality of the impl to be accessible all over the class,
  // we use separate variables (not impls) as class wide variables and the impl we only use in 
  // the constructor.
  tpcParameters = tpcParametersImpl;
  module = moduleImpl;
  localPadLayout = localPadLayoutImpl;

  // a homogeneous B-fiels 1 T in z
  bField = new ConstantBField( Vector3D( 0, 0, 1 ) );

  GearMgrImpl * gearMgrImpl = new GearMgrImpl;
  gearMgrImpl->setTPCParameters(tpcParameters);
  gearMgrImpl->setBField(bField);

  gearMgr = gearMgrImpl;

  kalTrackingHelix = new KalTrackingHelix(*gearMgr,
					  5 /*minTrackHits*/, 
					  1 /*maxSkipRows*/, 
					  20 /*maxDeltaChi2*/,
					  2 /*hitDistanceFactor*/,
					  0 /*startingHitMode*/,
					  79. /*vDrift*/,
					  "-1" /*modulesNotUsed*/);
}	

KalTrackingHelixTester::~KalTrackingHelixTester()
{
  delete kalTrackingHelix;

  // Do not delete the tpcParameters and the bField. They belong to the GearMgr, which 
  // deletes them when it is deleted itsef.
  delete gearMgr;
}

std::pair<unsigned int, unsigned int> KalTrackingHelixTester::getNearestPointInNextLayer()
{
  std::vector<double> localRowCentres(module->getNRows());
  for (int row = 0; row < module->getNRows(); ++row)
  {
    gear::Vector2D localPadCentre = localPadLayout->getPadCenter( module->getPadIndex(row, 0) );
    localRowCentres[row]=localPadCentre[0];
  }

  unsigned int nTestsFailed = 0;
  unsigned int nTestsTotal  = 0;
  for (int row = 0; row <  module->getNRows(); ++row )
  {
    ++nTestsTotal;

    // use an autopointer, which solves the problem of memory leaks in case of exceptions
    std::unique_ptr<TrackerHitImpl> hit(new TrackerHitImpl);
    double position[3];
    // test hits in all 4 quadrants
    Vector2D gearPosition=module->localToGlobal(localRowCentres[row], -0.1 + M_PI_2*row);
    position[0] = gearPosition[0];
    position[1] = gearPosition[1];
    position[2] = 0;
    hit->setPosition(position);
    
    ModuleRowHit mrh(module->getModuleID(), row, hit.get());

    gear::Vector2D nearestHitFromKalTracking;
    try{ 
      nearestHitFromKalTracking = kalTrackingHelix->getNearestPointInNextLayer(mrh); 
      
      if (row ==  module->getNRows()-1)
      {
	// for the last row this place should not be reached
	std::cout << "KalTrackingHelixTester::getNearestPointInNextLayer(): did not get "
		  << "get an exception when expected." << std::endl;
	++nTestsFailed;
	continue;
      }
    }
    catch ( gear::Exception & )
    {
      // this exception should only show up for the last row
      if (row !=  module->getNRows()-1)
      {
	// for the last row this place should not be reached
	std::cout << "KalTrackingHelixTester::getNearestPointInNextLayer(): unexpected  exception." 
		  << std::endl;
	++nTestsFailed;
      }
      continue;
    }
     
    gear::Vector2D expectedHit = module->localToGlobal(localRowCentres[row+1], -0.1 + M_PI_2*row);
    
    // check that the two hits are equal within rounding errors. I just use 1 nm, which should be ok
    if ( ( (expectedHit[0]-nearestHitFromKalTracking[0]) > 1e-6 ) || 
	 ( (expectedHit[1]-nearestHitFromKalTracking[1]) > 1e-6 ) )
    {
      std::cout << "KalTrackingHelixTester::getNearestPointInNextLayer():" 
		<< " Expected hit at (" << expectedHit[0] << ", " << expectedHit[1]
		<< ") but got hit at ("<< nearestHitFromKalTracking[0] 
		<< ", " << nearestHitFromKalTracking[1] << ")"
		<< std::endl;
      ++nTestsFailed;
    }

  }// for row
  
  std::cout << "KalTrackingHelixTester::getNearestPointInNextLayer(): "
	    << nTestsFailed << " out of " << nTestsTotal << " tests failed." << std::endl;
  return std::pair<unsigned int, unsigned int>(nTestsFailed, nTestsTotal);
}


std::pair<unsigned int, unsigned int> KalTrackingHelixTester::testCreateGearHit()
{
  // create a moduleRowHit at (local) phi = 0.1 rad for z = 1. and at 3.2 rad for z = 599.
  // Afterwards check that the measurement layer of the hit in row i is in the 
  // cradle at position i (which should be the case in this particular geometry),
  // and that the measurement errors, the hit pointer as well as magnetic field and
  // drift velocity are set correctly.
  
  unsigned int nTestsFailed=0;
  unsigned int nTestsTotal=0;

  std::list< std::pair<double, double> > angleZs; // pairs of angle and z to test with each row
  angleZs.push_back( std::pair<double, double>(0.1, 1.) );
  angleZs.push_back( std::pair<double, double>(3.2, 599.) );
  

  for (int row = 0; row < module->getNRows(); ++row)
  {
    // FIXME: what if there is no pad 0 in the row?
    double radius = localPadLayout->getPadCenter(module->getPadIndex(row,0) )[0];
    
    for (std::list< std::pair<double, double> >::iterator angleZIter = angleZs.begin();
	 angleZIter != angleZs.end(); ++angleZIter)
    {
      std::unique_ptr<TrackerHitImpl> hit(new TrackerHitImpl);
      double position[3];
      Vector2D gearPosition=module->localToGlobal(radius, angleZIter->first);
      position[0] = gearPosition[0];
      position[1] = gearPosition[1];
      position[2] = angleZIter->second;
      hit->setPosition(position);
      
      ModuleRowHit mrh(module->getModuleID(), row, hit.get());
    
      std::unique_ptr<GearTPCHit> gearTPCHit(kalTrackingHelix->createGearHit(mrh));

      ++nTestsTotal; // each call of createGearHit counts as one test
      
	  GearTPCMeasLayer * referenceMeasLayer = 
	dynamic_cast<GearTPCMeasLayer *>(kalTrackingHelix->_detectorCradle.At( row ));

      if ( &(gearTPCHit->GetMeasLayer()) != referenceMeasLayer )
      {
	std::cout << "KalTrackingHelixTester::testCreateGearHit(): Test failed: hit in row " 
		  << row << " at angle " << angleZIter->first  << " does not have the right "
		  << "measurement layer" << std::endl;
	++nTestsFailed;
	continue;
      }

      if( gearTPCHit->GetDX(0) != referenceMeasLayer->GetSigmaX( position[2] ) )
      {
	std::cout << "KalTrackingHelixTester::testCreateGearHit(): Test failed: hit in row " 
		  << row << " at angle " << angleZIter->first  << " does not have the right "
		  << "x-Error" << std::endl;
	++nTestsFailed;
	continue;
      }

      // std::cout << "z = " << angleZIter->second << " DZ = " << gearTPCHit->GetDX(1) << std::endl;
      if( gearTPCHit->GetDX(1) != referenceMeasLayer->GetSigmaZ( position[2] ) )
      {
	std::cout << "KalTrackingHelixTester::testCreateGearHit(): Test failed: hit in row " 
		  << row << " at angle " << angleZIter->first  << " does not have the right "
		  << "z-Error" << std::endl;
	++nTestsFailed;
	continue;
      }
      
      if (gearTPCHit->GetHitPtr() != hit.get() )
      {
	std::cout << "KalTrackingHelixTester::testCreateGearHit(): Test failed: hit in row " 
		  << row << " at angle " << angleZIter->first  << " points to the wrong tracker hit"
		  << std::endl;
	++nTestsFailed;
	continue;	
      }

      if (gearTPCHit->GetBfield() != 1.)
      {
	std::cout << "KalTrackingHelixTester::testCreateGearHit(): Test failed: hit in row " 
		  << row << " at angle " << angleZIter->first  << " has the wrong B-field."
		  << std::endl;
	++nTestsFailed;
	continue;	
      }      

      if (gearTPCHit->GetVdrift() != 79.)
      {
	std::cout << "KalTrackingHelixTester::testCreateGearHit(): Test failed: hit in row " 
		  << row << " at angle " << angleZIter->first  << " has the wron drift velocity."
		  << std::endl;
	++nTestsFailed;
	continue;	
      }      

    }// for agleZIter
    
  }// for row

  std::cout << "KalTrackingHelixTester::testCreateGearHit(): "
	    << nTestsFailed << " out of " << nTestsTotal << " tests failed." << std::endl;
  return std::pair<unsigned int, unsigned int>(nTestsFailed,nTestsTotal);
}

std::pair<unsigned int, unsigned int> KalTrackingHelixTester::testToBaseRange()
{
  unsigned int nSubTestsFailed = 0;
  for (unsigned int i = 0; i < 100; ++i)
  {
    // the argument will go from -
    double argument = 0.4*i - 20.;

    double argumentInBaseRange = kalTrackingHelix->toBaseRange(argument);
    
    if (argumentInBaseRange > M_PI || argumentInBaseRange < -M_PI)
    {
      nSubTestsFailed++;
    }
  }

  std::cout << "KalTrackingHelixTester::testToBaseRange() " << nSubTestsFailed
	    <<" out of 100 sub test failed. " << std::endl;

  // we consider the whole thing as 1 test, it's not such a serious function
  return std::pair<unsigned int, unsigned int>( (nSubTestsFailed == 0 ? 0 : 1 ) ,1);
}

std::pair<unsigned int, unsigned int> KalTrackingHelixTester::runAllTests()
{
  int nTestsFailed =0;
  int nTestsTotal  =0;

  std::pair<unsigned int, unsigned int> testResult = testToBaseRange();
  nTestsFailed += testResult.first;
  nTestsTotal  += testResult.second;

  testResult = testCreateGearHit();
  nTestsFailed += testResult.first;
  nTestsTotal  += testResult.second;

  testResult = getNearestPointInNextLayer();
  nTestsFailed += testResult.first;
  nTestsTotal  += testResult.second;
 

  std::cout << "KalTrackingHelixTester::runAllTests(): " << nTestsFailed
	    << " out of " << nTestsTotal << " tests failed." << std::endl;
  return std::pair<int, int>(nTestsFailed, nTestsTotal);
}    


}//namespace
