#include "PulseFinderTester.h"
// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

#include <cmath>

namespace marlintpc
{

  // the constructor. prepares the PulseFinder
  PulseFinderTester::PulseFinderTester()
  {
    // create an instance of the PulseFinder with the following parameters:
    // minimumPulseHeight = 12 ADC counts
    // minimumPulseLength = 5 ADC counts
    // pulseStartPedestalWidthFactor = 3 (three times the pedestal RMS)
    // pulseStartPedestalWidthFactor = 2 (two times the pedestal RMS)
    // pulseStartPedestalWidthFactor = 3 (three times the pedestal RMS)

    // pedestals and electronics parameter are fed in via override parameters. No need to register
    // a listener (listeners are not tested!)
    // Pedestals:
    // pedestalValue 2.4
    // pedestalWidth 1.1

    // 10 bit ADC => maxADCValue = 1023
    // polarity positive // FIXME: also test negative polarity
    // pedestal is subtracted // FIXME: also test non-subtracted data
    // readout frequency 40 MHz
    
    _pulseFinder = new PulseFinder(12, //minPulseHeight
				   5,  //minPulseLength
				   3,  //pulseStartPedestalWidthFactor
				   2,  //pulseEndPedestalWidthFactor
				   3,  //noiseValuePedestalWidthFactor
				   NULL, //pedestalListener
				   NULL); // adcElectronicsListener
    // default values are used for
    //  saveNBinsBeforeStart = 0,
    // int saveNBinsAfterEnd = 0,
    // bool useFitTime = false,
    // bool useFitCharge = false
			
    _pulseFinderPrePostPlateau = new PulseFinder(12, //minPulseHeight
						 5,  //minPulseLength
						 3,  //pulseStartPedestalWidthFactor
						 2,  //pulseEndPedestalWidthFactor
						 3,  //noiseValuePedestalWidthFactor
						 NULL, //pedestalListener
						 NULL, // adcElectronicsListener
						 2,  // saveNBinsBeforeStart
						 3); // saveNBinsAfterEnd

    _pulseFinderPrePostPlateau->setPlateauCutOff(4);// max 4 samples at the same height

    _readoutFrequency=40.;
    
    // As the listeners are not used we have to set the override values
    _pulseFinder->setPedestalValueOverride(2.4);
    _pulseFinderPrePostPlateau->setPedestalValueOverride(2.4);

    _pulseFinder->setPedestalWidthOverride(1.1);
    _pulseFinderPrePostPlateau->setPedestalWidthOverride(1.1);
  
    _pulseFinder->setMaxADCValueOverride(1023);
    _pulseFinderPrePostPlateau->setMaxADCValueOverride(1023);

    _pulseFinder->setReadoutFrequencyOverride(_readoutFrequency);
    _pulseFinderPrePostPlateau->setReadoutFrequencyOverride(_readoutFrequency);

    _pulseFinder->setPolarityOverride(1);
    _pulseFinderPrePostPlateau->setPolarityOverride(1);

    _pulseFinder->set_isPedestalSubtractedOverride(true);
    _pulseFinderPrePostPlateau->set_isPedestalSubtractedOverride(true);
    

  }
  
  PulseFinderTester::~PulseFinderTester()
  {
    delete _pulseFinder;
    delete _pulseFinderPrePostPlateau;
  }  



  //---------------------------------------------------------
  //---------------------------------------------------------
  // --------------------NOW WE START WITH THE TEST----------
  //---------------------------------------------------------
  //---------------------------------------------------------
 

 int PulseFinderTester::runAllTests()
  {
    // the number of failed tests. Should be 0.
    int nTestsFailed = 0;

    // test the individual member functions with tuned pulses
  // The thresholds translate to (pedestal is subtracted, so it's factor*width
  // startThreshold = 3.3 
  // end threshold = 2.2
  
  //---------------------------------------------------------
  // -----------------prepare some pulses-------------------
  //---------------------------------------------------------

  // the ints have been corrected by 2.4 as pedestal

  // pluse is ok, 8 samples should be in it
  float data1[] = {0.6,-0.4,1.6,-0.4,4.6,25.6,75.6,50.6,30.6,15.6,5.6, 2.6, 1.6, -0.4, 0.6};
  EVENT::FloatVec dataVec1(data1,data1 + sizeof(data1) / sizeof(float));
  IMPL::TrackerDataImpl trackerData1;
  trackerData1.setTime(10); // 10 adc samples time offset
  trackerData1.setChargeValues(dataVec1);

  // pulse is not high enough, although 5 samples are above the start/end thresholds
  float data2[] = {0.6,-0.4,1.6,-0.4,6.6,10.6,8.6,5.6, 2.6, 1.6, -0.4, 0.6};
  EVENT::FloatVec dataVec2(data2,data2 + sizeof(data2) / sizeof(float));
  IMPL::TrackerDataImpl trackerData2;
  trackerData2.setTime(10); // 10 adc samples time offset
  trackerData2.setChargeValues(dataVec2);
  
  // pulse is too short
  float data3[] = {0.6,-0.4,1.6,-0.4,10.6,18.6,8.6, 3.6, 1.6, -0.4, 0.6};
  EVENT::FloatVec dataVec3(data3,data3 + sizeof(data3) / sizeof(float));
  IMPL::TrackerDataImpl trackerData3;
  trackerData3.setTime(10); // 10 adc samples time offset
  trackerData3.setChargeValues(dataVec3);
  
  // pulse is out of range 
  // note: the value has not been pedestal subtracted (at least the offine version does it 
  // like this)
  float data4[] = {0.6,-0.4,1.6,-0.4,175.6,750.6,1023,834.6,565.6,122.6,1.6, -0.4, 0.6};
  EVENT::FloatVec dataVec4(data4,data4 + sizeof(data4) / sizeof(float));
  IMPL::TrackerDataImpl trackerData4;
  trackerData4.setTime(10); // 10 adc samples time offset
  trackerData4.setChargeValues(dataVec4);

  // two pulses in the data. The last sample is above the threshold, so the last result iterator
  // is end().
  float data5[] = {0.6,-0.4,1.6,-0.4,4.6,25.6,75.6,50.6,30.6,15.6,5.6, 2.6, 1.6, -0.4, 0.6,
		   0.6,-0.4,1.6,-0.4,4.6,25.6,75.6,50.6,30.6,15.6,5.6, 3.6};
  EVENT::FloatVec dataVec5(data5,data5 + sizeof(data5) / sizeof(float));
  IMPL::TrackerDataImpl trackerData5;
  trackerData5.setTime(10); // 10 adc samples time offset
  trackerData5.setChargeValues(dataVec5);

  // a double pulse
  float data6[] = {1.6,-0.4,4.6,25.6,75.6,50.6,30.6,15.6,25.6,75.6,50.6,30.6,15.6,
		  5.6, 2.6, 1.6, -0.4, 0.6};
  EVENT::FloatVec dataVec6(data6,data6 + sizeof(data6) / sizeof(float));
  IMPL::TrackerDataImpl trackerData6;
  trackerData6.setTime(10); // 10 adc samples time offset
  trackerData6.setChargeValues(dataVec6);

  // a pulse with a long plateau
  float data7[] = {0.6,1.6,-0.4,4.6,25.6,75.6,50.6,30.6,15.6,15.6,
		   14.6,15.6,14.6,14.6,15.6,14.6,13.6,5.6, 2.6, 1.6, -0.4, 0.6,1.6};
  EVENT::FloatVec dataVec7(data7,data7 + sizeof(data7) / sizeof(float));
  IMPL::TrackerDataImpl trackerData7;
  trackerData7.setTime(10); // 10 adc samples time offset
  trackerData7.setChargeValues(dataVec7);

  // Out of range pulse with plateau.
  // Usually the out of range also is a plateau.
  // This must be excluded from the cutting.
  float data8[] = {0.6,-0.4,1.6,-0.4,175.6,750.6,1023,1023, 1023, 1023, 1023, 1023,
		   834.6,565.6,122.6,124.6, // intenionally have the sample two ADC counts
		   // rising after the steep falling edge to test the signal splitter
		   122.6,123.6,124.6,122.6,123.6,124.6,122.6,123.6,124.6};
  EVENT::FloatVec dataVec8(data8,data8 + sizeof(data8) / sizeof(float));
  IMPL::TrackerDataImpl trackerData8;
  trackerData8.setTime(10); // 10 adc samples time offset
  trackerData8.setChargeValues(dataVec8);
  
  // pulse with plateau and second pulse on the plateau
  // a pulse with a long plateau
  float data9[] = {0.6,1.6,-0.4,4.6,25.6,75.6,50.6,30.6,15.6,15.6,
		   14.6,15.6,14.6,14.6,15.6,14.6,13.6,14.6,13.6,14.6,13.6,13.6,14.6,
		   35.6,125.6,75.6,45.6,15.6,
		   14.6,15.6,14.6,14.6,15.6,5.6, 2.6, 1.6, -0.4, 0.6,1.6};
  EVENT::FloatVec dataVec9(data9,data9 + sizeof(data9) / sizeof(float));
  IMPL::TrackerDataImpl trackerData9;
  trackerData9.setTime(10); // 10 adc samples time offset
  trackerData9.setChargeValues(dataVec9);
  
  // pulse with a plateau in the rising edge, should be split
  float data10[] = {5, 15, 25, 25, 25, 25, 35, 45, 35, 25, 15, 5};
  EVENT::FloatVec dataVec10(data10,data10 + sizeof(data10) / sizeof(float));
  IMPL::TrackerDataImpl trackerData10;
  trackerData10.setTime(10); // 10 adc samples time offset
  trackerData10.setChargeValues(dataVec10);

  // pulse with a slow rising part and one small dip, should not be split
  float data11[] = {5, 15, 25, 26, 27, 28, 35, 45, 44, 55, 45, 35, 25, 15, 5};
  EVENT::FloatVec dataVec11(data11,data11 + sizeof(data11) / sizeof(float));
  IMPL::TrackerDataImpl trackerData11;
  trackerData11.setTime(10); // 10 adc samples time offset
  trackerData11.setChargeValues(dataVec11);

  // pulse with a slow rising part at the beginning, and tree different plateaus in the rising edge 
  // should not be split 3 times
  float data12[] = {5, 6, 7, 15, 25, 26, 26, 35, 45, 44, 45, 55, 65, 65, 66, 67, 68, 88, 95, 75, 60, 45, 35, 25, 15, 5};
  EVENT::FloatVec dataVec12(data12,data12 + sizeof(data12) / sizeof(float));
  IMPL::TrackerDataImpl trackerData12;
  trackerData12.setTime(10); // 10 adc samples time offset
  trackerData12.setChargeValues(dataVec12);

  // perform the tests and count the number of failed tests
  
   //------------------------------------------------------------------
  //------------ first check the output of the brutePulseSearch--------
  //-------------------------------------------------------------------

   std::vector<std::pair<std::vector<float>::const_iterator, std::vector<float>::const_iterator> > 
     bruteSearchResult;
   
   // trackerData1
   _pulseFinder->brutePulseSearch(&trackerData1, &bruteSearchResult);
   // one pulse should have been returned, the first iterator pointing to entry nr 4 (value 4.6)
   // the second iterator pointing to entry nr 12 (value 1.6 = first bin below the threshold
   std::vector<IteratorProperties> referenceVec;
   referenceVec.push_back( IteratorProperties(4,12,4.6,1.6));
   if ( evaluateBruteSearchResult( bruteSearchResult, referenceVec,
			       trackerData1.getChargeValues().begin(),
			       trackerData1.getChargeValues().end()) == false)
     {
       std::cout << "brutePulseSearch on trackerData1 failed." << std::endl<< std::endl;
       nTestsFailed++;
     }

   //trackerData2
   bruteSearchResult.clear();
   referenceVec.clear();
   // this one should be empty
   _pulseFinder->brutePulseSearch(&trackerData2, &bruteSearchResult);
   if ( evaluateBruteSearchResult( bruteSearchResult, referenceVec,
			       trackerData2.getChargeValues().begin(),
			       trackerData2.getChargeValues().end()) == false)
     {
       std::cout << "brutePulseSearch on trackerData2 failed." << std::endl<< std::endl;
       nTestsFailed++;
     }
  
   // trackerData3
   bruteSearchResult.clear();
   referenceVec.clear();
   // this one should be empty
   _pulseFinder->brutePulseSearch(&trackerData3, &bruteSearchResult);
   if ( evaluateBruteSearchResult( bruteSearchResult, referenceVec,
			       trackerData3.getChargeValues().begin(),
			       trackerData3.getChargeValues().end()) == false)
     {
       std::cout << "brutePulseSearch on trackerData3 failed." << std::endl<< std::endl;
       nTestsFailed++;
     }
 
   // trackerData4
   bruteSearchResult.clear();
   referenceVec.clear();
   referenceVec.push_back( IteratorProperties(4,10,175.6,1.6));
   _pulseFinder->brutePulseSearch(&trackerData4, &bruteSearchResult);
   if ( evaluateBruteSearchResult( bruteSearchResult, referenceVec,
			       trackerData4.getChargeValues().begin(),
			       trackerData4.getChargeValues().end()) == false)
     {
       std::cout << "brutePulseSearch on trackerData4 failed." << std::endl<< std::endl;
       nTestsFailed++;
     }
 
  // trackerData5
   bruteSearchResult.clear();
   referenceVec.clear();
   referenceVec.push_back( IteratorProperties(4,12,4.6,1.6));
   referenceVec.push_back( IteratorProperties(19,27,4.6,0)); // the last value is irrelevant
   // because the iterator should be end(), with no valid value
   _pulseFinder->brutePulseSearch(&trackerData5, &bruteSearchResult);
   if ( evaluateBruteSearchResult( bruteSearchResult, referenceVec,
			       trackerData5.getChargeValues().begin(),
			       trackerData5.getChargeValues().end()) == false)
     {
       std::cout << "brutePulseSearch on trackerData5 failed." << std::endl<< std::endl;
       nTestsFailed++;
     }
 
  // trackerData6
   bruteSearchResult.clear();
   referenceVec.clear();
   referenceVec.push_back( IteratorProperties(2,15,4.6,1.6));
   _pulseFinder->brutePulseSearch(&trackerData6, &bruteSearchResult);
   if ( evaluateBruteSearchResult( bruteSearchResult, referenceVec,
			       trackerData6.getChargeValues().begin(),
			       trackerData6.getChargeValues().end()) == false)
     {
       std::cout << "brutePulseSearch on trackerData6 failed." << std::endl<< std::endl;
       nTestsFailed++;
     }
 
   //------------------------------------------------------------------
   //------------ TEST SPLITING-----------------------------------------
   //-------------------------------------------------------------------
   // continue with the splitting test. we use pulse 1 and 6 
   // the noise threshold for channel 0, module 0, as set in the TrackerData. Is dummy anyway since
   // we use the override parameters
   float noiseThreshold = _pulseFinder->getPedestalWidth(0,0) * 
                          _pulseFinder->_noiseValuePedestalWidthFactor;
   
   // trackerData1
   bruteSearchResult.clear();
   _pulseFinder->brutePulseSearch(&trackerData1, &bruteSearchResult);
   std::vector<PulseFinder::InternalSplitPulse> splitResult = 
     _pulseFinder->splitMultiplePulses( &bruteSearchResult, noiseThreshold);
   referenceVec.clear();
   // after splitting the resulst should still be the same
   referenceVec.push_back( IteratorProperties(4,12,4.6,1.6,false));
   if ( evaluateSplitResult(splitResult, referenceVec,
			trackerData1.getChargeValues().begin(),
			trackerData1.getChargeValues().end() ) == false)
   {
     std::cout << "splitMultiplePulses on trackerData1 failed." << std::endl<< std::endl;
     nTestsFailed++;    
   }
  
   // trackerData6
   bruteSearchResult.clear();
   _pulseFinder->brutePulseSearch(&trackerData6, &bruteSearchResult);
   splitResult = _pulseFinder->splitMultiplePulses( &bruteSearchResult, noiseThreshold);
   referenceVec.clear();
   // after splitting there should be two pulses. The second one starts with the first
   // bin after the minimum (which is above the noise threshold, which here is the case for 
   // the first bin on the second rising edge). Now the split flag should be set.
   referenceVec.push_back( IteratorProperties(2,8,4.6,25.6,true));
   referenceVec.push_back( IteratorProperties(8,15,25.6,1.6,true));
   if ( evaluateSplitResult(splitResult, referenceVec,
			trackerData6.getChargeValues().begin(),
			trackerData6.getChargeValues().end() ) == false)
   {
     std::cout << "splitMultiplePulses on trackerData6 failed." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   //trackerData10
   bruteSearchResult.clear();
   _pulseFinder->brutePulseSearch(&trackerData10, &bruteSearchResult);
   splitResult = _pulseFinder->splitMultiplePulses( &bruteSearchResult, noiseThreshold);
   referenceVec.clear();
   // we expect splitting before the samle with value 35 (rising again)
   referenceVec.push_back( IteratorProperties(0,6,5,35,true));
   referenceVec.push_back( IteratorProperties(6,12,35,5,true));
   if ( evaluateSplitResult(splitResult, referenceVec,
			trackerData10.getChargeValues().begin(),
			trackerData10.getChargeValues().end() ) == false)
   {
     std::cout << "splitMultiplePulses on trackerData10 failed." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   //trackerData11
   bruteSearchResult.clear();
   _pulseFinder->brutePulseSearch(&trackerData11, &bruteSearchResult);
   splitResult = _pulseFinder->splitMultiplePulses( &bruteSearchResult, noiseThreshold);
   referenceVec.clear();
   // we expect splitting before the samle with value 35 (rising again)
   referenceVec.push_back( IteratorProperties(0,15,5,5,false));
   if ( evaluateSplitResult(splitResult, referenceVec,
			trackerData11.getChargeValues().begin(),
			trackerData11.getChargeValues().end() ) == false)
   {
     std::cout << "splitMultiplePulses on trackerData11 failed." << std::endl<< std::endl;
     nTestsFailed++;    
   }
  
   //trackerData12
   bruteSearchResult.clear();
   _pulseFinder->brutePulseSearch(&trackerData12, &bruteSearchResult);
   splitResult = _pulseFinder->splitMultiplePulses( &bruteSearchResult, noiseThreshold);
   referenceVec.clear();
   // we expect splitting before the samle with value 35 (rising again)
   referenceVec.push_back( IteratorProperties(0,7,5,35,true));
   referenceVec.push_back( IteratorProperties(7,11,35,55,true));
   referenceVec.push_back( IteratorProperties(11,17,55,88,true));
   referenceVec.push_back( IteratorProperties(17,26,88,5,true));
   if ( evaluateSplitResult(splitResult, referenceVec,
			trackerData12.getChargeValues().begin(),
			trackerData12.getChargeValues().end() ) == false)
   {
     std::cout << "splitMultiplePulses on trackerData12 failed." << std::endl<< std::endl;
     nTestsFailed++;    
   }
  
   //------------------------------------------------------------------
   //------------ test the findPulseCandidates function------------------
   //-------------------------------------------------------------------

   // We use pulse1 (normal single pulse),
   // pulse4 (out of range), pulse 5 (separated double pulse) and pulse 6 (has to be split)

   // trackerData1
   std::vector<std::pair<lcio::TrackerDataImpl*, int> > pulseCandidates;
   _pulseFinder->findPulseCandidates(&trackerData1, &pulseCandidates);
   
   std::vector<CandidateProperties> referenceCandidateProperties;
   // nSamples = 8, firstSample = 4.6, lastSample = 2.6, time=14. (sample 4 in TrackerData 
   // with time = 10), quality = 0
   referenceCandidateProperties.push_back( CandidateProperties(8, 4.6, 2.6, 14, 0 ) );

   if ( evaluatePulseCandidate(pulseCandidates, referenceCandidateProperties) == false )
   {
      std::cout << "findPulseCandidates on trackerData1 failed." << std::endl<< std::endl;
      nTestsFailed++;        
   }
   // clean up the allocated TrackerData objects
   for (std::vector<std::pair<lcio::TrackerDataImpl*, int> >::iterator it = pulseCandidates.begin();
	it != pulseCandidates.begin(); ++it )
   {
     delete it->first;
   }
   pulseCandidates.clear();
   referenceCandidateProperties.clear();

   // trackerData4
   _pulseFinder->findPulseCandidates(&trackerData4, &pulseCandidates);
   referenceCandidateProperties.push_back( CandidateProperties(6, 175.6, 122.6, 14,
							       pulseflag::OVERFLOW_BIN_BIT ) );
   if ( evaluatePulseCandidate(pulseCandidates, referenceCandidateProperties) == false )
   {
      std::cout << "findPulseCandidates on trackerData4 failed." << std::endl<< std::endl;
      nTestsFailed++;        
   }
   // clean up the allocated TrackerData objects
   for (std::vector<std::pair<lcio::TrackerDataImpl*, int> >::iterator it = pulseCandidates.begin();
	it != pulseCandidates.begin(); ++it )
   {
     delete it->first;
   }
   pulseCandidates.clear();
   referenceCandidateProperties.clear();
  
   // trackerData5
   _pulseFinder->findPulseCandidates(&trackerData5, &pulseCandidates);
   referenceCandidateProperties.push_back( CandidateProperties(8, 4.6, 2.6, 14, 0 ) );
   referenceCandidateProperties.push_back( CandidateProperties(8, 4.6, 3.6, 29, 0 ) );
   if ( evaluatePulseCandidate(pulseCandidates, referenceCandidateProperties) == false )
   {
      std::cout << "findPulseCandidates on trackerData5 failed." << std::endl<< std::endl;
      nTestsFailed++;        
   }
   // clean up the allocated TrackerData objects
   for (std::vector<std::pair<lcio::TrackerDataImpl*, int> >::iterator it = pulseCandidates.begin();
	it != pulseCandidates.begin(); ++it )
   {
     delete it->first;
   }
   pulseCandidates.clear();
   referenceCandidateProperties.clear();

   // trackerData6
   _pulseFinder->findPulseCandidates(&trackerData6, &pulseCandidates);
   referenceCandidateProperties.push_back( CandidateProperties(6, 4.6, 15.6, 12,
							       pulseflag::WAS_SPLIT_BIT ) );
   referenceCandidateProperties.push_back( CandidateProperties(7, 25.6, 2.6, 18,
							       pulseflag::WAS_SPLIT_BIT ) );
   if ( evaluatePulseCandidate(pulseCandidates, referenceCandidateProperties) == false )
   {
      std::cout << "findPulseCandidates on trackerData6 failed." << std::endl<< std::endl;
      nTestsFailed++;        
   }
   // clean up the allocated TrackerData objects
   for (std::vector<std::pair<lcio::TrackerDataImpl*, int> >::iterator it = pulseCandidates.begin();
	it != pulseCandidates.begin(); ++it )
   {
     delete it->first;
   }
   pulseCandidates.clear();
   referenceCandidateProperties.clear();

   // some data where it's easy to calculate the values manually
   float calculationTestData[] = {5, 25, 125, 180, 125, 25, 5}; // nicely symmetrical
   EVENT::FloatVec calculationTestDataVec(calculationTestData,
					  calculationTestData +
					  sizeof(calculationTestData) / sizeof(float));
   IMPL::TrackerDataImpl calculationTestTrackerData;
   calculationTestTrackerData.setTime(5); // 5 adc samples time offset
   calculationTestTrackerData.setChargeValues(calculationTestDataVec);
  
   // test the time. With the derivative method it should be sum( q*t ) / sum(q) =
   // (20*1 + 100*2 + 55*3) / (20+100+55) + 5 time slices
   float meanBin = ( (20.*1 + 100.*2 + 55.*3) / (20.+100.+55.) + 5. ); //intermediate result   
   float referenceTimeDerivativeMethod = 
     meanBin* 1E3 / _readoutFrequency; //conversion to ns
   
   /** This is the variance, which is not the error. No idea how to calculate it correctly
   float referenceTimeErrorDerivativeMethod = sqrt( 20*20*(6.-meanBin)*(6.-meanBin)
						    + 100*100*(7.-meanBin)*(7.-meanBin)
						    + 55*55*(8.-meanBin)*(8.-meanBin) ) 
     / (20.+100.+55.)*  1E3 / _readoutFrequency; //conversion to ns
   */
   //maximum at 180 in bin=3 plus the offset of 5 and then conversion to ns
   float referenceTimeAtMaximum=8.* 1E3 / _readoutFrequency;

   //first bin above is already 0 plus the offset of 5 and then conversion to ns
   float referenceTimeFirstBinAboveThreshold=5.* 1E3 / _readoutFrequency;
 
   //center of gravity in bin 3 plus 5 offset
   float referenceTimeCentreOfGravity=8.* 1E3 / _readoutFrequency;
 
   PulseFinder::PulseTime pulseTimeInflexionPoint = 
     _pulseFinder->calcPulseTimeInflexionPoint(&calculationTestTrackerData);

   // allow for numerical uncertainties. 1e-3 ns = 1 ps should be ok as tolerance
   if ( fabs(pulseTimeInflexionPoint.getTime() - referenceTimeDerivativeMethod ) > 1e-3 )
   {
     std::cout << "calcPulseTimeInflexionPoint  test failed in time calculation." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   PulseFinder::PulseTime pulseTimeAtMaximum = 
     _pulseFinder->calcPulseTimeAtMaximum(&calculationTestTrackerData);
   // allow for numerical uncertainties. 1e-3 ns = 1 ps should be ok as tolerance
   if ( fabs(pulseTimeAtMaximum.getTime() - referenceTimeAtMaximum ) > 1e-3 )
   {
     std::cout << "calcPulseTimeAtMaximum test failed in time calculation." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   PulseFinder::PulseTime pulseTimeFirstBinAboveThreshold = 
     _pulseFinder->calcPulseTimeFirstBinAboveThreshold(&calculationTestTrackerData);
     // allow for numerical uncertainties. 1e-3 ns = 1 ps should be ok as tolerance
   if ( fabs(pulseTimeFirstBinAboveThreshold.getTime() - referenceTimeFirstBinAboveThreshold ) > 1e-3 )
   {
     std::cout << "calcPulseTimeFirstBinAboveThreshold test failed in time calculation." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   PulseFinder::PulseTime pulseTimeCentreOfGravity = 
     _pulseFinder->calcPulseTimeCentreOfGravity(&calculationTestTrackerData);
   // allow for numerical uncertainties. 1e-3 ns = 1 ps should be ok as tolerance
   if ( fabs(pulseTimeCentreOfGravity.getTime() - referenceTimeCentreOfGravity ) > 1e-3 )
   {
     std::cout << "calcPulseTimeCentreOfGravity test failed in time calculation." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   /// @todo:  PulseFinderTester: calculate correct time error and test it
   
   PulseFinder::PulseCharge pulseChargeSumQ = 
     _pulseFinder->calcPulseChargeSumQ(&calculationTestTrackerData);
   float referenceCharge= 5 + 25 + 125 + 180 + 125 + 25 + 5;

   // allow for numerical uncertainties. 1e-3 is well below 1 adc count.
   if ( fabs(pulseChargeSumQ.getCharge() - referenceCharge ) > 1e-3 )
   {
     std::cout << "calcPulseChargeSumQ test failed in charge calculation." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   /// @todo: PulseFinderTester: calculate correct charge error and test it

   PulseFinder::PulseCharge pulseChargeMaxQ = 
     _pulseFinder->calcPulseChargeMaxQ(&calculationTestTrackerData);

   if ( fabs(pulseChargeMaxQ.getCharge() !=  180 ) )
   {
     std::cout << "calcPulseChargeMaxQ test failed in charge calculation." << std::endl<< std::endl;
     nTestsFailed++;    
   }
  
   

   PulseFinder::Pulse fittedPulse = _pulseFinder->calculatePulseWithFit(&calculationTestTrackerData);

   // As the pulse is symmetrical the fitted time should be in the middle of the 4th bin
   // As we start from 0, with 5 bins offset, it is 3.5 + 5 = 8.5 time samples
   float referenceTimeFitMethod = 8.5 * 1E3 / _readoutFrequency; // in ns

   // allow for numerical uncertainties. 1e-3 ns = 1 ps should be ok as tolerance
   if ( fabs(fittedPulse.getTime() - referenceTimeFitMethod ) > 1e-3 )
   {
     std::cout << "calculatePulseWithFit test failed in time calculation." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   // the fitted charge depends on the pulse shape. As the shape is not a parabola one has to 
   // calculate the expected value for the minimisation. To be implemented...
   /// @todo:  PulseFinderTester: calculate expected charge fit value and test it

   // finally test the only public function: findPulses
   std::vector<TrackerPulse *> foundPulses = _pulseFinder->findPulses(&trackerData5);
   // we test with pulse 5 and 6. Pulse 5 should deliver two pulses without quality flag,
   // pulse 6 with the split flag.
   // As forceSpectrumSave is off, pulse 5 should not have TrackerData objects

   // the mean of the derivatves in the rising edge of the pulse
   float timeBinInPulse = (21.*1 + 50.*2)/(21.+50);

   std::vector<PulseProperties> referencePulseProperties;
   // nSamples, quality, time, charge
   // nSamples is 0 because the storeQuality flag is off, so there are no samples
   referencePulseProperties.push_back( PulseProperties(0,0,
		 (timeBinInPulse+14) * 1E3 / _readoutFrequency,
		 4.6+25.6+75.6+50.6+30.6+15.6+5.6+ 2.6 ) );
   referencePulseProperties.push_back( PulseProperties(0,0,
		 (timeBinInPulse+29) * 1E3 / _readoutFrequency,
		 4.6+25.6+75.6+50.6+30.6+15.6+5.6+ 3.6 ) );
   if (evaluatePulses(foundPulses, referencePulseProperties) == false)
   {
     std::cout << "findPulses test failed on trackerData5." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   for ( std::vector<TrackerPulse *>::iterator pulseIter = foundPulses.begin();
	 pulseIter != foundPulses.end(); ++pulseIter )
   {
     delete (*pulseIter)->getTrackerData();
     delete *pulseIter;
   }
   foundPulses.clear();
   referencePulseProperties.clear();

   // trackerData6
    foundPulses = _pulseFinder->findPulses(&trackerData6);
   // 8 samples should be above the threshold
   referencePulseProperties.push_back( PulseProperties(6,pulseflag::WAS_SPLIT_BIT,
		 (timeBinInPulse+12) * 1E3 / _readoutFrequency,
		 4.6+25.6+75.6+50.6+30.6+15.6) );
   referencePulseProperties.push_back( PulseProperties(7,pulseflag::WAS_SPLIT_BIT,
		 19 * 1E3 / _readoutFrequency,
		 25.6+75.6+50.6+30.6+15.6+5.6+2.6) );
   if (evaluatePulses(foundPulses, referencePulseProperties) == false)
   {
     std::cout << "findPulses test failed on trackerData6." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   for ( std::vector<TrackerPulse *>::iterator pulseIter = foundPulses.begin();
	 pulseIter != foundPulses.end(); ++pulseIter )
   {
     delete (*pulseIter)->getTrackerData();
     delete *pulseIter;
   }
   foundPulses.clear();
   referencePulseProperties.clear();

  // now test trackerData1 with forcing the spectrum save:
   _pulseFinder->setForceSpectrumSave();

   foundPulses = _pulseFinder->findPulses(&trackerData1);
   // 8 samples should be above the threshold
   referencePulseProperties.push_back( PulseProperties(8,0,
		 (timeBinInPulse+14) * 1E3 / _readoutFrequency,
		 4.6+25.6+75.6+50.6+30.6+15.6+5.6+ 2.6 ) );
   if (evaluatePulses(foundPulses, referencePulseProperties) == false)
   {
     std::cout << "findPulses test failed on trackerData1." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   for ( std::vector<TrackerPulse *>::iterator pulseIter = foundPulses.begin();
	 pulseIter != foundPulses.end(); ++pulseIter )
   {
     delete (*pulseIter)->getTrackerData();
     delete *pulseIter;
   }
   foundPulses.clear();
   referencePulseProperties.clear();

   //test with plateau cutoff, 2 pre and 3 post samples
   foundPulses = _pulseFinderPrePostPlateau->findPulses(&trackerData7);
   std::vector< std::pair<size_t, size_t> > referencePositions;
   referencePositions.push_back( std::pair<size_t, size_t>(1, 12) );
   if (evaluatePlateauCutoff(foundPulses, referencePositions , dataVec7 ) == false)
   {
     std::cout << "evaluatePlateauCutOff failed on trackerData7." << std::endl<< std::endl;
     nTestsFailed++;    
   }

   for ( std::vector<TrackerPulse *>::iterator pulseIter = foundPulses.begin();
	 pulseIter != foundPulses.end(); ++pulseIter )
   {
     delete (*pulseIter)->getTrackerData();
     delete *pulseIter;
   }
   foundPulses.clear();
   referencePositions.clear();

   //test with plateau cutoff, 2 pre and 3 post samples
   foundPulses = _pulseFinderPrePostPlateau->findPulses(&trackerData8);
   referencePositions.push_back( std::pair<size_t, size_t>(2, 18) );
   if (evaluatePlateauCutoff(foundPulses, referencePositions , dataVec8 ) == false)
   {
     std::cout << "evaluatePlateauCutOff failed on trackerData8." << std::endl;
     nTestsFailed++;    
   }
   else
   {
     // Spectrum ok so far. Test the quality flag (currently not testing charge and time)
     // The outOfRange and the anomalousShape flags should be on.
     // As the previous test succeeded there is one pulse, so directly indexing is safe.
     if (foundPulses[0]->getQuality() != 
	 (pulseflag::PLATEAU_CUTOFF_BIT |  pulseflag::OVERFLOW_BIN_BIT) )
     {
       
       std::cout << "QualityFlag wrong on trackerData8." 
		 << " It is "<< foundPulses[0]->getQuality()
		 << ", should be " 
		 << (pulseflag::PLATEAU_CUTOFF_BIT |  pulseflag::OVERFLOW_BIN_BIT)
		 << std::endl;
       nTestsFailed++;    
     }
     
   }

   for ( std::vector<TrackerPulse *>::iterator pulseIter = foundPulses.begin();
	 pulseIter != foundPulses.end(); ++pulseIter )
   {
     delete (*pulseIter)->getTrackerData();
     delete *pulseIter;
   }
   foundPulses.clear();
   referencePositions.clear();

    //test with plateau cutoff, 2 pre and 3 post samples
   foundPulses = _pulseFinderPrePostPlateau->findPulses(&trackerData9);
   referencePositions.push_back( std::pair<size_t, size_t>(1, 12) );
   referencePositions.push_back( std::pair<size_t, size_t>(23, 31) );
   if (evaluatePlateauCutoff(foundPulses, referencePositions , dataVec9 ) == false)
   {
     std::cout << "evaluatePlateauCutOff failed on trackerData9." << std::endl;
     nTestsFailed++;    
   }
   else
   {
     // Spectrum ok so far. Test the quality flag (currently not testing charge and time)
     // The wasSPlit and the anomalousShape flags should be on.
     // In addition the multiPulseCandidate flag is on because the distance calculation
     // is done before cutting the plateau (which makes sense because there is signal, which
     // is only artificially shortened, so the pulses in fact overlap)
     for (size_t i = 0; i < foundPulses.size(); i++)
     {
       if (foundPulses[i]->getQuality() != 
	   (pulseflag::MULTIPLE_PULSE_CANDIDATE_BIT |
	    pulseflag::PLATEAU_CUTOFF_BIT |
	    pulseflag::WAS_SPLIT_BIT) )
       {
	   
	 std::cout << "QualityFlag wrong on trackerData09, pulse" <<i
		   << ". It is "<< foundPulses[i]->getQuality()
		   << ", should be " 
		   << ( pulseflag::MULTIPLE_PULSE_CANDIDATE_BIT |
			pulseflag::PLATEAU_CUTOFF_BIT |  pulseflag::WAS_SPLIT_BIT)
		   << std::endl;
	 nTestsFailed++;           
       }
     }
   }

   for ( std::vector<TrackerPulse *>::iterator pulseIter = foundPulses.begin();
	 pulseIter != foundPulses.end(); ++pulseIter )
   {
     delete (*pulseIter)->getTrackerData();
     delete *pulseIter;
   }
   foundPulses.clear();
   referencePositions.clear();

   

   return nTestsFailed;
  }



  // --------------------HELPER FUNCTIONS-------------------------
  
  /** this function gets the search result of the brute search and compares it to the 
   *  reference values given. Returns true if all values are ok.
   */
  bool PulseFinderTester::evaluateBruteSearchResult( std::vector< std::pair<std::vector<float>::const_iterator, 
									std::vector<float>::const_iterator> > const & bruteSearchResult,
			      std::vector<IteratorProperties> const & referenceValues ,
			      std::vector<float>::const_iterator beginIterator,
			      std::vector<float>::const_iterator endIterator )
  {
    // several sub tests
    
    // number of pulses
    if (bruteSearchResult.size() != referenceValues.size()) 
    { 
	std::cout << "Number of pulses found is " << bruteSearchResult.size()
		  <<", should be "<< referenceValues.size() << std::endl;
	return false;
    }
    //    std::cout << "DEBUG: subtest ok" << std::endl;

    // pulse properties (are the iterators correkt?)
    for (size_t pulseIndex = 0; pulseIndex < bruteSearchResult.size() ; pulseIndex++)
    {

      if (std::distance(beginIterator, bruteSearchResult[pulseIndex].first) != 
	  referenceValues[pulseIndex].iterator1Distance)
      {
	std::cout << "first iterator of bruteSearchResult has distance " <<
	  std::distance(beginIterator, bruteSearchResult[pulseIndex].first)  
		  <<" to beginIterator, should be "
		  << referenceValues[pulseIndex].iterator1Distance <<  std::endl;
	return false;
      }
      //      std::cout << "DEBUG: subtest ok" << std::endl;

      if (std::distance(beginIterator, bruteSearchResult[pulseIndex].second) != 
	  referenceValues[pulseIndex].iterator2Distance)
      {
	std::cout << "second iterator of bruteSearchResult has distance " 
		  << std::distance(beginIterator, bruteSearchResult[pulseIndex].second)  
		  << " to beginIterator, should be "
		  << referenceValues[pulseIndex].iterator2Distance <<  std::endl;
	return false;
      }
      //      std::cout << "DEBUG: subtest ok" << std::endl;
   
      // I am not testing for the first iterator to point to be the end iterator
      if ( *(bruteSearchResult[pulseIndex].first) != referenceValues[pulseIndex].iterator1Value )
      {
	std::cout << "value of first iterator is " << *(bruteSearchResult[pulseIndex].first)
		  << ", should be "<< referenceValues[pulseIndex].iterator1Value << std::endl;
	return false;
      }
      //      std::cout << "DEBUG: subtest ok" << std::endl;

      // the second iterator might be end(), in this case we cannot test the value
      if (std::distance(beginIterator, endIterator) ==  
	  static_cast<unsigned int>(referenceValues[pulseIndex].iterator2Distance))
      {
	// the iterator must be end
	if (bruteSearchResult[pulseIndex].second != endIterator) 
	{
	  std::cout << "iterator should be the end iterator, but it isn't." <<std::endl;
	  return false;	  
	}
	//	std::cout << "DEBUG: subtest is end is ok" << std::endl;

      }
      else // second iterator is inside the pulse (first value below threshold)
      {
	if ( *(bruteSearchResult[pulseIndex].second) != referenceValues[pulseIndex].iterator2Value ) 
	{
	  std::cout << "value of second iterator is " << *(bruteSearchResult[pulseIndex].second)
		    << ", should be "<< referenceValues[pulseIndex].iterator2Value << std::endl;
	  return false;
	}
	//	std::cout << "DEBUG: subtest ok" << std::endl;
      }
      
    }// for pulseIndex

    return true;
  }

  bool PulseFinderTester::evaluateSplitResult(
	  std::vector<PulseFinder::InternalSplitPulse>  const & splitResult,
	  std::vector<IteratorProperties> const & referenceValues ,
	  std::vector<float>::const_iterator beginIterator,
	  std::vector<float>::const_iterator endIterator )
  {
    // create a pair of iterators from the InternalSplitPulse and fill a vector.
    // now the result looks like a bruteSearch, except for the split flag
    std::vector< std::pair<std::vector<float>::const_iterator, 
			   std::vector<float>::const_iterator> > bruteSearchEquivalent;
    for (std::vector<PulseFinder::InternalSplitPulse>::const_iterator splitPulseIter = splitResult.begin();
	 splitPulseIter != splitResult.end() ; ++splitPulseIter)
    {
      bruteSearchEquivalent.push_back( std::make_pair( splitPulseIter->startBin, 
						       splitPulseIter->endBin) );
    }
    // now we can reuse the test for the brute search
    if (evaluateBruteSearchResult( bruteSearchEquivalent, referenceValues, beginIterator, endIterator) == false)
    {
      return false;
    }

    // Now test the split flag.
    // As evaluateBruteSearchResult we know that splitResult and referenceValues have the same size,
    // so this is safe:
    for (size_t i=0; i < splitResult.size(); i++)
    {
      if (splitResult[i].isSplit != referenceValues[i].isSplit)
      {
	  std::cout << "Split flag is wrong. It is " << splitResult[i].isSplit
		    << ", should be "<< referenceValues[i].isSplit << std::endl;
	  return false;	
      }
    }

    return true;
  }

  bool PulseFinderTester::evaluatePulseCandidate(
	       std::vector<std::pair<TrackerDataImpl *, int> > candidates,
	       std::vector<PulseFinderTester::CandidateProperties> const & referenceCandidateProperties)
  {
    if (candidates.size() != referenceCandidateProperties.size() )
    {
      std::cout << "number of cadidate pulses is " << candidates.size()
		<< ", should be "<<  referenceCandidateProperties.size() << std::endl;
      return false;
    }

    for (size_t i = 0; i < candidates.size(); i++)
    {
      TrackerDataImpl * trackerData = candidates[i].first;
      if (trackerData->getChargeValues().size() != referenceCandidateProperties[i].nSamples)
      {
	std::cout << "number of samples in pulse candidate " << i << " is " 
		  << trackerData->getChargeValues().size() << ", should be " 
		  << referenceCandidateProperties[i].nSamples << std::endl;
	return false;
      }
      //      std::cout << "test ok" << std::endl;

      if ( *(trackerData->getChargeValues().begin()) != referenceCandidateProperties[i].firstSample)
      {
	std::cout << "fist sample in pulse candidate " << i << " is " 
		  << *(trackerData->getChargeValues().begin()) << ", should be " 
		  << referenceCandidateProperties[i].firstSample << std::endl;
	return false;
      }

      if ( *(trackerData->getChargeValues().rbegin()) != referenceCandidateProperties[i].lastSample)
      {
	std::cout << "fist sample in pulse candidate " << i << " is " 
		  << *(trackerData->getChargeValues().rbegin()) << ", should be " 
		  << referenceCandidateProperties[i].lastSample << std::endl;
	return false;
      }

      if ( trackerData->getTime() != referenceCandidateProperties[i].time)
      {
	std::cout << "time in pulse candidate " << i << " is " 
		  << trackerData->getTime() << ", should be " 
		  << referenceCandidateProperties[i].time << std::endl;
	return false;
      }

      if ( candidates[i].second != referenceCandidateProperties[i].quality)
      {
	std::cout << "time in pulse candidate " << i << " is " 
		  << candidates[i].second << ", should be " 
		  << referenceCandidateProperties[i].quality << std::endl;
	return false;
      }
    }

    return true;
  }

  

  bool PulseFinderTester::evaluatePulses(std::vector<TrackerPulse *> const & pulses,
		  std::vector<PulseFinderTester::PulseProperties> const & referencePulseProperties)
  {
    if (pulses.size() != referencePulseProperties.size())
    {
      std::cout << "number of pulses is " << pulses.size()
		<< ", should be "<<  referencePulseProperties.size() << std::endl;
      return false;
    }

    for (size_t i = 0; i < pulses.size(); i++)
    {
      if ( fabs(pulses[i]->getTime() - referencePulseProperties[i].time) > 1e-3 )
      {
	std::cout << "time in pulse " << i << " is " 
		  << pulses[i]->getTime() << ", should be " 
		  << referencePulseProperties[i].time << std::endl;
	return false;
      }

      if ( fabs(pulses[i]->getCharge() - referencePulseProperties[i].charge) > 1e-3 )
      {
	std::cout << "charge in pulse " << i << " is " 
		  << pulses[i]->getCharge() << ", should be " 
		  << referencePulseProperties[i].charge << std::endl;
	return false;
      }

      if (pulses[i]->getQuality() != referencePulseProperties[i].quality)
      {
	std::cout << "quality in pulse " << i << " is " 
		  << pulses[i]->getQuality() << ", should be " 
		  << referencePulseProperties[i].quality << std::endl;
	return false;
      }

      TrackerData const * trackerData = pulses[i]->getTrackerData();
      if (trackerData == NULL)
      {
	if (referencePulseProperties[i].nSamples != 0)
	{
	  std::cout << "trackerData object in pulse " << i << " is NULL while it should not."
		    << std::endl;
	  return false;
	}
	// if we did not expect the trackerData to be set the NULL pointer is ok
      }
      else
      {
	if (trackerData->getChargeValues().size() != referencePulseProperties[i].nSamples)
	{
	  std::cout << "Number of samples in pulse " << i << " is " 
		    << trackerData->getChargeValues().size() << ", should be " 
		    << referencePulseProperties[i].nSamples << std::endl;
	  return false;
	}
      }

    }// for i
    return true;
  }

  bool PulseFinderTester::evaluatePlateauCutoff(std::vector<TrackerPulse *> const & pulses,
			  std::vector<std::pair<size_t,size_t> > const & referencePositions,
			  EVENT::FloatVec const & referenceADCValues)
  {
    if (pulses.size() != referencePositions.size())
    {
      std::cout << "Number of pulses is " << pulses.size() 
		<< ", should be " <<  referencePositions.size() << std::endl;
      return false;
    }

    for (size_t i = 0; i < pulses.size(); i++)
    {
      // the PLATEAU_CUTOFF_BIT  has to be set
      if ( !pulseflag::isPlateauCutoff(pulses[i]->getQuality()) )
      {
	std::cout << "PLATEAU_CUTOFF_BIT not set in pulse " << i << std::endl;
	return false;
      }

      // the trackerdata has to be set
      TrackerData const * trackerData = pulses[i]->getTrackerData();
      if (trackerData==NULL)
      {
	std::cout << "TrackerData in pulse " << i 
		  << " is not set." << std::endl;
	return false;
      }
      
      size_t startIndex = referencePositions[i].first;
      size_t endIndex = referencePositions[i].second;

      if ( (endIndex-startIndex)!= trackerData->getChargeValues().size() )
      {
	std::cout << "Pulse length is " <<  trackerData->getChargeValues().size()
		  << " in pulse " << i << ", should be " << endIndex-startIndex << std::endl;
	return false;
      }

      if ( endIndex > referenceADCValues.size() )
      {
	std::cout << "Error in the test itself: endIndex out of range in pulse " << i << std::endl;
	return false;
      }

      // compare the spectrum
      for (size_t j = 0; j < endIndex-startIndex ; j++)
      {
	if ( trackerData->getChargeValues()[j] != referenceADCValues[startIndex+j] )
	{
	  std::cout << "sample " << j<< " in pulse " <<i << " is "
		    << trackerData->getChargeValues()[j]
		    << ", should be " << referenceADCValues[startIndex+j] << std::endl;
	  return false;
	}
	//else 
	//  {
	//    std::cout << "DEBUG: sample " << trackerData->getChargeValues()[j] << std::endl;
	//  }
      }
      
    }// for pulse index i 

    return true;
  }
  
}

